/*
** Module   : StatefulHttpClient.java
** Abstract : Implementation of the builtin class.
**
** Copyright (c) 2019-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------------Description----------------------------------------
** 001 IAS  20190923 First version.
** 002 ME   20200410 Added getter/setter methods for cookie jar.
** 003 MP   20200603 add constructors and methods like originar class.
** 004 ME   20201103 Implement cookie support OE12.2.
**     CA   20210221 Fixed parameters at the LegacySignature annotation.
** 005 CA   20210221 Fixed 'qualified', 'extent' and 'returns' annotations at the legacy signature.
**     CA  20220120 Do not used TypeFactory.object when the OO reference must not be tracked or registered.
**                  Do not use TypeFactory.object for internal usages, use ObjectVar if the reference must 
**                  be tracked.
**                  All TypeFactory.object variable definitions must be done outside of the top-level block.
**     CA  20220923 Variable definitions (including associated with parameters) must be done always outside of 
**                  the BlockManager API
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.oo.net.http;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.BlockManager.Action;
import com.goldencode.p2j.util.BlockManager.Condition;
import com.goldencode.p2j.util.InternalEntry.Type;

import static com.goldencode.p2j.report.ReportConstants.*;
import static com.goldencode.p2j.util.BlockManager.externalProcedure;
import static com.goldencode.p2j.util.BlockManager.function;
import static com.goldencode.p2j.util.BlockManager.internalProcedure;
import static com.goldencode.p2j.util.BlockManager.onBlockLevel;
import static com.goldencode.p2j.util.BlockManager.returnNormal;

import com.goldencode.p2j.oo.core.Assert;
import com.goldencode.p2j.oo.net.Uri;

/**
 * Http client with support for cookies
 *
 */
@LegacyResource(resource = "OpenEdge.Net.HTTP.StatefulHttpClient")
@LegacyResourceSupport(supportLvl = CVT_LVL_FULL | RT_LVL_FULL)
public class StatefulHttpClient extends HttpClientDecorator implements ISupportCookies
{

   /** Cookie Jar **/
   @LegacySignature(type = Type.PROPERTY, name = "CookieJar")
   @LegacyResourceSupport(supportLvl = CVT_LVL_FULL | RT_LVL_FULL)
   private final object<? extends ICookieJar> cookieJar = TypeFactory.object(ICookieJar.class);

   @LegacySignature(type = Type.METHOD, name = "Execute", parameters = {
            @LegacyParameter(name = "poRequest", type = "OBJECT", qualified = "openedge.net.http.ihttprequest", mode = "INPUT"),
            @LegacyParameter(name = "poResponse", type = "OBJECT", qualified = "openedge.net.http.ihttpresponse", mode = "INPUT") })
   @LegacyResourceSupport(supportLvl = CVT_LVL_FULL | RT_LVL_FULL)
   @Override
   public void execute(
            final object<? extends com.goldencode.p2j.oo.net.http.IhttpRequest> _poRequest,
            final object<? extends com.goldencode.p2j.oo.net.http.IhttpResponse> _poResponse)
   {
      object<? extends com.goldencode.p2j.oo.net.http.IhttpRequest> poRequest = TypeFactory
               .initInput(_poRequest);
      object<? extends com.goldencode.p2j.oo.net.http.IhttpResponse> poResponse = TypeFactory
               .initInput(_poResponse);
      object<? extends Uri> oUri = TypeFactory.object(Uri.class);
      object<? extends Cookie> oCookies[][] = new object[][] { TypeFactory.objectExtent(Cookie.class, false) };

      internalProcedure(StatefulHttpClient.class, this, "Execute", new Block((Body) () -> {
         
         oUri.assign(poRequest.ref().getUri());
         // check if there are cookies in jar for request url
         cookieJar.ref().getCookies(oUri, new OutputExtentParameter<object<? extends Cookie>>()
         {

            @Override
            public object<? extends Cookie>[] getVariable()
            {
               return oCookies[0];
            }

            @Override
            public void setVariable(object<? extends Cookie>[] reference)
            {
               oCookies[0] = reference;
            }
         });

         // if there are any set those on request
         if (oCookies[0].length > 0)
            poRequest.ref().setCookies(oCookies[0]);
         
         oCookies[0] = ArrayAssigner.resize(oCookies[0], new integer());
         
         getDecoratedHttpClient().ref().execute(poRequest, poResponse);
         
         // if we have cookies set in response save those in jar
         poResponse.ref().getCookies(new OutputExtentParameter<object<? extends Cookie>>()
         {

            @Override
            public object<? extends Cookie>[] getVariable()
            {
               return oCookies[0];
            }

            @Override
            public void setVariable(object<? extends Cookie>[] reference)
            {
               oCookies[0] = reference;
            }
         });
         
         if (oCookies[0].length > 0) {
            
            String host = oUri.ref().getHost().getValue().toLowerCase();
            
            for (object<? extends Cookie> cookie : oCookies[0])
            {
               // RFC6265 - defaults and domain validation
               character cpath = TextOps.isEmpty(cookie.ref().getPath()) ? oUri.ref().getPath() : cookie.ref().getPath();
               character cdomain = TextOps.isEmpty(cookie.ref().getDomain()) ? oUri.ref().getHost() : cookie.ref().getDomain();
               
               String domain = cdomain.getValue().toLowerCase();
               if (domain.startsWith("."))
                  domain = domain.substring(1);
               
               if (domain.equals(host) || host.endsWith("." + domain))
                  cookieJar.ref().addCookie(cdomain, cpath, cookie);
            }
         }
         
      }));
   }

   public void __net_http_StatefulHttpClient_execute__()
   {
      externalProcedure(StatefulHttpClient.class, StatefulHttpClient.this,
               new Block((Body) () -> {
                  onBlockLevel(Condition.ERROR, Action.THROW);
               }));
   }

   @LegacySignature(type = Type.CONSTRUCTOR, parameters = {
            @LegacyParameter(name = "poClient", type = "OBJECT", qualified = "openedge.net.http.ihttpclient", mode = "INPUT"),
            @LegacyParameter(name = "poCookieJar", type = "OBJECT", qualified = "openedge.net.http.icookiejar", mode = "INPUT") })
   @LegacyResourceSupport(supportLvl = CVT_LVL_FULL | RT_LVL_FULL)
   public void __net_http_StatefulHttpClient_constructor__(
            final object<? extends com.goldencode.p2j.oo.net.http.IhttpClient> _poClient,
            final object<? extends com.goldencode.p2j.oo.net.http.ICookieJar> _poCookieJar)
   {
      object<? extends com.goldencode.p2j.oo.net.http.ICookieJar> poCookieJar = TypeFactory
               .initInput(_poCookieJar);

      internalProcedure(StatefulHttpClient.class, this,
               "__net_http_StatefulHttpClient_constructor__", new Block((Body) () -> {
                  __net_http_HttpClientDecorator_constructor__(_poClient);

                  Assert.notNull(poCookieJar);
                  cookieJar.assign(poCookieJar);
               }));
   }

   @LegacySignature(type = Type.CONSTRUCTOR, parameters = {
            @LegacyParameter(name = "poClient", type = "OBJECT", qualified = "openedge.net.http.ihttpclient", mode = "INPUT") })
   @LegacyResourceSupport(supportLvl = CVT_LVL_FULL | RT_LVL_FULL)
   public void __net_http_StatefulHttpClient_constructor__(
            final object<? extends com.goldencode.p2j.oo.net.http.IhttpClient> _poClient)
   {
      internalProcedure(StatefulHttpClient.class, this,
               "__net_http_StatefulHttpClient_constructor__", new Block((Body) () -> {
                  __net_http_HttpClientDecorator_constructor__(_poClient);
               }));
   }

   /**
    * Get cookie storage.
    * 
    * @return cookie jar.
    */
   @LegacySignature(returns = "OBJECT", qualified = "OpenEdge.Net.HTTP.ICookieJar", type = Type.GETTER, name = "CookieJar")
   @LegacyResourceSupport(supportLvl = CVT_LVL_FULL | RT_LVL_FULL)
   @Override
   public object<? extends com.goldencode.p2j.oo.net.http.ICookieJar> getCookieJar()
   {
      return function(this, "CookieJar", object.class, new Block((Body) () -> {
         returnNormal(cookieJar);
      }));
   }

   /**
    * Set cookie storage.
    * 
    * @param _cookieJar The cookie jar.
    */
   @LegacySignature(type = Type.SETTER, name = "CookieJar", parameters = 
   {
      @LegacyParameter(name = "cookieJar", type = "OBJECT", mode = "INPUT", qualified = "OpenEdge.Net.HTTP.ICookieJar")
   })
   @LegacyResourceSupport(supportLvl = CVT_LVL_FULL | RT_LVL_FULL)
   public void setCookieJar(
            object<? extends com.goldencode.p2j.oo.net.http.ICookieJar> _cookieJar)
   {
      object<? extends ICookieJar> cookieJar = TypeFactory.initInput(_cookieJar);

      internalProcedure(this, "CookieJar", new Block((Body) () -> {
         this.cookieJar.assign(cookieJar);
      }));

   }

}
