/*
** Module   : MenuContainerWidget.java
** Abstract : The base class for MenuWidget and SubMenuWidget.
**
** Copyright (c) 2015-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 VIG 20150115 Created initial version.
** 002 HC  20150323 Removed business logic from config classes.
** 003 VIG 20150304 Added firstChild method, id's collision fix and other minor changes
** 004 CA  20150723 Made finishConfigProcessing() protected.
** 005 VIG 20150714 Changed base class to BaseEntity.
** 006 CA  20170130 Fixed a NPE in widgets(), for MENU containers with no widgets.
**     HC  20170131 Added runtime support for dynamic MENU* widgets.
** 007 CA  20170929 Fixed NEXT-SIBLING and PREV-SIBLING attributes.
**                  Improved menu registration with client-side.
** 008 ECF 20180102 Use ConfigManager instance from WidgetConfig instead of doing context-local
**                  lookup.
** 009 CA  20180201 Menus are pushed to client-side only when they are attached to a widget.
**                  Fixed pushWidgetAttr() for menu cases.
** 008 HC  20180419 Fixed an abend during menu setup when sub-menu or menu item wasn't attached to
**                  a menu widget.
** 009 HC  20180510 Fixed an NPE when attribute change pushed for a dynamic menu element without
**                  a parent.
** 010 CA  20180525 Added valid() and setSensitive().
** 011 CA  20180710 Fixed menu delete.
** 012 SBI 20180705 Fixed findMenuItemInt/findSubMenuInt to use case insensitive search.
** 013 EVL 20180811 Fix from CA for decline pushing widget attribute for deleted widget.
** 014 HC  20191205 Fixed ShowScrollBar not showing/hiding scroll bars.
** 015 HC  20200313 Javadoc fixes.
** 016 HC  20200926 Performance optimizations of server-client config state synchronization.
**     AIL 20201007 Override canPushWidgetAttr method in order to define custom constraints.
**     EVL 20201019 Added specific version of getId() for client driven attribute.
**     EVL 20201021 More optimization for getAttr() usage or widget ID getting.
**     EVL 20201022 Optimized attributes flush implementation.
**     CA  20201117 Do not use resourceDelete to delete the widget - use the delete() API for this code.
**     CA  20220901 Refactored scope notification support: ScopeableFactory was removed, and the registration 
**                  is now specific to each type of scopeable.  For each case, the block will be registered 
**                  for scope support (for that particular scopeable) only when the scopeable is 'active' 
**                  (i.e. unnamed streams or accumulators are used).  This allows a lazy registration of 
**                  scopeables, to avoid the unnecessary overhead of processing all the scopeables for each
**                  and every block.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.util.*;

import com.goldencode.p2j.util.*;

/**
 * This class represents sub-menu widget.
 */
abstract class MenuContainerWidget<T extends BaseConfig>
extends BaseEntity<T> 
implements CommonHandleTree,
           WidgetContainer
{
   /** Maps widget names to widget instances (from the menu definition). */
   private Map<String, GenericWidget<?>> n2w = null;

   /** List of widgets in ID order. */
   private List<GenericWidget<?>> widgets = new ArrayList<>();
   
   /** menu definition */
   private MenuDescription menuDef = null;
   
   /** Shared menu flag. */
   private boolean shared = false;

   /** batch modifications flag */
   private boolean batch = true;
   
   /**
    * Default constructor.  
    * 
    * @param    dynamic
    *           Flag indicating if this is a static or dynamic resource.
    * @param    menuConfig
    *           The menu config instance.
    */
   public MenuContainerWidget(boolean dynamic, T menuConfig)
   {
      super(dynamic, menuConfig);
      batch = !dynamic;
   }
   
   /**
    * Find the root menu for the specified menu element.
    *
    * @param    <V>
    *           Menu element config type.
    *
    * @param    menu
    *           The menu element.  Either a {@link MenuContainerWidget} or a 
    *           {@link MenuItemWidget}.
    * 
    * @return   The found root.
    */
   static <V extends BaseConfig> MenuContainerWidget<V> findRootMenu(BaseEntity<V> menu)
   {
      MenuContainerWidget<V> root = menu instanceof MenuContainerWidget 
                                       ? (MenuContainerWidget<V>) menu 
                                       : menu.getParent(MenuContainerWidget.class);
      MenuContainerWidget<V> parent;
      while(root != null && (parent = root.getParent(MenuContainerWidget.class)) != null)
      {
         root = parent;
      }
      
      return root;
   }
   
   /**
    * Check if the specified menu is attached to a widget.  This means its root must have the
    * {@link MenuConfig#ownerId} set to a value other than 0.
    *
    * @param    <V>
    *           Menu element config type.
    *
    * @param    menu
    *           The menu to check.  Either a {@link MenuContainerWidget} or a 
    *           {@link MenuItemWidget}.
    *           
    * @return   <code>true</code> if the menu is attached.
    */
   static <V extends BaseConfig> boolean isAttachedMenu(BaseEntity<V> menu)
   {
      MenuContainerWidget<V> root = (MenuContainerWidget<V>) findRootMenu(menu);

      return root instanceof MenuWidget && ((MenuConfig) root.config).ownerId != 0;
   }

   /**
    * Sets new value of SENSITIVE attribute. Overrides {@link GenericWidget} logic 
    * (that enables widget in its frame).
    *
    * @param    s
    *           The new value of the SENSITIVE attribute.
    */
   @Override
   public void setSensitive(boolean s)
   {
      // changing the SENSITIVE attribute does not affect the widget's visibility
      setAttr("enabled", config.enabled, s, (vv) -> config.enabled = vv);
   }
   
   /**
    * Get the <code>next-sibling</code> attribute for the specified child
    * 
    * @param    child
    *           The child for which the next-sibling is needed, in this {@link #widgets}.
    *           
    * @return   See above.
    */
   @Override
   public handle getNextSibling(HandleChain child)
   {
      if (child == null)
      {
         return new handle();
      }
      
      Iterator<GenericWidget<?>> iter = widgets.iterator();
      while (iter.hasNext())
      {
         GenericWidget<?> next = iter.next();
         
         if (next == child)
         {
            return (iter.hasNext() ? new handle(iter.next()) : new handle());
         }
      }
      
      return new handle();
   }
   
   /**
    * Get the <code>prev-sibling</code> attribute for the specified child
    * 
    * @param    child
    *           The child for which the prev-sibling is needed, in this {@link #widgets}.
    *           
    * @return   See above.
    */
   @Override
   public handle getPrevSibling(HandleChain child)
   {
      if (child == null)
      {
         return new handle();
      }

      Iterator<GenericWidget<?>> iter = widgets.iterator();
      GenericWidget<?> prev = null;
      while (iter.hasNext())
      {
         GenericWidget<?> next = iter.next();
         
         if (next == child)
         {
            return (prev == null ? new handle() : new handle(prev));
         }
      }
      
      return new handle();
   }

   /**
    * Get the <code>next-sibling</code> attribute for this widget.
    * 
    * @return   See above.
    */
   @Override
   public handle getNextSibling()
   {
      CommonWidget parent = getParent();
      
      return parent instanceof MenuContainerWidget 
                ? ((MenuContainerWidget<?>) parent).getNextSibling(this)
                : new handle();
   }

   /**
    * Get the <code>prev-sibling</code> attribute for this widget.
    * 
    * @return   See above.
    */
   @Override
   public handle getPrevSibling()
   {
      CommonWidget parent = getParent();
      
      return parent instanceof MenuContainerWidget 
                ? ((MenuContainerWidget<?>) parent).getPrevSibling(this)
                : new handle();
   }

   /**
    * Check if menu is shared.
    * 
    * @return  <code>true</code> if menu is shared.
    */
   public boolean isShared()
   {
      return shared;
   }
   
   /**
    * Set shared flag for menu.
    * 
    * @param   shared
    *          <code>true</code> if menu is shared.
    */
   void setShared(boolean shared)
   {
      this.shared = shared;
   }
   
   /**
    * Sets batching mode for frame modification.
    *
    * @param  on
    *         <code>true</code> sets batching mode
    */
   public void batch(boolean on)
   {
      batch(on, false);
   }

   /**
    * Queries batching mode for frame modification.
    *
    * @return 
    *         <code>true</code> if batching mode is in effect
    */
   public boolean isBatch()
   {
      return batch;
   }

   /**
    * Sets batching mode for frame modification and optionally pushes the
    * screen definition to the client. 
    *
    * @param  on
    *         <code>true</code> sets batching mode
    * @param  force
    *         <code>true</code> to ensure that the screen definition will be
    *         synchronously pushed to the client. <code>false</code> to
    *         defer the push. This is only honored when <code>on</code> is
    *         <code>false</code>.
    */
   public void batch(boolean on, boolean force)
   {
      batch = on;
      
      if (!batch)
         pushMenuDescription(force);
   }
   
   /**
    * Get all widgets stored in the container.
    *
    * @return  array of stored widgets.
    */
   public Collection<GenericWidget<?>> widgets()
   {
      return widgets == null ? Collections.emptyList() : Collections.unmodifiableCollection(widgets);
   }

   /**
    * Add the specified widget to this container.
    *
    * @param widget The widget to be added.
    */
   @Override
   public void addWidget(GenericWidget<?> widget)
   {
      if (widgets.contains(widget))
      {
         return;
      }

      widgets.add(widget);
   }

   /**
    * Remove the specified widget from this container.
    *
    * @param widget The widget to be removed.
    */
   @Override
   public void removeWidget(GenericWidget<?> widget)
   {
      widgets.remove(widget);
   }

   /**
    * Gets the numeric ID of this widget (this is used as an index into the
    * screen-buffer among other things).
    * 
    * @return   The numeric ID of this widget.
    */
   @Override
   public int getId()
   {
      // client driven ID getter case
      return getAttr("id", () -> (int)(config.id != null ? config.id.asInt() : -1), true);
   }

   /**
    * Core initialization processing. 
    *
    * @param    defClass
    *           The class of the definition implementation class.
    * @param    defInstance
    *           The instance of the definition implementation class.
    * @param    dynamic
    *           <code>true</code> If this menu is dynamic, <code>false</code> otherwise.
    */
   protected void coreInitialize(Class<?>         defClass,
                               Object             defInstance, 
                               boolean            dynamic)
   {
      WidgetListBase<?> wl = (WidgetListBase<?>) defInstance;
 
      // cache the widget map for this menu
      n2w = wl.getWidgets();
      
      // init the widget array
      widgets = new ArrayList<GenericWidget<?>>(n2w.size());
      
      // dynamic menu initialization should stop here
      if (dynamic)
      {
         return;
      }

      // add all the widgets in the order of their definition
      Iterator<String> iter = n2w.keySet().iterator();
      
      while (iter.hasNext())
      {
         
         // get the widget
         GenericWidget<?> widget = n2w.get(iter.next()); 
         widgets.add(widget);
         
         if (widget.getId() == -1) // widget instanceof SubMenuWidget
         {
            int wid = WidgetId.nextID().asInt();

            // initialize key state in the widget
            widget.setId(wid);

            // save new widget in session registry map
            LogicalTerminal.registerWidget(widget, wid);
         }
      }
   }
   
   /**
    * Conditionally pushes the screen definition instance down to the client. 
    * If modifications are currently being batched, no action is taken.
    * Otherwise, a new screen definition is built and sent to the client.
    * <p>
    * When the screen definition is built, a method of 
    * <code>LogicalTerminal</code> is called to send it down to the client.
    * 
    * Overrides {@link GenericWidget#pushScreenDefinition()} behavior.
    */
   protected void pushScreenDefinition()
   {
      pushMenuDescription(true);
   }

   /**
    * Conditionally pushes the screen definition instance down to the client.
    * If modifications are currently being batched, no action is taken.
    * Otherwise, a new screen definition is built and sent to the client.
    * <p>
    * When the screen definition is built, a method of
    * <code>LogicalTerminal</code> is called to send it down to the client.
    *
    * @param  force
    *         <code>true</code> to ensure that the screen definition will be
    *         synchronously pushed to the client. <code>false</code> to
    *         defer the push.
    */
   void pushMenuDescription(boolean force)
   {
      // batching is in progress, do not push individual changes
      if (batch || deleted)
         return;

      MenuContainerWidget<T> root = (MenuContainerWidget<T>) findRootMenu(this);

      if (root != this || ((MenuConfig) root.config).ownerId == 0)
      {
         return;
      }

      // this always pushes the whole menu-tree definition, if this shows to be a bottle-neck
      // push only the the menu items which really changed
      root.pushMenuDescriptionImpl(force);

      // push down the screen definition
      if (force)
      {
         TransactionManager.registerBlockScopeable(LogicalTerminal.getInstance());

         // synchronous push
         MenuDescription[] sd = new MenuDescription[] {root.menuDef};
         LogicalTerminal.pushMenuDescription(sd);
      }
      else
      {
         // deferred push
         LogicalTerminal.deferPush(root.menuDef);
      }
   }

   /**
    * Finish the config processing for this frame.
    * This includes registering the active configuration for each new widget.
    */
   @Override
   protected void finishConfigProcessing()
   {
      ConfigManager mgr = null;
      
      for (GenericWidget<?> w : widgets)
      {
         WidgetConfig cfg = w.config;
         if (mgr == null)
         {
            mgr = cfg.getConfigManager();
         }
         
         WidgetConfig active = mgr.getActiveConfig(cfg.id);
         if (active == null)
         {
            mgr.addWidgetConfig(cfg);
         }
         else
         {
            assert(cfg.getClass().equals(active.getClass()));
         }
         
         if (w instanceof MenuContainerWidget)
         {
            ((MenuContainerWidget<?>) w).finishConfigProcessing();
         }
      }
      
      // set the configuration this menu container
      if (mgr == null)
      {
         mgr = ConfigManager.getInstance();
      }
      WidgetConfig active = mgr.getActiveConfig(getAttr("id", () -> config.id, true));
      if (active == null)
      {
         mgr.addWidgetConfig(config);
      }
      else
      {
         assert(config.getClass().equals(active.getClass()));
      }
   }

   /**
    * Recursively finds menu-item in this menu tree. It can appear on different levels.
    * Menu-item is searched using some sort of depth-first algorithm.
    * 
    * <pre>
    * 1. Lets assume that cur is the current node in a MENU tree. Initial value is a DEFINE_MENU node.
    * 2. First we are searching the given MENU-ITEM in all immediate child MENU-ITEMs.
    * 3. If we've found the match than return success.
    * 4. If we've not found the match we are going into the all SUB-MENUs and for each of them we are repeating steps from the step 2.
    * 5. If we've searched all tree and didn't find a match - just return null.    
    * </pre>
    * 
    * @param   name
    *          MENU-ITEM name.
    *          
    * @return  concrete {@link MenuItemWidget} or null if not found.
    */
   protected MenuItemWidget findMenuItemInt(String name)
   {
      Collection<GenericWidget<?>> widgets = widgets();
      
      // check MENU-ITEMS of the first level
      for(GenericWidget<?> gw : widgets)
      {
         if (gw instanceof MenuItemWidget && name.equalsIgnoreCase(gw.name().getValue()))
         {
            return (MenuItemWidget) gw;
         }
      }
      
      // recursively check SUB-MENUs
      for(GenericWidget<?> gw : widgets)
      {
         if (gw instanceof SubMenuWidget)
         {
            MenuItemWidget mi = ((SubMenuWidget)gw).findMenuItemInt(name);
            
            if (mi != null)
            {
               return mi;
            }
         }
      }
      
      return null;
   }
   
   /**
    * Finds sub-menu in this menu tree. It can appear only once. 
    * Simple depth-first search is used.
    * 
    * @param   name
    *          Sub-menu name.
    * 
    * @return  concrete {@link SubMenuWidget} or null if not found.
    */
   protected SubMenuWidget findSubMenuInt(String name)
   {
      for (GenericWidget<?> gw : widgets())
      {
         if (!(gw instanceof SubMenuWidget))
            continue;
         
         if (name.equalsIgnoreCase(gw.name().getValue()))
            return (SubMenuWidget) gw;
         
         SubMenuWidget sm = ((SubMenuWidget)gw).findSubMenuInt(name);
         
         if (sm != null)
         {
            return sm;
         }
      }
      
      return null;
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      if (deleted || !canDelete())
      {
         return false;
      }

      // static menu widgets can't be deleted explicitly - the canDelete() will take care of 
      // prohibiting this, and allow it only when the creator procedure gets deleted
      
      if (isShared())
      {
         // shared menu can't be deleted until all are out of scope.
         return false;
      }

      return true;
   }
   
   /**
    * Perform actual delete of an resource. At the time of this call, it is assumed the resource
    * is valid for deletion (the handle and the resource are both valid).
    * <p>
    * The method first calls {@link #resourceDelete()}, if the method returns {@code false}, the deletion
    * is aborted.
    */
   @Override
   public void delete()
   {
      if (!resourceDelete())
      {
         return;
      }
      
      super.delete();

      // let the root delete the entire sub-tree - this is needed to avoid client-side trips
      // for each element in this sub-tree.
      MenuWidget root = (MenuWidget) findRootMenu(this);
      root.detachMenu(this);
      
      deleted = true;
      
      LogicalTerminal.deregisterMenu(getId());
      ConfigManager.getInstance().removeWidgetConfig(getAttr("id", () -> config.id, true));
   }   

   /**
    * Execute a DFS walk of the menu tree and collect all menu components rooted at this sub-tree.
    *
    * @param    result
    *           The list where to collect the menu components.
    */
   void listMenuTree(List<GenericWidget<?>> result)
   {
      result.add(this);
      
      for (GenericWidget<?> child : widgets())
      {
         if (child instanceof MenuContainerWidget)
         {
            ((MenuContainerWidget<?>) child).listMenuTree(result);
         }
         else
         {
            result.add(child);
         }
      }
   }

   /**
    * Accesses the first widget in the list of contained widgets.
    *
    * @return   The first widget or <code>null</code> if no widgets are
    *           contained.
    */
   public handle getFirstChild()
   {
      if (widgets.isEmpty())
      {
         return new handle();
      }
      else
      {
         return new handle(widgets.get(0));
      }
   }

   /**
    * Accesses the last widget in the list of contained widgets.
    *
    * @return   The last widget or <code>null</code> if no widgets are
    *           contained.
    */
   public handle getLastChild()
   {
      if (widgets.isEmpty())
      {
         return new handle();
      }
      else
      {
         return new handle(widgets.get(widgets.size() - 1));
      }
   }
   
   /**
    * Reports if this widget object is valid for use.
    * <p>
    * The validity of the widget object is determined solely by the scope of
    * the containing frame.
    * 
    * @return   <code>true</code> if we are valid (can be used).
    */
   @Override
   public boolean valid()
   {
      return !deleted;
   }
   
   /**
    * Returns true if the configuration can be pushed to the client-side.
    * 
    * @return   True if the configuration can be pushed to the client-side.
    */
   @Override
   protected boolean canPushWidgetAttr()
   {
      return !(deleted || !isAttachedMenu(this));
   }

   /**
    * Conditionally pushes the screen definition instance down to the client.
    * If modifications are currently being batched, no action is taken.
    * Otherwise, a new screen definition is built and sent to the client.
    * <p>
    * When the screen definition is built, a method of
    * <code>LogicalTerminal</code> is called to send it down to the client.
    *
    * @param  force
    *         <code>true</code> to ensure that the screen definition will be
    *         synchronously pushed to the client. <code>false</code> to
    *         defer the push.
    */
   private void pushMenuDescriptionImpl(boolean force)
   {
      int menuId = getId();
      menuDef = new MenuDescription(menuId);
      if (this.getParent() == null)
      {
         menuDef.addConfig(config, menuId);
      }

      // add all the widgets in the order of their definition

      for (GenericWidget<?> gw : widgets)
      {
         menuDef.addConfig(gw.config());
         if (gw instanceof SubMenuWidget)
         {
            MenuContainerWidget<?> sm = (MenuContainerWidget<?>) gw;
            // we are pushing, disable batch for all sub-menu's
            sm.batch = false;
            sm.pushMenuDescriptionImpl(force);
            menuDef.addChildDescription(sm.menuDef, sm.getId());
         }
      }

      if (isShared() && isVisible().booleanValue())
      {
         menuDef.setSuppressRedraw(true);
      }
   }
}
