/*
** Module   : AgentPool.java
** Abstract : Definition and implementation of an appserver agent.
**
** Copyright (c) 2013-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130529 Created initial version.
** 002 CA  20130628 Fixed context-local data reset for State-reset operating mode.
** 003 CA  20130701 Although official docs mention that State-reset doesn't call the shutdown or
**                  startup procedure, testing showed startup/shutdown procedures are called for  
**                  this operating mode too.
**                  Reworked appserver launching - implemented by the AppServerLauncher class.
** 004 CA  20130823 Allow each request to be identified via an unique ID. Changes related to 
**                  support async requests.
** 005 OM  20131018 Updated to conform the new SessionListener API.
** 006 SVL 20131203 Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 007 CA  20140218 Always use the full program name when interrogating SourceNameMapper.
** 008 CA  20140220 Argument postProcessing needs to be done before the procedure scope is popped.
**                  The propath needs to be re-initialized after context reset.
** 009 CA  20140223 Startup and Connect (State-aware and State-reset) procedures need to be ran 
**                  persistent.
** 010 CA  20140326 The remote persistent procedures need to be deleted by the Agent which created
**                  them.
** 011 CA  20140509 Fixed passing the parameter to the startup procedure.
** 012 CA  20140407 prepare() needs to force loading of the context-local DatabaseTriggerManager.
** 013 SVL 20140414 Implemented processing for TableParameters, added "appserver-agent" scope.
** 014 SVL 20140709 Added fall-through codes for table parameters errors.
** 015 CA  20150827 Release the memory for all memptr arguments, before returning to the caller.
** 016 CA  20151222 Any ERROR conditions raised during instantiation in obtainProxy are sent back
**                  to the caller (also, the pushed scope must be removed).
** 017 CA  20160404 Enhanced the appserver invocations, to allow implicit or explicit termination 
**                  of requests being executing on a certain connection.
** 018 CA  20160422 When resetting the Agent's context, make a bogus call to the P2J client-side,
**                  to clean up the frames (via the StateSynchronizer).
**                  Fixed deadlock when the Agent's java process is killed while it has control.
** 019 CA  20160626 Let the Agent recover from an UnstoppableExitException.
** 020 GES 20171220 Added TODO comment.
** 021 GES 20180102 Clarified an error handling comment.
** 022 CA  20180503 Added SESSION:SERVER-CONNECTION-ID support.
** 023 CA  20190327 If the command received by waitForCommand is on the same thread as the Agent's
**                  listening thread, execute immediately.
** 024 CA  20190612 Added APIs to create remote legacy objects and invoke methods on them.
** 025 CA  20190628 Proper SINGLETON and SINGLE-RUN implementation; misc improvements.
**     CA  20190703 Added WebHandler service support. 
** 026 CA  20190710 Hooked runtime for legacy WebHandler support.
**     CA  20190714 Added DATASET support.
**     CA  20190811 Fixed context reset.
** 027 CA  20191119 DATASET/TEMP-TABLE parameters register themselves as belonging to a bogus 
**                  class.
** 028 CA  20200427 Refactoring to allow support for CURRENT-REQUEST-INFO and CURRENT-RESPONSE-INFO
**                  transfer from the SERVER:REQUEST-INFO and to SERVER:RESPONSE-INFO.
**     CA  20200514 Fixed error handling for handleWebService.
**     CA  20200528 Fixes for extent arguments.
** 029 IAS 20200722 Improved logging.
**     CA  20200827 Reworked asynchronous invocations to perform all context-local work on the Conversation 
**                  thread, and let only the actual invocation be performed in an AssociatedThread.
**     RFB 20200926 Removed use of localMode since the 4GL doesn't care about the signature at an IN SUPER
**                  procedure decalaration. ControlFlowOps should validate the arguments.
**     CA  20201027 The remote side can delete the IN HANDLE used by the requester at the RUN statement.
**     SVL 20210127 Assign local RETURN-VALUE after an appserver call only if RETURN-VALUE of the remote
**                  procedure was set during this call.
**     SVL 20210202 Assign return value to an invocation result only if the value was set during this call.
**     CA  20210405 Again fixed the Agent context re-initialization in STATE-RESET mode.
**     IAS 20210520 Added support for session-based database auto-connect
**     CA  20211112 Fixed error management in PASOE appserver.
**     CA  20211214 For State-free appservers, the agents can be bound to remote persistent procedures; this
**                  binding must be made using the agent's ID, as a client can invoke multiple remote 
**                  persistent procedures, and using the procedure's ID for this binding can lead to 
**                  collisions, as the pair (connection ID, procedure ID) is not guaranteed to be unique for
**                  a connection.
**     CA  20211216 Fixed the agent's task execution timeout (is specified in seconds).
**     CA  20220104 Fixed Stateless appserver mode, which got broken in CA/20211214 - now the agent tracks any
**                  persistent procedure ran over it, 
**     CA  20220111 Fixed SESSION:SERVER-CONNECTION-CONTEXT - this value must be shared to any agent running
**                  requests for a connection.
**     CA  20220211 An OUTPUT memptr parameter must not be set to unknown if the MemoryBuffer has no bytes.
**     CA  20220310 Ensure the 'appserver-agent' root block is popped when the agent is reset and when it
**                  terminates - this is required because this block will allow dynamic temp-tables to be
**                  cleaned (i.e. the physical temp-table actually dropped).
**     CA  20220329 Added support for REST services written directly in Java.
**     CA  20220405 Added authentication and authorization for web requests.  When this is enabled, the target
**                  API call will be executed under the authenticated FWD context, and not the agent's context.
**     CA  20220901 Refactored scope notification support: ScopeableFactory was removed, and the registration 
**                  is now specific to each type of scopeable.  For each case, the block will be registered 
**                  for scope support (for that particular scopeable) only when the scopeable is 'active' 
**                  (i.e. unnamed streams or accumulators are used).  This allows a lazy registration of 
**                  scopeables, to avoid the unnecessary overhead of processing all the scopeables for each
**                  and every block.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;
import java.util.*;
import java.util.Queue;
import java.util.function.*;

import javax.servlet.http.*;

import org.apache.commons.logging.*;

import com.goldencode.p2j.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.oo.lang.*;
import com.goldencode.p2j.oo.web.WebHandler;
import com.goldencode.p2j.persist.*;
import com.goldencode.p2j.persist.trigger.*;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.security.SecurityManager;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.util.BlockManager.*;

/**
 * An apserver agent definition.
 */
class Agent
{
   /**
    * List of possible invocation modes: 
    * <ul>
    * <li>{@link #IMMEDIATELY} - in this thread (the thread must be an agent thread)</li>
    * <li>{@link #AGENT_WAIT} - in a different agent, wait for response</li>
    * <li>{@link #AGENT_NO_WAIT} in a different agent, do not wait for response</li>
    * </ul>
    */
   enum InvocationModes
   {
      IMMEDIATELY,
      AGENT_WAIT,
      AGENT_NO_WAIT
   }
   
   /** Logger */
   private static final Log LOG = LogFactory.getLog(AppServerManager.class);

   /**
    * The set of web service tokens currently being active (with an API call in progress).  As the FWD context
    * is not thread-safe, only one API call can be performed at a certain time.
    * TODO: remove this once the FWD context allows multiple concurrent requests. 
    */
   private static final Set<String> activeWebServiceTokens = new HashSet<>();

   /** The command queue for this agent. */
   private final Queue<AgentCommand> commands = new LinkedList<AgentCommand>();
   
   /** The agent's pool. */
   private final AgentPool pool;

   /** The agent's stopper worker. */
   private final AgentStopper stopperWorker = new AgentStopper();
   
   /** 
    * Dedicated thread to stop any job currently being executed by this Agent.  
    * The thread is running under the Agent's context.
    */
   private final AssociatedThread stopper = new AssociatedThread(stopperWorker);
   
   /** Flag indicating if the agent has been terminated. */
   private boolean terminated = false;
   
   /** Flag indicating the agent has been bound to the connection. */
   private boolean bound = false;
   
   /** The agent's listening thread. */
   private Thread thread;
   
   /** The agent's ID, unique. */
   private final int agentId;
   
   /** The remote procedures ran on this agent. */
   private Map<String, AgentRemoteProcedure> remoteProcedures = new HashMap<>();
   
   /**
    * Build a new agent and associate it with the given pool.
    * 
    * @param    pool
    *           The agent's pool.
    * @param    agentId
    *           The agent ID.
    */
   public Agent(AgentPool pool, int agentId)
   {
      this.pool = pool;
      this.agentId = agentId;
   }
   
   /**
    * Check if the specified web service token is currently in use (there is an active API call on an agent).
    * 
    * @param    token
    *           The web service token.
    *           
    * @return   <code>true</code> if {@link #activeWebServiceTokens} contains this token.
    */
   static boolean isWebServiceTokenInUse(String token)
   {
      synchronized (activeWebServiceTokens)
      {
         return activeWebServiceTokens.contains(token);
      }
   }

   /**
    * Invoke an external procedure with no parameters. 
    * 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    */
   private static void invokeScoped(character name)
   {
      executeScoped(() ->
      {
         // force invocation of an external procedure
         ControlFlowOps.invokeExternalProcedure(name, false, null, false, null, null);
         return ControlFlowOps.getReturnValue();
      });
   }

   /**
    * Execute the specified task scoped to a topmost, global block.
    * 
    * @param    task
    *           The task to execute.
    *           
    * @return   The value as returned by the task.
    */
   private static BaseDataType executeScoped(Supplier<BaseDataType> task)
   {
      // add the topmost scope to the TransactionManager
      TransactionManager.pushScope("startup",
                                   TransactionManager.NO_TRANSACTION,
                                   true,
                                   true,
                                   false,
                                   false);
      try
      {
         return task.get();
      }
      catch (StopConditionException e)
      {
         // check if this can be converted to error
         if (invokeFailure(e))
         {
            return new unknown();
         }
         
         throw e;
      }
      finally
      {
         // remove the topmost scope from the TransactionManager
         TransactionManager.popScope();
      }
   }

   /**
    * Check if the specified ERROR condition needs to be reported.
    * 
    * @param    invocationResult
    *           The result where to save the error.
    * @param    e
    *           The ERROR condition.
    */
   private void checkError(AppServerInvocationResult invocationResult, ErrorConditionException e)
   {
      if (ControlFlowOps.isReturnValueSetInRootScope())
      {
         if (isPasoe())
         {
            invocationResult.setError(e);
         }
         else
         {
            String msg = ControlFlowOps.getReturnValue().toStringMessage();
            invocationResult.setError(new ErrorConditionException(msg));
         }
         return;
      }
      
      Throwable cause = e.getCause();
      if (cause instanceof NumberedException)
      {
         // only the following codes will be sent to the remote side
         // TODO: this needs to be done only if the errors are caused by problems in
         // resolving the appserver call target, and not by nested RUN statements
         final Set<Integer> notifyErrs = new HashSet<Integer>();
         final Integer[] codes = 
            { 6349, 3230, 1005, 3234, 5729, 2570, 11428, 12290, 15472, 5639, 6445, 8030,
              12324 };
         notifyErrs.addAll(Arrays.asList(codes));

         NumberedException ne = (NumberedException) cause;
         int num = ne.getNumber();
         if (notifyErrs.contains(num))
         {
            invocationResult.setError(e);
         }
      }
      
      if (invocationResult.getError() == null)
      {
         cause = e;
         while (cause != null)
         {
            if (cause instanceof ConditionException)
            {
               invocationResult.setError(cause);
               break;
            }
            cause = cause.getCause();
         }
      }
      
      if (invocationResult.getError() == null)
      {
         invocationResult.setError(e);
      }
   }
   
   /**
    * Check if the given STOP condition is from an invoke failure.
    * 
    * @param    e
    *           The STOP condition.
    * 
    * @return   <code>true</code> if the cause if this STOP is a {@link NumberedException} known
    *           to be produced by invoke failures.
    */
   private static boolean invokeFailure(StopConditionException e)
   {
      Throwable cause = e.getCause();
      if (cause != null && cause instanceof NumberedException)
      {
         // this specific case treats the error raised by ControlFlowOps.invokeFailure, which will
         // always set the topmost cause as a NumberedException
         NumberedException ne = (NumberedException) cause;
         if (ne.getNumber() == 293)
         {
            // throws a ControlFlowOps.invokeFailure incoming error to allow the current command
            // to abort (which is a RUN of an external program) and will just 'eat' the exception
            // and copy it to the RemoteInvocationResult instance; it should be safe to directly 
            // instantiate the error because we are outside of 4GL normal behavior; anyway, the
            // idea is also 'this needs to be thrown'
            
            // raise ERROR condition if the external procedure was not found
            throw new ErrorConditionException(e);
         }
      }
      
      return false;
   }
   
   /**
    * Get the agentId for this session.
    * 
    * @return   The {@link #agentId}.
    */
   public int getAgentId()
   {
      return agentId;
   }
   
   /**
    * Remove the external procedure from this agent, as it was deleted.
    * 
    * @param    procId
    *           The remote procedure ID.
    */
   public synchronized void removeProcedure(String procId)
   {
      AgentRemoteProcedure remote = remoteProcedures.remove(procId);
      pool.removeProcedure(remote.requesterConnectionId, remote.extProg.get());
   }
   
   /**
    * Register the given external procedure with this agent, as a remote procedure.
    * 
    * @param    extProg
    *           The external procedure.
    */
   public synchronized void addProcedure(handle extProg)
   {
      AgentRemoteProcedure remote = new AgentRemoteProcedure(extProg);
            
      this.remoteProcedures.put(remote.extProgResourceId, remote);
      pool.addProcedure(remote.extProg.get(), this);
   }
   
   /**
    * Get the external procedure associated with this agent.
    * 
    * @return   The procedure associated for this ID or <code>null</code> if is not registered with 
    *           {@link #remoteProcedures}.
    */
   public synchronized ExternalProgramWrapper getProcedure(String resId)
   {
      AgentRemoteProcedure remote = remoteProcedures.get(resId);
      return remote == null ? null : remote.extProg;
   }
   
   /**
    * Check if the given resource ID is for a persistent procedure an on this agent.
    * 
    * @param    resId
    *           The procedure ID.
    *           
    * @return   See above.
    */
   public synchronized boolean hasProcedure(String resId)
   {
      return remoteProcedures.containsKey(resId);
   }
   
   /**
    * Get the procedure name for the given resource ID.
    * 
    * @param    resId
    *           The procedure ID.
    *           
    * @return   See above.
    */
   public synchronized String getProcedureName(String resId)
   {
      AgentRemoteProcedure remote = remoteProcedures.get(resId);
      return remote == null ? null : remote.procedureName;
   }
   
   /**
    * Start the agent and listen for commands.
    */
   public void listen()
   {
      this.thread = Thread.currentThread();
      
      // register a listener to terminate the agent, if its session terminates
      Session sess = SessionManager.get().getSession();
      sess.addSessionListener(new SessionListener()
      {
         /**
          * This callback will be called when the P2J client session is ending.
          * 
          * @param    session
          *           The session that is ending.
          */
         @Override
         public void terminate(Session session)
         {
            // terminate the stopper worker explicitly - otherwise it will deadlock when the queue
            // tears down
            stopperWorker.terminate();
            
            Agent.this.terminate();
         }

         /**
          * This callback will be called after the security context has been set up.
          *
          * @param    session
          *           The session that is starting.
          */
         @Override
         public void initialize(Session session)
         {
            // no-ops
         }
      });
      
      try
      {
         prepare();
   
         String startup = pool.getStartup();
         // although official docs mention that state-reset doesn't call the startup procedure, 
         // testing showed that startup procedure is called
         handle hStartup = new handle();
         if (startup != null && startup.length() > 0)
         {
            String startupParameter = pool.getStartupParameter();
            if (!invoke(null, null, InvocationModes.IMMEDIATELY, startup, hStartup, "Startup",
                        "I", new character(startupParameter)))
            {
               LOG.warn("There were problems invoking the startup procedure, terminate the agent: " 
                        + pool.getName());
               // if there were problems invoking the startup procedure, terminate the agent
               return;
            }
         }
      
         // notify the launch
         LOG.debug(String.format("About to notify launcher [%s]", pool.getName()), new Exception());
         AppServerLauncher.notifyLaunch(pool.getName(), true);
         
         // start the agent's stopper thread
         stopper.setName("Agent Stopper for " + Thread.currentThread().getName());
         stopper.setDaemon(true);
         stopper.start();
         
         while (!terminated())
         {
            AgentCommand command = null;
            
            // this loop will continue until someone sends the terminate command:
            // - in case the session is ended (P2J client gets killed)
            // - in case the manager is ended
            
            boolean interrupted = false;
            synchronized (this)
            {
               if (commands.size() == 0)
               {
                  try
                  {
                     wait();
                  }
                  catch (InterruptedException e)
                  {
                     interrupted = true;
                     if (LOG.isWarnEnabled())
                     {
                        LOG.warn("Agent was interrupted for appserver " + pool.getName(), e);
                     }
                  }
               }
      
               // get the last command (or null if none)
               command = commands.poll();
            }
   
            // someone has interrupted my waiting, check if the session is still alive
            if (interrupted && !sess.isRunning())
            {
               // mark the agent as terminated
               synchronized (this)
               {
                  terminated = true;
               }
   
               stopper.interrupt();
               return;
            }
            
            if (command == null)
            {
               // no command yet, continue
               continue;
            }
            
            // execute only if an explicit command is not already stopped
            int reqId = command.getRequestId();
            String connId = command.getConnectionId();
            
            if (reqId == 0 || AppServerManager.registerRequest(reqId, this))
            {
               AppServerManager.addActiveAgent(connId, this);
   
               try
               {
                  command.doExecute();
               }
               catch (Throwable t)
               {
                  if (LOG.isWarnEnabled())
                  {
                     LOG.warn("Agent encountered an error while executing a command for appserver " + 
                              pool.getName(), t);
                  }
                  command.setError(t);
               }
               finally
               {
                  AppServerManager.removeActiveAgent(connId, this);
   
                  if (reqId != 0)
                  {
                     AppServerManager.deregisterRequest(reqId);
                  }
                  command.finished();
               }
            }
         }
         
         String shutdown = pool.getShutdown();
         // although official docs mention that state-reset doesn't call the shutdown procedure, 
         // testing showed that shutdown procedure is called
         if (shutdown != null && shutdown.length() > 0)
         {
            // ignore errors here
            invoke(null, null, InvocationModes.IMMEDIATELY, shutdown, null, "Shutdown", null);
         }

         // the startup procedure handle is deleted automatically, as the agent context is gone
         // we can't force an explicit delete, because we are outside of a P2J external block.
      }
      finally
      {
         // pop the global 'appserver-agent' scope.  this scope is pushed in the 'prepare' method, and it needs
         // to be popped here, to allow processing of the dynamic DMOs and other global finalizables
         TransactionManager.popScope();

         // notify the agent's stopper thread
         stopper.interrupt();
      }
   }

   /**
    * Reset the context of this agent.  Called only if the appserver is in State-reset operating
    * mode.
    * 
    * @param    mode
    *           How to execute the reset context command - immediately or send it to the agent.
    */
   public void resetContext(InvocationModes mode)
   {
      AgentCommand command = new ResetContextCommand();
      switch (mode)
      {
         case IMMEDIATELY:
            command.doExecute();
            break;
         default:
            waitForCommand(null, command);
            break;
      }
   }
   /**
    * Create a new instance of the specified qualified class name.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    name
    *           The legacy qualified class name for the legacy class.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The constructor's arguments.
    *
    * @return   A {@link AppServerInvocationResult} instance, which holds the code for the remote
    *           object, all updated OUTPUT/INPUT-OUTPUT arguments and any returned value.
    */
   public AppServerInvocationResult newInstance(String    id,
                                                int       requestId,
                                                long      timeout,
                                                character name,
                                                String    modes,
                                                Object... args)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(invocationResult, new AgentCommand(invocationResult, id, requestId, timeout)
      {
         @Override
         public void execute()
         {
            final handle hProc = new handle();

            // add the topmost scope to the TransactionManager
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);
            object<? extends BaseObject> instance = new ObjectVar<>(BaseObject.class);
            
            // make a copy of the requester's args
            Object[] requesterArgs = new Object[args.length];
            System.arraycopy(args, 0, requesterArgs, 0, requesterArgs.length);

            try
            {
               preProcessArguments(modes, args);
               
               // this will 'pin' at least a reference, so that it will not be implicitly deleted
               instance.assign(ObjectOps.newDynamicInstance(name, modes, args));
               
               if (!instance.isUnknown())
               {
                  hProc.assign(new ExternalProgramWrapper(instance.ref()));
               }
            }
            catch (StopConditionException e)
            {
               if (invokeFailure(e))
               {
                  return;
               }
               
               throw e;
            }
            finally
            {
               Object[] finalArgs = postProcessArguments(modes, args, requesterArgs);
               invocationResult.setArguments(finalArgs);
               invocationResult.setModes(modes);

               // remove the topmost scope from the TransactionManager
               TransactionManager.popScope();
            }

            // make it bound only if the invocation was successful
            bind();
            
            // set the agent's procedure
            addProcedure(hProc);
            
            invocationResult.setProcedureId(hProc.toString());
            invocationResult.setAgentId(agentId);
         }
      });

      return invocationResult;
   }

   /**
    * Invoke a method in a remotely instantiated class.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    returnValue
    *           Flag indicating if the return value is required.
    * @param    name
    *           The legacy 4GL name for the class method.
    * @param    resId
    *           The code of a remotely instantiated object in which to invoke the method.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The method's arguments.
    * 
    * @return    A {@link AppServerInvocationResult} instance with all updated OUTPUT/INPUT-OUTPUT
    *            arguments and any returned value.
    */
   public AppServerInvocationResult invokeMethod(String    id,
                                                 int       requestId,
                                                 long      timeout,
                                                 boolean   returnValue,
                                                 character name,
                                                 String    resId,
                                                 String    modes,
                                                 Object... args)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(invocationResult, new AgentCommand(invocationResult, id, requestId, timeout)
      {
         @Override
         public void execute()
         {
            if (!isBound())
            {
               // TODO: error out?
               return;
            }

            // add the topmost scope to the TransactionManager
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);
            
            // resolve the procedure - this is actually 'extProg', validated when the agent was acquired.
            handle proc = handle.fromString(resId);
            object<? extends BaseObject> ref = new object<>((BaseObject) proc.get());

            try
            {
               invokeMethodImpl(invocationResult,
                                id,
                                name,
                                ref,
                                returnValue,
                                modes,
                                args);
            }
            finally
            {
               // remove the topmost scope from the TransactionManager
               TransactionManager.popScope();
            }
         }
      });
      
      return invocationResult;
   }
   
   /**
    * Invoke the specified external procedure as persistent and wait (or not) for it to complete.
    * This is used to invoke any connect/disconnect and startup/shutdown procedures.
    * <p>
    * The "kind" parameter will always receive one of the "Startup", "Shutdown", "Connect" or 
    * "Disconnect" values.
    * 
    * @param    res
    *           The instance where to save the invocation result. May be <code>null</code>.
    * @param    id
    *           The appserver connection id. 
    * @param    mode
    *           Flag indicating how to invoke the external procedure.
    * @param    proc
    *           The legacy 4GL name for the external procedure.
    * @param    hProc
    *           If not-null, the handle where to save this procedure.
    * @param    kind
    *           A helper text about this procedure.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   <code>true</code> if no error was raised during the procedure invocation.
    */
   public boolean invoke(AppServerInvocationResult res,
                         final String              id,
                         final InvocationModes     mode,
                         final String              proc,
                         final handle              hProc,
                         final String              kind,
                         final String              modes,
                         final Object...           args)
   {
      return invoke(null, res, id, mode, proc, hProc, kind, modes, args);
   }
   
   /**
    * Invoke the specified external procedure as persistent and wait (or not) for it to complete.
    * This is used to invoke any connect/disconnect and startup/shutdown procedures.
    * <p>
    * The "kind" parameter will always receive one of the "Startup", "Shutdown", "Connect" or 
    * "Disconnect" values.
    * 
    * @param    requestInfo
    *           The remote {@link com.goldencode.p2j.oo.lang.OerequestInfo} details, as set by
    *           {@link com.goldencode.p2j.oo.lang.OerequestInfo#toArray()}.
    * @param    res
    *           The instance where to save the invocation result. May be <code>null</code>.
    * @param    id
    *           The appserver connection id. 
    * @param    mode
    *           Flag indicating how to invoke the external procedure.
    * @param    proc
    *           The legacy 4GL name for the external procedure.
    * @param    hProc
    *           If not-null, the handle where to save this procedure.
    * @param    kind
    *           A helper text about this procedure.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   <code>true</code> if no error was raised during the procedure invocation.
    */
   public boolean invoke(Object[]                  requestInfo,
                         AppServerInvocationResult res,
                         final String              id,
                         final InvocationModes     mode,
                         final String              proc,
                         final handle              hProc,
                         final String              kind,
                         final String              modes,
                         final Object...           args)
   {
      final AppServerInvocationResult result = (res == null ? new AppServerInvocationResult() 
                                                            : res);

      AgentCommand command = new AgentCommand(requestInfo, result, id) 
      {
         @Override
         public void execute()
         {
            // the procedure is ran persistent only in certain cases:
            // 1. startup procedure 
            //   - automatically deleted when agent terminates
            //   - can be explicitly deleted via "delete object this-procedure".
            // 2. connect procedure
            //   - only for State-reset and State-aware modes
            //   - can be explicitly deleted via "delete object this-procedure".
            //   - automatically deleted on context reset, if State-reset
            
            // in both cases, the explicit delete is postponed until the program (startup or 
            // connect) terminates

            AppServerOperatingMode mode = pool.getOperatingMode();
            boolean persistentConnect = ("Connect".equals(kind) && 
                                         (mode == AppServerOperatingMode.STATE_RESET ||
                                          mode == AppServerOperatingMode.STATE_AWARE));
            boolean startup = ("Startup".equals(kind));
            
            boolean persistent = (persistentConnect || startup);
            
            // add the topmost scope to the TransactionManager
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);
            Throwable error = null;
            try
            {
               if (persistent)
               {
                  ControlFlowOps.invokePersistentSetWithMode(proc, hProc, modes, args);
                  
               }
               else
               {
                  ControlFlowOps.invokeExternalProcedure(new character(proc), false, null, false,  
                                                         null, modes, args);
               }
            }
            catch (Throwable t)
            {
               boolean ignore = false;

               Throwable cause = t;
               while (cause != null)
               {
                  if (cause instanceof UnstoppableExitException)
                  {
                     LogicalTerminal.activateBatchMode(true);
                     ignore = true;
                     break;
                  }
                  
                  cause = cause.getCause();
               }
               
               if (!ignore)
               {
                  error = t;
               }
            }
            finally
            {
               // remove the topmost scope from the TransactionManager
               TransactionManager.popScope();
            }
   
            if (error != null && LOG.isWarnEnabled())
            {
               // procedure has failed
               LOG.warn(String.format("%s procedure has failed for appserver %s:",
                                      kind, pool.getName()), 
                        error);
            }
            
            result.setError(error);
            if (ControlFlowOps.isReturnValueSetInRootScope())
            {
               result.setResult(ControlFlowOps.getReturnValue());
            }

            // check if we are calling the disconnect procedure
            if ("Disconnect".equals(kind))
            {
               pool.terminateConnection(id, true);
            }
         }
      };
      
      switch (mode)
      {
         case IMMEDIATELY:
            try
            {
               command.doExecute();
            }
            catch (Throwable t)
            {
               result.setError(t);
            }
            break;

         case AGENT_NO_WAIT:
            sendCommand(command);
            break;

         case AGENT_WAIT:
            waitForCommand(result, command);
            break;
      }
      
      return (result.getError() == null);
   }

   /**
    * Invoke a remote external procedure, internal procedure or user-defined function. The 
    * validation for the program name and arguments will be done on the remote side.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    cfg
    *           The invoke configuration.
    * 
    * @return    A {@link AppServerInvocationResult} instance with all updated OUTPUT/INPUT-OUTPUT
    *            arguments and any returned value.
    */
   public AppServerInvocationResult invoke(final String    id,
                                           final int       requestId,
                                           final long      timeout,
                                           final InvokeConfig cfg)
   {
      String token = cfg.getWebServiceToken();

      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();
      waitForCommand(invocationResult, token, 
      new AgentCommand(invocationResult, id, requestId, timeout)
      {
         @SuppressWarnings({ "unchecked", "rawtypes" })
         @Override
         public void execute()
         {
            object<? extends OerequestInfo> oreq = 
               (object) executeScoped(() -> SessionUtils.currentRequestInfo());
            object<? extends OerequestInfo> oresp = 
               (object) executeScoped(() -> SessionUtils.currentResponseInfo());

            OerequestInfo req = oreq.ref;
            OerequestInfo resp = oresp.ref;
            if (req != null)
            {
               executeScoped(() -> 
               {
                  cfg.restoreRequest(req, getAgentId(), getThreadId(), getSessionId());
                  resp.initialize(req, false);
                  return null;
               });
            }
            
            if (!invokeActivate(id, invocationResult))
            {
               return;
            }
            
            // resolve the procedure
            handle inHandle = cfg.getInHandle();
            String resId = null;
            if (cfg.isSingleRun() || cfg.isSingleton())
            {
               DeferredProgram deferred = (DeferredProgram) inHandle.get();
               deferred.setSingleRun(cfg.isSingleRun());
               deferred.setSingleton(cfg.isSingleton());

               try
               {
                  inHandle = (handle) executeScoped(() -> ControlFlowOps.resolveInstance(deferred, cfg.isClass()));
               }
               catch (ConditionException exc)
               {
                  invocationResult.setError(exc);
               }
               catch (Throwable t)
               {
                  t.printStackTrace();
                  invocationResult.setError(t);
               }

               if (inHandle == null || !inHandle._isValid())
               {
                  return;
               }
            }
            else if (inHandle != null)
            {
               resId = ((ProxyProcedureWrapper) inHandle.getResource()).getResourceId();
               // resolve the procedure - this is actually 'extProg', validated when the agent was acquired.
               inHandle = handle.fromString(resId);
            }
            
            boolean disconnect = false;
            try
            {
               handle finHandle = inHandle;
               Object[] args = cfg.getArguments() == null ? new Object[0] : cfg.getArguments();
               
               if (cfg.isClass())
               {
                  executeScoped(() ->
                  {
                     invokeMethodImpl(invocationResult, 
                                      id, 
                                      cfg.getTarget(),
                                      new object(((ObjectResource) finHandle.getResource()).ref()),
                                      cfg.isReturnValue(),
                                      cfg.getModes(),
                                      args);
                     return null;
                  });
               }
               else if (cfg.getProcedureHandle() != null)
               {
                  executeScoped(() ->
                  {
                     invokePersistentImpl(invocationResult, 
                                          id, 
                                          cfg.getTarget(), 
                                          cfg.getProcedureHandle(), 
                                          cfg.isTransactionDistinct(), 
                                          cfg.getModes(),
                                          args);
                     return null;
                  });
               }
               else
               {
                  disconnect = ((logical) executeScoped(() ->
                  {
                     return new logical(invokeImpl(invocationResult,
                                                   id,
                                                   cfg.getTarget(),
                                                   finHandle,
                                                   cfg.isFunction(),
                                                   cfg.isDynamicFunction(),
                                                   cfg.isSuperCall(),
                                                   cfg.isTransactionDistinct(),
                                                   cfg.getModes(),
                                                   args));
                  })).booleanValue();
               }
            }
            finally
            {
               if (cfg.isSingleRun())
               {
                  ProcedureManager.deleteInstance(inHandle, cfg.isClass());
               }
               if (resId != null && !inHandle._isValid())
               {
                  invocationResult.setInvalidInHandle(true);
               }
               
               invokeDeactivate(id);

               if (disconnect)
               {
                  AppServerManager.disconnect(pool, id);
               }
            }
         }
      });
      
      return invocationResult;
   }

   /**
    * Invoke a remote external procedure, internal procedure or user-defined function. The 
    * validation for the program name and arguments will be done on the remote side.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    resId
    *           The code of a remote procedure in which to invoke the procedure or function.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return    A {@link AppServerInvocationResult} instance with all updated OUTPUT/INPUT-OUTPUT
    *            arguments and any returned value.
    */
   public AppServerInvocationResult invoke(final String    id,
                                           final int       requestId,
                                           final long      timeout,
                                           final character name,
                                           final String    resId,
                                           final boolean   function,
                                           final boolean   dynamicFunction,
                                           final boolean   superCall,
                                           final boolean   transactionDistinct,
                                           final String    modes,
                                           final Object... args)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(invocationResult, new AgentCommand(invocationResult, id, requestId, timeout)
      {
         @Override
         public void execute()
         {
            invokeActivate(id, invocationResult);
            
            // resolve the procedure - this is actually 'extProg', validated when the agent was acquired.
            handle proc = (resId == null ? null : handle.fromString(resId));
            boolean disconnect = false;
            try
            {
               disconnect = ((logical) executeScoped(() ->
               {
                  return new logical(invokeImpl(invocationResult,
                                                id,
                                                name,
                                                proc,
                                                function,
                                                dynamicFunction,
                                                superCall,
                                                transactionDistinct,
                                                modes,
                                                args));
               })).booleanValue();
            }
            finally
            {
               invokeDeactivate(id);

               if (disconnect)
               {
                  AppServerManager.disconnect(pool, id);
               }
            }
         }
      });
      
      return invocationResult;
   }

   /**
    * Invoke the given Java method directly.
    * <p>
    * All arguments are sent by references: if it was marked as an OUTPUT or INPUT-OUTPUT, any change to that
    * instance will be reflected back to the caller.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    cfg
    *           The configuration with the Java method to be invoked.
    * 
    * @return   A {@link AppServerInvocationResult} instance with all updated OUTPUT/INPUT-OUTPUT
    *           arguments and any returned value.
    */
   public AppServerInvocationResult invoke(String id, int requestId, int timeout, JavaInvokeConfig cfg)
   {
      String token = cfg.getWebServiceToken();
      
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();
      waitForCommand(invocationResult, token, 
      new AgentCommand(invocationResult, id, requestId, timeout)
      {
         @SuppressWarnings({ "unchecked", "rawtypes" })
         @Override
         public void execute()
         {
            if (!invokeActivate(id, invocationResult))
            {
               return;
            }
            
            boolean disconnect = false;
            try
            {
               disconnect = ((logical) executeScoped(() ->
               {
                  return new logical(invokeJavaMethod(invocationResult, id, cfg));
               })).booleanValue();
            }
            finally
            {
               invokeDeactivate(id);
               
               if (disconnect)
               {
                  AppServerManager.disconnect(pool, id);
               }
            }
         }
      });
      
      return invocationResult;
   }

   /**
    * Obtain a proxy for the given external procedure. This will also reserve an agent which will
    * be used for all subsequent invocation (including external procedure initialization).
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           A name of a remote external procedure.
    *           
    * @return   A proxy to this procedure. May be invalid if the remote side can't resolve the
    *           procedure.
    */
   ExternalProgramWrapper obtainProxy(final String id, final String name)
   {
      final handle hProc = new handle();
      
      waitForCommand(null, new AgentCommand(null, id)
      {
         @Override
         public void execute()
         {
            // add the topmost scope to the TransactionManager; this will be popped only when:
            // a. the ext prog can't be resolved
            // b. the ext prog has been initialized
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);

            ExternalProgramWrapper extProg = null;
            
            try
            {
               extProg = ControlFlowOps.resolveExternalProcedure(name, pool.getExports());
               
               if (extProg != null)
               {
                  hProc.assign(extProg);
   
                  // make it bound only if the invocation was successful
                  bind();
                  
                  // associate the agent with this ext prog
                  addProcedure(hProc);
               }
            }
            
            finally
            {
               if (extProg == null)
               {
                  // remove the topmost scope from the TransactionManager, as the ext prog could
                  // not be resolved
                  TransactionManager.popScope();
               }
            }
         }
      });

      return (ExternalProgramWrapper) hProc.getResource();
   }

   /**
    * Initialize the specified proxy procedure, by calling the <code>execute</code> method.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    code
    *           The external procedure id.
    * @param    requestId
    *           The ID of the request.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    *           
    * @return   The result of the external program initialization.
    */
   public AppServerInvocationResult initializeProxy(final String   id,
                                                    final String   code,
                                                    final int      requestId,
                                                    final boolean  transactionDistinct,
                                                    final String   modes,
                                                    final Object[] args)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(invocationResult, new AgentCommand(invocationResult, id, requestId)
      {
         @Override
         public void execute()
         {
            ExternalProgramWrapper extProg = getProcedure(code);

            // make a copy of the requester's args
            Object[] requesterArgs = new Object[args.length];
            System.arraycopy(args, 0, requesterArgs, 0, requesterArgs.length);

            // WARNING: do not add a scope here - obtainProxy will add it and keep it on the stack
            
            try
            {
               preProcessArguments(modes, args);
               
               ControlFlowOps.initializeExternalProcedure(extProg,
                                                          transactionDistinct, 
                                                          modes,
                                                          args);
            }
            catch (StopConditionException e)
            {
               if (invokeFailure(e))
               {
                  return;
               }
               
               throw e;
            }
            finally
            {
               Object[] finalArgs = postProcessArguments(modes, args, requesterArgs);
               invocationResult.setArguments(finalArgs);
               invocationResult.setModes(modes);

               // remove the topmost scope from the TransactionManager.  this scope was added by obtainProxy
               TransactionManager.popScope();
            }

            if (ControlFlowOps.isReturnValueSetInRootScope())
            {
               invocationResult.setResult(ControlFlowOps.getReturnValue());
            }
            invocationResult.setProcedureId(new handle(extProg).toString());
            invocationResult.setAgentId(agentId);
         }
      });
      
      return invocationResult;
   }
   
   /**
    * Invoke a remote external procedure, persistently. The validation of  the program name and
    * arguments will be done on the remote side.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    *
    * @return    A {@link AppServerInvocationResult} instance, which holds the code for the remote
    *            procedure, all updated OUTPUT/INPUT-OUTPUT arguments and any returned value.
    */
   public AppServerInvocationResult invokePersistent(final String    id,
                                                     final int       requestId,
                                                     final long      timeout,
                                                     final character name,
                                                     final boolean   transactionDistinct,
                                                     final String    modes,
                                                     final Object... args)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(invocationResult, new AgentCommand(invocationResult, id, requestId, timeout)
      {
         @Override
         public void execute()
         {
            final handle hProc = new handle();

            // add the topmost scope to the TransactionManager
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);
            try
            {
               invokePersistentImpl(invocationResult, 
                                    id, 
                                    name, 
                                    hProc, 
                                    transactionDistinct, 
                                    modes, 
                                    args);
            }
            finally
            {
               // remove the topmost scope from the TransactionManager
               TransactionManager.popScope();
            }
         }
      });

      return invocationResult;
   }
   
   /**
    * Handle a generic web service request.  This exposes the Jetty request and response objects
    * to the Agent which will process it, so it is required for the appserver to be local.
    * 
    * @param    id
    *           The appserver connection id. 
    * @param    requestId
    *           The ID of the request.
    * @param    timeout
    *           The maximum allowed time for this invocation to complete.  In milliseconds.
    *           Use <code>0</code> to disable the timeout.
    * @param    handler
    *           The {@link com.goldencode.p2j.oo.web.WebHandler} implementation.
    * @param    basepath
    *           The service basepath.
    * @param    paths
    *           The split paths for this handler.
    * @param    token
    *           When not null, it represents the token of a FWD context created for an authenticated and 
    *           authorized web service call.  The API call will be performed in that context.
    * @param    target
    *           The request original target.
    * @param    request
    *           The Jetty request.
    * @param    response
    *           The Jetty response.
    *           
    * @return   A {@link AppServerInvocationResult} instance with any error throw during the web service
    *           processing.
    */
   public AppServerInvocationResult handleWebService(String                      id,
                                                     int                         requestId,
                                                     long                        timeout,
                                                     Class<? extends WebHandler> handler, 
                                                     String                      basepath,
                                                     String[]                    paths, 
                                                     String                      token,
                                                     String                      target, 
                                                     HttpServletRequest          request,
                                                     HttpServletResponse         response)
   {
      final AppServerInvocationResult invocationResult = new AppServerInvocationResult();

      waitForCommand(null, token, new AgentCommand(null, id, requestId, timeout)
      {
         @Override
         public void execute()
         {
            object<WebHandler> ohandler = TypeFactory.object(WebHandler.class);

            // add the topmost scope to the TransactionManager
            TransactionManager.pushScope("startup",
                                         TransactionManager.NO_TRANSACTION,
                                         true,
                                         true,
                                         false,
                                         false);

            try
            {
               try
               {
                  WebHandler.initialize(basepath, target, paths, request, response);
                  
                  ohandler.assign(ObjectOps.newInstance(handler));
                  
                  if (!ohandler._isValid())
                  {
                     // TODO:
                     return;
                  }
                  @SuppressWarnings("unused")
                  integer res = ohandler.ref().handleRequest();
                  // TODO: non-zero means a static error page must be shown, for that error...
               }
               catch (Throwable t)
               {
                  invocationResult.setError(t);
               }
               finally
               {
                  if (ohandler._isValid())
                  {
                     ObjectOps.delete(ohandler);
                  }
                  WebHandler.clear();
               }
            }
            finally
            {
               // remove the topmost scope from the TransactionManager
               TransactionManager.popScope();
            }
         }
      });
      
      return invocationResult;
   }


   /**
    * Send the terminate command to this agent.
    */
   public void terminate()
   {
      // interrupt current work
      stop();
      
      // send the terminate command
      sendCommand(new AgentCommand(null, null)
      {
         @Override
         public void execute()
         {
            synchronized (Agent.this)
            {
               terminated = true;
            }
         }
      });
   }

   /**
    * Check if this agent has been terminated.
    * 
    * @return   See above.
    */
   public synchronized boolean terminated()
   {
      return terminated;
   }
   
   /**
    * Bind the agent to the connection.
    */
   public synchronized void bind()
   {
      this.bound = true;
   }
   
   /**
    * Unbind the agent from the connection.
    */
   public synchronized void unbind()
   {
      this.bound = false;
   }
   
   /**
    * Check if the agent is bound to the connection.
    * 
    * @return   See above.
    */
   public synchronized boolean isBound()
   {
      return bound;
   }
   
   /**
    * Delete the agent's currently bound {@link #remoteProcedures procedures}.
    */
   public void deleteProcedures()
   {
      if (remoteProcedures.isEmpty())
      {
         return;
      }
      
      List<AgentRemoteProcedure> procs = new ArrayList<>(remoteProcedures.values());
      for (AgentRemoteProcedure remote : procs)
      {
         Object referent = remote.extProg.get();
         if (referent instanceof _BaseObject_)
         {
            deleteObject(remote.extProgResourceId);
         }
         else
         {
            deleteProcedure(remote.extProgResourceId, null);
         }
      }
      
      // clear them.
      remoteProcedures.clear();
   }
   
   /**
    * Delete the specified procedure, under the Agent's context.
    * 
    * @param    resId
    *           The resource ID for this procedure.
    * @param    token
    *           When not null, it represents the token of a FWD context created for an authenticated and 
    *           authorized web service call.  The API call will be performed in that context.
    */
   public void deleteProcedure(String resId, String token)
   {
      waitForCommand(null, token, new AgentCommand(null, null)
      {
         @Override
         public void execute()
         {
            ExternalProgramWrapper proc = getProcedure(resId);
            
            if (handle.knownResource(proc))
            {
               proc.delete();
            }
            
            removeProcedure(resId);
         }
      });
   }
   
   /**
    * Delete the specified legacy object, under the Agent's context.
    * 
    * @param    resId
    *           The resource ID for this legacy object.
    */
   public void deleteObject(String resId)
   {
      waitForCommand(null, new AgentCommand(null, null)
      {
         @Override
         public void execute()
         {
            ExternalProgramWrapper proc = getProcedure(resId);

            if (handle.knownResource(proc))
            {
               ObjectOps.delete((BaseObject) ((ExternalProgramWrapper) proc).get());
            }
            
            removeProcedure(resId);
         }
      });
   }
   
   /**
    * Set the current Agent's connection ID.
    * 
    * @param    id
    *           The connection ID.
    */
   void setConnectionId(String id)
   {
      waitForCommand(null, new AgentCommand(null, id)
      {
         @Override
         public void execute()
         {
            AppServerManager.setConnectionId(id);
         }
      });
   }

   /**
    * Validate the connection ID and procedure ID with this agent - the connection ID must be the same as
    * the {@link AgentRemoteProcedure#requesterConnectionId} which created the remote procedure, and the 
    * procedure ID must be {@link #remoteProcedures registered} with this agent.
    * 
    * @param    id
    *           The connection ID to validate.
    * @param    procId
    *           The procedure ID to validate.
    *           
    * @throws   IllegalArgumentException
    *           If the connection ID and procedure ID are not for this agent, as the client performs an illegal
    *           request.
    */
   synchronized void validateRequest(String id, String procId)
   {
      AgentRemoteProcedure remote = remoteProcedures.get(procId);
      
      if (remote == null || !remote.requesterConnectionId.equals(id))
      {
         throw new IllegalArgumentException("The procedure can not be accessed on this connection!");
      }
   }
   
   /**
    * Get the operating mode for this appserver.
    * 
    * @return   See above.
    */
   AppServerOperatingMode getOperatingMode()
   {
      return pool.getOperatingMode();
   }
   
   /**
    * Check if this agent has persistent procedures ran on that connection ID.
    * 
    * @param    id
    *           The connection ID.
    *           
    * @return   See above.
    */
   boolean hasProcedures(String id)
   {
      if (remoteProcedures.isEmpty())
      {
         return false;
      }
      
      for (AgentRemoteProcedure remote : remoteProcedures.values())
      {
         if (id.equals(remote.requesterConnectionId))
         {
            return true;
         }
      }
      
      return false;
   }

   /**
    * Check if this appserver definition is for PASOE.
    * 
    * @return    See above.
    */
   boolean isPasoe()
   {
      return pool.isPasoe();
   }

   /**
    * Send a STOP condition to the request currently being executed on this connection.
    * 
    * @throws   IllegalStateException
    *           If the appserver is not connected.
    */
   void stop()
   {
      // the STOP condition needs to be raised from a thread with the agent's context.  assuming
      // the agent's thread is busy executing another command, we can't use the agent's thread
      // to send this stop.  so an AssociatedThread with the same context as the Agent is used 
      // instead to wait for STOP commands
      synchronized (stopper)
      {
         stopper.notify();
      }
   }
   
   /**
    * Get the SERVER-CONNECTION-CONTEXT attribute of this session.
    * 
    * @param    connectionID
    *           The connection ID for this appserver.
    *           
    * @return   See above.
    */
   String getServerConnectionContext(String connectionID)
   {
      return pool.getServerConnectionContext(connectionID);
   }
   
   /**
    * Set the SERVER-CONNECTION-CONTEXT attribute of this session.
    * 
    * @param    connectionID
    *           The connection ID for this appserver.
    * @param    serverConnectionContext
    *           The attribute's value.
    */
   void setServerConnectionContext(String connectionID, String serverConnectionContext)
   {
      pool.setServerConnectionContext(connectionID, serverConnectionContext);
   }
   
   /**
    * Get the sessionId for this session.
    * 
    * @return   Always <code>null</code>.
    */
   private Integer getSessionId()
   {
      if (!pool.isPasoe())
      {
         return null;
      }
      
      // TODO: do we need a real value?
      return null;
   }
   
   /**
    * Get the threadId for this session.
    * 
    * @return   Always <code>null</code>.
    */
   private Integer getThreadId()
   {
      if (!pool.isPasoe())
      {
         return null;
      }
      
      // TODO: do we need a real value?
      return null;
   }
   
   /**
    * Invoke a legacy OO method.
    * 
    * @param    invocationResult
    *           The result where to save the arguments and any return value.
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           The legacy 4GL name for the OO method.
    * @param    ref
    *           The object where to invoke the method.
    * @param    returnValue
    *           Flag indicating if this OO method call must return a value.
    * @param    modes
    *           The method argument's modes, at the caller.
    * @param    args
    *           The arguments.
    */
   private void invokeMethodImpl(AppServerInvocationResult invocationResult,
                                 String       id,
                                 character    name,
                                 object<?>    ref,
                                 boolean      returnValue,
                                 String       modes,
                                 Object...    args)
   {
      // make a copy of the requester's args
      final Object[] requesterArgs = new Object[args.length];
      System.arraycopy(args, 0, requesterArgs, 0, requesterArgs.length);
      
      try
      {
         preProcessArguments(modes, args);

         try
         {
            Object result = null;
            if (returnValue)
            {
               result = ObjectOps.invoke(ref, name, modes, args);
            }
            else
            {
               ObjectOps.invokeStandalone(ref, name, modes, args);;

               if (result instanceof handle)
               {
                  // handle results are converted to a container holding only the remote resource
                  handle h = (handle) result;
                  result = new RemoteResourceImpl(h.isUnknown() ? null : h.toStringMessage());
               }
               invocationResult.setResult(result);
            }
         }
         catch (StopConditionException e)
         {
            throw e;
         }
         finally
         {
            Object[] finalArgs = postProcessArguments(modes, args, requesterArgs);
            invocationResult.setArguments(finalArgs);
            invocationResult.setModes(modes);
         }
      }
      catch (ErrorConditionException e)
      {
         checkError(invocationResult, e);
      }
      catch (QuitConditionException e)
      {
         invocationResult.setError(e);
      }
      catch (Throwable t)
      {
         boolean ignore = false;

         Throwable cause = t;
         while (cause != null)
         {
            if (cause instanceof UnstoppableExitException)
            {
               LogicalTerminal.activateBatchMode(true);
               ignore = true;
               break;
            }
            
            cause = cause.getCause();
         }
         
         if (!ignore)
         {
            invocationResult.setError(t);
         }
      }
   }
   
   /**
    * Invoke a remote external procedure, persistently. The validation of  the program name and
    * arguments will be done on the remote side.
    * 
    * @param    invocationResult
    *           The result to be sent back to the remote side.
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           The legacy 4GL name for the procedure.
    * @param    hProc
    *           The handle where to save the persistent procedure.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    */
   private void invokePersistentImpl(AppServerInvocationResult invocationResult,
                                     String    id,
                                     character name,
                                     handle    hProc,
                                     boolean   transactionDistinct,
                                     String    modes,
                                     Object... args)
   {
      // make a copy of the requester's args
      Object[] requesterArgs = new Object[args.length];
      System.arraycopy(args, 0, requesterArgs, 0, requesterArgs.length);

      try
      {
         preProcessArguments(modes, args);
         
         character exports = pool.getExports();
         ControlFlowOps.invokeExternalProcedure(name, true, hProc, transactionDistinct,
                                                exports, modes, args);
      }
      catch (StopConditionException e)
      {
         if (invokeFailure(e))
         {
            return;
         }
         
         throw e;
      }
      finally
      {
         Object[] finalArgs = postProcessArguments(modes, args, requesterArgs);
         invocationResult.setArguments(finalArgs);
         invocationResult.setModes(modes);
      }

      // make it bound only if the invocation was successful
      bind();
      
      // set the agent's procedure
      addProcedure(hProc);
      
      if (ControlFlowOps.isReturnValueSetInRootScope())
      {
         invocationResult.setResult(ControlFlowOps.getReturnValue());
      }

      invocationResult.setProcedureId(hProc.toString());
      invocationResult.setAgentId(agentId);
   }
   
   /**
    * Invoke an external procedure, internal procedure or user-defined function.
    * 
    * @param    invocationResult
    *           The result where to save the arguments and any return value.
    * @param    id
    *           The appserver connection id. 
    * @param    name
    *           The legacy 4GL name for the procedure/function.
    * @param    proc
    *           The external procedure in which to invoke the procedure or function.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May be <code>null</code>
    * @param    args
    *           The procedure's arguments.
    * 
    * @return    The result returned by this procedure or user-defined function call. 
    */
   private boolean invokeImpl(AppServerInvocationResult invocationResult,
                              String    id,
                              character name,
                              handle    proc,
                              boolean   function,
                              boolean   dynamicFunction,
                              boolean   superCall,
                              boolean   transactionDistinct,
                              String    modes,
                              Object... args)
   {
      character exports = pool.getExports();
      boolean disconnect = false;

      // make a copy of the requester's args
      final Object[] requesterArgs = new Object[args.length];
      System.arraycopy(args, 0, requesterArgs, 0, requesterArgs.length);
      
      try
      {
         preProcessArguments(modes, args);

         Object result = invocationResult.getResult();  // value could be set in appserver activate procedure
         if (proc == null)
         {
            // force invocation of an external procedure
            ControlFlowOps.invokeExternalProcedure(name, false, null, transactionDistinct,
                                                   exports, modes, args);
            if (ControlFlowOps.isReturnValueSetInRootScope())
            {
               result = ControlFlowOps.getReturnValue();
            }
         }
         else
         {
            // this is an internal-entry call
            result = ControlFlowOps.invoke(proc, name, function,
                                         dynamicFunction, superCall,
                                         transactionDistinct, modes, args);
            if (!function)
            {
               result = ControlFlowOps.isReturnValueSetInRootScope() ? ControlFlowOps.getReturnValue() : null;
            }
         }

         if (result instanceof handle)
         {
            // handle results are converted to a container holding only the remote resource
            handle h = (handle) result;
            result = new RemoteResourceImpl(h.isUnknown() ? null : h.toStringMessage());
         }
         invocationResult.setResult(result);
      }
      catch (ErrorConditionException e)
      {
         checkError(invocationResult, e);
      }
      catch (QuitConditionException e)
      {
         invocationResult.setError(e);
         disconnect = true;
      }
      catch (StopConditionException e)
      {
         if (invokeFailure(e))
         {
            return false;
         }
         
         throw e;
      }
      catch (Throwable t)
      {
         boolean ignore = false;

         Throwable cause = t;
         while (cause != null)
         {
            if (cause instanceof UnstoppableExitException)
            {
               LogicalTerminal.activateBatchMode(true);
               ignore = true;
               break;
            }
            
            cause = cause.getCause();
         }
         
         if (!ignore)
         {
            invocationResult.setError(t);
         }
      }
      finally
      {
         Object[] finalArgs = postProcessArguments(modes, args, requesterArgs);
         invocationResult.setArguments(finalArgs);
         invocationResult.setModes(modes);
      }
      
      return disconnect;
   }
   
   /**
    * Invoke the given Java method directly.
    * <p>
    * All arguments are sent by references: if it was marked as an OUTPUT or INPUT-OUTPUT, any change to that
    * instance will be reflected back to the caller.
    * 
    * @param    invocationResult
    *           The instance where to save the method's returned value or any error.
    * @param    id
    *           The appserver connection id. 
    * @param    cfg
    *           The configuration with the Java method to be invoked.
    *           
    * @return   <code>true</code> if no error was raised during the method invocation.
    */
   private boolean invokeJavaMethod(AppServerInvocationResult invocationResult, String id, JavaInvokeConfig cfg)
   {
      boolean disconnect = false;

      Object[] args = cfg.getArguments() == null ? new Object[0] : cfg.getArguments();
      
      try
      {
         Method mthd = cfg.getMethod();
         Class<?> clz = cfg.getClazz();
         Object result = null;
         Object instance = null;
         if (!Modifier.isStatic(mthd.getModifiers()))
         {
            instance = clz.newInstance();
         }
         result = mthd.invoke(instance, args);

         invocationResult.setResult(result);
      }
      catch (ErrorConditionException e)
      {
         checkError(invocationResult, e);
      }
      catch (QuitConditionException e)
      {
         invocationResult.setError(e);
         disconnect = true;
      }
      catch (StopConditionException e)
      {
         if (invokeFailure(e))
         {
            return false;
         }
         
         throw e;
      }
      catch (Throwable t)
      {
         boolean ignore = false;

         Throwable cause = t;
         while (cause != null)
         {
            if (cause instanceof UnstoppableExitException)
            {
               LogicalTerminal.activateBatchMode(true);
               ignore = true;
               break;
            }
            
            cause = cause.getCause();
         }
         
         if (!ignore)
         {
            invocationResult.setError(t);
         }
      }
      finally
      {
         invocationResult.setArguments(args);
      }
      
      return disconnect;
   }
   
   /**
    * Invoke the deactivate procedure on the specified connection ID.
    * 
    * @param    id
    *           The appserver connection id.
    */
   private void invokeDeactivate(String id)
   {
      String deactivate = pool.getDeactivate();
      if (!isBound() && deactivate != null && deactivate.length() > 0)
      {
         // call the deactivate procedure (only if the agent is not bound)
         try
         {
            invokeScoped(new character(deactivate));
         }
         catch (Throwable t)
         {
            // ignore all errors
         }
      }
   }
   
   /**
    * Invoke the activate procedure on the specified connection ID.
    * 
    * @param    id
    *           The appserver connection id.
    * @param    invocationResult
    *           The result where to save the arguments and any return value.
    */
   @SuppressWarnings("incomplete-switch")
   private boolean invokeActivate(String id, AppServerInvocationResult invocationResult)
   {
      String activate = pool.getActivate();
      if (!isBound() && activate != null && activate.length() > 0)
      {
         // call the activate procedure (only if the agent is not already bound)
         try
         {
            invokeScoped(new character(activate));
         }
         catch (ConditionException ce)
         {
            Condition c = BlockManager.exceptionToCondition(ce);
            String type = c.toString();
            int code = 0;
            switch (c)
            {
               case ERROR:
                  code = 8025;
                  break;
               case STOP:
                  code = 8026;
                  break;
               case QUIT:
                  code = 8027;
                  break;
            }
   
            final String msg = "%s  activation procedure ended with an %s condition";
            invocationResult.setError(
               new NumberedException(String.format(msg, activate, type), code));
         }
         catch (Throwable t)
         {
            boolean ignore = false;
   
            Throwable cause = t;
            while (cause != null)
            {
               if (cause instanceof UnstoppableExitException)
               {
                  LogicalTerminal.activateBatchMode(true);
                  ignore = true;
                  break;
               }
               
               cause = cause.getCause();
            }
            
            if (!ignore)
            {
               invocationResult.setError(t);
            }
         }
         finally
         {
            // pass return value
            if (ControlFlowOps.isReturnValueSetInRootScope())
            {
               invocationResult.setResult(ControlFlowOps.getReturnValue());
            }
            
            // if an error was encountered during activate, call deactivate and disconnect
            if (invocationResult.getError() != null)
            {
               String deactivate = pool.getDeactivate();
               if (!isBound() && deactivate != null && deactivate.length() > 0)
               {
                  // call the deactivate procedure (only if the agent is not bound)
                  try
                  {
                     invokeScoped(new character(deactivate));
                  }
                  catch (Throwable t)
                  {
                     // ignore all errors
                  }
               }
   
               AppServerManager.disconnect(pool, id);
               
               return false;
            }
         }
      }
      
      return true;
   }

   /**
    * Send the given command to this agent and return immediately.
    * 
    * @param    command
    *           The command to be executed.
    */
   private synchronized void sendCommand(AgentCommand command)
   {
      commands.add(command);
      
      this.notify();
   }
   
   /**
    * Send the given command to the agent and wait for it to complete.
    * 
    * @param    result
    *           The instance where to place the command result.
    * @param    token
    *           When not null, it represents the token of a FWD context created for an authenticated and 
    *           authorized web service call.  The API call will be performed in that context.
    * @param    command
    *           The command to be executed.
    */
   private void waitForCommand(AppServerInvocationResult result, String token, AgentCommand command)
   {
      command.token = token;
      waitForCommand(result, command);
   }
   
   /**
    * Send the given command to the agent and wait for it to complete.
    * 
    * @param    result
    *           The instance where to place the command result.
    * @param    command
    *           The command to be executed.
    */
   private void waitForCommand(AppServerInvocationResult result, AgentCommand command)
   {
      if (terminated)
      {
         LOG.error("Trying to execute a command on a terminated agent!");
         return;
      }
      
      if (Thread.currentThread() == thread)
      {
         // this command was posted from the listening thread, execute immediately.
         command.doExecute();
         return;
      }
      
      sendCommand(command);

      // we need to synchronize on the loop, else the command might finish between the test and
      // the wait and the notification will never be picked up
      synchronized (command)
      {
         long timeout = command.getTimeout() * 1000; // number of maximum seconds to wait
         boolean waitForFinish = (timeout == 0);
         long start = (!waitForFinish ? 0 : System.currentTimeMillis());
         
         while (!command.isFinished())
         {
            try
            {
               command.wait(timeout);
            }
            catch (InterruptedException e)
            {
               if (LOG.isWarnEnabled())
               {
                  LOG.warn("Command was interrupted for appserver " + pool.getName(), e);
               }
            }
            
            if (!waitForFinish)
            {
               long finish = System.currentTimeMillis();
               long elapsed = finish - start;
               
               if (!command.isFinished() && elapsed >= timeout)
               {
                  // terminate the agent's command, and wait for it to be marked as "finished"
                  stop();
                  timeout = 0;
                  waitForFinish = true;
               }
               else
               {
                  // wait again, for the remainder time
                  timeout = timeout - elapsed;
               }
            }
         }
      }

      if (result != null && command.getError() != null)
      {
         result.setError(command.getError());
      }
   }
   
   /**
    * Preprocess the arguments sent by the requester so that they are converted to instances
    * compatible with the P2J runtime.  This includes:
    * <ul>
    * <li>Creating the handles for all received resources</li>
    * <li>Creating the temp-tables for the received {@link TableWrapper} instances</li>
    * <li>Creating {@link memptr} instances from all {@link MemoryBuffer} instances</li>
    * </ul>
    * 
    * @param    requesterModes
    *           The parameter modes sent by the requester.
    * @param    args
    *           The argument list sent by the requester.  The elements in this array will be
    *           changed as needed.
    */
   private void preProcessArguments(String requesterModes, Object[] args)
   {
      ControlFlowOps.resetReturnValueSetInRootScope();
      
      String modes = requesterModes;
      if (modes == null || modes.length() != args.length)
      {
         // no modes specified or incompatible modes
         return;
      }

      // a bogus class... for dataset parameters 
      ProcedureManager.setInstantingExternalProgramClass(Object.class);
      for (int i = 0; i < modes.length(); i++)
      {
         Object arg = args[i];

         boolean inputOutput = modes.charAt(i) == 'U';
         boolean output = modes.charAt(i) == 'O';
         @SuppressWarnings("unused")
         boolean input = modes.charAt(i) == 'I';
         
         if (arg instanceof TableWrapper)
         {
            // a result set will be sent only if the requester and the appserver are not in the  
            // same JVM
            TableWrapper rs = (TableWrapper) arg;
            arg = new TableParameter(rs, rs.isAppend());
            ((TableParameter) arg).setParameterIndex(i + 1);
         }
         else if (arg instanceof DatasetWrapper)
         {
            DatasetWrapper ds = (DatasetWrapper) arg;
            arg = new DataSetParameter(ds);
            ((DataSetParameter) arg).setParameterIndex(i + 1);
         }
         else
         {
            boolean isArray = arg.getClass().isArray();
            Object[] data = null;
            
            if (isArray)
            {
               // check if this argument is an array
               data = (Object[]) Array.newInstance(arg.getClass().getComponentType(), Array.getLength(arg));
               
               for (int j = 0; j < data.length; j++)
               {
                  data[j] = Array.get(arg, j);
               }
            }
            else
            {
               data = new Object[] { arg };
            }
            
            for (int j = 0; j < data.length; j++)
            {
               Object darg = data[j];
               
               if (darg instanceof RemoteResource)
               {
                  String resid = ((RemoteResource) darg).getResourceId();
                  darg = (resid == null ? new handle() : handle.fromString(resid));
               }
               else if (darg instanceof MemoryBuffer)
               {
                  // convert MemoryBuffer instances to memptr instances
                  byte[] bytes = ((MemoryBuffer) darg).getValue();
                  darg = bytes == null ? new memptr() : new memptr(bytes);
               }
               
               data[j] = darg;
            }
            
            if (isArray)
            {
               if (inputOutput || output)
               {
                  BaseDataType[] extArg = (BaseDataType[]) data;
                  int idx = i;
                  
                  if (inputOutput)
                  {
                     arg = new InputOutputExtentParameter<BaseDataType>()
                     {
                        @Override
                        public BaseDataType[] getVariable()
                        {
                           return extArg;
                        }
   
                        @Override
                        public void setVariable(BaseDataType[] newRef)
                        {
                           args[idx] = newRef;
                        }
                     };
                  }
                  else
                  {
                     arg = new OutputExtentParameter<BaseDataType>()
                     {
                        @Override
                        public BaseDataType[] getVariable()
                        {
                           return extArg;
                        }

                        @Override
                        public void setVariable(BaseDataType[] newRef)
                        {
                           args[idx] = newRef;
                        }
                     };
                  }
               }
               else
               {
                  arg = data;
               }
            }
            else
            {
               arg = data[0];
            }
         }
         
         args[i] = arg;
      }
      ProcedureManager.setInstantingExternalProgramClass(null);
   }

   /**
    * Post-process the given arguments so that all OUTPUT and INPUT-OUTPUT arguments will be sent
    * back to the requester. This includes:
    * <ul>
    *    <li>Serializing all table parameters via {@link TableWrapper} instances</li>
    *    <li>Sending the resource IDs for all handle parameters</li>
    *    <li>Converting all {@link memptr} instances to {@link MemoryBuffer} instances</li>
    * </ul>
    * 
    * @param    requesterModes
    *           The parameter modes sent by the requester.
    * @param    localArgs
    *           The argument list after the call has been performed.
    * @param    requesterArgs
    *           The argument list sent by the requester.
    * 
    * @return   The postprocessed arguments, to be sent back to the requester.
    */
   private Object[] postProcessArguments(String   requesterModes,
                                         Object[] localArgs,
                                         Object[] requesterArgs)
   {
      try
      {
         String modes = requesterModes;
         if (modes == null || requesterArgs.length != modes.length())
         {
            // no modes specified, so no arguments can be returned - assume all are input
            return new Object[0];
         }
         
         Object[] arguments = new Object[localArgs.length];
         for (int i = 0; i < modes.length(); i++)
         {
            char mode = modes.charAt(i);
   
            // in input mode, the argument is not sent back
            Object localArg = (mode == 'O' || mode == 'U' ? localArgs[i] : null);
            
            if (localArg == null)
            {
               // nothing to post-process if the local arg was determined to be null
               arguments[i] = null;
               continue;
            }
            Object requesterArg = requesterArgs[i];
   
            // if output/input-output mode, send back the data. 
   
            // a table parameter is serialized to a result set only if the requester has sent a 
            // result set
            if (localArg instanceof TableParameter && requesterArg instanceof TableWrapper)
            {
               // we have a DMO on the remote side, use the P2J persistence layer to get the data
               TableParameter tableParam = (TableParameter) localArg;
               localArg = tableParam.getResultSet();
            }
            else if (localArg instanceof DataSetParameter && requesterArg instanceof DatasetWrapper)
            {
               DataSetParameter dsParam = (DataSetParameter) localArg;
               localArg = dsParam.getDatasetWrapper();
            }
            else
            {
               boolean isArray = localArg.getClass().isArray();
               Object[] data = null;
               
               if (isArray)
               {
                  // check if this argument is an array
                  data = new Object[Array.getLength(localArg)];
                  
                  for (int j = 0; j < data.length; j++)
                  {
                     data[j] = Array.get(localArg, j);
                  }
               }
               else
               {
                  data = new Object[] { localArg };
               }
               
               for (int j = 0; j < data.length; j++)
               {
                  Object dlocalArg = data[j];
                  
                  if (dlocalArg instanceof handle)
                  {
                     handle h = (handle) dlocalArg;
                     
                     dlocalArg = new RemoteResourceImpl(h.isUnknown() ? null : h.toStringMessage());
                  }
                  else if (dlocalArg instanceof memptr)
                  {
                     // convert memptr instances to MemoryBuffer instances
                     dlocalArg = new MemoryBuffer(((memptr) dlocalArg).asByteArray());
                  }
                  
                  data[j] = dlocalArg;
               }
               
               localArg = (isArray ? data : data[0]);
            }
            
            arguments[i] = localArg;
         }
         
         return arguments;
      }
      finally
      {
         for (int i = 0; i < localArgs.length; i++)
         {
            Object arg = localArgs[i];
            
            if (arg instanceof memptr)
            {
               // release the memory for all memptr arguments
               ((memptr) arg).setLength(0);
            }
         }
      }
   }

   /**
    * Prepare this agent by setting some batch-processing related configuration.
    */
   private void prepare()
   {
      // CA: 20210405: there were some changes in LT initialization (I think from the TM/PM helpers being 
      // added), and now is no longer needed to initialize TM before LT - this is done during PM helper
      // call done by LT.init()
      
      // avoid the end-of-program message
      LogicalTerminal.setRelogin(true);
      
      // CA: 20210405: this comment is no longer valid.  it is left as a 'warning' for future readers in case
      // the LT/TM initialization order is changed again and STATE-RESET mode gets broken...
      // CA: 20220831: ScopeableFactory was removed and the comment bellow is no longer valid, but still left
      // for historical purposes.
      
      // make sure the TM is loaded before LT, as the LT context-local instance is dependent on TM
      // when creating the ScopeableFactory instances, LT will be loaded by TM and this instance
      // will be used as the factory. if we don't do this and load LT first, TM will end up with
      // the wrong LT instance as ScopeableFactory.
      TransactionManager.isHeadless();
      
      // set this as batch mode
      EnvironmentOps.setBatchMode(true);
      LogicalTerminal.activateBatchMode(true);

      // set the propath
      EnvironmentOps.setSearchPath(pool.getPropath());
      
      // force DatabaseTriggerManager context-local instantiation
      DatabaseTriggerManager.get();

      ProcedureManager.load();

      // Global scope for keeping buffers states (like BufferManager.allBuffers). This scope must be explicitly
      // popped when the agent context gets reset or when the agent terminates, so the dynamic DMOs and any 
      // other global finalizables must be processed. 
      TransactionManager.pushScope("appserver-agent",
                                   TransactionManager.NO_TRANSACTION,
                                   true,
                                   true,
                                   false,
                                   false);
      DatabaseManager.autoConnect();
      AppServerManager.registerAgent(this);
   }
   
   /**
    * Defines some data which is required by the agent to know about any persistent procedure ran on it.
    */
   private static class AgentRemoteProcedure
   {
      /** The instance of an external program ran as remote on this agent. */
      private final ExternalProgramWrapper extProg;

      /** The {@link #extProg} resource ID. */
      private final String extProgResourceId;

      /** The requester's connection which ran {@link #extProg}. */
      private final String requesterConnectionId;
      
      /** The procedure name as it was resolved on the agent side. */
      private final String procedureName;
      
      /**
       * Initialize this instance. 
       */
      public AgentRemoteProcedure(handle proc)
      {
         this.extProg = (ExternalProgramWrapper) proc.getResource();
         this.extProgResourceId = proc.toStringMessage();
         this.requesterConnectionId = AppServerManager.getServerConnectionId();
         this.procedureName = this.extProg.getFileName().toStringMessage();
      }
   }
   
   /**
    * Implements the reset context command.
    */
   private class ResetContextCommand
   extends AgentCommand
   {
      public ResetContextCommand()
      {
         super(null, null);
      }

      @Override
      public void execute()
      {
         // before cleaning up, cleanup all still-alive dynamic frames
         LogicalTerminal.cleanupDynamicFrames();

         // pop the global 'appserver-agent' scope, which will be pushed again in the 'prepare' method.  
         TransactionManager.popScope();
         
         // reset the context
         try
         {
            SecurityManager.getInstance().resetContext();
         }
         catch (RestrictedUseException e)
         {
            // ignore
         }
         finally
         {
            // set some state again
            prepare();
         }
      }
   }
   
   /**
    * Class used to stop a command currently being executed by this Agent, by sendint an interrupt
    * request via {@link Control#interrupt()}.
    */
   private class AgentStopper
   implements Runnable
   {
      /** Flag indicating if the stopper was explicitly terminated. */
      private volatile boolean terminated = false;
      
      @Override
      public void run()
      {
         boolean doInterrupt = false;
         
         while (!terminated && !Agent.this.terminated())
         {
            synchronized (stopper)
            {
               try
               {
                  stopper.wait();
                  doInterrupt = !terminated;
               }
               catch (InterruptedException e)
               {
                  // ignore - if the thread gets interrupted, the loop will terminate only when
                  // the agent is terminated
               }
            }
            
            if (doInterrupt)
            {
               doInterrupt = false;
               
               // received a notification, if agent is still running, stop its command
               if (!Agent.this.terminated())
               {
                  Control.interrupt();
               }
               else
               {
                  // terminated, finish the loop
                  break;
               }
            }
         }
      }

      /**
       * Terminate the stopper thread.
       */
      public void terminate()
      {
         this.terminated = true;
         
         synchronized (stopper)
         {
            stopper.notify();
         }
      }
   }
   
   /**
    * Defines a command to be sent to an agent.
    */
   private static abstract class AgentCommand
   {
      /** The ID of the request. */
      private final int requestId;
      
      /**
       * The connection ID on which this command is executed.  May be <code>null</code> if this
       * command must terminate execution always (i.e. a procedure proxy initialization, procedure
       * deletion, etc).
       */
      private final String connectionId;
      
      /**
       * The maximum allowed time for this invocation to complete.  In milliseconds.
       * Use <code>0</code> to disable the timeout.
       */
      private final long timeout;
      
      /** Any error caught during the command execution. */
      private Throwable error = null;
      
      /** Flag indicating if this command has finished or not. */
      private boolean finished = false;
      
      /** The result for this command. */
      private AppServerInvocationResult result;
      
      /** The OERequestInfo details for this command. */
      private Object[] requestInfo;
      
      /** 
       * When not null, it represents the token of a FWD context created for an authenticated and authorized 
       * web service call.  The API call will be performed in that context.
       */
      private String token = null;
      
      /**
       * Initialize this command for no explicit request (an internal request).
       * 
       * @param    result
       *           The result to be sent back to the remote side.
       * @param    id
       *           The connection ID on which this command is executed.
       */
      public AgentCommand(AppServerInvocationResult result, String id)
      {
         this(result, id, -1);
      }
      
      /**
       * Initialize this command for no explicit request (an internal request).
       * 
       * @param    requestInfo
       *           The remote {@link com.goldencode.p2j.oo.lang.OerequestInfo} details, as set by
       *           {@link com.goldencode.p2j.oo.lang.OerequestInfo#toArray()}.
       * @param    result
       *           The result to be sent back to the remote side.
       * @param    id
       *           The connection ID on which this command is executed.
       */
      public AgentCommand(Object[] requestInfo, AppServerInvocationResult result, String id)
      {
         this(result, id, -1);
         
         this.requestInfo = requestInfo;
      }
      
      /**
       * Initialize this command for an explicit request.
       *  
       * @param    result
       *           The result to be sent back to the remote side.
       * @param    id
       *           The connection ID on which this command is executed.
       * @param    requestId
       *           The ID of the request.
       */
      public AgentCommand(AppServerInvocationResult result, String id, int requestId)
      {
         this(result, id, requestId, 0);
      }
      
      /**
       * Initialize this command for an explicit request.
       *  
       * @param    result
       *           The result to be sent back to the remote side.
       * @param    id
       *           The connection ID on which this command is executed.
       * @param    requestId
       *           The ID of the request.
       * @param    timeout
       *           The maximum allowed time for this invocation to complete.  In milliseconds.
       *           Use <code>0</code> to disable the timeout.
       */
      public AgentCommand(AppServerInvocationResult result, String id, int requestId, long timeout)
      {
         this.requestId = requestId;
         this.connectionId = id;
         this.timeout = timeout;
         this.result = result;
      }
      
      /** Override this to define the command body. */
      public abstract void execute();
      
      /**
       * Perform the actual command.  This will restore the SESSION:CURRENT-REQUEST-INFO from the 
       * {@link #requestInfo}, if set.
       */
      public final void doExecute()
      {
         if (token != null)
         {
            SecurityManager sm = SecurityManager.getInstance();
            if (sm.hasWebRequestContext(token)) 
            {
               while (true)
               {
                  synchronized (activeWebServiceTokens)
                  {
                     if (activeWebServiceTokens.contains(token))
                     {
                        try
                        {
                           activeWebServiceTokens.wait();
                        }
                        catch (InterruptedException e)
                        {
                           // ignore
                        }
                     }
                     else
                     {
                        // lock this token, so no other request can be performed
                        activeWebServiceTokens.add(token);
                        
                        break;
                     }
                  }
               }
               
               try
               {
                  sm.executeInContext(token, this::doExecuteImpl);
               }
               finally
               {
                  synchronized (activeWebServiceTokens)
                  {
                     // unlock this token, other requests can be performed
                     activeWebServiceTokens.remove(token);
                     activeWebServiceTokens.notifyAll();
                  }
               }
            }
            else
            {
               throw new RuntimeException("Attempt to execute a command with invalid token: " + token);
            }
         }
         else
         {
            doExecuteImpl();
         }
      }
      
      private void doExecuteImpl()
      {
         try
         {
            if (requestInfo != null)
            {
               executeScoped(() -> 
               {
                  SessionUtils.currentRequestInfo().ref().fromArray(requestInfo);
                  return null;
               });
               
               requestInfo = null;
            }
            
            execute();
         }
         finally
         {
            if (requestInfo != null && result != null)
            {
               executeScoped(() -> 
               {
                  result.storeResponse(SessionUtils.currentResponseInfo());
                  return null;
               });
            }
         }
      }
      
      /**
       * Get the request ID of this command.
       * 
       * @return   The request ID.
       */
      public int getRequestId()
      {
         return requestId;
      }
      
      /**
       * Get the {@link #connectionId connection ID} where this command is executed.
       * 
       * @return   The connection ID.
       */
      public String getConnectionId()
      {
         return connectionId;
      }
      
      /**
       * Get the {@link #timeout} of execution for this command.
       * 
       * @return   The connection ID.
       */
      public long getTimeout()
      {
         return timeout;
      }
      
      /**
       * Check the {@link #finished} flag.
       * 
       * @return   See above.
       */
      public synchronized boolean isFinished()
      {
         return finished;
      }

      /**
       * Notify that this command has finished.
       */
      public synchronized void finished()
      {
         this.finished = true;
         this.notify();
      }
      
      /**
       * Set the error caught during this command execution.
       * 
       * @param    t
       *           The caught error.
       */
      public synchronized void setError(Throwable t)
      {
         this.error = t;
      }
      
      /**
       * Get the error caught during this command execution.
       * 
       * @return   See above.
       */
      public synchronized Throwable getError()
      {
         return this.error;
      }
   }
}
