/*
** Module   : LegacyJavaAppserver.java
** Abstract : Helper APIs to work with legacy JavaOpenClient.
**
** Copyright (c) 2019-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------------Description----------------------------------------
** 001 CA  20190717 Created initial version.
** 002 CA  20191119 Added support for before-table.
**     CA  20191125 Added support for relations.
** 003 CA  20200923 API improvements. Added extent support, other data-types.  Allow access to the hidden 
**                  temp-table fields.
**     CA  20200927 Use IdentityHashMap instead of plain map when the key is a Class.
**     CA  20201008 Added HANDLE type.  Added addRelation API.
**     OM  20201120 Added support for xmlNodeName, errorString temp-table attributes.
**     CA  20210213 Modifying the signature of APIs in this class may result in incompatibilities with other
**                  existing code, either in TRPL rules (for the proxy generation code), or in 3rd party 
**                  customer apps.  Always leave the previous signature and mark it as deprecated. 
**     CA  20210626 Fixed computeUniqueConstraintsClauses - the dataset name must be null if is not defined.
**     CA  20210630 If the dataset has a table definition and is in OUTPUT mode, send the table definition. 
**     CA  20210816 rowid values can be passed as string.
**     CA  20210831 Allow clients to always enable the change logging, as for the .NET client OE uses a 
**                  DataTable, which always track changes.
**     CA  20210910 Changes to allow support for before-table returned from the remote request via an 
**                  OUTPUT/INPUT-OUTPUT table or dataset.
**     CA  20210917 Javadoc fixes.
**     OM  20211020 Added _DATASOURCE_ROWID property.
**     CA  20210104 Added support for denormalized extent fields, for .NET client compatibility.
**                  Fixed DATETIME-TZ field problem when the value is a Date, and not a GregorianCalendar.
**     CA  20220110 Refactored the temp-table's row-meta state to use constants for the indexes in the array 
**                  used for serialization.
**     CA  20220216 Fixed issues with denormalized extent field names: they can collide with other legacy 
**                  field names so the changes ensure they are unique.
**     CA  20220615 Added 'getBeforeObjects'.
**     CA  20220707 'getOutputParameter' must 'endLogging' only if changes are tracked by default.
**     CA  20220923 Another case of 'getOutputParameter' must 'endLogging' only if changes are tracked 
**                  by default.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import java.lang.reflect.*;
import java.util.*;

import com.goldencode.p2j.*;
import com.goldencode.p2j.cfg.*;
import com.goldencode.p2j.persist.*;

import commonj.sdo.*;

/**
 * The legacy Java Open Client for appserver relies on Tuscany SDO to create an in-memory DataSet.
 * This class provides helper APIs which do not map on Tuscany SDO existing APIs, and also 
 * additional APIs to connect to the FWD appserver and perform calls.
 * <p>
 * This will be extended as more features of Java Open Client are required.
 */
public class LegacyJavaAppserver
implements LegacyJavaAppserverApi
{
   /** Mapping of an integer type to the {@link BaseDataType} type. */
   private static final Map<Integer, Class<?>> legacyTypes = new IdentityHashMap<>();
   
   /** Mapping of a {@link BaseDataType} type to its numeric type. */
   private static final Map<Class<?>, Integer> types2Legacy = new IdentityHashMap<>();

   /** Map of integer modes to their string representation in FWD. */
   private static final Map<Integer, String> paramModes = new IdentityHashMap<>();
   
   static
   {
      legacyTypes.put(TYPE_HANDLE, handle.class);
      legacyTypes.put(TYPE_INTEGER, integer.class);
      legacyTypes.put(TYPE_INT64, int64.class);
      legacyTypes.put(TYPE_DECIMAL, decimal.class);
      legacyTypes.put(TYPE_MEMPTR, MemoryBuffer.class);
      legacyTypes.put(TYPE_CLOB, clob.class); 
      legacyTypes.put(TYPE_BLOB, blob.class); 
      legacyTypes.put(TYPE_CHARACTER, character.class);
      legacyTypes.put(TYPE_LONGCHAR, longchar.class);
      legacyTypes.put(TYPE_LOGICAL, logical.class);
      legacyTypes.put(TYPE_DATE, date.class);
      legacyTypes.put(TYPE_DATETIME, datetime.class);
      legacyTypes.put(TYPE_DATETIMETZ, datetimetz.class);
      legacyTypes.put(TYPE_RAW, raw.class); 
      legacyTypes.put(TYPE_ROWID, rowid.class); 
      legacyTypes.put(TYPE_RECID, recid.class); 
      legacyTypes.put(TYPE_UNKNOWN, unknown.class); 
      legacyTypes.put(TYPE_DATASET, DataSetContainer.class);
      legacyTypes.put(TYPE_DATASET_HANDLE, DataSetContainer.class);
      legacyTypes.put(TYPE_TABLE_HANDLE, TableWrapper.class);
      
      paramModes.put(INPUT, "I");
      paramModes.put(OUTPUT, "O");
      paramModes.put(INPUT_OUTPUT, "U");
      
      for (Integer key : legacyTypes.keySet())
      {
         types2Legacy.put(legacyTypes.get(key), key);
      }
      types2Legacy.put(DataSetContainer.class, TYPE_DATASET);
   }

   /** Flag indicating if the change logging is tracked by default, for all tables created. */
   private boolean defaultTrackChanges = false;
   
   /** Flag indicating if the extent fields are normalized or denormalized (expanded). */
   private boolean normalizedExtent = true;
   
   /**
    * Implicit contstructor, initialize a Java-style legacy client (with {@link #normalizedExtent} set to 
    * <code>true</code>). 
    */
   public LegacyJavaAppserver()
   {
      this(true);
   }
   
   /**
    * Initialize this client, by specifying the extent mode: normalized (for Java-style) or denormalized
    * (for .NET-style).
    * 
    * @param    normalizedExtent
    *           The extent mode.
    */
   public LegacyJavaAppserver(boolean normalizedExtent)
   {
      this.normalizedExtent = normalizedExtent;
   }

   /**
    * Set the {@link #defaultTrackChanges} flag.
    * 
    * @param    trackChanges
    *           <code>true</code> if all tables need to {@link #beginLogging}.
    */
   public void setDefaultTrackChanges(boolean trackChanges)
   {
      this.defaultTrackChanges = trackChanges;
   }

   /**
    * Get all the records for the table on the specified index, in the graph metadata.
    * 
    * @param    graph
    *           The graph.
    * @param    idx
    *           The table index.
    *           
    * @return   The list of rows.
    */
   @Override
   public List<DataObject> getDataObjects(DataGraph graph, int idx) 
   {
      return DataSetSDOHelper.getDataObjects(graph, idx);
   }

   /**
    * Get all the records for the table with the specified name, in this graph.
    * 
    * @param    graph
    *           The graph.
    * @param    table
    *           The table name.
    *           
    * @return   The list of rows.
    */
   @Override
   public List<DataObject> getDataObjects(DataGraph graph, String table)
   {
      return DataSetSDOHelper.getDataObjects(graph, table);
   }

   /**
    * Get all the records for the specified table.
    * 
    * @param    tableMeta
    *           The table metadata.
    *           
    * @return   The list of rows.
    */
   @Override
   public List<DataObject> getDataObjects(DataObject tableMeta)
   {
      return DataSetSDOHelper.getDataObjects(tableMeta);
   }

   /**
    * Get the before-table rows for this after-table.
    * 
    * @param    tableMeta
    *           The after-table metadata.
    *           
    * @return   The list of before-table rows.
    */
   @Override
   public List<DataObject> getBeforeObjects(DataObject tableMeta)
   {
      return DataSetSDOHelper.getBeforeObjects(tableMeta);
   }

   /**
    * Get the child rows for this record.
    * 
    * @param    row
    *           The row.
    * @param    relation
    *           The child relation.
    *           
    * @return   The child rows.
    */
   @Override
   public List<DataObject> getChildRows(DataObject row, String relation)
   {
      return DataSetSDOHelper.getChildRows(row, relation);
   }

   /**
    * Get the row's property on the given index as a {@link GregorianCalendar}.
    * 
    * @param    row
    *           The row.
    * @param    idx
    *           The property index.
    *           
    * @return   The calendar instance.
    */
   @Override
   public GregorianCalendar getGregorianCalendar(DataObject row, int idx)
   {
      Date d = row.getDate(idx);
      if (d == null)
      {
         return null;
      }
      
      GregorianCalendar gc = new GregorianCalendar();
      gc.setTime(d);
      return gc;
   }

   /**
    * Set the field's property on the given index from the {@link GregorianCalendar} instance.
    * 
    * @param    row
    *           The row.
    * @param    idx
    *           The property index.
    * @param    date
    *           The property's value.
    */
   @Override
   public void setGregorianCalendar(DataObject row, int idx, GregorianCalendar date)
   {
      row.setDate(idx, Date.from(date.toZonedDateTime().toInstant()));
   }

   /**
    * Get the row's property as a {@link GregorianCalendar}.
    * 
    * @param    row
    *           The row.
    * @param    name
    *           The property name.
    *           
    * @return   The calendar instance.
    */
   @Override
   public GregorianCalendar getGregorianCalendar(DataObject row, String name)
   {
      Date d = row.getDate(name);
      if (d == null)
      {
         return null;
      }

      GregorianCalendar gc = new GregorianCalendar();
      gc.setTime(d);
      return gc;
   }

   /**
    * Create a new graph to hold a dataset.  It holds just the root <code>graph</code> node.
    * 
    * @return   A graph with no metadata.
    */
   @Override
   public DataGraph createDataGraph()
   {
      // tracking-changes are not enabled for this graph - this API is used only for creating a dataset for 
      // output mode
      return DataSetSDOHelper.createDataGraph();
   }

   /**
    * Create a new metadata object to hold the table details.
    * 
    * @param    metadata
    *           The root graph node.
    *           
    * @return   The graph metadata object.
    */
   @Override
   public DataGraph createDataGraph(Object metadata)
   {
      DataObject origMetadata = null;
      if (metadata instanceof DataObject)
      {
         DataObject md = (DataObject) metadata;
         
         if (md.getContainer() == null)
         {
            DataGraph newGraph = DataSetSDOHelper.createDataGraph((DataObject) metadata);
            
            if (defaultTrackChanges)
            {
               DataSetSDOHelper.beginLogging(newGraph);
            }

            return newGraph;
         }
         
         origMetadata = (DataObject) metadata;
      }
      else
      {
         // this must be a graph
         DataGraph orig = (DataGraph) metadata;
         origMetadata = DataSetSDOHelper.getMetaData(orig);
      }
      
      String origName = DataSetSDOHelper.getDataGraphName(origMetadata);
      
      metadata = DataSetSDOHelper.createGraphMetaData(origName);
      DataGraph newGraph = DataSetSDOHelper.createDataGraph((DataObject) metadata);
      
      // copy all tables
      int nt = DataSetSDOHelper.getNumTables(origMetadata);
      for (int i = 0; i < nt; i++)
      {
         DataObject origTable = DataSetSDOHelper.getTableMetaData(origMetadata, i);
         String  name        = origTable.getString("name");
         int     numFields   = origTable.getInt("numFields");
         boolean b4img       = origTable.getBoolean("b4img");
         int     numIndexes  = origTable.getInt("numIndexes");
         String  indexes     = origTable.getString("indexes");
         String  xmlns       = origTable.getString("xmlns");
         String  xmlPrefix   = origTable.getString("xmlPrefix");
         String  xmlNodeName = origTable.getString("xmlNodeName");
         String  errorString = origTable.getString("errorString");
         boolean normalizedExtent = origTable.getBoolean("normalizedExtent");
         
         DataObject newTable = DataSetSDOHelper.createTableMetaData(name, numFields, b4img, 
                                                                    numIndexes, indexes, 
                                                                    xmlns, xmlPrefix, xmlNodeName, 
                                                                    errorString, 
                                                                    normalizedExtent);
         List<DataObject> origFields = DataSetSDOHelper.getFieldMetaData(origTable);
         for (DataObject origf : origFields)
         {
            int      idx        = origf.getInt("idx");
            String   fname      = origf.getString("name");
            int      extent     = origf.getInt("extent");
            Class<?> cls        = (Class<?>) origf.get("type");
            int      legacyType = origf.getInt("legacytype");
            int      order      = origf.getInt("order");
            int      xmlMapping = origf.getInt("xmlMapping");
            
            DataSetSDOHelper.setFieldMetaData(
                  newTable, idx, fname, extent, cls, legacyType, order, xmlMapping);
         }
         DataSetSDOHelper.addTableMetaData((DataObject) metadata, newTable);
      }
      
      if (defaultTrackChanges)
      {
         DataSetSDOHelper.beginLogging(newGraph);
      }
      
      return newGraph;
   }

   /**
    * Create a new row for the specified table.  This row is not attached to the table - use
    * {@link #addDataObject} to add it.
    *  
    * @param    graph
    *           The graph.
    * @param    table
    *           The table name.
    *           
    * @return   A new, empty, row.
    */
   @Override
   public DataObject createDataObject(DataGraph graph, String table)
   {
      return DataSetSDOHelper.createDataObject(graph, table);
   }

   /**
    * Create a new row for the specified table.    This row is not attached to the table - use
    * {@link #addDataObject(DataObject, DataObject)} to add it.
    * 
    * @param    tableMetaData
    *           The table metadata.
    *           
    * @return   The new row.
    */
   @Override
   public DataObject createDataObject(DataObject tableMetaData)
   {
      return DataSetSDOHelper.createDataObject(tableMetaData);
   }
   
   /**
    * Add the specified row to the graph.
    * 
    * @param    graph
    *           The field's
    * @param    row
    *           The row to add to the table.
    */
   @Override
   public void addDataObject(DataGraph graph, DataObject row)
   {
      DataSetSDOHelper.addDataObject(graph, row);
   }
   
   /**
    * Add the specified row to the graph.
    * 
    * @param    tableMetaData
    *           The table metadata.
    * @param    row
    *           The row to add to the table.
    */
   @Override
   public void addDataObject(DataObject tableMetaData, DataObject row)
   {
      DataSetSDOHelper.addDataObject(tableMetaData, row);
   }

   /**
    * Get the table name for a given row or table metadata.
    * 
    * @param    value
    *           The table row or table metadata.
    *           
    * @return   The table name.
    */
   @Override
   public String getTableName(DataObject value)
   {
      return DataSetSDOHelper.getTableName(value);
   }

   /**
    * Get the data graph name for the given table metadata.
    * 
    * @param     tableMetadata
    *            The table metadata.
    *            
    * @return    The data graph name or <code>null</code> if the table is not part of a graph.
    */
   @Override
   public String getGraphName(DataObject tableMetadata)
   {
      return DataSetSDOHelper.getGraphName(tableMetadata);
   }
   
   /**
    * Get the number of fields for a given row.
    * 
    * @param    row
    *           The table row.
    *           
    * @return   The number of fields as specified in the table metadata.
    */
   @Override
   public int getFieldCount(DataObject row)
   {
      return DataSetSDOHelper.getFieldCount(row);
   }

   /**
    * Get the name of the field on the given index in the metadata.
    *  
    * @param    row
    *           The table row.  Must be attached to a table.
    * @param    i
    *           The field index.
    *           
    * @return   The field's name.
    */
   @Override
   public String getFieldName(DataObject row, int i)
   {
      return DataSetSDOHelper.getFieldName(row, i);
   }

   /**
    * Get the type of the field on the given index in the metadata.
    *  
    * @param    row
    *           The table row.  Must be attached to a table.
    * @param    i
    *           The field index.
    *           
    * @return   The field's type.
    */
   @Override
   public int getFieldType(DataObject row, int i)
   {
      return DataSetSDOHelper.getFieldType(row, i);
   }

   /**
    * Check if this table has a field defined with that name.
    * 
    * @param    tableMetadata
    *           The table metadata.
    * @param    fieldName
    *           The field name.
    *           
    * @return   <code>true</code> if there is a field defined with this name.
    */
   @Override
   public boolean isTableField(DataObject tableMetadata, String fieldName)
   {
      return DataSetSDOHelper.isTableField(tableMetadata, fieldName);
   }
   
   /**
    * Get the field type.
    * 
    * @param    tableMetadata
    *           The table metadata.
    * @param    fieldName
    *           The field name.
    *           
    * @return   The field type.
    */
   @Override
   public int getTableFieldType(DataObject tableMetadata, String fieldName)
   {
      return DataSetSDOHelper.getTableFieldType(tableMetadata, fieldName);
   }

   /**
    * Get all table names from the given graph.
    * 
    * @param    graph
    *           The graph.
    *           
    * @return   See above.
    */
   @Override
   public String[] getTableNames(DataGraph graph)
   {
      return DataSetSDOHelper.getTableNames(graph);
   }

   /**
    * Get the number of tables for a graph.
    * 
    * @param    graph
    *           The graph.
    *           
    * @return   See above.
    */
   @Override
   public int getNumTables(DataGraph graph)
   {
      DataObject metadata = getMetaData(graph);
      return metadata == null ? 0 : DataSetSDOHelper.getNumTables(metadata);
   }

   /**
    * Get the number of tables for a graph metadata.
    * 
    * @param    metaData
    *           The graph metadata.
    *           
    * @return   See above.
    */
   @Override
   public int getNumTables(DataObject metaData)
   {
      return DataSetSDOHelper.getNumTables(metaData);
   }

   /**
    * Get the metadata object for this graph, from the <code>metadata</code> property of the root
    * object.
    * 
    * @param    graph
    *           The graph.
    *           
    * @return   The metadata object.
    */
   @Override
   public DataObject getMetaData(DataGraph graph)
   {
      return DataSetSDOHelper.getMetaData(graph);
   }

   /**
    * Get the metadata for the table registered on the given index in the graph.
    * 
    * @param    metaData
    *           The graph metadata.
    * @param    i
    *           The table index.
    *           
    * @return   The table metadata.
    */
   @Override
   public DataObject getTableMetaData(DataObject metaData, int i)
   {
      return DataSetSDOHelper.getTableMetaData(metaData, i);
   }

   /**
    * Get the metadata for the specified table.
    * 
    * @param    graph
    *           The graph.
    * @param    table
    *           The table name.
    *           
    * @return   The table metadata or <code>null</code> if the table is not part of this graph.
    */
   @Override
   public DataObject getTableMetaData(DataGraph graph, String table)
   {
      return DataSetSDOHelper.getTableMetaData(graph, table);
   }
   
   /**
    * Get the metadata list for all tables.
    * 
    * @param    graph
    *           The graph.
    *           
    * @return   The metadata for all tables.
    */
   @Override
   public List<DataObject> getTableMetaData(DataGraph graph)
   {
      return DataSetSDOHelper.getTableMetaData(graph);
   }

   /**
    * Create an object with a table metadata.  This object will have the {@code tablemetadata} type and is not
    * attached to the graph.
    * 
    * @param   name
    *          The table's name.
    * @param   numFields
    *          The number of fields.
    * @param   b4img
    *          Flag indicating if the table has a BEFORE-IMAGE.
    * @param   numIndexes
    *          The number of indexes.
    * @param   indexes
    *          The index specification (TODO: not validated/implemented at this time).
    * @param   xmlns
    *          The XML namespace (TODO: not validated/implemented at this time).
    * @param   xmlPrefix
    *          The XML prefix (TODO: not validated/implemented at this time).
    *
    * @return   An object with the table metadata.
    */
   @Override
   @Deprecated
   public DataObject createObjectMetaData(String  name,
                                          int     numFields,
                                          boolean b4img,
                                          int     numIndexes,
                                          String  indexes,
                                          String  xmlns,
                                          String  xmlPrefix)
   {
      return createObjectMetaData(name, numFields, b4img, numIndexes, indexes, xmlns, xmlPrefix, null, null);
   }
   
   /**
    * Create an object with a table metadata.  This object will have the {@code tablemetadata} type and is not
    * attached to the graph.
    * 
    * @param   name
    *          The table's name.
    * @param   numFields
    *          The number of fields.
    * @param   b4img
    *          Flag indicating if the table has a BEFORE-IMAGE.
    * @param   numIndexes
    *          The number of indexes.
    * @param   indexes
    *          The index specification (TODO: not validated/implemented at this time).
    * @param   xmlns
    *          The XML namespace (TODO: not validated/implemented at this time).
    * @param   xmlPrefix
    *          The XML prefix (TODO: not validated/implemented at this time).
    * @param   xmlNodeName
    *          The XML node name (TODO: not validated/implemented at this time).
    * @param   errorString
    *          The table's ERROR-STRING attribute  (TODO: not validated/implemented at this time).
    *
    * @return   An object with the table metadata.
    */
   @Override
   public DataObject createObjectMetaData(String  name,
                                          int     numFields,
                                          boolean b4img,
                                          int     numIndexes,
                                          String  indexes,
                                          String  xmlns,
                                          String  xmlPrefix,
                                          String  xmlNodeName,
                                          String  errorString)
   {
      return DataSetSDOHelper.createTableMetaData(name, numFields, b4img, 
                                                  numIndexes, indexes, 
                                                  xmlns, xmlPrefix, xmlNodeName, 
                                                  errorString, 
                                                  normalizedExtent);
   }

   /**
    * Add a new new field with the specified configuration.
    * 
    * @param    metaData
    *           The table metadata, of type <code>tablemetadata</code>.  The field will be added
    *           to the <code>fields</code> property.
    * @param    idx
    *           The field's index in the metadata.
    * @param    name
    *           The field's name.
    * @param    extent
    *           The field's extent.
    * @param    type
    *           The legacy type (numeric value based on {@link LegacyJavaAppserver} constants.
    * @param    order
    *           The field's order.
    * @param    xmlMapping
    *           The field's mapping.
    */
   @Override
   public void setFieldMetaData(DataObject metaData, 
                                int        idx, 
                                String     name, 
                                int        extent, 
                                int        type,
                                int        order, 
                                int        xmlMapping)
   {
      Class<?> cls = legacyTypes.get(type);
      DataSetSDOHelper.setFieldMetaData(metaData, idx, name, extent, cls, type, order, xmlMapping);
   }

   /**
    * Create a metadata object with the specified name.
    * 
    * @param    name
    *           The graph name.
    *           
    * @return   The metadata object.  Will be attached to a graph via {@link #createDataGraph(Object)}.
    */
   @Override
   public DataObject createGraphMetaData(String name)
   {
      return DataSetSDOHelper.createGraphMetaData(name);
   }

   /**
    * Add the specified table metadata to the graph.  This will create a custom type based on the
    * configured fields, so that each row will be created with this type.
    * 
    * @param    metaData
    *           The graph metadata.
    * @param    table
    *           The table metadata.
    */
   @Override
   public void addTableMetaData(DataObject metaData, DataObject table)
   {
      DataSetSDOHelper.addTableMetaData(metaData, table);
   }

   /**
    * Get the graph name from the specified metadata, of type <code>graphmetadata</code>.
    * 
    * @param    metadata
    *           The metadata.
    *           
    * @return   The graph name.
    */
   @Override
   public String getDataGraphName(DataObject metadata)
   {
      return DataSetSDOHelper.getDataGraphName(metadata);
   }

   /**
    * Configure a {@link LegacyJavaAppserverParameter} with the given details.
    * 
    * @param    mode
    *           The parameter mode.
    * @param    type
    *           The parameter type.
    * @param    value
    *           The parameter value.
    * @param    metadata
    *           The parameter metadata (in case of DataSet).
    *           
    * @return   A new parameter instance.
    */
   @Override
   public LegacyJavaAppserverParameter getJavaParameter(int mode, int type, Object value, Object metadata) 
   {
      Object orig = value;

      boolean input = mode == INPUT || mode == INPUT_OUTPUT;
      boolean output = mode == OUTPUT || mode == INPUT_OUTPUT;
      
      if (type == TYPE_DATASET || type == TYPE_DATASET_HANDLE)
      {
         List<DsTableDefinition> tableDefinitions = new ArrayList<>();

         DataObject graphMeta = (DataObject) metadata;
         String[] tables = graphMeta == null ? new String[0] 
                                             : DataSetSDOHelper.getTableNames(graphMeta);
         for (int i = 0; i < tables.length; i++)
         {
            DataObject tableMeta = DataSetSDOHelper.getTableMetaData(graphMeta, i);
            
            List<DataObject> rows = input ? DataSetSDOHelper.getDataObjects(tableMeta) 
                                          : Collections.emptyList();
            List<DataObject> fields = DataSetSDOHelper.getFieldMetaData(tableMeta);
            DsTableDefinition dstable = createDsTable(rows, fields, input, output);
            dstable.setName(tables[i]);
            
            dstable.setIndexes(tableMeta.getInt("numIndexes"), tableMeta.getString("indexes"));
            dstable.setXmlns(tableMeta.getString("xmlns"));
            dstable.setXmlPrefix(tableMeta.getString("xmlPrefix"));
            dstable.setXmlNodeName(tableMeta.getString("xmlNodeName"));
            dstable.setErrorString(tableMeta.getString("errorString"));
            tableDefinitions.add(dstable);
            
            if (tableMeta.getBoolean("b4img"))
            {
               List<DataObject> beforeRows = input ? DataSetSDOHelper.getBeforeObjects(tableMeta)
                                                   : Collections.emptyList();
               DsTableDefinition b4table = createDsTable(beforeRows, fields, input, output);
               // TODO: here we must not exceed 32 characters...
               b4table.setName("BI" + tables[i]);
               
               dstable.setPeerDef(b4table);
            }
         }
         
         DataSetContainer dsc = new DataSetContainer(input, output, false);
         String name = graphMeta == null ? null : DataSetSDOHelper.getDataGraphName(graphMeta);
         dsc.setStructureName(name);
         dsc.setTableDefinitions(tableDefinitions);
         dsc.setRelationDefinitions(null);
         value = dsc;
      }
      else if (type == TYPE_MEMPTR)
      {
         // this must always be MemoryBuffer
         if (value != null && !(value instanceof MemoryBuffer))
         {
            throw new IllegalArgumentException("memptr must be MemoryBuffer!");
         }
      }
      else if (type == TYPE_TABLE_HANDLE)
      {
         DataGraph graph = (DataGraph) value;
         DataObject graphMeta = DataSetSDOHelper.getMetaData(graph);
         
         String[] tables = graphMeta == null ? null 
                                             : DataSetSDOHelper.getTableNames(graphMeta);
         if (tables != null)
         {
            DataObject tableMeta = DataSetSDOHelper.getTableMetaData(graphMeta, 0);
            
            List<DataObject> rows = input ? DataSetSDOHelper.getDataObjects(tableMeta) 
                                          : Collections.emptyList();
            List<DataObject> fields = DataSetSDOHelper.getFieldMetaData(tableMeta);
            Iterator<DataObject> fieldIter = fields.iterator();
            Iterator<DataObject> rowIter = rows.iterator();

            TableResultSet ttrs = new TableResultSet(input, output, false)
            {
               DataObject nextRow;

               @Override
               public Object[] nextRow()
               {
                  nextRow = rowIter.next();
                  
                  int numFields = fields.size();
                  Object[] data = new Object[numFields];
                  for (int i = 0; i < data.length; i++)
                  {
                     int ftype = DataSetSDOHelper.getFieldType(fields.get(i));
                     int extent = DataSetSDOHelper.getFieldExtent(fields.get(i));
                     Object fval = nextRow.get(i);

                     data[i] = fromJavaType(fval, ftype, extent);
                  }
                  
                  return data;
               }
               
               @Override
               public Object[] nextRowMeta()
               {
                  FWDDataObject row = (FWDDataObject) nextRow;
                  
                  return row.toArray();
               }

               @Override
               public PropertyDefinition nextProperty()
               {
                  DataObject fieldMeta = fieldIter.next();
                  String name = DataSetSDOHelper.getFieldName(fieldMeta);
                  Class<?> type = legacyTypes.get(DataSetSDOHelper.getFieldType(fieldMeta));
                  int extent = DataSetSDOHelper.getFieldExtent(fieldMeta);
                  
                  return new PropertyDefinition(name, type, extent, name);
               }
               
               @Override
               public boolean hasMoreRows()
               {
                  return rowIter.hasNext();
               }
               
               @Override
               public boolean hasMoreProperties()
               {
                  return fieldIter.hasNext();
               }
            };
            ttrs.setStructureName(tables[0]);
            
            TableWrapper wrapper = new TableWrapper(ttrs);
            wrapper.setTableName(tables[0]);
            wrapper.setNumIndexes(tableMeta.getInt("numIndexes"));
            wrapper.setIndexes(tableMeta.getString("indexes"));
            wrapper.setXmlns(tableMeta.getString("xmlns"));
            wrapper.setXmlPrefix(tableMeta.getString("xmlPrefix"));
            wrapper.setXmlNodeName(tableMeta.getString("xmlNodeName"));
            wrapper.setErrorString(tableMeta.getString("errorString"));
            
            if (tableMeta.getBoolean("b4img"))
            {
               throw new RuntimeException("Using a TABLE-HANDLE with b4img, not yet supported in FWD!");
            }
            
            value = wrapper;
         }
         else
         {
            value = new TableWrapper();
         }
      }
      else
      {
         value = fromJavaType(value, type, SourceNameMapper.NO_EXTENT);
      }
      
      return new LegacyJavaAppserverParameter(type, mode, value, orig);
   }

   /**
    * Determine the output parameter value.
    * 
    * @param    param
    *           The parameter as received from the remote side.
    *           
    * @return   The parameter's output value.
    */
   @Override
   public Object getOutputParameter(LegacyJavaAppserverParameter param)
   {
      int type = param.getType();
      Object arg = param.getValue();
      
      switch (type)
      {
         case TYPE_DATASET:
         case TYPE_DATASET_HANDLE:
            DataGraph graph = (DataGraph) param.getOrig();
            DataSetContainer dsc = (DataSetContainer) param.getValue();
            
            DataObject metadata = DataSetSDOHelper.getMetaData(graph);
            boolean reuseMetadata = (metadata != null);
            if (!reuseMetadata)
            {
               //create metadata
               metadata = DataSetSDOHelper.createGraphMetaData(dsc.getStructureName());
               DataSetSDOHelper.setGraphMetaData(graph, metadata);
            }
            
            for (DsRelationDefinition dsr : dsc.getRelationDefinitions())
            {
               DataSetSDOHelper.addRelation(metadata, 
                                            dsr.getName(), 
                                            dsr.getParentBuffer(),
                                            dsr.getChildBuffer(),
                                            dsr.getWhereString(),
                                            dsr.isActive(),
                                            dsr.isParentId(),
                                            dsr.getFieldPairs(), 
                                            dsr.isFkeyHidden(),
                                            dsr.isRecursive(),
                                            dsr.isNested(),
                                            dsr.isReposition());
            }
            
            // end the logging
            if (defaultTrackChanges)
            {
               DataSetSDOHelper.endLogging(graph);
            }

            for (DsTableDefinition dst : dsc.getTableDefinitions())
            {
               createOutputTable(graph, dst, reuseMetadata, false);
               DsTableDefinition b4dst = dst.getPeerDef();
               if (b4dst != null)
               {
                  createOutputTable(graph, b4dst, reuseMetadata, true);
               }
            }
            
            if (defaultTrackChanges)
            {
               graph.getChangeSummary().beginLogging();
            }
            
            return graph;
            
         case TYPE_TABLE_HANDLE:
            graph = (DataGraph) param.getOrig();
            TableWrapper wrapper = (TableWrapper) param.getValue();
            
            metadata = DataSetSDOHelper.getMetaData(graph);
            reuseMetadata = (metadata != null);
            if (!reuseMetadata)
            {
               //create metadata
               metadata = DataSetSDOHelper.createGraphMetaData(wrapper.getTableName());
               DataSetSDOHelper.setGraphMetaData(graph, metadata);
            }
            
            // end the logging
            if (defaultTrackChanges)
            {
               DataSetSDOHelper.endLogging(graph);
            }
            
            createOutputTable(graph, wrapper, reuseMetadata, false);
            TableWrapper b4wrapper = wrapper.getPeerDef();
            if (b4wrapper != null)
            {
               createOutputTable(graph, b4wrapper, reuseMetadata, true);
            }
            
            if (defaultTrackChanges)
            {
               graph.getChangeSummary().beginLogging();
            }
            
            return graph;
      }
      
      return toJavaValue(arg, type, false, SourceNameMapper.NO_EXTENT);
   }
   
   /**
    * Determine the output parameter value on the given index.
    * 
    * @param    paramArray
    *           The parameter array.
    * @param    i
    *           The parameter's index.
    *           
    * @return   The parameter's output value.
    */
   @Override
   public Object getOutputParameter(LegacyJavaAppserverParameter[] paramArray, int i)
   {
      return getOutputParameter(paramArray[i]);
   }

   /**
    * Get a string representation of the argument's modes.
    * 
    * @param    paramArray
    *           The parameter array.
    *           
    * @return   The parameter modes.
    */
   @Override
   public String getArgumentModes(LegacyJavaAppserverParameter[] paramArray)
   {
      StringBuilder modes = new StringBuilder();
      for (LegacyJavaAppserverParameter ljcp : paramArray)
      {
         modes.append(paramModes.get(ljcp.getMode()));
      }
      
      return modes.toString();
   }

   /**
    * Resolve the arguments from the given parameter array.
    * 
    * @param    paramArray
    *           The parameter array.
    *           
    * @return   The arguments.
    */
   @Override
   public Object[] getArguments(LegacyJavaAppserverParameter[] paramArray)
   {
      Object[] params = new Object[paramArray.length];
      int idx = 0;
      for (LegacyJavaAppserverParameter ljcp : paramArray)
      {
         params[idx++] = ljcp.getValue();
      }
      
      return params;
   }

   /**
    * Create a memptr corresponding instance.
    * 
    * @param    value
    *           The bytes.
    *           
    * @return   A {@link MemoryBuffer}.
    */
   @Override
   public Object createMemptr(byte[] value)
   {
      return new MemoryBuffer(value);
   }

   /**
    * Check if the given parameter is a memptr.
    * 
    * @param    outputParameter
    *           The parameter.
    *           
    * @return   <code>true</code> if the value is a {@link MemoryBuffer}.
    */
   @Override
   public boolean isMemptr(Object outputParameter)
   {
      return outputParameter instanceof MemoryBuffer;
   }

   /**
    * Get the bytes from the given output parameter, which is a {@link MemoryBuffer}.
    * 
    * @param    outputParameter
    *           The output parameter.
    *           
    * @return   See above.
    */
   @Override
   public byte[] getBytes(Object outputParameter)
   {
      return ((MemoryBuffer) outputParameter).getValue();
   }

   /**
    * Create a new FWd connection with the specified configuration.
    * 
    * @param    config
    *           The configuration for the FWD server connection.
    * @param    account
    *           The FWD account to authenticate on the remote side.
    * @param    appServerName
    *           The name of the appserver to which the connection will be established.
    * @param    sessionFree
    *           <code>true</code> to indicate a session-free operating mode, <code>false</code>
    *           to indicate a session-managed operating mode.
    * @param    user
    *           The username passed to the appserver's connect procedure.
    * @param    pwd
    *           The password passed to the appserver's connect procedure.
    * @param    serverInfo
    *           The info passed to the appserver's connect procedure.
    *
    * @return   The helper with the connection result (which was established or not).
    * 
    * @throws   NumberedException
    *           If there are connection problems.
    */
   @Override
   public AppServerHelper getConnection(BootstrapConfig   config,
                                        String            account,
                                        String            appServerName,
                                        boolean           sessionFree,
                                        String            user,
                                        String            pwd,
                                        String            serverInfo)
   throws NumberedException
   {
      return AppServerHelper.connect(config, account, appServerName, sessionFree, user, pwd, serverInfo);
   }

   /**
    * Postprocess these arguments - this includes copying state back to the 
    * {@link LegacyJavaAppserverParameter} instance, for OUTPUT parameters.
    * 
    * @param    paramArray
    *           The parameter array. 
    * @param    args
    *           The returned arguments.
    */
   public void postProcessArguments(Object paramArray, Object[] args)
   {
      Object[] params = (Object[]) paramArray;
      for (int i = 0; i < params.length; i++)
      {
         LegacyJavaAppserverParameter ljcp = (LegacyJavaAppserverParameter) params[i];
         ljcp.setValue(args[i]);
      }
   }

   /**
    * Get the changes in this change summary, as a graph.
    * 
    * @param    changeSummary
    *           The change summary.
    *           
    * @return   See above.
    */
   @Override
   public DataGraph getChanges(Object changeSummary)
   {
      return ((ChangeSummary) changeSummary).getDataGraph();
   }

   /**
    * End the logging for changes in this object.
    * 
    * @param    changeSummary
    *           The change summary.
    */
   @Override
   public void endLogging(Object changeSummary)
   {
      DataSetSDOHelper.endLogging(((ChangeSummary) changeSummary).getDataGraph());
   }

   /**
    * Begin the logging for changes in this object.
    * 
    * @param    changeSummary
    *           The change summary.
    */
   @Override
   public void beginLogging(Object changeSummary)
   {
      DataSetSDOHelper.beginLogging(((ChangeSummary) changeSummary).getDataGraph());
   }

   /**
    * Get the change summary object for this graph.
    * 
    * @param    graph
    *           The graph.
    *           
    * @return   See above.
    */
   @Override
   public Object getChangeSummary(DataGraph graph)
   {
      return graph.getChangeSummary();
   }

   /**
    * Check if the given object has been deleted.
    * 
    * @param    changeSummary
    *           The change summary.
    * @param    object
    *           The object.
    *           
    * @return   See above.
    */
   @Override
   public boolean isDeleted(Object changeSummary, DataObject object)
   {
      return ((ChangeSummary) changeSummary).isDeleted(object);
   }

   /**
    * Check if the given object is newly created.
    * 
    * @param    changeSummary
    *           The change summary.
    * @param    object
    *           The object.
    *           
    * @return   See above.
    */
   @Override
   public boolean isCreated(Object changeSummary, DataObject object)
   {
      return ((ChangeSummary) changeSummary).isCreated(object);
   }

   /**
    * Get all the changed objects in this change summary.
    * 
    * @param    changeSummary
    *           The change summary.
    *           
    * @return   See above.
    */
   @Override
   public List<DataObject> getChangedDataObjects(Object changeSummary)
   {
      return ((ChangeSummary) changeSummary).getChangedDataObjects();
   }
   
   /**
    * Retrieve all rows which have an exact match for the specified fields.
    * 
    * @param    tableMetadata
    *           The table metadata.
    * @param    fields
    *           The field list.
    * @param    values
    *           The values to be matched.
    *           
    * @return   See above.
    */
   @Override
   public List<DataObject> selectRows(DataObject tableMetadata, String[] fields, Object[] values)
   {
      return DataSetSDOHelper.selectRows(tableMetadata, fields, values);
   }
   
   /**
    * Add a relation to the graph's metadata.
    * 
    * @param    metaData
    *           The graph metadata.
    * @param    name
    *           The relation name.
    * @param    parentBuffer
    *           The parent buffer.
    * @param    childBuffer
    *           The child buffer.
    * @param    whereString
    *           The where string.
    * @param    active
    *           Flag indicating if this relation is active.
    * @param    parentId
    *           Flag indicating if parent-id is used.
    * @param    pairs
    *           The field pairs, in <code>parent, child</code> format.
    * @param    fkeyHidden
    *           Flag indicating if fkey-hidden is used.
    * @param    recursive
    *           Flag indicating if recursive is used.
    * @param    nested
    *           Flag indicating if nested is used.
    * @param    reposition
    *           Flag indicating if reposition is used.
    */
   @Override
   public void addRelation(DataObject metaData, 
                           String name,
                           String parentBuffer,
                           String childBuffer,
                           String whereString,
                           boolean active,
                           boolean parentId,
                           String pairs,
                           boolean fkeyHidden,
                           boolean recursive,
                           boolean nested,
                           boolean reposition)
   {
      DataSetSDOHelper.addRelation(metaData, 
                                   name, 
                                   parentBuffer, 
                                   childBuffer, 
                                   whereString, 
                                   active, 
                                   parentId, 
                                   pairs, 
                                   fkeyHidden, 
                                   recursive, 
                                   nested, 
                                   reposition);
   }

   /**
    * Create and populate a table from the results received from the request.
    * <p>
    * If {@link #defaultTrackChanges} is set, then logging is enabled for all rows.
    * 
    * @param    graph
    *           The graph where to add the table.
    * @param    wrapper
    *           The table data.
    * @param    reuseMetadata
    *           Flag indicating if the table already exists in the graph.
    * @param    beforeTable
    *           Flag indicating if this is a before table.
    */
   private void createOutputTable(DataGraph    graph, 
                                  TableWrapper wrapper, 
                                  boolean      reuseMetadata, 
                                  boolean      beforeTable)
   {
      DataObject metadata = DataSetSDOHelper.getMetaData(graph);
      List<PropertyDefinition> props = wrapper.getProperties();
      String tableName = beforeTable ? wrapper.getPeerDef().getTableName() : wrapper.getTableName();
      
      if (!beforeTable && !reuseMetadata)
      {
         DataObject tableMeta = DataSetSDOHelper.createTableMetaData(tableName,
                                                                     props.size(),
                                                                     wrapper.getPeerDef() != null,
                                                                     wrapper.getNumIndexes(),
                                                                     wrapper.getIndexes(),
                                                                     wrapper.getXmlns(),
                                                                     wrapper.getXmlPrefix(),
                                                                     wrapper.getXmlNodeName(),
                                                                     wrapper.getErrorString(),
                                                                     normalizedExtent);
         
         int idx = 0;
         for (PropertyDefinition prop : props)
         {
            DataSetSDOHelper.setFieldMetaData(tableMeta, 
                                              idx, 
                                              prop.getLegacyName(), 
                                              prop.getExtent(), 
                                              prop.getType(), 
                                              types2Legacy.get(prop.getType()), 
                                              idx, 
                                              0);
            idx++;
         }
         // after we have the fields set
         DataSetSDOHelper.addTableMetaData(metadata, tableMeta);
      }
      else
      {
         DataObject table = getTableMetaData(graph, tableName);
         List<DataObject> rows = beforeTable ? DataSetSDOHelper.getBeforeObjects(table)
                                             : DataSetSDOHelper.getDataObjects(table);

         // clear the rows first
         for (DataObject row : rows)
         {
            row.delete();
         }
      }

      List<Object[]> rowsMeta = wrapper.getRowsMeta();
      List<Object[]> rows = wrapper.getRows();

      // set rows
      for (int recnum = 0; recnum < rows.size(); recnum++)
      {
         Object[] row = rows.get(recnum);
         Object[] rowMeta = rowsMeta.get(recnum);

         DataObject drow = DataSetSDOHelper.createDataObject(graph, tableName);
         if (!beforeTable && defaultTrackChanges)
         {
            ((FWDDataObject) drow).beginLogging();
         }

         try
         {
            ((FWDDataObject) drow).setForceNormalizedFields(true);
            // always use normalized fields here
            for (int idx = 0; idx < row.length; idx++)
            {
               Object rval = row[idx];
               PropertyDefinition prop = props.get(idx);
   
               Class<?> cls = prop.getType();
               Integer ltype = types2Legacy.get(cls);
               if (ltype == null)
               {
                  throw new NullPointerException(
                        "Unknown field type " + cls + " for " + prop.getLegacyName() + " in " + tableName);
               }
               
               Object val = toJavaValue(rval, ltype, true, prop.getExtent());
               
               drow.set(idx, val);
            }
         }
         finally
         {
            ((FWDDataObject) drow).setForceNormalizedFields(false);
         }
         
         ((FWDDataObject) drow).primaryKey((Long) rowMeta[TempRecord._PK_DATA_INDEX]);
         ((FWDDataObject) drow)._rowState((Integer) rowMeta[TempRecord._ROW_STATE_DATA_INDEX]);
         ((FWDDataObject) drow)._peerRowid((Long) rowMeta[TempRecord._PEER_ROWID_DATA_INDEX]);
         ((FWDDataObject) drow)._originRowid((Long) rowMeta[TempRecord._ORIGIN_ROWID_DATA_INDEX]);
         ((FWDDataObject) drow)._datasourceRowid((Long) rowMeta[TempRecord._DATASOURCE_ROWID_INDEX]);
         ((FWDDataObject) drow)._errorFlags((Integer) rowMeta[TempRecord._ERROR_FLAG_DATA_INDEX]);
         ((FWDDataObject) drow)._errorString((String) rowMeta[TempRecord._ERROR_STRING_DATA_INDEX]);
         
         if (beforeTable)
         {
            DataSetSDOHelper.addBeforeObject(graph, drow);
         }
         else
         {
            DataSetSDOHelper.addDataObject(graph, drow);
         }
      }
   }
   
   /**
    * Create and populate a table from the results received from the request.
    * <p>
    * If {@link #defaultTrackChanges} is set, then logging is enabled for all rows.
    * 
    * @param    graph
    *           The graph where to add the table.
    * @param    dst
    *           The table data.
    * @param    reuseMetadata
    *           Flag indicating if the table already exists in the graph.
    * @param    beforeTable
    *           Flag indicating if this is a before table.
    */
   private void createOutputTable(DataGraph         graph, 
                                  DsTableDefinition dst, 
                                  boolean           reuseMetadata, 
                                  boolean           beforeTable)
   {
      DataObject metadata = DataSetSDOHelper.getMetaData(graph);
      List<PropertyDefinition> props = dst.getProperties();
      String tableName = beforeTable ? dst.getPeerDef().getName() : dst.getName();
      
      if (!beforeTable && !reuseMetadata)
      {
         DataObject tableMeta = DataSetSDOHelper.createTableMetaData(tableName,
                                                                     props.size(),
                                                                     dst.getPeerDef() != null,
                                                                     dst.getNumIndexes(),
                                                                     dst.getIndexes(),
                                                                     dst.getXmlns(),
                                                                     dst.getXmlPrefix(),
                                                                     dst.getXmlNodeName(),
                                                                     dst.getErrorString(),
                                                                     normalizedExtent);
         
         int idx = 0;
         for (PropertyDefinition prop : props)
         {
            DataSetSDOHelper.setFieldMetaData(tableMeta, 
                                              idx, 
                                              prop.getLegacyName(), 
                                              prop.getExtent(), 
                                              prop.getType(), 
                                              types2Legacy.get(prop.getType()), 
                                              idx, 
                                              0);
            idx++;
         }
         // after we have the fields set
         DataSetSDOHelper.addTableMetaData(metadata, tableMeta);
      }
      else
      {
         DataObject table = getTableMetaData(graph, tableName);
         List<DataObject> rows = beforeTable ? DataSetSDOHelper.getBeforeObjects(table)
                                             : DataSetSDOHelper.getDataObjects(table);
         // clear the rows first
         for (DataObject row : rows)
         {
            row.delete();
         }
      }
      
      List<Object[]> rowsMeta = dst.getRowsMeta();
      List<Object[]> rows = dst.getRows();
      // set rows
      for (int recnum = 0; recnum < rows.size(); recnum++)
      {
         Object[] row = rows.get(recnum);
         Object[] rowMeta = rowsMeta.get(recnum);
         
         DataObject drow = DataSetSDOHelper.createDataObject(graph, tableName);
         if (!beforeTable && defaultTrackChanges)
         {
            ((FWDDataObject) drow).beginLogging();
         }

         try
         {
            // always use normalized fields here
            ((FWDDataObject) drow).setForceNormalizedFields(true);
            for (int idx = 0; idx < row.length; idx++)
            {
               Object rval = row[idx];
               PropertyDefinition prop = props.get(idx);
   
               Class<?> cls = prop.getType();
               Integer ltype = types2Legacy.get(cls);
               if (ltype == null)
               {
                  throw new NullPointerException(
                        "Unknown field type " + cls + " for " + prop.getLegacyName() + " in " + tableName);
               }
               
               Object val = toJavaValue(rval, ltype, true, prop.getExtent());
               drow.set(idx, val);
            }
         }
         finally
         {
            ((FWDDataObject) drow).setForceNormalizedFields(false);
         }
         
         ((FWDDataObject) drow).primaryKey((Long) rowMeta[TempRecord._PK_DATA_INDEX]);
         ((FWDDataObject) drow)._rowState((Integer) rowMeta[TempRecord._ROW_STATE_DATA_INDEX]);
         ((FWDDataObject) drow)._peerRowid((Long) rowMeta[TempRecord._PEER_ROWID_DATA_INDEX]);
         ((FWDDataObject) drow)._originRowid((Long) rowMeta[TempRecord._ORIGIN_ROWID_DATA_INDEX]);
         ((FWDDataObject) drow)._datasourceRowid((Long) rowMeta[TempRecord._DATASOURCE_ROWID_INDEX]);
         ((FWDDataObject) drow)._errorFlags((Integer) rowMeta[TempRecord._ERROR_FLAG_DATA_INDEX]);
         ((FWDDataObject) drow)._errorString((String) rowMeta[TempRecord._ERROR_STRING_DATA_INDEX]);
         
         if (beforeTable)
         {
            DataSetSDOHelper.addBeforeObject(graph, drow);
         }
         else
         {
            DataSetSDOHelper.addDataObject(graph, drow);
         }
      }
   }
   
   /**
    * Convert the given value to a {@link BaseDataType} compatible instance.
    * 
    * @param    value
    *           The Java value.
    * @param    type
    *           The value's type.
    * @param    extent
    *           The extent value.
    *           
    * @return   See above.
    */
   private Object fromJavaType(Object value, int type, int extent)
   {
      if (value == null)
      {
         Class<?> cls = legacyTypes.get(type);
         return BaseDataType.generateUnknown(cls);
      }
      
      if (extent != 0 && extent != SourceNameMapper.NO_EXTENT)
      {
         Class<?> elType = legacyTypes.get(type);
         int length = ((List) value).size();
         Object res = Array.newInstance(elType, length);
         
         for (int i = 0; i < length; i++)
         {
            Array.set(res, i, fromJavaType(((List) value).get(i), type, SourceNameMapper.NO_EXTENT));
         }
         
         return res;
         
      }
      switch (type)
      {
         case TYPE_LOGICAL:
            return new logical((Boolean) value);
            
         case TYPE_CHARACTER:
            return new character((String) value);
            
         case TYPE_HANDLE:
            return handle.fromResourceId((Long) value);

         case TYPE_LONGCHAR:
            return new longchar((String) value);
            
         case TYPE_INTEGER:
            return new integer((Number) value);
         
         case TYPE_INT64:
            return new int64((Number) value);
            
         case TYPE_DECIMAL:
            return new decimal((Number) value);
            
         case TYPE_BLOB:
            if (value.getClass().isArray() && value.getClass().getComponentType() == byte.class)
            {
               return new blob(((byte[]) value));
            }
            else if (value instanceof MemoryBuffer)
            {
               return new blob(((MemoryBuffer) value).getValue());
            }
            
            throw new IllegalStateException("Can not determine the blob value from " + value);
            
         case TYPE_CLOB:
            return new clob((String) value);
            
         case TYPE_RAW:
            return new raw((byte[]) value);
            
         case TYPE_RECID:
            return new recid((long) value);
            
         case TYPE_ROWID:
            if (value instanceof String)
            {
               String s = (String) value;
               s = s.trim();
               if (s.isEmpty())
               {
                  return new rowid();
               }
               value = Long.parseLong(s);
            }
            return new rowid((long) value);
            
         case TYPE_UNKNOWN:
            return new unknown();
            
         case TYPE_MEMPTR:
            if (value.getClass().isArray() && value.getClass().getComponentType() == byte.class)
            {
               return new MemoryBuffer(((byte[]) value));
            }
            else if (value instanceof MemoryBuffer)
            {
               return value;
            }
            
            throw new IllegalStateException("Can not determine the blob value from " + value);
         
         case TYPE_DATE:
            if (value instanceof GregorianCalendar)
            {
               value = Date.from(((GregorianCalendar) value).toZonedDateTime().toInstant());
            }

            return new date((Date) value);
            
         case TYPE_DATETIME:
            if (value instanceof GregorianCalendar)
            {
               value = Date.from(((GregorianCalendar) value).toZonedDateTime().toInstant());
            }

            return new datetime((Date) value);

         case TYPE_DATETIMETZ:
            if (value instanceof GregorianCalendar)
            {
               value = Date.from(((GregorianCalendar) value).toZonedDateTime().toInstant());
            }

            return new datetimetz((Date) value);
      }
      
      throw new IllegalArgumentException("Can not convert FWD value of type " + type);
   }
   
   /**
    * Convert a {@link BaseDataType} instance to a Java value.
    * 
    * @param    arg
    *           The value.
    * @param    type
    *           The value's type.
    * @param    field
    *           Flag indicating that this call is for a field value.
    * @param    extent
    *           The extent value.
    *           
    * @return   See above.
    */
   private Object toJavaValue(Object arg, int type, boolean field, int extent)
   {
      if (arg instanceof BaseDataType && ((BaseDataType) arg).isUnknown())
      {
         return null;
      }
      
      if (extent != 0 && extent != SourceNameMapper.NO_EXTENT)
      {
         List<Object> res = new ArrayList<>();
         
         int length = Array.getLength(arg);
         for (int i = 0; i < length; i++)
         {
            res.add(toJavaValue(Array.get(arg, i), type, field, SourceNameMapper.NO_EXTENT));
         }
         
         return res;
      }
      
      switch (type)
      {
         case TYPE_LOGICAL:
            return ((logical) arg).booleanValue();
            
         case TYPE_CHARACTER:
            return ((character) arg).toStringMessage();
            
         case TYPE_HANDLE:
            return handle.resourceId(((handle) arg).getResource());

         case TYPE_LONGCHAR:
            return ((longchar) arg).toStringMessage();
            
         case TYPE_INTEGER:
            return ((integer) arg).intValue();
         
         case TYPE_INT64:
            return ((int64) arg).longValue();
            
         case TYPE_DECIMAL:
            return ((decimal) arg).doubleValue(); // TODO: bigdecimal?
            
         case TYPE_BLOB:
            return ((blob) arg).getByteArray();
            
         case TYPE_CLOB:
            return ((clob) arg).toJavaType();
            
         case TYPE_RAW:
            return ((raw) arg).getByteArray();
            
         case TYPE_RECID:
            return ((recid) arg).longValue();
            
         case TYPE_ROWID:
            return ((rowid) arg).getValue();
            
         case TYPE_UNKNOWN:
            return null;
            
         case TYPE_MEMPTR:
            return (MemoryBuffer) arg;
         
         case TYPE_DATE:
         case TYPE_DATETIME:
            if (field)
            {
               return ((date) arg).dateValue();
            }
            else
            {
               GregorianCalendar gc = new GregorianCalendar();
               gc.setTime(((date) arg).dateValue());
               return gc;
            }
         case TYPE_DATETIMETZ:
            if (field)
            {
               return Date.from(((datetimetz) arg).calendarValue().toInstant());
            }
            else
            {
               return ((datetimetz) arg).calendarValue();
            }
      }

      throw new IllegalArgumentException("Could not resolve java value of type " + type);
   }
   
   /**
    * Create a table definition part of a dataset parameter.
    * 
    * @param    rows
    *           The table rows.
    * @param    fields
    *           The table fields.
    * @param    input
    *           Flag indicating if this parameter is in input or input-output mode.
    * @param    output
    *           Flag indicating if this parameter is in output or input-output mode.
    *           
    * @return   A {@link DsTableDefinition} instance with the specified metadata and records.
    */
   private DsTableDefinition createDsTable(List<DataObject> rows, 
                                           List<DataObject> fields, 
                                           boolean input, 
                                           boolean output)
   {
      Iterator<DataObject> fieldIter = fields.iterator();
      Iterator<DataObject> rowIter = rows.iterator();
      DsTableDefinition dstable = new DsTableDefinition(new TableResultSet(input, output, false)
      {
         DataObject nextRow;
         
         @Override
         public Object[] nextRow()
         {
            nextRow = rowIter.next();
            
            int numFields = fields.size();
            Object[] data = new Object[numFields];
            for (int i = 0; i < data.length; i++)
            {
               int ftype = DataSetSDOHelper.getFieldType(fields.get(i));
               int extent = DataSetSDOHelper.getFieldExtent(fields.get(i));
               Object fval = nextRow.get(i);

               data[i] = fromJavaType(fval, ftype, extent);
            }
            
            return data;
         }
         
         @Override
         public Object[] nextRowMeta()
         {
            FWDDataObject row = (FWDDataObject) nextRow;
            
            return row.toArray();
         }
         
         @Override
         public PropertyDefinition nextProperty()
         {
            DataObject fieldMeta = fieldIter.next();
            String name = DataSetSDOHelper.getFieldName(fieldMeta);
            Class<?> type = legacyTypes.get(DataSetSDOHelper.getFieldType(fieldMeta));
            int extent = DataSetSDOHelper.getFieldExtent(fieldMeta);
            
            PropertyDefinition prop = new PropertyDefinition(name, type, extent, name);
            
            return prop;
         }
         
         @Override
         public boolean hasMoreRows()
         {
            return rowIter.hasNext();
         }
         
         @Override
         public boolean hasMoreProperties()
         {
            return fieldIter.hasNext();
         }
      });
      
      return dstable;
   }
}
