/*
** Module   : ObjectOps.java
** Abstract : Defines APIs to implement 4GL object-specific operations.
**
** Copyright (c) 2018-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- ---------------------------------------Description----------------------------------------
** 001 CA  20181213 Created initial version.
**     OM  20181217 newInstance() returns the expected generic type.
** 002 OM  20190129 Remove reference after the destructors are invoked.
**     CA  20190219 Runtime implementaiton.
**     CA  20190326 Implemented ProcedureHelper and TransactionHelper instead of direct usage of
**                  the static API.  This allows the elimination of context local usage in 
**                  ProcedureManager and TransactionManager.
** 003 CA  20190418 Added no-arg DYNAMIC-NEW support.
** 004 CA  20190710 Fixed issues related to working with legacy OE classes for which FWD provides 
**                  support.  Do not delete a legacy instance if this is supposed to be returned 
**                  by a function call (this delete check is postponed in the parent scope).
**     CA  20190719 Plain object instance does not do reference counting.
**     CA  20190729 When loading a legacy class, backup and restore the output parameter assigners.
**     CA  20190995 Added cast(BDT) version.
** 005 AIL 20191120 Made the scope methods aware of DATABASE TRIGGER Block Type.
**     CA  20191112 Fixed instantiation of Progress.Lang.Class (via GetClass) - this must be a 
**                  unique instance for each invocation, and lifetime must match any other legacy
**                  4GL instance.
** 006 CA  20200116 Javadoc fixes.
** 007 CA  20200304 A 4GL object instance must delay its delete if is still on stack.
** 008 CA  20200330 More fixes related to OE-compatible error management, in constructor case
**                  and 'implicit' 4GL ERROR conditions.
** 009 CA  20200427 Replaced isValid().booleanValue() calls with _isValid().
**     CA  20200429 Added stubs for DYNAMIC-ENUM.
**     CA  20200503 Added stubs for GET-CLASS().
**     GES 20200506 Implemented DYNAMIC-ENUM().
**     GES 20200511 Bypassed life cycle management for legacy enums.
**     ME  20200521 Move getLegacyClass from LegacyClass, add class hierarchy registration.
**     GES 20200528 Fixed enum error handling.
**     GES 20200529 Added safety code to avoid life cycle processing for enums.
**     GES 20200603 Reworked legacy name/class mapping to use static data in a thread-safe manner.
** 010 ME  20200810 Added registerClassWorker() to getClassInstance().
**     CA  20201003 Replaced Guava identity HashSet with Collections.newSetFromMap(IdentityHashMap).
**     ME  20201009 Let 'invoke' return plain Object as it can be anything (including array).
**     CA  20201014 Reverted ME/20201009 - 'invoke' will return BDT, and the fix will either emit a 
**                  'invokeExtent' for calls where it is known for the lvalue to be an extent, or we add 
**                  BDT(Object) constructors for all sub-classes.
**     ME  20201127 Throw error instead of showing it if cast fails.
**     ME  20210128 Added cast/dynamic cast for object arrays.
**     CA  20210215 An INPUT parameter at the property SETTER must be registered as pending, so its reference
**                  can be tracked. 
**     CA  20210310 Performance improvement in hasDestructor - cache the result of this method.
**     CA  20210322 asResource(LegacyEnum) will always return an ObjectResource instance (and a resource ID
**                  computed), so temp-table fields can refer enums. 
**     ME  20210330 Added convenient increment/decrement methods for object references.
**     CA  20210804 Fixed a problem with caching available destructors - 'clsWithDtor' must have as key the 
**                  object's defining class paired with the destructor name (which may be from a super-class).
**     HC  20211011 Implemented i18n support.
**     CA  20211214 The legacy class instances are always singleton for the current context.
**     CA  20220106 THIS-OBJECT from a static method will return the legacy class object - this is required as
**                  in some cases (like 'clsEvt:subscribe(static-method)') FWD emits THIS-OBJECT as first
**                  argument and the parser does not resolve the static method name.
**     CA  20220120 Fixed some cases of pending assignment.  Some performance improvements.
**     CA  20220203 A small performance improvement for scopeFinished.
**     CA  20220208 Objects deserialized from a remote appserver call can be automatically registered as known
**                  instances.  Fixed a 'pendingAssign' bug when the root block exits. 
**     CA  20220901 Refactored scope notification support: ScopeableFactory was removed, and the registration 
**                  is now specific to each type of scopeable.  For each case, the block will be registered 
**                  for scope support (for that particular scopeable) only when the scopeable is 'active' 
**                  (i.e. unnamed streams or accumulators are used).  This allows a lazy registration of 
**                  scopeables, to avoid the unnecessary overhead of processing all the scopeables for each
**                  and every block.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

import com.goldencode.p2j.oo.lang.*;
import com.goldencode.p2j.security.*;

import java.lang.reflect.*;
import java.util.*;

import org.reflections.Reflections;

/**
 * Defines APIs to manage 4GL-style objects (instantiation, deletion, cast, etc).
 */
public class ObjectOps
{
   /** DYNAMIC-ENUM() error message specification. */
   private static final String DE_SPEC = "Could not evaluate '%s' parameter for DYNAMIC-ENUM";
   
   /** Legacy names for each loaded class. */
   private static final Map<Class<? extends _BaseObject_>, String> legacyNames = new IdentityHashMap<>();
   
   /** Map of legacy names converted classes. */
   private static final Map<String, Class<? extends _BaseObject_>> name2cls = new HashMap<>();
   
   /** Controls access to the static class/name lookup data. */
   private static final Object lock = new Object();
      
   /** Context-local data. */
   private static final ContextLocal<WorkArea> local = new ContextLocal<WorkArea>()
   {
      protected WorkArea initialValue()
      {
         return new WorkArea();
      }
   };
   
   
   
   static
   {
      // register all classes from 'oo sdk' implementation  
      Reflections reflections = new Reflections("com.goldencode.p2j.oo");
      Set<Class<? extends _BaseObject_>> legacyClasses = 
         reflections.getSubTypesOf(_BaseObject_.class);
      
      for (Class<? extends _BaseObject_> cls : legacyClasses) {
         registerClass(cls, false);
      }
   }

   /**
    * Register all specified object variables (scalar or extent) for scope notification.  The
    * notification will only process for top-level scopes.  When their defining top-level block
    * terminates, these will be processed, so that the reference count for the wrapped instance
    * will be decremented.
    * <p>
    * Important: this must not registered NEW GLOBAL SHARED or SHARED variables - only NEW SHARED
    * can be registered here.
    *  
    * @param    objects
    *           The objects to register.
    */
   public static void register(Object... objects)
   {
      WorkArea wa = local.get();

      wa.tm.registerBlockScopeable(wa);
      
      BlockType bt = wa.tm.getNearestTopLevelType();
      boolean extProg = bt.equals(BlockType.EXTERNAL_PROC);
      
      Set<Object> objs = (extProg ? wa.extProgDefinitions.get(wa.pm._thisProcedure()) 
                                  : wa.definitions.peek());
      if (objs == Collections.EMPTY_SET)
      {
         objs = Collections.newSetFromMap(new IdentityHashMap<>());
         if (extProg)
         {
            wa.extProgDefinitions.put(wa.pm._thisProcedure(), objs);
         }
         else
         {
            wa.definitions.pop();
            wa.definitions.push(objs);
         }
      }
      
      for (Object o : objects)
      {
         objs.add(o);
      }
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(String clsName, String mthdName)
   {
      return invoke(clsName, mthdName, null);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(character clsName, String mthdName)
   {
      return invoke(clsName.isUnknown() ? (String) null : clsName.toStringMessage(), 
                    mthdName,
                    null);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(String clsName, character mthdName)
   {
      return invoke(clsName, 
                    mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(), 
                    null);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(character clsName, character mthdName)
   {
      return invoke(clsName.isUnknown() ? (String) null : clsName.toStringMessage(), 
                    mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(),
                    null);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(character clsName, 
                                     String    mthdName, 
                                     String    modes, 
                                     Object... args)
   {
      return invoke(clsName.isUnknown() ? (String) null : clsName.toStringMessage(), 
                    mthdName,
                    modes,
                    args);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(String    clsName, 
                                     character mthdName, 
                                     String    modes, 
                                     Object... args)
   {
      return invoke(clsName, 
                    mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(), 
                    modes, 
                    args);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(character clsName,
                                     character mthdName,
                                     String    modes, 
                                     Object... args)
   {
      return invoke(clsName.isUnknown() ? (String) null : clsName.toStringMessage(), 
                    mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(),
                    modes,
                    args);
   }

   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(String clsName, String mthdName, String modes, Object... args)
   {
      // TODO: extent?
      return (BaseDataType) invoke(false, clsName, mthdName, modes, args);
   }

   /**
    * Invoke the specified instance method in the specified object.
    * 
    * @param     ref
    *            The legacy object reference.
    * @param     mthdName
    *            The method name.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(object<? extends _BaseObject_> ref, character mthdName)
   {
      return invoke(ref, mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(), null);
   }
   
   /**
    * Invoke the specified instance method in the specified object.
    * 
    * @param     ref
    *            The legacy object reference.
    * @param     mthdName
    *            The method name.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(object<? extends _BaseObject_> ref, String mthdName)
   {
      return invoke(ref, mthdName, null);
   }

   /**
    * Invoke the specified instance method in the specified object.
    * 
    * @param     ref
    *            The legacy object reference.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(object<? extends _BaseObject_> ref, 
                                     character                      mthdName, 
                                     String                         modes, 
                                     Object...                      args)
   {
      return invoke(ref,
                    mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(),
                    modes, 
                    args);
   }
   
   /**
    * Invoke the specified instance method in the specified object.
    * 
    * @param     ref
    *            The legacy object reference.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   public static BaseDataType invoke(object<? extends _BaseObject_> ref, 
                                     String                         mthdName, 
                                     String                         modes, 
                                     Object...                      args)
   {
      // TODO: extent?
      return (BaseDataType) invoke(false, ref, mthdName, modes, args);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    */
   public static void invokeStandalone(String clsName, String mthdName)
   {
      invoke(true, clsName, mthdName, null);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    */
   public static void invokeStandalone(character clsName, String mthdName)
   {
      invoke(true, 
             clsName.isUnknown() ? (String) null : clsName.toStringMessage(), 
             mthdName,
             null);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    */
   public static void invokeStandalone(String clsName, character mthdName)
   {
      invoke(true,
             clsName, 
             mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(), 
             null);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    */
   public static void invokeStandalone(character clsName, character mthdName)
   {
      invoke(true,
             clsName.isUnknown() ? (String) null : clsName.toStringMessage(), 
             mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(),
             null);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    */
   public static void invokeStandalone(character clsName, 
                                       String    mthdName, 
                                       String    modes, 
                                       Object... args)
   {
      invoke(true,
             clsName.isUnknown() ? (String) null : clsName.toStringMessage(), 
             mthdName,
             modes,
             args);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    */
   public static void invokeStandalone(String    clsName, 
                                       character mthdName, 
                                       String    modes, 
                                       Object... args)
   {
      invoke(true,
             clsName, 
             mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(), 
             modes, 
             args);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    */
   public static void invokeStandalone(character clsName,
                                       character mthdName,
                                       String    modes, 
                                       Object... args)
   {
      invoke(true,
             clsName.isUnknown() ? (String) null : clsName.toStringMessage(), 
             mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(),
             modes,
             args);
   }

   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    */
   public static void invokeStandalone(String    clsName, 
                                       String    mthdName,
                                       String    modes, 
                                       Object... args)
   {
      invoke(true, clsName, mthdName, modes, args);
   }

   /**
    * Invoke the specified instance method in the specified object.
    * 
    * @param     ref
    *            The legacy object reference.
    * @param     mthdName
    *            The static method name.
    */
   public static void invokeStandalone(object<? extends _BaseObject_> ref, character mthdName)
   {
      invoke(true, ref, mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(), null);
   }
   
   /**
    * Invoke the specified instance method in the specified object.
    * 
    * @param     ref
    *            The legacy object reference.
    * @param     mthdName
    *            The static method name.
    */
   public static void invokeStandalone(object<? extends _BaseObject_> ref, String mthdName)
   {
      invoke(true, ref, mthdName, null);
   }

   /**
    * Invoke the specified instance method in the specified object.
    * 
    * @param     ref
    *            The legacy object reference.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    */
   public static void invokeStandalone(object<? extends _BaseObject_> ref, 
                                       character                      mthdName, 
                                       String                         modes, 
                                       Object...                      args)
   {
      invoke(true,
             ref,
             mthdName.isUnknown() ? (String) null : mthdName.toStringMessage(),
             modes, 
             args);
   }
   
   /**
    * Invoke the specified instance method in the specified object.
    * 
    * @param     ref
    *            The legacy object reference.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    */
   public static void invokeStandalone(object<? extends _BaseObject_> ref, 
                                       String                         mthdName, 
                                       String                         modes, 
                                       Object...                      args)
   {
      invoke(true, ref, mthdName, modes, args);
   }

   /**
    * Deletes the given object.
    *
    * @param   o
    *          The object to be deleted.
    */
   public static void delete(object<?> o)
   {
      if (o.isUnknown() || !isValid(o.ref))
      {
         ErrorManager.recordOrThrowError(5425, 
                                         "Invalid or inappropriate handle value given to " +
                                         "DELETE OBJECT or DELETE PROCEDURE statement", 
                                         false);
         return;
      }
      
      // non-tracked instances cannot be deleted, this is a NOP
      if (!o.ref.isTracked())
      {
         return;
      }

      boolean deleted = false;
      
      try
      {
         deleted = delete(o.ref);
      }
      finally
      {
         if (deleted)
         {
            o.setUnknown();
         }
      }
   }

   /**
    * Create a new dynamic object.
    * 
    * @param    legacyClsName
    *           The legacy class name.  This may be qualified or unqualified.
    *           
    * @return   A new created object wrapped in a {@link object} instance.
    */
   public static <T extends _BaseObject_> object<T> newDynamicInstance(String legacyClsName)
   {
      return newDynamicInstance(legacyClsName, null);
   }
   
   /**
    * Create a new dynamic object.
    * 
    * @param    legacyClsName
    *           The legacy class name.  This may be qualified or unqualified.
    * @param    modes
    *           The argument's modes.
    * @param    args
    *           The arguments.
    *           
    * @return   A new created object wrapped in a {@link object} instance.
    */
   public static <T extends _BaseObject_> object<T> newDynamicInstance(String    legacyClsName,
                                                                       String    modes,
                                                                       Object... args)
   {
      Class<T> cls = resolveClass(legacyClsName);
      
      if (cls == null)
      {
         ErrorManager.recordOrThrowError(14284, 
                                         "DYNAMIC-NEW could not find class " + legacyClsName, 
                                         false);
         return new object<T>();
      }
      
      // enums cannot be instantiated
      if (LegacyEnum.failEnumInstantiation(cls))
      {
         return new object<T>();
      }
      
      return newInstanceInternal(cls, legacyClsName, modes, args);
   }

   /**
    * Create a new dynamic object.
    * 
    * @param    legacyClsName
    *           The legacy class name.  This may be qualified or unqualified.
    *           
    * @return   A new created object wrapped in a {@link object} instance.
    */
   public static <T extends _BaseObject_> object<T> newDynamicInstance(character legacyClsName)
   {
      return newDynamicInstance(legacyClsName.toStringMessage(), null);
   }

   /**
    * Create a new dynamic object.
    * 
    * @param    legacyClsName
    *           The legacy class name.  This may be qualified or unqualified.
    * @param    modes
    *           The argument's modes.
    * @param    args
    *           The arguments.
    *           
    * @return   A new created object wrapped in a {@link object} instance.
    */
   public static <T extends _BaseObject_> object<T> newDynamicInstance(character legacyClsName,
                                                                       String    modes,
                                                                       Object... args)
   {
      return newDynamicInstance(legacyClsName.toStringMessage(), modes, args);
   }
   
   /**
    * Create a new object.
    * 
    * @param    cls
    *           The object's class.
    *           
    * @return   A new created object wrapped in a {@link object} instance.
    */
   public static <T extends _BaseObject_> object<T> newInstance(Class<T>  cls)
   {
      return newInstance(cls, null);
   }
   
   /**
    * Create a new object.
    * 
    * @param    cls
    *           The object's class.
    * @param    modes
    *           The argument modes.
    * @param    args
    *           The arguments.
    *           
    * @return   A new created object wrapped in a {@link object} instance.
    */
   public static <T extends _BaseObject_> object<T> newInstance(Class<T>  cls,
                                                                String    modes,
                                                                Object... args)
   {
      String name = getLegacyName(cls);
      
      return newInstanceInternal(cls, name, modes, args);
   }

   /**
    * Create a new object.
    * 
    * @param    cls
    *           The object's class.
    * @param    name
    *           The legacy class name.  This may be qualified or unqualified.
    * @param    modes
    *           The argument modes.
    * @param    args
    *           The arguments.
    *           
    * @return   A new created object wrapped in a {@link object} instance.
    */
   public static <T extends _BaseObject_> object<T> newInstanceInternal(Class<T>  cls, 
                                                                        String    name, 
                                                                        String    modes,
                                                                        Object... args)
   {
      if (cls == _BaseObject_.class)
      {
         cls = (Class<T>) BaseObject.class;
      }

      // must not be an interface, enum or an abstract class

      if (LegacyEnum.class.isAssignableFrom(cls))
      {
         String cname = ObjectOps.getLegacyName(cls);
         
         // in the 4GL, this is a compile error but this provides runtime safety
         String msg = String.format("Cannot use NEW statement with '%s' because it is an enum",
                                    cname);
         
         ErrorManager.recordOrThrowError(17970, msg, false);
         
         return new object<T>();
      }
      if (Modifier.isAbstract(cls.getModifiers()))
      {
         ErrorManager.recordOrThrowError(15107, 
                                         "You cannot create an instance of " + name + 
                                         " because it is an abstract class",
                                         false);
         
         return new object<T>();
      }
      if (cls.isInterface())
      {
         ErrorManager.recordOrThrowError(12879, 
                                         name + " is an interface and cannot be used as a " +
                                         "procedure name for RUN or the type name for a NEW " +
                                         "statement", 
                                         false);
         
         return new object<T>();
      }

      ObjectResource res = null;
      try
      {
         //  before creating the resource, load the static constructor(s)
         if (!loadClass(cls))
         {
            return new object<>();
         }
         
         T ref = cls.newInstance();

         res = newInstanceInternal(ref, cls, name, modes, args);
         
         if (res == null)
         {
            return new object<T>();
         }
         
         return new object<T>(ref);
      }
      catch (InstantiationException   |
             IllegalAccessException   |
             IllegalArgumentException e)
      {
         if (res != null)
         {
            res.delete();
         }

         // TODO: log this?
         return new object<T>();
      }
   }
   
   /**
    * Create a new object.
    * 
    * @param    ref
    *           The Java reference for this {@link _BaseObject_} instance.
    * @param    cls
    *           The object's class.
    * @param    name
    *           The legacy class name.  This may be qualified or unqualified.
    * @param    modes
    *           The argument modes.
    * @param    args
    *           The arguments.
    *           
    * @return   A new created object wrapped in a {@link object} instance.
    */
   private static <T extends _BaseObject_> ObjectResource newInstanceInternal(T         ref, 
                                                                              Class<T>  cls, 
                                                                              String    name, 
                                                                              String    modes,
                                                                              Object... args)
   {
      // enums can't be instantiated
      if (LegacyEnum.class.isAssignableFrom(cls))
      {
         String cname = ObjectOps.getLegacyName(cls);
         
         // in the 4GL, this is a compile error but this provides runtime safety
         String msg = String.format("Cannot use NEW statement with '%s' because it is an enum",
                                    cname);
         
         ErrorManager.recordOrThrowError(17970, msg, false);
         
         return null;
      }

      // create a synthetic resource tracking this object; this is performed before anything 
      // else (static or instance constructor) gets executed
      ObjectResource res = new ObjectResource(ref);

      WorkArea wa = local.get();
      // this is saved regardless if instantiation passes or not - it will be cleared on delete
      wa.objects.put(ref, res);

      String ctor = getConstructorMethod(cls);
      String[] executes = getExecuteMethods(cls);

      // if an error was throw from within a c'tor (or UNDO, THROW with ROUTINE-LEVEL ON ERROR UNDO, THROW.)
      // then the initialization failed and call the destructors for all initialized classes
      
      pendingInitialize(wa, ref);
      
      boolean ok = false;
      boolean err = false;
      try
      {
         ok = ControlFlowOps.initializeLegacyObject(ref, name, executes, ctor, modes, args);
      }
      
      // allow Progress compatible exceptions to flow through
      catch (ConditionException | RetryUnwindException ce)
      {
         err = true;
         throw ce;
      }
      
      finally
      {
         endInitialization(ref);
         
         if (!ok)
         {
            // couldn't initialize... delete the resource
            res.delete();
            
            if (!err)
            {
               return null;
            }
         }
      }
      
      if (!isValid(ref))
      {
         // the object was deleted by the constructor - don't do anything else
         return null;
      }
      
      // 4GL deletes the NEW statement's reference immediately (if was not saved by other
      // means). we delay this check until the block exits.
      pendingAssign(ref);
      
      return res;
   }
   
   /**
    * Check if the given class is the same or a super-class of the given object.
    * 
    * @param    o
    *           The object instance.
    * @param    cls
    *           The type to check.
    *           
    * @return   <code>true</code> if the given class is the same or a super-class of the object's 
    *           type.
    */
   public static logical typeOf(object<?> o, Class<?> cls)
   {
      if (o.isUnknown())
      {
         // unknown value is returned
         return new logical();
      }
      
      Object ref = o.ref;
      
      return new logical(cls != null && ref != null && cls.isInstance(ref));
   }
   
   /**
    * Check if the given class is the same or a super-class of the given object.
    * 
    * @param    o
    *           The object instance.
    * @param    cls
    *           The type to check.
    *           
    * @return   <code>true</code> if the given class is the same or a super-class of the object's 
    *           type.
    */
   public static logical typeOf(object<?> o, String cls)
   {
      if (o.isUnknown())
      {
         // unknown value is returned
         return new logical();
      }
      
      Class<? extends _BaseObject_> lcls = resolveClass(cls);
      if (lcls == null)
      {
         return new logical();
      }
      
      return typeOf(o, lcls);
   }

   /**
    * Cast this object to the specified type.
    * 
    * @param    val
    *           The object instance.
    * @param    cls
    *           The type to check.
    *           
    * @return   An object of the new type.  If the value is not an {@link object}, then unknown
    *           is returned.
    */
   public static <T extends _BaseObject_> object<T> cast(BaseDataType val, Class<T> cls)
   {
      if (!(val instanceof object))
      {
         // if the value is not an object, then return unkown. 4GL doesn't show any error.
         return new object();
      }
      
      return cast((object) val, cls);
   }
   
   /**
    * Cast this object to the specified type.
    * 
    * @param    o
    *           The object instance.
    * @param    cls
    *           The type to check.
    *           
    * @return   A new {@link object} instance, if the cast is possible.
    */
   public static <T extends _BaseObject_> object<T> cast(object<?> o, Class<T> cls)
   {
      if (o.isUnknown())
      {
         // no error is thrown
         return new object<T>();
      }
      
      if (!cls.isInstance(o.ref))
      {
         String t1 = getLegacyName((Class<? extends _BaseObject_>) o.ref.getClass());
         String t2 = getLegacyName(cls);
         
         // at least for enums this is a warning, the error-status:error is false
         ErrorManager.recordOrThrowError(12869, "Invalid cast from " + t1 + " " + t2, false, false);
         
         // unknown object is returned
         return new object<T>();
      }
      
      return new object<T>((T) o.ref);
   }
   
   /**
    * Cast this object array to the specified type.
    * 
    * @param    o
    *           The object array instance.
    * @param    cls
    *           The type to check.
    *           
    * @return   A new {@link object} array instance, if the cast is possible.
    */
   public static <T extends _BaseObject_> object<T>[] cast(object<?>[] o, Class<T> cls)
   {
      object<T>[] objArr = TypeFactory.objectExtent(o.length, cls, false);
      
      for (int i = 0; i < objArr.length; i++)
      {
         objArr[i].assign(cast(o[i], cls));
      }
      
      return objArr;
   }

   /**
    * Find and return the specified enum.
    * 
    * @param    cname
    *           The enum class name.
    * @param    value
    *           The enum value.
    *           
    * @return   The enum instance.
    */
   public static object<? extends LegacyEnum> dynamicEnum(String cname, int64 value)
   {
      return dynamicEnum(new character(cname), value);
   }

   /**
    * Find and return the specified enum.
    * 
    * @param    cname
    *           The enum class name.
    * @param    value
    *           The enum value.
    *           
    * @return   The enum instance.
    */
   public static <T extends LegacyEnum> object<T> dynamicEnum(character cname, int64 value)
   {
      Class<T> cls = LegacyEnum.findClass(cname,
                                          ObjectOps::generateInvalidTypeDE,
                                          LegacyEnum::generateFailedToFindEnumType);
      
      if (cls == null)
      {
         return new object<T>();
      }
      
      return LegacyEnum.lookupWorker(cls, value, ObjectOps::generateInvalidValueDE); 

   }

   /**
    * Find and return the specified enum.
    * 
    * @param    cname
    *           The enum class name.
    * @param    name
    *           The enum name.
    *           
    * @return   The enum instance.
    */
   public static object<? extends LegacyEnum> dynamicEnum(String cname, character name)
   {
      return dynamicEnum(new character(cname), name);
   }

   /**
    * Find and return the specified enum.
    * 
    * @param    cname
    *           The enum class name.
    * @param    name
    *           The enum name.
    *           
    * @return   The enum instance.
    */
   public static <T extends LegacyEnum> object<T> dynamicEnum(character cname, character name)
   {
      Class<T> cls = LegacyEnum.findClass(cname,
                                          ObjectOps::generateInvalidTypeDE,
                                          LegacyEnum::generateFailedToFindEnumType);
      
      if (cls == null)
      {
         return new object<T>();
      }
      
      return LegacyEnum.lookupWorker(cls, name, ObjectOps::generateInvalidValueDE); 
   }

   /**
    * Find and return the specified enum.
    * 
    * @param    cname
    *           The enum class name.
    * @param    value
    *           The enum value.
    *           
    * @return   The enum instance.
    */
   public static object<? extends LegacyEnum> dynamicEnum(String cname, long value)
   {
      return dynamicEnum(new character(cname), new int64(value));
   }

   /**
    * Find and return the specified enum.
    * 
    * @param    cname
    *           The enum class name.
    * @param    value
    *           The enum value.
    *           
    * @return   The enum instance.
    */
   public static object<? extends LegacyEnum> dynamicEnum(character cname, long value)
   {
      return dynamicEnum(cname, new int64(value));
   }

   /**
    * Find and return the specified enum.
    * 
    * @param    cname
    *           The enum class name.
    * @param    name
    *           The enum name.
    *           
    * @return   The enum instance.
    */
   public static object<? extends LegacyEnum> dynamicEnum(String cname, String name)
   {
      return dynamicEnum(new character(cname), new character(name));
   }

   /**
    * Find and return the specified enum.
    * 
    * @param    cname
    *           The enum class name.
    * @param    name
    *           The enum name.
    *           
    * @return   The enum instance.
    */
   public static object<? extends LegacyEnum> dynamicEnum(character cname, String name)
   {
      return dynamicEnum(cname, new character(name));
   }

   /**
    * Cast this object to the specified type, resolved dynamically.
    * 
    * @param    o
    *           The object instance.
    * @param    legacyClsName
    *           The legacy class name.  This may be qualified or unqualified.
    *           
    * @return   A new {@link object} instance, if the cast is possible.  
    */
   public static object<? extends _BaseObject_> dynamicCast(object<?> o, String legacyClsName)
   {
      Class<? extends _BaseObject_> cls = resolveClass(legacyClsName);
      
      if (cls == null)
      {
         ErrorManager.recordOrThrowError(12900, 
                                        "CAST to '" + legacyClsName + "' not allowed. " + 
                                        "CAST target must be a user-defined type",
                                        false,
                                        false);
         return new object<>();
      }

      return cast(o, cls);
   }

   /**
    * Cast this object to the specified type, resolved dynamically.
    * 
    * @param    o
    *           The object instance.
    * @param    legacyClsName
    *           The legacy class name.  This may be qualified or unqualified.
    *           
    * @return   A new {@link object} instance, if the cast is possible.  
    */
   public static object<? extends _BaseObject_> dynamicCast(object<?> o, character legacyClsName)
   {
      if (legacyClsName.isUnknown() || legacyClsName.toStringMessage().isEmpty())
      {
         ErrorManager.recordOrThrowError(12900, 
                                        "CAST to 'UNKNOWN' not allowed. " + 
                                        "CAST target must be a user-defined type",
                                        false,
                                        false);

         return new object<>();
      }

      return dynamicCast(o, legacyClsName.toStringMessage());
   }
   
   /**
    * Cast this object array to the specified type, resolved dynamically.
    * 
    * @param    o
    *           The object array instance.
    * @param    legacyClsName
    *           The legacy class name.  This may be qualified or unqualified.
    *           
    * @return   A new {@link object} array instance, if the cast is possible.  
    */
   public static object<? extends _BaseObject_>[] dynamicCast(object<?>[] o, character legacyClsName)
   {
      if (legacyClsName.isUnknown() || legacyClsName.toStringMessage().isEmpty())
      {
         ErrorManager.recordOrThrowError(12900, 
                                        "CAST to 'UNKNOWN' not allowed. " + 
                                        "CAST target must be a user-defined type", 
                                        false,
                                        false);
      }

      return dynamicCast(o, legacyClsName.toStringMessage());
   }
   
   /**
    * Cast this object array to the specified type, resolved dynamically.
    * 
    * @param    o
    *           The object array instance.
    * @param    legacyClsName
    *           The legacy class name.  This may be qualified or unqualified.
    *           
    * @return   A new {@link object} array instance, if the cast is possible.  
    */
   public static object<? extends _BaseObject_>[] dynamicCast(object<?>[] o, String legacyClsName)
   {
      Class<? extends _BaseObject_> cls = resolveClass(legacyClsName);
      
      if (cls == null)
      {
         ErrorManager.recordOrThrowError(12900, 
                                        "CAST to '" + legacyClsName + "' not allowed. " + 
                                        "CAST target must be a user-defined type",
                                        false,
                                        false);
      }
      
      object<_BaseObject_>[] objArr = TypeFactory.objectExtent(o.length, _BaseObject_.class, false);

      int numErr = 0;
      int[] errNums = new int[o.length];
      String[] errMsgs = new String[o.length];

      for (int i = 0; i < objArr.length; i++)
      {
         final int idx = i;

         if (ErrorManager
                  .silent(() -> objArr[idx].assign(cast(o[idx], cls))))
         {
            errNums[numErr] = ErrorManager.getErrorNumber(1).intValue();
            errMsgs[numErr] = ErrorManager.getErrorText(1).getValue();
            numErr++;
         }
      }

      if (numErr > 0)
         ErrorManager.recordOrThrowError(Arrays.copyOf(errNums, numErr), Arrays.copyOf(errMsgs, numErr), false, false);
      
      return objArr;
   }
   
   /**
    * Get the THIS-OBJECT system reference.
    * 
    * @return   See above.
    */
   public static object<? extends _BaseObject_> thisObject()
   {
      object<? extends _BaseObject_> o = new object<>();
      
      WorkArea wa = local.get();
      Object ref = wa.pm._thisProcedure();
      
      if (ref instanceof _BaseObject_)
      {
         o.set((_BaseObject_) ref);
      }
      else if (ref.getClass() == Object.class)
      {
         Class<? extends _BaseObject_> cls = wa.referents2static.get(ref);
         o.set(getClassInstance(cls));
      }
      
      return o;
   }

   /**
    * Check if this object is valid.
    * 
    * @param    ref
    *           The legacy object.
    *           
    * @return   <code>true</code> if this object is valid.
    */
   public static boolean isValid(_BaseObject_ ref)
   {
      // not tracked instances are always valid
      if (ref != null && !ref.isTracked())
      {
         return true;
      }
      
      return ref != null && local.get().objects.containsKey(ref);
   }

   /**
    * Check if the given instance is serializable.
    * 
    * @param    ref
    *           The instance to check.
    *           
    * @return   See above.
    */
   public static boolean isSerializable(_BaseObject_ ref)
   {
      // TODO: check the annotation at the type, and not 'instanceof Serializable'
      // TODO: see RemoteInvocationResult.setError, the 'send' flag needs to be adjusted once this is complete.
      return false;
   }
   
   /**
    * Convert this object to its associated pseudo-resource.
    * 
    * @param    ref
    *           The legacy object.
    *           
    * @return   The pseudo-resource.
    */
   public static ObjectResource asResource(_BaseObject_ ref)
   {
      if (ref instanceof LegacyEnum)
      {
         return new ObjectResource(ref);
      }
      else
      {
         return local.get().objects.get(ref);
      }
   }

   /**
    * Load the given legacy class.  This will initialize it by recursively execute the static
    * c'tor for all super-classes (if defined).
    * 
    * @param    cls
    *           The class to initialize.
    */
   public static void load(Class<? extends _BaseObject_> cls)
   {
      loadClass(cls);
   }

   /**
    * Increment the reference counter for this instance.
    * 
    * @param    ref
    *           The legacy instance.
    */
   public static void increment(_BaseObject_ ref)
   {
      // non-tracked instances cannot be reference counted, this is a NOP
      if (ref != null && !ref.isTracked())
      {
         return;
      }
      
      if (isValid(ref))
      {
         local.get().objects.get(ref).increment();
      }
      else
      {
         // TODO: throw exception?
      }
   }
   
   /**
    * Increment the reference counter for this object instance.
    * 
    * @param    obj
    *           The legacy object instance.
    */
   public static void increment(object<? extends _BaseObject_> obj)
   {
      if (obj != null && obj._isValid())
         increment(obj.ref());
   }
   
   /**
    * Decrement the reference counter for this instance.
    * 
    * @param    ref
    *           The legacy instance.
    */
   public static void decrement(_BaseObject_ ref)
   {
      // non-tracked instances cannot be reference counted, this is a NOP
      if (ref != null && !ref.isTracked())
      {
         return;
      }
      
      if (isValid(ref))
      {
         local.get().objects.get(ref).decrement();
      }
      else
      {
         // TODO: throw exception?
      }
   }
   
   /**
    * Decrement the reference counter for this object instance.
    * 
    * @param    obj
    *           The legacy object instance.
    */
   public static void decrement(object<? extends _BaseObject_> obj)
   {
      if (obj != null && obj._isValid())
         decrement(obj.ref());
   }

   /**
    * Deregister the specified instance.
    * 
    * @param    ref
    *           The instance to deregister.
    */
   static void deregisterObject(_BaseObject_ ref)
   {
      WorkArea wa = local.get();
      wa.objects.remove(ref);
   }
   
   /**
    * Register this instance as a local resource.  This is used when the instance was created from a remote
    * side, and send via serialization.
    * 
    * @param    ref
    *           The instance to register.
    */
   static void registerObject(_BaseObject_ ref)
   {
      WorkArea wa = local.get();
      wa.objects.put(ref, new ObjectResource(ref));
   }
   
   /**
    * When an INPUT object variable is initialized in the top-level block via {@link TypeFactory#object},
    * it needs to be registered as 'pending'.  This will defer its registration with the 
    * {@link WorkArea#definitions} until the top-level block starts.
    * 
    * @param    objects
    *           The list of {@link object} definitions.
    */
   static void registerPending(Object... objects)
   {
      WorkArea wa = local.get();
      
      wa.pendingObjects.addAll(Arrays.asList(objects));
      
      registerPendingScopeable(wa);
   }

   /**
    * Register the {@link WorkArea} instance as pending scopeable, to be added to the next block.
    * 
    * @param    wa
    *           The {@link WorkArea} instance.
    */
   static void registerPendingScopeable(WorkArea wa)
   {
      if (wa == null)
      {
         wa = local.get();
      }
      
      wa.tm.registerPendingScopeable(wa);
   }
   
   /**
    * Check if the reference is marked as {@link WorkArea#pendingAssigned}.
    * 
    * @param    ref
    *           The reference to check.
    *           
    * @return   See above.
    */
   static boolean isPendingAssigned(_BaseObject_ ref)
   {
      WorkArea wa = local.get();
      
      return !wa.pendingAssigned.isEmpty() && wa.pendingAssigned.peek().contains(ref);
   }

   /**
    * Mark this instance as a pending assignment.
    * 
    * @param    ref
    *           The legacy instance.
    */
   static void pendingAssign(_BaseObject_ ref)
   {
      pendingAssign(ref, true);
   }
   
   /**
    * Mark this instance as a pending assignment.
    * 
    * @param    ref
    *           The legacy instance.
    * @param    register
    *           Flag indicating if the scopeable support must be registered.
    */
   static void pendingAssign(_BaseObject_ ref, boolean register)
   {
      if (ref == null)
      {
         return;
      }
      
      // pending initialization
      if (isInitializing(ref))
         return;
      
      WorkArea wa = local.get();
      if (register)
      {
         wa.tm.registerBlockScopeable(wa);
      }
      
      Set<_BaseObject_> pending = wa.pendingAssigned.peek();
      if (pending == Collections.EMPTY_SET)
      {
         wa.pendingAssigned.pop();
         wa.pendingAssigned.push(pending = Collections.newSetFromMap(new IdentityHashMap<>()));
      }
      pending.add(ref);
   }
   
   /**
    * Mark this instance as a pending initialization.
    * 
    * @param    wa
    *           The {@link WorkArea} instance.
    * @param    ref
    *           The legacy instance.
    */
   private static void pendingInitialize(WorkArea wa, _BaseObject_ ref)
   {
      List<Object> savePending = wa.pendingObjects.isEmpty() ? null : new ArrayList<>(wa.pendingObjects);
      try
      {
         // clear and restore the pending objects, as 'registerBlockScopeable' may register current block for
         // scope notifications, but the 'pending' objects are not ours.
         wa.pendingObjects.clear();
         wa.tm.registerBlockScopeable(wa);
         
         Set<_BaseObject_> pending = wa.pendingInitialized.peek();
         if (pending == Collections.EMPTY_SET)
         {
            wa.pendingInitialized.pop();
            wa.pendingInitialized.push(pending = Collections.newSetFromMap(new IdentityHashMap<>()));
         }
         pending.add(ref);
      }
      finally
      {
         wa.pendingObjects.clear();
         if (savePending != null)
         {
            wa.pendingObjects.addAll(savePending);
         }
      }
   }
   
   /**
    * Remove this instance from the pending assignments, as it was assigned.
    * 
    * @param    ref
    *           The legacy instance.
    */
   static void instanceAssigned(_BaseObject_ ref)
   {
      for (Set<_BaseObject_> pending : local.get().pendingAssigned)
      {
         if (pending.remove(ref))
         {
            break;
         }
      }
   }
   
   /**
    * Remove this instance from the pending initializations, as it was initialized.
    * 
    * @param    ref
    *           The legacy instance.
    */
   static void endInitialization(_BaseObject_ ref)
   {
      for (Set<_BaseObject_> pending : local.get().pendingInitialized)
      {
         if (pending.remove(ref))
         {
            break;
         }
      }
   }
   
   /**
    * Return true if this instance is pending initialization.
    * 
    * @param    ref
    *           The legacy instance.
    */
   static boolean isInitializing(_BaseObject_ ref)
   {
      for (Set<_BaseObject_> pending : local.get().pendingInitialized)
      {
         if (pending.contains(ref))
         {
            return true;
         }
      }
      
      return false;
   }
   
   /**
    * Check if this instance is a legacy object or a pseudo-instance associated with a static
    * class.
    * 
    * @param    ref
    *           The legacy object.
    * 
    * @return   See above.
    */
   static boolean isLegacyObject(Object ref)
   {
      return ref != null && (ref instanceof _BaseObject_ || ref.getClass() == Object.class);
   }
   
   /**
    * Deletes the given object.
    *
    * @param   ref
    *          The object to be deleted.
    * 
    * @return  <code>true</code> if the delete was possible (the instance is not on stack).
    */
   static boolean delete(_BaseObject_ ref)
   {
      if (ref instanceof LegacyEnum)
         return false;
      
      WorkArea wa = local.get();
      for (Set<_BaseObject_> pending : wa.pendingAssigned)
      {
         if (pending.contains(ref))
         {
            // this reference may be assigned - let it survive and will be cleaned up on the next
            // block exit
            return false;
         }
      }
      
      ObjectResource res = wa.objects.get(ref);
      if (res.isPendingDelete())
      {
         return false;
      }

      // allow the destructors to be called before actually removing the object
      res.delete();
      if (!res.valid())
      {
         wa.objects.remove(ref);
      }
      
      return !res.valid();
    }
   
   /**
    * Checks whether an object has any destructors pushed on stack ready to be executed when the
    * lifetime of the object ends.
    *
    * @param   ref
    *          The reference to the object to be tested.
    * 
    * @return  {@code true} if when the object is deleted at least one destructor will be
    *          executed. 
    */
   public static boolean hasDestructors(_BaseObject_ ref)
   {
      // non-tracked instances cannot have destructors
      if (ref != null && !ref.isTracked())
      {
         return false;
      }
      
      return isValid(ref) && local.get().objects.get(ref).hasActiveDestructors();
   }
   
   /**
    * When resizing an legacy object array with dynamic extent,  we must re-register the array 
    * reference.
    * 
    * @param    oldArray
    *           The old array.
    * @param    newArray
    *           The new array.
    */
   static void reRegister(BaseDataType[] oldArray, BaseDataType[] newArray)
   {
      WorkArea work = local.get();
      
      // go through all scopes
      for (Set<Object> scopeArrays : work.definitions)
      {
         if (scopeArrays.remove(oldArray))
         {
             // oldArray exists in the scope, replace with newArray
             scopeArrays.add(newArray);
             return;
         }
      }

      for (Set<Object> scopeArrays : work.extProgDefinitions.values())
      {
         if (scopeArrays.remove(oldArray))
         {
             // oldArray exists in the scope, replace with newArray
             scopeArrays.add(newArray);
             return;
         }
      }
   }

   /**
    * Get the {@link LegacyClass} mapped in FWD by the specified converted class.
    * 
    * @param    cls
    *           The converted class.
    *           
    * @return   A {@link LegacyClass} instance.
    */
   public static object<? extends LegacyClass> getLegacyClass(Class<? extends _BaseObject_> cls)
   {
      return new object<>(getClassInstance(cls));
   }

   /**
    * Get the {@link LegacyClass} mapped in FWD by the specified converted class name.
    * 
    * @param    typeName
    *           The converted class name.
    *           
    * @return   A {@link LegacyClass} instance.
    */
   public static object<? extends LegacyClass> getLegacyClass(String typeName)
   {
      Class<? extends _BaseObject_> cls = ObjectOps.resolveClass(typeName);
      
      if (cls == null)
      {
         ErrorManager.recordOrThrowError(15287, 
                                         "Could not dynamically find class'" + typeName + "'", 
                                         false, false);
         return new object<>();
      }
      
      registerClassWorker(typeName, cls);
      
      return getLegacyClass(cls);
   }

   /**
    * Get the {@link LegacyClass} mapped in FWD by the specified converted class name.
    * 
    * @param    cls
    *           The converted class name.
    *           
    * @return   A {@link LegacyClass} instance.
    */
   public static object<? extends LegacyClass> getLegacyClass(character cls)
   {
      return getLegacyClass(cls.isUnknown() ? "" : cls.getValue());
   }
   
   /**
    * Resolve the legacy name for this class.
    * 
    * @param    cls
    *           The class.
    *           
    * @return   The legacy name.
    */
   public static String getLegacyName(Class<? extends _BaseObject_> cls)
   {
      return legacyNameWorker(cls);
   }
   
   /**
    * Given a legacy class, get the pseudo-instance associated with the static members.
    * 
    * @param    cls
    *           The legacy class.
    * 
    * @return   The pseudo-instance for this class.
    */
   public static Object getStaticInstance(Class<? extends _BaseObject_> cls)
   {
      // enums are not tracked and have no static legacy state
      if (LegacyEnum.class.isAssignableFrom(cls))
      {
         // TODO: is there a better return value, this seems unsafe
         return null;
      }
      
      return local.get().staticReferents.get(cls);
   }
   
   /**
    * Given a static class pseudo-instance, get its associated legacy class.
    * 
    * @param    instance
    *           The pseudo-instance for a static class.
    * 
    * @return   The resolved legacy class.
    */
   public static Class<? extends _BaseObject_> getStaticClass(Object instance)
   {
      // enums are not tracked and have no static legacy state
      if (instance != null && LegacyEnum.class.isAssignableFrom(instance.getClass()))
      {
         // TODO: is there a better return value, this seems unsafe
         return null;
      }
      
      return local.get().referents2static.get(instance);
   }
   
   /**
    * Emulate an implicit static constructor.
    * 
    * @param    cls
    *           The legacy class for which the constructor will be executed.
    */
   static void defaultStaticConstructor(Class<?> cls)
   {
      // for enums, this is a NOP
      if (LegacyEnum.class.isAssignableFrom(cls))
      {
         return;
      }
      
      BlockManager.externalProcedure(cls, new Block());
   }

   /**
    * Load the given legacy class.  This will initialize it by recursively execute the static
    * c'tor for all super-classes (if defined).
    * 
    * @param    cls
    *           The class to initialize.
    *           
    * @return   <code>true</code> if the initialization completed.
    */
   static boolean loadClass(Class<? extends _BaseObject_> cls)
   {
      // enums are not tracked, legacy lifecycle processing is bypassed
      if (LegacyEnum.class.isAssignableFrom(cls))
      {
         return true;
      }
      
      WorkArea wa = local.get();
      // before trying to load this class, we need to backup (and restore) any state specified
      // for the invocation at the variables.  This includes the output parameter assigner
      OutputParameterAssigner opa = wa.tm.deregisterOutputParameterAssigner();
      try
      {
         return loadClassInt(cls);
      }
      finally
      {
         if (opa != null)
         {
            wa.tm.registerOutputParameterAssigner(opa);
         }
      }
   }
   
   /**
    * Perform recursive loading of all classes, up the hierarchy.
    * 
    * @param    cls
    *           The class to initialize.
    *           
    * @return   <code>true</code> if the initialization completed.
    */
   private static boolean loadClassInt(Class<? extends _BaseObject_> cls)
   {
      // enums are not tracked, legacy lifecycle processing is bypassed (we especially need to
      // avoid access to staticReferents and referents2static
      if (LegacyEnum.class.isAssignableFrom(cls))
      {
         return true;
      }
      
      // TODO: if first static c'tor fails, and second loading attempt doesn't, does 4GL allow
      // the class to load?
      
      WorkArea wa = local.get();
      if (wa.staticReferents.containsKey(cls) || wa.pendingLoadedClasses.contains(cls))
      {
         //  prevent recursive loading, if the static c'tor accesses static members
         return true;
      }
      
      wa.pendingLoadedClasses.add(cls);
      try
      {
         boolean parentLoaded = true;
         if (cls != BaseObject.class)
         {
            parentLoaded = loadClassInt((Class<? extends _BaseObject_>) cls.getSuperclass());
         }
         
         if (!parentLoaded)
         {
            return false;
         }
   
         String ctor = getConstructorMethod(cls, true);
         Object referent = new Object();
         wa.staticReferents.put(cls, referent);
         wa.referents2static.put(referent, cls);
         String name = getLegacyName(cls);

         // first load of a class causes its language to be applied, for good
         TranslationManager.getInstance().applyLanguage(cls);

         // execute the static c'tor
         if (!ControlFlowOps.initializeLegacyClass(cls, referent, name, ctor))
         {
            new ExternalProgramWrapper(referent).delete();
            wa.staticReferents.remove(cls);
            
            ErrorManager.recordOrThrowError(14631, 
                                            "Static instance failed to load. " +
                                            "Cannot reference class " + name,
                                            false, false);
            return false;
         }
         
         // if the class was loaded, then register it with the SourceNameMapper
         if (cls.getPackage().getName().startsWith("com.goldencode.p2j.oo"))
         {
            SourceNameMapper.registerLegacyClass(cls);
         }
         
         return true;
      }
      finally
      {
         wa.pendingLoadedClasses.remove(cls);
      }
   }

   /**
    * Resolve the associated converted class name for this legacy class.
    * 
    * @param    cname
    *           The fully qualified legacy class name.
    *           
    * @return   The found {@link Class} or <code>null</code> if it can't be found.
    */
   public static <T extends _BaseObject_> Class<T> resolveClass(String cname)
   {
      if (cname == null)
      {
         return null;
      }
      
      Class<T> type = (Class<T>) lookupClassWorker(cname);
      
      if (type != null)
      {
         return type;
      }
      
      String clsName = SourceNameMapper.convertOOToClass(cname);
      
      if (clsName == null)
      {
         return null;
      }
      
      try 
      {
         return (Class<T>) Class.forName(clsName);
      }
      catch (ClassNotFoundException ex)
      {
         return null;
      }
   }
   
   /**
    * Create a mapping between a given legacy class name and the converted Java class instance. 
    *  
    * @param    cname
    *           The fully qualified legacy class name.
    * @param    cls
    *           The converted type.
    */
   public static void registerClass(String cname, Class<? extends _BaseObject_> cls)
   {
      registerClassWorker(cname, cls);
   }
   
   /**
    * Create a mapping between a given legacy class name and the converted Java class instance. In this
    * the legacy class name will be looked up.
    * 
    * Complete class hierarchy is registered or not depending on the hierarchy flag.
    *  
    * @param    cls
    *           The converted type.
    * @param    hierarchy
    *           Flag to load the complete class hierarchy
    */
   public static void registerClass(Class<? extends _BaseObject_> cls, boolean hierarchy)
   {
      String clsName = legacyNameWorker(cls);

      if (clsName != null)
      {
         if (hierarchy)
         {
            if (cls.getSuperclass() != null && BaseObject.class.isAssignableFrom(cls.getSuperclass()))
            {
               registerClass((Class<? extends _BaseObject_>) cls.getSuperclass(), hierarchy);
            }

            for (Class i : cls.getInterfaces())
            {
               if (_BaseObject_.class.isAssignableFrom(i))
               {
                  registerClass(i, hierarchy);
               }
            }
         }
      }
   }
   
   /**
    * Resolve the {@link LegacyClass} of super class or interface of the class, register it as needed.
    * 
    * @param    superName
    *           name of assignable super class or interface
    * @param    cls
    *           The converted type.
    */
   public static Class<? extends _BaseObject_> getAssignableClass(String superName,
            Class<? extends _BaseObject_> cls)
   {
      String clsName = getLegacyName(cls);
      Class<? extends _BaseObject_> superCls = null;

      if (clsName != null && clsName.equalsIgnoreCase(superName))
      {
         registerClassWorker(superName, cls);
         return cls;
      }

      if (cls.getSuperclass() != null && BaseObject.class.isAssignableFrom(cls.getSuperclass()))
      {
         superCls = getAssignableClass(superName,
                  (Class<? extends _BaseObject_>) cls.getSuperclass());
      }

      if (superCls == null)
      {
         for (Class i : cls.getInterfaces())
         {
            if (_BaseObject_.class.isAssignableFrom(i))
            {
               superCls = getAssignableClass(superName, i);

               if (superCls != null)
                  break;
            }
         }

      }

      return superCls;

   }
   
   /**
    * Get the {@link LegacyClass} instance for the given type.
    * 
    * @param    cls
    *           The converted type.
    *           
    * @return   See above.
    */
   public static LegacyClass getClassInstance(Class<? extends _BaseObject_> cls)
   {
      WorkArea wa = local.get();
      LegacyClass lc = wa.legacyClasses.get(cls);
      
      if (lc == null)
      {
         
         LegacyResource lr = cls.getAnnotation(LegacyResource.class);
         String clsName = (lr != null ? lr.resource() : SourceNameMapper.convertJavaProg(cls));
         lc = new LegacyClass(clsName, cls);
   
         registerClassWorker(clsName, cls);
         loadClass(LegacyClass.class);
         newInstanceInternal(lc, LegacyClass.class, "Progress.Lang.Class", null);
         
         wa.legacyClasses.put(cls, lc);
      }
      
      return lc;
   }
   
   /**
    * Return the converted Java class instance for the given legacy class name. 
    *  
    * @param    cname
    *           The fully qualified legacy class name.
    *
    * @return   The converted Java class.
    */
   private static Class<? extends _BaseObject_> lookupClassWorker(String cname)
   {
      Class<? extends _BaseObject_> cls = null;
      
      synchronized (lock)
      {
         cls = name2cls.get(cname.toLowerCase());
      }
      
      return cls;
   }
   
   /**
    * Create a mapping between a given legacy class name and the converted Java class instance. 
    *  
    * @param    cname
    *           The fully qualified legacy class name.
    * @param    cls
    *           The converted type.
    */
   private static void registerClassWorker(String cname, Class<? extends _BaseObject_> cls)
   {
      if (cname == null || cls == null)
      {
         // something is wrong, this cannot work
         return;
      }
      
      synchronized (lock)
      {
         name2cls.putIfAbsent(cname.toLowerCase(), cls);
         legacyNames.putIfAbsent(cls, cname);
      }
   }
   
   /**
    * Resolve the legacy name for this class.
    * 
    * @param    cls
    *           The class.
    *           
    * @return   The legacy name.
    */
   private static String legacyNameWorker(Class<? extends _BaseObject_> cls)
   {
      if (cls == _BaseObject_.class)
      {
         cls = BaseObject.class;
      }
      
      String name = null;
      
      synchronized (lock)
      {
         name = legacyNames.get(cls);
         
         if (name == null)
         {
            if (cls.isAnnotationPresent(LegacyResource.class))
            {
               name = cls.getAnnotation(LegacyResource.class).resource();
            }
            else
            {
               name = SourceNameMapper.convertJavaProg(cls);
            }
            
            registerClassWorker(name, cls);
         }
      }
      
      return name;
   }
   
   /**
    * Resolve the constructor method name for this legacy class.
    * 
    * @param    cls
    *           The legacy class for which the constructor name must be resolved.
    *           
    * @return   The constructor names, under the form '__qualified_converted_name_constructor__'.
    */
   private static String getConstructorMethod(Class<?> cls)
   {
      return getConstructorMethod(cls, false);
   }
   
   /**
    * Resolve the constructor method name for this legacy class.
    * 
    * @param    cls
    *           The legacy class for which the constructor name must be resolved.
    * @param    staticDef
    *           Check if the static constructor name is required.
    *           
    * @return   The constructor names, under the form '__qualified_converted_name_constructor__'.
    */
   private static String getConstructorMethod(Class<?> cls, boolean staticDef)
   {
      // for enums, this really should never be called; this is just for safety purposes
      if (LegacyEnum.class.isAssignableFrom(cls))
      {
         // null seems unsafe so we return empty string
         return "";
      }
      
      String nm = cls.getName();
      if (nm.startsWith("com.goldencode.p2j.oo."))
      {
         nm = nm.substring("com.goldencode.p2j.oo".length() + 1);
      }
      else
      {
         nm = nm.substring(SourceNameMapper.getPackageRoot().length() + 1);
      }
      nm = nm.replace('.', '_');

      return String.format("__%s_constructor__" + (staticDef ? "static__" : ""), nm);
   }

   /**
    * Resolve the execute method names for this legacy class and its hierarchy.
    * 
    * @param    cls
    *           The legacy class for which the method name must be resolved.
    *           
    * @return   The method names, under the form '__qualified_converted_name_execute__'.
    */
   private static String[] getExecuteMethods(Class<?> cls)
   {
      // enums are not tracked and have no execute methods
      if (LegacyEnum.class.isAssignableFrom(cls))
      {
         return new String[0];
      }
      
      List<String> names = new ArrayList<>();
      while (cls != Object.class)
      {
         names.add(getExecuteMethod(cls));
         cls = cls.getSuperclass();
      }
      return names.toArray(new String[0]);
   }
   
   /**
    * Resolve the name of this legacy class converted 'execute' method.
    * 
    * @param    cls
    *           The legacy class for which the method name must be resolved.
    *           
    * @return   The method name, under the form '__qualified_converted_name_execute__'.
    */
   private static String getExecuteMethod(Class<?> cls)
   {
      // enums are not tracked and have no execute method
      if (LegacyEnum.class.isAssignableFrom(cls))
      {
         // null seems unsafe so we return empty string
         return "";
      }
      
      String nm = cls.getName();
      if (nm.startsWith("com.goldencode.p2j.oo."))
      {
         nm = nm.substring("com.goldencode.p2j.oo".length() + 1);
      }
      else
      {
         nm = nm.substring(SourceNameMapper.getPackageRoot().length() + 1);
      }
      nm = nm.replace('.', '_');

      return String.format("__%s_execute__", nm);
   }
   
   /**
    * Invoke the specified static method in the specified class.
    * 
    * @param     voidReturn
    *            Flag indicating that the method can return void.
    * @param     clsName
    *            The legacy class name.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   private static Object invoke(boolean   voidReturn,
                                String    clsName, 
                                String    mthdName, 
                                String    modes, 
                                Object... args)
   {
      if (mthdName == null)
      {
         ErrorManager.recordOrThrowError(15300, 
                                         "Could not evaluate method name parameter for Invoke",
                                         false);
         return new unknown();
      }

      Class<? extends _BaseObject_> cls = resolveClass(clsName);
      if (cls == null)
      {
         ErrorManager.recordOrThrowError(15285, 
                                         "A valid class instance or static class name is " +
                                         "required for dynamically invoking a method or " +
                                         "accessing a property",
                                         false);
         
         return new unknown();
      }

      if (!loadClass(cls))
      {
         return new unknown();
      }
      
      return ControlFlowOps.invokeLegacyMethod(voidReturn, 
                                               cls,
                                               clsName,
                                               mthdName,
                                               modes,
                                               args);
   }

   /**
    * Invoke the specified instance method in the specified object.
    * 
    * @param     voidReturn
    *            Flag indicating that the method can return void.
    * @param     ref
    *            The legacy object reference.
    * @param     mthdName
    *            The static method name.
    * @param     modes
    *            The argument modes.
    * @param     args
    *            The arguments.
    *            
    * @return    Any returned value, or unknown if the method is void.
    */
   private static Object invoke(boolean                        voidReturn,
                                object<? extends _BaseObject_> ref, 
                                String                         mthdName, 
                                String                         modes, 
                                Object...                      args)
   {
      if (mthdName == null)
      {
         ErrorManager.recordOrThrowError(15300, 
                                         "Could not evaluate method name parameter for Invoke",
                                         false);
         return new unknown();
      }
      
      if (ref.isUnknown())
      {
         ErrorManager.recordOrThrowError(15285, 
                                         "A valid class instance or static class name is " +
                                         "required for dynamically invoking a method or " +
                                         "accessing a property",
                                         false);

         return new unknown();
      }
      

      // TODO: invoking methods from builtin classes in some cases raises a
      // "Dynamic access to certain built-in methods and properties is not permitted (17008)"
      
      String clsName = getLegacyName(ref.type());
      return ControlFlowOps.invokeLegacyMethod(voidReturn, 
                                               ref.ref(), 
                                               clsName, 
                                               mthdName,
                                               modes,
                                               args);
   }

   /**
    * Generate the invalid type error for ToObject.
    */
   private static void generateInvalidTypeDE()
   {
      ErrorManager.recordOrThrowError(18007, String.format(DE_SPEC, "enum-type-name"));
   }
   
   /** 
    * Generate the invalid value error for ToObject.
    *
    * @param    cls
    *           The enum for which the value is invalid.
    */                              
   private static void generateInvalidValueDE(Class<?> cls)
   {
      ErrorManager.recordOrThrowError(18007, String.format(DE_SPEC, "member-name-or-value"));
   }
      
   /**
    * Context-local data.
    */
   private static class WorkArea
   implements Scopeable
   {
      /** Helper to use the ProcedureManager without any context local lookups. */
      private final ProcedureManager.ProcedureHelper pm = ProcedureManager.getProcedureHelper();

      /** Helper to use the TM without any context local lookups. */
      private final TransactionManager.TransactionHelper tm = TransactionManager.getTransactionHelper();

      /** The set of classes pending to be loaded. */
      private Set<Class<? extends _BaseObject_>> pendingLoadedClasses = Collections.newSetFromMap(new IdentityHashMap<>());

      /** Map of legacy objects to their pseudo-resource. */
      private Map<_BaseObject_, ObjectResource> objects = new IdentityHashMap<>();
      
      /** Pending object definitions which need to be registered in the next scope. */
      private Set<Object> pendingObjects = Collections.newSetFromMap(new IdentityHashMap<>()); 

      /** Variable definitions which are top-level block members. */
      private Deque<Set<Object>> definitions = new ArrayDeque<>();
      
      /** Map of variable definitions which are class members. */
      private Map<Object, Set<Object>> extProgDefinitions = new IdentityHashMap<>();
      
      /**
       * Bogus referents associated with a static class.  These are required to be able to track
       * this as a 'persistent procedure' which will never get deleted.
       */
      private Map<Class<? extends _BaseObject_>, Object> staticReferents = new IdentityHashMap<>();
      
      /** Bogus referents associated with a static class. */
      private Map<Object, Class<? extends _BaseObject_>> referents2static = new IdentityHashMap<>();
      
      /**
       * The set of newly created objects which wait to be assigned.  These will be deleted on
       * the next block's scope finish, if they are not yet assigned.
       */
      private Deque<Set<_BaseObject_>> pendingAssigned = new ArrayDeque<>();
      
      /**
       * The set of newly created objects which wait to be initialized.  These will be deleted
       * when the object initialization completes.
       */
      private Deque<Set<_BaseObject_>> pendingInitialized = new ArrayDeque<>();
      
      /** Indicate if a class has a destructor or not. */
      private Map<String, Boolean> clsWithDtor = new HashMap<>();
      
      /** The map of instantiated {@link LegacyClass} objects. */
      private Map<Class<? extends _BaseObject_>, LegacyClass> legacyClasses = new IdentityHashMap<>();
      
      /**
       * Provides a notification that a new scope is about to be entered.
       * <p>
       * Used to create new collections where to register the variable definitions for that block.
       */
      @Override
      public void scopeStart()
      {
         BlockType bt = tm.getBlockType();
         
         if (bt.equals(BlockType.INTERNAL_PROC) ||
             bt.equals(BlockType.FUNCTION)      ||
             bt.equals(BlockType.TRIGGER)       ||
             bt.equals(BlockType.DATABASETRIGGER))
         {
            if (!pendingObjects.isEmpty())
            {
               Set<Object> refs = Collections.newSetFromMap(new IdentityHashMap<>());
               refs.addAll(pendingObjects);
               definitions.push(refs);
            }
            else
            {
               definitions.push(Collections.emptySet());
            }
            
            pendingObjects.clear();
         }
         else if (bt.equals(BlockType.EXTERNAL_PROC))
         {
            Object referent = pm._thisProcedure();
            extProgDefinitions.computeIfAbsent(referent, (k) -> Collections.emptySet());
            Set<Object> extProgDefs = extProgDefinitions.get(referent);
            
            if (!pendingObjects.isEmpty())
            {
               if (extProgDefs == Collections.emptySet())
               {
                  extProgDefs = Collections.newSetFromMap(new IdentityHashMap<>());
                  extProgDefinitions.put(referent, extProgDefs);
               }
               
               extProgDefs.addAll(pendingObjects);
               
               pendingObjects.clear();
            }
         }
         
         pendingAssigned.push(Collections.emptySet());
         pendingInitialized.push(Collections.emptySet());
      }

      /**
       * Provides a notification that a scope is about to be exited.
       * <p>
       * All variable definitions defined in this block will have the reference counter decremented
       * for the wrapped object.
       */
      @Override
      public void scopeFinished()
      {
         BlockType bt = tm.getBlockType();
         
         if (bt == BlockType.INTERNAL_PROC && (pm._thisProcedure() instanceof _BaseObject_))
         {
            // check if this is a c'tor and if it completed normally - if so, save its destructor
            String name = pm.currentInternalEntry();

            if (name.startsWith(SourceNameMapper.CTOR_PREFIX) && name.endsWith(SourceNameMapper.CONSTRUCTOR))
            {
               // TODO: this was a c'tor - if it completed OK, then push it so that it will be 
               // invoked on delete.
               boolean normalCall = true;
               
               if (normalCall)
               {
                  String dname = name.substring(0, name.length() - SourceNameMapper.CONSTRUCTOR.length());
                  dname = dname + SourceNameMapper.DESTRUCTOR;
                  
                  _BaseObject_ ref = (_BaseObject_) pm._thisProcedure();

                  if (hasDestructor(ref.getClass(), dname))
                  {
                     objects.get(ref).pushDestructor(dname);
                  }
               }
            }
         }
         
         if (bt.equals(BlockType.INTERNAL_PROC) ||
             bt.equals(BlockType.FUNCTION)      ||
             bt.equals(BlockType.TRIGGER)       ||
             bt.equals(BlockType.DATABASETRIGGER))
         {
            Set<Object> defs = definitions.pop();
            decrement(defs);
         }
         
         if (!pendingInitialized.isEmpty())
            pendingInitialized.pop();
            
         // this is done AFTER any decrement is performed - because in case of functions returning
         // an object, that will be marked as 'pending assigned' and the object will not be
         // deleted (but the reference will be decremented).
         if (!pendingAssigned.isEmpty())
         {
            Set<_BaseObject_> pending = pendingAssigned.peek();
            if (pending != null && !pending.isEmpty())
            {
               // 4GL deletes an object immediately if it was not assigned - we track them here,
               // and if when this scope is exit their reference counter is 0, we delete them.
               Object funcRet = BlockManager.getFunctionReturn();
               Set<_BaseObject_> funcRetVals = null;
               if (funcRet != null && 
                   (funcRet instanceof object || 
                    (funcRet.getClass().isArray() && 
                     object.class.isAssignableFrom(funcRet.getClass().getComponentType()))))
               {
                  funcRetVals = Collections.newSetFromMap(new IdentityHashMap<>());
                  if (funcRet instanceof object)
                  {
                     object<?> obj = (object<?>) funcRet;
                     
                     if (obj._isValid() && obj.isTracked())
                     {
                        funcRetVals.add(obj.ref());
                     }
                  }
                  else
                  {
                     for (int i = 0; i < Array.getLength(funcRet); i++)
                     {
                        object<?> obj = (object<?>) Array.get(funcRet, i);
                        
                        if (obj != null && obj._isValid() && obj.isTracked())
                        {
                           funcRetVals.add(obj.ref());
                        }
                     }
                  }
               }
               
               for (_BaseObject_ obj : pending)
               {
                  if (!obj.isTracked())
                  {
                     continue;
                  }
                  
                  ObjectResource res = objects.get(obj);
                  
                  if (res.getReferences() == 0)
                  {
                     if (funcRetVals == null || !funcRetVals.contains(obj))
                     {
                        res.deleteSilent();
                     }
                     else
                     {
                        // move this one level down
                        if (pendingAssigned.size() > 1)
                        {
                           pendingAssigned.pop();
                           pendingAssign(obj, false);
                           pendingAssigned.push(pending);
                        }
                     }
                  }
               }
            }
            pendingAssigned.pop();
         }
      }

      /**
       * Provides a notification that an external scope is about to be deleted.
       * <p>
       * All variables defined at this external program will have the reference counter decremented.
       */
      @Override
      public void scopeDeleted()
      {
         Object referent = pm.getProcessedProcedure();
         if (referent == null           || 
             pm._isPersistent(referent) || 
             (referent == pm._thisProcedure() && !tm.isExternalBlock()))
         {
            // if no referent or procedure is still persistent, do not clean up yet
            return;
         }

         Set<Object> defs = extProgDefinitions.remove(referent);
         if (defs != null)
         {
            decrement(defs);
         }
      }

      /**
       * Get the {@link ScopeId} for the instance.
       * 
       * @return   {@link ScopeId#OBJECT_OPS}.
       */
      @Override
      public ScopeId getScopeId()
      {
         return ScopeId.OBJECT_OPS;
      }
      
      /**
       * Check if a destructor with the given name is defined in the specified class.
       * 
       * @param    cls
       *           The class name.
       * @param    dname
       *           The destructor name.
       *           
       * @return   See above.
       */
      private boolean hasDestructor(Class<? extends _BaseObject_> cls, String dname)
      {
         String key = cls.getName() + "." + dname;
         return clsWithDtor.computeIfAbsent(key, (k) -> 
         {
            try
            {
               cls.getMethod(dname);
               return true;
            }
            catch (Throwable t)
            {
               return false;
            }
         });
      }
      
      /**
       * Decrement the references for all this objects.
       * 
       * @param    defs
       *           Scalar or array object variables.
       */
      private void decrement(Set<Object> defs)
      {
         for (Object o : defs)
         {
            if (o instanceof object)
            {
               object<? extends _BaseObject_> obj = (object<?>) o;
               if (!obj.isUnknown())
               {
                  obj.decrement();
               }
            }
            else if (o instanceof object[])
            {
               object<? extends _BaseObject_>[] objs = (object<?>[]) o;
               for (object<? extends _BaseObject_> obj : objs)
               {
                  if (!obj.isUnknown())
                  {
                     obj.decrement();
                  }
               }
               
               if (objs.length == 0)
               {
                  // this was never resized, so remove its type
                  ArrayAssigner.removeObjectType((object[]) objs);
               }
            }
            else
            {
               // TODO: error out?
            }
         }
      }
   }
}
