/*
** Module   : Scopeable.java
** Abstract : the interface for objects whose changes must be externally 
**            initialized at scope start and finalized at scope exit
**
** Copyright (c) 2005-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- -----------------------------Description-----------------------------
** 001 GES 20050817   @22162 Created initial version, providing an interface for objects needing 
**                           initialization and termination callbacks at the start and end of 
**                           every block entered and exited.
** 002 CA  20131013          Added scopeDeleted() API, which will be called when the instantiating 
**                           procedure where this scopeable instance was registered gets deleted.
** 003 CA  20220901          Refactored scope notification support: ScopeableFactory was removed, and the 
**                           registration is now specific to each type of scopeable.  For each case, the block
**                           will be registered for scope support (for that particular scopeable) only when
**                           the scopeable is 'active' (i.e. unnamed streams or accumulators are used).  This
**                           allows a lazy registration of scopeables, to avoid the unnecessary overhead of
**                           processing all the scopeables for each and every block.
*/
/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.util;

/**
 * An interface for objects needing initialization and termination callbacks
 * at the start and end of every block entered and exited.  The exit is 
 * guaranteed to be called no matter how the flow of control exits the block
 * (exception, break, return or via the flow of control passing through the
 * natural end of the block). 
 */
public interface Scopeable
{
   /**
    * Provides a notification that a new scope is about to be entered.
    */
   public void scopeStart();
   
   /**
    * Provides a notification that a scope is about to be exited.
    */
   public void scopeFinished();
   
   /**
    * Provides a notification that an external scope is about to be deleted.
    */
   public void scopeDeleted();
   
   /**
    * Get the {@link ScopeId} for the instance.
    * 
    * @return   The scope ID, or <code>null</code> for cases when the scope notification is not done via
    *           {@link BlockDefinition#scopeList}.
    */
   public ScopeId getScopeId();
}