/*
** Module   : GenericFrame.java
** Abstract : implementation of UI frames using dynamic proxy feature
**
** Copyright (c) 2005-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------------Description-----------------------------------
** 001 NVS 20051026   @23332 Created initial version.
** 002 NVS 20060114   @24058 Extended hide() methods.
** 003 GES 20060115   @23919 Fixes to getters/setters and view.
** 004 NVS 20060116   @24059 hide() and view() methods change visibility
**                           flag in frame config instance.
** 005 GES 20060116   @23946 Implemented conversion of 1-based to 0-based positions.
** 006 NVS 20060125   @24060 up() methods were calling down() in LT.
** 007 NVS 20060125   @24069 Removed all traces of DownFrame class.
** 008 SIY 20060125   @24117 Frame manager support code.
** 009 GES 20060203   @24260 Fixed getter for isXYZ() methods. Added
**                           widget init code to cache the ID. Rewrote
**                           ID and widget lookup methods to be simpler 
**                           and faster.  Added some new interfaces for
**                           ID and widget access.  Simplified the data
**                           storage approach. Rewrote call sites to
**                           use common, faster implementations.
** 010 GES 20060204   @24270 Added display/update/set/prompt-for methods.
** 011 GES 20060208   @24383 Modifications for redirected I/O and using
**                           the UI statements on streams instead of the
**                           terminal.
** 012 SIY 20060210   @24512 Added getSelection() method.
** 013 SIY 20060213   @24582 Fix for the DialogBox support.
** 014 SIY 20060216   @24609 Removed unnecessary nested scopes handling.
** 015 SIY 20060222   @24630 Minor change in CHOOSE handling.
** 016 NVS 20060223   @24665 Within the UPDATE processing, the internal
**                           block should not catch the END condition.
** 017 GES 20060223   @24684 Final redirected output support.
** 018 NVS 20060224   @24727 High level I/O operations pass their codes
**                           into the WaitList so that the client has this
**                           information. Used for END-ERROR handling.
** 019 NVS 20060224   @24732 Since update() creates its own inner block,
**                           infinite loop protection should be disabled
**                           twice - on entry into update() and in the
**                           inner block, which is done in LT.waitFor().
** 020 GES 20060226   @24759 Added support for report headers/footers by
**                           honoring page-top and page-bottom attributes.
** 021 SIY 20060226   @24766 Widget arrays handling throws an exception
**                           if appropriate variable is undefined.
**                           Minor cleanups.
** 022 GES 20060228   @24777 Support prompt-for editing blocks.
** 023 GES 20060228   @24784 Support for a special behavior of the
**                           screen-value attribute and the input
**                           function.
** 024 GES 20060301   @24800 New feature to expose proper ASSIGN semantic
**                           in the case where the screen-buffer has no
**                           value for a given widget.
** 025 GES 20060303   @24877 Fixed error processing and missing error #.
** 026 NVS 20060306   @24905 WaitList does not need default events, they
**                           are recognized by ThinClient anyway.
** 027 NVS 20060308   @24940 ScreenBuffer is instantiated with state flags
** 028 NVS 20060309   @24963 Using a differential ScreenBuffer.
** 029 NVS 20060314   @25073 Resetting all widgets to the uninitialized
**                           state with every frame iteration.
** 030 SIY 20060316   @25132 Added method to retrieve widget name for the
**                           specified widget.
** 031 GES 20060317   @25084 Header support (includes some code written
**                           by SIY).
** 032 SIY 20060317   @23133 Merged support for delayed browse init.
** 033 SIY 20060318   @25134 Fixes for the header support.
** 034 NVS 20060321   @25177 Added getValidationSet() method.
** 035 GES 20060322   @25192 Added new signatures for title processing.
** 036 SIY 20060323   @25221 Fixed passing of widget values when list of
**                           widgets is provided explicitly.
** 037 GES 20060401   @25300 Added display sigs for the case where there
**                           are no data containing widgets in a frame.
** 038 SIY 20060401   @25306 Temporary workaround for the initial widget
**                           display.
** 039 GES 20060404   @25346 Moved enable/disable to a frame element
**                           approach.
** 040 SIY 20060404   @25356 Support for the text group.
** 041 SIY 20060409   @25434 Fixed handling of null items in the frame
**                           elements array.
** 042 GES 20060413   @25514 Added setScreenValue() method.
** 043 SIY 20060414   @25537 Added setVisible() methods.
** 044 SIY 20060415   @25549 Added enableExcept()/disableExcept() for 
**                           single widget.
** 045 GES 20060417   @25557 Rewrote setScreenValue() to handle all BDT
**                           types properly.
** 046 ECF 20060424   @25674 Added setScrollable() variant which accepts a
**                           logical.
** 047 SIY 20060424   @25685 Added getter/setter for parent and asWidget()
**                           method.
** 048 GES 20060426   @25731 Added underline() methods.
** 049 GES 20060427   @25820 Added setStatisticsColumn() method.
** 050 GES 20060428   @25868 Added accumulator support to display().
** 051 GES 20060502   @25931 Added possibleValidFrameId().
** 052 NVS 20060503   @25946 Repeating calls to setTitle have no effect.
**                           Title is set with the first call supplying a
**                           non-empty title string.
** 053 ECF 20060503   @25956 Completed first pass accumulator support for
**                           display aggregators.
** 054 NVS 20060504   @25978 setTitle() calls from the business logic are
**                           now called setDynamicTitle(). The original
**                           set is preserved for use in the frame
**                           creation logic.
** 055 ECF 20060504   @25980 Fix in AggregateHandler. List of full result
**                           aggregator elements was not being sorted.
** 056 GES 20060505   @25988 Detect when an error condition was caused by
**                           the END-ERROR key and rethrow the exception
**                           rather than retry in that case.
** 057 NVS 20060509   @26037 Implemented proper line spacing for totals
**                           in aggregated output.
** 058 SIY 20060508   @26038 Added notification of the client about
**                           opening of the frame scope.
** 059 GES 20060511   @26068 Reset the changed flags in setScreenBuffer()
**                           so that future getFrameBuffer() calls don't
**                           resend data as changed that has already been
**                           sent.  Widget state is modified on the 
**                           client when setting "changed" values and this 
**                           causes deviations from how Progress works.
** 060 GES 20060511   @26069 Fix for continueEditing() which was
**                           improperly disabling/copying back to the
**                           screen buffer on the first pass through the
**                           editing loop which doesn't have a pending GO. 
**                           Thus the logic was inverted when it should
**                           not have been.
** 061 NVS 20060511   @26089 Reimplemented conditional down() and up()
**                           calls which were simply short-circuited to
**                           down(1) and up(1) calls. Changed the 
**                           aggregated display accordingly. 
**                           Fixed scroll().
** 062 GES 20060517   @26195 Match new Finalizable interface requirement.
** 063 GES 20060523   @26422 Reworked internals to ensure that an abnormal
**                           exit from an editing block will still handle
**                           cleanup.
** 064 GES 20060523   @26450 SET/UPDATE/PROMPT-FOR must disable even when
**                           there is an abnormal exit.
** 065 GES 20060531   @26826 SET/UPDATE/PROMPT-FOR and editing blocks
**                           must force the focus to change to the first
**                           listed widget.
** 066 NVS 20060601   @26894 getFrameLine() now accounts for any pending
**                           conditional UP/DOWNs as it should.
** 067 SIY 20060601   @26915 promptFor() should clear messages at the end.
** 068 GES 20060612   @27120 Removed id list form of getFrameBuffer() as
**                           it should not be used because no matter what
**                           widgets are specified in a display, view,
**                           prompt-for, set or update, any changed data
**                           in the screen buffer is made visible on the
**                           client.
** 069 SIY 20060615   @27304 Clear frame buffer in clearAll().
** 070 GES 20060616   @27393 Clear frame buffer in clear().
** 071 SIY 20060622   @27536 Added some code for frame destroying mechanism
**                           (not functional yet).
**                           Frame name is stored in the config (can be
**                           overridden by frame initialization code).
** 072 GES 20060627   @27601 Added new setDown() signature to provide
**                           fully dynamic down frame control. Updated
**                           screen buffer processing to honor this new
**                           value.
** 073 NVS 20060627   @27608 Fixed all unconditional forms of up() and
**                           down(). Now they incorporate the function of
**                           conditionalUpDown() method but without making
**                           an extra trip to the client. Not only does it
**                           save some time, it fixes some inconsistency
**                           and eliminates unneeded pausing.
** 074 SIY 20060629   @27696 Fixed up() handling. Some refactoring for 
**                           better code transparency and maintainability.
** 075 GES 20060705   @27806 Moved setColors() from LogicalTerminal and
**                           added batching and conditionalUpDown.
** 076 SIY 20060705   @27838 Removed old workaround in scroll() which is
**                           not necessary anymore and causes regressions.
** 077 GES 20060707   @27906 Overhaul of setDynamicTitle to support a
**                           full callback model. The callback is stored
**                           and the resulting text is included in the
**                           screen buffer every time it is sent down to
**                           the thin client.
** 078 SIY 20060717   @28083 Call batch(close) after setUp().
** 079 GES 20060728   @28218 Moved conditional up/down processing to
**                           a state synch approach to avoid an extra
**                           round trip to the client for each view()
**                           or other language statement that must
**                           honor such things.
** 080 SIY 20060726   @28234 Clear the frame state when it goes of of 
**                           scope. Refactored frame hiding algorithm. 
** 081 GES 20060731   @28237 Minor changes to enable deferring of the
**                           destroyFrame() call.
** 082 SIY 20060731   @28250 Frame clearing now maintained more carefully.
** 083 SIY 20060801   @28334 Clear pending up/down on retry.
** 084 GES 20060803   @28378 Removed unnecessary cast to ComponentEntity.
** 085 GES 20060804   @28429 Shifted the aggregation handler's iterate
**                           processing to allow for the fact that iterate
**                           is only called at the top of the block. The
**                           previous version depended on being called at
**                           the bottom of the loop.
** 086 SIY 20060807   @28451 Changes in resetting of the frame.
** 087 GES 20060809   @28498 Added proper temporary terminal redirection
**                           to the up/down methods which take an explicit
**                           stream parameter.
** 088 SIY 20060810   @28517 Destroy frame when leaving current procedure
**                           or function, not only execute().
** 089 SIY 20060810   @28536 Partially rolled back #28517 to fix 
**                           regressions.
** 090 GES 20060816   @28626 Fix class cast problem by properly 
**                           dereferencing a stream's remote ID (one of
**                           the 2 cases was not being handled properly).
** 091 SIY 20060823   @28761 Added support for displaying static widgets
**                           in SET/UPDATE/PROMPT-FOR.
** 092 SIY 20060825   @28894 Protect from frame buffer resetting if frame 
**                           is not displayed yet.
** 093 NVS 20060830   @29021 Consolidated enable() and view() functions in
**                           the enable() method. No separate view() call
**                           is needed after call to enable().
** 094 GES 20060908   @29373 Avoid assignment of static widget values in
**                           assignOnly and widgetOnly cases.
** 095 GES 20060908   @29622 null as a stream reference must be treated
**                           as the terminal (remote stream id == -1).
** 096 GES 20060915   @29623 Honor overridden format strings for a given
**                           value to be placed in the screen buffer.
** 097 GES 20060918   @29663 Database fields now have J2SE-compatible
**                           variable names which is incompatible with the
**                           simplistic approach of "always decapitalize" 
**                           (the first char of a variable name). Now we 
**                           try the Java beans approach if the simplistic
**                           approach fails.
** 098 GES 20060919   @29687 Conversion is needed for cases where @ base
**                           field data is read from the screen buffer
**                           into a BDT of a different type.
** 099 GES 20060919   @29692 Data type conversion is needed in the
**                           assignScreenValue() method.  Now we use the
**                           common getter() code to handle this.
** 100 GES 20060920   @29721 Properly handle unknown literals as data
**                           type overrides in getter().
** 101 GES 20060926   @30025 Fixed TEXT group handling.
** 102 NVS 20061005   @30166 Added a full implementation of FRAME-VALUE
**                           statement as setFrameValue() method.
** 103 NVS 20061006   @30202 Added backing code for retrieving original
**                           widget names required by FRAME-FIELD.
** 104 GES 20061010   @30261 Implemented full support for shared frames.
** 105 GES 20061031   @30825 Safety code to ensure that in any failure
**                           a temporary terminal redirection is removed
**                           before continuing.  This gives us a cleaner
**                           abend and may fix some issues in the case
**                           where a failure is really a Progress 
**                           condition like ENDKEY or ERROR.
** 106 GES 20061031   @30830 Safety code to translate unknown to the
**                           proper type before updating the screen
**                           buffer.
** 107 GES 20061101   @30891 Deregister headers at end of block to which
**                           the frame is scoped.
** 108 SIY 20061102   @30899 Do not add Skip if there are no widgets in
**                           the frame body.
** 109 GES 20061103   @30930 Don't override visibility flag of a shared
**                           frame when we import that frame.
** 110 SIY 20061108   @31038 Support for special case of accessing array 
**                           widgets.
** 111 GES 20061120   @31294 Updated for support of dynamic row/column 
**                           and down.
** 112 GES 20061127   @31409 Added a shared flag to destroyFrame() to
**                           ensure that more deeply nested frames don't
**                           exit scope too soon.
** 113 SIY 20061128   @31459 Added scope nesting counter and excluded 
**                           nested scopes of shared frames from regular
**                           scope processing. 
** 114 NVS 20061129   @31473 Frame entering top level scope signals the
**                           fact to the client. Scope nesting level was
**                           decremented in a wrong place, fixed.
** 115 GES 20061206   @31595 Use new promptFor() helper.
** 116 GES 20061212   @31664 Eliminated most switchOutput() bracketing by 
**                           using new client exports that implement the
**                           temporary redirection in-line on the client
**                           rather than driving it via separate calls
**                           from the server.  This should have a major
**                           performance improvement for some forms of
**                           redirected output generation.
** 117 SIY 20061225   @31747 Renamed setUp() to setup(). 
** 118 SIY 20070116   @31843 Changed location of registration in TM for 
**                           top level finalization processing.
** 119 SIY 20070117   @31848 New frame ID assignment scheme.
** 120 NVS 20070118   @31887 Using WidgetId class and new trigger registry
**                           management based on EventList class. WaitList
**                           class is no longer in use.
** 121 SIY 20070118   @31899 More changes in finalization and shared
**                           frame handling.
** 122 SIY 20070124   @31972 Removed redundant registration of frame in TM.
** 123 SIY 20070306   @32282 Do not deregister frame in LogicalTerminal. 
** 124 SIY 20070314   @32399 Finish setup of browse widgets even in 
**                           widget-specific enable() call.
** 125 SIY 20070329   @32658 Fixed order of IDs returned by remainingIds().
** 126 SIY 20070409   @32874 Do not call down() in iterate() for regular 
**                           frames.
** 127 NVS 20070502   @33358 Made one of the getter() methods package
**                           private from private as it's used from
**                           LogicalTerminal.
** 128 NVS 20070507   @33418 Establishing proper connection between this
**                           instance of GenericFrame and an instance of
**                           FrameWidget.
** 129 SIY 20070521   @33751 Call conditionalUpDown() before calling 
**                           LogicalTerminal.refreshFrameWidget().
**                           Added signalling of closing scope.
** 130 SIY 20070601   @33925 Added display(BrowseWidget).
** 131 SIY 20070711   @34428 Fixed shared frame support.
** 132 NVS 20070716   @34537 Rerouted methods using unnamed streams to
**                           call matching methods using named streams.
**                           Added idsFromWidgets() method.
** 133 NVS 20070724   @34676 NOT THE FINAL CHANGE. All view(), display()
**                           calls pass the flag that tells the nature of
**                           the call to the workers. 
** 134 NVS 20070801   @34750 An attempt to fix conditional UP and DOWN in
**                           redirected mode.
** 135 NVS 20070807   @34797 DOWN for a non-down frame in redirected mode
**                           is a meaningful operation. It flushes the
**                           terminal buffer and inserts the required
**                           number of new lines into the stream.
** 136 NVS 20070808   @34818 Fixed display of page-top frames, which has
**                           been regressed recently.
** 137 NVS 20070808   @34822 display() call without parameters translates
**                           display(empty array) even in redirected mode.
** 138 NVS 20070817   @34879 Fixed display of page element frames without
**                           frame elements.
** 139 NVS 20070829   @34979 Added private method isRedirected(). Reworked
**                           implementations of up() and down() families
**                           of methods.
** 140 NVS 20070907   @35005 Fixed a bug in downWorker which affected up()
**                           method.
** 141 SIY 20070913   @35064 Fixed handling of LASTKEY when reading data
**                           from stream (redirected command/file).
**                           Ref: 35057-35062,35065
** 142 SIY 20070924   @35196 Support for new line feeds handling approach.
**                           Ref: 35198
** 143 NVS 20070925   @35208 Conditional down in redirected mode is not
**                           delayed only if the target frame is realized.
** 144 BRJ 20071003   @35360 Protected stopEditingMode from being called
**                           when startEditingMode was aborted due to no
**                           widgets being present.
** 145 NVS 20071102   @35764 Conditional down in interactive mode issued
**                           right after a redirected display for the same
**                           frame acts like a redirected down.
** 146 SIY 20071121   @35931 Support for Frame 'need pause' flag 
**                           maintenance.
**                           Ref: #35928
** 147 CA  20071122   @35949 Added setRetain(int) implementation. 
**                           Added RETAIN n and SCROLL n behavior for the
**                           down/up statements in method 
**                           downWorker(int,boolean). 
**                           Modified up(int) method to call method
**                           downWorker(int,boolean).      
** 148 CA  20071126   @36015 Replaced all (redirOut != null) conditions 
**                           with (redirOut!=null && redirOut.isOut()) 
**                           for the field GenericFrame.redirOut.      
** 149 SIY 20071126   @36038 Fix for up() handling in DOWN frames without 
**                           RETAIN and SCROLL options. 
**                           Some refactoring and cleanup.
**                           Ref: #36034
** 150 CA  20071129   @36077 Added the Progress caching behavior (for down 
**                           frames) for conditional up/down statements.
**                           Ref: #36076
** 151 NVS 20071203   @36138 Uncommented a condition in a down() call in
**                           iterate() which has been commented out by
**                           mistake
** 152 NVS 20071206   @36232 Aggregation handler is no longer scoped to
**                           the nearest loop. but rather to the scope of
**                           the frame it refers to. This is done by
**                           calling its iterate() and finished() methods
**                           from those methods called for the frame.
** 153 NVS 20071210   @36298 Moved call to finished() of aggregation
**                           handler before doing the rest of cleanup.
**                           Added missing down() call to compensate for
**                           the refactoring of iterate() method.
** 154 GES 20071213   @36327 Converted screen buffer caching to use
**                           Integer keys instead of integer (this is
**                           a bit lighter weight and allows ints to be
**                           used everywhere through autoboxing). Switch
**                           screen buffers in setScreenBuffer() if the
**                           frame-line changed during the last trip to
**                           the client.
** 155 GES 20071214   @36366 Don't swap screen buffers (setScreenBuffer)
**                           for non-down frames or when redirected, as
**                           no such swapping happens in the 4GL.
** 156 GES 20071214   @36374 Give the caller control over whether to allow
**                           screen buffer switching in setScreenBuffer.
** 157 GES 20071219   @36478 Honor dynamic down value in scroll to fix
**                           cache corruption.
** 158 SIY 20071219   @36512 Fixed displaying of totals.
**                           Ref: #36508
** 159 SIY 20071221   @36558 Fixed support of multiple accumulators in 
**                           some complex cases (conditional elements in 
**                           elements array).
**                           Ref: #36556
** 160 SIY 20071224   @36566 Do not produce accumulator results for
**                           sub-only accumulators.
**                           Ref: #36564
** 161 NVS 20071218   @36460 iterate() and retry() methods both should
**                           reset 1 down frames as well as down frames.
** 162 CA  20080104   @36670 DOWN statements in redirected mode (named or
**                           unnamed) must end any previous put output.
** 163 NVS 20080108   @36728 Only unconditional DOWNs flush any previous
**                           PUT output. Conditional DOWNs don't.
** 164 CA  20080110   @36781 Frames registered as a header inside an 
**                           iteration block must de-register themselves  
**                           from the stream's header list when the next  
**                           iteration is performed.
** 165 CA  20080115   @36814 The de-registration of a page-top/bottom frame
**                           must be done before the iterate() method is
**                           executed for that frame - it was moved to
**                           LogicalTerminal.ScopeProcessor.process. Access 
**                           modifier for some methods was changed from
**                           private to package.
** 166 CA  20080124   @36896 When a frame opens or closes a scope, this will
**                           be signaled to the client regardless of the
**                           scope nesting level. Ref: #36895
** 167 CA  20080219   @37128 When a DISPLAY ACCUM stmt accumulates its
**                           data (in AggregateHandler.accumulate), the 
**                           accumulated data must be retrieved from the
**                           AggregatorElement passed as a parameter - this
**                           is because, if a new complex expression is
**                           instantiated on each iteration, then the 
**                           accumulation would be done only with the first
**                           data (the one used to instantiate the first
**                           AggregatorElement, on the first iteration).
** 168 SIY 20080226   @37167 Editing block should handle conditional 
**                           up/down identically to regular update()/set()
**                           calls.
**                           Ref: #37166.
** 169 NVS 20080307   @37410 Added isBatch() which returns the batching
**                           mode status.
** 170 SIY 20080317   @37457 Added support for terminal stream.
**                           Ref: #37453
** 171 GES 20080317   @37591 Added FIRST-CHILD and TYPE attributes.
** 172 SIY 20080320   @37599 Added support for widgets which can have 
**                           screen values different than data value.
** 173 NVS 20080320   @37621 Changed LogicalTerminal.enable() call to
**                           match the new signature.
** 174 NVS 20080326   @37687 startEditingMode() method calls promptFor()
**                           client export since its functionality is now
**                           part of the modified promptFor().
** 175 SVL 20080402   @37787 updateDynamicTitle() call is removed from
**                           setDynamicTitle() method.
** 176 SIY 20080429   @38134 Ignore down() requests in interactive mode
**                           until frame is displayed.
**                           Ref: #38137
** 177 SIY 20080429   @38139 Preserve widget colors during setColors() in
**                           DOWN frames.
** 178 GES 20080429   @38146 Defer the push of screen definitions to do
**                           them in a batch instead of making a trip to
**                           the client for each one. This improves
**                           performance with no negative side effects.
** 179 SIY 20080430   @38155 Interactive PROMPT-FOR should set hadView 
**                           because it contains implicit VIEW.
** 180 GES 20080501   @38187 Merged endOutput calls into other client
**                           calls to avoid the extra trip.
** 181 NVS 20080507   @38231 Fixed makeFrameVisible(). Making the frame
**                           visible should follow making widgets visible
**                           or else browse, if it's in the frame, will
**                           cause pushScreenDefinition() and the frame
**                           will be made visible prematurely, which hides
**                           the frame title and may have other issues.
** 182 SIY 20080514   @38275 Push shared frame definition to client when 
**                           current incarnation finishes its life cycle.
** 183 SIY 20080604   @38546 Added accessor for frame name. Inferred 
**                           generic type arguments. 
** 184 SIY 20080609   @38648 Added accessor for hadView flag. 
** 185 NVS 20080516   @38791 Code cleanup in startEditingMode() method.
** 186 NVS 20080711   @39092 importSharedFrame() forces redrawing of
**                           dynamic title if it was assigned. Ignoring
**                           this was the root cause of the disappearing
**                           title.
** 187 SIY 20080715   @39111 Visible shared frame does not need to be 
**                           displayed again.
** 188 SIY 20080904   @39721 Inferred generics in frame creation methods.
** 189 CA  20080905   @39742 Added missing getter for VISIBLE attribute.
**                           When modifying the COL/ROW attributes of this 
**                           frame, send a flag to the client if the 
**                           location was changed.
** 190 CA  20080909   @39751 frameLineWorker will send the conditional
**                           down status to the client. This way, if the
**                           frame is hidden, FRAME-LINE will return 0; 
**                           else, the conditional down value will be 
**                           used by the client.
** 191 SIY 20080926   @39965 Track frame initialization and ignore invalid
**                           ROW/COL setting.
** 192 NVS 20081002   @39995 Fixed handling of interactive conditional
**                           DOWNs when previous output was redirected. 
** 193 CA  20080925   @39963 Support for dynamic browse title - the title
**                           expression is evaluated only the first time
**                           the buffer is sent to the client. Added new
**                           setDynamicTitle variants, which accept a BDT
**                           parameter.
** 194 CA  20080930   @39987 Added accessors for FRAME:WIDTH/HEIGHT-CHARS
**                           and FRAME-COL/ROW attributes.
** 195 CA  20081003   @40054 Fix the DISPLAY with ACCUM statement, when
**                           no accumulation is performed.
** 196 CA  20081017   @40138 The cached buffer is not sent to the client,
**                           if the frame had redirected output.
** 197 GES 20081106   @40358 On abnormal end in an editing block, don't
**                           copy changes back from the screen buffer.
** 198 SIY 20081111   @40413 Improved detection of abnormal end in editing
**                           block.
** 199 GES 20081112   @40414 Ensure that client side cleanup always gets
**                           called when an editing block exits via a
**                           path other than continueEditing().
** 200 SVL 20081107   @40392 Changes in frameLineWorker() due to API
**                           change. Changes in cache row calculation into
**                           up()/down().
** 201 CA  20081127   @40702 On scope enter, the frame must send a flag to
**                           the client, indicating that there is a parent
**                           block iterated at least once.
** 202 SIY 20081209   @40702 Reworked frame hiding implementation.
** 203 GES 20081210   @40863 Implemented a quirk of the UPDATE stmt which
**                           disables any accumulation of nested block's
**                           interactions counter. This can only happen
**                           in a trigger. Note that SET/PROMPT-FOR don't
**                           have or need this behavior.
** 204 NVS 20081211   @40874 The default field group is created right away
**                           having the frame as its parent.
**                           Fixed the NAME attribute getter for the frame
**                           as a widget.
** 205 SIY 20081226   @40995 Reset hadView flag in openScope(). 
** 206 SIY 20090122   @41194 Reset frame buffer in openScope(). 
** 207 SIY 20090127   @41231 Ignore display() calls for frame with no
**                           defined widgets.
** 208 SIY 20090206   @41266 Take into account presence of headers when  
**                           checking for empty display() call.
** 209 SIY 20090210   @41286 hadView now is set from single location (see
**                           LogicalTerminal.ScopeProcessor.frameViewed()).
**                           Added separate canHide flag which tracks
**                           frames' potential ability to hide other 
**                           frames during view() call.
** 210 SIY 20090213   @41308 display(BrowseWidget) now forces refresh.
** 211 SIY 20090220   @41341 Fixed handling of DOWN for stream redirected
**                           to interactive terminal. It should be handled
**                           just like regular interactive terminal.
** 212 CA  20090227   @41388 Added WrappedResources.isUnknown method; this
**                           implementation returns false.
** 213 ECF 20090312   @41535 Added safety code to updateDynamicTitle(). We
**                           now bracket the resolution of the title text
**                           with calls to enable/disable ErrorManager's
**                           warning mode to prevent erroring out if code
**                           invoked by this method raises an error
**                           condition.
** 214 SIY 20090316   @41615 Improved #213 in order to avoid generation of
**                           error messages during repeated frame views. 
** 215 SIY 20090407   @41730 Improved interaction of DOWN and UNDERLINE 
**                           in order to preserve correct cursor position
**                           when implicit and explicit down() calls are 
**                           mixed.
** 216 CA  20090413   @41781 Added handling for FRAME:VISIBLE/HIDDEN and 
**                           SENSITIVE flags.
** 217 CA  20090414   @41789 Added getHeaderWidgetForId, used by VISIBLE 
**                           flag change for header widgets. For shared 
**                           frames, the VISIBLE flag must be inherited
**                           from the original frame.
** 218 SIY 20090506   @42102 Honor users' attempt to override "unknown" date field.
** 219 SIY 20090602   @42547 Do not clear frame buffer upon retry() if 
**                           condition was observed.
** 220 GES 20090624   @42883 Protect streams access from closed streams.
** 221 SIY 20090701   @43015 Added server side part of format checking.
** 222 SIY 20090709   @43122 Fixed NPE in set().
** 223 SIY 20090716   @43204 Perform final check only for fields which were 
**                           entered during SET.
** 224 CA  20090727   @43392 When a new frame buffer is built, all static text
**                           widgets must be copied to the new frame buffer. 
** 225 SIY 20090817   @43685 Strip spaces from formatted index before using it
**                           for widget name generation.
** 226 GES 20090828   @43800 Make sure the disabling of ignore and warning
**                           modes always occurs even on exception.
** 227 SIY 20090915   @43915 Flush redirected frames in finish(), if they were  
**                           not flushed in iterate().
** 228 SIY 20090917   @43926 Retrieving values from screen buffer depends on 
**                           type of input operation. 
** 229 SIY 20091009   @44130 Changed order of checks in frameLineWorker(). 
**                           This allows significantly reduce number of 
**                           round-trips to client in certain cases.
** 230 SIY 20091117   @44403 Fixed NPE in display(FrameElement[]) for cases 
**                           when frame has no widgets and no headers (rare 
**                           case, most likely a bug in the 4GL code).
** 231 CA  20091209   @44494 The screenBufferCache must be cleared when the
**                           frame opens the first scope, as left-overs may
**                           may exist in cache.
** 233 SIY 20100316   @44749 Redirected frames should have conditional DOWN
**                           processed with their stream.
** 234 SVL 20100519   @44861 Mark the added skip as a header element.
** 235 SVL 20101004          Added support for WIDTH-PIXELS attribute.
** 236 SIY 20101004          Changes related to CONTROL-FRAME support.
** 237 SIY 20101005          Added 'UNLESS-HIDDEN' versions of enable() and
**                           display() methods (stubs).
** 238 SVL 20101006          Added support for HEIGHT, WIDTH, THREE-D, X, Y
**                           attributes and VALIDATE method.
** 239 ECF 20101006          Added asWidgetHandle method.
** 240 SIY 20101012          Added more variants for view() and hide() methods.
** 241 SIY 20101029          Changed setX/setY signatures.
** 242 SIY 20101122          Removed unused methods/variables, cleaned up 
**                           javadocs.
** 243 EVL 20101124          Implementing the approach for tracking reusage of
**                           the same internal data between file report output
**                           iteration. The objective is to fix the problems
**                           when P2J clean up the state and display value of
**                           the reusable FillIns. The change includes:
**                              - introducing the new boolean flag to store
**                                the state we want to consider
**                              - changing the flag vaue in view() method calls
**                                depending on the new incoming data presence.
**                              - bypass the screen buffer resetting in
**                                resetFrameBuf() for such cases and making
**                                cleanup in finished() method
**                              - downWorker() is modified to disable screen
**                                buffer reset and calling down() on the
**                                client side for this case
** 244 ECF 20101209          Replaced standard, J2SE dynamic proxy use with
**                           GCD proxy implementation in createFrameWorker.
**                           This required major refactoring of the code which
**                           creates standard and shared frames, and which
**                           imports shared frames, as well as the inner class
**                           SharedFrameCleaner. The new proxy implementation
**                           allows a superclass to be specified for the proxy
**                           class. This allows us to leverage many existing
**                           methods in GenericFrame, so that we do not need
**                           to re-implement these methods in every proxy.
**                           This is necessary to save PermGen memory, which
**                           otherwise would grow with each addition to the
**                           CommonFrame API, using the standard, J2SE dynamic
**                           proxy implementation. The increased PermGen for
**                           each new CommonFrame method would be multiplied
**                           by the total number of proxies created. With the
**                           new implementation, those methods are simply
**                           invoked as virtual methods in GenericFrame.
** 245 CA  20110314          Fixed SCREEN-VALUE behavior for logical widgets:
**                           the returned value must be the formatted 
**                           representation of the value currently held by the
**                           widget. Not confirmed in a 4GL environment.
** 246 ECF 20110502          Revised createFrameProxy() to proxy all of the
**                           given interface's superinterfaces, up to (but not
**                           including) CommonFrame. This is necessary for
**                           some hand-written frame code which uses a custom
**                           frame inheritance hierarchy.
** 247 SIY 20110112          Changes related to new color scheme.
** 248 SIY 20111101          Updated imports.
** 249 SIY 20111202          Updated imports.
** 250 CS  20121031          Removed client specific references
** 251 CA  20130221          Refactored usage of TYPE attribute.
** 252 EVL 20130222          Adding UI related attributes and methods.
** 253 CA  20130223          Added support for STREAM-IO, FONT and IN WINDOW clauses in 
**                           frame-phrase.
** 254 OM  20130304          Refactored isValid and isUnknown of WrappedResource to valid 
**                           and unknown.
** 255 CA  20130603          resourceDelete must return a boolean value.
** 256 OM  20130606          Added datetime/-tz support in setScreenValue().
** 257 CA  20130927          GenericFrame is not a legacy resource; when legacy resource is
**                           needed, asWidget() will be called, to get its compatible legacy
**                           resource.
** 258 SVL 20131203          Changes caused by renaming of Nameable.get/setName to Nameable.name.
** 259 CA  20131013          Adding support for implicit deletion of a static frame; not yet 
**                           complete, as the static frame actually gets deleted when HIDE is
**                           called for the frame and its INSTANTIATING-PROCEDURE attr is unknown.
** 260 VIG 20131204          Implemented display(.., hWin) methods,
**                           Added method to check if window handle is valid.
** 261 EVL 20131127          Adding support for ENABLE stop in batch mode.
** 262 EVL 20140115          Adding background and foreground differences support.
** 263 VIG 20140121          Added new method currentIteration() and field-groups processing.
** 264 EVL 20140514          Fix for widget underline when frame is not yet on the screen.  This
**                           case need the implicit down processing to be done as if the frame
**                           is on the screen.
** 265 CA  20140526          handle unwrap APIs were switched to instance method calls.
** 266 EVL 20140522          Adding implementations for UNLESS-HIDDEN option.  The statements
**                           involved: DISABLE, DISPLAY, ENABLE, PROMPT-FOR, SET, UPDATE.
** 267 VIG 20140510          Adding setFrameHandle() and getFrameHandle().           
** 268 EVL 20140619          Adding the code to validate widget realize action to the method
**                           copyToScreenBuffer().  The errors have cumulative effect and at the
**                           end of the realization the client side session should be stopped.
** 269 GES 20140627          Changed color attribute method names to match naming standards.
** 270 CA  20140701          Added support for font-related attributes: FONT, TITLE-FONT.
** 271 CA  20140711          Enhanced font support, for GUI clients.
** 272 EVL 20140801          Rework for widget ID usage approach.  Adding methods to support
**                           dynamic widgets.
** 273 HC  20140911          GUI coordinate support - data types refactoring.
** 274 EVL 20140926          Adding more functionality to support dynamic frames and widgets.
** 275 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 276 EVL 20141024          The scope opening for dynamic frames should be done as the final
**                           step of the dynamic frame creation.  Also for dynamic frames the ID
**                           in config file already has the correct value. Do not change it in
**                           coreInitialize() method.
** 277 VIG 20141025          Added multi-window implementation for: view, display, enable.
** 278 EVL 20141029          Fix for infinite loop in FieldGroup processing issue inside
**                           recursive updateFieldGroups() method call.  Literal field group
**                           should be restored.
** 279 MAG 20141014          Added [get/set]WidthChars and [get/set]HeightChars. inherited from
**                           Sizable API. Added set/get BGCOLOR and FGCOLOR
** 280 HC  20141021          Implemented window move/position support.
** 281 CA  20141106          Added window ID handling for PROMPT-FOR/UPDATE/SET statements.
** 282 HC  20141207          Implemented runtime support for widget coordinate attributes.
** 283 GES 20141124          Added scrollbar-vertical and v6frame options.
** 284 CA  20150105          Changes for SIDE-LABEL-HANDLE runtime support.
** 285 GES 20150111          Eliminated unnecessary logging.
** 286 GES 20150115          Rewrote getScreenValue() to always return the formatted version of
**                           the widget's value. The implementation is separated from the getter
**                           to ensure this does not cause regressions in unrelated code paths.
** 287 CA  20150123          Fixed handling of side-labels in shared frames.
** 288 GES 20150131          Delegate format string length calcs to the BDT type.
** 289 GES 20150204          Clear GO-PENDING when a NEXT statement is used in an EDITING block.
** 290 GES 20150207          Set the event list operation for editing blocks.
** 291 GES 20150212          Made convertScreenValue() public.
** 292 HC  20150221          Improvements to frame tab order.
** 293 CA  20150223          Fixed label's COLUMN and WIDTH-CHARS attributes (in ChUI).
** 294 HC  20150226          Improved frame definition dynamic COL/ROW behavior.
** 295 CA  20150224          If in EDITING mode, INPUT function behaves like the SCREEN-VALUE 
**                           attribute.
** 296 CA  20150310          Setting SCREEN-VALUE for logical fields must honor their format.
** 297 CA  20150323          If a widget's value is set while the frame setup is in progress, save
**                           it as the "frame definition value", in the config.frameDefValue field.
** 298 CA  20150402          If the name of a new shared frame collides with the name of an 
**                           existing frame, register a finalizable to restore the hidden frame
**                           once the scope of this new shared frame has finished.
** 299 SVL 20150331          Added special handling for browse column widgets.
** 300 HC  20150323          Removed business logic from config classes.
** 301 CA  20150406          Added implementation for frame's X and Y attribute getters.  Fixed
**                           the setters and getters for the HEIGHT/WIDTH-PIXELS attributes.
** 302 CA  20150416          Fixed server-side support for PARENT attribute.
**                           Fixed support for NEXT/PREV-SIBLING in relation with tree widget
**                           structures.
** 303 IAS 20150604          Added support for the ENABLE UNLESS-HIDDEN ALL EXCEPT clause
** 304 CA  20150722          Changed firstChild to return a handle instance, otherwise chained 
**                           FRAME:FIRST-CHILD:<attr> will not work.
**                           When there is a pending UP/DOWN in a down frame, the server-side's
**                           widget configs need to be in sync with the client-side, as AFTER the
**                           pending UP/DOWN was performed; thus, APIs like setColors, scroll,
**                           setFrameValue or setScreenValue need to have the widget configs
**                           updated before the main logic is executed - this is resolved by 
**                           forcing a UP/DOWN, before the main logic, if a conditional down is
**                           active.
**                           Browse column widget IDs must be saved at the BrowseConfig, so the 
**                           client-side is aware of them.
** 305 CA  20150724          For shared frames, the DOWN value is inherited from the master frame;
**                           any value set at the config is ignored!
** 306 CA  20150725          Realized frames can't have their dynamic down changed when their 
**                           scope is reopened (related to H305).
** 307 SVL 20150821          Fixed deletion of dynamic widgets.
** 308 EVL 20150821          Modification for getScreenValue() method to reflect selection-list
**                           specific call.  We need to have the ability to get the value outside
**                           the given format length in multiple selection.
** 309 IAS 20150922          Added support for the TOP-ONLY and OVERLAY attributes
** 310 SVL 20151013          In deleteDynamicWidget cleanup is performed before widget
**                           deregistration. Pushing screen definitions is optional for deleted
**                           child widgets.
** 311 OM  20150922          Added setters for NO-AUTO-VALIDATE, CANCEL-BUTTON attributes.
**                           Added disableUnlessHidden() method corresponding to DISABLE ALL
**                           UNLESS-HIDDEN Progress syntax. Replaced UnsupportedOperationException
**                           with UnimplementedFeature. Handled NPE when widget is not in frame. 
** 312  IAS 20151015          Added setters for CANCEL-BUTTON/DEFAULT-BUTTON options, removed
**                           setters for corresponding attributes (they are in the FrameWidget
**                           now). Added support for SCREEN-IO/STREAM-IO/USE-TEXT options and
**                           THREE-D attribute.
** 313 EVL 20151016          Explicit setting frame width or height by attribute should be noted
**                           in respective fixed(Width|Height) config flag.  Adding TODO entry to
**                           explain size settings in attributes different considerations between
**                           frames and other widgets.
** 314 IAS 20151030          Added support for the BACKGROUND attribute, 
**                           NO-AUTO-VALIDATE option and VALIDATE() method
**     IAS 20151106          Removed setters for the NO-HELP and NO-VALIDATE flags
** 316 CA  20151026          Moved getParentHandle() impl to FrameWidget class.
**                           Fixed VISIBLE,HIDDEN,DOWN attributes when set via the frame's handle.
**                           A fix for HIDDEN attribute: distinguish between HIDE statement and 
**                           setting the HIDDEN attribute to true.  Also, if the frame's or any
**                           of its parents is marked as HIDDEN, then DISPLAY is a no-op.
** 317 CA  20151123          Replaced root frame check with dedicated API.
**                           displayWorker will send all DISPLAY stmts to the client-side: this 
**                           will take care of not "view-ing" the parent frames, if the nested 
**                           frame is hidden.
** 318 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 319 OM  20160111          Fixed core initialization of FRAMEs with SHARED HEADER widgets.
**                           Fixed (typo) error number for 3138 condition.
**     CA  20151221          Added check if shared frame matches the master.
** 320 HC  20160203          Improved runtime support of FRAME:VIRTUAL* and FRAME:SCROLLABLE
**                           attributes. Frame layout improvements.
** 321 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 322 CA  20160303          Made attribute setters be a no-op if the value doesn't change.
** 323 OM  20160205          Matching frames conditions were incorrect. Added cleanup on error.
** 324 CA  20160315          Fixed H323 regression in initLiteralFG - an empty array needs to be
**                           passed to wrapWidgetsToFieldGroup, and not the current widget list.
** 325 CA  20160327          Fixed FRAME:X/Y attribute setters and getters.
** 326 EVL 20160513          Fix for dialog-box size getting issue.
** 327 OM  20160531          Reworked frame core initialisation to match the IDs of space widgets.
** 328 GES 20160616          Added getFrameBufferRaw() to allow callers to avoid the unnecessary
**                           header and dynamic expr (title, col, row...) evaluation.
** 329 HC  20160628          Javadoc fix.
** 330 EVL 20160621          Adding stub for CURRENT-ITERATION setter.
** 331 SVL 20160701          Background field group is created for dynamic frames too.
** 332 HC  20160714          Fixed nested frame not made visible when outer enabled. Fixed frame
**                           widgets being properly added and linked with the frame's primary
**                           field group.
** 332 EVL 20160722          Changed LogicalTerminal.view() parameter list to take screen buffer
**                           array.
** 333 IAS 20160829          displayAndDown() operation support
** 334 CA  20160915          Fixed deletion of frames when they contain static frames or dynamic
**                           widgets.
** 335 EVL 20160914          In addition to space and skip the browse column widget can not be
**                           the part of the field group too.
** 336 CA  20160930          SCREEN-VALUE set/get and initial value can behave different for 
**                           each widget.
** 337 CA  20161014          Unified the setters for the FRAME's width/height attributes.
** 338 HC  20161024          An NPE prevented in enable().
** 339 OM  20161111          Implemented SHARED FRAMEs hierarchy.
**         20161205          Cleanup initialization process of imported shared frames.
** 340 CA  20161123          Centralized the logic for the COL/ROW/X/Y setters in GenericFrame, to
**                           use the APIs from FrameWidget.
** 341 OM  20161205          Cleanup initialization process of imported shared frames.
** 342 CA  20170130          Fixed widgetListFromFrameElements to ignore bypass elements.
**                           Fixed FRAME:BGCOLOR setter.
**     CA  20161101          Fixed X/Y attribute setters for dynamic widgets, which have their 
**                           FRAME set after these attributes are set.
** 343 CA  20170320          Removed the color save/restoration  code - it's no longer useful,
**                           and it breaks down frames after the improvements for down frame
**                           configs.
** 344 EVL 20170530          Adding DROP-TARGET frame option support.
** 345 CA  20170914          Fixed getOriginalName() for frame widgets.
** 346 SBI 20161101          Added hasFieldGroup test method.
** 347 CA  20171006          For imported shared frames, HELP and VALEXP are local to that usage 
**                           and don't affect master frame.
** 348 SBI 20171016          Changed to create the default field group for dynamic and static 
**                           frames uniformly.
** 349 CA  20171111          Added WIDGET:BGCOLOR-RGB and WIDGET:FGCOLOR-RGB FWD extension
**                           attributes
**     SBI 20171120          Exposes toTop/toBottom methods to move an owned widget to the top or
**                           to the bottom of its widget list.
** 350 GES 20171220          Added a TODO comment.
** 351 CA  20171128          Added support for DISPLAY UNLESS-HIDDEN with empty widget list.
**                           Also, if the frame is HIDDEN, then it will be a no-op.
**     OM  20171126          Updated local cache when a widget SCREEN-VALUE is updated.
**     SVL 20171204          Replaced forceRefresh call with refreshCurrentRow call.
** 352 CA  20180130          Replaced pushScreenDefinition() with pushWidgetAttr() for cases when
**                           business logic is assigning a widget attribute.
** 353 ECF 20180102          Use ConfigManager instance from WidgetConfig instead of doing
**                           context-local lookup.
** 354 HC  20180223          Improved assignment of SCREEN-VALUE value, the value may contain
**                           fill characters according to the widget format and data type.
** 355 SVL 20180301          Do not show browse columns on ENABLE.
** 356 SBI 20180503          Removed validationLayout(widgets).
** 357 GES 20180508          getHWND() now returns the frameId instead of unknown value.
** 358 CA  20180501          Added SIDE-LABELS attribute.
**     CA  20180502          Fixed COLOR option, COLOR statement and other color-related issues.
**     CA  20180514          Added CAREFUL-PAINT.
**     CA  20180525          Fixed NPE in removeHiddenWidgets.
**     CA  20180527          Honor widget's format in convertScreenValue. TODO: not all BDTs 
**                           accept the format as second argument for the c'tor.
**     CA  20180528          Added UNLESS-HIDDEN to the UPDATE/SET/PROMPT-FOR ... EDITING blocks.
**     CA  20180529          Accessing frame's dimension or location attributes, or changing the
**                           frame's location forces the frame to be placed.
**     CA  20180530          Fixed unknown and empty FRAME:TITLE.
** 359 EVL 20180621          Fix for empty frame report generation.
** 360 HC  20180711          Fixed ENABLE statement not showing up the frame's parent window.
** 361 OM  20180731          Added empty-equivalent SCREEN-VALUE quirk for date-related widgets.
** 362 SVL 20180813          finishSetup() for widgets is called on view/enable of the parent
**                           frame *before* this frame is marked as visible. The reason is that
**                           screen definitions may be pushed in finishSetup() and if that happens
**                           after the frame is marked as visible, frame definitions are pushed
**                           too and in the follow-up call of TC.viewWorker() the frame is not
**                           recognized to be changing its state (invisible -> visible).
**     EVL 20180814          Fix the cases when widget does not have the format string.
** 363 SBI 20181031          Changed getter to get the screen value from the widget.
**     EVL 20181130          Undo for the previous change due to serious regressions in ChUI.
**                           Need to implement some better approach.
** 364 SBI 20190117          Fixed sendWidgetScreenValue to check FrameElement implements get().
** 365 SBI 20190207          Changed display(BrowseWidget browse) in order to eliminate browse
**                           refreshing if there is an invisible frame containing this browse.
** 366 HC  20190226          FieldGroup must be dynamic when its parent frame is dynamic, too.
**     HC  20190311          Fixed parsing of decimal screen value.
**     SVL 20190317          Added cleanNonTabItems.
** 367 SBI 20190510          Fixed addDynamicWidget by adding a hit test.
** 368 EVL 20190809          When setting new parent for widget is is possible to add dynamic
**                           widget that was already added previously.  In this case we need to
**                           consider this case in addDynamicWidget() method.
** 369 SVL 20191031          Change in parseScreenValue in order to make setScreenValue set
**                           a character parameter by value, not by reference.
**     SBI 20191206          Fixed registerHeader in order to collect widgets for the frame header.
**     EVL 20191206          Fix for StackOverflow caused by recursive frame delete-hide calls.
**                           The deletion causes frame hide which in turn calls frame delete. This
**                           continues until stack space exhausting.
** 370 HC  20200313          Javadoc fixes.
**     GES 20200221          Delegated NEXT-TAB-ITEM/PREV-TAB-ITEM. to the frame widget.
**     RFB 20200326          Finished off TitledElement implementations for frames.
**     RFB 20200327          Moved real implementations of TitledElements to FrameWidget, and just
**                           invoke those methods from here. Also, no getters needed here.
**                           Removed unused setters.
**     RFB 20200424          Because we are now passing the actual 4GL name for a frame to 
**                           createFrame, in coreInitialize we will set to lowercase before
**                           storing in the frame class.
**     RFB 20200427          Updated javadoc to be more helpful with respect to frame name and
**                           frameName in both the members, and coreInitialize().
**     CA  20200503          FRAME option DCOLOR <number> stub.
** 371 GES 20200806          Eliminated ErrorManager context-local usage.
**     EVL 20200828          Added method to change internal frame name variable.
**     SBI 20200916          Changed stopEditingMode for the case of UPDATE with editing phrase,
**                           changed continueEditing and fixed setWorker (refs: #4804-61).
**     GES 20201010          Pushed processing into screen-buffer to improve encapsulation.
**     EVL 20201022          Optimized attributes flush implementation.
**     CA  20201112          Avoid pushing the entire frame definition when setting the FRAME or PARENT 
**                           attribute.
**     CA  20201117          Fixed a memory leak - FIELD-GROUP widgets were not being deleted.
**     SVL 20210105          When the state of a shared frame is restored, call pushScreenDefinition only
**                           once per frame. This eliminates flickering in ChUI.
**     SVL 20210112          Do not reset ENTERED flag when exiting an EDITING cycle.
**     CA  20210111          If the widget rejects the SCREEN-VALUE, do not update the local cache in the
**                           ScreenBuffer.
**     CA  20210123          For DATE, DATETIME and DATETIME-TZ, parseScreenValue must yield an unknown date
**                           instead of an empty string, when SCREEN-VALUE is empty.
**                           For CHARACTER values, parseScreenValue must use the default format, but only if
**                           the widget supports FORMAT (EDITOR does not support this attribute).
**     CA  20210127          Fixed a problem in DOWN frames - the literal field-group gets reconstructed, and
**                           it must be forced to be deleted.
**     VVT 20210317          parseScreenValue(Class, character, String): format string check added
**     VVT 20210322          Fixed after the review. See #4880-107.
**     VVT 20210323          Fixed after the review. See #4880-108.
**     CA  20210511          Fixed DISPLAY UNLESS-HIDDEN when the frame has HIDDEN=TRUE (the setScreenValue 
**                           must know that is from an UI statement).
**     HC  20210524          Fixed successive UNDERLINE statements to land on the same line.
**     CA  20210525          After SCREEN-VALUE has been set, the buffer for this widget must be marked as
**                           UNCHANGED.
**     HC  20210602          Made UP, DOWN, DISPLAY, ENABLE ALL and DISABLE ALL no-ops for default frame
**                           with no widgets.
**     CA  20210711          An 'ASSIGN cbWidget.' will leave the cbWidget variable as empty string if the
**                           COMBO-BOX:SCREEN-VALUE is unknown.
**     CA  20210711          Child frames need to know if it was visible before actually making them visible.
**                           This allows a weird quirk in 4GL, where the first visible child frame will not be 
**                           'moved to top' when setting the VISIBLE attribute (explicitly or implicitly via
**                           an UI statement).
**     CA  20210811          For a DIALOG-BOX, the NO-BOX attribute is always set (FrameConfig.box is false).
**                           Also, the BOX attribute for a DIALOG-BOX frame is not queryable/settable.
**     VVT 20210917          Javadoc fixed, missing @Override annotations added. 
**     CA  20210919          When reading from a file, the "?" literal must be treated as a string, and the ?
**                           literal must be treated as an unknown value.
**     CA  20211008          Fixed DISPLAY UNLESS-HIDDEN - push the screen-buffer for the widgets.
**     CA  20211018          The FRAME:BOX attribute for ChUI DIALOG-BOX needs to be kept set to TRUE, as in 
**                           ChUI the frame's border needs to be drawn.
**     VVT 20211018          New implementation methods added: getWidth(), setWidth(), getHeight() and
**                           setHeight() to match the changes in the Sizeable interface. See: #5594-45.
**     CA  20211029          Refactored FIELD-GROUP management for DOWN frames: now they are created only once
**                           and the static widgets are moved to the proper CURRENT-ITERATION field-group,
**                           when the frame changes the row.
**     ME  20211210          Skip assign for browse column if value is unknown but no rows available.
**     VVT 20211215          New methods added: setX(int), setY(int). See #5899.
**     ME  20220105          Skip assign for browse column if value is unknown but no rows available.
**                           Screen value for combo/list returns unknown if value not in list.
**     SVL 20220107          Partially reverted previous entry.
**     ME  20220110          Screen value is unknown only for combo/selection list if not in list.
**     AL2 20220121          Moved combo/selection screen-value conditional after core frame buffer checking.
**     ME  20220209          Replace getAttr for "realized" with isRealized method.
**     SVL 20220215          Implemented partial support for FRAME:SCROLL option.
**     OM  20220305          SCREEN-VALUE accepts ISO format for DATETIME(-TZ) widgets.
**                           General code maintenance and optimisations.
**     ME  20220306          Added getRealized. 
**     VVT 20220525          moveAfterTab method renamed to moveAfterTabItem in sake of consistence with
**                           moveBeforeTabItem. See #6383.
**     IAS 20220617          Added support for the FIELD-VALUE assignemnt  
**     EVL 20220808          The DISPLAY WIDGET statement should reset HIDDEN attribute to FALSE when the
**                           widget is explicitly specified.
**     EVL 20220809          Fixed the regression from recent HIDDEN attribute reset change.  Only widgets
**                           defined in DISPLAY statement should be affected.
**     CA  20220912          For editable frames, only the screen buffers which have changed are transferred
**                           to the client-side.
**     CA  20220901          Refactored scope notification support: ScopeableFactory was removed, and the 
**                           registration is now specific to each type of scopeable.  For each case, the block
**                           will be registered for scope support (for that particular scopeable) only when
**                           the scopeable is 'active' (i.e. unnamed streams or accumulators are used).  This
**                           allows a lazy registration of scopeables, to avoid the unnecessary overhead of
**                           processing all the scopeables for each and every block.
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;
 
import java.beans.*;
import java.lang.reflect.*;
import java.math.*;
import java.util.*;
import java.util.Map.Entry;
import java.util.logging.*;

import com.goldencode.p2j.ui.FieldGroup.UpdatesOp;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.format.*;
import com.goldencode.p2j.ui.client.format.DisplayFormat.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.BlockManager.Condition;
import com.goldencode.p2j.util.ErrorManager.ErrorEntry;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.proxy.*;
import com.goldencode.util.*;

/**
 * This class implements a frame factory method which produces instances of 
 * interfaces generated at conversion time.
 * <p>
 * The implementation is based on the GCD dynamic proxy library. Factory
 * generated instances are proxies. They forward method calls back to this
 * class.
 * <p>
 * The forwarded methods are:
 * <ul>
 *   <li>widget value getters;
 *   <li>widget value setters;
 *   <li>widget accessors;
 * </ul>
 * Widgets corresponds to either single variables or those defined in Progress
 * as having extents (arrays).
 * <p>
 * Arrays implement two types of methods: for single individually named 
 * elements or for generic names that require a runtime calculated index. 
 */
public class GenericFrame
implements InvocationHandler,
           CommonFrame,
           Finalizable
{
   /** proxy getter call */
   private static final int INVOKE_GET = 0;

   /** proxy setter call */
   private static final int INVOKE_SET = 1;

   /** proxy accessor call */
   private static final int INVOKE_ACCESS = 2;

   /** Error message for enable statement during input redirection. */
   private static final String ENABLE_INPUT_REDIRECTED =
      "Cannot execute ENABLE statement when input stream is not the screen";
   
   /** Error message for set/prompt-for statements in batch mode. */
   private static final String NO_INPUT_SOURCE_ATTEMPT =
      "Attempt to read with no current source of input";

   /** Error message for output related statements in batch mode. */
   private static final String NO_OUTPUT_DESTINATION_ATTEMPT =
      "Attempt to write with no current output destination";
   
   /** logger */
   private static final Logger LOG = LogHelper.getLogger(GenericFrame.class.getName());
   
   /** Context-local error state. */
   private ErrorManager.ErrorHelper errHlp = null;

   /** batch modifications flag */
   private boolean batch = true;

   /** instance of the frame configuration implementation class */
   private Object configInstance = null;

   /** class of the frame configuration implementation class */
   private Class<?> configClass = null;

   /** legacy frame name as found in the 4GL with case-preserved. */
   private String frameName = null;

   /** frame ID */
   private int frameId = -1;

   /** Maps widget names to widget instances (from the frame definition). */
   private Map<String, GenericWidget<?>> n2w = null;

   /** Maps converted widget names to original ones */
   private Map<String, String> n2n = null;
   
   /** Maps header widget names to widget instances (from the frame definition). */
   private Map<String, GenericWidget<?>> n2h = null;
   
   /** Maps background widget names to widget instances (from the frame definition). */
   private Map<String, GenericWidget<?>> n2b = null;
   
   /** List of widgets in ID order. */
   private GenericWidget<?>[] widgets = null;
   
   /** Frame headers. */
   private FrameElement[] headers = null;
   
   /** frame config */
   private FrameWidget frame = null;

   /** frame definition */
   private ScreenDefinition frameDef = null;
      
   /** screen buffer for this frame */
   private ScreenBuffer frameBuf = null;
   
   /** Identifies type of editing block (PROMPT_FOR, SET or UPDATE). */
   private int currentOp = NONE;
   
   /** Lock object for synchronization of shared editing block data. */
   private final Object editLock = new Object();

   /** The widget used for displaying a statistic's description. */
   private GenericWidget<?> statsColumn = null;
   
   /** Stream to which redirected aggregator element output is sent */
   private Stream redirOut = null;
   
   /** The aggregate handler currently active for this frame */
   private AggregateHandler aggregateHandler = null;
   
   /**
    * Keeps state of the conditional DOWN and/or UP cursor movements.
    * Can be set to {-1, 0, +1}. -1 means there is an outstanding request
    * to move the cursor up. 0 means no outstanding requests. +1 means there
    * is an outstanding request to move the cursor down.
    */
   private int pendingUpDown = 0;
   
   /** Callback expression for the number of iterations in a down frame. */
   private Resolvable dynDown = null;

   /** Callback expression for the frame column. */
   private Resolvable dynColumn = null;

   /** Callback expression for the frame row. */
   private Resolvable dynRow = null;

   /** The registered storage for the current frame title. */
   private CharacterExpression dynTitle = null;

   /** Set to true when frame is viewed first time. */
   private boolean hadView = false;

   /** List of widgets in text groups in current editing block. */
   private List<FrameElement> textGroup = null;

   /** Nesting scope counter. */
   private int scopeNesting = 0;

   /** Shared flame flag. */
   private boolean shared = false;
   
   /** A screen-buffer cache to be used with a down frame. 
    * key   - a row number
    * value - a ScreenBuffer instance
    */
   private Map<Integer, ScreenBuffer> screenBufferCache = new HashMap<>();
   
   /** Dynamic DOWN value. */
   private int dynamicDown = 0;
   
   /** The first field group in the frame (holds all non-header widgets). */
   private FieldGroup fieldGroup = null;
   
   /** The current field-group. Corresponds to CURRENT-ITERATION attribute */
   private FieldGroup currentFieldGroup;
   
   /** The field-group for the literal widget */
   private FieldGroup literalFieldGroup;
   
   /** The mapping from row number to corresponding field-group */
   private Map<Integer, FieldGroup> rows2fg = new HashMap<>();

   /**
    * Set to <code>true</code> when frame is initialized by appropriate
    * {@link Settable} instance.
    */
   private boolean insideSetup = false;

   /** 
    * List of aggregators, deferred to be registered on the first iteration. 
    */
   private FrameElement[] deferredAggregators = null;

   /** Can hide other frames. */
   private boolean canHide = true;

   /** Set to <code>true</code> if last operation was DOWN. */
   private boolean afterDown = false;

   /** Set to <code>true</code> if last down() was internal underline() call. */
   private boolean afterUnderline = false;

   /** Track presence of condition during frame processing. */
   private boolean hadCondition = false;

   /** Track implicit validation nesting. */
   private int triggerActive = 0;

   /** Final format check control. */
   private boolean finalCheck = false;

   /** Second and subsequent formatting errors are suppressed. */
   private boolean silentFormatErrors = false;

   /** Widget IDs stored in this map are not checked for formatting errors. */
   private Set<Integer> formatFilter = new HashSet<>();

   /** Track frame data flushing in redirected mode. */
   private boolean flushed = true;
   
   /** Track whether we change data from one iteration to another */
   private boolean sameViewDataRedirected = false;
   
   /**
    * The current UI statement in use for this frame. <code>null</code> if no statement is in 
    * effect.
    */
   private UIStatement currentStatement = null;
   
   /** Flag indicating we are inside a {@link #placeFrame()} call. */
   private boolean framePlacing = false;

   /** {@code true} for a default frame (a static frame without name) */
   private boolean defaultFrame;
   
   /**
    * Default constructor; creates a new instance of a frame. This constructor
    * is meant to be invoked by subclasses. To initialize the frame, the
    * caller must invoke {@link #initialize(Class, Object, String)}.
    */
   protected GenericFrame()
   {
   }
   
   /**
    * Creates new frame as a proxy object (as a local resource).  This loads
    * the static frame definition.
    *
    * @param    <T> 
    *           Frame interface type.
    * @param    iface
    *           The class of the interface definition. 
    * @param    name
    *           The frame name.
    *
    * @return   The created frame.
    */
   public static <T extends CommonFrame> T createFrame(Class<T> iface, String name)
   {
      T frame = createFrameProxy(iface);

      // initialize static frame
      initializeNewFrame((GenericFrame) frame, iface, name, null, false);
      setupFrame(frame);
      ((GenericFrame) frame).finishConfigProcessing();
      ((GenericFrame) frame).literalFieldGroup = ((GenericFrame) frame).initLiteralFG();
      
      return frame;
   }

   /**
    * Creates new dynamic frame as a proxy object (as a local resource).  This loads
    * the static frame definition.
    *
    * @param    <T> 
    *           Frame interface type.
    * @param    iface
    *           The class of the interface definition. 
    * @param    frameWidget
    *           The frame widget to be the base for new dynamic frame.
    *
    * @return   The created frame.
    */
   public static <T extends CommonFrame> T
   createDynamicFrame(final Class<T> iface, final FrameWidget frameWidget)
   {
      final T frame = createFrameProxy(iface);
      
      // initialize dynamic frame
      initializeNewFrame((GenericFrame) frame, iface, null, frameWidget, true);
      setupFrame(frame);
      ((GenericFrame) frame).finishConfigProcessing();
      
      // part of the scope opening calls need to be done for dynamic frame
      
      // push down any pending screen definitions
      LogicalTerminal.processDeferredPush(true);
      // signal to client about scope opening
      LogicalTerminal.signalFrameScope((GenericFrame) frame, MarkEntry.OPEN_SCOPE);
      
      return frame;
   }

   /**
    * Creates new frame as a proxy object and exports that new frame as a
    * shared resource (non-global).  This loads the static frame definition.
    *
    * @param    <T> 
    *           Frame interface type.
    * @param    iface
    *           The class of the interface definition. 
    * @param    name
    *           The frame name.
    *
    * @return   The created frame.
    */
   public static <T extends CommonFrame> T createSharedFrame(Class<T> iface, String name)
   {
      T frame = createFrame(iface, name);
      
      final GenericFrame gFrame = (GenericFrame) frame;
      gFrame.shared = true;
      
      // export the frame (there is no global shared frame)
      final GenericFrame old = SharedVariableManager.addFrame(name, gFrame);
      
      if (old != null)
      {
         // remove it from the registry, but add it back once the work is done with the new frame
         // which uses the same name (thus it hides this one).
         LogicalTerminal.releaseSharedFrame(old.getFrameId());
         Finalizable clean = new Finalizable()
         {
            
            @Override
            public void retry()
            {
               // no-op
            }
            
            @Override
            public void iterate()
            {
               // no-op
            }
            
            @Override
            public void finished()
            {
               LogicalTerminal.registerSharedFrame(old);
            }
            
            @Override
            public void initFailure()
            {
               finished();
            }
            
            @Override
            public void deleted()
            {
               // no-op
            }
         };
         TransactionManager.registerNextExternal(clean);
      }
      
      // return the specific proxy that matches the requested interface
      return frame;
   }

   /**
    * Import an existing frame as a proxy object from the shared resource
    * manager.  Create a new proxy and copy the state of the retrieved, shared
    * frame to this new proxy for temporary use, until the scope in which the
    * new proxy is used ends.  When the associated external procedure exits,
    * the original frame definition will be restored, and any relevant state
    * from the exiting scope will be copied back to it.
    * 
    * @param    <T> 
    *           Frame interface type.
    * @param    iface
    *           The class of the interface definition. 
    * @param    name
    *           The frame name.
    *
    * @return   The imported frame.
    */
   public static <T extends CommonFrame> T importSharedFrame(Class<T> iface, String name)
   {
      GenericFrame mFrame = SharedVariableManager.lookupFrame(name);
      
      // if [mFrame] is not of the correct type then we have detected at conversion time
      // if it is from a wrong frame partition, [mFrame] is [null] or not assignable to [iface]
      if (mFrame == null || !iface.isAssignableFrom(mFrame.getClass()))
      {
         // TODO: this is not safe for silent error mode, ErrorManager.recordOrThrowError()
         //       should be used OR ErrorManager.throwError(NE, boolean) but generally we
         //       should not be instantiating this exception directly
         
         String msg = "%s Shared frame layouts do not match for frame %s";
         String pname = ProcedureManager.getInstantiatingExternalProgram();
         throw new ErrorConditionException(683, String.format(msg, pname, name));
      }
      
      // there needs to be a cleaner which saves the HELP for all widgets and restores it once
      // its done...
      for (GenericWidget<?> gw : mFrame.widgets)
      {
         gw.saveSharedState();
      }
      Finalizable clean = new Finalizable()
      {
         
         @Override
         public void retry()
         {
            // no-op
         }
         
         @Override
         public void iterate()
         {
            // no-op
         }
         
         @Override
         public void finished()
         {
            boolean update = false;
            for (GenericWidget<?> gw : mFrame.widgets)
            {
               update = gw.restoreSharedState() || update;
            }

            if (update)
            {
               mFrame.pushScreenDefinition();
            }
         }
         
         @Override
         public void initFailure()
         {
            finished();
         }
         
         @Override
         public void deleted()
         {
            // no-op
         }
      };
      TransactionManager.registerNextExternal(clean);
      
      return (T) mFrame;
   }
   
   /**
    * Processes validation rules (if they exist), error checking and any other constraint
    * processing that needs to be applied.  This is used to check if a proposed update to an
    * object will pass validation.
    *
    * @param    enabledOnly
    *           If <code>true</code>, VALIDATE will
    *           check enabled fields only. Otherwise, the VALIDATE method validates all fields,
    *           whether enabled or not.
    *
    * @return   {@code true} if the validation is successful.
    */
   public logical validateFields(boolean enabledOnly)
   {
      if (!frame._isRealized())
      {
         return new logical(true);
      }
      for (GenericWidget<?> w : widgets)
      {
         if (!(w instanceof ControlEntity<?>) || (w instanceof BrowseWidget))
         {
            continue;
         }
         logical r = w.validateFields(enabledOnly);
         if (!r.booleanValue())
         {
            return r;
         }
      }
      return new logical(true);
   }

   /**
    * Retrieve the frame definition class and create and return a new instance
    * of this class.  This method relies on the convention of storing the
    * frame definition class in a static field of the frame interface, named
    * "configClass".  Reflection is used to retrieve that field and
    * instantiate the object from it.
    * 
    * @param   iface
    *          Frame interface type.
    * 
    * @return  Instance of the frame definition class.
    * 
    * @throws  RuntimeException
    *          if the configClass field does not exist, is inaccessible, or
    *          the class cannot be instantiated.
    */
   private static Object getConfigInstance(Class<?> iface)
   {
      Class<?> cfgCls  = null;
      Object cfgInst = null;

      try
      {
         // each frame interface has a static field "configClass", which
         // contains the implementation class associated with that interface
         Field link = iface.getDeclaredField("configClass");
         
         cfgCls  = (Class<?>) link.get(null);
         cfgInst = cfgCls.newInstance();
      }
      catch (NoSuchFieldException exc)
      {
         throw new RuntimeException("configClass is undefined ", exc);
      }
      catch (IllegalAccessException exc)
      {
         throw new RuntimeException("configClass is inaccessible ", exc);
      }
      catch (InstantiationException exc)
      {
         throw new RuntimeException("cannot instantiate configClass ", exc);
      }
      
      return cfgInst;
   }
   
   /**
    * Initialize the frame and register it with the transaction manager.
    *
    * @param    <T>
    *           Frame type.
    *
    * @param    frame
    *           Frame to be initialized (also the proxy for the given
    *           interface).
    * @param    iface
    *           The class of the interface definition. 
    * @param    name
    *           The frame name.
    * @param    frameWidget
    *           The frame widget to be the base for new dynamic frame.
    * @param    dynamic
    *           The flag to differentiate processing for static and dynamic frames.
    */
   private static <T extends CommonFrame>
   void initializeNewFrame(GenericFrame frame,
                           Class<T> iface,
                           String name,
                           FrameWidget frameWidget,
                           boolean dynamic)
   {
      Object configInstance = getConfigInstance(iface);
      Class<?> impl = configInstance.getClass();
      
      if (dynamic)
      {
         // initialize the frame for all we can init at this time
         frame.initializeEmptyFrame(impl, configInstance, frameWidget);
      }
      else
      {
         // initialize the frame (aka, invocation handler) for the implementation class
         frame.initialize(impl, configInstance, name);
      }
   }
   
   /**
    * Checks if the parameter - valid window handle.
    * 
    * @param    hWin
    *           Handle of window to display 
    * @return   <code>TRUE</code> - if handle is valid, 
    *           <code>FALSE</code> - otherwise
    */
   private boolean isAbleToDisplayWindow(handle hWin)
   {
      // current-window is used (not the frame's window)
      if (hWin == null || !hWin._isValid() || hWin.isUnknown())
      {
         String ms = "Could not find widget handle for window from expression";
         ErrorManager.recordOrThrowError(3138, ms);
         return false;
      }
      
      if (!(hWin.getResource() instanceof WindowWidget))
      {
         String widgetType = hWin.unwrap().getResourceType().getTypeName();
         String ms = "Wrong type of object handle (" + widgetType + ")  " +
                     "-- must be a window handle";
         ErrorManager.recordOrThrowError(3139, ms);
         return false;
      }
      
      return true;
   }

   /**
    * Create and initialize a frame as a proxy object, loading the static
    * frame definition.
    * 
    * @param    <T> 
    *           Frame interface.
    * @param    iface
    *           The class of the interface definition. 
    *
    * @return   The created frame.
    */
   @SuppressWarnings("unchecked")
   private static <T extends CommonFrame> T createFrameProxy(Class<T> iface)
   {
      List<Class<?>> ifaceList = new ArrayList<>();
      
      // proxy the interface and all of its ancestor superinterfaces, except CommonFrame
      Class<?> next = iface;
      while (next != null && !CommonFrame.class.equals(next))
      {
         ifaceList.add(next);
         Class<?>[] ext = next.getInterfaces();
         next = (ext.length == 1 ? ext[0] : null);
      }
      
      // also proxy Settable
      ifaceList.add(Settable.class);
      
      Class<?>[] ifaces = new Class<?>[ifaceList.size()];
      ifaces = ifaceList.toArray(ifaces);
      
      // manufacture an instance that will redirect method calls to our invocation handler
      T proxy = ProxyFactory.getProxy(GenericFrame.class, ifaces);
      
      return proxy;
   }
   
   /**
    * Invoke the frame's setup logic and set the batch flag to false.
    * 
    * @param   <T>
    *          Frame interface type.
    * @param   frame
    *          Frame proxy instance.
    */
   private static <T extends CommonFrame> void setupFrame(T frame)
   {
      frame.setInsideSetup(true);
      ((Settable) frame).setup(frame);
      frame.setInsideSetup(false);
      frame.cleanNonTabItems();
      frame.batch(false);
   }
   
   /**
    * Set the current UI statement in use for this frame.
    * <p>
    * If {@link #currentStatement} is already set, do nothing and return <code>false</code>. 
    * Otherwise return <code>true</code>.
    * 
    * @param    stmt
    *           The UI statement being executed.
    * 
    * @return   See above.
    */
   public boolean setStatement(UIStatement stmt)
   {
      boolean save = (currentStatement == null);
      if (save)
      {
         currentStatement = stmt;
      }
      
      return save;
   }
   
   /**
    * Set the {@link #currentStatement} back to <code>null</code>.
    */
   public void resetStatement()
   {
      currentStatement = null;
   }
   
   /**
    * Get the {@link #currentStatement}.
    * 
    * @return   See above.
    */
   public UIStatement getStatement()
   {
      return currentStatement;
   }
   
   /**
    * @see com.goldencode.p2j.ui.CommonFrame#setInsideSetup(boolean)
    */
   @Override
   public void setInsideSetup(boolean flag)
   {
      insideSetup = flag;
   }
   
   /**
    * Check if the frame is currently being initialized
    * 
    * @return   See above.
    */
   @Override
   public boolean isInsideSetup()
   {
      return insideSetup;
   }
   
   /**
    * Get frame name.
    * 
    * @return frame name.
    */
   public String _getName()
   {
      return frameName;
   }
   
   /**
    * Set frame name.  Internal version.
    * 
    * @param    name
    *           The new name for the frame.
    */
   public void _setName(String name)
   {
      frameName = name;
   }
   
   /**
    * Check if frame is shared.
    * 
    * @return  <code>true</code> if frame is shared frame.
    */
   public boolean isShared()
   {
      return shared;
   }
   
   /**
    * Getter for the redirOut field.
    * 
    * @return  the redirOut field.
    */
   public Stream getRedirOut() 
   {
      return redirOut;
   }
   
   /**
    * Register header frame elements.
    * <p>
    * This method corresponds to these Progress statements:
    * FORM HEADER
    * 
    * @param   fe
    *          Array of headers to register.
    */
   @Override
   public void registerHeader(FrameElement[] fe)
   {
      int size = 0;
      
      if (headers != null)
      {
         size += headers.length;
      }
      if (fe != null)
      {
         size += fe.length;
      }
      
      FrameElement[] newHeaders = new FrameElement[size];
      
      int destPos = 0;
      
      if (headers != null && headers.length > 0)
      {
         System.arraycopy(headers, 0, newHeaders, destPos, headers.length);
         destPos += headers.length;
      }
      
      if (fe != null && fe.length > 0)
      {
         System.arraycopy(fe, 0, newHeaders, destPos, fe.length);
      }
      
      headers = newHeaders;
   }

   /**
    * Register all the widgets which need accumulation with this frame.
    * 
    * @param    fe
    *           The list of widgets that may need accumulation.
    */
   @Override
   public void registerAccumulators(FrameElement[] fe)
   {
      this.deferredAggregators = fe;
   }
   
   /**
    * This method acts as an interceptor which is called every time a
    * method of the proxy object is invoked. It allows us to perform pre-
    * and post-processing for these methods, and to delegate the regular
    * work of the method to the backing implementation DMO.
    *
    * @param   proxy
    *          Proxy object upon which the intercepted method is invoked.
    * @param   method
    *          Method to be invoked through reflection.
    * @param   args
    *          Arguments, if any, to <code>method</code>.
    *
    * @return  value returned from the invoked method. 
    * 
    * @throws  Throwable
    *          if an exception/error occurs upon the invocation target,
    *          or if any problem occurs during pre- or post-processing.
    */
   @Override
   public Object invoke(Object proxy, Method method, Object[] args)
   throws Throwable
   {
      // frame is accessed, register scopeable
      TransactionManager.registerBlockScopeable(LogicalTerminal.getInstance());
      
      int    op         = -1;
      String widgetName = null;

      try
      {
         Class<?> declClass = method.getDeclaringClass();
         
         // handle special (reserved) methods
         if (declClass.equals(Object.class)      ||
             declClass.equals(Settable.class))
         {
            return method.invoke(configInstance, args);
         }
         
         // handle CommonFrame interface methods
         if (declClass.equals(CommonFrame.class))
         {
            return method.invoke(this, args);
         }
         
         int remove = -1;
         
         String wname = method.getName();
         
         // must be a getter/setter/widget accessor
         if (wname.startsWith("get"))
         {
            // standard getter
            op = INVOKE_GET;
            remove = 3;
         }
         else if (wname.startsWith("is"))
         {
            // boolean getter
            op = INVOKE_GET;
            remove = 2;
         }
         else if (wname.startsWith("set"))
         {
            // setter
            op = INVOKE_SET;
            remove = 3;
         }
         else if (wname.startsWith("widget"))
         {
            // widget access
            op = INVOKE_ACCESS;
            remove = 6;
         }
         else
         {
            throw new RuntimeException("unknown proxy method call " + wname);         
         }
         
         widgetName = wname.substring(remove);
      }
      
      catch (InvocationTargetException exc)
      {
         // This is the result of a programming error.  Don't wrap in a
         // legacy-style exception type.
         throw exc.getTargetException();
      }

      GenericWidget<?> widget = getWidget(widgetName,
                                          op, 
                                          (args == null) ? 0    : args.length,
                                          (args == null) ? null : args[0]);
      
      if (widget == null)
      {
         ErrorManager.recordOrThrowError(71, 
                                         "The value input for " + widgetName + " is not valid", 
                                         true);
      }
      
      // execute getter
      if (op == INVOKE_GET)
      {
         return getter(getWidgetId(widget), method.getReturnType(), true, true);
      }

      // execute setter
      if (op == INVOKE_SET)
      {
         Object parm = (args != null && args.length == 2) ? args[1] : args[0];
         return setter(getWidgetId(widget), parm);
      }

      // must be the accessor
      
      if (op == INVOKE_ACCESS && widget == null)
      {
         if (args != null && args.length == 1)
         {
            String err = "Array index " + indexAsString(args[0]) +
                         " is not formatted in the frame";
            ErrorManager.recordOrThrowError(622, err);
            
            // silent error case
            return null;
         }
      }
      
      return widget;
   }
   
   /**
    * Sets the V6FRAME option.
    *
    * @param    v6frame
    *           <code>true</code> if the option is set to ON.
    */
   @Override
   public void setV6Frame(boolean v6frame)
   {
      if (!LogicalTerminal.isChui())
         UnimplementedFeature.partial("V6FRAME client support missing (only matters in GUI).");
      
      frame.config.v6frame = v6frame;
      frame.pushWidgetAttr("v6frame", frame.config.v6frame);
   }
   
   /**
    * Sets the VISIBLE attribute.
    *
    * @param    visible
    *           <code>true</code> if the attribute must be set to ON.
    */
   @Override
   public void setVisible(boolean visible)
   {
      frame.setVisible(visible);
   }

   /**
    * Sets the VISIBLE attribute.
    *
    * @param    visible
    *           <code>true</code> if the attribute must be set to ON.
    */
   @Override
   public void setVisible(logical visible)
   {
      if (visible.isUnknown())
         return;
      setVisible(visible.booleanValue());
   }

   /**
    * Gets the VISIBLE attribute of the frame.
    *
    * @return   <code>true</code> if the attribute is set to ON.
    */
   public boolean _isVisible()
   {
      return asWidget()._isVisible();
   }

   /**
    * Gets the VISIBLE attribute of the frame.
    *
    * @return   PROGRESS <code>true</code> value if the attribute is set to ON.
    */
   @Override
   public logical isVisible()
   {
      return new logical(_isVisible());
   }

   /**
    * Queries the SENSITIVE attribute of this frame.
    *
    * @return <code>true</code> if SENSITIVE is set.
    */
   @Override
   public logical isSensitive()
   {
      return new logical(frame.isSensitive());
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(logical s)
   {
      if (s.isUnknown())
         return;
      setSensitive(s.booleanValue());
   }

   /**
    * Sets new value of SENSITIVE attribute.
    *
    * @param s
    *        new value of SENSITIVE attribute
    */
   @Override
   public void setSensitive(boolean s)
   {
      frame.setSensitive(s);
   }

   /**
    * Sets the HIDDEN writable attribute.
    *
    * @param    hidden
    *           The new value for the HIDDEN attribute.
    */
   @Override
   public void setHidden(logical hidden) 
   {
      if (hidden.isUnknown())
         return;
      setHidden(hidden.booleanValue());
   }

   /**
    * Sets the HIDDEN writable attribute.
    *
    * @param    hidden
    *           The new value for the HIDDEN attribute.
    */
   @Override
   public void setHidden(boolean hidden) 
   {
      frame.setHidden(hidden);
   }

   /**
    * Queries the HIDDEN attribute of this frame.
    *
    * @return <code>true</code> if HIDDEN is set.
    */
   @Override
   public logical isHidden()
   {
      return new logical(frame.isHidden());
   }
   
   /**
    * Obtain the current value in the screen buffer for the given widget and 
    * assign this into the backing variable. If there is no value in the
    * screen buffer (if it is <code>null</code>) no assignment will occur.
    *
    * @param    data
    *           Represents the variable to assign and the widget for which
    *           the screen value is to be retrieved.
    */
   @Override
   public void assignScreenValue(FrameElement data)
   {
      GenericWidget<?> widget = data.getWidget();
      int              wid    = getWidgetId(widget);
      Class<?>         type   = ((ControlEntity<?>) widget).getDataClass();
      
      // get the value out of the screen buffer (and make sure it is 
      // converted to the right type if needed)
      BaseDataType value = (BaseDataType) getter(wid, type, false, false);
      
      if (value != null && value.isUnknown() && widget instanceof BrowseColumnWidget)
      {
         BrowseWidget browse = ((BrowseColumnWidget) widget).getBrowse();
         if (browse.getCurrentRow() == -1)
         {
            value = null;
         }
      }
      
      // only assign if there is a value in the screen buffer
      if (value != null)
      {
         if (value.isUnknown() && widget instanceof ComboBoxWidget)
         {
            data.set(value.instantiateDefault());
         }
         else
         {
            data.set(value);
         }
      }
   }
   
   /**
    * Get this frame as instance of <code>GenericFrame</code>.
    * 
    * @return  reference to this frame as <code>GenericFrame</code>
    *          instance.
    */
   @Override
   public FrameWidget asWidget()
   {
      return frame;
   }
   
   /**
    * Get this frame wrapped in a handle.
    * 
    * @return  handle to this frame.
    */
   @Override
   public handle asWidgetHandle()
   {
      return new handle(frame);
   }
   
   /**
    * Applies an event to the widget in focus.
    * <p>
    * This method corresponds to these Progress statements:
    * APPLY FRAME
    *
    * @param event
    *        event to be applied
    */
   @Override
   public void apply(character event)
   {
      if (event.isUnknown())
         throw new RuntimeException("event cannot be unknown");

      apply(event.getValue());
   }

   /**
    * Applies an event to the widget in focus.
    * <p>
    * This method corresponds to these Progress statements:
    * APPLY FRAME
    *
    * @param event
    *        event to be applied
    */
   @Override
   public void apply(String event)
   {
      LogicalTerminal.apply(this, event);
   }

   /**
    * Applies an event to the widget in focus.
    * <p>
    * This method corresponds to these Progress statements:
    * APPLY FRAME
    *
    * @param eventCode
    *        event to be applied
    */
   @Override
   public void apply(integer eventCode)
   {
      if (eventCode.isUnknown())
         throw new RuntimeException("event cannot be unknown");

      LogicalTerminal.apply(this, eventCode.intValue());
   }

   /**
    * Applies an event to the widget in focus.
    * <p>
    * This method corresponds to these Progress statements:
    * APPLY FRAME
    *
    * @param eventCode
    *        event to be applied
    */
   @Override
   public void apply(double eventCode)
   {
      LogicalTerminal.apply(this, (int)eventCode);
   }

   /**
    * Sets batching mode for frame modification.
    *
    * @param  on
    *         <code>true</code> sets batching mode
    */
   @Override
   public void batch(boolean on)
   {
      batch(on, false);
   }

   /**
    * Queries batching mode for frame modification.
    *
    * @return 
    *         <code>true</code> if batching mode is in effect
    */
   public boolean isBatch()
   {
      return batch;
   }

   /**
    * Sets batching mode for frame modification and optionally pushes the
    * screen definition to the client. 
    *
    * @param  on
    *         <code>true</code> sets batching mode
    * @param  force
    *         <code>true</code> to ensure that the screen definition will be
    *         synchronously pushed to the client. <code>false</code> to
    *         defer the push. This is only honored when <code>on</code> is
    *         <code>false</code>.
    */
   public void batch(boolean on, boolean force)
   {
      batch = on;
      
      if (!batch)
         pushScreenDefinition(force);
   }
   
   /**
    * Performs CHOOSE processing.
    * <p>
    * This method corresponds to these Progress statements:
    * CHOOSE
    *
    * @param  c
    *         CHOOSE statement options
    */
   @Override
   public void choose(Choose c)
   {
      conditionalUpDown();
      LogicalTerminal.choose(this, c, null);
   }
      
   /**
    * Performs CHOOSE processing.
    * <p>
    * This method corresponds to these Progress statements:
    * CHOOSE COLOR
    *
    * @param  c
    *         CHOOSE statement options
    * @param  cs
    *         color specification for use with CHOOSE
    */
   @Override
   public void choose(Choose c, ColorSpec cs)
   {
      conditionalUpDown();
      LogicalTerminal.choose(this, c, cs);
   }
      
   /**
    * Clears the data and colors for all fill-in fields of the frame.
    * For down frames, also clears side labels.
    * <p>
    * This method corresponds to these Progress statements:
    * CLEAR
    *
    * @param  noPause
    *         <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void clear(boolean noPause)
   {
      conditionalUpDown();
      resetFrameBuf();
      LogicalTerminal.clear(this, noPause);
   }

   /**
    * Clears all occurrences for the down frame and resets the current display
    * position to the top of the frame.
    * <p>
    * This method corresponds to these Progress statements:
    * CLEAR ALL
    *
    * @param  noPause
    *         <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void clearAll(boolean noPause)
   {
      pendingUpDown = 0;
      resetFrameBuf();
      LogicalTerminal.clearAll(this, noPause);
      
      // the screen buffer cache needs also to be cleared
      screenBufferCache.clear();

      currentFieldGroup = fieldGroup;
   }
   
   /**
    * Conversion of CURRENT-ITERATION attribute (KW_CUR_ITER) of frame widget.
    *
    * Getter of CURRENT-ITERATION attribute.
    *
    * @return current value of CURRENT-ITERATION attribute.
    */
   @Override
   public handle currentIteration()
   {
      return currentFieldGroup == null ? new handle() : currentFieldGroup.asWidgetHandle();
   }
   
   /**
    * Set the CURRENT-ITERATION widget handle for this frame. This attribute is writable only
    * for frames. 
    *
    * @param   newIter
    *          A {@link handle} for the current iteration of the frame.
    */
   @Override
   public void setCurrentIteration(handle newIter)
   {
      UnimplementedFeature.missing("CURRENT-ITERATION setter is not implemented for frames.");
   }
   
   /**
    * Disables input for all field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements:
    * DISABLE ALL
    */
   @Override
   public void disable()
   {
      if (defaultFrame && widgets.length == 0)
      {
         // for default frame with no widgets this is a no-op
         return;
      }

      LogicalTerminal.disable(this, null);
   }

   /**
    * Disables input for all field-level widgets of the frame, except the
    * specified ones.
    * <p>
    * This method corresponds to these Progress statements:
    * DISABLE ALL EXCEPT
    *
    * @param    xcpt
    *           Array ow widgets to be excluded from the operation.
    */
   @Override
   public void disableExcept(GenericWidget<?>[] xcpt)
   {
      int[] ids = remainingIds(xcpt);
      
      if (ids == null)
      {
         // all widgets are excluded
         return;
      }
      
      // disable remaining widgets
      LogicalTerminal.disable(this, ids);
   }

   /**
    * Disables input for all field-level widgets of the frame, except the
    * specified one.
    * <p>
    * This method corresponds to these Progress statements:
    * DISABLE ALL EXCEPT
    *
    * @param    widget
    *           Widgets which will be excluded from the operation.
    */
   @Override
   public void disableExcept(GenericWidget<?> widget)
   {
      disableExcept(new GenericWidget<?>[] { widget });
   }
   
   /**
    * Disables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: DISABLE
    * 
    * @param    list
    *           The widgets to disable.
    */
   @Override
   public void disable(GenericWidget<?>[] list)
   {
      int[] ids = validIds(list);
      
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      // disable widgets
      LogicalTerminal.disable(this, ids);
   }
   
   /**
    * Disables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: DISABLE
    * 
    * @param   list
    *          The widgets to disable.
    */
   @Override
   public void disable(FrameElement[] list)
   {
      GenericWidget<?>[] wids = widgetListFromFrameElements(list);
      disable(wids);
   }
   
   /**
    * Disables input for all field-level widgets of the frame.
    * <p>
    * This method corresponds to this Progress statements: 
    * DISABLE ALL UNLESS-HIDDEN
    */
   @Override
   public void disableUnlessHidden()
   {
      disableUnlessHidden((GenericWidget<?>[]) n2w.values().toArray());
   }
   
   /**
    * Disables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * DISABLE UNLESS-HIDDEN
    * 
    * @param    list
    *           The widgets to disable.
    */
   @Override
   public void disableUnlessHidden(GenericWidget<?>[] list)
   {
      GenericWidget<?>[] listChanged = removeHiddenWidgets(list);
      
      if (listChanged != null && listChanged.length > 0)
      {
         disable(listChanged);
      }
   }
   
   /**
    * Disables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * DISABLE UNLESS-HIDDEN
    * 
    * @param    list
    *           The widgets to disable.
    */
   @Override
   public void disableUnlessHidden(FrameElement[] list)
   {
      FrameElement[] listChanged = removeHiddenWidgets(list);
      
      if (listChanged != null && listChanged.length > 0)
      {
         disable(listChanged);
      }
   }

   /**
    * Displays a frame with no "data container" fields on the screen.
    * <p>
    * As part of the processing, the frame marked as visible.  The frame will
    * be brought to the top of the Z-order and refreshed.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    * 
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void display(handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         // chui has only one window - default
         display(new FrameElement[]{}, hWin);
      }
   }

   /**
    * Refresh specified browse widget.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY WITH BROWSE
    *
    * @param    browse
    *           A widget to refresh. 
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void display(BrowseWidget browse, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         // TODO test behaviour in multi-window env
         display(browse);
      }
   }

   /**
    * Displays (writes) non-data-containing fields from the frame to the
    * given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void display(Stream out, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         viewWorker(out, null, false, true, hWin);
      }
   }
   
   /**
    * Displays a list of fields from the frame on the screen.  As part of
    * this operation, the list of data values are copied to the screen
    * buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    *
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void display(FrameElement[] data, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         displayWorker(data, true, hWin);
      }
   }
   
   /**
    * Displays a list of fields from the frame on the screen.  As part of
    * this operation, the list of data values are copied to the screen
    * buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY UNLESS-HIDDEN
    *
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void displayUnlessHidden(FrameElement[] data, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         displayUnlessHidden(data, true, hWin);
      }
   }
   
   /**
    * Displays (writes) a list of fields from the frame on the given stream.
    * As part of this operation, the list of data values are copied to the
    * screen buffer.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void display(Stream out, FrameElement[] data, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         display(out, data, true, hWin);
      }
   }
   
   /**
    * Displays a frame with no "data container" fields on the screen.
    * <p>
    * As part of the processing, the frame marked as visible.  The frame will
    * be brought to the top of the Z-order and refreshed.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    */
   @Override
   public void display()
   {
      display(new FrameElement[]{});
   }

   /**
    * Refresh specified browse widget.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY WITH BROWSE
    *
    * @param    browse
    *           A widget to refresh. 
    */
   @Override
   public void display(BrowseWidget browse)
   {
      if (asWidget().allFramesVisible())
      {
         browse.refreshCurrentRow();
      }
   }
   
   /**
    * Displays (writes) non-data-containing fields from the frame to the
    * given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    *
    * @param    out
    *           A stream instance which is open for writing. 
    */
   @Override
   public void display(Stream out)   
   {
      viewWorker(out, null, false, false, null);
   }

   /**
    * Displays a list of fields from the frame on the screen.  As part of
    * this operation, the list of data values are copied to the screen
    * buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    *
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    */
   @Override
   public void display(FrameElement[] data)
   {
      displayWorker(data, false, null);
   }

   /**
    * Displays (writes) a list of fields from the frame on the given stream.
    * As part of this operation, the list of data values are copied to the
    * screen buffer.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    */
   @Override
   public void display(Stream out, FrameElement[] data)
   {
      display(out, data, false, null);   
   }

   /**
    * Displays a list of fields from the frame on the screen.  As part of
    * this operation, the list of data values are copied to the screen
    * buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY UNLESS-HIDDEN
    *
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    */
   @Override
   public void displayUnlessHidden(FrameElement[] data)
   {
      displayUnlessHidden(data, false, null);
   }
   
   /**
    * Displays this frame (emulates a view) unless it is hidden.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY UNLESS-HIDDEN
    */
   @Override
   public void displayUnlessHidden()
   {
      if (frame.getAttr("hidden", () -> frame.config.hidden, true))
      {
         return;
      }
      
      view();
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: ENABLE
    * 
    * @param    list
    *           The widgets to enable.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void enable(GenericWidget<?>[] list, handle hWin)
   {
      enable(list, true, hWin);
   }
   
   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: ENABLE
    * 
    * @param    list
    *           The widgets to enable.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void enable(FrameElement[] list, handle hWin)
   {
      GenericWidget<?>[] wids = widgetListFromFrameElements(list);
      enable(wids, hWin);
   }
   
   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN
    * 
    * @param    widget
    *           The widget to enable.
    */
   @Override
   public void enableUnlessHidden(GenericWidget<?> widget ) 
   {
      enableUnlessHidden(new GenericWidget<?>[] { widget });
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN
    * 
    * @param    widget
    *           The widget to enable.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void enableUnlessHidden(GenericWidget<?> widget, handle hWin) 
   {
      enableUnlessHidden(new GenericWidget<?>[] { widget }, hWin);
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN
    * 
    * @param    list
    *           The widgets to enable.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void enableUnlessHidden(GenericWidget<?>[] list, handle hWin)
   {
      GenericWidget<?>[] listChanged = removeHiddenWidgets(list);
      
      if (listChanged != null && listChanged.length > 0)
      {
         enable(listChanged, hWin);
      }
   }
   
   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN
    * 
    * @param    list
    *           The widgets to enable.
    */
   @Override
   public void enableUnlessHidden(GenericWidget<?>[] list)
   {
      GenericWidget<?>[] listChanged = removeHiddenWidgets(list);
      
      if (listChanged != null && listChanged.length > 0)
      {
         enable(listChanged);
      }
   }
   
   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE ALL UNLESS-HIDDEN
    * 
    */
   @Override
   public void enableUnlessHidden()
   {
      enableUnlessHidden(widgets);
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN
    * 
    * @param    list
    *           The widgets to enable.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void enableUnlessHidden(FrameElement[] list, handle hWin)
   {
      FrameElement[] listChanged = removeHiddenWidgets(list); 

      if (listChanged != null && listChanged.length > 0)
      {
         enable(listChanged, hWin);
      }
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN
    * 
    * @param    list
    *           The widgets to enable.
    */
   @Override
   public void enableUnlessHidden(FrameElement[] list)
   {
      FrameElement[] listChanged = removeHiddenWidgets(list); 

      if (listChanged != null && listChanged.length > 0)
      {
         enable(listChanged);
      }
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN ALL EXCEPT
    * 
    * @param    widget
    *           The widget to enable.
    */
   @Override
   public void enableUnlessHiddenExcept(GenericWidget<?> widget ) 
   {
      enableUnlessHiddenExcept(new GenericWidget<?>[] {widget} );
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN ALL EXCEPT
    * 
    * @param    widget
    *           The widget to enable.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void enableUnlessHiddenExcept(GenericWidget<?> widget, handle hWin) 
   {
      enableUnlessHiddenExcept(new GenericWidget<?>[] {widget}, hWin );
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN ALL EXCEPT
    * 
    * @param    xcpt
    *           The widgets to enable.
    */
   @Override
   public void enableUnlessHiddenExcept(GenericWidget<?>[] xcpt) 
   {
      enableExcept(removeHiddenWidgets(widgets), xcpt, false, null);
   }
   
   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: 
    * ENABLE UNLESS-HIDDEN ALL EXCEPT
    * 
    * @param    xcpt
    *           The widgets to enable.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void enableUnlessHiddenExcept(GenericWidget<?>[] xcpt, handle hWin) 
   {
      enableExcept(removeHiddenWidgets(widgets), xcpt, true, hWin);
   }
   
   /**
    * Positions the cursor on a new line in a down frame. This is a
    * conditional call. Multiple conditional down() calls have the effect of a
    * single down(1) call, which is delayed till the nearest terminal I/O
    * operation and can be cancelled by a conditional up() call.
    * <p>
    * Note that this call can be ignored in interactive mode if frame is not
    * yet viewed at least once before this call. After first VIEW
    * (explicit or implicit) hiding frame (either via HIDE statement or by
    * changing VISIBLE attribute) does not affect behavior and calls to this
    * method are honored. This corresponds to 4GL behavior.
    * <p>
    * This method corresponds to these Progress statements: DOWN
    */
   @Override
   public void down()
   {
      afterUnderline = false;
      downInt();
   }

   /**
    * Positions the cursor on a new line in a down frame.
    * <p>
    * This method corresponds to these Progress statements:
    * DOWN n
    *
    * @param lines
    *        number of iterations of the data to move down
    */
   @Override
   public void down(double lines)
   {
      afterUnderline = false;
      downWorker((int) lines, false);
   }

   /**
    * Positions the cursor on a new line in a down frame.
    * <p>
    * This method corresponds to these Progress statements:
    * DOWN n
    *
    * @param lines
    *        number of iterations of the data to move down
    */
   @Override
   public void down(NumberType lines)
   {
      afterUnderline = false;
      downWorker(lines.intValue(), false);
   }

   /**
    * Writes a single new-line to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * DOWN STREAM s
    *
    * @param    out
    *           A stream instance which is open for writing. 
    */
   @Override
   public void down(Stream out)
   {
      // must be before any other checks
      if (out != null && out.protectWrites())
      {
         return;
      }

      afterUnderline = false;
      downWorker(1, getRemoteStreamId(out), true, false);
   }

   /**
    * Writes <code>lines</code> number of new-lines to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * DOWN STREAM s n
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    lines
    *           Number of new-lines to output to the stream.
    */
   @Override
   public void down(Stream out, double lines)
   {
      // must be before any other checks
      if (out != null && out.protectWrites())
      {
         return;
      }
      
      if (lines < 1)
         return;

      afterUnderline = false;
      downWorker((int) lines, getRemoteStreamId(out), false, true);
   }

   /**
    * Writes <code>lines</code> number of new-lines to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * DOWN STREAM s n
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    lines
    *           Number of new-lines to output to the stream.
    */
   @Override
   public void down(Stream out, NumberType lines)
   {
      // must be before any other checks
      if (out != null && out.protectWrites())
      {
         return;
      }
      
      int l = lines.intValue();

      if (l < 1)
         return;

      downWorker(l, getRemoteStreamId(out), false, true);
   }

   /**
    * Enables input for all field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements:
    * ENABLE ALL
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   @Override
   public void enable()
   {
      // Special processing for ENABLE statement in batch mode.
      if (errorOnBatchModeEnable(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in batch mode
         return;
      }

      if (errorOnRedirectedInput(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in silent error mode
         return;
      }

      if (defaultFrame && widgets.length == 0)
      {
         // for default frame with no widgets this is a no-op
         return;
      }

      makeFrameEnable();
      LogicalTerminal.enable(this, null);
   }

   /**
    * Enables input for all field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements:
    * ENABLE ALL IN WINDOW
    * 
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void enable(handle hWin)
   {
      // Special processing for ENABLE statement in batch mode.
      if (errorOnBatchModeEnable(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in batch mode
         return;
      }

      if (errorOnRedirectedInput(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in silent error mode
         return;
      }
      
      makeFrameEnable();
      LogicalTerminal.enable(this, null, hWin);
   }

   /**
    * Enables input for all field-level widgets of the frame, except the
    * specified ones.
    * <p>
    * This method corresponds to these Progress statements:
    * ENABLE ALL
    *
    * @param    xcpt
    *           Array ow widgets to be excluded from the operation.
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   @Override
   public void enableExcept(GenericWidget<?>[] xcpt)
   {
      enableExcept(widgets, xcpt, false, null);
   }

   /**
    * Enables input for all field-level widgets of the frame, except the
    * specified ones.
    * <p>
    * This method corresponds to these Progress statements: ENABLE ALL IN WINDOW
    * 
    * @param    xcpt
    *           List of widgets which will not be enabled.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void enableExcept(GenericWidget<?>[] xcpt, handle hWin)
   {
      enableExcept(widgets, xcpt, true, hWin);
   }

   /**
    * Enables input for given field-level widgets of the frame, except the
    * specified ones.
    * <p>
    * This method corresponds to these Progress statements: ENABLE ALL IN WINDOW
    * 
    * @param    all
    *           List of widgets to be processed.
    * @param    xcpt
    *           List of widgets which will not be enabled.
    * @param    windowClause
    *           <code>true</code> if IN WINDOW clause is presented in ENABLE statement.
    *           <code>false</code> otherwise.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
    public void enableExcept(GenericWidget<?>[] all, GenericWidget<?>[] xcpt, 
         boolean windowClause, handle hWin)
   {

      if (errorOnBatchModeEnable(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in batch mode
         return;
      }

      if (errorOnRedirectedInput(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in silent error mode
         return;
      }
      
      int[] ids = remainingIds(all, xcpt);

      if (ids == null)
      {
         // all widgets are excluded
         return;
      }
      
      // enable remaining widgets
      if (windowClause)
      {
         LogicalTerminal.enable(this, ids, hWin);
      }
      else 
      {
         LogicalTerminal.enable(this, ids);
      }
   }

   /**
    * Enables input for all field-level widgets of the frame, except the
    * specified one.
    * <p>
    * This method corresponds to these Progress statements:
    * ENABLE ALL
    *
    * @param    widget
    *           Widget which will be excluded from the operation.
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   @Override
   public void enableExcept(GenericWidget<?> widget)
   {
      enableExcept(new GenericWidget<?>[] { widget });
   }
   
   /**
    * Enables input for all field-level widgets of the frame, except the
    * specified one.
    * <p>
    * This method corresponds to these Progress statements: ENABLE ALL IN WINDOW
    * 
    * @param    widget
    *           Widget which will not be enabled.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void enableExcept(GenericWidget<?> widget, handle hWin)
   {
      enableExcept(new GenericWidget<?>[] { widget }, hWin);
   }
   
   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: ENABLE
    * 
    * @param    list
    *           The widgets to enable.
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   @Override
   public void enable(GenericWidget<?>[] list)
   {
      enable(list, false, null);   
   }

   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: ENABLE
    * 
    * @param   list
    *          The widgets to enable.
    * 
    * @throws  ErrorConditionException
    *          If the unnamed input is redirected and silent error mode is
    *          off.
    */
   @Override
   public void enable(FrameElement[] list)
   {
      GenericWidget<?>[] wids = widgetListFromFrameElements(list);
      enable(wids);
   }
   
   /**
    * Obtain the current value in the screen buffer of the backing data for
    * the given widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @param    widget
    *           The widget for which the value is to be retrieved.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   @Override
   public character getScreenValue(GenericWidget<?> widget)
   {
      return getScreenValue(widget, false);
   }
   
   /**
    * Obtain the current value in the screen buffer of the backing data for
    * the given widget and return it as a <code>character</code> type. If
    * there is no value in the screen buffer (if it is <code>null</code>
    * which is the representation for an uninitialized value), then return
    * the empty string.
    *
    * @param    widget
    *           The widget for which the value is to be retrieved.
    * @param    ignoreFormat
    *           Ignore the widget format to form return value.
    *
    * @return   The value as a <code>character</code> type or the empty
    *           string if the screen buffer value is uninitialized.
    */
   public character getScreenValue(GenericWidget<?> widget, boolean ignoreFormat)
   {
      checkFieldGroup(widget);
      
      // safety first, should never happen
      if (frameBuf == null || widget == null)
      {
         return new character();
      }
      
      int wid = widget.getId();
      
      // quick out for uninitialized widgets except browse columns
      if (frameBuf.isUninitialized(wid) && !(widget instanceof BrowseColumnWidget))
      {
         // always empty string
         return widget.getScreenValue(false, null, null, true);
      }
      
      BaseDataType bdt = null;
      String       fmt = null;
      
      // try to find a screen-value override (this only handles a sub-set of the cases
      // at this time, such as combo-box and selection list)
      Object value = frameBuf.getScreenValue(wid);
      
      if (value == null)
      {
         // this is the common case where we lookup the actual widget value as stored
         // in the screen buffer
         value = frameBuf.getWidgetValue(wid);
         
         if (value == null)
         {
            if (widget instanceof SelectionListWidget || widget instanceof ComboBoxWidget)
            {
               // for combo-box and selection-list if value is not in list
               // screen-value is unknown
               return new character();
            }
            
            // hmmm, at this point we should have had a valid widget that was initialized
            // so encountering null here is unexpected. See below how we deal with unknown value.
            String wt = widget.getDataType().toStringMessage();
            // return new character(wt.startsWith("date") ? "" : "?");
            return new character("");
         }
      }
         
      if (value instanceof FormattedValue)
      {
         // these cases occur when an @ base field is used and there is
         // a call to the getter before any editing occurs on the client
         // OR an edit on the client was done but no changes were made 
         FormattedValue fv = (FormattedValue) value; 
         
         bdt = fv.getValue();
         fmt = fv.getFormat();
      }
      
      if (bdt == null && value instanceof BaseDataType)
      {
         // this could be either a regular value or an @ base field use
         bdt = (BaseDataType) value;
      }
      
      if (bdt == null)
      {
         String spec = "Unexpected screen-buffer contents %s (%s)";
         String msg  = String.format(spec, value, value.getClass().getName());
         
         UnimplementedFeature.todo(msg);
         
         // can this happen? for now, we treat it as a case of uninitialized
         return widget.getScreenValue(false, null, null, true);
      }
      
      // at this point we have a valid BDT value that needs to be converted to a
      // formatted character value
      
      // deal with unknown value
      if (bdt.isUnknown())
      {
         // for some reason, date/datetime/datetime-tz all return empty string for unknown
         // while all other types return the string "?" (not actually unknown val); since
         // datetime/datetime-tz both inherit from date, the use of isAssignableFrom() should
         // deal with all 3 cases
         return widget.getScreenValue(true, bdt, null, true);
      }
      
      // handle empty text values (the 4GL does not emit any static format string chars in
      // this case, so we can't let it call through)
      if (bdt instanceof com.goldencode.p2j.util.Text)
      {
         if (TextOps.lengthOf((com.goldencode.p2j.util.Text) bdt) == 0)
         {
            return new character("");
         }
      }
      
      // at this point we have a non-null, non-unknown BDT value
      
      // use the pre-configured widget format if we don't have an override
      if (fmt == null)
      {
         fmt = widget._getFormat();
      }

      return widget.getScreenValue(true, bdt, fmt, ignoreFormat);
   }
   
   /**
    * Set the current value in the screen buffer of the backing data for
    * the given widget.  If the given value is <code>null</code> then this
    * widget will be set to the uninitialized value.
    *
    * @param    widget
    *           The widget for which the value is to be retrieved.
    * @param    value
    *           The new value for the widget, use <code>null</code> to set
    *           the value as uninitialized.
    */
   @Override
   public void setScreenValue(GenericWidget<?> widget, character value)
   {
      checkFieldGroup(widget);
      
      int wid = widget.getId();
    
      if (frameDef != null && frameDef.isSideLabel(wid))
      {
         GenericWidget<?> parent = getWidgetForId(frameDef.getSideLabelParent(wid));
         parent.setLabel(value);
      }
      
      if (value == null)
      {
         // force the value to be uninitialized
         frameBuf.setState(wid, ScreenBuffer.UNINITIALIZED);
      }
      else
      {
         // honor the type of the widget's data
         Class<? extends BaseDataType> type;
         if (widget instanceof BrowseColumnWidget)
         {
            BrowseColumnWidget column = (BrowseColumnWidget) widget;
            BaseDataType var = DisplayFormat.instanceOfType(column.getDataType().getValue());
            type = var.getClass();
         }
         else
         {
            type = (Class<? extends BaseDataType>) ((ControlEntity<?>) widget).getDataClass();
         }

         // some widgets like editor do not have the format option
         String format = null;
         if (widget.hasFormat())
         {
            format = widget._getFormat();
            if (format == null && type.equals(character.class))
            {
               format = new character().defaultFormatString();
            }
         }
         BaseDataType val;
         try
         {
            val = parseScreenValue(type, value, format);
         }
         catch (DisplayFormatCheckException e)
         {
            // FIXME: we need to find out which error number to use here.
            throw new ErrorConditionException(e.getMessage());
         }

         // set the value in the screen buffer for this widget
         if (widget.setScreenValue(frameBuf, val, false))
         {
            conditionalUpDown(true);
            
            // refresh the screen
            LogicalTerminal.refreshFrameWidget(this, wid, val);

            // update local cache
            frameBuf.putScreenValue(wid, val);
            
            frameBuf.setState(wid, ScreenBuffer.UNCHANGED);
         }
      }
   }

   /**
    * Parses screen value (i.e. the value assigned to SCREEN-VALUE) and returns the expected data type.
    * The method handles string values in the normalized form as well as values containing fill chars as
    * permitted by the format and the actual data type.
    *
    * TODO: some negative cases are not handled, see uast/format_fill_chars.p and the section "test case 3"
    *
    * @param   type
    *          The expected value type.
    * @param   value
    *          The input screen value.
    * @param   format
    *          The widget format, may be {@code null}.
    *
    * @return  Returns the parsed string value in the expected data type instance.
    * 
    * @throws  ErrorConditionException
    *          then the format provided is invalid 
    * @throws  DisplayFormatCheckException
    *          then the format provided cannot be applied is not compatible with the type.
    */
   public static <T extends BaseDataType> T
   parseScreenValue(Class<T> type, character value, String format)
   throws DisplayFormatCheckException,
          ErrorConditionException
   {
      return parseValue(type, value, format, false);
   }
   
   /**
    * Parses field value (i.e. the value assigned to FIELD-VALUE) and returns the expected data type.
    * The method handles string values in the normalized form as well as values containing fill chars as
    * permitted by the format and the actual data type.
    *
    * TODO: some negative cases are not handled, see uast/format_fill_chars.p and the section "test case 3"
    *
    * @param   type
    *          The expected value type.
    * @param   value
    *          The input screen value.
    * @param   format
    *          The widget format, may be {@code null}.
    *
    * @return  Returns the parsed string value in the expected data type instance.
    * 
    * @throws  ErrorConditionException
    *          then the format provided is invalid 
    * @throws  DisplayFormatCheckException
    *          then the format provided cannot be applied is not compatible with the type.
    */
   public static <T extends BaseDataType> T
   parseFieldValue(Class<T> type, character value, String format)
   throws DisplayFormatCheckException,
          ErrorConditionException
   {
      return parseValue(type, value, format, true);
   }

   /**
    * Parses screen/field value (i.e. the value assigned to SCREEN-VALUE/FIELD-VALUE) and returns the expected data type.
    * The method handles string values in the normalized form as well as values containing fill chars as
    * permitted by the format and the actual data type.
    *
    * TODO: some negative cases are not handled, see uast/format_fill_chars.p and the section "test case 3"
    *
    * @param   type
    *          The expected value type.
    * @param   value
    *          The input screen value.
    * @param   format
    *          The widget format, may be {@code null}.
    * @param   fieldValue
    *          Indicates the parsing is for FIELD-VALUE
    *
    * @return  Returns the parsed string value in the expected data type instance.
    * 
    * @throws  ErrorConditionException
    *          then the format provided is invalid 
    * @throws  DisplayFormatCheckException
    *          then the format provided cannot be applied is not compatible with the type.
    */
   public static <T extends BaseDataType> T
   parseValue(Class<T> type, character value, String format, boolean fieldValue)
   throws DisplayFormatCheckException,
          ErrorConditionException
   {
      if (value.isUnknown())
      {
         try
         {
            return type.newInstance();
         }
         catch (ReflectiveOperationException e)
         {
            throw new RuntimeException(e);
         }
      }
      
      BaseDataType result = value;
      
      if (type.equals(character.class))
      {
         result = new character(value);
         if (format != null)
         {
            DisplayFormat displayFormat = DisplayFormatFactory.makeFormat(value, format);
            Presentation pres = fieldValue ? displayFormat.fromVar4FieldValue(value) : 
                                             displayFormat.fromVar(value);
            pres.checkFormat();
            pres.toVar(result);
         }
      }
      else if (type.equals(datetimetz.class))
      {
         if (isEmptyFormat(value.toStringMessage(), "-/:.+"))
         {
            result = new datetimetz();
         }
         else
         {
            String valStr = value.getValue();
            result = valStr.contains("T") ? date.parseIso8601(valStr) : new datetimetz(value);
         }
      }
      else if (type.equals(datetime.class))
      {
         if (isEmptyFormat(value.toStringMessage(), "-/:.+"))
         {
            result = new datetime();
         }
         else
         {
            String valStr = value.getValue();
            result = valStr.contains("T") ? date.parseIso8601(valStr) : new datetime(value);
         }
      }
      else if (type.equals(date.class))
      {
         result = isEmptyFormat(value.toStringMessage(), "-/") ? new date() : new date(value);
      }
      else if (type.equals(decimal.class))
      {
         String parsed = NumberType.parseDecimal(value.getValue(), false, true);
         BigDecimal dec = new BigDecimal(parsed);
         // do not pass the parsed string to decimal constructor, it expects a current number format and
         // not the Java format
         result = new decimal(dec);
      }
      else if (type.equals(integer.class))
      {
         String parsed = NumberType.parseDecimal(value.getValue(), true, true);
         result = new integer(parsed);
      }
      else if (type.equals(int64.class))
      {
         String parsed = NumberType.parseDecimal(value.getValue(), true, true);
         result = new int64(parsed);
      }
      else if (type.equals(logical.class))
      {
         result = new logical(value, format);
      }

      return (T)result;
   }
   
   /**
    * Sets the containing frame instance.
    * 
    * @param    hFrame
    *           handle which is expected to contain a <code>FrameWidget</code>
    *           instance
    */
   @Override
   public void setFrame(handle hFrame)
   {
      // TODO:  no idea if this works...
      this.frame = ((FrameWidget) hFrame.get());
   }

   /**
    * Set the current value in the screen buffer of the backing data for
    * the given widget.  If the given value is <code>null</code> then this
    * widget will be set to the uninitialized value.
    *
    * @param    wid
    *           The widget for which the value is to be set
    * @param    value
    *           The new value for the widget, use <code>null</code> to set
    *           the value as uninitialized.
    */
   public void setFrameValue(int wid, BaseDataType value)
   {
      GenericWidget<?> widget = getWidgetForId(wid);
      
      if (value != null && !(value instanceof character))
      {
         // honor the type of the widget's data
         Class<?> type = ((ControlEntity<?>) widget).getDataClass();
            
         // ensure the right type is being used
         if (!type.equals(value.getClass()))
         {
            String err = "Incompatible data types for frame-value assignment";
            ErrorManager.recordOrThrowError(676, err);
            return;
         }
      }
         
      // make a copy
      BaseDataType val = value.duplicate();
         
      if (value instanceof unknown)
      {
         val = translateUnknown(value, (ControlEntity<?>) widget);
      }
      
      // set the value in the screen buffer by index
      frameBuf.putWidgetValue(wid, val);
      
      conditionalUpDown(true);
      
      // refresh the screen and own copy of frame-value
      LogicalTerminal.refreshFrameWidget(this, wid, val);
   }
   
   /**
    * Makes the frame invisible with the default pause.
    * <p>
    * This method corresponds to these Progress statements:
    * HIDE FRAME
    * 
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void hide(handle hWin)
   {
      LogicalTerminal.hide(this.getFrameId(), false, hWin);
   }
   
   /**
    * Makes the frame invisible.
    * <p>
    * This method corresponds to these Progress statements:
    * HIDE FRAME
    *
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void hide(handle hWin, boolean noPause)
   {
      LogicalTerminal.hide(this.getFrameId(), noPause, hWin);
   }

   /**
    * If the frame is a PAGE-TOP or PAGE-BOTTOM and the stream is paged,
    * the frame is removed from the list of those used for the header or
    * footer.  Otherwise this is a no-operation.
    * <p>
    * This method corresponds to these Progress statements:
    * HIDE STREAM s FRAME
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void hide(Stream out, handle hWin, boolean noPause)
   {
      // TODO: implement this
   }

   /**
    * Makes widgets in the frame invisible.
    * 
    * @param    data
    *           Array of widgets to hide.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void hide(FrameElement[] data, handle hWin)
   {
      // TODO: implement this
   }
   
   /**
    * Makes widgets in the frame invisible.
    * 
    * @param    data
    *           Array of widgets to hide.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void hide(FrameElement[] data, handle hWin, boolean noPause)
   {
      // TODO: implement this
   }
   
   /**
    * Makes the frame invisible with the default pause.
    * <p>
    * This method corresponds to these Progress statements:
    * HIDE FRAME
    */
   @Override
   public void hide()
   {
       hide(false);
   }

   /**
    * Makes the frame invisible.
    * <p>
    * This method corresponds to these Progress statements:
    * HIDE FRAME
    *
    * @param  noPause
    *         <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void hide(boolean noPause)
   {
      hideWorker(noPause, false);
   }

   /**
    * If the frame is a PAGE-TOP or PAGE-BOTTOM and the stream is paged,
    * the frame is removed from the list of those used for the header or
    * footer.  Otherwise this is a no-operation.
    * <p>
    * This method corresponds to these Progress statements:
    * HIDE STREAM s FRAME
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void hide(Stream out, boolean noPause)
   {
      if (out != null)
      {
         // the version of HIDE that takes a widget list has a strange case
         // where the error messages are doubled (must be something deep inside
         // the Progress UI implementation) which occurs whenever there is > 1
         // widget in the list; since this method is the "all frame" version,
         // the same problem doesn't occur here; to implement that quirk please
         // see viewWorker(Stream...) since VIEW/DISPLAY have a similar issue
         
         // normal safety code here, exit in NO-ERROR mode if needed
         if (out.protectWrites())
            return;
      }
      
      // check if we must deregister our page element
      if (isPageElement(out))
      {
         // deregister the page element
         LogicalTerminal.deregisterPageElement(frameId,
                                               getPageTop(),
                                               getRemoteStreamId(out));
      }
   }
   
   /**
    * Specifies the field to be put in focus during the next input operation.
    * <p>
    * This method corresponds to these Progress statements:
    * NEXT-PROMPT
    *
    * @param  field
    *         widget to be put in focus
    */
   @Override
   public void nextPrompt(GenericWidget<?> field) 
   {
      int id = getWidgetId(field);
      
      if (id != -1)
      {
         LogicalTerminal.nextPrompt(this, id);
      }
   }

   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is not used at all.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void promptFor(FrameElement[] widgets, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         Stream in = UnnamedStreams.input();
         
         setStaticWidgets(widgets);
         
         if (in != null)
         {
            // redirected input case
            promptFor(in, widgets, hWin);
         }
         else
         {
            // interactive case
            promptFor(widgets, null, hWin);
         }
      }
   }

   /**
    * For each entry in the widget array, the input stream will be read and 
    * the widget's associated field in the screen buffer will be set. Data
    * will not be copied to or from the user's variables.
    * <p>
    * Any error during stream reading or data parsing will cause an error
    * condition to be generated.  Reaching the end of the stream will cause
    * an end condition to be generated.
    * <p>
    * Both the input stream and the current default output stream (if output
    * is redirected) are checked.  If both are set to <code>echo</code> mode,
    * then after all stream input is done, the screen buffer's contents will  
    * be echoed to the current output destination using {@link #view}. This
    * will occur whether or not an error occurs during stream processing and
    * data parsing.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    in
    *           A stream instance which is open for reading. 
    * @param    widgets
    *           The list of widgets which are to receive input from the
    *           stream.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void promptFor(Stream in, FrameElement[] widgets, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         processStreamInput(in, widgets, false, true, hWin);
      }
   }

   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is not used at all.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void promptFor(FrameElement[] widgets, EventList wait, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         promptForWorker(widgets, wait, true, hWin);
      }
   }

   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void promptFor(GenericWidget<?>[] widgets, handle hWin)
   {
      promptFor(widgets, null, hWin);
   }

   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void promptFor(GenericWidget<?>[] widgets, EventList wait, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         promptForWorker(widgets, wait, true, hWin);
      }
   }
   
   /**
    * Enables a list of widgets from the frame on the screen, shifts into input blocking mode
    * via {@link #waitFor} and after the user's edits are complete the list of widgets are
    * disabled.  No data is copied to or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be marked as visible.
    * The frame will be brought to the top of the Z-order and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own execution to the
    * stream form.  In such a case, all interactive processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR UNLESS-HIDDEN
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion of these elements is
    *           not used at all.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void promptForUnlessHidden(FrameElement[] widgets, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         FrameElement[] widgetsChanged = removeHiddenWidgets(widgets);
         
         if (widgetsChanged != null && widgetsChanged.length > 0)
         {
            promptFor(widgetsChanged, hWin);
         }
      }
   }
   
   /**
    * Enables a list of widgets from the frame on the screen, shifts into input blocking mode
    * via {@link #waitFor} and after the user's edits are complete the list of widgets are
    * disabled.  No data is copied to or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be marked as visible.
    * The frame will be brought to the top of the Z-order and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR UNLESS-HIDDEN
    *
    * @param    widgets
    *           The list of widgets which is to be edited.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void promptForUnlessHidden(GenericWidget<?>[] widgets, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         GenericWidget<?>[] widgetsChanged = removeHiddenWidgets(widgets);
         
         if (widgetsChanged != null && widgetsChanged.length > 0)
         {
            promptFor(widgetsChanged, hWin);
         }
      }
   }

   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is not used at all.
    */
   @Override
   public void promptFor(FrameElement[] widgets)
   {
      Stream in = UnnamedStreams.input();
      
      setStaticWidgets(widgets);
      
      if (in != null)
      {
         // redirected input case
         promptFor(in, widgets);
      }
      else
      {
         // interactive case
         promptFor(widgets, (EventList) null);
      }
   }
   
   /**
    * For each entry in the widget array, the input stream will be read and 
    * the widget's associated field in the screen buffer will be set. Data
    * will not be copied to or from the user's variables.
    * <p>
    * Any error during stream reading or data parsing will cause an error
    * condition to be generated.  Reaching the end of the stream will cause
    * an end condition to be generated.
    * <p>
    * Both the input stream and the current default output stream (if output
    * is redirected) are checked.  If both are set to <code>echo</code> mode,
    * then after all stream input is done, the screen buffer's contents will  
    * be echoed to the current output destination using {@link #view}. This
    * will occur whether or not an error occurs during stream processing and
    * data parsing.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * Note that this method does change the LASTKEY value to 13 if input 
    * stream contains at least one character. If input stream is empty then
    * END condition is triggered. 
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    in
    *           A stream instance which is open for reading. 
    * @param    widgets
    *           The list of widgets which are to receive input from the
    *           stream.
    */
   @Override
   public void promptFor(Stream in, FrameElement[] widgets)
   {
      processStreamInput(in, widgets, false, false, null);
   }
   
   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is not used at all.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    */
   @Override
   public void promptFor(FrameElement[] widgets, EventList wait)   
   {
      promptForWorker(widgets, wait, false, null);
   }

   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.
    */
   @Override
   public void promptFor(GenericWidget<?>[] widgets)
   {
      promptFor(widgets, (EventList) null);
   }
   
   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    *
    * @param    widgets
    *           The list of widgets which is to be edited.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    */
   @Override
   public void promptFor(GenericWidget<?>[] widgets, EventList wait)
   {
      promptForWorker(widgets, wait, false, null);
   }
   
   /**
    * Enables a list of widgets from the frame on the screen, shifts into input blocking mode
    * via {@link #waitFor} and after the user's edits are complete the list of widgets are
    * disabled.  No data is copied to or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be marked as visible.
    * The frame will be brought to the top of the Z-order and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own execution to the
    * stream form.  In such a case, all interactive processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR UNLESS-HIDDEN
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion of these elements is
    *           not used at all.
    */
   @Override
   public void promptForUnlessHidden(FrameElement[] widgets)
   {
      FrameElement[] widgetsChanged = removeHiddenWidgets(widgets);
      
      if (widgetsChanged != null && widgetsChanged.length > 0)
      {
         promptFor(widgetsChanged);
      }
   }
   
   /**
    * Enables a list of widgets from the frame on the screen, shifts into input blocking mode
    * via {@link #waitFor} and after the user's edits are complete the list of widgets are
    * disabled.  No data is copied to or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be marked as visible.
    * The frame will be brought to the top of the Z-order and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR UNLESS-HIDDEN
    *
    * @param    widgets
    *           The list of widgets which is to be edited.
    */
   @Override
   public void promptForUnlessHidden(GenericWidget<?>[] widgets)
   {
      GenericWidget<?>[] widgetsChanged = removeHiddenWidgets(widgets);
      
      if (widgetsChanged != null && widgetsChanged.length > 0)
      {
         promptFor(widgetsChanged);
      }
   }
   
   /**
    * Scrolls the lines of the down frame up or down.
    * <p>
    * This method corresponds to these Progress statements:
    * SCROLL
    * <p>
    * The server-side screen buffer will be updated accordingly so it will
    * map the right values for the rows. 
    *
    * @param  fromCurrent
    *         <code>true</code> if FROM-CURRENT option was used
    * @param  upDown
    *         <code>true</code> if scrolling up
    */
   @Override
   public void scroll(boolean fromCurrent, boolean upDown)
   {
      int oldFrameLine = frameLineWorker();
      
      conditionalUpDown(true);
      LogicalTerminal.scroll(this, fromCurrent, upDown);
      
      // implement the scrolling behavior for the cached buffers
      // no row-drawing is needed because this is implemented on the
      // client-side
      int down  = dynamicDown;
      int start = 0;
      int end   = 0;
      int incr  = 0;

      int frameLine = 1;
      
      if (fromCurrent)
         frameLine = oldFrameLine;
      
      if (upDown)
      {
         incr = 1;
         start = frameLine;
         end = down - 1;
      }
      else
      {
         incr = -1;
         start = frameLine + 1;
         end = down;  
      }
      
      Map<Integer, ScreenBuffer> oldScreenBuffer = screenBufferCache;
      screenBufferCache = new HashMap<>();
      
      for (int i = 1; i < start - 1; i++)
      {         
         ScreenBuffer oldBuf = oldScreenBuffer.get(i);
         addScreenBufferToCache(oldBuf, i);
      }
      
      for (int newkey = start; newkey <= end; newkey++)
      {
         int oldkey = newkey + incr;
         
         ScreenBuffer oldBuf = oldScreenBuffer.get(oldkey);
         addScreenBufferToCache(oldBuf, newkey);
      }
      
      showScreenBuffer(frameLineWorker());
   }

   /**
    * Enables a list of widgets, shifts into input blocking mode via 
    * {@link #waitFor}, after the user's edits are complete the list of
    * widgets is disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements. 
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    data
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is only used on output (they are assigned
    *           to the values provided by the user).
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void set(FrameElement[] data, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         Stream in = UnnamedStreams.input();
         
         if (in != null)
         {
            // redirected input case
            set(in, data, hWin);
         }
         else
         {
            // interactive case
            set(data, null, hWin);
         }
      }
   }

   /**
    * For each entry in the widget array, the input stream will be read and 
    * the widget's associated field in the screen buffer will be set. Data
    * will be copied back from the screen buffer to the user's variables.
    * <p>
    * Any error during stream reading or data parsing will cause an error
    * condition to be generated.  Reaching the end of the stream will cause
    * an end condition to be generated.
    * <p>
    * Both the input stream and the current default output stream (if output
    * is redirected) are checked.  If both are set to <code>echo</code> mode,
    * then after all stream input is done, the screen buffer's contents will  
    * be echoed to the current output destination using {@link #view}. This
    * will occur whether or not an error occurs during stream processing and
    * data parsing.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    in
    *           A stream instance which is open for reading. 
    * @param    data
    *           The list of widgets which are to receive input from the
    *           stream.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void set(Stream in, FrameElement[] data, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         processStreamInput(in, data, true, true, hWin);
      }
   }
   
   /**
    * Enables a list of widgets, shifts into input blocking mode via 
    * {@link #waitFor}, after the user's edits are complete the list of
    * widgets is disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements. 
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is only used on output (they are assigned
    *           to the values provided by the user).
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void set(FrameElement[] widgets, EventList wait, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         setWorker(widgets, wait, false, true, hWin);
      }
   }
   
   /**
    * Enables a list of widgets, shifts into input blocking mode via {@link #waitFor}, after
    * the user's edits are complete the list of widgets is disabled and the edited data is copied
    * from the screen buffer to the listed frame elements.  The UNLESS-HIDDEN version of the
    * method. 
    * <p>
    * As part of the processing, the frame and each listed widget will be marked as visible.
    * The frame will be brought to the top of the Z-order and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the frame element list,
    * in the order in which they appear.  Otherwise, these do not affect the user interface
    * processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own execution to the
    * stream form.  In such a case, all interactive processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * SET UNLESS-HIDDEN
    *
    * @param    data
    *           The list of widgets which is to be edited.  The data portion of these elements
    *           is only used on output (they are assigned to the values provided by the user).
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void setUnlessHidden(FrameElement[] data, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         FrameElement[] dataChanged = removeHiddenWidgets(data);
         
         if (dataChanged != null && dataChanged.length > 0)
         {
            set(dataChanged, hWin);
         }
      }
   }

   /**
    * Enables a list of widgets, shifts into input blocking mode via 
    * {@link #waitFor}, after the user's edits are complete the list of
    * widgets is disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements. 
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    data
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is only used on output (they are assigned
    *           to the values provided by the user).
    */
   @Override
   public void set(FrameElement[] data)
   {
      Stream in = UnnamedStreams.input();
      
      if (in != null)
      {
         // redirected input case
         set(in, data);
      }
      else
      {
         // interactive case
         set(data, (EventList) null);
      }
   }
   
   /**
    * For each entry in the widget array, the input stream will be read and 
    * the widget's associated field in the screen buffer will be set. Data
    * will be copied back from the screen buffer to the user's variables.
    * <p>
    * Any error during stream reading or data parsing will cause an error
    * condition to be generated.  Reaching the end of the stream will cause
    * an end condition to be generated.
    * <p>
    * Both the input stream and the current default output stream (if output
    * is redirected) are checked.  If both are set to <code>echo</code> mode,
    * then after all stream input is done, the screen buffer's contents will  
    * be echoed to the current output destination using {@link #view}. This
    * will occur whether or not an error occurs during stream processing and
    * data parsing.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * Note that this method does change the LASTKEY value to 13 if input 
    * stream contains at least one character. If input stream is empty then
    * END condition is triggered. 
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    in
    *           A stream instance which is open for reading. 
    * @param    data
    *           The list of widgets which are to receive input from the
    *           stream.
    */
   @Override
   public void set(Stream in, FrameElement[] data)
   {
      processStreamInput(in, data, true, false, null);
   }
   
   /**
    * Enables a list of widgets, shifts into input blocking mode via 
    * {@link #waitFor}, after the user's edits are complete the list of
    * widgets is disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements. 
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is only used on output (they are assigned
    *           to the values provided by the user).
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    */
   @Override
   public void set(FrameElement[] widgets, EventList wait)
   {
      setWorker(widgets, wait, false, false, null);
   }
   
   /**
    * Enables a list of widgets, shifts into input blocking mode via {@link #waitFor}, after
    * the user's edits are complete the list of widgets is disabled and the edited data is copied
    * from the screen buffer to the listed frame elements.  The UNLESS-HIDDEN version of the
    * method. 
    * <p>
    * As part of the processing, the frame and each listed widget will be marked as visible.
    * The frame will be brought to the top of the Z-order and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the frame element list,
    * in the order in which they appear.  Otherwise, these do not affect the user interface
    * processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own execution to the
    * stream form.  In such a case, all interactive processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * SET UNLESS-HIDDEN
    *
    * @param    data
    *           The list of widgets which is to be edited.  The data portion of these elements
    *           is only used on output (they are assigned to the values provided by the user).
    */
   @Override
   public void setUnlessHidden(FrameElement[] data)
   {
      FrameElement[] dataChanged = removeHiddenWidgets(data);
      
      if (dataChanged != null && dataChanged.length > 0)
      {
         set(dataChanged);
      }
   }
   
   /**
    * Positions the cursor on a new line in a down frame.
    * This is a conditional call. Multiple conditional up() calls have
    * the effect of a single up(1) call, which is delayed till the nearest
    * terminal I/O operation and can be cancelled by a conditional down()
    * call.
    * <p>
    * This method corresponds to these Progress statements:
    * UP
    */
   @Override
   public void up()
   {
      if (isRedirected())
      {
         // this is a NOP
         return;
      }

      if (hadView && pendingUpDown > -1)
      {
         int frameLine = LogicalTerminal.getFrameLine(this, pendingUpDown).intValue();
         
         pendingUpDown--;
         
         // non-down frames maintain the pending value but don't swap buffers
         if (!frame.isDown())
            return;
         
         // we add the screen buffer for the previous frame line
         addScreenBufferToCache(this.frameBuf, frameLine);

         // restore the buffer for the next row, but don't send it to the 
         // client
         int nextLine = frameLine - 1;
         frameBuf = getScreenBufferFromCache(nextLine);
      }
   }

   /**
    * Positions the cursor on a new line of the down frame.
    * <p>
    * This method corresponds to these Progress statements:
    * UP n
    *
    * @param  lines
    *         number of iterations of the data to scroll up (positive) or
    *         down (negative).
    */
   @Override
   public void up(double lines)
   {
      up((int) lines);
   }

   /**
    * Positions the cursor on a new line of the down frame.
    * <p>
    * This method corresponds to these Progress statements:
    * UP n
    *
    * @param  lines
    *         number of iterations of the data to scroll up (positive) or
    *         down (negative).
   */
   @Override
   public void up(NumberType lines)
   {
      up(lines.intValue());
   }

   /**
    * This method exists to support a valid form of the Progress 
    * <code>UP</code> statement which has no effect.
    * <p>
    * This method corresponds to these Progress statements:
    * UP STREAM s
    *
    * @param    out
    *           A stream instance which is open for writing. 
    */
   @Override
   public void up(Stream out)
   {
      // must be before any other checks
      if (out != null && out.protectWrites())
      {
         return;
      }
      
      // otherwise this is a NOP
      return;
   }

   /**
    * If <code>lines</code> is a negative number, the absolute value of that
    * number of new-lines will be written to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * UP STREAM s n
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    lines
    *           If negative, the number of new-lines to output to the stream.
    *           Otherwise, this has no effect.
    */
   @Override
   public void up(Stream out, double lines)
   {
      // must be before any other checks
      if (out != null && out.protectWrites())
      {
         return;
      }
      
      if (lines > -1)
      {
         // this is a NOP
         return;
      }
      
      up((int) lines, getRemoteStreamId(out));
   }

   /**
    * If <code>lines</code> is a negative number, the absolute value of that
    * number of new-lines will be written to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * UP STREAM s n
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    lines
    *           If negative, the number of new-lines to output to the stream.
    *           Otherwise, this has no effect.
    */
   @Override
   public void up(Stream out, NumberType lines)
   {
      up(out, lines.intValue());
   }

   /**
    * Copies data from the frame elements to each corresponding widget, then
    * enables the list of widgets, shifts into input blocking mode via
    * {@link #waitFor}, after the user's edits are complete the list of 
    * widgets are disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements.  
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method uses {@link #display} and {@link #set} to do its real
    * work.  The only difference is that the <code>set</code> is embedded
    * inside a "hidden" <code>TransactionManager</code> block which handles
    * undo/retry on <code>error</code> conditions and undo/leave on 
    * <code>end</code> conditions.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    *
    * @param    data
    *           The list of data and widgets which is to be updated.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void update(FrameElement[] data, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         Stream in = UnnamedStreams.input();
         
         if (in != null)
         {
            // redirected input case
            updateWorker(in, data, true, hWin);
         }
         else
         {
            // interactive case
            updateWorker(data, null, true, hWin);
         }
      }
   }

   /**
    * For each entry in the widget array, the input stream will be read and 
    * the widget's associated field in the screen buffer will be set. The 
    * data passed in will be copied from the user's variables into the screen
    * buffer.  After the input is read from the stream into the screen buffer,
    * that data will be copied back from the screen buffer to the user's
    * variables.  
    * <p>
    * Any error during stream reading or data parsing will cause an error
    * condition to be generated.  Reaching the end of the stream will cause
    * an end condition to be generated.  Unlike the <code>update</code>
    * version in this class which has a "hidden" retry block around the
    * core "set" processing, this streams version has no such feature. Errors
    * will be immediately seen by the caller.
    * <p>
    * Both the input stream and the current default output stream (if output
    * is redirected) are checked.  If both are set to <code>echo</code> mode,
    * then after all stream input is done, the screen buffer's contents will  
    * be echoed to the current output destination using {@link #view}. This
    * will occur whether or not an error occurs during stream processing and
    * data parsing.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    *
    * @param    in
    *           A stream instance which is open for reading. 
    * @param    data
    *           The list of widgets which are to receive input from the
    *           stream.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void update(Stream in, FrameElement[] data, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         updateWorker(in, data, true, hWin);
      }
   }

   /**
    * Copies data from the frame elements to each corresponding widget, then
    * enables the list of widgets, shifts into input blocking mode via
    * {@link #waitFor}, after the user's edits are complete the list of 
    * widgets are disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements.  
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method uses {@link #display} and {@link #set} to do its real
    * work.  The only difference is that the <code>set</code> is embedded
    * inside a "hidden" <code>TransactionManager</code> block which handles
    * undo/retry on <code>error</code> conditions and undo/leave on 
    * <code>end</code> conditions.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    *
    * @param    data
    *           The list of data and widgets which is to be updated.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void update(FrameElement[] data, EventList wait, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         updateWorker(data, wait, true, hWin);
      }
   }
   
   /**
    * Copies data from the frame elements to each corresponding widget, then enables the list of
    * widgets, shifts into input blocking mode via {@link #waitFor}, after the user's edits are
    * complete the list of widgets are disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements.  
    * <p>
    * As part of the processing, the frame and each listed widget will be marked as visible.
    * The frame will be brought to the top of the Z-order and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method uses {@link #display} and {@link #set} to do its real work.  The only difference
    * is that the <code>set</code> is embedded inside a "hidden" <code>TransactionManager</code>
    * block which handles undo/retry on <code>error</code> conditions and undo/leave on
    * <code>end</code> conditions.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the frame element list,
    * in the order in which they appear.  Otherwise, these do not affect the user interface
    * processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own execution to the
    * stream form.  In such a case, all interactive processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE UNLESS-HIDDEN
    *
    * @param    data
    *           The list of data and widgets which is to be updated.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void updateUnlessHidden(FrameElement[] data, handle hWin)
   {
      if (isAbleToDisplayWindow(hWin))
      {
         FrameElement[] dataChanged = removeHiddenWidgets(data);
         
         if (dataChanged != null && dataChanged.length > 0)
         {
            updateWorker(dataChanged, null, true, hWin);
         }
      }
   }
   
   /**
    * Copies data from the frame elements to each corresponding widget, then
    * enables the list of widgets, shifts into input blocking mode via
    * {@link #waitFor}, after the user's edits are complete the list of 
    * widgets are disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements.  
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method uses {@link #display} and {@link #set} to do its real
    * work.  The only difference is that the <code>set</code> is embedded
    * inside a "hidden" <code>TransactionManager</code> block which handles
    * undo/retry on <code>error</code> conditions and undo/leave on 
    * <code>end</code> conditions.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    *
    * @param    data
    *           The list of data and widgets which is to be updated.
    */
   @Override
   public void update(FrameElement[] data)
   {
      Stream in = UnnamedStreams.input();
      
      if (in != null)
      {
         // redirected input case
         updateWorker(in, data, false, null);
      }
      else
      {
         // interactive case
         updateWorker(data, null, false, null);
      }
   }
   
   /**
    * For each entry in the widget array, the input stream will be read and 
    * the widget's associated field in the screen buffer will be set. The 
    * data passed in will be copied from the user's variables into the screen
    * buffer.  After the input is read from the stream into the screen buffer,
    * that data will be copied back from the screen buffer to the user's
    * variables.  
    * <p>
    * Any error during stream reading or data parsing will cause an error
    * condition to be generated.  Reaching the end of the stream will cause
    * an end condition to be generated.  Unlike the <code>update</code>
    * version in this class which has a "hidden" retry block around the
    * core "set" processing, this streams version has no such feature. Errors
    * will be immediately seen by the caller.
    * <p>
    * Both the input stream and the current default output stream (if output
    * is redirected) are checked.  If both are set to <code>echo</code> mode,
    * then after all stream input is done, the screen buffer's contents will  
    * be echoed to the current output destination using {@link #view}. This
    * will occur whether or not an error occurs during stream processing and
    * data parsing.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * Note that this method does change the LASTKEY value to 13 if input 
    * stream contains at least one character. If input stream is empty then
    * END condition is triggered. 
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    *
    * @param    in
    *           A stream instance which is open for reading. 
    * @param    data
    *           The list of widgets which are to receive input from the
    *           stream.
    */
   @Override
   public void update(Stream in, FrameElement[] data)
   {
      updateWorker(in, data, false, null);
   }

   /**
    * Copies data from the frame elements to each corresponding widget, then
    * enables the list of widgets, shifts into input blocking mode via
    * {@link #waitFor}, after the user's edits are complete the list of 
    * widgets are disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements.  
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method uses {@link #display} and {@link #set} to do its real
    * work.  The only difference is that the <code>set</code> is embedded
    * inside a "hidden" <code>TransactionManager</code> block which handles
    * undo/retry on <code>error</code> conditions and undo/leave on 
    * <code>end</code> conditions.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    *
    * @param    widgets
    *           The list of data and widgets which is to be updated.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    */
   @Override
   public void update(FrameElement[] widgets, EventList wait)
   {
      updateWorker(widgets, wait, false, null);
   }
   
   /**
    * Copies data from the frame elements to each corresponding widget, then enables the list of
    * widgets, shifts into input blocking mode via {@link #waitFor}, after the user's edits are
    * complete the list of widgets are disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements.  
    * <p>
    * As part of the processing, the frame and each listed widget will be marked as visible.
    * The frame will be brought to the top of the Z-order and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method uses {@link #display} and {@link #set} to do its real work.  The only difference
    * is that the <code>set</code> is embedded inside a "hidden" <code>TransactionManager</code>
    * block which handles undo/retry on <code>error</code> conditions and undo/leave on
    * <code>end</code> conditions.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the frame element list,
    * in the order in which they appear.  Otherwise, these do not affect the user interface
    * processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own execution to the
    * stream form.  In such a case, all interactive processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE UNLESS-HIDDEN
    *
    * @param    data
    *           The list of data and widgets which is to be updated.
    */
   @Override
   public void updateUnlessHidden(FrameElement[] data)
   {
      FrameElement[] dataChanged = removeHiddenWidgets(data);
      
      if (dataChanged != null && dataChanged.length > 0)
      {
         updateWorker(dataChanged, null, false, null);
      }
   }
   
   /**
    * Makes the frame visible.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW FRAME
    * 
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void view(handle hWin)
   {
      conditionalUpDown();

      Stream  out    = UnnamedStreams.output();
      boolean bypass = false;
      
      // do we have an interactive terminal (not redirected output)
      
      if (out == null)
      {
         makeFrameVisible();
      }
      else
      {
         bypass = isPageElement(out);
         
         if (bypass)
         {
            LogicalTerminal.registerPageElement(frameId, getPageTop());
         }
      }
      
      if (!bypass)
      {
         LogicalTerminal.view(this, new int[] {frameId}, hWin);
      }
   }
   
   /**
    * Makes widgets in frame visible.
    * 
    * @param    data
    *           The list of data and widgets which is to be viewed.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void view(FrameElement[] data, handle hWin)
   {
      view(data, true, hWin);
   }
   
   /**
    * Writes the current screen-buffer contents to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW STREAM s FRAME
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   @Override
   public void view(Stream out, handle hWin)
   {
      // TODO: implement this
   }
   
   /**
    * Makes the frame visible.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW FRAME
    */
   @Override
   public void view()
   {
      if (defaultFrame && widgets.length == 0)
      {
         // for default frame with no widgets this is a no-op
         return;
      }

      conditionalUpDown();

      Stream  out    = UnnamedStreams.output();
      boolean bypass = false;
      
      // do we have an interactive terminal (not redirected output)
      
      if (out == null)
      {
         makeFrameVisible();
      }
      else
      {
         bypass = isPageElement(out);
         
         if (bypass)
         {
            LogicalTerminal.registerPageElement(frameId, getPageTop(), widgets.length == 0);
         }
      }
      
      if (!bypass)
      {
         LogicalTerminal.view(this, new int[] {frameId});
      }
   }
   
   /**
    * Makes visible specified widgets.
    * 
    * @param   data
    *          Array of FrameElement instances which need to be made visible.
    */
   @Override
   public void view(FrameElement[] data)
   {
      view(data, false, null);
   }

   /**
    * Writes the current screen-buffer contents to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW STREAM s FRAME
    *
    * @param    out
    *           A stream instance which is open for writing. 
    */
   @Override
   public void view(Stream out)
   {
      // No input data means we use the current frame data
      sameViewDataRedirected = true;
      viewWorker(out, null, true, false, null);
   }
   
   /**
    * Writes the current screen-buffer contents to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW STREAM s FRAME
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param   data
    *          Array of FrameElement instances which need to be made visible.
    */
   public void view(Stream out, FrameElement[] data)
   {
      // New input data - reset the flag
      sameViewDataRedirected = false;
      viewWorker(out, data, true, false, null);
   }
   
   /**
    * Makes visible specified widgets. Prepares those Ids for client call.
    * 
    * @param    data
    *           Array of FrameElement instances which need to be made visible.
    *           
    * @return   Ids of specified widgets.
    */
   private int[] prepareView(FrameElement[] data)
   {
      if (!frame._isHidden() || currentStatement != UIStatement.DISPLAY)
      {
         makeFrameVisible();
      }

      int sz = 0;
      
      for (int i = 0; i < data.length; i++)
         if (data[i] != null)
            sz++;
      
      int[] ids = new int[sz];
      
      boolean inDisplayStmt = currentStatement == UIStatement.DISPLAY;
      
      for (int i = 0, j = 0; i < data.length; i++)
      {
         if (data[i] != null)
         {
            GenericWidget gw = data[i].getWidget();
            // we need to reset the HIDDEN attribute to FALSE within DISPLAY statement
            if (inDisplayStmt)
            {
               final WidgetConfig wcfg = gw.config();
               gw.setAttr("hidden", wcfg.hidden, false, (vv) -> wcfg.hidden = vv);
            }
            ids[j++] = gw.getId();
         }
      }
      return ids;
   }
   
   /**
    * Writes the current screen-buffer contents to the given stream.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW [STREAM s] FRAME [IN WINDOW] and DISPLAY [STREAM s] ... [IN WINDOW].
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    data
    *           Array of FrameElement instances which need to be made visible.
    * @param    view
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    * @param    windowClause
    *           <code>true</code> if IN WINDOW clause is presented in ENABLE statement.
    *           <code>false</code> otherwise.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    *           <code>null</code> means that LogicalTerminal method without hWin will be used. 
    */
   private void viewWorker(Stream         out, 
                           FrameElement[] data, 
                           boolean        view, 
                           boolean        windowClause, 
                           handle         hWin)
   {
      if (out != null)
      {
         // strange case where the error messages are doubled (must be 
         // something deep inside the Progress UI implementation) which occurs
         // for all DISPLAY and any non-frame VIEW that has > 1 widget in the
         // list
         if (!view || (data != null && data.length > 1))
         {
            try
            {
               if (out.protectWrites())
               {
                  // silent error mode is ON (this is NO-ERROR) so the error
                  // output is suppressed and doesn't need to be doubled,
                  // just exit
                  return;
               }
            }
            
            catch (ErrorConditionException ece)
            {
               // ignore because we have to do this again
            }
         }
         
         // normal safety code here, exit in NO-ERROR mode if needed
         if (out.protectWrites())
            return;
      }
      
      conditionalUpDown();

      // if needed, temporarily switch the terminal to a new output
      // destination on the client side
      int sid = getRemoteStreamId(out); 
      
      boolean displayBypass = false;
      boolean register      = false;
      boolean pageTop       = false;

      redirOut = out;
      
      // check if we must register as a page element
      if (isPageElement(out))
      {
         displayBypass = view;
         register      = true;
         pageTop       = getPageTop();
         
         // if we will bypass the view() below, then we need to at least
         // register the page element here
         if (displayBypass)
         {
            LogicalTerminal.registerPageElement(frameId, pageTop, sid);
         }
      }
      
      // output the screen-buffer to the current destination
      if (!displayBypass)
      {
         int[] ids = null;
         if (data != null)
         {
            GenericWidget<?>[] wids = widgetListFromFrameElements(data);
            ids = idsFromWidgets(wids);
         }
         else
         {
            ids = new int[] {frameId};
         }

         if (hWin != null)
         {
            LogicalTerminal.view(this, ids, sid, register, pageTop, view, hWin);
         }
         else
         {
            LogicalTerminal.view(this, ids, sid, register, pageTop, view);
         }
      }
   }
   
   /**
    * Displays (writes) a list of fields from the frame on the given stream.
    * As part of this operation, the list of data values are copied to the
    * screen buffer.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY ... [IN WINDOW].
    *
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    * @param    windowClause
    *           <code>true</code> if IN WINDOW clause is presented in ENABLE statement.
    *           <code>false</code> otherwise.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   private void display(Stream out, FrameElement[] data, boolean windowClause, handle hWin)
   {
      // set the value in the screen buffer for each widget using the widget
      // id as the index
      copyToScreenBuffer(data);

      viewWorker(out, data, false, windowClause, hWin);

      // handle aggregation if needed
      processAggregates(data, true);
   }
   
   /**
    * Enables input for the given field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: ENABLE
    * 
    * @param    list
    *           The widgets to enable.
    * @param    windowClause
    *           <code>true</code> if IN WINDOW clause is presented in ENABLE statement.
    *           <code>false</code> otherwise.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   private void enable(GenericWidget<?>[] list, boolean windowClause, handle hWin)
   {
      // Special processing for ENABLE statement in batch mode.
      if (errorOnBatchModeEnable(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in batch mode
         return;
      }

      if (errorOnRedirectedInput(4018, ENABLE_INPUT_REDIRECTED))
      {
         // failure in silent error mode
         return;
      }
          
      if (list == null || list.length == 0)
      {
         return;
      }
      
      // translate the passed-in widgets to widget ids,
      // we must keep the duplicate widget references and order
      List<Integer> widgetIds = new ArrayList<>(list.length);
      
      for (int i = 0; i < list.length; i++)
      {
         // null values are silently ignored
         if (list[i] == null)
         {
            continue;
         }
         
         int id = getWidgetId(list[i]);
         
         // invalid widget
         if (id < 0)
         {
            continue;
         }
         
         widgetIds.add(id);
      }
      
      if (widgetIds.isEmpty())
      {
         // no valid widgets
         return;
      }
      
      // enable widgets
      
      for (int i = 0; i < list.length; i++)
      {
         // null values are silently ignored
         if (list[i] == null)
         {
            continue;
         }

         list[i].finishSetup();
      }
      
      // translate ids to int[]
      int[] ids = new int[widgetIds.size()];
      int i = 0;
      for (Integer id : widgetIds)
      {
         ids[i++] = id;
      }

      if (windowClause)
         LogicalTerminal.enable(this, ids, hWin);
      else
         LogicalTerminal.enable(this, ids);
   }
   
   /**
    * Makes widgets in frame visible.
    * 
    * @param    data
    *           The list of data and widgets which is to be viewed.
    * @param    windowClause
    *           <code>true</code> if IN WINDOW clause is presented in VIEW statement.
    *           <code>false</code> otherwise.
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   private void view(FrameElement[] data, boolean windowClause, handle hWin)
   {
      conditionalUpDown();

      if (UnnamedStreams.output() != null || data == null)
      {
         // TODO: This call probably will need to be targeted at a specific window, but for now
         // we don't have redirected mode working for GUI, so it is unclear what to do.         
         view();
         
         return;
      }

      int[] ids = prepareView(data);
    
      if (windowClause)
         LogicalTerminal.view(this, ids, hWin);
      else
         LogicalTerminal.view(this, ids);
   }
   
   /**
    * Returns the column position of the left corner of the frame within its
    * window.
    * <p>
    * This method corresponds to the Progress built-in function
    * FRAME-COL
    * 
    * @return  current value of the frame column.
    */
   @Override
   public decimal frameCol()
   {
      return LogicalTerminal.getFrameCol(this);
   }

   /**
    * Returns the number of iterations in a down frame.
    * <p>
    * This method corresponds to the Progress built-in function
    * FRAME-DOWN
    * 
    * @return  current value of the frame DOWN value.
    */
   @Override
   public integer frameDown()
   {
      integer fd = LogicalTerminal.getFrameDown(this);
      
      return fd;
   }

   /**
    * Returns the current logical line number in a frame.
    * <p>
    * This method corresponds to the Progress built-in function
    * FRAME-LINE
    * 
    * @return   1-based ordinal of the current data iteration of the frame.
    *           If the frame is redirected to a stream or if the frame is
    *           not visible, then 0 is returned. If the frame is not a down
    *           frame then 1 will be returned.
    */
   @Override
   public integer frameLine()
   {
      return  new integer(frameLineWorker());
   }

   /**
    * Returns the row position of the upper-left corner of the frame within 
    * its window.
    * <p>
    * This method corresponds to the Progress built-in function
    * FRAME-ROW
    * 
    * @return  current value of the frame row.
    */
   @Override
   public decimal frameRow()
   {
      return LogicalTerminal.getFrameRow(this);
   }

   /**
    * Sets the ATTR-SPACE option.
    * 
    * @param   attr_space
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setAttrSpace(boolean attr_space)
   {
      frame.config.attr_space = attr_space;
      frame.pushWidgetAttr("attr_space", frame.config.attr_space);
   }
   
   /**
    * Get the CAREFUL-PAINT attribute.
    * 
    * @return   <code>true</code> if the attribute is ON.
    */
   public logical isCarefulPaint()
   {
      return frame.isCarefulPaint();
   }
   
   /**
    * Sets the CAREFUL-PAINTattribute.
    * 
    * @param   careful
    *          <code>true</code> if the attribute is set to ON.
    */
   public void setCarefulPaint(logical careful)
   {
      frame.setCarefulPaint(careful);
   }
   
   /**
    * Sets the CAREFUL-PAINTattribute.
    * 
    * @param   careful
    *          <code>true</code> if the attribute is set to ON.
    */
   public void setCarefulPaint(boolean careful)
   {
      frame.setCarefulPaint(careful);
   }
   
   /**
    * Sets the CENTERED option.
    * 
    * @param   centered
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setCentered(boolean centered)
   {
      frame.config.centered = centered;
      frame.pushWidgetAttr("centered", frame.config.centered);
   }

   /**
    * Sets the display color.
    * 
    * @param   color
    *          color number
    */
   @Override
   public void setDColor(Color color)
   {
      frame.config.dcolor = color;
      frame.pushWidgetAttr("dcolor", frame.config.dcolor);
   }

   /**
    * Sets the display color.
    * 
    * @param   color
    *          color number
    */
   @Override
   public void setDColor(int color)
   {
      UnimplementedFeature.missing("DCOLOR n");
   }

   /**
    * Sets the input (prompt-for) color.
    * 
    * @param   color
    *          color number
    */
   @Override
   public void setPfColor(Color color)
   {
      frame.config.pfcolor = color;
      frame.pushWidgetAttr("pfcolor", frame.config.pfcolor);
   }

   /**
    * Sets the display color.
    * 
    * @param   cs
    *          color specification
    */
   @Override
   public void setDColor(ColorSpec cs)
   {
      frame.colorDisplay(cs);
   }

   /**
    * Sets the input (prompt-for) color.
    * 
    * @param   cs
    *          color specification
    */
   @Override
   public void setPfColor(ColorSpec cs)
   {
      frame.colorPrompt(cs);
   }

   /**
    * Get display color.
    * 
    * @return  display color.
    */
   @Override
   public integer getBgColor()
   {
      int color = frame.getAttr("bgcolor", () -> frame.config.bgcolor);
      return color == -1 ? new integer() : new integer(color);
   }

   /**
    * Sets the display color.
    * 
    * @param   color
    *          color number
    */
   @Override
   public void setBgColor(NumberType color)
   {
      setBgColor(color == null || color.isUnknown() ? -1 : color.intValue());
   }

   /**
    * Sets the display color.
    * 
    * @param   color
    *          color number
    */
   @Override
   public void setBgColor(int color)
   {
      frame.setBgColor(color);
   }

   /**
    * Get input (prompt-for) color.
    * 
    * @return  input (prompt-for) color.
    */
   @Override
   public integer getFgColor()
   {
      int color = frame.getAttr("fgcolor", () -> frame.config.fgcolor);
      return color == -1 ? new integer() : new integer(color);
   }

   /**
    * Sets the input (prompt-for) color.
    * 
    * @param   color
    *          color number
    */
   @Override
   public void setFgColor(NumberType color)
   {
      setFgColor(color == null || color.isUnknown() ? 0 : color.intValue());
   }

   /**
    * Sets the input (prompt-for) color.
    * 
    * @param   color
    *          color number
    */
   @Override
   public void setFgColor(int color)
   {
      frame.setFgColor(color);
   }

   /**
    * Get display color RGB option.
    * 
    * @return  display color RGB.
    */
   @Override
   public integer getBgColorRgb()
   {
      int color = frame.getAttr("bgcolorRgb", () -> frame.config.bgcolorRgb);
      return color == -1 ? new integer() : new integer(color);
   }

   /**
    * Sets the display color RGB option.
    * 
    * @param   colorRgb
    *          color RGB
    */
   @Override
   public void setBgColorRgb(NumberType colorRgb)
   {
      setBgColorRgb(colorRgb == null || colorRgb.isUnknown() ? -1 : colorRgb.intValue());
   }

   /**
    * Sets the display color RGB option.
    * 
    * @param   colorRgb
    *          color number
    */
   @Override
   public void setBgColorRgb(int colorRgb)
   {
      frame.setBgColorRgb(colorRgb);
   }

   /**
    * Get input (prompt-for) color RGB option.
    * 
    * @return  input (prompt-for) color RGB.
    */
   @Override
   public integer getFgColorRgb()
   {
      int color = frame.getAttr("fgcolorRgb", () -> frame.config.fgcolorRgb);
      return color == -1 ? new integer() : new integer(color);
   }

   /**
    * Sets the input (prompt-for) color RGB option.
    * 
    * @param   colorRgb
    *          color RGB
    */
   @Override
   public void setFgColorRgb(NumberType colorRgb)
   {
      setFgColorRgb(colorRgb == null || colorRgb.isUnknown() ? 0 : colorRgb.intValue());
   }

   /**
    * Sets the input (prompt-for) color RGB option.
    * 
    * @param   colorRgb
    *          color RGB
    */
   @Override
   public void setFgColorRgb(int colorRgb)
   {
      frame.setFgColorRgb(colorRgb);
   }

   /**
    * Sets the frame column as a callback to read the current value
    * calculated by the given expression. This value is honored at the time
    * that the frame is made visible (for the first time or after being
    * hidden).
    * 
    * @param    column
    *           The expression to evaluate for the frame column.
    */
   @Override
   public void setColumn(Resolvable column)
   {
      dynColumn = column;
   }
   
   /**
    * Sets the frame column as a callback to read the current value stored in
    * the given <code>Accessor</code>. This value is honored at the time that
    * the frame is made visible (for the first time or after being hidden).
    * 
    * @param    column
    *           The instance variable to read for the frame column.
    */
   @Override
   public void setColumn(Accessor column)
   {
      setColumn(new ExpressionWrapper(column));
   }
   
   /**
    * Sets the column position for this frame (statically).
    *
    * @param    column
    *           The 1-based column number at which to start displaying the 
    *           frame.  This is converted to a 0-based column number for the
    *           internal runtime.
    */
   @Override
   public void setColumn(NumberType column)
   {
      frame.setColumn(column);
   }

   /**
    * Sets the column position for this frame (statically).
    *
    * @param    column
    *           The 1-based column number at which to start displaying the 
    *           frame.  This is converted to a 0-based column number for the
    *           internal runtime.
    */
   @Override
   public void setColumn(double column)
   {
      frame.setColumn(column);
   }

   /**
    * Returns the column position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress <code>FRAME:COLUMN</code> 
    * attribute.
    * 
    * @return   The column position for the upper-left corner of the frame.
    */
   @Override
   public decimal getColumn()
   {
      return frame.getColumn();
   }

   /**
    * Returns the row position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress <code>FRAME:ROW</code>
    * attribute.
    * 
    * @return   The row position for the upper-left corner of the frame.
    */
   @Override
   public decimal getRow()
   {
      return frame.getRow();
   }
   
   /**
    * Sets the number of columns.
    * 
    * @param   columns
    *          number of columns
    */
   @Override
   public void setColumns(int columns)
   {
      FrameConfig c = frame.config;
      c.columns = columns;
      
      // This statement is intended to implicitly turn on side label mode.
      // Without this the numeric FillIns will have wrong alignment
      // to the right side
      if (columns > 0)
      {
         c.sideLabels = true;
      }

      frame.pushWidgetAttr(new String[] {"columns", "sideLabels"},
                           new Object[] {c.columns, c.sideLabels});
   }

   /**
    * Get current state of VIEW-AS DIALOG-BOX option.
    *
    * @return  <code>true</code> if config is a DialogBox config.
    */
   public boolean isDialogBox()
   {
      return frame.getAttr("dialog", () -> frame.config.dialog);
   }
   
   /**
    * Sets the VIEW-AS DIALOG-BOX option.
    * 
    * @param   dialogBox
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setDialogBox(boolean dialogBox)
   {
      frame.config.dialog = dialogBox;
      if (dialogBox)
      {
         frame.config.box = LogicalTerminal.isChui();
      }
   }

   /**
    * Sets the size of a DOWN frame as a callback to read the current value
    * calculated by the given expression.  This value is then honored at the
    * time that the frame is made visible (for the first time or after being
    * hidden).
    *
    * @param    down
    *           The expression to evaluate for the number of data iterations
    *           in the current (DOWN) frame.
    */
   @Override
   public void setDown(Resolvable down)
   {
      if (frame._isRealized())
      {
         // this is a no-op in case of shared frames - the DOWN value is inherited from the source
         // frame.  DOWN can only be overridden by explicitly setting the FRAME's DOWN attribute
         return;
      }
      
      dynDown = down;
      frame.setDown(0);
   }
   
   /**
    * Sets the size of a DOWN frame as a callback to read the current value
    * stored in the given <code>Accessor</code>.  This value is then honored
    * at the time that the frame is made visible (for the first time or after
    * being hidden).
    *
    * @param    down
    *           The simple instance variable to read for the number of data
    *           iterations in the current (DOWN) frame.
    */
   @Override
   public void setDown(Accessor down)
   {
      setDown(new ExpressionWrapper(down));
   }

   /**
    * Sets the size of a DOWN frame (statically).
    *
    * @param    down
    *           The number of data iterations in the DOWN frame.
    */
   @Override
   public void setDown(NumberType down)
   {
      setDown(down.intValue());
   }

   /**
    * Sets the size of a DOWN frame (statically).
    *
    * @param    down
    *           The number of data iterations in the DOWN frame.
    */
   @Override
   public void setDown(double down)
   {
      frame.setDown((int) down);
   }

   /**
    * Sets the height of the widget.
    *
    * @param   height
    *          height of the widget
    */
   @Override
   public void setHeightChars(double height)
   {
      frame.setHeightChars(height);
   }

   /**
    * Sets the height of the widget.
    *
    * @param   height
    *          height of the widget
    */
   @Override
   public void setHeightChars(NumberType height)
   {
      frame.setHeightChars(height);
   }

   /**
    * Returns the height of the widget.
    *
    * @return  height of the widget, retrieved from the 
    *          <code>FRAME:HEIGHT-CHARS</code> attribute.
    */
   @Override
   public decimal getHeightChars()
   {
      return frame.getHeightChars();
   }

   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param   heightPixels
    *          new value of WIDTH-PIXELS attribute
    */
   @Override
   public void setHeightPixels(int heightPixels)
   {
      frame.setHeightPixels(heightPixels);
   }

   /**
    * Sets new value of HEIGHT-PIXELS attribute.
    *
    * @param   heightPixels
    *          new value of WIDTH-PIXELS attribute
    */
   @Override
   public void setHeightPixels(NumberType heightPixels)
   {
      frame.setHeightPixels(heightPixels);
   }

   /**
    * Returns the width of the widget in pixels. Corresponds Progress
    * HEIGHT-PIXELS attribute.
    *
    * @return The width of the widget in pixels.
    */
   @Override
   public integer getHeightPixels()
   {
      return frame.getHeightPixels();
   }

   /**
    * Sets the KEEP-TAB-ORDER option.
    * 
    * @param   keepTabOrder
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setKeepTabOrder(boolean keepTabOrder)
   {
      frame.config.keepTabOrder = keepTabOrder;
      frame.pushWidgetAttr("keepTabOrder", frame.config.keepTabOrder);
   }

   /**
    * Sets the THREE-D option.
    *
    * @param threeD
    *        <code>true</code> if the option is set to ON.
    */
   @Override
   public void setThreeD(boolean threeD)
   {
      if (frame.config.threeD == threeD)
      {
         return;
      }
      
      frame.config.threeD = threeD;
      frame.pushWidgetAttr("threeD", frame.config.threeD);
   }

   /**
    * Sets DROP-TARGET option.
    * 
    * @param dropTarget
    *        <code>true</code> if the option is set to ON.
    */
   @Override
   public void setDropTarget(boolean dropTarget)
   {
      if (frame.config.dropTarget == dropTarget)
      {
         return;
      }
      
      frame.config.dropTarget = dropTarget;
      frame.pushWidgetAttr("dropTarget", frame.config.dropTarget);
   }
   
   /**
    * Sets the NO-BOX option.
    * 
    * @param   noBox
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoBox(boolean noBox)
   {
      if (frame.config.dialog)
      {
         // can't set this for DIALOG-BOX, even if it appears at the frame definition.
         noBox = !LogicalTerminal.isChui();
      }
      
      frame.config.box = !noBox;
      frame.pushWidgetAttr("box", frame.config.box);
   }

   /**
    * Sets the NO-HIDE option.
    * 
    * @param   noHide
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoHide(boolean noHide)
   {
      FrameConfig c = frame.config;  
      c.noHide = noHide;
      c.noHideConfig = noHide;
      frame.pushWidgetAttr(new String[] {"noHide", "noHideConfig"},
                           new Object[] {c.noHide, c.noHideConfig});
   }

   /**
    * Sets the NO-LABELS option.
    * 
    * @param   noLabels
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoLabels(boolean noLabels)
   {
      frame.config.labels = !noLabels;
      frame.pushWidgetAttr("labels", frame.config.labels);
   }

   /**
    * Sets the NO-UNDERLINE option.
    * 
    * @param   noUnderline
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoUnderline(boolean noUnderline)
   {
      frame.config.noUnderline = noUnderline;
      frame.pushWidgetAttr("noUnderline", frame.config.noUnderline);
   }

   /**
    * Sets the NO-AUTO-VALIDATE option.
    *
    * @param noAutoValidate
    *        <code>true</code> if the option is set to ON.
    */
   @Override
   public void setNoAutoValidate(boolean noAutoValidate)
   {
      frame.config.noAutoValidate = noAutoValidate;
      frame.pushWidgetAttr("noAutoValidate", frame.config.noAutoValidate);
   }
   
   /**
    * Sets the OVERLAY option.
    * 
    * @param   overlay
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setOverlay(boolean overlay)
   {
      frame.setOverlay(overlay);
   }

   /**
    * Gets the PAGE-BOTTOM option.
    * 
    * @return   <code>true</code> if the option is set to ON.
    */
   public boolean getPageBottom()
   {
      return frame.getAttr("pageBottom", () -> frame.config.pageBottom);
   }

   /**
    * Sets the PAGE-BOTTOM option.
    * 
    * @param   pageBottom
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setPageBottom(boolean pageBottom)
   {
      frame.config.pageBottom = pageBottom;
      frame.pushWidgetAttr("pageBottom", frame.config.pageBottom);
   }

   /**
    * Gets the PAGE-TOP option.
    * 
    * @return   <code>true</code> if the option is set to ON.
    */
   public boolean getPageTop()
   {
      return frame.getAttr("pageTop", () -> frame.config.pageTop);
   }

   /**
    * Sets the PAGE-TOP option.
    * 
    * @param   pageTop
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setPageTop(boolean pageTop)
   {
      frame.config.pageTop = pageTop;
      frame.pushWidgetAttr("pageTop", frame.config.pageTop);
   }

   /**
    * Sets the PARENT option.
    * 
    * @param   widget
    *          New parent widget for frame.
    */
   @Override
   public void setParent(GenericWidget<?> widget)
   {
      frame.config.parentId = widget.getId();
      pushScreenDefinition();
   }
   
   /**
    * Get the PARENT option.
    * 
    * @return  handle object for the frame handle.
    * 
    */
   @Override
   public handle getParentHandle()
   {
      return frame.getParentHandle();
   }
   
   /**
    * Get the first child of the frame which will be the field group that
    * contains all non-header widgets.
    *
    * @return   The first field group.
    */
   @Override
   public handle firstChild()
   {
      return new handle(fieldGroup);
   }
   
   /**
    * Tests if the field group is initialized for this frame.
    * 
    * @return   True if its field group has been initialized.
    */
   public boolean hasFieldGroup()
   {
      return fieldGroup != null;
   }
   
   /**
    * Wraps widgets to the FieldGroup - pseudo-widget used as container of row widgets  
    *
    * @param    dynamic
    *           If {@code true} the created field group will be created as dynamic.
    * @param    widgets
    *           Source widget list
    * @param    beforeFg
    *           The right neighbor of the field-group been created or null if new field-group
    *           will be the last in the chain. If <b>beforeFg</b> is the first element in the chain
    *           the new BaseEntity will become the new first element in the chain.
    *
    * @return  FieldGroup with specified widgets
    */
   private FieldGroup wrapWidgetsToFieldGroup(boolean            dynamic,
                                              GenericWidget<?>[] widgets,
                                              FieldGroup         beforeFg)
   {
      FieldGroup fg = new FieldGroup(dynamic, beforeFg);
      fg.setFrame(this);
      
      setFieldGroupWidgets(fg, widgets);
      
      // set an ID for this field-group; do not register it with the ConfigManager, as the 
      // field-groups are a kind of virtual widget on server-side,  with no client-side counterpart.
      fg.config.id = WidgetId.nextID();
      
      // register it as widget
      LogicalTerminal.registerWidget(fg, fg.config.id.asInt());
      
      return fg;
   }
   
   /**
    * Add the specified widgets to that FIELD-GROUP.  All widgets will report that they are part of the 
    * specified group.
    * 
    * @param    fg
    *           The FIELD-GROUP.
    * @param    widgets
    *           Source widget list
    * 
    * @see      #checkFieldGroup(GenericWidget) for limitations.
    */
   private void setFieldGroupWidgets(FieldGroup fg, GenericWidget<?>[] widgets)
   {
      LinkedList<GenericWidget<?>> list = new LinkedList<>();
      
      // iterate though all non-header widgets in the proper order (the
      // n2w must return an ordered iterator)
      for (GenericWidget<?> next : widgets)
      {
         // SPACE and SKIP aren't real widgets in Progress terms
         // browse column widget is also artificial one
         if (next instanceof SkipEntity || next instanceof BrowseColumnWidget)
         {
            continue;
         }
         
         // create the specific list of widgets (in the correct order)
         list.add(next);
      }
      
      fg.setWidgets(list);
   }
   
   /**
    * Check the widget's FIELD-GROUP and log a message if is not part of CURRENT-ITERATION.
    * 
    * @param    widget
    *           The widget to check.
    */
   private void checkFieldGroup(GenericWidget<?> widget)
   {
      GenericFrame frame = widget.getFrame();
      if (!widget._dynamic() && frame.frame.config.labels && !frame.frame.config.sideLabels)
      {
         // CA: the SCREEN-VALUE issue may be solved by using a proxy instance which will be used to route
         // all method calls for that widget.  Additionally, this instance will hold the field-group to which
         // the widget is added.  In SCREEN-VALUE case, we can determine the field's row, and from that, 
         // the correct ScreenBuffer from screenBufferCache.
         
         // the same for widget's PARENT field, it must be set at the proxy instance and resolved from that.
         
         handle fg = widget.getParentHandle();
         if (frame.currentFieldGroup != null && fg._isValid() && fg.getResource() != frame.currentFieldGroup)
         {
            LOG.severe("UNSUPPORTED: SCREEN-VALUE is accessed for a widget not part of the CURRENT-ITERATION!");
         }
      }
   }
   
   /**
    * Sets the frame retain size.
    * 
    * @param   retain
    *          retain size
    */
   @Override
   public void setRetain(int retain)
   {
      frame.config.retain = retain;
   }
   
   /**
    * Sets the frame row as a callback to read the current value calculated
    * by the given expression. This value is honored at the time that the
    * frame is made visible (for the first time or after being hidden).
    * 
    * @param    row
    *           The expression to evaluate for the frame row.
    */
   @Override
   public void setRow(Resolvable row)
   {
      dynRow = row;
   }
   
   /**
    * Sets the frame row as a callback to read the current value stored in the
    * given <code>Accessor</code>. This value is honored at the time that the
    * frame is made visible (for the first time or after being hidden).
    * 
    * @param    row
    *           The instance variable to read for the frame row.
    */
   @Override
   public void setRow(Accessor row)
   {
      setRow(new ExpressionWrapper(row));
   }
   
   /**
    * Sets the row position for this frame (statically).
    * 
    * @param    row
    *           The 1-based row number at which to start displaying the frame.
    *           This is converted to a 0-based row number for the internal
    *           runtime.
    */
   @Override
   public void setRow(NumberType row)
   {
      frame.setRow(row);
   }

   /**
    * Sets the row position for this frame (statically).
    * 
    * @param    row
    *           The 1-based row number at which to start displaying the frame.
    *           This is converted to a 0-based row number for the internal
    *           runtime.
    */
   @Override
   public void setRow(double row)
   {
      frame.setRow(row);
   }

   /**
    * Sets the frame scroll size.
    * 
    * @param   scroll
    *          scroll size
    */
   @Override
   public void setScroll(int scroll)
   {
      frame.config.scroll = scroll;
   }

   /**
    * Sets the SCROLLABLE option.
    * 
    * @param   scrollable
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setScrollable(boolean scrollable)
   {
      frame.setScrollable(scrollable);
   }

   /**
    * Sets the SCROLLABLE option.
    *
    * @param scrollable
    *        <code>true</code> if the option is set to ON.
    */
   @Override
   public void setScrollable(logical scrollable)
   {
      frame.setScrollable(scrollable);   
   }
   
   /**
    * Sets the SCROLLBAR-VERTICAL option.
    *
    * @param    vert
    *           <code>true</code> if the option is set to ON.
    */
   @Override
   public void setScrollbarVertical(boolean vert)
   {
      frame.config.scrollbarVertical = vert;
      frame.pushWidgetAttr("scrollbarVertical", frame.config.scrollbarVertical);
   }
   
   /**
    * Sets the SIDE-LABELS option.
    * 
    * @param   sideLabels
    *          <code>true</code> if the option is set to ON.
    */
   @Override
   public void setSideLabels(boolean sideLabels)
   {
      frame.setSideLabels(sideLabels);
   }

   /**
    * Set the FONT specification of this frame.
    * 
    * @param    fontNum
    *           The font.
    */
   @Override
   public void setFont(long fontNum)
   {
      setFont(new int64(fontNum));
   }

   /**
    * Set the FONT specification of this frame.
    * 
    * @param    fontNum
    *           The font.
    */
   @Override
   public void setFont(int64 fontNum)
   {
      // WARNING: this can't be a no-op, as setting the font (even to the same value), the widget 
      // gets realized/resized

      // if the font is unknown, it gets set to UNDEFINED.
      int font = (fontNum.isUnknown() ? FontManager.UNDEFINED_FONT : fontNum.intValue());
      
      frame.config.font = font;
      frame.pushWidgetAttr("font", frame.config.font);
   }

   /**
    * Get the FONT specification of this frame.
    * 
    * @return   See above.
    */
   @Override
   public integer getFont()
   {
      int font = frame.getAttr("font", () -> frame.config.font);
      
      return (font == FontManager.UNDEFINED_FONT ? new integer() : new integer(font));
   }
   
   /**
    * Gets the frame title.
    *
    * @return   The configured title text or <code>unknown</code>.
    */
   @Override
   public character getTitle()
   {
      return frame.getTitle();
   }

   /**
    * Sets the frame title.
    *
    * @param    title
    *           The title text to use.
    */
   @Override
   public void setTitle(character title)
   {
      frame.setTitle(title);
   }

   /**
    * Sets the frame title.
    *
    * @param    title
    *           The title text to use.
    */
   @Override
   public void setTitle(String title)
   {
      frame.setTitle(title);
   }

   /**
    * Sets the frame title and title color.
    *
    * @param    title
    *           The title text to use.
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setTitle(character title, ColorSpec titleColor)
   {
      setTitle(title.isUnknown() ? "" : title.getValue(), titleColor);
   }

   /**
    * Sets the frame title and title color.
    *
    * @param    title
    *           The title text to use.
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setTitle(String title, ColorSpec titleColor)
   {
      FrameConfig fc = frame.config;
      fc.title = title;
      fc.titleDColor = titleColor.convert();
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback to the given
    * parameter).  The value at the instant of this call is not the value that 
    * will necessarily be seen when the frame is viewed.
    *
    * @param    title
    *           The simple variable or field reference that contains the
    *           title text to be used.
    */
   @Override
   public void setDynamicTitle(BaseDataType title)
   {
      setDynamicTitle(new AccessorWrapper(title));
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback) using the
    * given parameter.  The value at the instant of this call is not the
    * value that will necessarily be seen when the frame is viewed.
    *
    * @param    dyntitle
    *           The simple variable or field reference that contains the
    *           title text to be used.
    */
   @Override
   public void setDynamicTitle(Accessor dyntitle)
   {
      class TitleExpr
      extends CharacterExpression
      {
         private Accessor title = null;
         
         /**
          * Construct an instance for given {@link Accessor} instance.
          * 
          * @param    title
          *           Dynamic title variable.
          */
         public TitleExpr(Accessor title)
         {
            this.title = title;
         }

         @Override
         public character execute()
         {
            return (character) title.get();
         }
      }
      
      setDynamicTitle(new TitleExpr(dyntitle));
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback) using the
    * given parameter.  The value at the instant of this call is not the
    * value that will necessarily be seen when the frame is viewed.
    *
    * @param    title
    *           The expression callback that generates the title text to be
    *           used.
    */
   @Override
   public void setDynamicTitle(CharacterExpression title)
   {
      dynTitle = title;
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback to the given
    * parameter).  The value at the instant of this call is not the value that 
    * will necessarily be seen when the frame is viewed.
    *
    * @param    title
    *           The simple variable or field reference that contains the
    *           title text to be used.
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setDynamicTitle(BaseDataType title, ColorSpec titleColor)
   {
      setDynamicTitle(new AccessorWrapper(title), titleColor);
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback) using the
    * given parameter.  The value at the instant of this call is not the
    * value that will necessarily be seen when the frame is viewed.
    *
    * @param    title
    *           The simple variable or field reference that contains the
    *           title text to be used.
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setDynamicTitle(Accessor title, ColorSpec titleColor)
   {
      setDynamicTitle(title);
      setTitleColor(titleColor);
   }

   /**
    * Sets the frame title to be used when the frame is viewed for the first
    * time or when the frame is viewed for the first time after being hidden.
    * <p>
    * The title's value will be read dynamically (as a callback) using the
    * given parameter.  The value at the instant of this call is not the
    * value that will necessarily be seen when the frame is viewed.
    *
    * @param    title
    *           The expression callback that generates the title text to be
    *           used.
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setDynamicTitle(CharacterExpression title,
                               ColorSpec           titleColor)
   {
      setDynamicTitle(title);
      setTitleColor(titleColor);
   }

   /**
    * Sets the color for the frame title.
    *
    * @param    titleColor
    *           The overridden color for the title.
    */
   @Override
   public void setTitleColor(ColorSpec titleColor)
   {
      frame.config.titleDColor = titleColor.convert();
      frame.pushWidgetAttr("titleDColor", frame.config.titleDColor);
   }

   /**
    * Sets the TITLE-BGCOLOR writable attribute.
    *
    * @param    bgcolor
    *           The new value for the TITLE-BGCOLOR attribute.
    */
   @Override
   public void setTitleBgColor(long bgcolor)
   {
      asWidget().setTitleBgColor(bgcolor);
   }

   /**
    * Sets the TITLE-DCOLOR writable attribute.
    *
    * @param    dcolor
    *           The new value for the TITLE-DCOLOR attribute.
    */
   @Override
   public void setTitleDColor(long dcolor)
   {
      asWidget().setTitleDColor(dcolor);
   }

   /**
    * Sets the TITLE-FGCOLOR writable attribute.
    *
    * @param    fgcolor
    *           The new value for the TITLE-FGCOLOR attribute.
    */
   @Override
   public void setTitleFgColor(long fgcolor)
   {
      asWidget().setTitleFgColor(fgcolor);
   }

   /**
    * Set the TITLE-FONT attribute of this browse widget.
    *
    * @param    fontNum
    *           An entry in the font-table or unknown to refer the default font.
    */
   @Override
   public void setTitleFont(long fontNum)
   {
      asWidget().setTitleFont(fontNum);
   }
   
   /**
    * Sets the TOP-ONLY option.
    *
    * @param topOnly
    *        <code>true</code> if the option is set to ON.
    */
   @Override
   public void setTopOnly(boolean topOnly)
   {
      frame.setTopOnly(topOnly);
   }

   /**
    * Sets the width of the widget.
    *
    * @param width
    *        width of the widget
    */
   @Override
   public void setWidthChars(double width)
   {
      frame.setWidthChars(width);
   }

   /**
    * Set the WIDTH Progress attribute.
    *
    * This attribute is an alias for <tt>WIDTH-CHARS attribute</tt>, this method shall
    * never be called directly, its solely purpose is to assert the call
    * <tt>CAN-SET(widget, "WIDTH")</tt> works properly. 
    *
    * @param    widthChars
    *           The integer value for the new WIDTH value.
    */
   @Override
   public void setWidth(NumberType widthChars)
   {
      setWidthChars(widthChars);
   }

   /**
    * Sets the width of the widget.
    *
    * @param width
    *        width of the widget
    */
   @Override
   public void setWidthChars(NumberType width)
   {
      frame.setWidthChars(width);
   }
   
   /**
    * Returns the width of the widget.
    *
    * @return  width of the widget, retrieved from the 
    *          <code>FRAME:WIDTH-CHARS</code> attribute.
    */
   @Override
   public decimal getWidthChars()
   {
      return frame.getWidthChars();
   }

   /**
    * Get the WIDTH Progress attribute. The value is equal to the current component char
    * width.
    * 
    * This attribute is an alias for <tt>WIDTH-CHARS attribute</tt>, this method shall
    * never be called directly, its solely purpose is to assert the call
    * <tt>CAN-QUERY(widget, "WIDTH")</tt> works properly. 
    *
    * @return    The integer value of the WIDTH attribute.
    */
   @Override
   public decimal getWidth()
   {
      return getWidthChars();
   }

   /**
    * Get the HEIGHT Progress attribute. The value is equal to the current component char
    * height.
    * 
    * This attribute is an alias for <tt>HEIGHT-CHARS attribute</tt>, this method shall
    * never be called directly, its solely purpose is to assert the call
    * <tt>CAN-QUERY(widget, "HEIGHT")</tt> works properly. 
    *
    * @return    The integer value of the HEIGHT attribute.
    */
   @Override
   public decimal getHeight()
   {
      return getHeightChars();
   }

   /**
    * Set the HEIGHT Progress attribute.
    *
    * This attribute is an alias for <tt>HEIGHT-CHARS attribute</tt>, this method shall
    * never be called directly, its solely purpose is to assert the call
    * <tt>CAN-SET(widget, "HEIGHT")</tt> works properly. 
    *
    * @param    heightChars
    *           The integer value for the new HEIGHT value.
    */
   @Override
   public void setHeight(NumberType heightChars)
   {
      setHeightChars(heightChars);
   }


   /**
    * Returns the width of the widget in pixels. Corresponds Progress
    * WIDTH-PIXELS attribute.
    *
    * @return The width of the widget in pixels.
    */
   @Override
   public integer getWidthPixels()
   {
      return frame.getWidthPixels();
   }

   /**
    * Sets new value of WIDTH-PIXELS attribute.
    *
    * @param widthPixels
    *        new value of WIDTH-PIXELS attribute
    */
   @Override
   public void setWidthPixels(NumberType widthPixels)
   {
      frame.setWidthPixels(widthPixels);
   }

   /**
    * Sets new value of WIDTH-PIXELS attribute.
    *
    * @param   widthPixels
    *          new value of WIDTH-PIXELS attribute
    */
   @Override
   public void setWidthPixels(int widthPixels)
   {
      frame.setWidthPixels(widthPixels);
   }

   /**
    * Returns the horisontal position of the frame. Corresponds Progress X
    * attribute.
    *
    * @return the horisontal position of the frame.
    */
   @Override
   public integer getX()
   {
      return frame.getX();
   }

   /**
    * Returns the vertical position of the frame. Corresponds Progress Y
    * attribute.
    *
    * @return the vertical position of the frame.
    */
   @Override
   public integer getY()
   {
      return frame.getY();
   }

   /**
    * Sets the horizontal position of the frame (Progress X
    * attribute).
    *
    * @param x
    *        horizontal position to be set.
    */
   @Override
   public void setX(final NumberType x)
   {
      frame.setX(x);
   }

   /**
    * Sets the horizontal position of the frame (Progress X
    * attribute).
    *
    * @param x
    *        horizontal position to be set.
    */
   @Override
   public void setX(final int x)
   {
      frame.setX(x);
   }

   /**
    * Sets the vertical position of the frame (Progress Y
    * attribute).
    *
    * @param y
    *        vertical position to be set.
    */
   @Override
   public void setY(final NumberType y)
   {
      frame.setY(y);
   }

   /**
    * Sets the vertical position of the frame (Progress Y
    * attribute).
    *
    * @param y
    *        vertical position to be set.
    */
   @Override
   public void setY(final int y)
   {
      frame.setY(y);
   }

   /**
    * Sets the default button for this frame (an option in the frame phrase).
    *
    * @param    defaultButton
    *           button widget that works as the default
    */
   @Override
   public void setDefaultButtonOption(GenericWidget<?> defaultButton)
   {
      if (defaultButton == null)
      {
         return;
      }

      int id = getFrameWidgetId(defaultButton);
      
      if (id == -1)
         return;

      frame.config.defaultButtonId = id;
   }
   
   /**
    * Sets the cancel button for this frame (an option in the frame phrase).
    *
    * @param    cancelButton
    *           button widget that receives the CHOOSE event when a user cancels the current frame
    *           or dialog box by pressing the ESC key.
    */
   @Override
   public void setCancelButtonOption(GenericWidget<?> cancelButton)
   {
      if (cancelButton == null)
      {
         return;
      }

      int id = getFrameWidgetId(cancelButton);
      
      if (id == -1)
         return;

      frame.config.cancelButtonId = id;
   }

//   /**
//    * Sets the default button for this frame.
//    *
//    * @param defaultButton
//    *        button widget that works as the default
//    */
//   @Override
//   public void setDefaultButton(ButtonWidget defaultButton)
//   {
//      int id = getWidgetId(defaultButton);
//      
//      if (id == -1)
//         return;
//
//      frame.config.defaultButtonId = id;
//      pushScreenDefinition();
//   }
//   
//   /**
//    * Sets the cancel button for this frame.
//    *
//    * @param    cancelButton
//    *           button widget that receives the CHOOSE event when a user cancels the current frame
//    *           or dialog box by pressing the ESC key.
//    */
//   @Override
//   public void setCancelButton(ButtonWidget cancelButton)
//   {
//      int id = getWidgetId(cancelButton);
//      
//      if (id == -1)
//         return;
//
//      frame.config.cancelButtonId = id;
//      pushScreenDefinition();
//   }
   
   /**
    * Sets the widget which will be used to display the description for 
    * any given statistic that is being automatically aggregated by the
    * {@link #display} processing.
    *
    * @param    widget
    *           The widget to use as the statistics column.
    */
   @Override
   public void setStatisticsColumn(GenericWidget<?> widget)
   {
      statsColumn = widget;
   }
   
   /**
    * Notifies the frame about entering a new frame scope.  This may occur
    * more than once for a given frame due to retries or (more likely) due
    * to a frame being scoped to a block that is nested inside a loop. In
    * the case where this is a shared frame, this will be called more than
    * once while a previous frame scope is still open.
    * <p>
    * This method registers a callback for the block exit event. The callback
    * finalizes the frame and all contained widgets.
    * <p>
    * On block iterate or retry, frame state will also be maintained.
    */
   @Override
   public void openScope()
   {
      // push down any pending screen definitions
      LogicalTerminal.processDeferredPush();
      
      // register for iterate, retry and finished notifications
      TransactionManager.registerFinalizable(this, false);
      
      // associate this frame with a scope in the LT (this is safe to do
      // multiple times and must be safe because of multiple nested calls
      // to this method)
      
      if (scopeNesting == 0)
      {
         LogicalTerminal.registerFrameInScope(this);
         
         // reset frame to its initial state
         resetFrameBuf();
         pendingUpDown = 0;
         hadView       = false;
         canHide       = true;
         hadCondition  = false;

         // the screen buffer cache needs also to be cleared
         screenBufferCache.clear();
      }
      
      LogicalTerminal.signalFrameScope(this, MarkEntry.OPEN_SCOPE);

      scopeNesting ++;
   }

   /**
    * Cleanup that must occur at the end of the block to which the frame is
    * scoped.
    */
   @Override
   public void finished()
   {
      // If we previously bypassed the clean up - do this now
      if (hadView && !frame.isDown() &&
          isOutRedirected() && sameViewDataRedirected)
      {
         frameBuf.reset();
      }
      
      if (hadView && isRedirected() || (isOutRedirected() && !isOutTerm()))
      {
         if (!isFlushed() && getRedirOut() != null)
         {
            getRedirOut().flush();
         }
         flushed = true;
      }
      
      // aggregation needs this to finish aggregation processing 
      if (aggregateHandler != null)
      {
         aggregateHandler.finished();
      }
      
      // clear the list, in case no iteration was performed
      deferredAggregators = null;
      
      scopeNesting--;
      
      LogicalTerminal.signalFrameScope(this, MarkEntry.CLOSE_SCOPE);
   }
   
   /**
    * This is a no-op.
    */
   @Override
   public void deleted()
   {
      // no-op
   }
   
   /**
    * Final cleanups when frame completely leaves outer execute() scope (only
    * occurs once for a frame, even if this is a shared frame which has
    * nested <code>openScope()</code> invocations). 
    */
   public void destroy()
   {
      // Destroy frame at client. Deregistration will happen later,
      // when client will actually hide frame and deregister it. 
      LogicalTerminal.destroyFrame(this);
      
      // reset redirection state
      // TODO: check to see if there are other frame events which should trigger clearing
      //       this state; it seems wrong that the only way to clear this is when the
      //       frame goes out of scope; perhaps if the output destination is already closed
      //       or when it is switched...
      redirOut = null;
      
      // for dynamic frames we need to send message for life time finished here because there
      // is no other way to do this.
      if (frame._dynamic())
      {
         LogicalTerminal.signalFrameScope(this, MarkEntry.CLOSE_SCOPE);
      }
   }
   
   /**
    * Provides a notification that the block whose scope in which the object
    * is registered is about to iterate and attempt another pass.  This
    * provides an opportunity to preserve state from the previous pass and
    * make preparations for the coming pass.
    */
   @Override
   public void iterate()
   {
      if (deferredAggregators != null)
      {
         processAggregates(deferredAggregators, false);
         deferredAggregators = null;
      }

      if (aggregateHandler != null)
      {
         aggregateHandler.iterate();
      }

      canHide = true;
      resetFrameBuf();
      
      if (isOutRedirected())
         down(getRedirOut());
      else
         down();
   }

   /**
    * Provides a notification that the block whose scope in which the object
    * is registered is about to retry and re-attempt the current iteration.
    * This provides an opportunity to clear or reset state from the previous
    * pass.
    */
   @Override
   public void retry()
   {
      if (!hadCondition )
      {
         resetFrameBuf();
         LogicalTerminal.resetFrame(this);
      }
      hadCondition = false;
   }
   
   /**
    * Marks the beginning of the editing block.
    *
    * @param    list
    *           List of frame elements controlled by this editing block. This
    *           is used to ensure that the proper semantics of update or set
    *           are honored at editing block start (optionally copying data
    *           into the screen buffer, enabling widgets and view).
    * @param    type
    *           Specifies if this editing block must operates like
    *           {@link #promptFor} (no copying to or from the screen buffer),
    *           {@link #set} (no copying to the screen buffer) or 
    *           {@link #update} (copying both directions).  Use the constants
    *           PROMPT_FOR, SET and UPDATE to specify this parameter. 
    */
   @Override
   public void startEditingMode(FrameElement[] list, int type)
   {
      // create a default GO-ON list
      EventList eventList = new EventList(true);

      startEditingMode(list, type, eventList);
   }
   
   /**
    * Marks the beginning of the editing block with caller control over the
    * exit conditions for the loop.
    *
    * @param    list
    *           List of frame elements controlled by this editing block. This
    *           is used to ensure that the proper semantics of update or set
    *           are honored at editing block start (optionally copying data
    *           into the screen buffer, enabling widgets and view).
    * @param    type
    *           Specifies if this editing block must operates like
    *           {@link #promptFor} (no copying to or from the screen buffer),
    *           {@link #set} (no copying to the screen buffer) or 
    *           {@link #update} (copying both directions).  Use the constants
    *           PROMPT_FOR, SET and UPDATE to specify this parameter. 
    * @param    wait
    *           List of events to wait for.
    *
    */
   @Override
   @SuppressWarnings("unused")
   public void startEditingMode(FrameElement[] list, 
                                int            type,
                                EventList      wait)
   {
      // this object will register itself for exit notification from the 
      // current block, if the exit is abnormal it will trigger cleanup
      EditingCleaner ec = new EditingCleaner(list);
      
      synchronized (editLock)
      {
         if (currentOp != NONE)
         {
            throw new IllegalStateException("Cannot nest editing blocks!");
         }
         
         currentOp = type;
      }
      
      // process text grouping (word wrap mode)
      textGroup  = textGroup(list);
      
      if (type == UPDATE)
      {
         copyToScreenBuffer(list);
      }
      
      GenericWidget<?>[] wids = widgetListFromFrameElements(list);
      
      int[] ids = validIds(wids);
      
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      conditionalUpDown();
      
      // set the wait-list operation type (it MUST not be WAIT-FOR since the client will
      // not be dropping into that worker)
      switch (type)
      {
         case PROMPT_FOR:
            wait.setOperation(EventList.PROMPT_FOR);
            break;
         case SET:
            wait.setOperation(EventList.SET);
            break;
         case UPDATE:
            wait.setOperation(EventList.UPDATE);
            break;
      }
      
      LogicalTerminal.promptFor(this, ids, wait, false, null);
   }
   
   /**
    * Displays an underline for the specified field-level widgets of the
    * frame. Does not permanently change the attributes for the given
    * widgets!
    * <p>
    * This method corresponds to these Progress statements:
    * UNDERLINE
    * 
    * @param    list 
    *           List of widgets.
    */
   @Override
   public void underline(GenericWidget<?>[] list)
   {
      int[] ids = validIds(list);
      
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      underline(ids);
   }
   
   /**
    * Displays an underline for the specified field-level widgets of the
    * frame. Does not permanently change the attributes for the given
    * widgets!
    * <p>
    * This method corresponds to these Progress statements:
    * UNDERLINE
    * 
    * @param    list 
    *           List of widget ids.
    */
   @Override
   public void underline(int[] list)
   {
      // store the current had view flag
      boolean oldHadView = hadView;
      boolean oldAfterUnderline = afterUnderline;

      // consider the frame as visible for non-redirected cases
      if (!isRedirected())
      {
         hadView = true;
      }
      
      if (!afterDown && !oldAfterUnderline)
      {
         downInt();
      }
      else if (oldAfterUnderline && pendingUpDown > 0)
      {
         // successive UNDERLINE statements must land on the same line
         pendingUpDown--;
      }
      
      // TODO: when UNDERLINE is used BEFORE a DISPLAY (or in sequence with a DOWN), P2J still
      // has problems, in redirected mode. see testcases/uast/frame_layout/redirected_underline.p
      
      conditionalUpDown();
      
      // drive the process
      LogicalTerminal.underline(this, list);
      downInt();
      afterUnderline = true;
      
      // restore the current had view flag
      hadView = oldHadView;
   }
   
   /**
    * Displays an underline for the specified field-level widgets of the
    * frame. Does not permanently change the attributes for the given
    * widgets!
    * <p>
    * This method corresponds to these Progress statements:
    * UNDERLINE
    * 
    * @param    stream
    *           An explicit stream which will be used for the output.
    * @param    list 
    *           List of widgets.
    */
   @Override
   public void underline(Stream stream, GenericWidget<?>[] list)
   {
      int[] ids = validIds(list);
      
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      underline(stream, ids);
   }
   
   /**
    * Displays an underline for the specified field-level widgets of the
    * frame. Does not permanently change the attributes for the given
    * widgets!
    * <p>
    * This method corresponds to these Progress statements:
    * UNDERLINE
    * 
    * @param    stream
    *           An explicit stream which will be used for the output.
    * @param    list 
    *           List of widget ids.
    */
   @Override
   public void underline(Stream stream, int[] list)
   {
      if (stream != null && stream.protectWrites())
      {
         return;
      }
      
      // temporarily switch the terminal to a new output destination
      int id      = getRemoteStreamId(stream); 
      int tempOut = LogicalTerminal.switchOutput(id, true, true);
      
      try
      {
         // because of the internal down/conditionalUpDown behavior of this
         // next method call, we cannot easily move the switchOutput() call
         // to the client side
         underline(list);
      }
      
      finally
      {
         // put the old terminal back (either interactive or a different
         // redirection)
         LogicalTerminal.switchOutput(tempOut, false);
      }
   }
   
   /**
    * Waits for a default list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR </b>
    */
   @Override
   public void waitFor()
   {
      conditionalUpDown();
      LogicalTerminal.waitFor(this);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR </b>
    * 
    * @param   el
    *          List of events to wait for.
    */
   @Override
   public void waitFor(EventList el)
   {
      conditionalUpDown();
      LogicalTerminal.waitFor(this, el);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget </b>
    * 
    * @param   el
    *          List of events to wait for.
    * @param   focusWidget
    *          Widget which is in focus.
    */
   @Override
   public void waitFor(EventList el, GenericWidget<?> focusWidget)
   {
      conditionalUpDown();
      LogicalTerminal.waitFor(this, el, focusWidget);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR PAUSE n </b>
    * 
    * @param   el
    *          List of events to wait for.
    * @param   seconds
    *          Time to wait for event.
    */
   @Override
   public void waitFor(EventList el, int seconds)
   {
      conditionalUpDown();
      LogicalTerminal.waitFor(this, el, seconds);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget PAUSE n </b>
    * 
    * @param   el
    *          List of events to wait for.
    * @param   focusWidget
    *          Widget which is in focus.
    * @param   seconds
    *          Time to wait for event.
    */
   @Override
   public void waitFor(EventList el, GenericWidget<?> focusWidget,
                       int seconds)
   {
      conditionalUpDown();
      LogicalTerminal.waitFor(this, el, focusWidget, seconds);
   }
      
   /**
    * Blocks until a next UI event is available (returns immediately if an
    * event is already available).
    */
   @Override
   public void waitForNextKey()
   {
      LogicalTerminal.waitForNextKey();
   }

   /**
    * Remove from the list of tab items the items that became non-tab items.
    */
   @Override
   public void cleanNonTabItems()
   {
      getFieldGroup().cleanNonTabItems();
   }

   /**
    * Marks the end of the editing block when triggered by any abnormal exit
    * from the block on the server.  This method does NOT need to be called
    * if the exit is naturally triggered via {@link #continueEditing}. This
    * method allows the client's state to be cleaned up in the abnormal
    * termination case.
    *
    * @param    list
    *           List of frame elements controlled by this editing block. This
    *           is used to ensure that the proper semantics of update or set
    *           are honored at editing block exit (disabling widgets and
    *           assigning data back).
    * @param    normal
    *           <code>true</code> if this was triggered as the result of
    *           the natural pending GO exit from the editing block (if this 
    *           was the result of a call to {@link #continueEditing}).
    *           <code>false</code> if this was due to an abnormal exit and
    *           the client's state needs cleanup.
    * @param    cleanup
    *           <code>true</code> if the client side cleanup is needed.
    */
   public void stopEditingMode(FrameElement[] list,
                               boolean        normal,
                               boolean        cleanup)
   {
      GenericWidget<?>[] wids = widgetListFromFrameElements(list);
      
      int[] ids = validIds(wids);
      
      // only notify the client of the stop and disable if we actually
      // enabled and started editing mode in the first place, if there are
      // no widgets specified, then we never entered editing mode in the
      // first place
      if (ids != null)
      {
         if (cleanup)
         {
            //protect the screen widget states as the client sends only modified values but 
            // the business logic can modify screen widget values too.
            frameBuf.markAllChangedStatesAsEntered();
            // clear the client's state
            LogicalTerminal.stopEditingMode();
         }
      }
      
      textUngroup(textGroup);
      textGroup = null;
      
      synchronized (editLock)
      {
         int op = currentOp;
         currentOp = NONE;

         if (ids != null)
         {
            // only copy changes back if we exited normally (and we are one
            // of the op types that copy back)
            if (normal && (op == SET || op == UPDATE))
            {
               boolean defaults = (op == UPDATE);
               copyFromScreenBuffer(list, defaults, op == SET);
            }
         }
      }

      if (ids != null)
      {
         LogicalTerminal.disable(this, ids, false);
      }
   }
   
   /**
    * Checks to see whether a pending GO event is about to end the editing block.
    *
    * @param    list
    *           List of frame elements controlled by this editing block. This is used to ensure
    *           that the proper semantics of update or set are honored at editing block exit
    *           (disabling widgets and assigning data back).
    * @param    hadNext
    *           <code>true</code> if the executing business logic executed a <code>NEXT</code>
    *           statement causing the editing block to iterate. This will result in clearing
    *           any current <code>GO-PENDING</code> condition.  This is passed as a parameter
    *           here so that a separate trip down to the client is not needed to process the
    *           editing block iteration state update.
    *
    * @return   <code>true</code> if there is no <code>GO-PENDING</code> and the editing mode
    *           continues.
    */
   @Override
   public boolean continueEditing(FrameElement[] list, boolean hadNext)
   {
      boolean cont = LogicalTerminal.continueEditing(hadNext);
      
      if (!cont)
      {
         stopEditingMode(list, true, false);
      }
      
      return cont;
   }

   /**
    * Executes the validation tests for every supported field-level widget
    * in the frame (except the browse, which you must validate explicitly).
    * Corresponds Progress VALIDATE method.
    *
    * @return <code>true</code> if validation was successful.
    */
   @Override
   public logical validate()
   {
      UnimplementedFeature.missing("VALIDATE method not implemented");
      return new logical();
   }

   /**
    * Sets custom colors for output fields, input fields, or both in
    * all specified widgets.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>COLOR DISPLAY [PROMPT]</b> <br>
    * <b>COLOR PROMPT</b>
    * 
    * @param    display
    *           A color specification for regular output or <code>null</code>
    *           if no change is required.
    * @param    prompt
    *           A color specification for input or <code>null</code> if
    *           no change is required.
    * @param    widgets
    *           Array of target widgets.  All elements must be in the same
    *           frame.
    */
   @Override
   public void setColors(ColorSpec       display,
                         ColorSpec       prompt, 
                         GenericWidget<?>[] widgets)
   {
      // honor any pending conditional down/up processing
      conditionalUpDown(true);
      
      // batch up the changes (defer processing until later)
      batch(true);
      
      for (GenericWidget<?> widget : widgets)
      {
         if (widget.getAttr("enabled", () -> widget.config.enabled, true))
         {
            // TODO: this needs more testing for ChUI - some color state gets cleared (like INPUT 
            // for editable fill-in) when widgets are enabled. 

            // skip enabled widgets
            continue;
         }
         
         if (display != null)
            widget.colorDisplay(display);
         
         if (prompt != null)
            widget.colorPrompt(prompt);
      }
      
      // suppress redrawing of the entire frame
      frameDef.setSuppressRedraw(true);
      
      // push the changes
      batch(false, true);
   }

   /**
    * Sets custom colors for the given output field, input field, or both.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>COLOR DISPLAY [PROMPT]</b> <br>
    * <b>COLOR PROMPT</b>
    * 
    * @param    display
    *           A color specification for regular output or <code>null</code>
    *           if no change is required.
    * @param    prompt
    *           A color specification for input or <code>null</code> if
    *           no change is required.
    * @param    widget
    *           The target widget.
    */
   @Override
   public void setColors(ColorSpec     display,
                         ColorSpec     prompt, 
                         GenericWidget<?> widget)
   {
      setColors(display, prompt, new GenericWidget<?>[] {widget});
   }

   /**
    * Reports if this frame object is valid for use.
    * <p>
    * The validity of the frame object is determined solely by being in 
    * scope.
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   boolean valid()
   {
      return !frame.deleted;
   }

   /**
    * Gets the unique numeric ID of this frame.
    *
    * @return  frame ID
    */
   @Override
   public int getFrameId()
   {
      return frameId;
   }

   /**
    * Gets the screen buffer reference for this frame, without any update to dynamic content.
    *
    * @return   The screen buffer reference.
    */
   public ScreenBuffer getFrameBufferRaw()
   {
      return frameBuf;
   }

   /**
    * Gets the differential screen buffer of this frame.
    *
    * @param    force
    *           Flag indicating if the screen buffer is resolved regardless if was changed or not.
    *
    * @return   The screen buffer modifications.
    */
   public ScreenBuffer getFrameBuffer(boolean force)
   {
      updateDynamicContent();
      return frameBuf.getChanged(force);
   }

   /**
    * Gets the frame widget of this frame.
    * 
    * @return  <code>FrameWidget</code>
    */
   public FrameWidget getFrameWidget()
   {
      return frame;
   }

   /**
    * Request indexes of selected items in RadioSet, ComboBox or SelectionList
    * widgets.
    * 
    * @param   widget
    *          Widget for which request should be performed.
    * 
    * @return  array of 0-based indexes of selected items or <code>null</code>
    *          if specified widget does not exists or does not support such a
    *          request.
    */
   public int[] getSelection(GenericWidget<?> widget)
   {
      return LogicalTerminal.getSelection(widget.getId());
   }

   /**
    * Get widget name for the given widget.
    * 
    * @param   widget
    *          Widget to look name for.
    * 
    * @return  Widget name.
    */
   public String getName(GenericWidget<?> widget)
   {
      if (widget.equals(frame))
      {
         return frameName;
      }
      
      if (widget instanceof FieldGroup)
      {
         return widget.name() != null ? widget.name().toStringMessage() : "";
      }
      
      for (String name : n2w.keySet())
      {
         if (n2w.get(name) == widget)
         {
            return name;
         }
      }
      
      return "Unknown";
   }

   /**
    * Get original widget name for the given widget.
    * 
    * @param   widget
    *          Widget to look name for.
    * 
    * @return  original widget name.
    */
   public String getOriginalName(GenericWidget<?> widget)
   {
      String cvtName = getName(widget);
      if (n2n.containsKey(cvtName))
      {
         return n2n.get(cvtName);
      }
      
      if (widget instanceof FrameWidget)
      {
         return widget.getFrame().frameName;
      }
      
      return null;
   }

   /**
    * Gets a set of widgets of this frame that require validation.
    * 
    * @return  a <code>Set</code> of widgets, which may be empty. 
    */
   public Set<GenericWidget<?>> getValidationSet()
   {
      Set<GenericWidget<?>> set = new HashSet<>();
      Iterator<String> iter = n2w.keySet().iterator();
      
      while (iter.hasNext())
      {
         GenericWidget<?> gw = n2w.get(iter.next());
         
         if (gw.isValidatable())
            set.add(gw);
      }
      
      return set;
   }

   /**
    * Mark frame as viewed.
    */
   public void setHadView()
   {
      hadView = true;
      canHide = false;
      flushed = false;
   }
   
   /**
    * Get state of internal flag which is set to <code>true</code> if frame
    * was viewed at client at least once.
    * 
    * @return  <code>true</code> if frame was viewed.
    */
   public boolean isViewed()
   {
      return hadView;
   }
   
   /**
    * Check if frame can hide other frames.
    * 
    * @return <code>true</code> if frame can hide other frames.
    */
   public boolean isCanHide()
   {
      return canHide;
   }
   
   /**
    * This method implements MOVE-AFTER-TAB function.
    * 
    * @param   widget
    *          Widget reference.
    *          
    * @return  <code>true</code> if operation was successful.
    */
   @Override
   public boolean moveAfterTabItem(GenericWidget<?> widget)
   {
      UnimplementedFeature.missing("MOVE-AFTER-TAB method not implemented");
      return false;
   }

   /**
    * Convert given value into proper form for further conversion onto SCREEN-VALUE.
    * <p>
    * In cases when <code>dataType</code> is {@link character} and the <code>value</code> is 
    * a logical value, then the result will be the logical value formatted using the given
    * <code>format</code>.
    * 
    * @param   dataType
    *          Desired data type.
    * @param   value
    *          Source value to convert from.
    * @param   format
    *          The widget's format. Currently used only for logical values.
    *
    * @return  Converted value.
    */
   public Object convertScreenValue(Class<?> dataType, Object value, String format)
   {
      if (value != null)
      {
         // handle type conversion if needed
         
         BaseDataType bdt = null;
         
         if (value instanceof FormattedValue)
         {
            // these cases occur when an @ base field is used and there is
            // a call to the getter before any editing occurs on the client
            // OR an edit on the client was done but no changes were made 
            FormattedValue fv = (FormattedValue) value; 
            
            bdt   = fv.getValue(); 
         }
         else if (value instanceof BaseDataType && !dataType.isInstance(value))
         {
            // some use of @ base fields may not use the FormattedValue but
            // the type conversion is still needed
            bdt = (BaseDataType) value;
         }
         
         if (bdt != null)
         {
            value = null;
            
            if (dataType.isInstance(bdt)) 
            {
               // the types are the same so we can simply return the value
               value = bdt;
            }
            else
            {
               // hmm, we need to convert this into the proper type
               if (dataType.equals(character.class))
               {
                  if (bdt.getClass().equals(logical.class))
                  {
                     // in this case, use the formatted representation
                     value = character.valueOf(bdt, format);
                  }
                  else
                  {
                     // use the common/safe way to convert from any BDT to the character type
                     // TODO: shouldn't this use the format string too?
                     value = character.valueOf(bdt);
                  }
               }
               else
               {
                  if (bdt instanceof unknown)
                  {
                     value = BaseDataType.generateUnknown(dataType);
                  }
                  else
                  {
                     if (format != null)
                     {
                        // TODO: not all BDT's have c'tors with (BaseDataType, String) params
                        
                        // try to find a constructor that matches our type
                        Class<?>[] sig = new Class[] { bdt.getClass(), String.class };
                        
                        try
                        {
                           Constructor<?> cvt = dataType.getConstructor(sig);
                           
                           // best efforts to convert
                           if (cvt != null)
                           {
                              value = cvt.newInstance(bdt, format);
                           }

                           return value;
                        }
                        
                        catch (Exception exc)
                        {
                           // security problems, no such method... force default
                           // path below
                        }
                     }
                     // try to find a constructor that matches our type
                     Class<?>[] sig = new Class[] { bdt.getClass() };
                     
                     try
                     {
                        Constructor<?> cvt = dataType.getConstructor(sig);
                        
                        // best efforts to convert
                        if (cvt != null)
                        {
                           value = cvt.newInstance(bdt);
                        }
                     }
                     
                     catch (Exception exc)
                     {
                        // security problems, no such method... force default path below
                     }
                  }
               }
            }
         }
      }
      
      return value;
   }
   
   /**
    * A worker to perform frame hide. This is possible in two ways: via the HIDE statement or by 
    * setting the frame's attribute.   The <code>hidden</code> flag will be set to 
    * <code>true</code> if frame is hidden by setting the frame's HIDDEN attribute to 
    * <code>true</code>.  Otherwise, it is assumed a HIDE statement is in effect.
    * 
    * @param  noPause
    *         <code>true</code> if NO-PAUSE option was used
    * @param  hidden
    *         <code>true</code> if HIDDEN attribute was set to <code>true</code>.
    */
   void hideWorker(boolean noPause, boolean hidden)
   {
      Stream out = UnnamedStreams.output();
      
      if (out != null)
      {
         hide(out, noPause);
      }
      else
      {
         hidden = hidden || !frame.isRootFrame();
         boolean didHide = LogicalTerminal.hide(frameId, noPause, hidden);
         
         if (!didHide)
         {
            return;
         }
         
         frame._setVisible(false);
         
         // reset the title so next VIEW/DISPLAY will start from scratch
         if (dynTitle != null)
            frameBuf.putTitle(null);
      }
      
      frameDelete();
   }
   
   /**
    * Get the current scope nesting.
    * 
    * @return   The {@link #scopeNesting} value.
    */
   int getScopeNesting()
   {
      return scopeNesting;
   }
   
   /**
    * Delete this frame, if is not dynamic and its associated instantiated procedure no longer
    * exists.
    */
   void frameDelete()
   {
      if ((!frame.deleted || !frame._dynamic()) && frame.instantiatingProcedure().isUnknown())
      {
         // delete all widgets
         for (GenericWidget<?> widget : n2h.values())
         {
            widget.delete();
         }

         Collection<GenericWidget<?>> values = n2w.values();
         GenericWidget<?>[] widgets = values.toArray(new GenericWidget<?>[values.size()]);

         for (GenericWidget<?> widget : widgets)
         {
            // check if the widget was already deleted; also, all static child frames will survive
            if (n2w.containsValue(widget) && 
                (widget._dynamic() || !(widget instanceof FrameWidget)))
            {
               widget.delete();
            }
         }

         // force deletion of a static frame for which its instantiating procedure is no longer
         // available
         frame.delete();
      }
   }
   
   /**
    * Cleanup this frame, as it was deleted.
    */
   void frameCleanup()
   {
      // frame cleanup must be delayed until HIDE is called for a "zombie" static frame (i.e. a 
      // frame with an unknown INSTANTIATING-PROCEDURE) or a dynamic frame is explicitly deleted

      LogicalTerminal.signalFrameScope(this, MarkEntry.DELETED_SCOPE);

      // WARNING: can there be issues with shared frames that have multiple
      //          openScope() calls?  In that case it is possible we may
      //          clear something registered in an outer scope.  WATCH THIS!
      
      if (headers != null)
      {
         headers = null;
         frameBuf.clearHeaders();
      }
      
      destroy();
   }
   
   /**
    * Mark beginning of trigger processing. 
    */
   void triggerEnter()
   {
      triggerActive++;
   }

   /**
    * Mark end of trigger processing. 
    */
   void triggerLeave()
   {
      triggerActive--;
      silentFormatErrors = false;
   }
   
   /**
    * Conditionally pushes the screen definition instance down to the client. 
    * If modifications are currently being batched, no action is taken.
    * Otherwise, a new screen definition is built and sent to the client.
    * <p>
    * When the screen definition is built, a method of 
    * <code>LogicalTerminal</code> is called to send it down to the client.
    */
   void pushScreenDefinition()
   {
      pushScreenDefinition(true);
   }
   
   /**
    * Conditionally pushes the screen definition instance down to the client. 
    * If modifications are currently being batched, no action is taken.
    * Otherwise, a new screen definition is built and sent to the client.
    * <p>
    * When the screen definition is built, a method of 
    * <code>LogicalTerminal</code> is called to send it down to the client.
    * 
    * @param  force
    *         <code>true</code> to ensure that the screen definition will be
    *         synchronously pushed to the client. <code>false</code> to
    *         defer the push.
    */
   void pushScreenDefinition(boolean force)
   {
      // batching is in progress, do not push individual changes
      if (batch)
      {
         return;
      }
      
      // time to push the changes down
      if (frameDef == null)
      {
         initScreenDefinition();
      }
      
      // push down the screen definition
      if (force)
      {
         TransactionManager.registerBlockScopeable(LogicalTerminal.getInstance());

         // synchronous push
         ScreenDefinition[] sd = new ScreenDefinition[] {frameDef};
         LogicalTerminal.pushScreenDefinition(sd);
      }
      else
      {
         // deferred push
         LogicalTerminal.deferPush(frameDef);
      }
   }
   
   /**
    * Sets a new ScreenBuffer for this frame.
    * 
    * @param   sb
    *          new ScreenBuffer for this frame
    * @param   allow
    *          <code>true</code> to allow the switching of screen buffers if
    *          the frame is a non-redirected down frame and the frame-line is
    *          detected as changing.  Otherwise this switching will be
    *          disabled.
    */
   void setScreenBuffer(ScreenBuffer sb, boolean allow)
   {
      if (allow)
      {
         int oldline = frameBuf.getLine();
         int newline = sb.getLine();
         
         // check if the frame line is different than before AND this is a
         // down frame that is not redirected
         if (oldline != newline && frame.isDown() && !isRedirected())
         {
            // the buffer cache is 1-based so we can only dereference an
            // existing screen buffer if the index is 1 or above
            if (newline > 0)
            {
               // switch screen buffers because the client switched frame
               // lines
               frameBuf = getScreenBufferFromCache(newline);
            }
         }
      }
      frameBuf.mergeChanged(sb);
      frameBuf.resetChanged();
   }
   
   /**
    * Returns -1 if the widget name is invalid or <code>null</code>, otherwise 
    * it returns the assigned widget ID.
    *
    * @param    widgetName
    *           A widget name.
    *
    * @return   The numeric ID assigned to the widget or -1 if no such widget
    *           exists.
    */
   int getWidgetId(String widgetName)
   {
       return getWidgetId(n2w.get(widgetName));
   }
   
   /**
    * Returns -1 if the widget is invalid or <code>null</code>, otherwise 
    * it returns the assigned widget ID.
    *
    * @param    widget
    *           A widget instance in the frame.
    *
    * @return   The numeric ID assigned to the widget or -1 if no such widget
    *           exists.
    */
   int getWidgetId(GenericWidget<?> widget)
   {
      if (widget != null)
      {
         return widget.getId();
      }
      
      return -1;
   }

   /**
    * Returns -1 if the widget is invalid, otherwise it returns the assigned widget ID.
    *
    * @param    widget
    *           A widget instance in the frame.
    *
    * @return   The numeric ID assigned to the widget or -1 if no such widget
    *           exists.
    */
   int getFrameWidgetId(GenericWidget<?> widget)
   {
         return widget != null && widgets != null &&
               Arrays.stream(widgets).anyMatch(w -> w.getId() == widget.getId()) ?
                     widget.getId() : -1;
   }

   /**
    * Looks up the widget by its ID.
    * 
    * @param    id
    *           The numeric ID assigned to the widget.
    * 
    * @return   The associated widget instance or <code>null</code> if a
    *           widget by that ID does not exist in this frame.
    */
   GenericWidget<?> getWidgetForId(int id)
   {
      return LogicalTerminal.getWidgetForId(id);
   }

   /**
    * Get an ID of window owning this frame or -1 when there is no owning window.
    * 
    * @return   See above.
    */
   int getWindowId()
   {
      return frame.getAttr("windowID", () -> frame.config.windowID);
   }

   /**
    * Looks up the header widget by its ID.
    * 
    * @param    id
    *           The numeric ID assigned to the header widget.
    * 
    * @return   The associated header widget instance or <code>null</code> if
    *           a widget by that ID does not exist in this frame.
    */
   GenericWidget<?> getHeaderWidgetForId(int id)
   {
      Iterator<GenericWidget<?>> itr = n2h.values().iterator();
      while (itr.hasNext())
      {
         GenericWidget<?> w = itr.next();
         if (w.getId() == id)
         {
            return w;
         }
      }

      return null;
   }
   
   /**
    * Let frame know that there was a condition during frame processing. 
    */
   void setCondition()
   {
      hadCondition = true;
   }
   
   /**
    * Read the next field from the given stream using any whitespace as the 
    * delimiter and validate the data against the widget's format string 
    * before assigning the data into the screen-buffer and back to the
    * user's variable.  Screen-buffer and widget processing can be bypassed
    * by passing <code>null</code> for these values and in this case, the
    * stream reading will not operate upon any of the user interface state.
    * <p>
    * If the current line buffer is empty, a new line will be read from the
    * stream.
    * <p>
    * In the following conditions, no assignment will be made:
    * <ul>
    *   <li> a field is requested after the entire line has already been
    *        processed
    *   <li> an unquoted hyphen is used as the contents of the field AND the 
    *        delimiter is NOT a hyphen (the index is still incremented past
    *        the field)
    *   <li> the first char of a field starts with a double quote but there
    *        is no closing quote (the index is incremented to the end of the
    *        buffer)
    * </ul>
    * <p>
    * After all field reads for a given line are complete, 
    * <code>Stream.resetCurrentLine</code> must be called to enable input for
    * the next line.  Otherwise, all subsequent reads will continue trying to
    * read from the current line. This will result in reading data left on
    * the current line or reading nothing if the line has already been fully 
    * read.  <code>Stream.resetCurrentLine</code> notifies the stream
    * environment that it should move to processing the next line.
    *
    * @param    in
    *           The input stream from which to read.
    * @param    data
    *           The variable to process.  If <code>null</code>, then the
    *           field must be read but the data should be discarded.
    * @param    frameBuf
    *           The frame buffer to update.  Should be <code>null</code> if
    *           there is no frame buffer to update.
    * @param    wid
    *           The widget to process.  Should be <code>null</code> if the
    *           <code>data</code> parameter is <code>null</code> or if there
    *           is no associated widget.
    * @param    fmt
    *           Format string to use for validation.  If <code>null</code> 
    *           obtain it from the widget (if specified) or use the default
    *           for the data type if there is no widget.
    * @param    assign
    *           <code>true</code> if the read data should be assigned back to
    *           the caller's <code>data</code> instance.
    *
    * @return   <code>true</code> if no error occurred, <code>false</code> if
    *           further processing should be aborted due to a failure in this
    *           method.
    *           
    * @throws   EndConditionException 
    * @throws   ErrorConditionException 
    */
   static boolean getField(Stream           in,
                           Accessor         data,
                           ScreenBuffer     frameBuf,
                           GenericWidget<?> wid,
                           String           fmt,
                           boolean          assign)
   throws EndConditionException,
          ErrorConditionException
   {   
      BaseDataType var = (data == null ? null : data.get().duplicate());
      Object    result = in.readFieldWorker(var);

      if (result != null)
      {         
         // first try to get the format string
         if (fmt == null && wid != null)
         {
            fmt = wid._getFormat();
         }
         
         // fallback plan
         if (fmt == null)
         {
            fmt = var.defaultFormatString();
         }
         
         // let the data type calculate the number of output chars in this format
         int len = var.formatLength(fmt);
         
         // truncate the text if the read data can't fit into the format
         if ((result instanceof String) && ((String) result).length() > len)
         {
            result = ((String) result).substring(0, len);
            
            if (var instanceof logical)
            {
               // no error in this case keep going
            }
            else
            {
               // for character, we can save the data into the screen buffer
               // like Progress does (we can't do this for the other data
               // types at this time)
               if (var instanceof character &&
                   frameBuf != null         &&
                   wid      != null)
               {
                  frameBuf.putWidgetValue(wid.getId(), new character((String) result));
               } 
               
               // throw the error
               String errmsg = "File input is too long: " + result; 
               ErrorManager.recordOrThrowError(557, errmsg);
               
               // if we are in silent error mode return
               return false;
            }
         }
         
         // handle unknown value in a generic way
         if (result == Stream.UNKNOWN_VALUE)
         {
            var.setUnknown();
         }
         else
         {         
            // convert the value to the proper type
            if (var instanceof integer)
            {
               ((integer) var).assign(new integer((String) result));
            }
            else if (var instanceof decimal)
            {
               ((decimal) var).assign(new decimal((String) result));
            }
            else if (var instanceof date)
            {
               ((date) var).assign(new date((String) result));
            }
            else if (var instanceof logical)
            {
               ((logical) var).assign(new logical((String) result, fmt));
            }
            else if (var instanceof character)
            {
               ((character) var).assign(new character((String) result));
            }            
         }
         
         // assign the value to the screen buffer
         if (frameBuf != null && wid != null)
         {
            frameBuf.putWidgetValue(wid.getId(), var);
         }
         
         // assign back to the caller's variable
         if (assign)
         {
            data.set(var);
         }
      }      
      
      return true;
   }
   
   /**
    * Adding new widget to the current frame.  This is the "inverted" version of the
    * <code>widget.setFrame()</code>.
    *
    * @param    widget
    *           The widget to be added to current frame.
    */
   void addWidget(GenericWidget<?> widget)
   {
      if (widget == null)
      {
         return;
      }
      widget.setFrame(this);
   }

   /**
    * Remove the specified widget from this frame.
    * 
    * @param    widget
    *           The widget to remove.
    */
   public void removeWidget(GenericWidget<?> widget)
   {
      List<GenericWidget<?>> newWidgets = new ArrayList<>(widgets.length);
      for (int i = 0; i < widgets.length; i++)
      {
         if (widgets[i] != widget)
         {
            newWidgets.add(widgets[i]);
         }
      }
      
      widgets = newWidgets.toArray(new GenericWidget[0]);
      
      Iterator<Map.Entry<String, GenericWidget<?>>> iter = n2w.entrySet().iterator();
      while (iter.hasNext())
      {
         Map.Entry<String, GenericWidget<?>> entry = iter.next();
         if (entry.getValue() == widget)
         {
            iter.remove();
         }
      }
   }

   /**
    * Adding new dynamic widget to the current frame.
    *
    * @param    dynWidget  The widget to be added to current frame.
    */
   void addDynamicWidget(GenericWidget<?> dynWidget)
   {
      // adding new widget to the name map
      boolean hitTest = n2w.put(getDynamicWidgetName(dynWidget), dynWidget) != null;
      
      if (hitTest)
      {
         return;
      }
      
      // add new widget to the widgets array of the current frame
      GenericWidget<?>[] newWidgets = new GenericWidget[n2w.size()];
      // at this time it is possible we just need to replace already existed widget
      boolean replaceExisted = n2w.size() == widgets.length;
      for (int i = 0; i < widgets.length; i++)
      {
         if (replaceExisted && widgets[i].getId() == dynWidget.getId())
         {
            newWidgets[i] = dynWidget;
         }
         else
         {
            newWidgets[i] = widgets[i];
         }
      }
      // the case when we append the new one
      if (!replaceExisted)
      {
         newWidgets[widgets.length] = dynWidget;
      }
      widgets = newWidgets;

      // register widget with global registry ID must be valid at this time
      // this cal is neutral to already registered widget 
      LogicalTerminal.registerWidget(dynWidget, dynWidget.getId());
      
      // refresh frame definition if exists
      // but if for the given ID we already have definition - skip this step
      if (frameDef != null && (!replaceExisted || frameDef.getConfig(dynWidget.getId()) == null))
      {
         frameDef.addConfig(dynWidget.config());
         if (frameBuf != null)
         {
            frameBuf.addWidget(dynWidget.getId());
         }
      }
      
      LogicalTerminal.attachRuntimeWidget(this, dynWidget);

      if (!this.getTitle().isUnknown() && dynWidget instanceof BaseEntity)
      {
         // adjust the widget's coordinates by re-setting them, if they were explicitly set
         // before the dynamic widget's frame was set - this is required to include the frame's
         // title
         
         // needs to be done after the FRAME attr was pushed to the client-side, so that the 
         // title-related offsets are set in the BaseConfig.frame*Offset fields

         BaseConfig bc = (BaseConfig) dynWidget.config;
         
         if (dynWidget.getAttr("y", () -> bc.y) != BaseConfig.INV_COORD)
         {
            ((BaseEntity) dynWidget).setXOrYWorker(false, bc.y, false);
         }
         
         if (dynWidget.getAttr("x", () -> bc.x) != BaseConfig.INV_COORD)
         {
            ((BaseEntity) dynWidget).setXOrYWorker(false, bc.x, true);
         }
      }
   }

   /**
    * Removing dynamic widget from the current frame.
    *
    * @param    dynWidget
    *           The widget to be deleted from the current frame.
    */
   void deleteDynamicWidget(GenericWidget<?> dynWidget)
   {
      deleteDynamicWidget(dynWidget, true);
   }

   /**
    * Removing dynamic widget from the current frame.
    *
    * @param    dynWidget
    *           The widget to be deleted from the current frame.
    * @param    pushScreenDefinitions
    *           <code>true</code> to push screen definitions, <code>false</code> to avoid pushing
    *           in order to prevent recursive invocations.
    */
   void deleteDynamicWidget(GenericWidget<?> dynWidget, boolean pushScreenDefinitions)
   {
      dynWidget.cleanup();

      // hide from the screen
      dynWidget.hide();

      // removing widget from name map
      n2w.remove(getDynamicWidgetName(dynWidget));

      int dynId = dynWidget.getId();

      // remove new widget from the widgets array of the current frame
      GenericWidget<?>[] newWidgets = new GenericWidget[n2w.size()];
      int j = 0;
      for (int i = 0; i < widgets.length; i++)
      {
         if (widgets[i].getId() != dynId)
         {
            newWidgets[j++] = widgets[i];
         }
      }
      widgets = newWidgets;
      
      if (fieldGroup != null)
      {
         fieldGroup.removeWidget(dynWidget);
      }

      // deregister widget from global registry
      LogicalTerminal.deregisterWidget(dynId);
      
      // refresh frame definition if exists
      if (frameDef != null)
      {
         frameDef.removeConfig(dynWidget.config());
         if (frameBuf != null)
         {
            frameBuf.removeWidget(dynId);
         }
      }

      if (pushScreenDefinitions && !frame.deleted)
      {
         LogicalTerminal.deleteDynamicWidget(this, dynWidget);
      }
   }

   /**
    * Place the frame.
    */
   void placeFrame() 
   {
      if (framePlacing)
      {
         return;
      }
      
      framePlacing = true;

      // everything is evaluated here - if the title is from a field and there is no record avail,
      // then an error will be shown.
      updateDynamicContent();
      
      if (!frame.getAttr("framePlaced", () -> frame.config.framePlaced))
      {
         // frame is not realized, just placed
         frame.config.framePlaced = true;
         LogicalTerminal.placeFrame(frame.config.id.asInt());
      }
      
      framePlacing = false;
   }
   
   /**
    * Construct the string name for dynamic widget based on unique ID number.
    *
    * @param    dynWidget  The widget the name will be generated to.
    *
    * @return   The String for dynamic widget name in string form of "ID".
    */
   private String getDynamicWidgetName(GenericWidget<?> dynWidget)
   {
      return Integer.toString(dynWidget.getId());
   }

   /**
    * Reset frame buffer. Make sure that static widgets are not reset if frame
    * was not displayed yet.
    * <p>
    * NOTE: Current approach assumes that frame is either, at least once
    * displayed or not displayed but only static widgets are assigned. This
    * may result to differences in behavior with Progress when regular
    * (non-static) widget is updated, frame is reset and only then is
    * displayed.
    */
   private void resetFrameBuf()
   {
      // Frame buffer resetting is not required if we use the current data
      // from one iteration to another
      if (hadView && !sameViewDataRedirected &&
          (frame.isDown() || !isOutRedirected()))
         frameBuf.reset();
   }

   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    * 
    * @param    widgets
    *           The list of widgets which is to be edited.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    * @param    windowClause
    *           <code>true</code> if windowClause is presented in DISPLAY stmt.
    *           <code>false</code> otherwise
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   private void promptForWorker(GenericWidget<?>[] widgets, 
                                EventList          wait,
                                boolean            windowClause,
                                handle             hWin)
   {
      int[] ids      = validIds(widgets);
      int   tempTerm = -1;
      
      // Special processing for SET and PROMPT-FOR statements in batch mode.
      if (errorOnBatchModeSetPrompt(513, NO_INPUT_SOURCE_ATTEMPT))
      {
         return;
      }
   
      if (ids == null)
      {
         // no valid widgets
         return;
      }
      
      conditionalUpDown();
   
      // special "double-vision" mode (no input redirection but output is
      // redirected, so we must allow interactive editing but then ensure
      // that the output is echoed to the redirected output stream
      if (UnnamedStreams.output() != null)
      {
         tempTerm = LogicalTerminal.switchOutput(-1, false);
      }
      
      Set<Integer> save = formatFilter;
      
      try
      {
         if (wait == null)
         {
            wait = new EventList(true);
         }
         wait.setOperation(EventList.PROMPT_FOR);
         
         formatFilter = new HashSet<>();
         
         for (int i = 0; i < ids.length; i++)
         {
            Object value = frameBuf.getWidgetValue(ids[i]);
            
            if (value instanceof character)
            {
               character chr = (character) value;
               
               if (!chr.isUnknown())
               {
                  if (chr.toStringMessage().trim().isEmpty())
                     formatFilter.add(ids[i]);
                  else
                     enableFinalFormatCheck();
               }
            }
         }
         
         LogicalTerminal.promptFor(this, ids, wait, true, hWin);
      }
      
      finally
      {
         formatFilter = save;
         
         // here is the extra view (puts the double in double-vision)      
         if (tempTerm != -1)
         {
            LogicalTerminal.switchOutput(tempTerm, false);
            
            if (windowClause)
            {
               view(hWin);
            }
            else
            {
               view();
            }
         }
      }
   }

   /**
    * Enables a list of widgets from the frame on the screen, shifts into
    * input blocking mode via {@link #waitFor} and after the user's edits
    * are complete the list of widgets are disabled.  No data is copied to
    * or from the screen buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * PROMPT-FOR
    * 
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is not used at all.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    * @param    windowClause
    *           <code>true</code> if windowClause is presented in DISPLAY stmt.
    *           <code>false</code> otherwise
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   private void promptForWorker(FrameElement[] widgets, 
                                EventList      wait, 
                                boolean        windowClause, 
                                handle         hWin)
   {
      Stream in = UnnamedStreams.input();
      
      if (in != null)
      {
         // redirected input case
         if (windowClause)
         {
            promptFor(in, widgets, hWin);
         }
         else
         {
            promptFor(in, widgets);
         }
      }
      else
      {
         setStaticWidgets(widgets);
         
         // process text grouping (word wrap mode)
         List<FrameElement> list = textGroup(widgets);
         
         // interactive case
         GenericWidget<?>[] wids = widgetListFromFrameElements(widgets);
         if (windowClause)
         {
            promptFor(wids, wait, hWin);
         }
         else
         {
            promptFor(wids, wait);
         }
         
         // reset text grouping (if any)
         textUngroup(list);
      }
   }

   /**
   /**
    * For each entry in the widget array, the input stream will be read and 
    * the widget's associated field in the screen buffer will be set. The 
    * data passed in will be copied from the user's variables into the screen
    * buffer.  After the input is read from the stream into the screen buffer,
    * that data will be copied back from the screen buffer to the user's
    * variables.  
    * <p>
    * Any error during stream reading or data parsing will cause an error
    * condition to be generated.  Reaching the end of the stream will cause
    * an end condition to be generated.  Unlike the <code>update</code>
    * version in this class which has a "hidden" retry block around the
    * core "set" processing, this streams version has no such feature. Errors
    * will be immediately seen by the caller.
    * <p>
    * Both the input stream and the current default output stream (if output
    * is redirected) are checked.  If both are set to <code>echo</code> mode,
    * then after all stream input is done, the screen buffer's contents will  
    * be echoed to the current output destination using {@link #view}. This
    * will occur whether or not an error occurs during stream processing and
    * data parsing.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * Note that this method does change the LASTKEY value to 13 if input 
    * stream contains at least one character. If input stream is empty then
    * END condition is triggered. 
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    * 
    * @param    in
    *           A stream instance which is open for reading. 
    * @param    data
    *           The list of widgets which are to receive input from the
    *           stream.
    * @param    windowClause
    *           <code>true</code> if windowClause is presented in DISPLAY stmt.
    *           <code>false</code> otherwise
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   private void updateWorker(Stream in, FrameElement[] data, boolean windowClause, handle hWin)
   {
      // Special processing for UPDATE statement in batch mode for redirected case.
      if (LogicalTerminal.isInBatchMode())
      {
         LogicalTerminal.activateBatchMode(false);
      }
   
      setStaticWidgets(data);
      copyToScreenBuffer(data);
      processStreamInput(in, data, true, windowClause, hWin);
   }

   /**
    * Copies data from the frame elements to each corresponding widget, then
    * enables the list of widgets, shifts into input blocking mode via
    * {@link #waitFor}, after the user's edits are complete the list of 
    * widgets are disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements.  
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method uses {@link #display} and {@link #set} to do its real
    * work.  The only difference is that the <code>set</code> is embedded
    * inside a "hidden" <code>TransactionManager</code> block which handles
    * undo/retry on <code>error</code> conditions and undo/leave on 
    * <code>end</code> conditions.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * UPDATE
    * 
    * @param    widgets
    *           The list of data and widgets which is to be updated.
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    * @param    windowClause
    *           <code>true</code> if windowClause is presented in DISPLAY stmt.
    *           <code>false</code> otherwise
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   private void updateWorker(FrameElement[] widgets, 
                             EventList      wait, 
                             boolean        windowClause,
                             handle         hWin)
   {
      // turning off infinite loop protection
      TransactionManager.disableLoopProtection(null);
   
      Stream in = UnnamedStreams.input();
      
      if (in != null)
      {
         // redirected input case
         if (windowClause)
         {
            update(in, widgets, hWin);
         }
         else
         {
            update(in, widgets);
         }
         
         return;
      }
   
      // Special processing for UPDATE statement in batch mode.
      if (LogicalTerminal.isInBatchMode() && UnnamedStreams.input() == null)
      {
         // no window is required
         LogicalTerminal.activateBatchMode(false);
         ErrorManager.displayError(516, NO_OUTPUT_DESTINATION_ATTEMPT, true);
         LogicalTerminal.stopBatchSession();
   
         return;
      }
      
      // interactive case
      setStaticWidgets(widgets);
      copyToScreenBuffer(widgets);
      
      // the set processing is done inside a "hidden" block that has retry
      // and undo support on errors, no end-key property
      TransactionManager.pushScope("_hiddenUpdateBlock_",
                                   TransactionManager.SUB_TRANSACTION,
                                   false,
                                   false,
                                   false,
                                   false);
      try
      {
         update:
         {
            TransactionManager.blockSetup();
            
            // force the TM to avoid accumulating any contained block's
            // interactions counter (this is a quirk of UPDATE that is not
            // duplicated in SET/PROMPT-FOR); note that only a nested trigger
            // can cause such an accumulation... so we must bypass that case
            // here
            TransactionManager.setRollup(false);
            
            do
            {
               try
               {
                  if (wait == null)
                  {
                     wait = new EventList(true);
                  }
                  wait.setOperation(EventList.UPDATE);
                  
                  // the real work is done here!
                  setWorker(widgets, wait, true, windowClause, hWin);
               }
               
               catch (ErrorConditionException err_except)
               {
                  // special case where we exit if this was caused by the 
                  // END-ERROR key, otherwise we retry
                  if (err_except.isEndError())
                  {
                     throw err_except;
                  }
                  
                  TransactionManager.rollback("_hiddenUpdateBlock_");
                  TransactionManager.triggerRetry(null);
               }
            }
            while (TransactionManager.needsRetry());
            
            if (TransactionManager.isBreakPending())
            {
               break update;
            }
         }
      }
      
      finally
      {
         // we don't need to reset our rollup flag since this hidden block
         // is exiting anyway and that flag only affects accumulation of
         // contained blocks NOT how our current block's interactions counter
         // is accumulated
         
         TransactionManager.popScope();
      }      
   }

   /**
    * Enables a list of widgets, shifts into input blocking mode via 
    * {@link #waitFor}, after the user's edits are complete the list of
    * widgets is disabled and the edited data is copied from the screen 
    * buffer to the listed frame elements. 
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * The assignment honors any {@link EmbeddedAssignment} instances in the
    * frame element list, in the order in which they appear.  Otherwise, 
    * these do not affect the user interface processing in any way.
    * <p>
    * If the unnamed input is redirected, this method will redirect its own
    * execution to the stream form.  In such a case, all interactive 
    * processing is bypassed.
    * <p>
    * This method corresponds to these Progress statements:
    * SET
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is only used on output (they are assigned
    *           to the values provided by the user).
    * @param    wait
    *           The list of events to treat as a <code>GO</code> event.
    * @param    def
    *           <code>true</code> means that default values will be used
    *           if screen buffer does not contain updated values.
    * @param    windowClause
    *           <code>true</code> if windowClause is presented in DISPLAY stmt.
    *           <code>false</code> otherwise
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   private void setWorker(FrameElement[] widgets, 
                          EventList      wait, 
                          boolean        def, 
                          boolean        windowClause, 
                          handle         hWin)
   {
      Stream in = UnnamedStreams.input();
      
      if (in != null)
      {
         // redirected input case
         if (windowClause)
         {
            set(in, widgets, hWin);
         }
         else
         {
            set(in, widgets);
         }
      }
      else
      {
         // Special processing for SET and PROMPT-FOR statements in batch mode.
         if (errorOnBatchModeSetPrompt(513, NO_INPUT_SOURCE_ATTEMPT))
         {
            return;
         }
         // interactive case
         if (wait == null)
         {
            wait = new EventList(true);
         }
         wait.setOperation(EventList.SET);
         
         boolean retry;
         int i = 0;
         
         do
         {
            if (windowClause)
            {
               promptFor(widgets, wait, hWin);
            }
            else
            {
               promptFor(widgets, wait);
            }

            retry = false;
            
            if (finalCheck)
            {
               finalCheck = false;
               
               for (i = 0; i < widgets.length; i++)
               {
                  if (widgets[i] == null)
                     continue;
                  
                  GenericWidget<?> widget = widgets[i].getWidget();
                  String format = widget._getFormat();
                  Object value = frameBuf.getWidgetValue(widget.getId());
                  
                  if (value == null || !frameBuf.isEntered(widget.getId()))
                     continue;
                  
                  // finally perform check
                  ErrorEntry res = DisplayFormatFactory.checkFormat((BaseDataType) value, 
                                                                    format);

                  if (res == null)
                     continue;
                  
                  LogicalTerminal.message(res, 
                                          new ErrorEntry(143, 
                                                         configClass.getSimpleName() + 
                                                         ": Unable to evaluate " +
                                                         "field for assignment"));
                  retry = true;
                  break;
               }
            }
            
            if (retry)
            {
               // i points to failed widget
               Utils.scroll(widgets, i);
            }
         }
         while (retry);
         
         // get the value in the screen buffer for each widget using the widget
         // id as the index, then assign this back to the variable from the caller
         
         copyFromScreenBuffer(widgets, def, currentOp == SET);
      }
   }
   
   /**
    * Returns the current logical line number in a frame.
    * <p>
    * This method corresponds to the Progress built-in function
    * FRAME-LINE
    * 
    * @return   1-based ordinal of the current data iteration of the frame.
    *           If the frame is redirected to a stream or if the frame is
    *           not visible, then 0 is returned. If the frame is not a down
    *           frame then 1 will be returned.
    *           Can return the value outside of the frame. For example for this
    *           sequence of calls for 4-down frame: 
    *           <pre>
    *             down(3). // postions on the 4th row 
    *             down().  // +1 
    *             down(1).  // in downWorker this method returns 5 
    *           </pre>
    *           this method will return 5. It also confroms to FRAME-LINE behavior.
    */
   private int frameLineWorker()
   {
      int line = -1;
      
      /*
      if (frameBuf != null)
      {
         line = frameBuf.getLine();
      }
      */
      
      // taking a trip to the client for this information is REALLY
      // expensive so we should avoid it, but to avoid it we must properly
      // maintain the frame-line value independently on the server... which
      // is tricky so for now we do it the safe way
      if (line == -1)
      {
         // couldn't find it locally, go down to the client to get it
         if (isRedirected() || !_isVisible())
            line = 0;
         else
            line = LogicalTerminal.getFrameLine(this, pendingUpDown).intValue();
      }
      
      return line;
   }
   
   /**
    * Processes any aggregator elements in the given array of frame elements.
    * The first time this method is invoked within a particular display loop,
    * an <code>AggregateHandler</code> instance is created.  This instance is
    * stored in an instance member, which is cleared when the display loop is
    * terminated.  Each time this method is invoked, the handler is requested
    * to accumulate data in all of the associated <code>Accumulator</code>
    * instances (if any) associated with the given frame elements.
    *
    * @param    fe
    *           The list of widgets that may need accumulation.
    * @param    accumulate
    *           Flag indicating if accumulation should be performed. If false,
    *           the widgets which have aggregators will be registered with
    *           this frame.
    */
   private void processAggregates(FrameElement[] fe, boolean accumulate)
   {
      // make a set of the accumulator widgets
      Set<FrameElement> wids = null;

      int len = fe.length;

      for (int i = 0; i < len; i++)
      {
         if (fe[i] != null && fe[i].statsColumn())
         {
            if (wids == null)
            {
               wids = new HashSet<>();
            }
            wids.add(fe[i]);
         }
      }

      if (wids != null)
      {
         // create or update the handler
         AggregatorElement[] accums = wids.toArray(new AggregatorElement[0]);

         try
         {
            if (aggregateHandler == null)
               aggregateHandler = new AggregateHandler(accums);
            else
               aggregateHandler.verifyAccumulators(accums);
         }
         catch (IllegalStateException exc)
         {
            if (LOG.isLoggable(Level.WARNING))
            {
               LOG.warning(exc.getMessage());
            }
         }
      }
      
      // accumulate data
      if (accumulate && aggregateHandler != null)
      {
         aggregateHandler.accumulate(fe);
      }
   }
   
   /**
    * Reports if this frame is a page element (a header or footer). This is
    * only <code>true</code> in the case that the output stream is paged AND
    * this frame is marked as a page-top (header) or a page-bottom (footer).
    *
    * @param    out
    *           The output destination.
    *
    * @return   <code>true</code> if this is a header or footer.
    */
   boolean isPageElement(Stream out)
   {
      if (out == null)
         return false;
      
      if (out instanceof StreamWrapper)
      {
         StreamWrapper sw = (StreamWrapper) out;
         
         if (sw.dereference() == null)
            return false;
      }
      
      return (out.rawPageSize() > 0 && (getPageBottom() || getPageTop()));
   }
   
   /**
    * Reports if the unnamed stream is redirected at this time.
    *
    * @return   <code>true</code> if the unnamed stream is redirected
    */
   boolean isRedirected()
   {
      return UnnamedStreams.isOut();
   }
   
   /**
    * Reports if the frame was already realized.
    *
    * @return   <code>true</code> if the frame was realized
    */
   boolean isRealized()
   {
      return frame._isRealized();
   }

   /**
    * Reports if there is a cached output stream for use with redirecting
    * the output from aggregation.
    *
    * @return   <code>true</code> if aggregation output is redirected
    */
   boolean isOutRedirected()
   {
      return redirOut != null && redirOut.isOut();
   }
   
   /**
    * Check if output stream is a terminal.
    * 
    * @return   <code>true</code> if output stream is terminal.
    */
   boolean isOutTerm()
   {
      return redirOut != null && redirOut.isTerm();
   }
   
   /**
    * Finish the config processing for this frame.
    * <p>
    * This includes registering the active configuration for each new widget, and synchronizing
    * the widget configuration with the actual widget state.
    */
   void finishConfigProcessing()
   {
      ConfigManager mgr = null;
      
      for (GenericWidget<?> w : allWidgets())
      {
         WidgetConfig cfg = w.config;
         if (mgr == null)
         {
            mgr = cfg.getConfigManager();
         }
         
         int sideLabelId = w.processSideLabel(false);
         frameDef.addSideLabel(sideLabelId, cfg.id.asInt());
         
         WidgetConfig active = mgr.getActiveConfig(cfg.id);
         if (active == null)
         {
            mgr.addWidgetConfig(cfg);
         }
         else
         {
            assert(cfg.getClass().equals(active.getClass()));
         }
         
         w.finishConfigProcessing();
      }
   }

   /**
    * Full initialization processing. 
    *
    * @param    configClass
    *           The class of the configuration implementation class.
    * @param    configInstance
    *           The instance of the configuration implementation class.
    * @param    frameName
    *           The frame name.
    */
   private void initialize(Class<?> configClass, Object configInstance, String frameName)
   {
      // register frame with LogicalTerminal
      frameId = LogicalTerminal.registerFrame(this);
      
      coreInitialize(configClass, configInstance, frameName, false);
      
      // create the associated screen buffer which is initially empty
      frameBuf = new ScreenBuffer(frameId, validIds(widgets));
   }
   
   /**
    * Core initialization processing. This stores both the legacy 4GL with case preserved as
    * well as the internal frame name, which is modified to be all lowercase. This allows for
    * easier case-insensive frame comparisons, as well as being compatible with the 4GL 
    * FRAME-NAME attribute, which corresponds to the first use of a frame's name.
    *
    * @param   configClass
    *          The class of the configuration implementation class.
    * @param   configInstance
    *          The instance of the configuration implementation class.
    * @param   frameName
    *          The frame name
    * @param   dynamic
    *          Flags the dynamic frame preparation.
    *
    * @return  {@code true} on success. If initialization could not be done (E.g. because the new
    *          frame definition does not match the master frame) the method returns {@code false}.
    */
   private boolean coreInitialize(Class<?> configClass,
                                  Object configInstance,
                                  String frameName,
                                  boolean dynamic)
   {
      this.configClass    = configClass;
      this.configInstance = configInstance;
      this.frameName      = frameName;
      
      WidgetList wl = (WidgetList) configInstance;
      n2w = wl.getWidgets();        // cache the widget map for this frame
      n2n = wl.getNames();          // cache the widget name map for this frame
      n2h = wl.getHeaderWidgets();  // cache the header widget map for this frame
      n2b = wl.getBackgroundWidgets();
      
      // init the widget array (note: the background widgets are also stored in 'widgets' field)
      widgets = new GenericWidget[n2w.size()];
      
      // dynamic frames should assign frame member to existed widget before this call
      if (!dynamic)
      {
         frame = new FrameWidget(false);
      }
      
      // set up name and object for frame widget
      frame.name(this.frameName.toLowerCase() + "##" + this.configClass.getName());
      frame.setFrame(this);
      
      initBackgroundGroup();
      
      createFieldGroup();
      
      // initialize the default field group
      currentFieldGroup = fieldGroup;
      rows2fg.put(1, fieldGroup);

      // dynamic frame initialization should stop here
      if (dynamic)
      {
         return true;
      }
      
      // this is only static frames valid call
      frame.config.id = new WidgetId(frameId);
      
      int idx = 0;  // widget array counter (this frame)
      
      // add all the widgets in the order of their definition
      Iterator<String> iter = n2w.keySet().iterator();
      while (iter.hasNext())
      {
         String wName = iter.next();            // widget's name
         int wid = WidgetId.nextID().asInt();   // request the next widget ID
         
         // initialize key state in the widget
         widgets[idx] = n2w.get(wName);         // get the widget
         widgets[idx].setFrame(this);
         widgets[idx].setId(wid);
         
         // save new widget in session registry map
         LogicalTerminal.registerWidget(widgets[idx], wid);
         
         idx++; // advance to next index in 'widgets' array 
      }
      
      // insert additional skip if headers are not empty
      if (n2h.size() > 0 && n2w.size() > 0)
      {
         SkipEntity skip = new SkipEntity(true);
         skip.setHeader(true);
         n2h.put("", skip);
      }
      
      idx = 0;
      // assign ids to header widgets in the order of their definition
      iter = n2h.keySet().iterator();
      while (iter.hasNext())
      {
         // get the widget
         GenericWidget<?> hw = n2h.get(iter.next());
         int wid = hw.getId();
         
         if (wid == -1)
         {
            wid = WidgetId.nextID().asInt(); // request the next widget ID
         }
         // initialize key state in the widget
         hw.setFrame(this);
         hw.setId(wid);
         
         // we don't add these to the widgets[] since header widgets are not real
         // however they occupy the real ID so save new widget in session registry map
         LogicalTerminal.registerWidget(hw, wid);
         idx++;
      }
      
      // this will be created after the frame setup.
      literalFieldGroup = null;
      defaultFrame = !dynamic && frameName.isEmpty();
      
      return true;
   }

   /**
    * Create BACKGROUND field group for the frame and add background widgets to it, if any.
    */
   private void initBackgroundGroup()
   {
      Collection<GenericWidget<?>> bg = n2b.values();
      FieldGroup fg = wrapWidgetsToFieldGroup(frame._dynamic(),
                                              bg.toArray(new GenericWidget[bg.size()]),
                                              null);
      fg.setWidgetAdder(this::addWidget);
      frame.setBackground(fg);
   }
   
   /**
    * Literal field group initialization procedure.  Used to return literal group to initial
    * state after modifications. 
    * 
    * @return   The literal field group for given frame.
    */
   private FieldGroup initLiteralFG()
   {
      if (!(frame.config.labels && !frame.config.sideLabels))
      {
         return null;
      }
      
      GenericWidget<?>[] litWidgets = new GenericWidget[0];
      
      // add real literal widget(s) to this list
      String[] labels = new String[widgets.length];
      int labelWidgets = 0;
      for (int i = 0; i < widgets.length; i++)
      {
         GenericWidget<?> widget  = widgets[i];
         if (widget._dynamic())
         {
            continue;
         }
         
         WidgetConfig cfg = widget.config();
         if (cfg instanceof ControlTextConfig)
         {
            ControlTextConfig cc = (ControlTextConfig) cfg;
            String lblStr = widget.getLabelStr();
            if (cc.labels && lblStr != null)
            {
               labels[i] = lblStr;
               labelWidgets = labelWidgets + 1;
            }
         }
      }
      
      if (labelWidgets > 0)
      {
         litWidgets = new GenericWidget[labelWidgets + 1];
         int idx = 1;
         StringBuilder sep = new StringBuilder();
         for (int i = 0; i < widgets.length; i++)
         {
            if (labels[i] == null)
            {
               continue;
            }
            
            GenericWidget<?> widget = widgets[i];
            String lbl = labels[i];
            StringHelper.repeatChar(lbl == null ? ' ' : '-', widget.getWidthChars().intValue(), sep);
            sep.append(i == widgets.length ? "" : " ");
            
            litWidgets[idx] = createLiteralWidget(widget, widget.getLabelStr());
            idx = idx + 1;
         }
         
         litWidgets[0] = createLiteralWidget(null, sep.toString());
      }
      
      return wrapWidgetsToFieldGroup(frame._dynamic(), litWidgets, null);
   }
   
   /**
    * Create a LITERAL widget to be added to the {@link #literalFieldGroup}.
    * 
    * @param    widget
    *           The widget from which the LABEL is resolved - is <code>null</code> for the separator.
    * @param    txt
    *           The text for the LITERAL widget.
    *           
    * @return   A new LITERAL widget which dispatches the SCREEN-VALUE to the associated widget's LABEL, or 
    *           uses the passed text for the separator LITERAL.
    */
   private LiteralWidget createLiteralWidget(GenericWidget<?> widget, String txt)
   {
      LiteralWidget lit = new LiteralWidget()
      {
         public character getScreenValue() 
         {
            return widget == null ? new character(txt) : new character(widget.getLabelStr());
         }
         
         @Override
         public void setScreenValue(BaseDataType value)
         {
            if (widget == null)
            {
               // OE abends when changing the separator...
            }
            else
            {
               // TODO: is this correct?
               widget.setLabel(value.toStringMessage());
            }
         }
      };
      
      lit.config.id = WidgetId.nextID();
      
      LogicalTerminal.registerWidget(lit, lit.config.id.asInt());
      lit.setFrame(this);
      
      return lit;
   }
   
   /**
    * Full initialization processing for empty frame.  Used for dynamic frame preparation.
    *
    * @param    configClass
    *           The class of the configuration implementation class.
    * @param    configInstance
    *           The instance of the configuration implementation class.
    * @param    frameWidget
    *           The frame widget to use as the base for new generic frame.
    */
   private void initializeEmptyFrame(Class<?> configClass,
                                     Object configInstance,
                                     FrameWidget frameWidget)
   {
      // this should be done first to avoid NPE in following call
      frame = frameWidget;
      
      // initialization common for dynamic and static frames
      coreInitialize(configClass, configInstance, getDynamicWidgetName(frameWidget), true);
      
      // special dynamic frames initialization part
      
      // assign the frame ID 
      frameId = frameWidget.getId();
      
      // create the associated screen buffer which is initially empty
      frameBuf = new ScreenBuffer(frameId, 0);
      
      // add frame to logical terminal registry
      LogicalTerminal.addDynamicFrame(this, frameId);
   }
   
   /**
    * Drive the input stream reading process, where each frame element in the
    * list resolves to the read of a field from the stream, a bypass of the
    * field in the stream or an embedded assignment. This is an alternative
    * process to the traditional {@link #promptFor} processing.  In 
    * particular, while the screen buffer is modified during this processing,
    * the state of the UI is not otherwise modified.  This is just a way to  
    * redirect input from a stream and read that input from user interface
    * specific APIs.
    * <p>
    * Note that this method does change the LASTKEY value to 13 if input 
    * stream contains at least one character. If input stream is empty then
    * END condition is triggered. 
    *
    * @param    in
    *           The input stream from which to read.
    * @param    wids
    *           The variables and widgets to process. Any <code>null</code>
    *           elements and widget only elements are ignored.
    * @param    assign
    *           <code>true</code> if the read data should be assigned back to
    *           the caller's <code>data</code> instance.
    * @param    windowClause
    *           <code>true</code> if windowClause is presented in DISPLAY stmt.
    *           <code>false</code> otherwise
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    *           
    * @throws   EndConditionException 
    * @throws   ErrorConditionException 
    */
   private void processStreamInput(Stream          in,
                                   FrameElement[]  wids,
                                   boolean         assign,
                                   boolean         windowClause,
                                   handle          hWin)
   throws EndConditionException,
          ErrorConditionException
   {
      if (in != null)
      {
         boolean abort = false;
         
         // errors are always doubled (see viewWorker()), we don't know why
         // Progress does this, but we duplicate it here
         try
         {
            abort = in.protectReads();
            
            if (abort)
            {
               // this only can occur in silent error mode (otherwise an
               // exception would have been thrown) so there are no errors to
               // double, just return
               return;
            }
         }
         
         catch (ErrorConditionException ece)
         {
            // ignore and mark the flag to allow the 2nd error to be generated
            abort = true;
         }
         
         // implement the second error if needed
         if (abort && in.protectReads())
            return;
      }
         
      in.resetCurrentLine();
      
      try
      {
         driver:
         for (int i = 0; i < wids.length; i++)
         {
            if (wids[i] != null && !wids[i].widgetOnly())
            {
               if (wids[i].assignOnly())
               {
                  if (assign)
                  {
                     // implement embedded assignments
                     wids[i].set(null);
                  }
               }
               else
               {
                  // honor bypass (the CARET in Progress)
                  Accessor access = (wids[i].bypass() ? null : wids[i]);
                  GenericWidget<?> ww = wids[i].bypass() ? null : wids[i].getWidget();
                  
                  if (!getField(in, access, frameBuf, ww, null, assign))
                  {
                     break driver;
                  }
               }
            }
         }
         
      }
      
      finally
      {
         in.resetCurrentLine();
         
         Stream out = UnnamedStreams.output();
         
         // only echo if the input stream is marked for echoing AND if
         // no output redirection is active OR output redirection is active
         // AND that stream is marked for echoing
         if (in.getEcho() && (out == null || out.getEcho()))
         {         
            // cause echo of our data (whatever made it to the screen-buffer)
            if (windowClause)
            {
               view(hWin);
            }
            else
            {
               view();
            }
         }
         
         // set LASTKEY to 13 as if there was full line
         KeyReader.setLastKey(13);
      }
   }

   /**
    * Core widget accessor processing, which handles the various widget name
    * forms and the lookup of the widget including proper processing of array
    * elements.
    * <p>
    * This method handles name processing in both the J2SE-compatible form
    * as well as the "simplified" naming that always decapitalizes the first
    * letter of all widget names.
    * <p>
    * This method also handles the case when one widget represents entire 
    * array of widgets.
    *
    * @param    widgetName
    *           The base name of the widget to lookup, taken directly from
    *           the proxy method name.
    * @param    op
    *           The type of the operation (get, set or accessor).
    * @param    numArgs
    *           The number of arguments passed to the proxy handler. 
    * @param    index
    *           The index of an array element or <code>null</code> if no
    *           array processing is needed.
    *
    * @return   The widget by that name or <code>null</code> if no widget
    *           can be found by that name.
    */
   private GenericWidget<?> getWidget(String widgetName,
                                      int    op,
                                      int    numArgs,
                                      Object index)   
   {
      // detect if this is an array form
      boolean array = (op == INVOKE_GET    && numArgs == 1) ||
                      (op == INVOKE_SET    && numArgs == 2) ||
                      (op == INVOKE_ACCESS && numArgs == 1);
                      
      // first try our "simplistic" but non-J2SE-compatible widget name which
      // always has the first char converted to lower case
      char[] cname = widgetName.toCharArray();

      if (cname.length > 0)
         cname[0] = Character.toLowerCase(cname[0]);
 
      String name = new String(cname);
      
      GenericWidget<?> result = null; 
      
      if (array)    
      {
         // try non-indexed access
         result = getWidgetByName(name);
         
         if (result != null)
            return result;
         
         // for an indexed array element, recalculate the target widget name
         name = generateWidgetName(name, op, index);
      }
      
      result = getWidgetByName(name);

      if (result != null)
         return result;

      // the simplistic form failed, try the J2SE-compatible form
      name = Introspector.decapitalize(widgetName);
      
      if (array)  
      {
         // try non-indexed access
         result = getWidgetByName(name); 
         
         if (result != null)
            return result;
         
         name = generateWidgetName(name, op, index);
      }
      
      result = getWidgetByName(name);
      
      return result;
   }
   
   /**
    * Simple helper to lookup a widget by a given name.
    *
    * @param    name
    *           The case-sensitive widget name to retrieve.
    *
    * @return   The widget by that name or <code>null</code> if no widget
    *           can be found by that name.
    */
   private GenericWidget<?> getWidgetByName(String name)
   {
      GenericWidget<?> res = n2w.get(name);
      return (res != null) ? res : n2h.get(name);
   }
   
   /**
    * Returns the current screen value of the specified widget. 
    * 
    * @param    widgetName
    *           The widget name. 
    * @param    dataType
    *           The widget data type; a class representing one of the
    *           following widget data types:
    *           <ul>
    *             <li><code>character</code>;
    *             <li><code>integer</code>;
    *             <li><code>decimal</code>;
    *             <li><code>logical</code>;
    *             <li><code>date</code>.
    *           </ul>
    *
    * @return    <code>Object</code> currently in the screen buffer as the
    *            value of the widget or one of these defaults, if no value
    *            has been set:
    *           <ul>
    *             <li><code>character</code> - "";
    *             <li><code>integer</code> - 0;
    *             <li><code>decimal</code> - 0;
    *             <li><code>logical</code> - no (false);
    *             <li><code>date</code> - ? (unknown).
    *           </ul>
    */
   @SuppressWarnings("unused")
   private Object getter(String widgetName, Class<?> dataType)
   {
      int index = getWidgetId(widgetName);
      
      if (index < 0)
      {
         throw new RuntimeException("Invalid widget name " + widgetName);
      }
      
      return getter(index, dataType, true, false);
   }
   
   /**
    * Returns the current screen value of the specified widget. 
    * 
    * @param    widgetId
    *           The widget ID. 
    * @param    dataType
    *           The widget data type; a class representing one of the
    *           following widget data types:
    *           <ul>
    *             <li><code>character</code>;
    *             <li><code>integer</code>;
    *             <li><code>decimal</code>;
    *             <li><code>logical</code>;
    *             <li><code>date</code>.
    *           </ul>
    * @param    defaults
    *           <code>true</code> if a default value should be returned when
    *           the field is uninitialized in the screen buffer.
    * @param    validate 
    *           Perform validation.
    *
    * @return    <code>Object</code> currently in the screen buffer as the
    *            value of the widget or one of these defaults, if no value
    *            has been set:
    *           <ul>
    *             <li><code>character</code> - "";
    *             <li><code>integer</code> - 0;
    *             <li><code>decimal</code> - 0;
    *             <li><code>logical</code> - no (false);
    *             <li><code>date</code> - ? (unknown).
    *           </ul>
    */
   Object getter(int widgetId, Class<?> dataType, boolean defaults, boolean validate)
   {  
      GenericWidget<?> widget = getWidgetForId(widgetId);

      // if we are in an EDITING block, all getters resolve to the SCREEN-VALUE
      Object value = (currentOp == NONE ? frameBuf.getWidgetValue(widgetId)
                                        : getScreenValue(widget));

      // get the value from the screen buffer by index
      String format = widget._getFormat();
      
      value = convertScreenValue(dataType, value, format);

      if (value == null)
      {
         // cut off early if value is null and no need to generate defaults
         if (!defaults)
         {
            // try to get the screen value from the widget itself
// TODO: the commented out code below causes the serious regressions in ChUI mode
//       need more investigation and better approach to implement the idea behind
//            if (currentOp == NONE)
//            {
//               return convertScreenValue(dataType, widget.getScreenValue(), format);
//            }
            
            return null;
         }
            
         // generate a default value if nothing is in the screen buffer
         if (BaseDataType.class.isAssignableFrom(dataType))
         {
            value = BaseDataType.generateDefault(dataType);
         }
         else
         {
            throw new RuntimeException("unsupported data type " +  dataType);
         }
      }
      
      if (validate && !formatFilter.contains(widgetId))
      {
         // finally perform check
         ErrorEntry res = DisplayFormatFactory.checkFormat((BaseDataType) value, format);
               
         if (res != null)
         {
            try
            {
               // if NO-ERROR is active or there are too many messages, 
               // then postpone error messaging and force final format check
               if (!ErrorManager.isSilent() && !silentFormatErrors)
               {
                  // current-window is used

                  LogicalTerminal.message(res, 
                                          new ErrorEntry(71, 
                                                         "The value input for " + 
                                                         widget.getLabelStr() + 
                                                         " is not valid"));
                  silentFormatErrors = true;
               }
               else
               {
                  enableFinalFormatCheck();
               }
            }
            catch (Throwable e)
            {
               if (e instanceof StopConditionException)
                  throw (ConditionException) e;

               if (e instanceof QuitConditionException)
                  throw (ConditionException) e;

               // ignore everything else
            }
         }
      }
      
      return value;
   }

   /**
    * Enable format check at the end of input loop.  
    */
   private void enableFinalFormatCheck()
   {
      finalCheck = true;
   }

   /**
    * Sets a new value for the specified widget in the screen buffer.
    * 
    * @param   widgetName
    *          widget name
    * @param   value
    *          An object representing the widget value. Since this code is
    *          called from the Java Reflection API, any time a primitive
    *          (e.g. <code>int</code>) is passed, Reflection automatically
    *          wraps it in an <code>Object</code> of the correct type (e.g.
    *          <code>Integer</code>).  Thus it is safe to always expect the
    *          value to be an <code>Object</code>.
    * @return  <code>null</code> to comply with <code>void</code>
    */
   @SuppressWarnings("unused")
   private Object setter(String widgetName, Object value)
   {
      int index = getWidgetId(widgetName);

      if (index < 0)
      {
         throw new RuntimeException("Invalid widget name " + widgetName);
      }      
      
      return setter(index, value);
   }
      
   /**
    * Sets a new value for the specified widget in the screen buffer.
    * 
    * @param   index
    *          The widget ID.
    * @param   value
    *          An object representing the widget value. Since this code is
    *          called from the Java Reflection API, any time a primitive
    *          (e.g. <code>int</code>) is passed, Reflection automatically
    *          wraps it in an <code>Object</code> of the correct type (e.g.
    *          <code>Integer</code>).  Thus it is safe to always expect the
    *          value to be an <code>Object</code>.
    * @return  <code>null</code> to comply with <code>void</code>
    */
   private Object setter(int index, Object value)
   {
      if (index < 0)
      {
         throw new RuntimeException("Invalid widget!");
      }      
      
      Object val = null;

      // wrapper the value according to these rules:
      // String  -> character
      // Integer -> integer
      // Double  -> decimal
      // Boolean -> logical
      if (value instanceof String)
      {
         val = new character((String) value);
      }
      else if (value instanceof Integer)
      {
         val = new integer(((Integer) value).intValue());
      }
      else if (value instanceof Double)
      {
         val = new decimal(((Double) value).doubleValue());
      }
      else if (value instanceof Boolean)
      {
         val = new logical(((Boolean) value).booleanValue());
      }
      else if (value instanceof BaseDataType)
      {
         // we need to make a copy of the original value because otherwise
         // any assign() in the calling code will actually modify the value
         // in the screen buffer, which would be incorrect behavior
         val = ((BaseDataType) value).duplicate();
         
         if (value instanceof unknown)
         {
            ControlEntity<?> widget = (ControlEntity<?>) getWidgetForId(index);
            val = translateUnknown((BaseDataType) value, widget);
         }
      }
      else
      {
         // basic Object support
         val = value;
      }
         
      // setting the value in the screen buffer by index
      frameBuf.putWidgetValue(index, val);

      if (insideSetup && val instanceof BaseDataType)
      {
         getWidgetForId(index).config.frameDefValue = (BaseDataType) val;
      }
      
      return null;
   }
   
   /**
    * Generate an indexed widget name from a simple base name and a numeric
    * object that represents the index.  This method converts the numeric
    * object into a string using simple integer based formatting and then
    * concatenates this to the base name.
    *
    * @param    base
    *           The widget's base name.
    * @param    op
    *           The type of the method name (getter/setter/accessor).  Only
    *           used for error text.
    * @param    index
    *           The number to convert into a string representation.
    *
    * @return   The resulting name.
    *
    * @throws   RuntimeException
    *           If the index object is not a supported numeric object.
    */
   private String generateWidgetName(String base, int op, Object index)
   throws RuntimeException
   {
      String result = indexAsString(index);
      
      if (result == null)
      {
         String type = "accessor";
         
         if (op == INVOKE_GET)
         {
            type = "getter";
         }
         else if (op == INVOKE_SET)
         {
            type = "setter";
         }
         
         throw new RuntimeException("illegal " + type + " call");
      }
      
      return base + result.trim();
   }
   
   /**
    * Takes a numeric type (<code>NumberType</code>, <code>Integer</code> or
    * <code>Double</code>) and converts this into a string that can be used
    * as a name component for a widget getter/setter/accessor.
    *
    * @param    index
    *           The number to convert into a string representation.
    *
    * @return   The resulting string or <code>null</code> if the type is not
    *           one of the supported numeric types.
    */
   private String indexAsString(Object index)
   {
      String result = null;
      
      if (NumberType.class.isInstance(index))
      {
         result = ((NumberType)index).toString(">>>>>>>9");
      }
      else if (Integer.class.isInstance(index))
      {
         result = ((Integer)index).toString();
      }
      else if (Double.class.isInstance(index))
      {
         int idx = ((Double)index).intValue();
         
         result = Integer.toString(idx); 
      }
      
      return result;
   }
   
   /**
    * Ensure that any instance of the <code>unknown literal</code> is
    * converted into the proper type as expected by the given widget. 
    * <p>
    * The <code>unknown literal</code> is represented by the class
    * {@link unknown} which is a subclass of {@link BaseDataType}.  
    * This is only used to represent the literal and it needs to be converted 
    * to the proper type for use by the widget.
    * <p>
    * This must not be confused with a {@link BaseDataType} such as
    * an {@link integer} which is set to the <code>unknown value</code>.
    * Such cases are not literals.  Note that this method's purpose is to
    * convert the literal form into the proper non-literal type.
    *
    * @param    value
    *           The value to be converted if necessary.
    * @param    widget
    *           The widget being processed.
    *
    * @return   The original value if the type is not <code>unknown</code> or
    *           an instance of the proper type (set as 
    *           <code>unknown value</code>) otherwise.
    */
   private BaseDataType translateUnknown(BaseDataType  value, ControlEntity<?> widget)
   {
      if (value instanceof unknown)
      {
         // honor the type of the widget's data
         Class<?> type = widget.getDataClass();
         
         // honor user attempt to override "unknown" date field behavior 
         // (empty string) 
         if (type == date.class)
            type = character.class;
         
         // get an instance that represents unknown value
         value = BaseDataType.generateUnknown(type);
      }
      
      return value;
   }

   /**
    * Set each value in the screen buffer for each widget using the widget
    * id as the index.  Please note that assign only, widget only and 
    * <code>null</code> entries in the array are ignored.
    *
    * @param    data
    *           The list of elements to copy into the screen buffer.
    */
   private void copyToScreenBuffer(FrameElement[] data)
   {
      boolean validToRealize = true;
      
      for (int i = 0; i < data.length; i++)
      {  
         if ( data[i] != null      && 
             !data[i].assignOnly() &&
             !data[i].widgetOnly() &&
             !data[i].bypass())
         {
            Object value = (data[i].hasFormat()) 
                           ? data[i].getFormattedValue()
                           : data[i].get();
                           
            ControlEntity<?> widget = (ControlEntity<?>) data[i].getWidget();
            
            // check if the widget can be realized with it's parameters
            if (!widget.preRealizeCheck() && validToRealize)
            {
               validToRealize = false;
            }
            
            if (value instanceof unknown)
            {
               value = translateUnknown((BaseDataType) value, widget);
            }

            widget.setScreenValue(frameBuf, value, true);
         }
      }
      
      // there are problems in widget realize, stop the client session
      if (!validToRealize)
      {
         LogicalTerminal.pause();
         throw new StopConditionException("");
      }
      
      if (data.length > 0 && triggerActive > 0)
      {
         LogicalTerminal.setHardFormatCheck(this.getFrameId());
      }
   }

   /**
    * Set value of each static widget mentioned in the list.
    * 
    * @param    data
    *           The list of elements.
    */
   private void setStaticWidgets(FrameElement[] data)
   {
      for (int i = 0; i < data.length; i++)
      {  
         if (data[i] == null                             || 
             data[i].assignOnly()                        ||
             data[i].widgetOnly()                        ||
             data[i].getWidget() instanceof FillInWidget ||
             !(data[i].getWidget() instanceof ControlTextWidget))
         {
            continue;
         }

         BaseDataType     value  = data[i].get(); 
         ControlEntity<?> widget = (ControlEntity<?>) data[i].getWidget();
         
         if (value instanceof unknown)
         {
            value = translateUnknown(value, widget);
         }
         
         // assuming one cannot use the @ base field feature with a static
         // widget, we shouldn't need to process format strings here so we 
         // keep this simple
         frameBuf.putWidgetValue(widget.getId(), value);
      }
   }
   
   /**
    * This method should be called anytime the frame buffer is sent down
    * to the thin client.
    * <p>
    * The latest header values will be obtained and stored in the header map.
    * Any subsequent display on the client will reflect the latest headers.
    * <p>
    * Any dynamic down value that has been registered will be queried and
    * stored in the frame buffer at this time.  The user of the frame may
    * have changed this value at any time.  We read the latest value from
    * the accessor (the delegated access to this variable) and if it is
    * positive it is stored in the screen buffer directly.  Any non-positive
    * value is stored as 0.  If there is no registered dynamic down value,
    * then this processing is bypassed completely.
    * <p>
    * Any dynamic title expression that has been registered will be evaluated
    * and the result will be stored in the frame buffer. If there is no 
    * registered dynamic title expression, then this processing is bypassed
    * completely.
    * <p>
    * There is no counterpart to this method upon return from the client 
    * since headers and dynamic down values can never be edited, they
    * are only sent down to the client (one way rather than bidirectional).
    */
   private void updateDynamicContent()
   {
      if (headers != null)
      {
         frameBuf.allocateHeaders(headers);
      }

      frameBuf.setCurrentStatement(currentStatement);

      if (frame._isVisible())
      {
         return;
      }

      if (dynDown != null)
      {
         NumberType num  = (NumberType) dynDown.resolve();
         int        down = num.intValue();
         
         frameBuf.putDown((down < 0) ? 0 : down);
      }
      
      if (dynColumn != null)
      {
         NumberType num    = (NumberType) dynColumn.resolve();
         
         if (!(num.isUnknown() || Double.compare(num.doubleValue(), 0.0d) <= 0))
         {
            frame.setColumn(num);
         }
      }

      if (dynRow != null)
      {
         NumberType num = (NumberType) dynRow.resolve();
         
         if (!(num.isUnknown() || Double.compare(num.doubleValue(), 0.0d) <= 0))
         {
            frame.setRow(num);
         }
      }
      
      updateDynamicTitle();
      updateDynamicBrowseTitle();
   }
   
   /**
    * Update dynamic browse title.
    */
   private void updateDynamicBrowseTitle()
   {
      for (int i = 0; i < widgets.length; i++)
      {
         if (widgets[i] instanceof BrowseWidget)
         {
            BrowseWidget brws = (BrowseWidget) widgets[i];
            
            brws.updateDynamicTitle();
         }
      }
   }

   /**
    * Update dynamic frame title.
    */
   protected void updateDynamicTitle()
   {
      if (dynTitle != null)
      {
         character[] title = new character[1];
         
         Runnable code = () -> title[0] = dynTitle.execute();
         
         if (frameBuf.getTitle() != null)
         {
            if (errHlp == null)
            {
               errHlp = ErrorManager.getErrorHelper();
            }
            
            errHlp.ignore(code);
         }
         else
         {
            // Bracket the dynamic title expression execute() in silent error
            // mode if not already enabled.
            boolean warning = ErrorManager.isWarningMode();
            
            if (!warning)
            {
               ErrorManager.warningModeEnable();
            }
            
            try
            {
               title[0] = dynTitle.execute();
            }
            
            finally
            {
               if (!warning)
               {
                  ErrorManager.warningModeDisable();
               }
            }
         }
         
         String txt = (title[0].isUnknown() ? "" : title[0].toStringMessage());
         
         // force the title to go down with the next screen buffer
         frameBuf.putTitle(txt);
         
         // keep the screen definition in sync BUT don't trigger a push here
         FrameConfig fc = frame.config;
         fc.title = txt;
      }      
   }

   /**
    * Process list of frame elements and link widgets into text (word wrap)
    * groups if necessary. Resulting set of widgets is returned as list for
    * use by other methods.
    * 
    * @param   list
    *          Array of <code>FrameElement</code> instances.
    * 
    * @return  List of frame elements which refer widgets linked into text
    *          groups. All frame elements are placed in one list regardless
    *          from group.
    */
   private List<FrameElement> textGroup(FrameElement[] list)
   {
      ArrayList<FrameElement> res = new ArrayList<>();
      
      FillInWidget prev = null;
      int currGroup = -1;
      
      for (int i = 0; i < list.length; i++)
      {
         if (list[i] == null)
            continue;
         if (list[i].wordWrap())
         {
            res.add(list[i]);
            
            int id = ((WordWrapElement) list[i]).getIndex();

            if (id != currGroup)
               prev = null;
            
            currGroup = id;
            
            FillInWidget fin = (FillInWidget) list[i].getWidget();
            fin.startGrouping(prev);
            prev = fin;
         }
      }
      
      if (res.size() > 0)
         pushScreenDefinition();
      
      return res;
   }

   /**
    * Reset text grouping (word wrap mode).
    * 
    * @param   list
    *          List of <code>FrameElement</code> instances.
    */
   private void textUngroup(List<FrameElement> list)
   {
      for (Iterator<FrameElement> iter = list.iterator(); iter.hasNext();)
      {
         FrameElement item = iter.next();
         
         ((FillInWidget) item.getWidget()).resetGrouping();
      }
      
      if (list.size() > 0)
         pushScreenDefinition();
   }
   
   /**
    * Get the value in the screen buffer for each widget using the widget
    * id as the index, then assign this back to the variable from the 
    * caller.  Please note that widget only and <code>null</code> entries
    * in the array are ignored.
    * <p>
    * This method honors {@link EmbeddedAssignment} elements in the proper
    * order in the array.
    *
    * @param    data
    *           The list of elements to copy from the screen buffer.
    * @param    defaults
    *           <code>true</code> means that default values will be used
    *           if screen buffer does not contain updated values.
    * @param    calledBySet
    *           The true value indicates that the current screen operation is SET, otherwise 
    *           the current screen operation is UPDATE.
    */
   private void copyFromScreenBuffer(FrameElement[] data, boolean defaults, boolean calledBySet)
   {
      // bracket the assignment with a notification
      TransactionManager.batchStart();
      
      for (int i = 0; i < data.length; i++)
      {  
         if (data[i] != null && !data[i].widgetOnly() && !data[i].bypass())
         {
            BaseDataType value = null; 
            
            // read from screen buffer if we are not assign only
            if (!data[i].assignOnly())
            {
               GenericWidget<?> wid  = data[i].getWidget();
               Class<?>         type = ((ControlEntity<?>) wid).getDataClass();
               
               // get the value out of the screen buffer
               value = (BaseDataType) getter(wid.getId(), type, defaults, false);
               
               if (value == null)
               {
                  // no value in the screen buffer, so don't assign back
                  continue;
               }
               
               // assign the value back (embedded assignments handle their
               // own data value, so we pass null in these cases)
               if (wid != null &&
                        ((calledBySet && !frameBuf.isUninitialized(wid.getId())) ||
                         (!calledBySet && frameBuf.isChanged(wid.getId()))))
               {
                  data[i].set(value);
               }
            }
         }
      }
      
      // end of assign bracket
      TransactionManager.batchStop();
   }
   
   /**
    * Makes the frame and all its field-level widgets enabled and visible.
    */
   private void makeFrameEnable()
   {
      if (makeFrameEnableImpl())
      {
         pushScreenDefinition();
      }
   }
   
   /**
    * Makes the frame and all its field-level widgets enabled and visible.
    * 
    * @return   the flag indicating whether {@link #pushScreenDefinition()}
    *           must be called upon return from the method.
    */
   private boolean makeFrameEnableImpl()
   {
      finishSetup(true, false);

      // flag indicating we need to refresh screen definition on client side
      boolean needRefreshScreen = false;

      for (int i = 0; i < widgets.length; i++)
      {
         if (widgets[i] instanceof BrowseColumnWidget)
         {
            continue;
         }
         widgets[i].config().visible = true;

         // statement "ENABLE ALL" cleans the hidden flag for widget
         if (widgets[i]._isHidden())
         {
            needRefreshScreen = true;
            widgets[i].config().hidden = false;
         }
         
         if (widgets[i] instanceof FrameWidget)
         {
            needRefreshScreen = widgets[i].frame.makeFrameEnableImpl() ||
                                needRefreshScreen;
         }
         
      }

      if (!frame.getAttr("hidden", () -> frame.config.hidden, true))
      {
         // make the frame visible
         frame.config.wasVisible = frame.config.visible;
         frame._setVisible(true);
      }
      
      return needRefreshScreen;
   }

   /**
    * Finish initialization for all widgets in the frame: {@link GenericWidget#finishSetup()} is
    * called for each widget.
    *
    * @param withHidden
    *        If <code>true</code> then {@link GenericWidget#finishSetup()} is called for hidden
    *        widgets too. If <code>false</code>, only visible widgets are affected.
    * @param recursive
    *        If <code>true</code> then this function is recursively called for the child frames
    *        (if <code>withHidden</code> if <code>false</code>, only visible child frames are
    *        affected).
    */
   private void finishSetup(boolean withHidden, boolean recursive)
   {
      for (GenericWidget<?> widget : widgets)
      {
         if (withHidden || !widget._isHidden())
         {
            if (recursive && (widget instanceof FrameWidget))
            {
               widget.frame.finishSetup(withHidden, true);
            }
            widget.finishSetup();
         }
      }
   }

   /**
    * Makes the frame and all its field-level widgets visible.
    */
   private void makeFrameVisible()
   {
      finishSetup(false, true);
      updateDynamicContent();

      // make all field-level widgets visible, unless are marked as HIDDEN
      for (int i = 0; i < widgets.length; i++)
      {
         if (!widgets[i]._isHidden())
         {
            if (widgets[i] instanceof FrameWidget)
            {
               if (!frame.config.visible)
               {
                  // child frames are made visible only if the parent frame is not visible.
                  widgets[i].frame.makeFrameVisible();
               }
            }
            else
            {
               // Direct access; for some reason, this code doesn't push the visibility change to
               // the client.
               // Update: if pushScreenDefinition will be used, that will ruin the following
               // case:
               //   form i j with frame f1.
               //   display i with frame f1.
               // j value will be displayed, which is incorrect.
               widgets[i].config().visible = true;
            }
         }
      }

      // CA: if you push here visibility attributes, then frame visibility state will be changed when the
      // attribute is processed by TC.pushWidgetAttr - this is not OK, we need to let the UI statement do the
      // actual work on the FWD client side, and leave these 'unchanged' until then.

      // make the frame visible
      boolean wasVisible = frame.getAttr("visible", () -> frame.config.visible, true);
      frame.config.wasVisible = wasVisible;
      frame.pushWidgetAttr("wasVisible", frame.config.wasVisible);
      frame._setVisible(true);
      // set the hidden flag to OFF
      frame._setHidden(false);
   }
   
   /**
    * Create an array of all widgets in this frame that are in the given
    * array.
    *
    * @param    list
    *           The widgets to include in the resulting list.
    *
    * @return   The list of remaining ids or <code>null</code> if no ids
    *           were included.
    */
   public int[] validIds(GenericWidget<?>[] list)  
   {
      if (list == null || list.length == 0)
         return null;

      Set<Integer> widgetIds = new LinkedHashSet<>();

      // add included widget IDs
      for (int i = 0; i < list.length; i++)
      {
         // null values are silently ignored
         if (list[i] == null)
            continue;
         
         int id = getWidgetId(list[i]);
         
         // invalid widget
         if (id == -1)
            continue;
         
         widgetIds.add(id);
      }
     
      return idSetToArray(widgetIds);
   }

   /**
    * Create an array of widgets in this frame that are static including possible header ones.
    *
    * @param    headerWidgetsIncluded
    *           If <code>true</code> header widgets will be included.
    *
    * @return   The list of widget ids for this frame.
    */
   public int[] getStaticWidgetIds(boolean headerWidgetsIncluded)
   {
      Set<Integer> widgetIds = new LinkedHashSet<>();

      // first add mandatory usual widgets (widgets may be null if frame failed initialization)
      if (widgets != null)
      {
         for (GenericWidget<?> widget : widgets)
         {
            // all widgets are collected, except child frames, as when a frame gets deleted
            // (implicitly or explicitly) its direct children are deleted, except for static frame 
            // children; the frame children are not removed here because (if dynamic) they will
            // be explicitly deleted via FrameWidget.frameDelete
            if (!(widget instanceof FrameWidget))
            {
               widgetIds.add(widget.getId());
            }
         }
      }
      
      // header widgets should be included too
      if (headerWidgetsIncluded)
      {
         // enumerate all the header widgets in the order of their definition
         for (String s : n2h.keySet())
         {
            // get the widget
            GenericWidget<?> gw = n2h.get(s);
            
            if (!gw._dynamic())
            {
               widgetIds.add(gw.getId());
            }
         }
      }

      return idSetToArray(widgetIds);
   }
   
   /**
    * Makes widgets in the frame invisible.
    * 
    * @param    data
    *           Array of widgets to hide.
    */
   @Override
   public void hide(FrameElement[] data)
   {
      hide(data, false);
   }

   /**
    * Makes widgets in the frame invisible.
    * 
    * @param    data
    *           Array of widgets to hide.
    * @param    noPause
    *           <code>true</code> if NO-PAUSE option was used
    */
   @Override
   public void hide(FrameElement[] data, boolean noPause)
   {
      //TODO: add batch hide to ClientExports and use it
      for (FrameElement frameElement : data)
      {
         frameElement.getWidget().hide(noPause);
      }
   }
   
   /**
    * Refresh current DOWN value.
    * 
    * @param    down
    *           New DOWN value reported by client.
    */
   void refreshDown(int down)
   {
      dynamicDown = down;
   }

   /**
    * Reset 'flushed' flag. This means that frame need to be flushed manually 
    * if there is an unexpected exit from loop. 
    */
   void resetFlushed()
   {
      flushed = false;
   }

   /**
    * Return <code>true</code> if frame content was flushed since last VIEW.
    * 
    * @return   <code>true</code> if frame was flushed.
    */
   boolean isFlushed()
   {
      return flushed;
   }
   
   /**
    * Displays a list of fields from the frame on the screen.  As part of
    * this operation, the list of data values are copied to the screen
    * buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY UNLESS-HIDDEN data [IN WINDOW].
    *
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    * @param    windowClause
    *           <code>true</code> if windowClause is presented in DISPLAY stmt.
    *           <code>false</code> otherwise
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   private void displayUnlessHidden(FrameElement[] data, boolean windowClause, handle hWin)
   {
      if (frame.getAttr("hidden", () -> frame.config.hidden, true))
      {
         sendWidgetScreenValue(data);
         
         return;
      }
      
      FrameElement[] dataChanged = removeHiddenWidgets(data); 

      if (dataChanged != null && dataChanged.length > 0)
      {
         displayWorker(dataChanged, windowClause, hWin);
      }
   }

   /**
    * Displays a list of fields from the frame on the screen.  As part of
    * this operation, the list of data values are copied to the screen
    * buffer.
    * <p>
    * As part of the processing, the frame and each listed widget will be
    * marked as visible.  The frame will be brought to the top of the Z-order
    * and refreshed.
    * <p>
    * Any <code>null</code> entry in the array is silently ignored.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY data [IN WINDOW].
    *
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    * @param    windowClause
    *           <code>true</code> if windowClause is presented in DISPLAY stmt.
    *           <code>false</code> otherwise
    * @param    hWin
    *           A handle to the window where the frame is displayed.
    */
   private void displayWorker(FrameElement[] data, boolean windowClause, handle hWin)
   {
      // ignore call if there are no widgets defined in frame
      if (data != null        && 
          data.length == 0    && 
          (widgets == null || widgets.length == 0) &&
          (headers == null || headers.length == 0))
      {
         return;
      }

      if (!isRedirected() && frame.getAttr("scroll", () -> frame.config.scroll) > 0)
      {
         if (pendingUpDown != 0)
         {
            //TODO SVL: Only SCROLL = 1 is supported, we need support an arbitrary SCROLL number.
            // For the testcase and discussion on SCROLL, see https://proj.goldencode.com/issues/5474#note-7
            // and further.
            Integer currentRow = getFGRow(currentFieldGroup.asWidgetHandle());
            if (currentRow != null)
            {
               if (pendingUpDown < 0 && currentRow - 1 <= 0)
               {
                  down();
                  scroll(false, false);
               }
               else if (pendingUpDown > 0 && currentRow + 1 > dynamicDown)
               {
                  up();
                  scroll(false, true);
               }
            }
         }
      }

      conditionalUpDown();
      
      Stream out = UnnamedStreams.output(); 
      
      if (out != null)
      {
         if (windowClause)
         {
            display(out, data, hWin);
         }
         else
         {
            display(out, data);
         }
      }
      else
      {
         // set the value in the screen buffer for each widget using the widget
         // id as the index
         copyToScreenBuffer(data);
      
         // now make the frame visible, bring it to the top and refresh the
         // contents
         if (windowClause)
         {
            view(data, hWin);
         }
         else
         {
            view(data);
         }
      }
      
      // handle aggregation if needed
      processAggregates(data, true);
   }
   
   /**
    * Create an array of all widgets in this frame that are NOT in the 
    * given array.
    *
    * @param    xcpt
    *           The widgets to exclude from the resulting list.
    *
    * @return   The list of remaining ids or <code>null</code> if all ids
    *           were excluded.
    */
   private int[] remainingIds(GenericWidget<?>[] xcpt)
   {
      return remainingIds(widgets, xcpt);
   }
   
   /**
    * Create an array of given widgets in this frame that are NOT in the 
    * given array.
    *
    * @param    all
    *           The widgets to be processed.
    * @param    xcpt
    *           The widgets to exclude from the resulting list.
    *
    * @return   The list of remaining ids or <code>null</code> if all ids
    *           were excluded.
    */
   private int[] remainingIds(GenericWidget<?>[] all, GenericWidget<?>[] xcpt)
   {
      Map<Integer, GenericWidget<?>> widgetIds = new LinkedHashMap<>();
      
      // inspect widget array for IDs
      for (int j = 0; j < all.length; j++)
      {
         widgetIds.put(all[j].getId(), all[j]);
      }
      
      // remove excluded widget IDs
      for (int i = 0; i < xcpt.length; i ++)
      {
         // null values are silently ignored
         if (xcpt[i] == null)
            continue;
         
         int id = getWidgetId(xcpt[i]);
         
         // invalid widget
         if (id == -1)
            continue;
         
         widgetIds.remove(id);
      }
      
      return idSetToArray(widgetIds.keySet());
   }

   /**
    * Converts the given set of <code>Integer</code> values into an array
    * of primitive <code>int</code>.
    *
    * @param    widgetIds
    *           The set of ids to convert.
    *
    * @return   The <code>int[]</code> or <code>null</code> if there are no
    *           ids in the set.
    */
   private int[] idSetToArray(Set<Integer> widgetIds)
   {  
      // calculate the widgets
      int size = widgetIds.size();
      
      if (size == 0)
      {
         return null;
      }

      int[] ids = new int[size];
      Integer[] wraps = widgetIds.toArray(new Integer[0]);

      for (int i = 0; i < size; i ++)
      {
         ids[i] = wraps[i].intValue();
      }
      
      return ids;
   }
   
   /**
    * Converts a widget array into an array of widget IDs.
    *
    * @param    widgets
    *           array of widgets which is to be edited.
    *
    * @return   The array of widget IDs.
    */
   private int[] idsFromWidgets(GenericWidget<?>[] widgets)
   {
      if (widgets == null)
         return null;

      int[] ids = new int[widgets.length];

      for (int i = 0; i < widgets.length; i ++)
      {
         ids[i] = getWidgetId(widgets[i]);
      }
      
      return ids;
   }

   /**
    * Converts a frame element array into a array of widgets.  Any nodes that
    * are assignment only result in a <code>null</code> entry in the array.
    *
    * @param    widgets
    *           The list of widgets which is to be edited.  The data portion
    *           of these elements is not used at all.
    *
    * @return   The array of widgets.
    */
   private GenericWidget<?>[] widgetListFromFrameElements(FrameElement[] widgets)
   {
      GenericWidget<?>   current = null;
      GenericWidget<?>[] wids    = new GenericWidget[widgets.length];
      
      for (int i = 0; i < widgets.length; i++)
      {
         if (widgets[i] == null || widgets[i].assignOnly() || widgets[i].bypass())
         {
            current = null;
         }
         else
         {
            current = widgets[i].getWidget();
         }
         
         wids[i] = current;
      }
      
      return wids;
   }
   
   /**
    * Raise an error condition with the given error number and text IF
    * unnamed input has been redirected.  If silent error mode is enabled,
    * any error condition will be logged and this method will return.
    *
    * @param    num
    *           Error number.
    * @param    txt
    *           Error text.
    *
    * @return   <code>true</code> if in silent error mode and an error was
    *           raise.  <code>false</code> if the unnamed input is not
    *           redirected.
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   private boolean errorOnRedirectedInput(int num, String txt)
   {
      if (UnnamedStreams.input() != null)
      {
         // current-window is used (not the frame's window)
         ErrorManager.recordOrThrowError(num, txt);
         
         return true;
      }
      
      return false;
   }
   
   /**
    * Raise an error condition with the given error number and text IF
    * unnamed input has been redirected.  If silent error mode is enabled,
    * any error condition will be logged and this method will return.
    *
    * @param    num
    *           Error number.
    * @param    txt
    *           Error text.
    *
    * @return   <code>true</code> if in silent error mode and an error was
    *           raise.  <code>false</code> if the unnamed input is not
    *           redirected.
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   private boolean errorOnBatchModeEnable(int num, String txt)
   {
      // It depends on underlying OS and|or background mode.
      if (LogicalTerminal.isInBatchMode() &&
          !PlatformHelper.isUnderWindowsFamily())
      {
         // no window is required

         LogicalTerminal.activateBatchMode(false);
         // We have to display message and stop the client execution
         if (LogicalTerminal.isBatchInBackground())
         {
            ErrorManager.displayErrorRedirected(num, txt, true);
         }
         else // Foreground mode shows another message
         {
            ErrorManager.displayErrorRedirected(513, NO_INPUT_SOURCE_ATTEMPT, true);
         }
         LogicalTerminal.stopBatchSession();
         return true;
      }
      
      return false;
   }

   /**
    * Raise an error condition with the given error number and text IF
    * unnamed input has been redirected.  If silent error mode is enabled,
    * any error condition will be logged and this method will return.
    *
    * @param    num
    *           Error number.
    * @param    txt
    *           Error text.
    *
    * @return   <code>true</code> if in silent error mode and an error was
    *           raise.  <code>false</code> if the unnamed input is not
    *           redirected.
    *
    * @throws   ErrorConditionException
    *           If the unnamed input is redirected and silent error mode is
    *           off.
    */
   private boolean errorOnBatchModeSetPrompt(int num, String txt)
   {
      // It depends on internal I/O redirection.
      if (LogicalTerminal.isInBatchMode() && UnnamedStreams.input() == null)
      {
         // no window is required

         LogicalTerminal.activateBatchMode(false);
         // It depends on underlying OS and|or background mode.
         if (PlatformHelper.isUnderWindowsFamily() || !LogicalTerminal.isBatchInBackground())
         {
            // We have to display message before stop the client execution
            ErrorManager.displayError(num, txt, true);
         }
         LogicalTerminal.stopBatchSession();
         return true;
      }
      
      return false;
   }
   
   /**
    * Executes delayed conditional cursor movement.
    */
   private void conditionalUpDown()
   {
      conditionalUpDown(false);
   }
   
   /**
    * Executes delayed conditional cursor movement.
    * 
    * @param    force
    *           If the pending up/down is processed immediately.
    */
   private void conditionalUpDown(boolean force)
   {
      if (pendingUpDown != 0)
      {
         if (force)
         {
            int down = pendingUpDown;
            pendingUpDown = 0;
            downWorker(down, false);
            return;
         }
         
         int frameLine = frameLineWorker();
         
         LogicalTerminal.setConditionalUpDown(pendingUpDown, frameId);
         
         // frameLine can be out of the frame borders and therefore it
         // can't be used as a corresponding parameter for updateFieldGroups method.
         // we need the real row associated with the current field-group.
         Integer currentRow = getFGRow(currentFieldGroup.asWidgetHandle());
         if (currentRow == null)
         {
            throw new IllegalStateException(
                     "Current field-group is not assigned to a row in this frame");
         }
         
         updateFieldGroups(pendingUpDown, currentRow, dynamicDown, Math.abs(pendingUpDown));
         
         pendingUpDown = 0;
         
         if (frameLine < 0 || frameLine > dynamicDown)
         {
            // if after the conditional up/down we land outside the frame,
            // the cache must be reset
            screenBufferCache.clear();
         }
      }
      
      afterDown  = false;
      afterUnderline = false;
   }
   
   /**
    * DOWN worker routine for the unnamed stream, redirected or not.
    * 
    * @param    lines
    *           Number of lines to scroll down.
    *
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    */
   private void downWorker(int lines, boolean cond)
   {
      if (defaultFrame && widgets.length == 0)
      {
         // for default frame with no widgets this is a no-op
         return;
      }

      if (cond)
         lines = 1;

      if (aggregateHandler != null)
            aggregateHandler.notifyDown((cond) ? 0: lines);
      
      if (lines < 0 && isRedirected())
      {
         // this is a NOP
         return;
      }
      
      // for unnamed streams, we must end any PUT or EXPORT output
      int sid = (isRedirected() && !cond)
                ? getRemoteStreamId(UnnamedStreams.output())
                : -1;

      // save the current FRAME-LINE value to be used with generic down. 
      // It can be incorrect like at can in conditionalUpDown() method 
      int savedFrameLine = frameLineWorker();
      
      if (pendingUpDown > 0 && afterUnderline)
         pendingUpDown--;
      
      int tmp = pendingUpDown;
      pendingUpDown = 0;

      LogicalTerminal.down(this, lines + tmp, cond, sid);
      afterDown = true;
      afterUnderline = false;
      
      if (isRedirected())
      {
         // redirected streams don't use the cache algorithm
         if (lines > 0) 
         {
            // reset the screen buffer if it is a down instruction
            frameBuf = new ScreenBuffer(frameId, validIds(widgets));
         }
         return;
      }
      
      if (dynamicDown == 0) 
      {
         // action that forces p2j client to calculate frame down 
         LogicalTerminal.view(frameId);
      }
      
      // save down, retain and scroll values for this frame, to be used later
      FrameConfig frameConfig = frame.config;
      
      // avoid going to the client with frameDown() since it is expensive
      int down   = dynamicDown; 
      int retain = frame.getAttr("retain", () -> frameConfig.retain);
      int scroll = frame.getAttr("scroll", () -> frameConfig.scroll);

      // check if both retain and scroll are zero
      boolean noRetainOrScroll = retain == 0 && scroll == 0;
      
      // check if the up/down is within the frame row-boundaries
      int     newFrameLine  = savedFrameLine + lines;
      boolean defaultUpDown = newFrameLine >= 1 && newFrameLine <= down;

      if (defaultUpDown)
      {
         // handle a generic up/down
         // cache the current buffer in the frame line as saved
         // before the up/down
         addScreenBufferToCache(this.frameBuf, savedFrameLine);

         // show the buffer associated to the frame line row
         // also make visible this screen buffer to the screen
         showScreenBuffer(frameLineWorker());                     
      }
      else
      {
         newFrameLine = lines < 0 ? down : 1;
         if (noRetainOrScroll)
         {
            // clear all row buffers
            this.screenBufferCache.clear();
            // cache the buffer for the current line
            showScreenBuffer(frameLineWorker());
            // all frame is reset, no additional display is needed
         }
         else 
         {    
            // data about the first part
            int start1 = 0;
            int end1   = 0;
            
            // data about second part
            int incr2  = 0;
            int start2 = 0;
            int end2   = 0;

            // data about third part
            int dir3  = 0;
            int end3  = 0;
            
            if (lines < 0)
            {
               // an up with retain or scroll
               start1 = 1;

               incr2  = -1;
               start2 = down;
               end2   = 1;

               if (retain != 0)
                  end1 = retain;
              
               if (scroll != 0)
                  end1 = down - scroll;
            }
            else
            {
               // a down with retain or scroll
               end1 = down;
               
               incr2  = 1;
               start2 = 1;
               end2   = down;

               if (retain != 0)
                  start1 = down - retain + 1;
               
               if (scroll != 0)
                  start1 = scroll + 1;
            }
            
            if (retain > 0)
            {
               dir3  = -1; 
               end3  = down - retain;
            }
            if (scroll > 0)
            {
               dir3  = 1; 
               end3  = scroll;
            }

            // save old cache
            Map<Integer, ScreenBuffer> oldCache = screenBufferCache;

            // new cache in which to store the new rows
            screenBufferCache = new HashMap<>();
            
            // first part
            
            for (int row = start1; row <= end1; row += 1)
            {
               int newKey = 0;
               
               // up
               if (lines < 0)
               {
                  if (retain != 0)
                     newKey = down - retain + row;
                  if (scroll != 0)
                     newKey = down - row + 1;
               }
               
               // down
               if (lines > 0)
               {
                  if (retain != 0)
                     newKey = row - (down - retain);
                  if (scroll != 0)
                     newKey = row - scroll;
               }
               
               ScreenBuffer oldBuf = oldCache.get(row);
               addScreenBufferToCache(oldBuf, newKey);
            }
            
            if (lines < 0)
            {
               addScreenBufferToCache(frameBuf, frameLineWorker());
            }
               
            // second part - to make the new rows visible, iterate 
            // through the cached rows
            for (int row = start2; 
                (incr2 > 0 && row <= end2) || (incr2 < 0 && row >= end2); 
                row += incr2)
            {
               showScreenBuffer(row);

               if (row != end2)
                  LogicalTerminal.down(this, incr2, cond);
            }
            
            // third part - position the cursor to the correct row
            for (int i = 1; i < end3; i++)
            {
               LogicalTerminal.down(this, dir3, cond);
            }                  

            // make the current screen buffer visible
            showScreenBuffer(frameLineWorker());
         }
      }

      // correct if it really needed
      savedFrameLine = Math.min(savedFrameLine, dynamicDown);
      savedFrameLine = Math.max(1, savedFrameLine);

      int realstep;
      if (!defaultUpDown)
      {
         realstep = (lines + tmp) > 0 ? down + 1 - savedFrameLine : savedFrameLine;
      }
      else
      {
         realstep = lines + tmp;
      }
      realstep = Math.abs(realstep) % (down + 1);
      updateFieldGroups(lines, savedFrameLine, down, realstep);
   }

   /**
    * This method will send the screen buffer for row at given index
    * to be displayed by the client, only if no redirected output occurred.
    * <p>
    * All widgets state must be set to changed so they will be updated. 
    * 
    * @param    idx
    *           The index of the line to be shown to the screen by the 
    *           client. The client MUST be positioned to the right row.
    */
   private void showScreenBuffer(int idx) 
   {
      if (isOutRedirected())
      {
         return;
      }

      ScreenBuffer buff = getScreenBufferFromCache(idx);
      this.frameBuf = buff;

      ConfigHelper.setModified(frame.config, true);
      
      // markAllChanged() will not return null, this is intentional since LT.view() treats that as "all
      // widgets" instead of "no widgets"; so if there are no widgets the result will be a 0 sized array
      LogicalTerminal.view(this, buff.markAllChanged());
   }

   /**
    * DOWN worker routine for named streams.
    * <p>
    * For a named stream, UP statements have no effect on the FRAME-LINE 
    * index or on the screen buffer. DOWN statements will advance to the 
    * next row (even past the FRAME-DOWN row) and will reset the screen
    * buffer.
    * 
    * @param    lines
    *           Number of lines to scroll down.
    * @param    id
    *           The stream ID to which the terminal will be redirected during
    *           this processing or -1 if no temporary redirection is needed.
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    * @param    endOut                
    *           <code>true</code> to force any pending PUT or EXPORT output
    *           to be ended for the given stream (so long as that stream is
    *           not the terminal).
    */
   private void downWorker(int lines, int id, boolean cond, boolean endOut)
   {
      if (defaultFrame && widgets.length == 0)
      {
         // for default frame with no widgets this is a no-op
         return;
      }

      if (cond)
         lines = 1;
      
      if (aggregateHandler != null)
            aggregateHandler.notifyDown((cond) ? 0 : lines);

      if (lines < 0 && id != -1)
      {
         // this is a NOP
         return;
      }

      int tmp = pendingUpDown;
      pendingUpDown = 0;

      if (id == -1)
      {
         // just for safety
         LogicalTerminal.down(this, lines + tmp, cond);
      }
      else
      {
         // Same data for every iteration - bypass down on client side
         if (!sameViewDataRedirected)
            LogicalTerminal.down(this, (lines + tmp), id, cond, endOut);
         
         // For non-down report related frames we have to preserve current
         // screen buffer
         if (!frame.isDown() && isOutRedirected() && sameViewDataRedirected)
            return;

         if (lines > 0) 
         {
            // reset the screen buffer if it is a down instruction
            frameBuf = new ScreenBuffer(frameId, validIds(widgets));
         }
         
         // the UP instruction has no effect on the FRAME-LINE index
         // or on the buffer
      }
   }

   /**
    * UP worker routine.
    * 
    * @param    lines
    *           Number of lines to scroll up.
    */
   private void up(int lines)
   {
      downWorker(-lines, false);
   }
   
   /**
    * UP worker routine.
    * 
    * @param    lines
    *           Number of lines to scroll up.
    * @param    id
    *           The stream ID to which the terminal will be redirected during
    *           this processing or -1 if no temporary redirection is needed.
    */
   private void up(int lines, int id)
   {
      downWorker(-lines, id, false, false);
   }
   
   /**
    * Dereferences the given stream to find and return the remote stream ID.
    *
    * @param    out
    *           The stream (may be a <code>StreamWrapper</code> that has
    *           had a <code>RemoteStream</code> assigned OR it may be a
    *           <code>RemoteStream</code> instance).  If <code>null</code>
    *           or any wrapped reference is <code>null</code>, then the
    *           result will be treated as referencing the terminal and a
    *           -1 will be returned.
    *
    * @return   The remote stream ID.
    */
   int getRemoteStreamId(Stream out)
   {
      RemoteStream rs = null;
      
      if (out instanceof RemoteStream)
      {
         rs = (RemoteStream) out;
      }
      else if (out instanceof StreamWrapper)
      {
         StreamWrapper wrap = (StreamWrapper) out; 
         rs = (RemoteStream) wrap.dereference();
      }
      
      if (rs == null)
         return -1;

      // terminal stream with page size 0 is a regular interactive terminal
      // which does not require special handling
      if (rs.isTerm() && rs.getPageSize().intValue() == 0)
         return -1;
      
      return rs.getId();
   }
   
   /**
    * Adds a screen buffer instance to the cache for the specified row.
    * 
    * @param   buff
    *          The buffer to be added to the cache.
    * @param   row
    *          The 1-based row to which this buffer corresponds to.
    */
   private void addScreenBufferToCache(ScreenBuffer buff, int row) 
   {
      this.screenBufferCache.put(row, buff);
   }

   /**
    * Gets the screen buffer associated for this row. 
    * If there is no screen buffer associated with this row, then the method
    * will return a new empty buffer.
    * 
    * @param   row
    *          The 1-based row for which to get the current screen buffer.
    *          
    * @return  The buffer associated with this row.
    */
   private ScreenBuffer getScreenBufferFromCache(int row) 
   {
      ScreenBuffer buff = null;
      
      if (!this.screenBufferCache.containsKey(row)) 
      {
         buff = new ScreenBuffer(frameId, validIds(widgets));

         // all static text widgets need to be copied from the previous buffer
         for (int i = 0; i < widgets.length; i++)
         {
            int wid = widgets[i].getId();
            GenericWidget<?> widget = this.getWidgetForId(wid);
            
            if (widget instanceof ControlTextWidget)
            {
               // if this is a static widget
               ControlTextWidget<?> text = (ControlTextWidget<?>) widget;
               ControlTextConfig config = text.config();
               if (text.getAttr("flagStatic", () -> config.flagStatic))
               {
                  buff.putWidgetValue(wid, frameBuf.getWidgetValue(wid));
               }
            }
         }
         
         addScreenBufferToCache(buff, row);
      }
      else 
      {
         buff = this.screenBufferCache.get(row);
         buff.setBufferChanged(true);
      }
      
      return buff;
   }
   
   /**
    * Removes the widgets with hidden attribute value set as TRUE.
    * 
    * @param    inputList
    *           The widgets to process.
    *
    * @return   The array of the GenericWidget&lt;?&gt; values with HIDDEN attrubute is FALSE.
    */
   private GenericWidget<?>[] removeHiddenWidgets(GenericWidget<?>[] inputList)
   {
      List<GenericWidget<?>> tmpList = new ArrayList<>();
      
      for (int i = 0; i < inputList.length; i++)
      {
         if (!inputList[i].isHidden().booleanValue())
         {
            tmpList.add(inputList[i]);
         }
      }
      
      return tmpList.toArray(new GenericWidget[tmpList.size()]);
   }
   
   /**
    * Removes the widgets with hidden attribute value set as TRUE.
    * 
    * @param    inputList
    *           The widgets to process.
    *
    * @return   The array of the GenericWidget values with HIDDEN attribute is FALSE.
    */
   public static FrameElement[] removeHiddenWidgets(FrameElement[] inputList)
   {
      List<FrameElement> tmpList = new ArrayList<>();
      
      for (int i = 0; i < inputList.length; i++)
      {
         if (inputList[i] != null && !inputList[i].getWidget().isHidden().booleanValue())
         {
            tmpList.add(inputList[i]);
         }
      }
      
      return tmpList.toArray(new FrameElement[tmpList.size()]);
   }
   
   /**
    * Restores the widget registry for static content of this frame based on header and
    * non-header widget info.
    */
   private void restoreWidgetRegistry()
   {
      // non-header widgets first
      for (int i = 0; i < widgets.length; i++)
      {
         if (!widgets[i]._dynamic())
         {
            LogicalTerminal.registerWidget(widgets[i], widgets[i].getId());
         }
      }
      
      // enumerate all the header widgets in the order of their definition
      Iterator<String> iter = n2h.keySet().iterator();

      while (iter.hasNext())
      {
         // get the widget
         GenericWidget<?> gw = n2h.get(iter.next());
            
         if (!gw._dynamic())
         {
            LogicalTerminal.registerWidget(gw, gw.getId());
         }
      }
   }
   
   /**
    * Provides a cleanup mechanism for editing blocks that exits abnormally
    * (without a natural exit via {@link #continueEditing}).
    */
   private class EditingCleaner
   implements ConditionListener
   {
      /** List of widgets that need cleanup. */
      private FrameElement[] list = null;
      
      /** Track clean exit. */
      private boolean condition = false;
      
      /**
       * Create an instance, storing key data for cleanup purposes.
       * 
       * @param   list
       *          Array of frame elements. 
       */
      EditingCleaner(FrameElement[] list)
      {
         this.list = list;
         TransactionManager.registerFinalizable(this, false);
      }
      
      /**
       * Uses the finish notification provided by the {@link
       * TransactionManager} to ensure that any abnormal exit is detected
       * and cleanup occurs.
       */
      @Override
      public void finished()
      {
         boolean clean = false;
         
         synchronized (editLock)
         {
            if (currentOp != NONE)
            {
               clean = true;
            }
         }
         
         if (clean)
         {
            stopEditingMode(list, !condition, true);
         }
      }
      
      /**
       * Processes the deleted notification provided by the {@link ProcedureManager}.  
       * This implementation does nothing.
       */
      @Override
      public void deleted()
      {
         // no-op
      }
      
      /**
       * Processes the iterate notification provided by the {@link
       * TransactionManager}.  This implementation does nothing.
       */
      @Override
      public void iterate()
      {
         // no-op
      }      
      
      /**
       * Processes the retry notification provided by the {@link
       * TransactionManager}.  This implementation does nothing.
       */
      @Override
      public void retry()
      {
         // no-op
      }
      
      /**
       * Track all conditions which may break the editing mode.
       * 
       * @param   condition
       *          Unused.
       */
      @Override
      public void condition(Condition condition)
      {
         this.condition = true; 
      }            
   }

   /**
    * Manages <code>AggregatorElements</code> associated with the enclosing
    * frame instance.  An array of such elements is provided at construction
    * and is stored.  The constructor registers a new instance with the
    * <code>TransactionManager</code> at the nearest enclosing loop.  Each
    * invocation of the enclosing frame's <code>display</code> method causes
    * all associated accumulators to accumulate the data associated with that
    * pass of the loop.  At the bottom of each iteration of the loop, this
    * object checks whether any of its managed accumulators is currently at
    * the last row in a break group.  If so, sub-results are displayed.  At
    * the end of loop, final results of all accumulators which support such a
    * result are displayed.
    */
   private class AggregateHandler
   implements Finalizable
   {
      /** All aggregator elements handled by this object */
      private LinkedHashSet<AggregatorElement> accums = new LinkedHashSet<>();
      
      /** Elements which need to be displayed on finished processing */
      private List<AggregatorElement> fulls = new ArrayList<>();
      
      /** Elements which need to be displayed on next iteration */
      private List<AggregatorElement> subs = new ArrayList<>();
      
      /** Comparator used to sort AggregatorElements */
      private Comparator<AggregatorElement> comparator = new AggregatorElementComparator();

      /** Number of lines in last DOWN call. */
      private int lastDown = 0;
      
      /**
       * If <code>true</code> then last operation was display() rather than
       * accumulate().
       */
      private boolean displayed = false; 
      
      /**
       * Construct a new instance, detecting which of the given aggregator
       * elements must produce a "final" result line at loop termination.
       * This objects registers itself with the transaction manager at the
       * nearest enclosing loop block.
       *
       * @param   accums
       *          All aggregator elements to be managed by this handler.
       *
       * @throws  IllegalStateException
       *          if no enclosing, looping block can be identified to which
       *          to register this <code>Finalizable</code>.
       */
      AggregateHandler(AggregatorElement[] accums)
      {
         verifyAccumulators(accums);
      }
      
      /**
       * Frame calls this method to notify handler about last down() call.
       * 
       * @param   lines
       *          Number of lines passed to down().
       */
      public void notifyDown(int lines)
      {
         lastDown  = lines;
      }
      
      /**
       * Compare provided array of accumulator elements with set of known ones
       * and add ones which are not yet known.
       * 
       * @param   elem
       *          Array of {@link AggregatorElement} instances to verify.
       */
      public void verifyAccumulators(AggregatorElement[] elem)
      {
         for (AggregatorElement next : elem)
         {
            if (accums.contains(next))
               continue;
            accums.add(next);
            
            if (!next.getAccumulator().isSubgroupOnly())
            {
               fulls.add(next);
            }
         }
         
         Collections.sort(fulls, comparator);
      }

      /**
       * Get the current value for all data being accumulated and cause each
       * registered accumulator to accumulate these data.  Detect and sort
       * the subset of aggregators (if any) which need to display their break
       * group results at the bottom of the current loop iteration.
       * <p>
       * If this handler is not managing any aggregators, this method returns
       * immediately.
       * 
       * @param   fe
       *          Array of frame elements.  
       */
      void accumulate(FrameElement[] fe)
      {
         HashMap<FrameElement, FrameElement> feMap = new HashMap<>();
         for (int i = 0; i < fe.length; i ++)
         {
            feMap.put(fe[i], fe[i]);
         }

         for (AggregatorElement next : accums)
         {
            Accumulator accum  = next.getAccumulator();
                        
            if (feMap.containsKey(next))
            {
               FrameElement frameElem = feMap.get(next);
               BaseDataType datum = frameElem.get();
               accum.accumulate(datum);
            }
         }
         
         displayed = false;
      }
      
      /**
       * Display break group results (if any).  This method is invoked by the
       * <code>TransactionManager</code> at the top of the second and 
       * subsequent loop iteration.  This means that the handler has had a
       * chance to accumulate the current row's results.
       */
      @Override
      public void iterate()
      {
         subs.clear();
         
         for (AggregatorElement next : accums)
         {
            Accumulator accum  = next.getAccumulator();
                        
            if (accum.isLastOfGroup())
            {
               subs.add(next);
            }
         }
         
         if (!subs.isEmpty())
         {
            Collections.sort(subs, comparator);

            display(subs);
         }
         else
         {
            displayed = false;            
         }
      }
      
      /**
       * Processes the retry notification provided by the {@link
       * TransactionManager}.  This implementation does nothing.
       */
      @Override
      public void retry()
      {
         // no-op
      }      
      
      /**
       * Processes the deleted notification provided by the {@link ProcedureManager}.  
       * This implementation does nothing.
       */
      @Override
      public void deleted()
      {
         // no-op
      }
      
      /**
       * Display full results (if any).  This method is invoked by the
       * <code>TransactionManager</code> after the display loop terminates,
       * after the handler has had a chance to accumulate the final row's
       * results.
       * <p>
       * This method resets the enclosing frame's aggregate handler reference.
       */
      @Override
      public void finished()
      {
         try
         {
            // force any pending subs out first (since iterate() is not
            // called at the bottom of the block)

            if (!fulls.isEmpty())
               display(fulls);
            else
               down();
         }
         finally
         {
            // reset aggregate handler
            aggregateHandler = null;
         }
      }
      
      /**
       * Convert list of {@link AggregatorElement} instances into array of
       * widget IDs.
       * 
       * @param   accums
       *          List of elements.
       * @param   handleFulls
       *          Flag which indicates processing of final results.
       * 
       * @return  The array of widget IDs.
       */
      private int[] getIDs(List<AggregatorElement> accums, boolean handleFulls)
      {
         LinkedHashSet<Integer> ids = new LinkedHashSet<>(5);
         
         for (Element element : accums)
         {
            if (handleFulls && element.getAccumulator().isSubOnly())
               continue;
            
            ids.add(element.getWidget().getId());
         }

         return Utils.integerCollectionToPrimitive(ids);
      }
      
      /**
       * Display aggregation results, either for a single break group, or for
       * the entire set of rows displayed by a display loop.
       *
       * @param   accums
       *          List of aggregator elements to display.
       */
      private void display(List<AggregatorElement> accums)
      {
         boolean redirect = isOutRedirected();
         
         boolean handleFulls = accums == fulls;
         
         // check for accumulators which should 
         // produce output for entire data set
         if (handleFulls)
         {
            boolean produceOut = false;
            
            for (AggregatorElement elem : accums)
            {
               if (elem.getAccumulator().isSubOnly())
                  continue;
               
               produceOut = true;
               break;
            }
            
            if (!produceOut)
               return;
         }
         
         int tempOut = -1;
         
         if (redirect)
         {
            // temporarily switch the terminal to a new output destination
            int id = getRemoteStreamId(redirOut); 
            tempOut = LogicalTerminal.switchOutput(id, true, true);
         }
         
         try
         {
            if (displayed)
            {
               up();
               up();
            }
               
            if (lastDown > 0)
               up();
            
            // underline the columns being accumulated
            int len = accums.size();
            Iterator<AggregatorElement> iter;
            int[] ids = getIDs(accums, handleFulls);
            
            if (!hadView)
            {
               List<AggregatorElement> wset = new ArrayList<>(ids.length);
               Set<Integer> idSet = new HashSet<>(ids.length);
               for (AggregatorElement elem : accums)
               {
                  if (idSet.contains(elem.getWidget().getId()) == false)
                  {
                     wset.add(elem);
                     idSet.add(elem.getWidget().getId());
                  }
               }
               view(wset.toArray(new AggregatorElement[0]));
               up();
            }
            underline(ids);
            
            ScreenBuffer sb = new ScreenBuffer(frameId, validIds(widgets));
            
            // process all columns with accumulation
            
            iter = accums.iterator();
            
            for (int i = 0; iter.hasNext(); i++)
            {
               AggregatorElement next = iter.next();
               GenericWidget<?> wid = next.getWidget();
               Accumulator accum = next.getAccumulator();
               
               // skip accumulators which produce only subtotals
               if (handleFulls && accum.isSubOnly())
                  continue;
               
               String fmt   = next.hasFormat() ? next.getFormat() : null;
               Object value = (accums == subs) ? accum.getSubResult()
                                               : accum.getFullResult();
               
               if (fmt != null)
               {
                  value = new FormattedValue((BaseDataType) value, fmt);
               }
               
               int[] widgetIds = new int[] { statsColumn.getId(), wid.getId()};
               
               // add the statistic to the screen buffer for the correct 
               // widget
               sb.putWidgetValue(wid.getId(), value);
               
               int weight = accum.getWeight();
               int nweight = (i == len - 1)
                             ? Integer.MAX_VALUE
                             : accums.get(i + 1).getAccumulator().getWeight();
               
               // this is the last set of outputs or a new set will begin on
               // the next iteration
               
               if (weight != nweight)
               {
                  // add the statistic label to the to the extra stats widget
                  character label = new character(accum.getLabel());
                  sb.putWidgetValue(statsColumn.getId(), label);
                  
                  // output the stats line
                  conditionalUpDown();
                  
                  // the following call is qualified as DISPLAY in LogicalTerm
                  // note that we display exactly two widgets to avoid 
                  // triggering of repainting of all widgets in the frame
                  LogicalTerminal.view(GenericFrame.this, new ScreenBuffer[] {sb}, widgetIds);
                  
                  // force a new line
                  
                  if (accums == subs)
                     down(1);
                  else
                     down();
                  
                  // clear the screen buffer
                  sb.reset();
               }
            }
         }
         
         finally
         {
            if (redirect)
            {
               // put the old terminal back (either interactive or a
               // different redirection)
               LogicalTerminal.switchOutput(tempOut, false);
            }
            displayed = true;
         }
      }

      /**
       * A comparator implementation which compares
       * <code>AggregatorElement</code>s using only the weights of their
       * contained <code>Accumulator</code> instances.
       *
       * @see  com.goldencode.p2j.util.Accumulator#getWeight
       */
      private class AggregatorElementComparator
      implements Comparator<AggregatorElement>
      {
         /**
          * Compares two <code>AggregatorElement</code>s, based upon their
          * contained <code>Accumulator</code>'s weights.  Lower weights sort
          * first.
          *
          * @param   o1
          *          AggregatorElement 1.
          * @param   o2
          *          AggregatorElement 2.
          *
          * @return  Negative value if o1's weight is lower than o2's;
          *          positive value if o2's weight is lower than o1's;
          *          0 if they have equal weights.
          */
         @Override
         public int compare(AggregatorElement o1, AggregatorElement o2)
         {
            int weight1 = o1.getAccumulator().getWeight();
            int weight2 = o2.getAccumulator().getWeight();
            
            return (weight1 - weight2);
         }
      }
   }
   
   /**
    * Provide description for GenericFrame instance.
    * 
    * @see  java.lang.Object#toString()
    */
   @Override
   public String toString()
   {
      return frameName + "#" + frameId;
   }

   /**
    * Support for the DISPLAY followed by the DOWN combination.
    */
   @Override
   public void displayAndDown()
   {
      displayAndDown(new FrameElement[0]);
   }

   /**
    * Support for the DISPLAY followed by the DOWN combination.
    * 
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    */
   @Override
   public void displayAndDown(FrameElement[] data)
   {
      if (!isRedirected())
      {
         displayWorker(data, false, null);
         down();
         return;
      }
      displayAndDownWorker(null, data);
   }

   /**
    * Support for the DISPLAY UNLESS HIDDEN followed by the DOWN combination.
    * 
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    */
   @Override
   public void displayUnlessHiddenAndDown(FrameElement[] data)
   {
      if (frame.getAttr("hidden", () -> frame.config.hidden, true))
      {
         sendWidgetScreenValue(data);
         
         return;
      }
      
      FrameElement[] dataChanged = removeHiddenWidgets(data); 

      if (!isRedirected())
      {
         if (dataChanged != null && dataChanged.length > 0)
         {
            displayWorker(dataChanged, false, null);
         }
         down();
         return;
      }
      displayAndDownWorker(null, data);
   }

   /**
    * Support for the DISPLAY UNLESS HIDDEN followed by the DOWN combination.
    * 
    * @see    #displayUnlessHidden()
    */
   @Override
   public void displayUnlessHiddenAndDown()
   {
      if (frame.getAttr("hidden", () -> frame.config.hidden, true))
      {
         return;
      }
      
      view();
      down();
   }

   /**
    * Support for the DISPLAY STREAM ... followed by the DOWN STREAM.. combination.
    * 
    * @param    out
    *           A stream instance which is open for writing. 
    */
   @Override
   public void displayAndDown(Stream out)
   {
      displayAndDown(out, new FrameElement[0]);
   }

   /**
    * Support for the DISPLAY STREAM ... followed by the DOWN STREAM.. combination.
    * 
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    */
   @Override
   public void displayAndDown(Stream out, FrameElement[] data)
   {
      displayAndDownWorker(out, data);
   }

   /**
    * Set the screen-value for each widget in the given list, if is not hidden.
    * 
    * @param    data
    *           The list of data (and associated widgets).
    */
   private void sendWidgetScreenValue(FrameElement[] data)
   {
      if (data == null)
      {
         return;
      }
   
      List<Integer> wids = new ArrayList<>();
      List<BaseDataType> values = new ArrayList<>();
      
      for (FrameElement el : data)
      {
         if(el == null || el.assignOnly() || el.widgetOnly() || el.bypass())
         {
            continue;
         }

         GenericWidget widget = el.getWidget();
         if (!widget.getAttr("hidden", () -> widget.config.hidden, true))
         {
            BaseDataType value = el.get();
            widget.setScreenValue(frameBuf, value, true);
            wids.add(widget.getId());
            values.add(value);
         }
      }
      
      if (!wids.isEmpty())
      {
         int[] awids = Utils.integerCollectionToPrimitive(wids);
         BaseDataType[] avalues = values.toArray(new BaseDataType[values.size()]);
         LogicalTerminal.refreshFrameWidgets(this, awids, avalues);
      }
   }
   
   /**
    * Support for the DISPLAY STREAM ... followed by the DOWN STREAM.. combination.
    * 
    * @param    out
    *           A stream instance which is open for writing. 
    * @param    data
    *           The list of data (and associated widgets) which is to be
    *           displayed.
    */
   private void displayAndDownWorker(Stream out, FrameElement[] data)
   {
      copyToScreenBuffer(data);

      // ignore call if there are no widgets defined in frame
      if (data != null && data.length == 0 && 
          (widgets == null || widgets.length == 0) &&
          (headers == null || headers.length == 0))
      {
         return;
      }

      if (out == null)
      {
         out = UnnamedStreams.output();
      }

      if (data != null && data.length > 1)
      {
         try
         {
            if (out.protectWrites())
            {
               // silent error mode is ON (this is NO-ERROR) so the error
               // output is suppressed and doesn't need to be doubled,
               // just exit
               return;
            }
         }

         catch (ErrorConditionException ece)
         {
            // ignore because we have to do this again
         }
      }

      // normal safety code here, exit in NO-ERROR mode if needed
      if (out.protectWrites())
         return;

      conditionalUpDown();
      pendingUpDown = 0;

      int sid = getRemoteStreamId(out);


      redirOut = out;

      boolean register = false;
      boolean pageTop = false;
      if (isPageElement(out))
      {
         register = true;
         pageTop = getPageTop();
      }

      int[] ids = null;
      if (data != null)
      {
         GenericWidget<?>[] wids = widgetListFromFrameElements(data);
         ids = idsFromWidgets(wids);
      }
      else
      {
         ids = new int[] { frameId };
      }

      LogicalTerminal.displayAndDown(this, ids, sid, register, pageTop);

      processAggregates(data, true);
      if (aggregateHandler != null)
         aggregateHandler.notifyDown(0);

      afterDown = true;
      afterUnderline = false;
      frameBuf = new ScreenBuffer(frameId, validIds(widgets));
   }
   
   /**
    * Gets the integer value of the window containing the widget.
    * 
    * @return   The integer value, which will be the same as the frame ID.
    */
   @Override
   public integer getHWND()
   {
      // we use this with LogicalTerminal.disableRedraw() to implement a full replacement for
      // the WIN32 API call LockWindowUpdate; disableRedraw() will break if the returned HWND
      // cannot be mapped back into the specific frame
      return new integer(frameId);
   }
   
   /**
    * Getting the NEXT-TAB-ITEM attribute.
    * 
    * @return  nothing.
    */
   @Override
   public handle getNextTabItem()
   {
      return asWidget().getNextTabItem();
   }
   
   /**
    * Getting the PREV-TAB-ITEM attribute.
    * 
    * @return  nothing.
    */
   @Override
   public handle getPreviousTabItem()
   {
      return asWidget().getPreviousTabItem();
   }

   /**
    * Sets "parent" frame for this frame.
    * TODO implement client-side logic
    *
    * @param    hFrame
    *           The handle of the frame that will be a "parent" frame. 
    */
   @Override
   public void setFrameHandle(handle hFrame)
   {
      if (!(hFrame.getResource() instanceof GenericFrame))
      {
         throw new IllegalArgumentException("Specified handle does not point to frame widget");
      }
      
      asWidget().setFrame((GenericFrame) hFrame.getResource());
   }

   /**
    * Gets "parent" frame for this frame.
    * TODO implement client-side logic
    *
    * @return   The handle of the parent frame.
    */
   @Override
   public handle getFrameHandle()
   {
      return asWidget().asWidgetHandle();
   }
   
   /**
    * Set the STREAM-IO option of this frame.
    * 
    * @param    value
    *           <code>true</code> if the STREAM-IO option is present.
    */
   @Override
   public void setStreamIO(boolean value)
   {
      frame.config.screenIO = !value;
      if (value)
      {
         frame.config.useText = true;
      }
   }

   /**
    * Set the SCREEN-IO option of this frame.
    * 
    * @param    value
    *           <code>true</code> if the SCREEN-IO option is present.
    */
   @Override
   public void setScreenIO(boolean value)
   {
      frame.config.screenIO = value;
   }

   /**
    * Set the USE-TEXT option of this frame.
    * 
    * @param    value
    *           <code>true</code> if the USE-TEXT option is present.
    */
   @Override
   public void setUseText(boolean value)
   {
      FrameConfig fc = frame.config;
      fc.useText = value;
   }
   
   /**
    * Returns the frame's field group.
    * 
    * @return   See above.
    */
   FieldGroup getFieldGroup()
   {
      return fieldGroup;
   }

   /**
    * Delete all field-groups created for this frame.
    */
   void deleteFieldGroups()
   {
      if (fieldGroup != null)
      {
         fieldGroup.delete();
      }
      
      if (literalFieldGroup != null)
      {
         // delete all widgets from this field-group
         for (GenericWidget<?> wid : literalFieldGroup.getWidgets())
         {
            wid.delete();
         }
         literalFieldGroup.delete();
      }
      
      for (FieldGroup fg : rows2fg.values())
      {
         fg.delete();
      }
      
      if (frame != null && frame.getBackground() != null)
      {
         frame.getBackground().delete();
      }
   }
   
   /**
    * This method prints field-groups of the frame and is needed only for debugging purposes.
    * 
    * Here is a sample output (each triplet group is for a distinct call):
    * <pre>
    * 3
    *  gr1 [null, gr2]     gr2 [gr1, gr3]     gr3 [gr2, gr4]     gr4 [gr3, gr5]     gr5 [gr4, UN]
    * 1 -&gt; gr5, 
    * 
    * 3
    *  gr1 [null, gr2]     gr2 [gr1, gr3]     gr3 [gr2, gr4]     gr4 [gr3, gr5]     gr5 [gr4, UN]
    * 1 -&gt; gr5, 2 -&gt; gr1, 
    * 
    * 3
    *  gr1 [null, gr2]     gr2 [gr1, gr3]     gr3 [gr2, gr4]     gr4 [gr3, gr5]     gr5 [gr4, UN]
    * 1 -&gt; gr5, 2 -&gt; gr1, 3 -&gt; gr2, 
    * </pre>
    * 
    * Here are groups of several triplets of strings.
    * The first line in a triplet is a number of lines to step. Can be &lt; 0.
    * The second row displays field groups, each with its neighbors in square brackets:
    * [left, right].
    * The third row is a mapping from the row number to field-group, only for rows with an
    * associated field-group.
    * 
    * @param   lines
    *          Number of lines to step. Can be &lt;0 (UP statement) or &gt;0 (DOWN statement).
    */
   private void printFieldGroups(int lines)
   {
      System.out.println(lines);
      String result = "";
      for (FieldGroup fg = fieldGroup; 
           fg != null; 
           fg = (FieldGroup) fg.getNextSibling().getResource())
      {
         result += " " + (fg.name().getValue() == null ? "UN" : fg.name().getValue()) + " [";
         if(fg.getPrevSibling() == null || fg.getPrevSibling().isUnknown()) 
            result += null;
         else
         {
            String fgName = ((FieldGroup)fg.getPrevSibling().getResource()).name().getValue();
            result += fgName == null ? "UN" : fgName; 
         }
         result += ", ";
         if(fg.getNextSibling() == null || fg.getNextSibling().isUnknown()) 
            result += null;
         else
         {
            String fgName = ((FieldGroup)fg.getNextSibling().getResource()).name().getValue();
            result += fgName == null ? "UN" : fgName;
         }
         result += "]     ";
         if (fg.getNextSibling() == null || fg.getNextSibling().isUnknown())
            break;
      }
      System.out.println(result);
      
      result = "";
      for(Entry<Integer, FieldGroup> entry : rows2fg.entrySet())
      {
         result += entry.getKey() + " -> " + entry.getValue().name().getValue() + ", ";
      }
      System.out.println(result);
      System.out.println();
   }

   /**
    * This method actualizes field-groups of this frame in correspondance with 
    * the current iteration (DOWN/UP step). For the cases DOWN-SIZE != 1 this method will
    * iterate steps one by one as it would be several DOWN/UP 1 statements.
    * 
    * @param    lines
    *           Number of lines to step. Can be &lt;0 (UP statement) or &gt;0 (DOWN statement).
    * @param    currentRow
    *           The valid current line in this frame (after current step).
    * @param    down
    *           Frame down size.
    * @param    realstep
    *           Current step in the internal iteration for cases DOWN-SIZE != 1.
    *           For example for DOWN 5 values of this parameter will be 5, 4, 3, 2, 1, 0          
    */
   private void updateFieldGroups(int lines, int currentRow, int down, int realstep)
   {
      if (lines == 0 || realstep == 0)
      {
         return;
      }

      // debuggin field-groups
      // printFieldGroups(lines);

      int nextRow;
      if (lines < 0)
      {
         nextRow = currentRow - 1;
         if (nextRow == 0)
         {
            nextRow = down;
         }
      }
      else
      {
         nextRow = currentRow + 1;
         if (nextRow > down)
         {
            nextRow = 1;
         }
      }

      FieldGroup fg = rows2fg.get(nextRow);
      GenericWidget<?>[] widgetsArr = n2w.values().toArray(new GenericWidget[n2w.values().size()]);
      if (fg == null)
      {
         FieldGroup before = literalFieldGroup;
         if (lines < 0                                  && 
             currentFieldGroup.getPrevSibling() != null && 
             !currentFieldGroup.getPrevSibling().isUnknown())
         {
            before = currentFieldGroup;
         }
         
         fg = wrapWidgetsToFieldGroup(frame._dynamic(), widgetsArr, before);
         rows2fg.put(nextRow, fg);
      }
      else
      {
         // update the widgets so now they report this field-group
         // CA: see checkFieldGroup for limitations.
         setFieldGroupWidgets(fg, widgetsArr);
      }
      
      // TODO: OE creates a 'hidden fieldgroup' when going beyond the first or last row...
      
      currentFieldGroup = fg;
      
      updateFieldGroups(lines, nextRow, down, --realstep);
   }

   /**
    * This method checks if the specified field-group is used for some row in this frame.
    * 
    * @param    groupHandle
    *           Handle of specified field-group.
    * 
    * @return   true - if field-group is associated with some row in this frame
    *           false - otherwise.
    */
   private boolean isFGAssigned(handle groupHandle)
   {
      return getFGRow(groupHandle) != null;
   }
   
   /**
    * This method searches the frame row for specified field-group.
    * TODO: use bidirectional map
    * 
    * @param    groupHandle
    *           Handle of specified field-group.
    * 
    * @return   row number - if field-group is associated with some row in this frame
    *           null - otherwise.
    * @throws   IllegalArgumentException
    *           if groupHandle is invalid or is not FieldGroup handle.     
    */
   private Integer getFGRow(handle groupHandle)
   {
      if (groupHandle == null || groupHandle.isUnknown())
      {
         throw new IllegalArgumentException("group handle must not be null");
      }

      if (!(groupHandle.getResource() instanceof FieldGroup))
      {
         throw new IllegalArgumentException("groupHandle must be a FieldGroup handle");
      }
      
      FieldGroup fg = (FieldGroup) groupHandle.getResource();
      for (Entry<Integer, FieldGroup> entry : rows2fg.entrySet())
      {
         if (fg.equals(entry.getValue()))
         {
            return entry.getKey();
         }
      }

      return null;
   }
   
   /**
    * Get all the widgets associated with this frame.
    * 
    * @return   See above.
    */
   private List<GenericWidget<?>> allWidgets()
   {
      List<GenericWidget<?>> allWidgets = new ArrayList<>();
      allWidgets.addAll(n2h.values());
      allWidgets.addAll(n2w.values());
      allWidgets.add(frame);
      
      // sort them by their IDs, as callers of this method expect to process them in the order of
      // their creation (caused by BrowseColumnWidget needing to be processed AFTER the parent
      // BrowseWidget).
      Collections.sort(allWidgets, (w1, w2) -> w1.config.id.compareTo(w2.config.id));
      
      // TODO: this works until the widget IDs start wrapping... another way of keeping the 
      //       widgets in the proper order is needed.
      
      return allWidgets;
   }

   /**
    * Gets the current tab item list.
    * 
    * @return   The current tab item list
    */
   public int[] getTabItemsList()
   {
      if (currentFieldGroup != null)
      {
         return currentFieldGroup.getTabItemList();
      }
      else if (fieldGroup != null)
      {
         return fieldGroup.getTabItemList();
      }
      
      return null;
   }
   
   /**
    * Updates the current tab item list if KEEP-TAB-ORDER is not set.
    * 
    * @param    widgetIds
    *           The widget ids listed in the enable, update, set, prompt-for statements
    * @param    op
    *           The current tab item list rearranging operation
    * 
    * @return   The updated tab item list
    */
   public int[] updateTabItemsList(int[] widgetIds, UpdatesOp op)
   {
      if (frame.getAttr("keepTabOrder", () -> frame.config.keepTabOrder))
      {
         return getTabItemsList();
      }
      
      if (currentFieldGroup != null)
      {
         return currentFieldGroup.updateTabItemList(widgetIds, op);
      }
      else if (fieldGroup != null)
      {
         return fieldGroup.updateTabItemList(widgetIds, op);
      }
      
      return null;
   }
   
   /**
    * Updates the current tab item list if KEEP-TAB-ORDER is not set.
    * 
    * @param    widgets
    *           The widgets listed in the enable, update, set, prompt-for statements
    * @param    op
    *           The current tab item list rearranging operation
    * 
    * @return   The updated tab item list
    */
   public int[] updateTabItemsList(List<GenericWidget<?>> widgets, UpdatesOp op)
   {
      if (frame.getAttr("keepTabOrder", () -> frame.config.keepTabOrder))
      {
         return getTabItemsList();
      }
      
      if (currentFieldGroup != null)
      {
         return currentFieldGroup.updateTabItemList(widgets, op);
      }
      else if (fieldGroup != null)
      {
         return fieldGroup.updateTabItemList(widgets, op);
      }
      
      return null;
   }
   
   /**
    * Moves the target widget to the top of its widgets list if it is in this list.
    * 
    * @param    widget
    *           The target widget
    */
   protected void toTop(GenericWidget<?> widget)
   {
      if (currentFieldGroup != null)
      {
         currentFieldGroup.toTop(widget);
      }
   }
   
   /**
    * Moves the target widget to the bottom of its widgets list if it is in this list.
    * 
    * @param    widget
    *           The target widget
    */
   protected void toBottom(GenericWidget<?> widget)
   {
      if (currentFieldGroup != null)
      {
         currentFieldGroup.toBottom(widget);
      }
   }
   
   /**
    * Creates and initializes the new field group and returns it.
    */
   private void createFieldGroup()
   {
      // TODO this field is initialized in #coreInitialize method. Is this situation legal?
      GenericWidget<?>[] widgetsArr = 
            n2w.values().toArray(new GenericWidget[n2w.values().size()]);
      fieldGroup = wrapWidgetsToFieldGroup(frame._dynamic(), widgetsArr, null);
      fieldGroup.setWidgetAdder(this::addWidget);
   }
   
   /**
    * Utility method used by {@code parseScreenValue()} method to identify the empty screen
    * values. This are NOT unknown values but are displayed as such in a {@code date/time/-tz}
    * fill-in widget.
    *
    * @param   s
    *          The string to pe assigned to SCREEN-VALUE attribute. 
    * @param   ignoredTokens
    *          The ignored tokens specific to this data type. These may occur in any combination
    *          and all their occurrences are ignored.
    *
    * @return  {@code true} when the string to be set as SCREEN-VALUE is equivalent to empty
    *          string.
    */
   private static boolean isEmptyFormat(String s, String ignoredTokens)
   {
      if (s.equals("?"))
      {
         return true; // well known case, note that it is not rtrimmed!
      }
      
      for (int k = 0; k < ignoredTokens.length(); k++)
      {
         s = s.replace(ignoredTokens.charAt(k), ' ');
      }
      
      return s.trim().isEmpty();
   }

   /**
    * Positions the cursor on a new line in a down frame. This is a
    * conditional call. Multiple conditional down() calls have the effect of a
    * single down(1) call, which is delayed till the nearest terminal I/O
    * operation and can be cancelled by a conditional up() call.
    * <p>
    * Note that this call can be ignored in interactive mode if frame is not
    * yet viewed at least once before this call. After first VIEW
    * (explicit or implicit) hiding frame (either via HIDE statement or by
    * changing VISIBLE attribute) does not affect behavior and calls to this
    * method are honored. This corresponds to 4GL behavior.
    * <p>
    * This method corresponds to these Progress statements: DOWN
    */
   private void downInt()
   {
      afterUnderline = false;

      if (aggregateHandler != null)
         aggregateHandler.notifyDown(0);

      flushed  = true;

      // If either the unnamed stream is redirected or the last used stream
      // was something else, this call has to be passed down to the client
      // to switch streams properly.
      // Note that we assume that TERMINAL stream behaves just like interactive
      // mode and does not process DOWN immediately.
      if (hadView && isRedirected() || (isOutRedirected() && !isOutTerm()))
      {
         pendingUpDown = 0;

         if (isRedirected())
         {
            // the unnamed stream is redirected
            downWorker(1,           // down for 1 line
                       true);       // conditional down
         }
         else
         {
            // the last used stream was something else, needs switching
            downWorker(1,           // down for 1 line
                       -1,          // TODO: why not use redirOut, in case of named streams?
                       true,        // conditional down
                       false);      // does not end output (?)
         }

         return;
      }

      if (hadView && pendingUpDown < 1)
      {
         int frameLine =
            LogicalTerminal.getFrameLine(this, pendingUpDown).intValue();

         pendingUpDown++;

         // non-down frames maintain the pending value but don't swap buffers
         if (!frame.isDown())
            return;

         // we add the screen buffer for the previous frame line
         addScreenBufferToCache(this.frameBuf, frameLine);

         // restore the buffer for the next row, but don't send it to the
         // client
         int nextLine = frameLine + 1;
         frameBuf = getScreenBufferFromCache(nextLine);
      }
   }

   /**
    * Simple functional interface which accepts two parameters and returns a boolean.
    */
   @FunctionalInterface
   private static interface WidgetCheck<One, Two>
   {
      public boolean check(One one, Two two);
   }

   /**
    * Unconditionally initializes the frame's screen definition.
    */
   private void initScreenDefinition()
   {
      // create a new frameDef
      frameDef = new ScreenDefinition(frameId);

      // add the frame config as the first item
      FrameConfig fc = frame.config;

      frameDef.addConfig(fc, frameId);

      // add headers first

      WidgetConfig cfg = null;

      for (String s : n2h.keySet())
      {
         GenericWidget<?> gw = n2h.get(s);

         cfg = gw.config();
         cfg.frameId = frameId;

         frameDef.addConfig(cfg);
      }

      // add all the widgets in the order of their definition

      for (GenericWidget<?> gw : widgets)
      {
         cfg = gw.config();
         cfg.frameId = frameId;
         frameDef.addConfig(cfg);
      }

      if (isShared() && frame.isVisible().booleanValue())
      {
         frameDef.setSuppressRedraw(true);
      }
   }
}
