/*
** Module   : LogicalTerminal.java
** Abstract : server runtime implementation of Progress UI concepts
**
** Copyright (c) 2005-2022, Golden Code Development Corporation.
**
** -#- -I- --Date-- --JPRM-- ----------------------------------Description-----------------------------------
** 001 NVS 20051026   @23493 Created initial version.
** 002 NVS 20060114   @24089 Added hide() methods.
** 003 GES 20060116   @23923 Safety code for active frame queries.
** 004 GES 20060117   @23954 PAUSE rework and fixes.
** 005 NVS 20060125   @24070 Removed all traces of DownFrame class.
** 006 NVS 20060127   @24090 Removed debug output.
** 007 GES 20060128   @24115 Added global callback support.
** 008 GES 20060128   @24116 Added final "Procedure complete..." pause.
** 009 NVS 20060130   @24144 All client and server exports returning 
**                           screen buffer, now take screen buffer as well.
**                           These are: waitFor(), apply(), trigger() and
**                           validate.
** 010 SIY 20060130   @24155 Now pauseBeforeEnd() export is used to perform 
**                           controlled pause before application quit.
** 011 SIY 20060201   @24190 Added versions of message(...) methods with 
**                           color support. Added returning of value from
**                           message() where appropriate.
** 012 GES 20060202   @24214 Removed BDT return from message statements 
**                           (put in w #011) because this does not match 
**                           the API.
** 013 SIY 20060202   @24215 Iteration notifications for frame manager.
** 014 GES 20060203   @24259 Rewrote frame lookup method to be simpler
**                           and faster.
** 015 GES 20060204   @24271 Converted message() to use Accessor instead
**                           of BaseDataType since the delegator approach
**                           is needed to support database fields. Updated
**                           waitFor() to handle null input.
** 016 GES 20060210   @24531 New approach to importing client exports.
** 017 SIY 20060210   @24513 Added getSelection() method.
** 018 NVS 20060208   @24508 Added initialization of control flow stack.
** 019 SIY 20060213   @24583 Added support for block start/end/iterate
**                           tracking; javadoc cleanups; fixed infinite
**                           loop in consumeEvent().
** 020 SIY 20060215   @24600 Changes for messageBox() implementation.
** 021 NVS 20060221   @24621 Default waitFor events are synonyms for GO,
**                           other special cases like STOP, END-ERROR etc
**                           are well known to the ThinClient and handled
**                           there.
** 022 GES 20060214   @24685 Stream redirection and some general cleanup.
** 023 NVS 20060224   @24731 Infinite loop protection needs to be turned
**                           off when a blocking terminal I/O operation is
**                           issued. Modified: choose(), message(),
**                           messageBox(), waitFor(), readKey(),
**                           pauseWorker().
** 024 GES 20060226   @24745 Added a query allowing the client to request
**                           the latest screen buffer from the server.
**                           This is useful in cases where the client
**                           initiates the view() of a frame without an
**                           explicit call from the server (e.g. page
**                           headers and footers in redirected output
**                           mode).
** 025 NVS 20060226   @24748 HELP request invokes a directory configured
**                           method which can be a converted Progress
**                           application's execute().
** 026 GES 20060226   @24758 Header and footer support.
** 027 SIY 20060220   @24631 Finished choose() support.
** 028 SIY 20060226   @24767 Fixed handling of SET/UPDATE in message().
** 029 GES 20060227   @24769 Fix for refresh() signature.
** 030 GES 20060227   @24775 Added passing a screen-buffer for widget
**                           level view (otherwise the latest screen
**                           values won't be visible like they should).
** 031 NVS 20060302   @24849 processEvents() export now takes and returns
**                           the ScreenBuffer.
** 032 NVS 20060306   @24906 WaitList does not need default events, they
**                           are recognized by ThinClient anyway.
** 033 NVS 20060306   @24919 Removed debug output from trigger() method.
** 034 NVS 20060308   @24941 Using new ScreenBuffer with widget state.
** 035 NVS 20060309   @24965 Managing ENTERED state through ScreenBuffer's
**                           state.
** 036 SIY 20060228   @25137 Changed signature for the getColumns() and
**                           getRowCount(), other changes related to browse
**                           support.
** 037 NVS 20060320   @25160 Simplified validation protocol by using new
**                           class ValidationResult.
** 038 NVS 20060321   @25179 WaitList is stuffed with the validation list
**                           before traveling down to the client.
** 039 GES 20060322   @25187 Proper implementation for the terminal type
**                           methods.
** 040 NVS 20060323   @25210 Fixed NullPointerException in validation when
**                           the widget being validated hasn't changed.
** 041 SIY 20060323   @25223 view() now passes specified widgets if list
**                           of widgets is provided explicitly.
** 042 GES 20060404   @25352 Fix for wait-for with the active frame as the
**                           target.
** 043 GES 20060404   @25360 Added an no parameter version of message()
**                           which writes an empty string to the msg line.
** 044 GES 20060411   @25472 Added waitFor() version with NumericType
**                           in addition to simple int for wait time.
** 045 GES 20060413   @25520 Added messageBox() signatures.
** 046 SIY 20060424   @25686 Added method for obtaining handle for given
**                           frame ID.
** 047 GES 20060426   @25732 Removed underline convenience method.
** 048 GES 20060428   @25835 New view() entry point to allow overrides of
**                           the screen buffer.
** 049 GES 20060502   @25932 Safety code for cases when the client sends
**                           back a frame ID that is not valid on the 
**                           server.
** 050 GES 20060505   @25989 messageBox should only disable infinite
**                           loop protection in set/update mode.
** 051 GES 20060510   @26060 Slight modification to blockIterate() call
**                           to pass an additional flag.
** 052 NVS 20060511   @26090 Reimplemented conditional down() and up()
**                           calls in GenericFrame and now this class does
**                           not need neither method.
** 053 GES 20060517   @26196 Match new Finalizable interface requirement.
** 054 SIY 20060517   @26235 Added notification of the client that block 
**                           is retried.
** 055 NVS 20060522   @26430 Added missing methods clear(), clear(boolean)
** 056 GES 20060523   @26440 Added new client export stopEditingMode().
** 057 ECF 20060524   @26534 Added missing message() method variant.
** 058 GES 20060531   @26821 Support the new enable() signature on the
**                           thin client.
** 059 NVS 20060611   @27115 WaitList merging in waitFor() now marks
**                           trigger related events vs exit conditions and
**                           passes both kinds down correctly.
** 060 GES 20060612   @27121 view() must pass a screen buffer with all
**                           changed data, not just the data of the
**                           listed fields.
** 061 NVS 20060612   @27139 Fixed some NullPointerExceptions in trigger
**                           handling code for ANYWHERE triggers with no
**                           frame involved.
** 062 NVS 20060612   @27168 Fixed another NullPointerException in waitFor
**                           method receiving no ScreenBuffer from the
**                           client.
** 063 GES 20060613   @27185 Avoid null pointer in apply processing since
**                           it is legal to get a null back instead of
**                           a screen buffer in the case where the given
**                           frame has no data containing widgets.
** 064 GES 20060614   @27190 Converted the status related signatures to
**                           String parms instead of character. Fixed
**                           statusInput() (no parms version) to call
**                           statusInputRevert() as is documented (it
**                           just set the input message to empty string
**                           previously).  Unknown value processing is
**                           now handled here, not in the thin client.
** 065 GES 20060616   @27394 clear() and clearAll() forms that operate
**                           on a queried frame (getActiveFrame) must
**                           call their respective counterparts in the
**                           frame to allow the frame's state to be 
**                           properly maintained.
** 066 NVS 20060619   @27430 Returning empty ScreenBuffer instead of null
**                           from trigger() method.
** 067 GES 20060619   @27475 Safety code to avoid null pointer exceptions.
**                           WARNING: it is still TBD as to why the
**                           getWidgetForId() method is called with IDs 
**                           that seem valid (e.g. frame ID 31000 and
**                           widget ID 31005) but are not in the frame
**                           registry. At this point no adverse effect
**                           has been found so no further work is being
**                           done.
** 068 GES 20060623   @27556 Set lastkey after message, choose and pause.
**                           Also now ensure that if a condition is raised
**                           that the lastkey will still be set. This was
**                           needed in the above places and in waitFor.
** 069 GES 20060627   @27605 Signature change to pass a screen buffer
**                           to enable() to allow truly dynamic down
**                           frame sizing.
** 070 SIY 20060629   @27697 Removed redundant up() method.
** 071 GES 20060628   @27694 Removed resetChanged() call as it is now
**                           redundant.
** 072 GES 20060630   @27735 Support for choose() passing a screen buffer
**                           to the thin client and processing one on
**                           return.
** 073 GES 20060705   @27807 Moved setColors() to GenericFrame.
** 074 GES 20060712   @28005 Resolved class-cast error in wait-for.
** 075 GES 20060712   @28031 New waitFor signatures to accommodate a frame
**                           as the focused widget.
** 076 GES 20060714   @28060 Apply for current-window (and any case where
**                           the specified target widget has no frame)
**                           will now pass a null screen buffer to the
**                           client.
** 077 SIY 20060714   @28071 More waitFor signatures.
** 078 SIY 20060720   @28145 FRAME-INDEX/FRAME-VALUE handling is changed.
** 079 GES 20060726   @28203 Moved interactions counter to server side
**                           and implemented bidirectional state synch
**                           with the client.
** 080 GES 20060728   @28219 Shifted conditional up/down processing to
**                           use state synch in a "one-shot" approach
**                           that is simple (no arrays) but by design it
**                           must be highly coupled with a subsequent
**                           call to a client export that needs to honor
**                           a conditional up/down.
** 081 SIY 20060726   @28233 Complete rewrite of the frame hiding handling. 
**                           Many minor changes.
** 082 GES 20060731   @28238 Added support for deferral of markFrames()
**                           and destroyFrame().
** 083 GES 20060802   @28297 Switched to a common helper method.
** 084 SIY 20060802   @28337 Improving handling of the retry.
** 085 GES 20060803   @28380 Removed unneeded pushWindow() at startup.
** 086 GES 20060807   @28449 Moved message list content processing to
**                           server to reduce traffic over the protocol.
** 087 SIY 20060807   @28458 Reworked frame hiding algorithm.
** 088 GES 20060808   @28485 Shifted APIs to use Java primitives instead
**                           of the Progress compatible wrappers, where
**                           possible.  Moved the wrapper processing to
**                           the server from the client.  Made some fixes
**                           to messageBox() to add missing functionality.
**                           Major javadoc improvements.
** 089 GES 20060808   @28491 Use int instead of long for the backing 
**                           methods that support the server exports.
** 090 SIY 20060810   @28519 Added frame hiding processing for lightweight
**                           blocks.
** 091 GES 20060824   @28876 Make apply() actually call the client even
**                           if getActiveFrame() returns -1. This 
**                           corresponds to the situation where there is
**                           no frame and/or no focused widget. On the
**                           client there is default processing using the
**                           main Window itself. The previous approach
**                           had just silently returned.
** 092 NVS 20060914   @29605 Changed the interactions counter tracking
**                           logic. Nested blocks conditionally add to the
**                           current block's counter the cumulative number
**                           of interactions detected within the blocks.
**                           Any ConditionException that causes the block
**                           termination (END, ERROR, STOP), clears the
**                           cumulative counter. This logic works to any
**                           nesting level.
** 093 NVS 20060927   @29996 pause(0) calls are now passed down to the
**                           client as well. The reason is these calls
**                           reset the pending pause flag.
** 094 GES 20061005   @30157 Force the END-ERROR key to always generate
**                           ERROR (never ENDKEY) when in a message stmt
**                           during nested error mode. This only matters
**                           if there is a pause caused by the message
**                           display and the user presses END-ERROR. In
**                           nested mode ENDKEY can never occur.
** 095 NVS 20061005   @30167 Reimplemented setScreenValue() method and
**                           implemented setFrameValue() method in support
**                           of FRAME-VALUE statement.
** 096 NVS 20061006   @30203 Using new getOriginalName() method to get the
**                           original widget name for FRAME-FIELD.
** 097 GES 20061009   @30224 Support for headless mode.
** 098 GES 20061010   @30280 Handle deep frames separately in case of
**                           RETRY.
** 099 GES 20061017   @30451 Match server exports API change.
** 100 NVS 20061017   @30467 pauseBeforeEnd() takes no parameter and is
**                           called conditionally, only when necessary.
** 101 GES 20061020   @30554 On CTRL-C during an input blocking operation,
**                           set lastkey to -1.
** 102 GES 20061030   @30789 Changed the return value from trigger() to
**                           eliminate the extra call to the client
**                           (to mark an event as "consume") and to make
**                           the trigger processing synchronous. This
**                           also makes it safer for nested triggers.
** 103 GES 20061102   @30892 Added forceDeregisterPageElement().
** 104 SIY 20061105   @30979 Changed signatures of Browse related methods
**                           and other Browse-related methods.
** 105 GES 20061114   @31124 Enable apply to update the list of the
**                           currently active triggers.  This is
**                           especially important in the case where apply
**                           is used outside of a wait-for.
** 106 GES 20061127   @31410 Added a shared flag to destroyFrame() to
**                           ensure that more deeply nested frames don't
**                           exit scope too soon.
** 107 NVS 20061129   @31474 Added new method signalFrameScope() to signal
**                           frames entering new scope, which is important
**                           for proper clearing of the down frame state.
** 108 GES 20061206   @31596 Added support for promptFor().
** 109 GES 20061212   @31663 Implemented new client exports that implement
**                           the temporary redirection in-line on the
**                           client rather than driving it via separate
**                           calls from the server.
** 110 SIY 20070117   @31849 New frame ID management scheme.
** 111 NVS 20070118   @31888 Using WidgetId class and new trigger registry
**                           management based on EventList class. WaitList
**                           class is no longer in use.
** 112 SIY 20070119   @31900 Minor changes in frame ID reusing scheme.
**                           Removed separate handling of shared frames.
** 113 SIY 20070123   @31926 Changes in frame ID reusing scheme.
** 114 SIY 20070124   @31973 Restored frame ID reusing and cleaned up 
**                           identifyFrame().
** 115 SIY 20070306   @32283 Fixed handling of out of scope frames. 
**                           Added client driven frame deregistration. 
** 116 NVS 20070328   @32632 help() request from the client now reports
**                           the lastKey variable.
** 117 SIY 20070418   @33099 Fixed handling ENTERED widget state.
**                           Also, see @33100 for more details.
** 118 NVS 20070426   @33245 Implemented infinite recursion protection for
**                           triggers by ignoring trigger requests for the
**                           already active pair {event, widget}.
**                           Added new methods: nextTabItem, isSensitive,
**                           setSensitive which call new client exports.
** 119 NVS 20070502   @33359 Validation should use a default value for the
**                           widget in case there is nothing in the frame
**                           buffer for it.
** 120 NVS 20070507   @33430 Added new method getAnyWidgetForId() which
**                           may return frame widget as well as field
**                           level widgets. Used this new method in
**                           trigger() method where a frame ID can come as
**                           trigger target.
** 121 SIY 20070506   @33424 Updated getRowCount(). Other minor cleanups.
** 122 SIY 20070521   @33752 Added support for open/close frame scope
**                           marking.
** 123 EVL 20070609   @34005 Adding explicit import of the class
**                           com.goldencode.p2j.net.Queue to eliminate
**                           conflict with the same class from java.util
**                           package to be able to compile for Java 6
** 124 SIY 20070711   @34429 Fixed shared frame support.
** 125 NVS 20070716   @34538 Client export view() has changed signature.
** 126 NVS 20070723   @34677 Added corresponding view/display flags to all
**                           calls of the view() client exports.
** 127 NVS 20070817   @34880 Quick return for no-op call of redirectOutput
** 128 NVS 20070830   @34980 down() calls to the client now pass the flag
**                           that tells conditional downs.
** 129 ECF 20070912   @35054 Partial rollback of #023 (@24731). Infinite
**                           loop protection must not be disabled from
**                           pauseWorker(). Progress will not disable ILP
**                           for a pause.
** 130 ECF 20071112   @35901 Made changes required by net package
**                           refactoring. Replaced Queue with Session.
**                           Replaced deprecated SecurityManager context
**                           token calls with ContextLocal implementation.
** 131 SIY 20071126   @36039 Added support for frame DOWN maintenance
**                           using StateSynchronizer.
**                           Ref: #36034
** 132 GES 20071214   @36375 Further restrict screen buffer switching to
**                           choose.
** 133 CA  20080115   @36815 ScopeProcess.process will deregister any 
**                           page-top/bottom frames when the next iteration 
**                           is performed.
** 134 ECF 20080208   @37031 Added getSelectedRowIndex(). Gets the row
**                           index of a selected row in a browse widget.
** 135 NVS 20080213   @37041 Added the ability to wait for events in the
**                           waitFor() method on the client where there is
**                           no frame at all.
** 136 CA  20080305   @37297 Moved the message output logic about terminal
**                           redirection to the ThinClient.message method.
**                           Ref: #37225
** 137 NVS 20080307   @37411 Added setLocation() for dynamic ROW and COL
**                           widget attributes support.
** 138 GES 20080320   @37597 Method name change.
** 139 NVS 20080320   @37621 Changed call to enable() export of ThinClient
**                           to match the new signature.
** 140 NVS 20080326   @37686 Changed call to promptFor() export of 
**                           ThinClient to match the new signature.
** 141 SIY 20080331   @37741 Check color validness before using it for 
**                           messages.
**                           Ref: #37732
** 142 SIY 20080409   @37875 Return NullWidget instance instead of null
**                           from focus() method.
** 143 GES 20080421   @38072 Added setFormat().
** 144 NVS 20080423   @38101 To make triggers work while in CHOOSE, server
**                           now passes the combined event list to client.
** 145 GES 20080429   @38147 Support for deferring the push of screen
**                           definitions to do them in a batch instead of
**                           making a trip to the client for each one.
**                           This improves performance with no negative
**                           side effects.
** 146 GES 20080501   @38186 Added parameters to allow multiple calls to
**                           be consolidated into a single call.
** 147 GES 20080613   @38879 Refactored client-to-server state sync and
**                           added pause monitoring. Explicit or implicit
**                           pauses on the client will be tracked and the
**                           transaction manager will be notified.
** 148 SIY 20080715   @39110 Reset 'suppress redraw' flag after pushing
**                           screen definitions to client. Inferred generic 
**                           type arguments.
** 149 CA  20080905   @39743 Added missing getter for VISIBLE attribute.
** 150 SIY 20080910   @39767 Cleanup list of MarkEntries before adding 
**                           CLOSE_SCOPE for specified frame if OPEN_SCOPE 
**                           for this frame is still in the list.
** 151 GES 20080911   @39792 Added signatures to allow primitives to be
**                           passed to setFrameValue().
** 152 SIY 20080912   @39796 When cleaning up list of MarkEntries keep 
**                           DESTROY entries in the list to preserve proper
**                           cleanup of frame definitions at client.
** 153 CA  20080915   @39808 getFrameLine was changed to receive the 
**                           pendingUpDown value as a parameter.
** 154 SIY 20081003   @40017 Added handling of special case for FOR loops. 
** 155 SIY 20081006   @40044 REPEAT loops seems behave like FOR loops. 
** 156 CA  20080930   @39985 Added accessors for FRAME:WIDTH/HEIGHT-CHARS
**                           and FRAME-COL/ROW attributes.
** 157 CA  20081027   @40258 For a message displayed as an alert-box, the 
**                           default message is computed from the passed
**                           variable, when UPDATE or SET is used.
** 158 GES 20081106   @40359 Editing blocks must use the interactions
**                           counter of the containing block rather than
**                           having its own counter.
** 159 GES 20081126   @40738 Massive rework and cleanup of the
**                           interactions counter processing. Only blocks
**                           that have the ENDKEY property will maintain
**                           a block-level counter.  This explains why
**                           EDITING blocks (and some DO blocks/loops)
**                           have unusual handling of END-ERROR.
** 160 SIY 20081201   @40754 Updated handling of global 'can hide' flag
**                           according to discovered rules.
** 161 CA  20081204   @40818 Implemented ROW-DISPLAY event: these events
**                           are registered with the client at the same
**                           time as their definition.
** 162 SVL 20081208   @40825 Match client exports API change.
** 163 CA  20081210   @40831 Match BrowseWidget API change.
** 164 SIY 20081209   @40852 Reworked frame hiding implementation.
** 165 CA  20081210   @40844 Implemented CLOSE QUERY statement - on close,
**                           the browse widget will be notified to reset
**                           its state.
** 166 GES 20081211   @40864 Avoid rollup of contained block's
**                           interactions counter depending on the
**                           type of enclosing block.
** 167 SIY 20081215   @40904 Frames from outer scopes can hide other 
**                           frames just like frames from current scope.
** 168 SIY 20081216   @40919 Narrowed condition when #167 can take place
**                           (frames from outer scopes can hide other
**                           frames when viewed in inner scopes only if
**                           current scope has no frames scoped to it).
** 169 GES 20081217   @40925 Rework of trigger registration to pass the
**                           necessary metadata to allow trigger
**                           instantiation to be delegated to the runtime.
**                           This is needed to allow a "fresh" trigger to
**                           be created for every invocation. That allows
**                           the trigger's data members to be set to the
**                           proper default values each time. Trigger
**                           execution was also modified to handle this
**                           new requirement.
** 170 SIY 20081224   @40987 Added hideWindow() method.
** 171 SIY 20090210   @41285 Improved frame hiding - frame which was not 
**                           iterated nor just entered its scope does not
**                           trigger frame hiding.
** 172 SIY 20090213   @41307 Added 'force' parameter into refreshStartRow
**                           signature.
** 173 CA  20090227   @41386 Fixed SELF handle - if outside a trigger, a
**                           NullWidget instance must be returned.
** 174 SIY 20090326   @41709 Support for status saving/restoring logic.
** 175 CA  20090326   @41724 Modified apply(...) signature to identify an
**                           APPLY call with or without explicit target.
**                           For implicit APPLY target, focus must be used
**                           as target, if exists. Else, APPLY ends.
** 176 CA  20090413   @41777 Adjust state of HIDDEN/VISIBLE attributes for
**                           widgets received from the client.
** 177 CA  20090414   @41788 Modified getAnyWidgetForIdInt to handle H176 
**                           missed case for VISIBLE/HIDDEN flag when  
**                           header widgets exist.
** 178 CA  20090415   @41796 APPLY ... FOCUS, when FOCUS does not exist, 
**                           must not execute and will show an error.
** 179 CA  20090421   @41819 APPLY with implicit target, when there is no
**                           focused widget, must send the event to the
**                           client, to be processed - but, no triggers
**                           will be invoked.
** 180 SIY 20090421   @41833 Handle frame hiding on new procedure entry.
** 181 SIY 20090422   @41843 Limit cases when frame hiding triggers wiping of
**                           the screen.
** 182 SIY 20090602   @42546 Notify frames that condition was thrown in current scope.
** 183 SVL 20090612   @42678 Added resetBufferSizeManager() method.
** 184 SVL 20090701   @43014 Added helper method for error reporting. Improved
**                           a bit ServerState encapsulation. Minor cleanups.
** 185 CA  20090901   @43819 In c'tor, Control.init no longer receives the 
**                           InterruptHandler implementation - now is set 
**                           globally per entire server in the SessionManager.
** 186 SIY 20090903   @43828 Non-down frame displayed in inner loop and which 
**                           marked for hiding must retain this markup in order
**                           to prevent "stuttering" in frame hiding. 
** 187 NVS 20090918   @43955 Scopeable factory produces the scopeable only if
**                           the session is not set to the headless mode. Now
**                           the "headless" session context token is checked
**                           firsts. It's supposed to be a Boolean instructing
**                           the runtime whether to initialize the UI or not.
**                           If no token found, then the algorithm falls back
**                           to the directory entry check.
** 188 SIY 20091012   @44134 Ignore too fast readKey and inputClear requests. 
** 189 SIY 20091119   @44439 Fixed ClassCastException in validate(). 
** 190 CA  20091111   @44328 Added a statusDefault(String, boolean) export 
**                           which disables 63-char trimming, if the
**                           noLengthLimit flag is set.
** 191 SIY 20100405   @44791 Fixed frame hiding/pausing after hideAll().
** 192 EVL 20100527   @44891 Maintain proper balance when a block with ENDKEY
**                           support is nested in a parent block which does not
**                           rollup the interactions counter. This can occur in
**                           2 cases. 1st: in a trigger that is called while in
**                           an UPDATE statement.  2nd: in a validation expr
**                           which calls a user-defined function while in an
**                           UPDATE stmt. There were really two problems here. 
**                           The topmost blocks stack value was not popped as it
**                           should have been. But also, the counter value
**                           should only be restored in certain cases and in
**                           others it should be popped but then forgotten.
** 193 EVL 20100616   @44908 RETRY causes non-ENDKEY blocks to have the
**                           interactions counter restored to the value that was
**                           there on entry to the block. This may be needed for
**                           iterate too, but the fix is limited at this time to
**                           reduce risk.
** 194 SIY 20101004          Added necessary method stubs.
** 195 ECF 20101005          Added stubs for some SESSION handle attributes.
** 196 SIY 20101005          More method stubs.
** 197 SVL 20101006          Added systemHelp functions.
** 198 SIY 20101006          Added stubs for CURRENT-WINDOW assignments in
**                           session and in THIS-PROCEDURE.
** 198 SIY 20101006          Added stub for THIS-PROCEDURE:CURRENT-WINDOW. 
** 199 SIY 20101122          Small cleanups in javadoc and other minor changes.
** 200 ECF 20110107          Added replaceFrame to support shared frame
**                           changes in GenericFrame.
** 201 SIY 20110112          Changes related to new color handling scheme.
** 202 SIY 20111101          Updated imports.
** 203 GES 20111020          Modified per-session initialization to detect
**                           when the client code is running inside the server
**                           process AND to initialize for local redirection
**                           instead of using a remote object.
** 204 SIY 20111202          Updated imports.
** 205 CA  20130112          Use handle instead of phandle (this is no longer used).
** 206 CA  20130131          Implement ON ... PERSISTENT RUN triggers.
** 207 CA  20130215          Removed deleteWidget.
** 208 CA  20130221          Added HEIGHT/WIDTH-CHARS/PIXELS attribute support.
** 209 CA  20130222          Added stubs for more SESSION-related attributes and methods.
** 210 CA  20130222          Renamed self() to _self(), to disambiguate from SelfManager.self.
**                           On trigger invocation, call SelfManager.push to save the target 
**                           widget.
** 211 CS  20130223          Added conversion support for CURRENT-WINDOW, ACTIVE-WINDOW and 
**                           DEFAULT-WINDOW system handles features
** 212 CA  20130223          Fixed CURRENT/ACTIVE/DEFAULT-WINDOW handles, as they need to act as 
**                           true handles.
** 213 CA  20130308          Added conversion support for HIDE STREAM stream ALL statement.
** 214 OM  20130516          Changed apply() to accept 64-bit integer values.
** 215 CA  20130822          Added support of appserver async requests: posting and executing of
**                           server-side events on the client side.
** 216 CA  20130908          Added support for socket runtime.
** 217 CA  20131001          Added API to change the numeric separators on client-side.
** 218 CA  20131002          Fixed frame as resource usage.
** 219 CA  20131002          Implemented CLIENT-TYPE and WINDOW-SYSTEM attribute; first pass at 
**                           the SET-WAIT-STATE and GET-WAIT-STATE methods (for CHUI mode).
** 220 EVL 20131014          Adding annotation for legacy attributes to be ignored for the class.
** 221 HC  20131031          Changed resource ID data type to Long, see issue #2183.
** 222 OM  20131113          Fixed javadoc for getWaitState() (eliminate warning).
** 223 CA  2013113           Triggers in a persistent procedure must survive until the procedure 
**                           gets deleted.  Fixes related to implicit deletion of static frames.
** 224 VIG 20131127          Added a check of client driver type (isChui)
** 225 CA  20131218          Fixed H221 - the signature for validResource, readResponseEvent and 
**                           connectEvent must match the ones at ServerExports interface.
** 226 EVL 20131127          Adding support for batch mode.
** 227 CA  20140110          Removed caching of TransactionManager - this was required only when
**                           state sync was done in a Thread without a security context, which is 
**                           not the case anymore.
** 228 VIG 20131225          Added LAST-EVENT:FUNCTION key caching (from ClientState)
** 229 VIG 20131226          Added implementation of setCurrentWindow(..) and 
**                           currentWindow(..) methods.
** 230 CA  20140115          Replaced NullWidget usage with unknown handles.
** 231 EVL 20140115          Adding support for background simulation mode.
** 232 HC  20140119          Implemented runtime support for the COLOR 
**                           protermcap-attribute.
** 233 CA  20140218          Fixed TARGET- and SOURCE-PROCEDURE for trigger case.
** 234 GES 20140228          Honor SKIP in normal MESSAGE statements (this is an undocumented
**                           "feature" since the 4GL docs state this is only supposed to work
**                           for VIEW-AS ALERT-BOX).
** 235 CA  20140513          Added a weight for the context-local var, to ensure predetermined 
**                           order during context reset.
** 236 VIG 20140601          Adding moveBeforeTabItem, getFirstTabItem, setFirstTabItem, 
**                           getNextTabItem, extractWidgetIDs tab order management methods.
** 237 VIG 20140612          Fixing getFirstTabItem and moveBeforeTabItem with new signatures.
** 238 MAG 20140512          Implemented CREATE WINDOW statement (only used in GUI).  Assign 
**                           window ID and register windows.
** 239 CA  20140711          Enhanced font support, for GUI clients.
** 240 CA  20140728          Added notification that a frame was hidden by the client-side.
** 241 VIG 20140805          Added active-window and default-window and those implementation 
**                           for GUI. Added pause* methods for cases using IN WINDOW clause.
**                           Added pauseBeforeHide* methods for cases using IN WINDOW clause.
** 242 CA  20140805          Changes related to multi-window support;  added API to load a image 
**                           from server-side.  The removeWindow API was removed, as window 
**                           removing needs to be done on the server-side, too (by 
**                           deregisterWindow).
** 243 MAG 20140812          Implements COLOR-TABLE, for GUI clients.
** 244 VIG 20140828          Changed argument type of targetWindowWorker method argument.
**                           Current window added to server state.
** 245 MAG 20140909          Implements window environments assignment.
**                           Added load environment method to load INI files from server.
** 246 EVL 20140801          Moving widget ID code to the WidgetID class to isolate all ID
**                           there in new approach for ID distribution.  Adding the widgetRegisty
**                           map to store widget-ID relationship for all widgets except default
**                           window.
** 247 HC  20140911          GUI coordinate support - data types refactoring.
** 248 EVL 20140926          More support for dynamic widgets and frames.
** 249 CA  20140926          Added shared widget configuration support, to allow GUI/ChUI
**                           concrete implementation for the same widget ID.  Refactored the
**                           widget configuration classes: all fields were made public; these
**                           classes need to be as dumb as possible, all logic related to
**                           setting/getting a certain field should be at the widget or at the
**                           caller. Refs #2254
** 250 HC  20141005          Prevented multiple instantiation of LogicalTerminal during session
**                           initialization.
** 251 CA  20141021          Added NPE protection during image loading.  Added a way to inform the
**                           server-side about client-side events incoming from a non-keyboard 
**                           device (i.e. WINDOW or MOUSE).
** 252 VIG 20140911          Added status* and message* methods for IN WINDOW cases.
** 253 HC  20141021          Implemented window move/position support.
** 254 CA  20141106          Added window ID to promptFor APIs.  The frame's window ID attribute 
**                           must be set by the client-side.
** 255 CA  20141119          Refactored drawing code to enable multiple OS windows, for GUI
**                           support.  Misc changes/fixes related to drawing on multiple windows.
** 256 CA  20141123          Removed a TODO in pushWindow.
** 257 CA  20141125          Fixed problem in deregisterWindow - client-side needs to be processed
**                           first, and server-side will deregister the window last.
** 258 EVL 20141204          Differentiate window icon loading from other images.
** 259 SVL 20141209          Implemented some browse-related functions.
** 260 HC  20141207          Implemented runtime support for several SESSION attributes. Improved
**                           server-client state transfer of the default window.
** 261 GES 20141120          Added the FRAME-NAME built-in function. Doubled the number of
**                           messageBox() variants to deal with character type as a title.
** 262 EVL 20141216          Adding SESSION:TOOLTIPS attribute handling.
** 263 GES 20141228          Cache window system value and modified to deal with API changes.
**                           Added deleteEditorLine().
** 264 HC  20150111          Fixed widget position change in a frame.
** 265 GES 20150109          Use a cached chui/gui flag from client parameters if it exists
**                           (which will avoid a client round-trip).  It probably always will
**                           exist.  If so we could remove isChui() from client exports.
** 266 GES 20150111          Triggers and validation processing must call handleScreenBuffer()
**                           to set the frame-name, frame-field and related state.
** 267 EVL 20150108          Fix the default application icon to be set as gclogo.ico embedded
**                           into runtime jar file.  Rework for icon/image implementation.
** 268 CA  20150122          Added runtime for APPLY ... TO <handle>.
** 269 HC  20150128          Added support for frame widget position changes induced by
**                           setting ROW/COL/X/Y attributes.
** 270 CA  20150202          Added missing apply(character) API.
** 271 GES 20150204          Clear GO-PENDING when a NEXT statement is used in an EDITING block.
** 272 CA  20150212          Allow events to be captured and processed by the widgets without
**                           server-side or main event-loop notification.  This fixes problems in
**                           the EDITING and WAIT-FOR event loops, for i.e. COMBO-BOX and EDITOR
**                           widgets.
** 273 GES 20150212          For validation of at-base fields, convert the value to the right
**                           data type if needed.  This can happen if the value was not edited
**                           and was previously DISPLAY'd as the wrong type. Without this the
**                           validation expression would get a ClassCastException.
** 274 CA  20150218          On trigger invocation, the frame for the screen buffer to be applied
**                           is identified from the screen buffer's base ID.  This allows SELF and
**                           FOCUS to vary independently, when a trigger is invoked.
** 275 VIG 20150116          Added menu registration/destroying methods.
** 276 CA  20150220          Fixed the FRAME-VALUE statement: if HELP is active, it references the
**                           FOCUSed frame/widget as before the HELP is invoked.
** 277 HC  20150223          Improved trigger recursion detection to also consider events with
**                           multiple numeric codes.
** 278 GES 20150223          Eliminated bogus screen-buffer error messages when negative widget
**                           IDs are processed.  Instrumented trigger processing with logging.
** 279 SVL 20150212          Modified getRows signature. Fixed warning for an editable browse.
** 280 CA  20150225          Fixed SCREEN-VALUE usage for DISPLAY/PROMPT-FOR/... EDITING stmts. 
** 281 SVL 20150305          Updated refresh() signature.
** 282 CA  20150305          CLEAR and CLEAR ALL APIs must update the frame's screen buffer.
** 283 CA  20150312          Fixed mem leak in ScopeProcessor.  Added a warning if isChui calls 
**                           the client-side to determine the ChUI or GUI mode.
** 284 CA  20150317          Added API to determine if a legacy font is fixed-width.
** 285 EVK 20150318          Added SAVE-FILE and READ-FILE support for editor.
** 286 CA  20150402          Added registerSharedFrame, which can add back a frame to the registry. 
** 287 SVL 20150331          Added some browse-related functions. refreshStartRow become
**                           refreshScrollRow with more options.
** 288 IAS 20150404          Added support for new EDITOR attributes/methods
** 289 CA  20150408          Added findDirectChildren.
** 290 SVL 20150411          Added select{Next|Previous}Row, isRowAvailable.
** 291 HC  20150323          Removed business logic from config classes.
** 292 CA  20150416          Changes related to fixing the frame's PARENT attribute.
** 293 VIG 20150225          Added getAllMenus method.
** 294 EVL 20150415          Adding support for prepackaged button widget images.
** 295 SVL 20150420          Added executePendingRefresh, notifyColumnsUpdated.
** 296 HC  20150508          Improvements to GUI window scrolling support.
** 297 CA  20150519          Refactored so that the font specifications are read in only one trip.
**                           Added API to transfer the entire text metrics DB from the server to
**                           the client-side, in only one trip.
** 298 SVL 20150508          Added some browse-related functions.
** 299 HC  20150612          Javadoc fixes.
** 300 SVL 20150625          Added getRowCount().
** 301 IAS 20150626          Added 'debug' method.
** 302 CA  20150722          APIs which affect the current frame row (like CLEAR, DOWN, pending
**                           DOWN in pushScreenDefinition) must refresh the server-side widget 
**                           configs so that they are in sync with the widgets on the current 
**                           iteration.
** 303 SVL 20150722          Updated browse-related API.
** 304 CA  20150725          ClientExports.view APIs now return the frame's down configs, too.
** 305 IAS 20150801          Added support for the WIDGET-LEAVE/WIDGET-ENTER attributes.
** 306 ECF 20150801          Replace StringBuffer with StringBuilder.
** 307 EVL 20150826          Adding support for SCROLL-TO-ITEM() method for selection-list.
** 308 HC  20150831          Fixed the value of config realized field for default window.
** 309 IAS 20150922          Added support for the MOVE-TO-TOP()/MOVE-TO-BOTTOM() methods.
** 310 CA  20151006          The DEFAULT-WINDOW initially is not visible; it gets shown on the 
**                           first UI statement targeting this window, which shows a message or 
**                           widget.
** 311 IAS 20151029          VALIDATE() method implementation
** 312 CA  20151026          PERSISTENT RUN triggers need to pass THIS-PROCEDURE to the trigger.
**                           Added support for MOVE-AFTER-TAB-ITEM.
**     EVL 20151026          Fix for server-side image loading.  New approach to scan all
**                           application jar files with case sensitivity capable mode match.
**                           Jar class loader cache implemented and moved to separate class.
**     CA  20151029          Triggers created in a persistent procedure while a event processing
**                           loop is active need to be pushed to the event list for the nearest
**                           event processing loop, so they will be available in the UI.
**     SVL 20151029          Added queryRepositioned.
**     EVL 20151103          Adding handing image names without extensions while loading from jar.
**     CA  20151106          A fix for HIDDEN attribute: distinguish between HIDE statement and 
**                           setting the HIDDEN attribute to true.  DEFAULT-WINDOW is made visible
**                           only in ChUI.  Save/restore the UI statements in use for each frame, 
**                           when triggers are called.
** 313 HC  20151213          Improved DIALOG-BOX window parenting logic and related changes.
** 314 OM  20160126          Called setDefaultWindow() of ProcedureManager. The server context is
**                           automatically headless, too. Small code maintenance.
** 315 EVL 20160224          Javadoc fixes to make compatible with Oracle Java 8 for Solaris 10.
** 316 IAS 20160229          LOAD-MOUSE-POINTER and SET-WAIT-STATE support.
** 317 HC  20160222          Fixed error reporting when assigning session read-only attributes.
**                           Fixed killing a frame when in setup and remote call to the client
**                           side made.
** 318 OM  20160205          Implemented cleanupPending() for cleaning pending frame definitions
**                           when an external procedure fails to initialize.
** 319 IAS 20160302          Support for the APPL-ALERT-BOXES, DATA-ENTRY-RETURN, DEBUG-ALERT
**                           and IMMEDIATE-DISPLAY SESSION attributes
** 320 IAS 20160328          Client log support
** 321 CA  20160327          Added support for LAST-EVENT:X and LAST-EVENT:Y attributes.
** 322 IAS 20160331          Fixed potential NPE 
** 323 OM  20160401          Fixed method name collision for static proxy.
** 324 SVL 20160406          Added moveColumn.
** 325 CA  20160422          Fixed cleanup of LogicalTerminal instances - they need to be 
**                           deregistered from the protocol/session.  Added explicit cleanup of
**                           dynamic frames, to be deleted from the client-side, too.
** 326 CA  20160418          Added infrastructure for message passing from converted code to a 
**                           customer-specific application or calls from the customer-specific 
**                           application into the converted code (i.e. invoke external programs,
**                           procedures or functions).  Can be used only when the P2J client runs 
**                           embedded in a customer-specific application.
** 327 SVL 20160419          Added moveColumnServer.
** 328 GES 20160616          Reworked screen-buffer access to eliminate unnecessary processing.
** 329 CA  20160627          In appserver case, PUT SCREEN is a no-op and READKEY returns -1, if
**                           unnamed input stream is not redirected. 
** 330 EVL 20160722          Using ScreenBuffer array for editable frames instead of only current
**                           buffer for frame under processing to pass the actual screen buffer
**                           data between server and client.  The methods involved for server
**                           side are: trigger(), validate(), help(), refresh() and
**                           refreshBuffers().  Adding helper method to process array of screen
**                           buffers to separate common code.
** 331 SBI 20160811          Changed deregisterWindow() to assign CURRENT-WINDOW to be
**                           DEFAULT-WINDOW if the current window should be deleted.
** 332 IAS 20160829          displayAndDown() operation support
** 333 CA  20160915          Frames which are being destroyed and not visible are destroyed 
**                           immediately.
** 334 OM  20160919          Fixed infinite pause in batch mode (server-side).
** 335 GES 20161004          Changes to support lambda trigger mode (inner class mode is still
**                           used in some cases).
** 336 CA  20161018          Any client-to-server code invocations must maintain the trigger 
**                           registries, not just explicit trigger invocation.
** 337 HC  20161110          Added getParentFrames().
** 338 CA  20161123          Overlay frames must be destroyed and hidden when their scope ends.
** 339 OM  20161215          In batch mode, hide-all is NO-OP (hWin is invalid/unknown).
** 340 SVL 20170126          Added overloaded version of statusDefault.
**     GES 20170129          Added FRAME-FILE and FRAME-DB support.
**     GES 20170131          Don't eat fatal condition exceptions when executing a trigger on
**                           a handle.
**     CA  20170228          Added PUBLISH/SUBSCRIBE/UNSUBSCRIBE extensions for global support and
**                           for external applications.
** 341 OM  20170320          Added support for Themes.
** 342 SVL 20170417          Added getRows() for search by key.
** 343 EVL 20170510          Changing the fonts passed from current theme, enabling default
**                           aliases to be obtained as well.
** 344 OM  20170504          Support for MOVE-TO-EOF, SEARCH and INSERT-FILE for editor widget.
** 345 SVL 20170812          Added scrollToCurrent(Selected)Row, calcBrowseDown,
**                           pushIntermediateRows, functions related to row display draw listener.
** 346 SVL 20170831          getEditingFillIns refactored to getEditorsConfigs.
** 347 EVL 20170831          Getting the file size limits for uploading used in web client.
** 348 CA  20170914          Fix for walking window children via the first-child/next-sibling
**                           attributes.
** 349 HC  20171003          Implemented SESSION:SUPPRESS-WARNINGS runtime support and the
**                           ability to handle (display or output depending on the runtime state)
**                           legacy warnings.
** 350 CA  20170929          Fixed MENU:POPUP-ONLY attribute getter return type.
** 351 SBI 20161101          Moved tab item processing to the server side, changed to provide
**                           the client side with tab item list.
** 352 SBI 20171020          Changed updateTabItemList to use getFrame as a field access method.
** 353 SVL 20171024          Added displayError(errmsg, noMessageBox).
** 354 CA  20171026          Added invokeProcedure, currently used by native COM automation, when
**                           an event is fired.
** 355 CA  20171122          Added getFrame.
**     GES 20171123          Reworked help() to execute the applhelp.p program via the propath
**                           instead of reading a configured directory entry.
** 356 GES 20171206          Update javadoc and code to match changes to silent error mode.
** 357 SVL 20171204          Added refreshCurrentRow.
**     SBI 20171204          Changed to update tab item list only if the owner frame has been
**                           realized.
**     EVL 20171218          Getting the GUI menu delay value from server config file.
** 358 CA  20180130          Replaced pushScreenDefinition() with pushWidgetAttr() for cases when
**                           business logic is assigning a widget attribute.
**                           Small optimization: avoid getTriggerId() to be called multiple times.
** 359 HC  20171229          Any modal window unconditionally disables wait state.
**     ECF 20180102          Use String.replace instead of more expensive String.replaceAll in
**                           methods called frequently
**     HC  20180116          LASTKEY must report -1 in some trigger cases.
** 360 SVL 20180209          Added runReport and setEnhancedParameter.
** 361 SVL 20180321          Improved javadoc.
** 362 HC  20180410          Renamed ReportOutputFormat to MediaType.
** 363 SBI 20180419          Added current windows stack and changed deregisterWindow.
** 364 CA  20180313          Added FWD extension SESSION:DISABLE-REDRAW.
** 365 ECF 20180420          Minimize context-local lookups.
** 366 EVL 20180507          Adding method to get input stream for arbitrary resource located in
**                           application jar file.
** 367 GES 20180508          Added general purpose disableRedraw() as a built-in function (FWD
**                           extension).  These methods are a general replacement for the WIN32
**                           API LockWindowUpdate.
** 368 SBI 20180518          Renamed file-system/pkgroot into legacy-system/pkgroot, changed
**                           getImageStreamFromApplication in order to resolve legacy absolute
**                           path names.
** 369 ECF 20180613          Performance improvement: limit the search for resources in jars to
**                           only those jars specified as "resource-jars" in the directory.
**         20180615          Refactored resource loading code and moved all jar search code to
**                           FileSystemOps.
** 370 CA  20180503          SESSION:SYSTEM-ALERT-BOXES is by default on for GUI and off for ChUI.
**                           Renamed/moved some SESSION attribute getters/setters to fix issues
**                           with using the SESSION via a handle.
**     CA  20180523          Fixed environment name, when used in maps: registries are 
**                           case-insensitive, stanza INI are case-sensitive.
**     CA  20180529          Accessing frame's dimension or location attributes, or changing the
**                           frame's location forces the frame to be placed.
** 371 CA  20180608          Protected cleanupTriggers for headless mode.
** 372 EVL 20180621          Fix for empty frame report generation.
** 373 CA  20180710          Fixed issues related to dynamic menu deregistration.
** 374 ECF 20180710          Performance: use cached, context-local headless check instead of
**                           figuring it out each time during trigger cleanup.
** 375 SVL 20180529          Added functions for managing enhanced browse configuration.
** 376 CA  20180727          Changes to allow APPLY statement to be executed from the remote 
**                           embedded app.
** 377 CA  20180727          Added isWindow.
** 378 CA  20180731          Added HWND function (FWD extension), which returns a handle from the
**                           received HWND value (as read from the widget's HWND attribute).
** 379 CA  20180820          Fixed 'apply "event" to h' where 'h' is a window widget.
**     SVL 20180901          Added setDynamicSorting/Filtering.
** 380 SBI 20181007          Added getDesktopSettings().
** 381 EVL 20181126          Adding last key cleanup more cases to fix issues with browse column
**                           widget and SE_LEAVE event when last event before was mouse generated.
** 382 HC  20181124          Implemented TREEVIEW widget and related changes.
** 383 CA  20181204          Added support for V6DISPLAY attribute (just access, it does not
**                           affect the UI).
**     ECF 20181215          Javadoc fix.
**     HC  20181212          Implemented support for asynchronous key state read emulation.
** 384 EVL 20190219          Fix for NPE in trigger execution for procedure that does not exist
**                           anymore.
** 385 HC  20190306          Implemented limited Windows API emulation for GetSystemMetrics
**                           function.
**     SVL 20190325          Updated signature of notifyColumnsUpdated.
**     CA  20190326          Implemented ProcedureHelper and TransactionHelper instead of direct 
**                           usage of the static API.  This allows the elimination of context  
**                           local usage in ProcedureManager and TransactionManager.
**     CA  20190328          Triggers are emitted as anon inner classes, instantiated via a lambda 
**                           expression.
** 386 EVL 20190321          Adding CLIPBOARD related text methods: EDIT-(COPY|CUT|PASTE|UNDO).
**                           Adding support for EDIT-CAN-UNDO, EDIT-CAN-PASTE attributes.
** 387 OM  20190411          Fixed unwrapping to expected type.
** 388 CA  20190423          Better approach for H395 - triggers are either emitted as singleton 
**                           or as an anon inner class, and always inline where they are defined.
** 389 HC  20190423          Implemented direct font control, AKA custom font support.
** 390 EVL 20190629          Adding ability to get current user name.
** 391 SBI 20190514          Implemented getResourceIdByWidgetId(int).
** 392 EVL 20190709          Adding client machine name/IP getting.  Fix for setCurrentWindow()
**                           call with FrameWidget handle as input parameter.
**     CA  20190811          Optimized ScopeProcessor.process.
**     EVL 20190817          Changing approach of handling request for browse widget already gone.
**                           We need to avoid simple RuntimeException throwing.  This completely
**                           crashes the application.  The logging is now used instead.
**     HC  20190821          Added new trigger method for invoking triggers directly on the server.
** 393 SBI 20190925          Added client storage.
** 394 HC  20190911          Implemented COM event emit originated from client.
**                           Triggering events directly from server must lookup the trigger on
**                           the client.
** 395 SVL 20191002          Added rowEntryCreated/Deleted and checkDeleteEntryRowIndex.
** 396 HC  20191019          Improved the logic preventing recursive triggers to consider the
**                           cases when the widget matching the trigger is different from the
**                           source widget.
** 397 IAS 20191022          Added DEBUG-ALERT support to the displayError method.
**     SVL 20191111          Added hasRows().
** 398 CA  20191119          Track the executing legacy class, to be able to register any buffer  
**                           to its source definition class.
** 399 MAG 20191015          Added refreshWidget(), oleDrag().
**                           Added IVBDataObject and IVBDataObject methods.
** 400 HC  20200128          Removed client-server widget config definitions synchronization.
**                           Client and server config managers can now be dynamically extended
**                           with new widget configurations.
** 401 EVL 20200203          Change method access type to public for register/deregister widgets.
**     CA  20200211          Allow fractional seconds (convert to millis based on the legacy 
**                           version).
**     CA  20200213          A small change in cleanupTriggers, to allow calls from related 
**                           threads.
** 402 VVT 20200218          Code cleanup: unused variable(s) removed.
**                           Missing template arguments added.
**                           Missing @Override annotation(s) added.
**                           @SuppressWarnings annotations added.
**                           Unused private method removed: extractWidgetIDs()
** 403 HC  20200313          Javadoc fixes.
** 404 RFB 20200319          Added getBrowseNumEntries for NUM-ENTRIES to get the value from 
**                           the client using getBrowseCacheSize.
**     HC  20200331          Improved triggering events from the server, now all the events from
**                           all the scopes are included in the search set.
**     SBI 20200331          Implemented findWidgetByCoordinates.
**     RFB 20200427          Simplified toMilliseconds() by using new EnvironmentOps.versionNumber
**                           to determine the current version.
**     SBI 20200529          Added removeGlobalImages.
**     SVL 20200417          Added functions related to enhanced browse configurations.
** 405 CA  20200722          getChanges() looks for explicitly registered dead frames, instead of iterating 
**                           the entire registry each time.
**     VVT 20200731          Added invokeInProcedure().
** 406 GES 20200627          Added getEventColumn()/getEventRow() for LAST-EVENT:COLUMN/ROW attributes.
**     HC  20200726          Initial implementation of SPREADSHEET widget and related changes.
**     VVT 20200731          Added invokeInProcedure().
**     AIL 20200821          Added suppressWarnings to the server state.
**     GES 20200821          Avoid context-local lookup in ErrorManager query.
**     VVT 20200910          Removed message(handle). See #4622-17 to #4622-19.
**     CA  20200923          Changed StateSynchronizer to return Externalizable instances.
**     HC  20200927          Batching of server config changes.
**     CA  20200927          Avoid context-local lookups by relying on the helper instance.
**     CA  20200930          Added a 'trigger invalidation counter' which will force LT.invokeOnServer to 
**                           compute and send trigger differences only if the trigger has altered the trigger 
**                           registry.
**     CA  20201003          Cache the result of the TriggerManager.mergeTriggerRegistry, until the trigger
**                           registry is modified - this allows to reuse the EventList instance.
**                           Delay the creation of the ScopeProcessor collections, until they are used.  This 
**                           will improve the performance of blocks which do not use frames.
**     CA  20201011          Replaced integer sets with a fast access bitmap.
**                           Improved getRowDisplayEvents.
**     HC  20201013          Server-pushed config field changes are serialized with field ids instead of
**                           names.
**     CA  20201015          Do not invalidate the cache if there were no events in the current scope.
**     CA  20201112          Avoid pushing the entire frame definition when setting the FRAME or PARENT 
**                           attribute.
**     CA  20201117          Added API to delete a widget at runtime.
**     SVL 20210112          Added disable(frame, int[], boolean).
**     EVL 20210126          Considering -1 event code as no-op for apply();
**     CA  20210217          SESSION:SYSTEM-ALERT-BOXES is false for GUI (must be tested from command line and
**                           not from procedure editor).
**     SVL 20210303          Process down configs in CHOOSE statement.
**     SVL 20210303          Added protection against negative event numbers in apply(long).
**     SVL 20210316          Track persistent/non-persistent trigger types.
**     SVL 20210316          Added isFocusedRowSelected().
**     VVT 20210317          displayError() fixed: now honors the ErrorManager.isSystemAlertBoxes()
**     SVL 20210321          Dropped unreachable code (WindowWidget is an instance of GenericWidget).
**     EVL 20210330          Fixed SKIP widget processing in alert boxes display.
**     CA  20210405          PROCESS-EVENTS must execute even if there is no active frame.
**     SVL 20210407          Added notifyBrowseLayoutPerformed.
**     SBI 20210414          Changed getResourceStreamFromApplication due to JarResource moved to Utils.
**     SVL 20210415          Browse not found error handled using MissingWidgetException.
**     CA  20210423          Do not post socket events (read-response/connect) if the socket resource has been
**                           deleted.
**     CA  20210505          If an event name can't be resolved to a code other than -1, APPLY is a no-op.
**     CA  20210505          A trigger can't be registered with an invalid resource.
**     EVL 20210512          Fix for incorrect message ignoring for key codes taken from String messages in
**                           APPLY statements without explicit target definition.  Renaming the worker method.
**     HC  20210514          Fixed stale screen buffer after READKEY.
**     VVT 20210601          Added: removeGlobalImages(). See #5282.
**     SBI 20210621          Implemented getKeyboardLayoutsSettings().
**     EVL 20210621          Adding method to activate popup menu on client side.
**     SBI 20210702          Changed getUploadFileSizeLimits() to return the file size threshold for writing data
**                           to the disk.
**     SVL 20210814          Added getColumnFiltersMappedDataTypes.
**     SBI 20210815          Added invokeWidgetCommand.
**     VVT 20210819          Fixed wrong parameter descriptions order in Javadoc in multiple
**                           registerTrigger() methods.
**     VVT 20210820          New API call introduced: requestNodeExpand()
**     VVT 20210820          Minor fixes for #5575-19.
**     VVT 20210820          Partly reverted the previous change. See #5575-26.
**     HC  20210902          Fixed default window lookup when CURRENT-WINDOW not set to the default one.
**     SVL 20211004          Dropped unnecessary parameter.
**     CA  20211008          Fixed DISPLAY UNLESS-HIDDEN - push the screen-buffer for the widgets.
**     RFB 20211008          Updates per review. Ref #5589.
**     AL2 20211011          Added findTrigger helper.
**         20211013          Handle client screen buffer at report generation.
**     CA  20211015          Fixed a regression in CA/20201003 - the logic was not preserved in a case for the
**                           frameViewed() code.
**     CA  20211018          For ChUI, a SKIP in a MESSAGE is converted to integer, and not '\n' (as ChUI has 
**                           special processing for SKIP).
**     VVT 20211018          New implementaton methods added: getWidth(), setWidth(), getHeight() and setHeight() to match
**                           the changes in the Sizeable interface. See: #5594-45.
**     CA  20211029          Review improvement for CA/20211018 in messageBox() code.
**     VVT 20211116          getWindowById() split into two (getWindowById() and lookupWindowById())
**                           to be able to lookup window without throwing exceptions. Unused and incorrect isWindow()
**                           method removed.
**     IAS 20211125          Added check for <code>null</code> in the targetWindowWorkerInt to avoid NPE.
**     OM  20211201          Improved transaction support for finally blocks.
**     IAS 20211202          Set SESSION:SYSTEM-ALERT-BOXES = true in GUI mode.
**     SVL 20211202          Added getBrowseColumnY.
**     VVT 20211203          Fixed: handleScreenBuffer: key modifiers mask now passed to KeyReader. See #5850.
**     VVT 20211208          LogicalTerminal.setImage() API and error handling fixed. See #5211.
**     CA  20211222          Fixed memory leak: the global finalizer (to prompt for a key) must not be 
**                           registered for appserver scopes.
**     SVL 20211223          Don't update default browse configuration when column movement is performed by
**                           user.
**     IAS 20211223          Special processing of errors raised by UDFs. 
**     VVT 20220119          LogicalTerminal.message() in SET mode: default values must be assigned instead
**                           of 'undefined' value in case no text was entered. See #5989.
**     CA  20220201          The numeric format is lazily pushed to the client via state sync, to avoid 
**                           unnecessary calls if the client-side is not interested of these. 
**     VVT 20220330          Minor improvements: local variable added for repeatable computed expression;
**                           incomplete Javadoc fixed; static method call made static;
**     VVT 20220401          keyboard state synchronization support added. Typo in Javadoc fixed.
**     AL2 20220509          Implement the getTempDirectory interface.
**     CA  20220515          The active session is now set for leaf sessions, too - this is required for a
**                           change in RemoteObject.obtainInstance, where first a check is done for an existing
**                           local proxy, and after that a network instance is obtained (a requirement for the
**                           server-side OS resources support, like memptr).
**     AL2 20220512          Replaced getTempDirectory with readHtmlBrowserWhitelist.
**     IAS 20220527          Added deleteFromControlSet operation
**     AL2 20220606          Added getHTMLBrowserLoadTimeout.
**     SVL 20220620          Changed readDefaultFonts() signature.
**     ME  20220722          Let deleteDynamicWidget call the client side even for 'special' widgets.
**     VVT 20220728          Javadocs minor addition.
**     TW  20220803          Error-status:error must be set for return error in method, while return error
**                           in old-style ABL function must remain the same. Refs: #6567.
**     TW  20220817          Error-status:error must be set for return error in static method.
**     CA  20220907          Batch the updateTabItemList information, to allow calls for multiple frames.
**                           If there is no trigger or WAIT-FOR for one of the U1..U10 dev events, then do not
**                           call the client-side.
**     CA  20220912          For editable frames, only the screen buffers which have changed are transferred
**                           to the client-side.
**     HC  20220915          Performance: Eliminated integer auto-boxing when enqueuing widget attributes.
**     CA  20220901          Refactored scope notification support: ScopeableFactory was removed, and the 
**                           registration is now specific to each type of scopeable.  For each case, the block
**                           will be registered for scope support (for that particular scopeable) only when
**                           the scopeable is 'active' (i.e. unnamed streams or accumulators are used).  This
**                           allows a lazy registration of scopeables, to avoid the unnecessary overhead of
**                           processing all the scopeables for each and every block.
**     CA  20220906          Moved the interactive client cleaner to LogicalTerminal.registerCleaner().
*/

/*
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU Affero General Public License as
** published by the Free Software Foundation, either version 3 of the
** License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Affero General Public License for more details.
**
** You may find a copy of the GNU Affero GPL version 3 at the following
** location: https://www.gnu.org/licenses/agpl-3.0.en.html
** 
** Additional terms under GNU Affero GPL version 3 section 7:
** 
**   Under Section 7 of the GNU Affero GPL version 3, the following additional
**   terms apply to the works covered under the License.  These additional terms
**   are non-permissive additional terms allowed under Section 7 of the GNU
**   Affero GPL version 3 and may not be removed by you.
** 
**   0. Attribution Requirement.
** 
**     You must preserve all legal notices or author attributions in the covered
**     work or Appropriate Legal Notices displayed by works containing the covered
**     work.  You may not remove from the covered work any author or developer
**     credit already included within the covered work.
** 
**   1. No License To Use Trademarks.
** 
**     This license does not grant any license or rights to use the trademarks
**     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
**     of Golden Code Development Corporation. You are not authorized to use the
**     name Golden Code, FWD, or the names of any author or contributor, for
**     publicity purposes without written authorization.
** 
**   2. No Misrepresentation of Affiliation.
** 
**     You may not represent yourself as Golden Code Development Corporation or FWD.
** 
**     You may not represent yourself for publicity purposes as associated with
**     Golden Code Development Corporation, FWD, or any author or contributor to
**     the covered work, without written authorization.
** 
**   3. No Misrepresentation of Source or Origin.
** 
**     You may not represent the covered work as solely your work.  All modified
**     versions of the covered work must be marked in a reasonable way to make it
**     clear that the modified work is not originating from Golden Code Development
**     Corporation or FWD.  All modified versions must contain the notices of
**     attribution required in this license.
*/

package com.goldencode.p2j.ui;

import java.io.*;
import java.net.*;
import java.util.*;
import java.util.function.*;
import java.util.logging.*;

import com.carrotsearch.hppc.*;
import org.roaringbitmap.*;
import org.roaringbitmap.IntConsumer;
import org.roaringbitmap.longlong.*;
import org.roaringbitmap.longlong.LongConsumer;

import com.goldencode.p2j.comauto.*;
import com.goldencode.p2j.directory.*;
import com.goldencode.p2j.extension.*;
import com.goldencode.p2j.jmx.*;
import com.goldencode.p2j.main.*;
import com.goldencode.p2j.net.*;
import com.goldencode.p2j.net.Session;
import com.goldencode.p2j.security.*;
import com.goldencode.p2j.ui.ColorTable.*;
import com.goldencode.p2j.ui.FieldGroup.*;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.util.*;
import com.goldencode.p2j.util.ErrorManager;
import com.goldencode.p2j.util.ErrorManager.*;
import com.goldencode.p2j.util.LogHelper;
import com.goldencode.p2j.util.SocketImpl;
import com.goldencode.p2j.util.Text;
import com.goldencode.util.*;
import com.google.common.collect.*;
import com.google.common.io.*;

/**
 * This class represents the state of the logical terminal associated with the
 * current user session. There is one instance of this class per session
 * context.
 */
public class LogicalTerminal
implements Initializable,
           Scopeable,
           StateSynchronizer
{
   /** Alert type MESSAGE */
   public static final int ALERT_MESSAGE = 0;
   
   /** Alert type QUESTION */
   public static final int ALERT_QUESTION = 1;
   
   /** Alert type INFO */
   public static final int ALERT_INFO = 2;
   
   /** Alert type ERROR */
   public static final int ALERT_ERROR = 3;
   
   /** Alert type WARNING */
   public static final int ALERT_WARNING = 4;
   
   /** Alert for MESSAGE with APPL-ALERT-BOXES == true */
   public static final int APPL_ALERT = 5;
   
   /** Alert as in batch mode (no dialogs) flag */
   public static final int BATCH_ALERT = 128;

   /** Alert type mask (clean BATCH_ALERT flag) */
   public static final int BATCH_ALERT_MASK = -1 ^ 128;

   /** Button set YES_NO */
   public static final int BTN_YES_NO = 0;
   
   /** Button set YES_NO_CANCEL */
   public static final int BTN_YES_NO_CANCEL = 1;
   
   /** Button set OK */
   public static final int BTN_OK = 2;
   
   /** Button set OK_CANCEL */
   public static final int BTN_OK_CANCEL = 3;
   
   /** Button set RETRY_CANCEL */
   public static final int BTN_RETRY_CANCEL = 4;
   
   /** logger */
   private static final Logger LOG = LogHelper.getLogger(LogicalTerminal.class.getName());

   /** Widget attributes flushes counter */
   private static final SimpleCounter SIMPLE_COUNTER =
      SimpleCounter.getInstance(FwdJMX.Counter.WidgetAttrFlushes);

   /** Widget attributes enqueues counter */
   private static final ValuesCounter VALUES_COUNTER =
      ValuesCounter.getInstance(FwdJMX.MapCounter.WidgetAttrEnqueues);

   /** Context-local instance of LogicalTerminal */
   private static final ContextLocal<LogicalTerminal> instance =
   new ContextLocal<LogicalTerminal>()
   {
      @Override
      public WeightFactor getWeight()
      {
         return WeightFactor.LEVEL_3;
      };
      
      @Override
      protected LogicalTerminal initialValue()
      {
         return (new LogicalTerminal());
      }
      
      @Override
      protected void cleanup(LogicalTerminal value)
      {
         // as the LT instance is being cleaned, remove its references.
         
         if (!SessionManager.get().isLeaf())
         {
            Session session = SessionManager.get().getSession();
            session.registerSynchronizer(null);
         }
         
         RemoteObject.registerLeftSynchronizer(null);
      };
   };
   
   /**
    * Listener is a browse in the process of getting new rows. If an event such as pause or
    * message box occurs in ROW-DISPLAY trigger, browse needs to be notified and redrawn with the
    * new rows.
    */
   private static final ContextLocal<RowDisplayDrawListener> rowDisplayDrawListener =
                                                                           new ContextLocal<>();
   
   /** Cache this exemplar to make array processing faster. */
   private static final ScreenDefinition[] EXEMPLAR = new ScreenDefinition[0];
   
   /** Application help program name (as hard coded in the 4GL). */
   private static final String APPLHELP = "applhelp.p";
   
   /** The coordinates conversion implementation. */
   private static CoordinatesConversion coordinates = 
      new CoordinatesConversion(new LogicalTerminalConversionParams());

   /** Current active frame */
   private GenericFrame frameRef = null;
   
   /** Flag identifying that HELP is active. */
   private boolean helpActive = false;
   
   /** The frame field targeted by FRAME-VALUE before HELP is invoked. */
   private int hlpFrameFieldId = -1;
   
   /** The frame targeted by FRAME-VALUE before HELP is invoked. */
   private GenericFrame hlpFrameRef = null;
   
   /** Current value of FRAME-VALUE attribute. */
   private String frameValue = "";
   
   /** Current value of FRAME-FIELD attribute. */
   private String frameField = "";
   
   /** Current value of FRAME-NAME built-in function. */
   private String frameName = "";
   
   /** Current value of FRAME-DB built-in function. */
   private String frameDb = "";
   
   /** Current value of FRAME-FILE built-in function. */
   private String frameFile = "";
   
   /** Current value of FRAME-INDEX attribute. */
   private integer frameIndex = new integer(0);
   
   /** Current ID of FRAME-FIELD */
   private int frameFieldId = -1;
   
   /** Current Logical Terminal MESSAGES color */
   private ColorSpec messageColor = new ColorSpec(ColorSpec.COLOR_MESSAGES);
   
   /** Current logical color used for VIEW-AS ALERT-BOX */
   private ColorSpec msgBoxColor = new ColorSpec(ColorSpec.COLOR_NORMAL);
   
   /** Default timeout for the waitFor() methods. */
   private double defaultTimeout = -1;
   
   /**
    * Flag indicating if milliseconds are honored, when seconds have a fractional part for PAUSE, 
    * READKEY, CHOOSE and WAIT-FOR statements. 
    */
   private Boolean useMillis = null;
   
   /** Client interface */
   private ClientExports client = null; 
   
   /** Registry for all frames for this user's session. */
   private Map<Integer, GenericFrame> frameRegistry = new HashMap<>();
   
   /** Set of dead frames since the last client-side trip.  Used by {@link #getChanges()}. */
   private RoaringBitmap deadFrameIds = new RoaringBitmap();
   
   /** Registry of the currently editable frames to get the screen buffers to operate. */
   private Map<GenericFrame, RoaringBitmap> editableFrameRegistry = new HashMap<>();
   
   /** Registry for all menus for this user's session. */
   private Map<Integer, MenuContainerWidget<?>> menuRegistry = new HashMap<>();
   
   /** Registry for all triggers for this user's session. */
   private Map<Integer, TriggerDefinition> triggerRegistry = new HashMap<>();
   
   /**
    * A counter for the trigger (de)registration.  It allows {@link #invokeOnServer} to not spend time 
    * computing differences if the trigger has not altered the {@link #triggerRegistry}.
    */
   private long triggerChangeCount = 0;
   
   /** 
    * A cache of the {@link TriggerManager#mergeTriggerRegistry}.  It gets invalidated whenever the 
    * {@link #scope} gets changed.
    */
   private EventList cachedEventList = null;
   
   /** scoped dictionary for all possible targets */
   @SuppressWarnings("rawtypes")
   private ScopedDictionary scope = new ScopedDictionary(new LinkedList());
   
   /** A set containing the IDs of all ON .. PERSISTENT RUN triggers */
   private RoaringBitmap persistentTriggers = new RoaringBitmap();
   
   /** CURRENT-WINDOW widget */
   private WindowWidget currentWindow = null;
   
   /** The windows stack defined by CURRENT-WINDOW assingments */
   private LinkedList<WindowWidget> currentWindowStack = new LinkedList<>();
   
   /** ACTIVE-WINDOW widget. Can be set only internally */
   private WindowWidget activeWindow = null;
   
   /** DEFAULT-WINDOW widget. Set only once when Progress starts. */
   private WindowWidget defaultWindow = null;
   
   /** 
    * Holds the state of the SESSION:DISABLE-REDRAW attribute. 
    * This attribute is a FWD extension, not real 4GL!
    */
   private boolean disableRedraw = false;
   
   /** The window that is currently disabled for redrawing. */ 
   private WindowWidget locked = null;
   
   /** active triggers widgets/procedures stack */
   private LinkedList<WrappedResource> triggerResource = new LinkedList<>();
   
   /** active triggers events stack (synced to widgets stack) */
   private LinkedList<Integer> triggerEvent = new LinkedList<>();

   /** active triggers matched widgets stack (synced to widgets stack) */
   private LinkedList<Integer> triggerMatchedWidgets = new LinkedList<>();

   /** Active triggers persistent/non-persistent type stack. */
   private LinkedList<Boolean> triggerPersistent = new LinkedList<>();

   /** Cached number of screen lines. */
   private integer screenLines = null;
   
   /** Tracks data that is associated with specific blocks. */
   private Deque<Integer> blocks = new ArrayDeque<>();
   
   /** Tracks screen definitions that have yet to be pushed to the client. */
   private PartitionedArray<ScreenDefinition> definitions = new PartitionedArray<>();
   
   /** Tracks menu descriptions that have yet to be pushed to the client. */
   private PartitionedArray<MenuDescription> descriptions = new PartitionedArray<>();
   
   /** Temporary storage for screen definitions. */
   private ArrayList<ScreenDefinition> level2definitions = new ArrayList<>();
   
   /** Temporary storage for screen definitions. */
   private ArrayList<MenuDescription> level2descriptions = new ArrayList<>();
   
   /** 
    * Temporary collect widget IDs for menus whose description were not forcefully pushed to
    * client. It is only used  for cleanup in the event that the external procedure in which they
    * are defined fails to instantiate properly. If the new procedure scope starts successfully,
    * this collection is discarded without other processing, the menus will be cleaned up 
    * naturally, at the end of the scope.
    */
   private RoaringBitmap pendingMenus = new RoaringBitmap();
   
   /** The code of last event function (LAST-EVENT:FUNCTION) */
   private int functionKey;
   
   /** The value of the LAST-EVENT:X attribute. */
   private Integer eventX = null;
   
   /** The value of the LAST-EVENT:Y attribute. */
   private Integer eventY = null;
   
   /** Scope chain. */
   private ScopeProcessor scopeTrack = null;
   
   /** State to synchronize on the next trip to the client. */
   private ServerState state = new ServerState();
   
   /** Flag that reflects relogin procedure. */
   private boolean relogin = false;
   
   /** Flag for the executing trigger's <code>NO-APPLY</code> status. */
   private boolean consumeEvent = false;
   
   /** Current external procedures call nesting. */
   private int nesting = 0;
   
   /** Timestamp for last readKey operation. */
   private long millis = 0;
   
   /** Cached terminal type value **/
   private String terminalType = null;
   
   /** Window widgets dynamic created */
   private Map<Integer, WindowWidget> windowRegistry = new LinkedHashMap<>();
   
   /**
    * All widgets map for given class instance except default window including dynamic and static.
    */
   private Map<Integer, Object> widgetRegistry = new HashMap<>();
   
   /** Mutex semaphore to serialize data access. */
   private Object registryLock = new Object();
   
   /** Cached value of <code>SESSION:WINDOW-SYSTEM</code>. */
   private String windowSystem = null;
   
   /**
    * Data which are meaningful only within the trigger
    * Introduced for the WIDGET-LEAVE/WIDGET-ENTRY support
    */
   private EventInfo eventInfo = null;
   
   /** WAIT STATE name */
   private String waitState = "";
   
   /** The state of SESSION:IMMEDIATE-DISPLAY attribute */
   private boolean isImmediateDisplay = false;
   
   /** The state of SESSION:DATA-ENTRY-RETURN attribute */
   private Boolean isDataEntryReturn = null;
   
   /** The state of SESSION:APPL-ALERT-BOXES attribute */
   private boolean isUsingAlertBoxes = false;
   
   /** Flag indicating if the client is running in embedded mode or not. */
   private Boolean embedded = null;
   
   /** Flag tracking the SESSION:V6DISPLAY attribute.  No side-effects on the UI at this time. */
   private boolean v6display = false;

   /** Client metrics */
   private ClientMetrics clientMetrics;
   
   /** Helper to use the ProcedureManager without any context local lookups. */
   private ProcedureManager.ProcedureHelper pm;
   
   /** Helper to use the TM without any context local lookups. */
   private TransactionManager.TransactionHelper tm;
   
   /** The client local storage proxy */
   private ObjectStorage clientLocalStorage;

   /** Enqueued widget attribute ids */
   private IntArrayList widgetAttrIds = new IntArrayList();

   /** Enqueued widget attribute values */
   private List<Object> widgetAttrValues = new ArrayList<>();

   /** The current client numeric format, set when the server-side first wants to change it. */
   private String currentClientNumericFormat = null;
   
   /** The latest client numeric format which needs to be pushed to the client via {@link ServerState}. */
   private String newClientNumericFormat = null;

   /** Pressed keys */
   private int[] keyboardState;
   
   /** The stack of WAIT-FOR termination events. */
   private Deque<EventList> waitStack = new ArrayDeque<>();
   
  /**
    * Constructor.
    */
   private LogicalTerminal()
   {
   }

   /**
    * Returns the context instance.
    *
    * @return  See above.
    */
   public static LogicalTerminal getInstance()
   {
      return locate();
   }
   
   @SuppressWarnings({ "unchecked", "rawtypes" })
   @Override
   public void init()
   {
      pm = ProcedureManager.getProcedureHelper();
      tm = TransactionManager.getTransactionHelper();

      // initialize the conversation stack for this context
      Session session = SessionManager.get().getSession();
      
      // resolve client exports, state sync and CTRL-C handling
      if (!SessionManager.get().isLeaf())
      {
         client = (ClientExports) RemoteObject.obtainNetworkInstance(ClientExports.class);
   
         session.registerSynchronizer(this);
         RemoteObject.registerLeftSynchronizer(this);
         Control.init(session);
      }
      else
      {
         // server-side code is left-side by convention
         
         // MUST do this first!
         RemoteObject.registerLeftSynchronizer(this);
         
         client = (ClientExports) RemoteObject.obtainLocalInstance(ClientExports.class, true);
      }
      
      clientLocalStorage = new ClientStorageProxy(client);

      // create the main window BUT DO NOT PUSH IT since a duplicate instance
      // already exists on the client side with the same exact values
      currentWindow = new WindowWidget(false);
      if (session != null)
      {
         currentWindow.config.visible = isChui();
      }
      
      activeWindow = defaultWindow = currentWindow;
      currentWindow.setId(WidgetId._DEFAULT_WINDOW_ID);
      
      // let ProcedureManager know the DEFAULT-WINDOW in order to prevent releasing/deleting it.
      ProcedureManager.deregisterPendingResource(defaultWindow);
      
      // register the server side widget ID helper
      WidgetIdHelper.initWorker(new ServerIdHelper());

      // register external COMs
      WidgetDescriptorHelper.getDescriptors().stream().forEach(d ->
      {
         String comName = d.getCOMName();
         Class comClass = d.getCOMClass();

         if (comName != null && comClass != null)
         {
            ComServer.registerComObject(comName, comClass);
            LOG.log(Level.INFO,
                    "Registered COM class '" + comName + "' [" + comClass + "].");
         }
      });
      
      if (session != null && session.isRunning())
      {
         // LT can be used during server start too (i.e. by MetadataManager.populateDatabase)
         // so ensure we have a proper client session
         ConfigManager mgr = ConfigManager.getInstance();
         
         if (mgr.getActiveConfig(defaultWindow.config().id) == null)
         {
            // TODO: when the context is reset (in an appserver agent), what happens with the 
            // client-side? do the widgets gets destroyed too, and the window's configuration 
            // reinitialized? (i.e. it needs to be pushed by the client-side again?).
            
            // register only once
            mgr.addWidgetConfig(defaultWindow.config());

            ConfigSyncManager.markScopeStart();
            try
            {
               // the window was registered, activate the config updates
               client.activateConfigUpdates();
               defaultWindow.config().realized = true;
            }
            finally
            {
               ConfigSyncManager.markScopeEnd();
            }
         }
         
         if (!client.isInBatchMode())
         {
            // initialize client side PROPATH value 
            FileSystemOps.init();
            // setting up default application icon on the client side
            if (!isChui())
            {
               // set default window icon on the client side
               setWindowDefaultIcon();
               ErrorManager.setSystemAlertBoxes(true);
            }

            // push the default window initial state
            defaultWindow.pushScreenDefinition();
         }
      }
   }

   /**
    * Set the value of the eventIfo
    * to be used before the trigger activation
    * 
    * @param eventInfo
    *             the new eventInfo value 
    */
   private void setEventInfo(EventInfo eventInfo)
   {
      this.eventInfo = eventInfo;
   }
   
   /**
    * Clear the value of the eventInfo
    * to be used after trigger activation
    * (typically in the <code>finally</code> block)
    */
   private void clearEventInfo()
   {
      this.eventInfo = null;
   }
   
   /**
    * Get the value of the event info
    * 
    * @return the value of the eventInfo
    */
   public static EventInfo getEventInfo()
   {
      return locate().eventInfo;
   }
 
   /**
    * Pushes the window configuration instance down to the client.
    * 
    * @param   window
    *          instance of <code>WindowConfig</code>
    */
   public static void pushWindow(WindowConfig window)
   {
      locate().client.pushWindow(window);
   }
   
   /**
    * Get default timeout used by this session for processing <b>WAIT-FOR </b>
    * requests when no explicit timeout is specified.
    * 
    * @return  Default timeout used for <b>WAIT-FOR </b>.
    */
   public double getDefaultTimeout()
   {
      return defaultTimeout;
   }

   /**
    * Set default timeout used by this session for processing <b>WAIT-FOR </b>
    * requests when no explicit timeout is specified.
    * 
    * @param   timeout
    *          New timeout value.
    */
   public void setDefaultTimeout(double timeout)
   {
      defaultTimeout = timeout;
   }

   /**
    * Register a cleaner for interactive sessions, where a pause is shown before terminating the client.
    */
   public static void registerCleaner()
   {
      if (!AppServerManager.isRemote() && !TransactionManager._isHeadless())
      {
         Finalizable cleaner = new Finalizable()
         {
            @Override
            public void finished()
            {
               LogicalTerminal.getInstance().pauseBeforeEnd(TransactionManager.isProcessingQuit());
            }
            
            @Override
            public void iterate()
            {
               // nothing to do here 
            }
            
            @Override
            public void retry()
            {
               // nothing to do here 
            }
            
            @Override
            public void deleted()
            {
               // nothing to do here
            }
         };
            
         TransactionManager.registerFinalizable(cleaner, true);
      }
   }
   
   /**
    * Rings a bell on the terminal.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>BELL </b>
    */
   public static void bell()
   {
      locate().client.bell();
   }
   
   /**
    * Convert the given (fractional) seconds to milliseconds.
    * <p>
    * This uses the current version as returned by {@link EnvironmentOps#getVersion()} to determine
    * if the fractional seconds are used(when version is after 11.3) or ignored (otherwise).
    *  
    * @param    seconds
    *           The (fractional) seconds to convert.
    *           
    * @return   The milliseconds.
    */
   public static int toMilliseconds(double seconds)
   {
      if (Double.compare(seconds, 0) < 0)
      {
         // any negative value means 'no wait'
         return -1;
      }
      
      LogicalTerminal lt = locate();
      if (lt.useMillis == null)
      {
         lt.useMillis = EnvironmentOps.versionNumber() >= 11.3f;
      }
      
      int millis = lt.useMillis ? (int) Math.round(seconds * 1000.0) : ((int) seconds) * 1000;
      
      return millis;
   }

   /**
    * Hides all widgets and messages displayed in the message area. If pausing
    * is in effect, the standard message "Press space bar to continue" is
    * displayed on the status line and the user is required to press a key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>HIDE ALL [NO_PAUSE] </b>
    * 
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding
    *          messages
    */
   public static void hideAll(boolean noPause)
   {
      hideAll(currentWindow(), noPause);
   }

   /**
    * Hides all widgets and messages displayed in the message area, in the given stream. 
    * If pausing is in effect, the standard message "Press space bar to continue" is
    * displayed on the status line and the user is required to press a key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>HIDE STREAM stream ALL [NO_PAUSE] IN WINDOW hWin </b>
    * 
    * @param   hWin
    *          Target window handle or null for default window.
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding
    *          messages
    */
   public static void hideAll(handle hWin, boolean noPause)
   {
      if (EnvironmentOps.isBatchMode().booleanValue())
      {
         // in batch mode, hide is NO-OP (hWin is invalid/unknown)
         return;
      }
      
      LogicalTerminal lt = locate();
      lt.forceHiding();
      lt.client.hideAll(noPause, targetWindowWorker(hWin, false));
      
      // TODO get frames from target window?
      
      // force frame deletion
      Set<GenericFrame> frames = new HashSet<>(lt.frameRegistry.values());
      for (GenericFrame frame : frames)
      {
         frame.frameDelete();
      }
   }

   /**
    * Hides all widgets and messages displayed in the message area, in the given stream. 
    * If pausing is in effect, the standard message "Press space bar to continue" is
    * displayed on the status line and the user is required to press a key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>HIDE STREAM stream ALL [NO_PAUSE] </b>
    * 
    * @param   stream
    *          The stream for which the frames need to be hidden.
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding
    *          messages
    */
   public static void hideAll(Stream stream, boolean noPause)
   {
      // TODO: implement this.
   }

   /**
    * Reset buffer size manager to its initial state.
    */
   public static void resetBufferSizeManager()
   {
      locate().client.resetBufferSizeManager();
   }
   
   /**
    * Stores the value in a field of the active frame, which is in focus and
    * redisplays the field.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>FRAME-VALUE = expression </b>
    * 
    * @param   value
    *          the value to store in the frame field
    */
   public static void setFrameValue(int value)
   {
      setFrameValue(new integer(value));
   }
   
   /**
    * Stores the value in a field of the active frame, which is in focus and
    * redisplays the field.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>FRAME-VALUE = expression </b>
    * 
    * @param   value
    *          the value to store in the frame field
    */
   public static void setFrameValue(double value)
   {
      setFrameValue(new decimal(value));
   }
   
   /**
    * Stores the value in a field of the active frame, which is in focus and
    * redisplays the field.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>FRAME-VALUE = expression </b>
    * 
    * @param   value
    *          the value to store in the frame field
    */
   public static void setFrameValue(String value)
   {
      setFrameValue(new character(value));
   }
   
   /**
    * Stores the value in a field of the active frame, which is in focus and
    * redisplays the field.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>FRAME-VALUE = expression </b>
    * 
    * @param   value
    *          the value to store in the frame field
    */
   public static void setFrameValue(boolean value)
   {
      setFrameValue(new logical(value));
   }
   
   /**
    * Stores the value in a field of the active frame, which is in focus and
    * redisplays the field.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>FRAME-VALUE = expression </b>
    * 
    * @param   value
    *          the value to store in the frame field
    */
   public static void setFrameValue(BaseDataType value)
   {
      LogicalTerminal lt = locate();
      boolean hlpError = false;
      
      if (lt.helpActive)
      {
         if (lt.hlpFrameRef != null && lt.hlpFrameFieldId != -1)
         {
            // if help is active, use the backed up frame!
            lt.hlpFrameRef.setFrameValue(lt.hlpFrameFieldId, value);
            
            return;
         }
         hlpError = true;
      }
      
      if (hlpError || lt.frameRef == null || lt.frameFieldId == -1)
      {
         ErrorManager.recordOrShowError(674, 
                                        "Frame-value assignment attempted with no frame active", 
                                        false,
                                        true,
                                        false);
         return;
      }
      
      lt.frameRef.setFrameValue(lt.frameFieldId, value);
   }

   /**
    * Gets the value from a field of the active frame, which is in focus.
    * <p>
    * This method corresponds to this Progress built-in function <br>
    * <b>FRAME-VALUE </b>
    * 
    * @return  the value from the frame field in focus
    */
   public static character getFrameValue()
   {
      return new character(locate().frameValue);
   }

   /**
    * Refreshes the specified frame and widget with a new screen value.
    * 
    * @param   frame
    *          frame to refresh
    * @param   wid
    *          widget to refresh
    * @param   value
    *          new widget value
    */
   public static void refreshFrameWidget(GenericFrame frame, int wid, BaseDataType value)
   {
      refreshFrameWidgets(frame, new int[] { wid }, new BaseDataType[] { value });
   }

   /**
    * Refreshes the specified frame and widgets with a new screen value.
    * 
    * @param   frame
    *          frame to refresh
    * @param   wids
    *          widgets to refresh
    * @param   values
    *          new values for the widgets
    */
   public static void refreshFrameWidgets(GenericFrame frame, int[] wids, BaseDataType[] values)
   {
      LogicalTerminal lt = locate();

      String val = lt.client.refreshFrameWidgets(frame.getFrameId(), wids, values);
      
      // update the frame-value as well if the widget is in focus
      for (int wid : wids)
      {
         if (lt.frameFieldId == wid)
         {
            lt.frameValue = val;
            break;
         }
      }
   }

   /**
    * Hides all messages displayed in the message area. If pausing is in
    * effect, the standard message "Press space bar to continue" is displayed
    * on the status line and the user is required to press a key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>HIDE MESSAGE [NO_PAUSE] </b>
    * 
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding
    *          messages
    */
   public static void hideMessage(boolean noPause)
   {
      locate().client.hideMessage(noPause, WindowConfig.RESOLVE_WINDOW);
   }

   /**
    * Hides all messages displayed in the message area. If pausing is in
    * effect, the standard message "Press space bar to continue" is displayed
    * on the status line and the user is required to press a key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>HIDE MESSAGE [NO_PAUSE] </b>
    * 
    * @param   stream
    *          The stream for which the frames need to be hidden.
    * @param   hWin
    *          Target window handle.
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding
    *          messages
    */
   public static void hideMessage(Stream stream, handle hWin, boolean noPause)
   {
      // TODO implement
   }

   /**
    * Hides all messages displayed in the message area. If pausing is in
    * effect, the standard message "Press space bar to continue" is displayed
    * on the status line and the user is required to press a key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>HIDE MESSAGE [NO_PAUSE] </b>
    * 
    * @param   hWin
    *          Target window handle.
    * @param   noPause
    *          <code>true</code> if no pause is required before hiding
    *          messages
    */
   public static void hideMessage(handle hWin, boolean noPause)
   {
      locate().client.hideMessage(noPause, targetWindowWorker(hWin, false));
   }
   
   /**
    * Clears the type-ahead buffer if the input source is the keyboard.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>INPUT CLEAR </b>
    */
   public static void inputClear()
   {
      LogicalTerminal lt = locate();
      
      // calling inputClear immediately after readKey which returned no 
      // events has no sense and can be safely ignored
      if (System.currentTimeMillis() - lt.millis < 10)
         return;
      
      lt.client.inputClear();
   }

   /**
    * Displays the empty string in the message line of the client.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE </b>
    */
   public static void message()
   {
      message(new Object[] {""}, (handle)null);
   }

   /**
    * Displays a standard message.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    */
   public static void message(Object text)
   {
      message(new Object[] {text}, (handle)null);
   }

   /**
    * Displays a standard message.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object text, handle hWin)
   {
      message(new Object[] {text}, hWin);
   }
   
   /**
    * Displays a message.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE </b>
    * 
    * @param   list
    *          list of message text components
    */
   public static void message(Object[] list)
   {
      message(list, false, null, false, (handle)null);
   }

   /**
    * Displays a message.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE </b>
    * 
    * @param   list
    *          list of message text components
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object[] list, handle hWin)
   {
      message(list, false, null, false, hWin);
   }
   
   /**
    * Displays a standard message and sets or updates a character variable.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    */
   public static void message(Object       text,
                              boolean      set, 
                              Accessor     var) 
   {
      message(new Object[] {text}, set, var, (handle)null);
   }

   /**
    * Displays a standard message and sets or updates a character variable.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object       text,
                              boolean      set, 
                              Accessor     var,
                              handle       hWin) 
   {
      message(new Object[] {text}, set, var, hWin);
   }
   
   /**
    * Displays a message and sets or updates a character variable.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    */
   public static void message(Object[]     list,
                              boolean      set, 
                              Accessor     var) 
   {
      message(list, set, var, false, (handle)null);
   }

   /**
    * Displays a message and sets or updates a character variable.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object[]     list,
                              boolean      set, 
                              Accessor     var,
                              handle       hWin) 
   {
      message(list, set, var, false, hWin);
   }

   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    */
   public static void message(Object       text, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto)
   {
      message(new Object[] {text}, set, var, auto, (handle)null);
   }

   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object       text, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto,
                              handle       hWin)
   {
      message(new Object[] {text}, set, var, auto, hWin);
   }
   
   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    */
   public static void message(Object[]     list,
                              boolean      set, 
                              Accessor     var,
                              boolean      auto)
   {
      message(list, set, var, auto, null, null, null);
   }

   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object[]     list,
                              boolean      set, 
                              Accessor     var,
                              boolean      auto,
                              handle       hWin)
   {
      message(list, set, var, auto, null, null, hWin);
   }
   
   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable FORMAT </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   format
    *          temporary FillIn format specification
    */
   public static void message(Object       text, 
                              boolean      set,
                              Accessor     var,
                              String       format)
   {
      message(new Object[] {text}, set, var, format, (handle)null);
   }

   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable FORMAT </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   format
    *          temporary FillIn format specification
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object       text, 
                              boolean      set,
                              Accessor     var,
                              String       format,
                              handle       hWin)
   {
      message(new Object[] {text}, set, var, format, hWin);
   }
   
   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable FORMAT </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   format
    *          temporary FillIn format specification
    */
   public static void message(Object[]     list,
                              boolean      set, 
                              Accessor     var,
                              String       format)
   {
      message(list, set, var, false, format, (ColorSpec)null, null);
   }

   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable FORMAT </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   format
    *          temporary FillIn format specification
    * @param   hWin
    *          handle of target window.        
    */
   public static void message(Object[]     list,
                              boolean      set, 
                              Accessor     var,
                              String       format,
                              handle       hWin)
   {
      message(list, set, var, false, format, (ColorSpec)null, hWin);
   }
   
   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   format
    *          temporary FillIn format specification
    */
   public static void message(Object       text, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto,
                              String       format)
   {
      message(new Object[] {text}, set, var, auto, format, (ColorSpec)null, null);
   }

   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   format
    *          temporary FillIn format specification
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object       text, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto,
                              String       format, 
                              handle       hWin)
   {
      message(new Object[] {text}, set, var, auto, format, (ColorSpec)null, hWin);
   }
   
   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    * 
    * @param   list
    *          Array of message text components either as
    *          <code>character</code> or <code>String</code> instances.
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   format
    *          temporary FillIn format specification
    */
   public static void message(Object[]     list, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto,
                              String       format)
   {
      message(list, set, var, auto, format, null, null);
   }

   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    * 
    * @param   list
    *          Array of message text components either as
    *          <code>character</code> or <code>String</code> instances.
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   format
    *          temporary FillIn format specification
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object[]     list, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto,
                              String       format, 
                              handle       hWin)
   {
      message(list, set, var, auto, format, null, hWin);
   }
   
   /**
    * Displays a standard message.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   cs
    *          color specification.
    */
   public static void message(Object text, ColorSpec cs)
   {
      message(new Object[] {text}, cs, null);
   }

   /**
    * Displays a standard message.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   cs
    *          color specification.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object text, ColorSpec cs, handle hWin)
   {
      message(new Object[] {text}, cs, hWin);
   }
   
   /**
    * Displays a message.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE </b>
    * 
    * @param   list
    *          list of message text components
    * @param   cs
    *          color specification.
    */
   public static void message(Object[] list, ColorSpec cs)
   {
      message(list, false, null, false, cs, null);
   }

   /**
    * Displays a message.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE </b>
    * 
    * @param   list
    *          list of message text components
    * @param   cs
    *          color specification.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object[] list, ColorSpec cs, handle hWin)
   {
      message(list, false, null, false, cs, hWin);
   }
   
   /**
    * Displays a standard message and sets or updates a character variable.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   cs
    *          color specification.
    */
   public static void message(Object       text,
                              boolean      set, 
                              Accessor     var,
                              ColorSpec    cs) 
   {
      message(new Object[] {text}, set, var, cs, null);
   }

   /**
    * Displays a standard message and sets or updates a character variable.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   cs
    *          color specification.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object       text,
                              boolean      set, 
                              Accessor     var,
                              ColorSpec    cs, 
                              handle       hWin) 
   {
      message(new Object[] {text}, set, var, cs, hWin);
   }
   
   /**
    * Displays a message and sets or updates a character variable.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   cs
    *          color specification.
    */
   public static void message(Object[]     list,
                              boolean      set, 
                              Accessor     var,
                              ColorSpec    cs) 
   {
      message(list, set, var, false, cs, null);
   }

   /**
    * Displays a message and sets or updates a character variable.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   cs
    *          color specification.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object[]     list,
                              boolean      set, 
                              Accessor     var,
                              ColorSpec    cs,
                              handle hWin) 
   {
      message(list, set, var, false, cs, hWin);
   }
   
   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   cs
    *          color specification.
    */
   public static void message(Object       text, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto,
                              ColorSpec    cs)
   {
      message(new Object[] {text}, set, var, auto, cs, null);
   }

   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   cs
    *          color specification.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object       text, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto,
                              ColorSpec    cs,
                              handle       hWin)
   {
      message(new Object[] {text}, set, var, auto, cs, hWin);
   }
   
   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   cs
    *          color specification.
    */
   public static void message(Object[]     list,
                              boolean      set, 
                              Accessor     var,
                              boolean      auto,
                              ColorSpec    cs)
   {
      message(list, set, var, auto, null, cs, null);
   }

   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   cs
    *          color specification.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object[]     list,
                              boolean      set, 
                              Accessor     var,
                              boolean      auto,
                              ColorSpec    cs,
                              handle       hWin)
   {
      message(list, set, var, auto, null, cs, hWin);
   }
   
   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable FORMAT </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   format
    *          temporary FillIn format specification
    * @param   cs
    *          color specification.
    */
   public static void message(Object       text,
                              boolean      set, 
                              Accessor     var,
                              String       format,
                              ColorSpec    cs)
   {
      message(new Object[] {text}, set, var, format, cs, null);
   }

   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable FORMAT </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   format
    *          temporary FillIn format specification
    * @param   cs
    *          color specification.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object       text,
                              boolean      set, 
                              Accessor     var,
                              String       format,
                              ColorSpec    cs,
                              handle       hWin)
   {
      message(new Object[] {text}, set, var, format, cs, hWin);
   }
   
   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable FORMAT </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   format
    *          temporary FillIn format specification
    * @param   cs
    *          color specification.
    */
   public static void message(Object[]     list, 
                              boolean      set,
                              Accessor     var,
                              String       format,
                              ColorSpec    cs)
   {
      message(list, set, var, false, format, cs, null);
   }

   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable FORMAT </b>
    * 
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   format
    *          temporary FillIn format specification
    * @param   cs
    *          color specification.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object[]     list, 
                              boolean      set,
                              Accessor     var,
                              String       format,
                              ColorSpec    cs,
                              handle       hWin)
   {
      message(list, set, var, false, format, cs, hWin);
   }
   
   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   format
    *          temporary FillIn format specification
    * @param   cs
    *          color specification.
    */
   public static void message(Object       text, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto, 
                              String       format,
                              ColorSpec    cs)
   {
      message(new Object[] {text}, set, var, auto, format, cs);
   }
   
   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   format
    *          temporary FillIn format specification
    * @param   cs
    *          color specification.
    * @param   hWin
    *          handle of target window.
    */
   public static void message(Object       text, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto, 
                              String       format,
                              ColorSpec    cs,
                              handle       hWin)
   {
      message(new Object[] {text}, set, var, auto, format, cs, hWin);
   }
   
   /**
    * Displays a standard message and sets or updates a character variable
    * using a specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    * 
    * @param   text
    *          message text either as <code>character</code> or
    *          <code>String</code>
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param   var
    *          Variable to set or update.  Must not be <code>null</code>.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   format
    *          temporary FillIn format specification
    * @param   cs
    *          color specification.
    */
   public static void message(Object[]       text, 
                              boolean      set,
                              Accessor     var,
                              boolean      auto, 
                              String       format,
                              ColorSpec    cs)
   {
      message(text, set, var, auto, format, cs, null);
   }
   
   /**
    * Displays a message and sets or updates a character variable using a
    * specified format for the temporary FillIn widget.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>MESSAGE SET/UPDATE variable AUTO-RETURN FORMAT </b>
    * @param   list
    *          list of message text components
    * @param   set
    *          <code>true</code> for SET, <code>false</code> for UPDATE,
    *          not used if var parameter is <code>null</code>.
    * @param   var
    *          Variable to set or update.  If <code>null</code>, this is a
    *          regular MESSAGE without SET or UPDATE.
    * @param   auto
    *          <code>true</code> if AUTO-RETURN option is specified.
    * @param   format
    *          temporary FillIn format specification
    * @param   cs 
    *          color specification
    * @param   hWin
    *          Handle of the target window
    */
   public static void message(Object[]     list,
                              boolean      set,
                              Accessor     var,
                              boolean      auto,
                              String       format, 
                              ColorSpec    cs,
                              handle       hWin)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      if (lt.isUsingAlertBoxes && var == null)
      {
         messageBox(list, set, var, APPL_ALERT, BTN_OK, "Message", cs, hWin);
         return;
      }
      
      if (var != null)
      {
         // disable loop protection and increment interactions counter
         lt.markInteractive();
      }
      
      Color clr = color(cs, lt.messageColor);
      
      Stream in  = UnnamedStreams.input();
      Stream out = UnnamedStreams.output();
      
      if (in != null)
      {
         // possible stream reading mode
         if (var != null)
         {
            in.resetCurrentLine();
            
            try
            {
               // read from the stream
               if (!GenericFrame.getField(in, var, null, null, format, true))
               {
                  // failure in silent error mode
                  return;
               }
            }
            
            finally
            {
               in.resetCurrentLine();
            }
            
            if (out == null)
            {
               // output is not redirected, we are done
               return;
            }
         }
      }
      // TODO: if the input is read from a device like /dev/null, then the  
      // message will not be displayed to the terminal. if the terminal is 
      // also redirected, then the message is output only to the stream.
      
      // normal processing
      BaseDataType bdt = (var == null) ? null : var.get();
      
      if (set && bdt != null)
      {
         bdt.setUnknown();
      }
      
      StringBuilder sb = new StringBuilder();
      
      // generate our text from the input list, process each element in the 
      // array (each one represents the result of a standalone expression)
      for (int i = 0; i < list.length; i++)
      {
         // put a spacer in between separate expressions
         if (i > 0)
            sb.append(" ");
         
         // handle each type differently
         if (list[i] instanceof java.lang.String)
         {
            // no conversion needed
            sb.append((String) list[i]);
         }
         else if (list[i] instanceof BaseDataType)
         {
            // a BDT has a special data type aware message format
            sb.append(((BaseDataType)list[i]).toStringMessage());
         }
         else if (list[i] instanceof SkipEntity)
         {
            // just another way to embed newlines; this is an undocumented feature which is
            // only supposed to work with VIEW-AS ALERT-BOX, but it does work in the 4GL
            SkipEntity skip = (SkipEntity) list[i];
            StringHelper.repeatChar('\n', skip.getSkip() + 1, sb); 
         }
         else
         {
            // handle it as an Object
            sb.append(list[i].toString());
         }
      }
      
      // OK, now we have our text to display
      String text = sb.toString();
      
      // check for the need to handle the special nested mode quirk where
      // during any pause that is caused by this message, the END-ERROR key
      // will ALWAYS generate ERROR not ENDKEY
      boolean nestedMode = lt.tm.isNestedMode();
      int     oldint     = lt.state.interactions;
      
      if (nestedMode)
      {
         // force END-ERROR to always be ERROR
         lt.state.interactions = 2;
      }
      
      try
      {
         MessageReturnValue mrv = lt.client.message(text,
                                                    set,
                                                    bdt,
                                                    auto,
                                                    format,
                                                    clr,
                                                    targetWindowWorker(hWin, false));
         
         if (mrv != null)
         {
            // update the caller's var with the result of the edit operation
            if (var != null)
            {
               final BaseDataType retVar = mrv.var;
               if (set && retVar.isUnknown())
               {
                  if(retVar instanceof Text)
                  {
                     retVar.assign("");
                  }
                  else if(retVar instanceof NumberType)
                  {
                     retVar.assign(0);
                  }
                  else if(retVar instanceof logical)
                  {
                     retVar.assign(false);
                  }
               }
               
               var.set(retVar);      
            }
            
            // maintain lastkey if a key was pressed during pause or any edit
            if (mrv.keycode != -1)
            {
               KeyReader.setLastKey(mrv.keycode);      
            }
         }
      }
      
      catch (ConditionException ce)
      {
         // make sure that the lastkey is properly maintained
         int kc = ce.getKeyCode();
         
         // anything other than -1 is a valid key press AND even -1 is a
         // valid indicator of CTRL-C if this was a STOP 
         if (kc != -1 || ce instanceof StopConditionException)
         {
            KeyReader.setLastKey(kc);
         }
         
         // continue the unwinding of the stack
         throw ce;
      }
      
      finally
      {
         if (nestedMode)
         {
            // reset to saved value
            lt.state.interactions = oldint;
         }      
      }
   }

   /**
    * Checks if the client is running in batch mode.
    * 
    * @return  <code>true</code> for batch mode, <code>false</code> otherwise.
    * 
    */
   public static boolean isInBatchMode()
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      return (lt != null && lt.client != null) ? lt.client.isInBatchMode() : false;
   }

   /**
    * Checks if the client is running in batch background mode.
    * 
    * @return  <code>true</code> for background mode, <code>false</code> otherwise.
    * 
    */
   public static boolean isBatchInBackground()
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      return (lt != null && lt.client != null) ? lt.client.isBatchInBackground() : false;
   }

   /**
    * Stops the associated client session by generating UnstoppableExitException.
    */
   public static void stopBatchSession()
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      if (lt == null)
      {
         return;
      }
      
      if (lt.client != null)
      {
         lt.client.stopBatchSession();
      }
   }

   /**
    * Displays the error message via current client side error writer.  This does not use
    * the client UI code tools, usually meaning standard output channel and further client
    * termination.
    *
    * @param   errmsg
    *          The error message to display
    */
   public static void displayError(String errmsg)
   {
      displayError(errmsg, !ErrorManager.isSystemAlertBoxes());
   }

   /**
    * Displays the error message via current client side error writer.  This does not use
    * the client UI code tools, usually meaning standard output channel and further client
    * termination.
    *
    * @param   errmsg
    *          The error message to display
    * @param   noMessageBox
    *          <code>true</code> if error is always displayed in message area or logged rather
    *          than displayed as a message box.
    */
   public static void displayError(String errmsg, boolean noMessageBox)
   {
      noMessageBox |= !ErrorManager.isSystemAlertBoxes();
      
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      List<String> stack = null;
      if (SessionUtils.isDebugAlert().booleanValue() && !noMessageBox)
      {
         stack = ProcedureManager.getStackTrace();
      }

      if (lt == null)
      {
         return;
      }

      if (lt.client != null)
      {
         lt.client.displayError(errmsg, noMessageBox,
            stack == null ? null : 
            stack.toArray(new String[stack.size()]));

      }
   }

   /**
    * Displays the warning message via current client side error writer.  This does not use
    * the client UI code tools, usually meaning standard output channel and further client
    * termination.
    *
    * @param   wrnmsg
    *          The warning message to display
    */
   public static void displayWarning(String wrnmsg)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      if (lt == null)
      {
         return;
      }

      if (lt.client != null)
      {
         lt.client.displayWarning(wrnmsg);
      }
   }

   /**
    * Displays the error message via current client side error writer.  This does not use
    * the client UI code tools, usually meaning standard output channel anf further client
    * termination.  This is the redirected version of the displayError.  If the output is
    * redirected to file the message is writing to the file.
    *
    * @param   errmsg
    *          The error message to display
    */
   public static void displayErrorRedirected(String errmsg)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      if (lt == null)
      {
         return;
      }
      
      if (lt.client != null)
      {
         lt.client.displayErrorRedirected(errmsg);
      }
   }
   
   /**
    * Changes the activation flag for batch mode driver if loaded.
    *
    * @param   on
    *          The new status of the active flag for batch driver.
    */
   public static void activateBatchMode(boolean on)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      if (lt == null)
      {
         return;
      }
      
      if (lt.client != null)
      {
         lt.client.activateBatchMode(on);
      }
   }
   
   /**
    * Displays a text item in an alert box.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    */
   public static void messageBox(Object text) 
   {
      messageBox(new Object[] {text});
   }

   /**
    * Displays a text item in an alert box.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object text, handle hWin) 
   {
      messageBox(new Object[] {text}, hWin);
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box.  The
    * message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    */
   public static void messageBox(Object[] list) 
   {
      messageBox(list, ALERT_MESSAGE, BTN_OK, (String)null);
   }

   /**
    * Displays a list of text items and empty lines in an alert box.  The
    * message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object[] list, handle hWin) 
   {
      messageBox(list, ALERT_MESSAGE, BTN_OK, (String)null, hWin);
   }

   /**
    * Displays a text item in an alert box.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX COLOR color
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    col
    *           Color scheme to use.
    */
   public static void messageBox(Object text, ColorSpec col) 
   {
      messageBox(new Object[] {text}, col);
   }

   /**
    * Displays a text item in an alert box.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX COLOR color
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    col
    *           Color scheme to use.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object text, ColorSpec col, handle hWin) 
   {
      messageBox(new Object[] {text}, col, hWin);
   }

   /**
    * Displays a list of text items and empty lines in an alert box.  The
    * message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX COLOR color
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    col
    *           Color scheme to use.
    */
   public static void messageBox(Object[] list, ColorSpec col) 
   {
      messageBox(list, ALERT_MESSAGE, BTN_OK, (String)null, col);
   }

   /**
    * Displays a list of text items and empty lines in an alert box.  The
    * message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX COLOR color
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    col
    *           Color scheme to use.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object[] list, ColorSpec col, handle hWin) 
   {
      messageBox(list, ALERT_MESSAGE, BTN_OK, (String)null, col, hWin);
   }

   /**
    * Displays a text item in an alert box. This is a modal dialog with the
    * specified title and button configuration.
    * <p>
    * This method corresponds to the Progress statements:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    */
   public static void messageBox(Object text, int alertType, int buttons, String title)
   {
      messageBox(new Object[] {text}, alertType, buttons, title);
   }

   /**
    * Displays a text item in an alert box. This is a modal dialog with the
    * specified title and button configuration.
    * <p>
    * This method corresponds to the Progress statements:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    */
   public static void messageBox(Object text, int alertType, int buttons, character title)
   {
      messageBox(new Object[] {text}, alertType, buttons, title);
   }

   /**
    * Displays a text item in an alert box. This is a modal dialog with the
    * specified title and button configuration.
    * <p>
    * This method corresponds to the Progress statements:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object text,
                                 int    alertType, 
                                 int    buttons, 
                                 String title,
                                 handle hWin)
   {
      messageBox(new Object[] {text}, alertType, buttons, title, hWin);
   }
   
   /**
    * Displays a text item in an alert box. This is a modal dialog with the
    * specified title and button configuration.
    * <p>
    * This method corresponds to the Progress statements:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object    text,
                                 int       alertType, 
                                 int       buttons, 
                                 character title,
                                 handle    hWin)
   {
      messageBox(new Object[] {text}, alertType, buttons, title, hWin);
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box. This 
    * is a modal dialog with the specified title and button configuration.
    * The message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    */
   public static void messageBox(Object[] list, int alertType, int buttons, String title)
   {
      messageBox(list, false, (Accessor)null, alertType, buttons, title);
   }

   /**
    * Displays a list of text items and empty lines in an alert box. This 
    * is a modal dialog with the specified title and button configuration.
    * The message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    */
   public static void messageBox(Object[] list, int alertType, int buttons, character title)
   {
      messageBox(list, false, (Accessor) null, alertType, buttons, title);
   }

   /**
    * Displays a list of text items and empty lines in an alert box. This 
    * is a modal dialog with the specified title and button configuration.
    * The message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object[] list,
                                 int      alertType, 
                                 int      buttons, 
                                 String   title,
                                 handle   hWin)
   {
      messageBox(list, false, (Accessor)null, alertType, buttons, title, hWin);
   }

   /**
    * Displays a list of text items and empty lines in an alert box. This 
    * is a modal dialog with the specified title and button configuration.
    * The message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object[]  list,
                                 int       alertType, 
                                 int       buttons, 
                                 character title,
                                 handle    hWin)
   {
      messageBox(list, false, (Accessor)null, alertType, buttons, title, hWin);
   }

   /**
    * Displays a text item in an alert box.  This is a modal dialog with the
    * specified title and button configuration.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    *         COLOR color
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    */
   public static void messageBox(Object    text,
                                 int       alertType, 
                                 int       buttons, 
                                 String    title,
                                 ColorSpec col)
   {
      messageBox(new Object[] {text}, alertType, buttons, title, col);
   }

   /**
    * Displays a text item in an alert box.  This is a modal dialog with the
    * specified title and button configuration.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    *         COLOR color
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    */
   public static void messageBox(Object    text,
                                 int       alertType, 
                                 int       buttons, 
                                 character title,
                                 ColorSpec col)
   {
      messageBox(new Object[] {text}, alertType, buttons, title, col);
   }

   /**
    * Displays a text item in an alert box.  This is a modal dialog with the
    * specified title and button configuration.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    *         COLOR color
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object    text,
                                 int       alertType, 
                                 int       buttons, 
                                 String    title,
                                 ColorSpec col,
                                 handle    hWin)
   {
      messageBox(new Object[] {text}, alertType, buttons, title, hWin);
   }

   /**
    * Displays a text item in an alert box.  This is a modal dialog with the
    * specified title and button configuration.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    *         COLOR color
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object    text,
                                 int       alertType, 
                                 int       buttons, 
                                 character title,
                                 ColorSpec col,
                                 handle    hWin)
   {
      messageBox(new Object[] {text}, alertType, buttons, title, hWin);
   }

   /**
    * Displays a list of text items and empty lines in an alert box.  This 
    * is a modal dialog with the specified title and button configuration. 
    * The message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    *         COLOR color
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    */
   public static void messageBox(Object[]  list,
                                 int       alertType, 
                                 int       buttons, 
                                 String    title,
                                 ColorSpec col)
   {
      messageBox(list, false, (Accessor)null, alertType, buttons, title, col);
   }

   /**
    * Displays a list of text items and empty lines in an alert box.  This 
    * is a modal dialog with the specified title and button configuration. 
    * The message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    *         COLOR color
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    */
   public static void messageBox(Object[]  list,
                                 int       alertType, 
                                 int       buttons, 
                                 character title,
                                 ColorSpec col)
   {
      messageBox(list, false, (Accessor)null, alertType, buttons, title, col);
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box.  This 
    * is a modal dialog with the specified title and button configuration. 
    * The message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    *         COLOR color
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object[]  list,
                                 int       alertType, 
                                 int       buttons, 
                                 String    title,
                                 ColorSpec col,
                                 handle    hWin)
   {
      messageBox(list, false, (Accessor)null, alertType, buttons, title, col, hWin);
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box.  This 
    * is a modal dialog with the specified title and button configuration. 
    * The message body is represented by a content array. Each line of the
    * message including any number of empty lines (skip) are elements of the
    * content array.  Text elements of the content array must be
    * <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType BUTTONS buttons TITLE title
    *         COLOR color
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object[]  list,
                                 int       alertType, 
                                 int       buttons, 
                                 character title,
                                 ColorSpec col,
                                 handle    hWin)
   {
      messageBox(list, false, (Accessor)null, alertType, buttons, title, col, hWin);
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    */
   public static void messageBox(Object    text,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 String    title)
   {
      messageBox(new Object[] {text}, set, var, alertType, buttons, title);
   }

   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    */
   public static void messageBox(Object    text,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 character title)
   {
      messageBox(new Object[] {text}, set, var, alertType, buttons, title);
   }

   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object    text,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 String    title,
                                 handle    hWin)
   {
      messageBox(new Object[] {text}, set, var, alertType, buttons, title, hWin);
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object    text,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 character title,
                                 handle    hWin)
   {
      messageBox(new Object[] {text}, set, var, alertType, buttons, title, hWin);
   }
   
   /**
    * Displays a text item in an alert box and sets/updates a 
    * <code>logical</code> variable as a result of the action taken.  This
    * is a modal dialog with the specified title and button configuration.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statements:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title COLOR color
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    */
   public static void messageBox(Object    text,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 String    title,
                                 ColorSpec col)
   {
      messageBox(new Object[] {text}, set, var, alertType, buttons, title, col);
   }

   /**
    * Displays a text item in an alert box and sets/updates a 
    * <code>logical</code> variable as a result of the action taken.  This
    * is a modal dialog with the specified title and button configuration.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statements:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title COLOR color
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    */
   public static void messageBox(Object    text,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 character title,
                                 ColorSpec col)
   {
      messageBox(new Object[] {text}, set, var, alertType, buttons, title, col);
   }

   /**
    * Displays a text item in an alert box and sets/updates a 
    * <code>logical</code> variable as a result of the action taken.  This
    * is a modal dialog with the specified title and button configuration.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statements:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title COLOR color
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object    text,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 String    title,
                                 ColorSpec col,
                                 handle    hWin)
   {
      messageBox(new Object[] {text}, set, var, alertType, buttons, title, col, hWin);
   }
   
   /**
    * Displays a text item in an alert box and sets/updates a 
    * <code>logical</code> variable as a result of the action taken.  This
    * is a modal dialog with the specified title and button configuration.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statements:
    * <p>
    * <pre>
    * MESSAGE text VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title COLOR color
    * </pre>
    * 
    * @param    text
    *           A single message text component to display.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    col
    *           Color scheme to use.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object    text,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 character title,
                                 ColorSpec col,
                                 handle    hWin)
   {
      messageBox(new Object[] {text}, set, var, alertType, buttons, title, col, hWin);
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    */
   public static void messageBox(Object[]  list,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 String    title)
   {
      messageBox(list, set, var, alertType, buttons, title, (ColorSpec)null);      
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    */
   public static void messageBox(Object[]  list,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 character title)
   {
      messageBox(list, set, var, alertType, buttons, title, (ColorSpec)null);      
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object[]  list,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 String    title,
                                 handle    hWin)
   {
      messageBox(list, set, var, alertType, buttons, title, (ColorSpec)null, hWin);      
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object[]  list,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 character title,
                                 handle    hWin)
   {
      messageBox(list, set, var, alertType, buttons, title, (ColorSpec)null, hWin);      
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title COLOR color
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    cs
    *           Color specification.
    */
   public static void messageBox(Object[]  list,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 String    title,
                                 ColorSpec cs)
   {
      messageBox(list, set, var, alertType, buttons, title, cs, (handle)null);
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title COLOR color
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    cs
    *           Color specification.
    */
   public static void messageBox(Object[]  list,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 character title,
                                 ColorSpec cs)
   {
      messageBox(list, set, var, alertType, buttons, title, cs, (handle)null);
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title COLOR color
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    cs
    *           Color specification.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object[]  list,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 character title,
                                 ColorSpec cs,
                                 handle    hWin)
   {
      if (title == null || title.isUnknown())
      {
         UnimplementedFeature.todo("MESSAGE-BOX title unknown behavior not determined");
         title = new character("");
      }
      
      messageBox(list, set, var, alertType, buttons, title.toStringMessage(), cs, hWin);
   }
   
   /**
    * Displays a list of text items and empty lines in an alert box and
    * sets/updates a <code>logical</code> variable as a result of the action
    * taken.  This is a modal dialog with the specified title and button
    * configuration. The message body is represented by a content array. Each
    * line of the message including any number of empty lines (skip) are
    * elements of the content array.  Text elements of the content array must
    * be <code>Strings</code>.  In addition, there may be any number of
    * <code>skip</code> elements which are encoded as a Java
    * <code>Integer</code> instance where the integral value is equivalent
    * to the number of lines to skip.
    * <p>
    * The value of the <code>var</code> being edited determines the default
    * button if it is not <code>null</code>.  <code>true</code> is the first
    * button, <code>false</code> is the second button and for a 3-button
    * dialog, <code>unknown</code> is the third button.
    * <p>
    * The result of the user's selection is returned and this result value
    * can be mapped to a button following the same mapping as used for
    * setting the default button (see above).
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * MESSAGE list VIEW-AS ALERT-BOX alertType SET/UPDATE var BUTTONS buttons
    *         TITLE title COLOR color
    * </pre>
    * 
    * @param    list
    *           The list of message text components.
    * @param    set
    *           <code>true</code> for SET, <code>false</code> for UPDATE.
    * @param    var
    *           The variable to update if not <code>null</code>.
    * @param    alertType
    *           One of the ALERT_* constants specifying message type.
    * @param    buttons
    *           One of the BTN_* constants specifying used button set.
    * @param    title
    *           The dialog title.
    * @param    cs
    *           Color specification.
    * @param    hWin
    *           handle of target window.
    */
   public static void messageBox(Object[]  list,
                                 boolean   set,
                                 Accessor  var, 
                                 int       alertType,
                                 int       buttons,
                                 String    title,
                                 ColorSpec cs,
                                 handle    hWin)
   {
      notifyRowDisplayDrawListener();

      List<String> stack = null;
      if (SessionUtils.isDebugAlert().booleanValue())
      {
         stack = ProcedureManager.getStackTrace();
      }
      
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      // are we in edit mode?
      if (var != null)
      {
         // disable loop protection and increment interactions counter
         lt.markInteractive();
      }

      logical src = null;
            
      // determine color
      Color clr = color(cs, lt.msgBoxColor);
      boolean isChui = isChui();
      
      // filter the content array down to String instances and Integer 
      // instances
      for (int i = 0; i < list.length; i++)
      {
         if (list[i] instanceof BaseDataType)
         {
            // convert any BDT to a string
            BaseDataType bdt = (BaseDataType) list[i];
            list[i] = bdt.toStringMessage();
         }
         else if (list[i] instanceof SkipEntity)
         {
            int skips = ((SkipEntity) list[i]).getSkip();
            
            if (isChui)
            {
               // convert any SKIP into an Integer
               list[i] = skips;
            }
            else
            {
               // convert any SKIP into required number of the new line characters
               list[i] = StringHelper.repeatChar('\n', skips + 1);
            }
         }
      }
      
      // only send an initial value down if there is an edit going on
      if (var != null)
      {
         src = (logical) var.get();
      }
      
      logical res = lt.client.messageBox(list,
                                         src,
                                         alertType & BATCH_ALERT_MASK, 
                                         buttons,
                                         title,
                                         clr,
                                         targetWindowWorker(hWin, false),
                                         stack == null ? null : 
                                            stack.toArray(new String[stack.size()]),
                                          (alertType & BATCH_ALERT) != 0);
                                           
      // if the user requested an edit operation, return the value
      if (var != null)
      {
         var.set(res);
      }
      
      // we MUST NOT maintain lastkey here because Progress does not 
   }

   /**
    * Defines a specific action for the key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>ON key-label key-function </b>
    * 
    * @param   keyLabel
    *          the label of the key
    * @param   keyFunction
    *          the action to take when the user presses the key
    */
   public static void remapKey(character keyLabel, character  keyFunction)
   {
      if (keyLabel.isUnknown() || keyFunction.isUnknown())
         return;
      
      String kl = keyLabel.getValue();
      String kf = keyFunction.getValue();
      Keyboard.remapKey(kl, kf);
      locate().client.remapKey(kl, kf);
   }

   /**
    * Defines a specific action for the key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>ON key-label key-function </b>
    * 
    * @param   keyLabel
    *          the label of the key
    * @param   keyFunction
    *          the action to take when the user presses the key
    */
   public static void remapKey(String keyLabel, String keyFunction)
   {
      remapKey(new character(keyLabel), new character(keyFunction));
   }

   /**
    * Defines a specific action for the key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>ON key-label key-function </b>
    * 
    * @param   keyLabel
    *          the label of the key
    * @param   keyFunction
    *          the action to take when the user presses the key
    */
   public static void remapKey(character keyLabel, String keyFunction)
   {
      remapKey(keyLabel, new character(keyFunction));
   }

   /**
    * Defines a specific action for the key.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>ON key-label key-function </b>
    * 
    * @param   keyLabel
    *          the label of the key
    * @param   keyFunction
    *          the action to take when the user presses the key
    */
   public static void remapKey(String  keyLabel, character keyFunction)
   {
      remapKey(new character(keyLabel), keyFunction);
   }

   /**
    * Suspends processing until the user presses any key. The type-ahead
    * buffer is cleared. The default message "Press space bar to continue"
    * appears on the status line.
    * <p>
    * This method corresponds to the Progress statement: <b>PAUSE</b>
    */
   public static void pause()
   {
      pauseWorker(-1, null, false, null);
   }

   /**
    * Suspends processing until the user presses any key. The type-ahead
    * buffer is cleared. The specified (or no) message appears on the status
    * line.
    * <p>
    * This method corresponds to the Progress statement:
    * <b>PAUSE MESSAGE/NO-MESSAGE</b>
    * 
    * @param   text
    *          Either <code>null</code> if NO-MESSAGE is in effect, or a
    *          message text.
    */
   public static void pause(character text)
   {
      pause(pauseText(text));
   }
   
   /**
    * Suspends processing until the user presses any key. The type-ahead
    * buffer is cleared. The default message "Press space bar to continue"
    * appears on the status line.
    * <p>
    * This method corresponds to the Progress statement: <b>PAUSE IN WINDOW h</b>
    * 
    * @param   hWin
    *          Handle of target window.
    */
   public static void pause(handle hWin)
   {
      pauseWorker(-1, null, false, hWin);
   }

   /**
    * Suspends processing for the specified number of seconds or until the
    * user presses any key, whichever comes first. The type-ahead buffer is
    * cleared. The default message "Pausing for n seconds. Press space bar to
    * continue" appears on the status line.
    * <p>
    * This method corresponds to the Progress statement: <b>PAUSE n IN WINDOW h</b>
    * where fractional seconds are used (as millis).
    * 
    * @param    seconds
    *           The number of seconds to pause. If specified as a
    *           non-positive number this method immediately returns.
    * @param    hWin
    *           Handle of target window.
    */
   public static void pause(double seconds, handle hWin)
   {
      pauseWorker(seconds, null, false, hWin);
   }
   
   /**
    * Suspends processing until the user presses any key. The type-ahead
    * buffer is cleared. The specified (or no) message appears on the status
    * line.
    * <p>
    * This method corresponds to the Progress statement:
    * <b>PAUSE MESSAGE/NO-MESSAGE IN WINDOW h</b>
    * 
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect, or a
    *           message text.
    * @param    hWin
    *           Handle of target window.
    */
   public static void pause(String text, handle hWin)
   {
      pauseWorker(-1, text, false, hWin);
   }
   
   /**
    * Suspends processing for the specified number of seconds or until the user 
    * presses any key. The type-ahead buffer is cleared. The specified (or no) 
    * message appears on the status line.
    * <p>
    * This method corresponds to the Progress statement:
    * <b>PAUSE n MESSAGE/NO-MESSAGE IN WINDOW h</b>
    * where fractional seconds are used (as millis).
    * 
    * @param    seconds
    *           The number of seconds to pause. If specified as a
    *           non-positive number this method immediately returns.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect, or a
    *           message text.
    * @param    hWin
    *           Handle of target window.
    */
   public static void pause(double seconds, String text, handle hWin)
   {
      pauseWorker(seconds, text, false, hWin);
   }
   
   /**
    * Suspends processing until the user presses any key. The type-ahead
    * buffer is cleared. The specified (or no) message appears on the status
    * line.
    * <p>
    * This method corresponds to the Progress statement:
    * <b>PAUSE MESSAGE/NO-MESSAGE</b>
    * 
    * @param   text
    *          Either <code>null</code> if NO-MESSAGE is in effect, or a
    *          message text.
    */
   public static void pause(String text)
   {
      pauseWorker(-1, text, (text == null), null);
   }

   /**
    * Suspends processing for the specified number of seconds or until the
    * user presses any key, whichever comes first. The type-ahead buffer is
    * cleared. The default message "Pausing for n seconds. Press space bar to
    * continue" appears on the status line.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PAUSE n </b>
    * where fractional seconds are used (as millis).
    * 
    * @param    seconds
    *           The number of seconds to pause. If specified as a
    *           non-positive number this method immediately returns.
    */
   public static void pause(double seconds)
   {
      pauseWorker(seconds, null, false, null);
   }

   /**
    * Suspends processing for the specified number of seconds or until the
    * user presses any key, whichever comes first. The type-ahead buffer is
    * cleared. The default message "Pausing for n seconds. Press space bar to
    * continue" appears on the status line.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PAUSE n </b>
    * where fractional seconds are used (as millis).
    * 
    * @param    seconds
    *           The number of seconds to pause. If specified as a
    *           non-positive number this method immediately returns.
    *
    * @throws   ErrorConditionException
    *           If the number of seconds is the <code>unknown value</code>
    *           or is <code>null</code>.
    */
   public static void pause(NumberType seconds)
   throws ErrorConditionException
   {
      if (seconds == null || seconds.isUnknown())
      {
         String err = "Could not evaluate number of seconds in PAUSE";
         ErrorManager.recordOrThrowError(353, err);
         return;
      }
      
      pause(seconds.doubleValue());
   }

   /**
    * Suspends processing for the specified number of seconds or until the
    * user presses any key, whichever comes first. The type-ahead buffer is
    * cleared. The specified (or no) message appears on the status line.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PAUSE n MESSAGE/NO-MESSAGE </b>
    * where fractional seconds are used (as millis).
    * 
    * @param    seconds
    *           The number of seconds to pause. If specified as a
    *           non-positive number this method immediately returns.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect, or the
    *           message text.
    */
   public static void pause(double seconds, character text)
   {
      pause(seconds, pauseText(text));
   }

   /**
    * Suspends processing for the specified number of seconds or until the
    * user presses any key, whichever comes first. The type-ahead buffer is
    * cleared. The specified (or no) message appears on the status line.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PAUSE n MESSAGE/NO-MESSAGE </b>
    * where fractional seconds are used (as millis).
    * 
    * @param    seconds
    *           The number of seconds to pause. If specified as a
    *           non-positive number this method immediately returns.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect, or the
    *           message text.
    *
    * @throws   ErrorConditionException
    *           If the number of seconds is the <code>unknown value</code>
    *           or is <code>null</code>.
    */
   public static void pause(NumberType seconds, character text)
   throws ErrorConditionException
   {
      if (seconds == null || seconds.isUnknown())
      {
         String err = "Could not evaluate number of seconds in PAUSE";
         ErrorManager.recordOrThrowError(353, err);
         return;
      }
         
      pause(seconds.doubleValue(), text);  
   }

   /**
    * Suspends processing for the specified number of seconds or until the
    * user presses any key, whichever comes first. The type-ahead buffer is
    * cleared. The specified (or no) message appears on the status line.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PAUSE n MESSAGE/NO-MESSAGE </b>
    * where fractional seconds are used (as millis).
    * 
    * @param    seconds
    *           The number of seconds to pause. If specified as a
    *           non-positive number this method immediately returns.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect, or the
    *           message text.
    */
   public static void pause(double seconds, String text)
   {
      pauseWorker(seconds, text, (text == null), null);
   }

   /**
    * Suspends processing for the specified number of seconds or until the
    * user presses any key, whichever comes first. The type-ahead buffer is
    * cleared. The specified (or no) message appears on the status line.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PAUSE n MESSAGE/NO-MESSAGE </b>
    * where fractional seconds are used (as millis).
    * 
    * @param    seconds
    *           The number of seconds to pause. If specified as a
    *           non-positive number this method immediately returns.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect, or the
    *           message text.
    */
   public static void pause(NumberType seconds, String text)
   {
      if (seconds == null || seconds.isUnknown())
      {
         String err = "Could not evaluate number of seconds in PAUSE";
         ErrorManager.recordOrThrowError(353, err);
         return;
      }
         
      pause(seconds.doubleValue(), text);  
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts until the user presses a key. The default
    * message "Press space bar to continue" appears on the status line.
    * <p>
    * This method corresponds to Progress statement:
    * <p>
    * <pre>
    * PAUSE BEFORE-HIDE
    * </pre>
    */
   public static void pauseBeforeHide()
   {
      pauseBeforeHide(-1, (String) null, null);
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts until the user presses a key. The
    * specified (or no) message appears on the status line.
    * <p>
    * This method corresponds to Progress statement:
    * <p>
    * <pre>
    * PAUSE BEFORE-HIDE MESSAGE text
    * PAUSE BEFORE-HIDE NO-MESSAGE
    * </pre>
    * 
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect or the
    *           message text to display.
    */
   public static void pauseBeforeHide(character text)
   {
      pauseBeforeHide(-1, text);
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts until the user presses a key. The
    * specified (or no) message appears on the status line.
    * <p>
    * This method corresponds to Progress statement:
    * <p>
    * <pre>
    * PAUSE BEFORE-HIDE MESSAGE text
    * PAUSE BEFORE-HIDE NO-MESSAGE
    * </pre>
    * 
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect or the
    *           message text to display.
    */
   public static void pauseBeforeHide(String text)
   {
      pauseBeforeHide(-1, text, null);
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts the specified number of seconds or until
    * the user presses a key, whichever comes first. The default message
    * "Pausing for n seconds. Press space bar to continue" appears on the
    * status line.
    * <p>
    * This method corresponds to Progress statement:
    * <p>
    * <pre>
    * PAUSE seconds BEFORE-HIDE
    * </pre>
    * where fractional seconds are used (as millis).
    * 
    * @param    seconds
    *           Pause duration in seconds.
    */
   public static void pauseBeforeHide(double seconds)
   {
      pauseBeforeHide(seconds, (String) null, null);
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts the specified number of seconds or until
    * the user presses a key, whichever comes first. The default message
    * "Pausing for n seconds. Press space bar to continue" appears on the
    * status line.
    * <p>
    * This method corresponds to Progress statement:
    * <p>
    * <pre>
    * PAUSE seconds BEFORE-HIDE MESSAGE/NO-MESSAGE
    * </pre>
    * 
    * @param    seconds
    *           Pause duration in seconds.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect or the
    *           message text to display.
    */
   public static void pauseBeforeHide(double seconds, String text)
   {
      pauseBeforeHide(seconds, (String) null, null);
   }
   
   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts the specified number of seconds or until
    * the user presses a key, whichever comes first. The default message
    * "Pausing for n seconds. Press space bar to continue" appears on the
    * status line.
    * <p>
    * This method corresponds to Progress statement:
    * <p>
    * <pre>
    * PAUSE seconds BEFORE-HIDE
    * </pre>
    * 
    * @param    seconds
    *           Pause duration in seconds.
    */
   public static void pauseBeforeHide(NumberType seconds)
   {
      if (seconds == null || seconds.isUnknown())
      {
         String err = "Could not evaluate number of seconds in PAUSE";
         ErrorManager.recordOrThrowError(353, err);
         return;
      }

      pauseBeforeHide(seconds.doubleValue(), (String) null, null);
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts the specified number of seconds or until
    * the user presses a key, whichever comes first. The specified (or no)
    * message appears on the status line.
    * <p>
    * This method corresponds to Progress statements:
    * <p>
    * <pre>
    * PAUSE seconds BEFORE-HIDE MESSAGE text
    * PAUSE seconds BEFORE-HIDE NO-MESSAGE
    * </pre>
    * 
    * @param    seconds
    *           Pause duration in seconds.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect or the
    *           message text to display.
    */
   public static void pauseBeforeHide(double seconds, character text)
   {
      String msg = text.isUnknown() ? null : text.toStringMessage();
      pauseBeforeHide(seconds, msg, null);
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts until the user presses a key. The default
    * message "Press space bar to continue" appears on the status line.
    * <p>
    * This method corresponds to Progress statement:
    * <p>
    * <pre>
    * PAUSE BEFORE-HIDE IN WINDOW h.
    * </pre>
    * 
    * @param    hWin
    *           Handle of target window.
    */
   public static void pauseBeforeHide(handle hWin)
   {
      pauseBeforeHide(-1, null, hWin);
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts the specified number of seconds or until
    * the user presses a key, whichever comes first. The default message
    * "Pausing for n seconds. Press space bar to continue" appears on the
    * status line.
    * <p>
    * This method corresponds to Progress statement:
    * <p>
    * <pre>
    * PAUSE seconds BEFORE-HIDE IN WINDOW h.
    * </pre>
    * 
    * @param    seconds
    *           Pause duration in seconds.
    */   
   public static void pauseBeforeHide(double seconds, handle hWin)
   {
      pauseBeforeHide(seconds, null, hWin);
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts the specified number of seconds or until
    * the user presses a key, whichever comes first. The specified (or no)
    * message appears on the status line.
    * <p>
    * This method corresponds to Progress statements:
    * <p>
    * <pre>
    * PAUSE seconds BEFORE-HIDE MESSAGE text
    * PAUSE seconds BEFORE-HIDE NO-MESSAGE
    * </pre>
    * 
    * @param    seconds
    *           Pause duration in seconds.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect or the
    *           message text to display.
    */
   public static void pauseBeforeHide(NumberType seconds, character text)
   {
      if (seconds == null || seconds.isUnknown())
      {
         String err = "Could not evaluate number of seconds in PAUSE";
         ErrorManager.recordOrThrowError(353, err);
         return;
      }

      pauseBeforeHide(seconds.doubleValue(), text);
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts the specified number of seconds or until
    * the user presses a key, whichever comes first. The specified (or no)
    * message appears on the status line.
    * <p>
    * This method corresponds to Progress statements:
    * <p>
    * <pre>
    * PAUSE seconds BEFORE-HIDE MESSAGE text
    * PAUSE seconds BEFORE-HIDE NO-MESSAGE
    * </pre>
    * 
    * @param    seconds
    *           Pause duration in seconds.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect or the
    *           message text to display.
    * @param    hWin
    *           target window for IN WINDOW clause.
    */
   public static void pauseBeforeHide(double seconds, String text, handle hWin)
   {
      locate().client.pauseBeforeHide(toMilliseconds(seconds), 
                                      text, 
                                      targetWindowWorker(hWin, true));
   }

   /**
    * Specifies the pause action the user must take whenever frames are hidden
    * automatically. The pause lasts the specified number of seconds or until
    * the user presses a key, whichever comes first. The specified (or no)
    * message appears on the status line.
    * <p>
    * This method corresponds to Progress statements:
    * <p>
    * <pre>
    * PAUSE seconds BEFORE-HIDE MESSAGE text
    * PAUSE seconds BEFORE-HIDE NO-MESSAGE
    * </pre>
    * where fractional seconds are used (as millis).
    * 
    * @param    seconds
    *           Pause duration in seconds.
    * @param    text
    *           Either <code>null</code> if NO-MESSAGE is in effect or the
    *           message text to display.
    */
   public static void pauseBeforeHide(NumberType seconds, String text)
   {
      if (seconds == null || seconds.isUnknown())
      {
         String err = "Could not evaluate number of seconds in PAUSE";
         ErrorManager.recordOrThrowError(353, err);
         return;
      }

      pauseBeforeHide(seconds.doubleValue(), text, null);
   }

   /**
    * Processes all outstanding events without blocking for user input.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PROCESS EVENTS </b>
    */
   public static void processEvents()
   {
      LogicalTerminal lt = locate();

      // identify the active frame
      int frameId = lt.client.getActiveFrame();
      
      if (frameId == -1)
      {
         lt.client.processEvents(null);
         return;
      }
            
      GenericFrame frame = lt.identifyFrame(frameId);
      
      if (frame == null)
      {
         throw new RuntimeException("Invalid frame ID " + frameId);
      }
      
      ScreenBuffer[] sb = lt.client.processEvents(lt.getEditableServerScreenBuffers(frame));
      // special processing for default screen buffer
      if (sb != null)
      {
         final ScreenBuffer firstSb = sb[0];
         frame.setScreenBuffer(firstSb, false);
         KeyReader.setLastKey(firstSb.getKeyCode(), firstSb.getKeyState());
         
         // and handling the rest of the screen buffers as usual
         if (sb.length > 1)
         {
            handleScreenBuffers(sb, 1);
         }
      }
      lt.frameRef = frame;
   }

   /**
    * Turns cursor display on or off.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>PUT CURSOR </b> <br>
    * <b>PUT CURSOR OFF </b>
    * 
    * @param   onOff
    *          <code>true</code> if the cursor display is being turned on.
    */
   public static void putCursor(boolean onOff)
   {
      locate().client.putCursor(onOff);
   }

   /**
    * Puts cursor to the specified screen position.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PUT CURSOR ROW COLUMN </b>
    * 
    * @param   row
    *          1-based screen row number
    * @param   column
    *          1-based screen column number
    */
   public static void putCursor(double row, double column)
   {
      locate().client.putCursor(row, column);
   }

   /**
    * Puts cursor to the specified screen position.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PUT CURSOR ROW COLUMN </b>
    * 
    * @param   row
    *          1-based screen row number
    * @param   column
    *          1-based screen column number
    */
   public static void putCursor(NumberType row, NumberType column)
   {
      putCursor(row.doubleValue(), column.doubleValue());
   }

   /**
    * Puts cursor to the specified screen position.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PUT CURSOR ROW COLUMN </b>
    * 
    * @param   row
    *          1-based screen row number
    * @param   column
    *          1-based screen column number
    */
   public static void putCursor(double row, NumberType column)
   {
      putCursor(row, column.doubleValue());
   }

   /**
    * Puts cursor to the specified screen position.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PUT CURSOR ROW COLUMN </b>
    * 
    * @param   row
    *          1-based screen row number
    * @param   column
    *          1-based screen column number
    */
   public static void putCursor(NumberType row, double column)
   {
      putCursor(row.doubleValue(), column);
   }

   /**
    * Puts text on the screen at the (1,1) position and with the default
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text
    * </pre>
    * 
    * @param    text
    *           The text to display.  The output will be generated using
    *           <code>character.toStringMessage</code>
    */
   public static void putScreen(character text)
   {
      putScreen(text.toStringMessage());
   }

   /**
    * Puts text on the screen at the (1,1) position and with the default
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text
    * </pre>
    * 
    * @param    text
    *           The text to display.
    */
   public static void putScreen(String text)
   {
      putScreen(text, (ColorSpec) null);
   }

   /**
    * Puts text on the screen at the specified position and with the default
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    */
   public static void putScreen(character text, double row, double column)
   {
      putScreen(text.toStringMessage(), row, column, (ColorSpec) null);
   }

   /**
    * Puts text on the screen at the specified position and with the default
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    */
   public static void putScreen(String text, double row, double column)
   {
      putScreen(text, row, column, (ColorSpec) null);
   }

   /**
    * Puts text on the screen at the specified position and with the default
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    */
   public static void putScreen(character  text,
                                NumberType row, 
                                NumberType column)
   {
      putScreen(text.toStringMessage(),
                row.doubleValue(),
                column.doubleValue(),
                (ColorSpec) null);
   }

   /**
    * Puts text on the screen at the specified position and with the default
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    */
   public static void putScreen(String     text,
                                NumberType row, 
                                NumberType column)
   {
      putScreen(text,
                row.doubleValue(),
                column.doubleValue(),
                (ColorSpec) null);
   }

   /**
    * Puts text on the screen at the specified position and with the default
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    */
   public static void putScreen(character text, double row, NumberType column)
   {
      putScreen(text.toStringMessage(),
                row,
                column.doubleValue(),
                (ColorSpec) null);
   }

   /**
    * Puts text on the screen at the specified position and with the default
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    */
   public static void putScreen(String text, double row, NumberType column)
   {
      putScreen(text, row, column.doubleValue(), (ColorSpec) null);
   }

   /**
    * Puts text on the screen at the specified position and with the default
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    */
   public static void putScreen(character text, NumberType row, double column)
   {
      putScreen(text.toStringMessage(),
                row.doubleValue(),
                column,
                (ColorSpec) null);
   }

   /**
    * Puts text on the screen at the specified position.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>PUT SCREEN ROW COLUMN </b>
    * 
    * @param   text
    *          string of text to display
    * @param   row
    *          1-based screen row number
    * @param   column
    *          1-based screen column number
    */
   public static void putScreen(String text, NumberType row, double column)
   {
      putScreen(text, row, column, (ColorSpec) null);
   }

   /**
    * Puts text on the screen at the (1,1) position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    cs
    *           The color specification or -1 for the default color.
    */
   public static void putScreen(character text, ColorSpec cs)
   {
      putScreen(text.toStringMessage(), 1, 1, cs);
   }

   /**
    * Puts text on the screen at the (1,1) position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    cs
    *           The color specification or -1 for the default color.
    */
   public static void putScreen(String text, ColorSpec cs)
   {
      putScreen(text, 1, 1, cs);
   }

   /**
    * Puts text on the screen at the specified position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    * @param    cs
    *           The color specification or -1 for the default color.
    */
   public static void putScreen(character text,
                                double    row,
                                double    column, 
                                ColorSpec cs)
   {
      putScreen(text.toStringMessage(), row, column, cs);
   }

   /**
    * Puts text on the screen at the specified position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    * @param    cs
    *           The color specification or <code>null</code> for the default color.
    */
   public static void putScreen(String    text,
                                double    row,
                                double    column, 
                                ColorSpec cs)
   {
      if (AppServerManager.isRemote())
      {
         // appserver PUT SCREEN statements are silently ignored, even if terminal is redirected
         return;
      }
      
      Color color = cs == null ? null : cs.convert();
      locate().client.putScreen(text, row, column, color);
   }

   /**
    * Puts text on the screen at the specified position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    * @param    cs
    *           The color specification or <code>null</code> for the default color.
    */
   public static void putScreen(character  text,
                                NumberType row, 
                                NumberType column,
                                ColorSpec  cs)
   {
      putScreen(text.toStringMessage(),
                row.doubleValue(),
                column.doubleValue(),
                cs);
   }

   /**
    * Puts text on the screen at the specified position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    * @param    cs
    *           The color specification or <code>null</code> for the default color.
    */
   public static void putScreen(String     text,
                                NumberType row, 
                                NumberType column,
                                ColorSpec  cs)
   {
      putScreen(text, row.doubleValue(), column.doubleValue(), cs);
   }

   /**
    * Puts text on the screen at the specified position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    * @param    cs
    *           The color specification or <code>null</code> for the default color.
    */
   public static void putScreen(character  text,
                                double     row,
                                NumberType column, 
                                ColorSpec  cs)
   {
      putScreen(text.toStringMessage(), row, column.doubleValue(), cs);
   }

   /**
    * Puts text on the screen at the specified position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    * @param    cs
    *           The color specification or <code>null</code> for the default color.
    */
   public static void putScreen(String     text,
                                double     row,
                                NumberType column, 
                                ColorSpec  cs)
   {
      putScreen(text, row, column.doubleValue(), cs);
   }

   /**
    * Puts text on the screen at the specified position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    * @param    cs
    *           The color specification or -1 for the default color.
    */
   public static void putScreen(character  text,
                                NumberType row,
                                double     column,
                                ColorSpec  cs)
   {
      putScreen(text.toStringMessage(), row.doubleValue(), column, cs);
   }

   /**
    * Puts text on the screen at the specified position and with the given
    * text attributes.  Such text is directly drawn on top of the Z-order
    * and portions of that text will be automatically removed as those 
    * portions are obscured by other drawing on the screen (be it other
    * calls to this method or any other drawing).  Any partial obscuring of
    * the output will leave any unobscured text on the screen until drawing
    * occurs over those screen positions.
    * <p>
    * This method corresponds to the Progress statement:
    * <p>
    * <pre>
    * PUT SCREEN text ROW y COLUMN x
    * </pre>
    * 
    * @param    text
    *           The text to display.
    * @param    row
    *           1-based screen row number.
    * @param    column
    *           1-based screen column number.
    * @param    cs
    *           The color specification or -1 for the default color.
    */
   public static void putScreen(String     text,
                                NumberType row,
                                double     column, 
                                ColorSpec  cs)
   {
      putScreen(text, row.doubleValue(), column, cs);
   }

   /**
    * Reads and returns one keystroke from the keyboard type-ahead buffer.
    * <p>
    * This method corresponds to these Progress statements:<br>
    * <b>READKEY PAUSE n</b><br>
    * <b>READKEY</b>
    * 
    * @param    millis
    *           -1 means indefinite wait, 0 means no wait and any positive
    *           value is the number of milliseconds to wait.
    * 
    * @return   -1 for time-out, otherwise the event code.
    *
    * @throws   StopConditionException
    *           If CTRL-C is input.
    */
   public static int readKey(int millis)
   throws StopConditionException
   {
      if (AppServerManager.isRemote())
      {
         // never wait in case of appserver agents
         millis = 0;
         
         if (UnnamedStreams.input() == null)
         {
            // in case of appserver Agents reading from STDIN, READKEY will return -1.
            return -1;
         }
      }
      
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      // disable loop protection and increment interactions counter
      lt.markInteractive();

      if (millis == 0)
      {
         // Time distance is small enough to make assumption that user had no
         // chance to press key reasonable, but big enough to make it work  
         // even under heavy load.
         if (System.currentTimeMillis() - lt.millis < 350)
            return -1;
      }

      int key = -1;
      
      try
      {
         ReadKeyResult res = lt.client.readKey(millis);
         key = res.keyCode;
         
         if (millis == 0 && key == -1)
            lt.millis = System.currentTimeMillis();
         else
            lt.millis = 0;

         ScreenBuffer[] sb = res.editableBuffers;
         if (sb != null && sb.length > 0)
         {
            GenericFrame gfr = getFrame(sb[0].getBaseId());
            handleScreenBuffers(gfr, sb, false);
         }
      }
      
      catch (StopConditionException sce)
      {
         // a CTRL-C must have been input, force lastkey to -1
         KeyReader.setLastKey(-1);
         throw sce;
      }
      
      return key;
   }

   /**
    * Resets the status message for the running procedure back to the 
    * standard message (which is blank or the empty string).
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS DEFAULT</code>.
    */
   public static void statusDefault()
   {
      statusDefault((String) null);
   }

   /**
    * Specifies a new default status message for the running procedure when
    * not in INPUT mode (editing).
    * <p>
    * This method corresponds to the Progress statement 
    * <code>STATUS DEFAULT text</code>.
    * 
    * @param    text
    *           The new status message.  <code>unknown value</code> is the
    *           equivalent of the empty string.
    */
   public static void statusDefault(character text)
   {
      statusDefault(text.isUnknown() ? null : text.toStringMessage());
   }

   /**
    * Specifies a new default status message for the running procedure when
    * not in INPUT mode (when editing).
    * <p>
    * This method corresponds to the Progress statement 
    * <code>STATUS DEFAULT text</code>.
    * 
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    */
   public static void statusDefault(String text)
   {
      locate().client.statusDefault(text, false, WindowConfig.RESOLVE_WINDOW);
   }
   
   /**
    * Resets the status message for the running procedure back to the standard message 
    * (which is blank or the empty string).
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS DEFAULT IN WINDOW ...</code>.
    * 
    * @param    hWin
    *           Handle of target window.
    */
   public static void statusDefault(handle hWin)
   {
      locate().client.statusDefault(null, false, targetWindowWorker(hWin, false));
   }
   
   /**
    * Specifies a new default status message for the running procedure when
    * not in INPUT mode (when editing).
    * <p>
    * This method corresponds to the Progress statement 
    * <code>STATUS DEFAULT text IN WINDOW wnd</code>.
    * 
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    * @param    hWin
    *           Handle of target window.
    */
   public static void statusDefault(String text, handle hWin)
   {
      locate().client.statusDefault(text, false, targetWindowWorker(hWin, false));
   }

   /**
    * Specifies a new default status message for the running procedure when
    * not in INPUT mode (when editing).
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS DEFAULT text IN WINDOW wnd</code>.
    *
    * @param    text
    *           The new status message.  Unknown value is the equivalent
    *           of the empty string.
    * @param    hWin
    *           Handle of target window.
    */
   public static void statusDefault(character text, handle hWin)
   {
      statusDefault(text.isUnknown() ? null : text.toStringMessage(), hWin);
   }
   
   /**
    * Specifies a new default status message for the running procedure when
    * not in INPUT mode (when editing).
    * <p>
    * This method is used to display a status message, which is trimmed to
    * 63 chars only if the <code>noLengthLimit</code> flag is not set.
    * 
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    * @param    noLengthLimit
    *           flag indicating the status message should not be trimmed
    */
   public static void statusDefault(String text, boolean noLengthLimit)
   {
      locate().client.statusDefault(text, noLengthLimit, WindowConfig.RESOLVE_WINDOW);
   }

   /**
    * Resets the status message for the blocked on input procedure back to
    * the default, which is "Enter data or press 'END-ERROR' to end".
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT</code>.
    */
   public static void statusInput()
   {
      locate().client.statusInputRevert(WindowConfig.RESOLVE_WINDOW);
   }

   /**
    * Specifies a new status message used when the user interface is blocked
    * on input.
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT text</code>.
    * 
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    */
   public static void statusInput(String text)
   {
      locate().client.statusInput(text, WindowConfig.RESOLVE_WINDOW);
   }

   /**
    * Specifies a new status message used when the user interface is blocked
    * on input.
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT text IN WINDOW wnd.</code>.
    * 
    * @param    text
    *           The new status message.  <code>null</code> is the equivalent
    *           of the empty string.
    * @param    hWin
    *           Handle of the target window
    */
   public static void statusInput(String text, handle hWin)
   {
      locate().client.statusInput(text, targetWindowWorker(hWin, false));
   }
   
   /**
    * Specifies a new status message used when the user interface is blocked
    * on input.
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT text</code>.
    * 
    * @param    text
    *           The new status message.  <code>unknown</code> is the
    *           equivalent of the empty string.
    */
   public static void statusInput(character text)
   {
      statusInput(text.isUnknown() ? null : text.toStringMessage());
   }

   /**
    * Disables display of the input status message.
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT OFF</code>.
    */
   public static void statusInputOff()
   {
      locate().client.statusInputOff(WindowConfig.RESOLVE_WINDOW);
   }

   /**
    * Disables display of the input status message in specified window.
    * <p>
    * This method corresponds to the Progress statement
    * <code>STATUS INPUT OFF IN WINDOW wnd.</code>.

    * @param   hWin
    *          Handle of target window.
    */
   public static void statusInputOff(handle hWin)
   {
      locate().client.statusInputOff(targetWindowWorker(hWin, false));
   }
   
   /**
    * Underlines a set of widgets of the specified frame. If the target
    * frame is not a down frame, the frame is simply cleared.
    * 
    * @param    frame
    *           The frame for which this request came.
    * @param    widgetIds
    *           The array of field level widget IDs to be underlined.
    */
   public static void underline(GenericFrame frame, int[] widgetIds)
   {
      locate().client.underline(frame.getFrameId(), widgetIds);
   }

   /**
    * Waits for a default list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR </b>
    *
    * @param    frame
    *           The frame for which this request came or <code>null</code>
    *           if the currently focused frame should be used.
    */
   public static void waitFor(CommonFrame frame)
   {
      waitFor(frame, null);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR </b>
    * 
    * @param    frame
    *           The frame for which this request came or <code>null</code>
    *           if the currently focused frame should be used.
    * @param    el
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    */
   public static void waitFor(CommonFrame frame, EventList el)
   {
      if (el == null)
      {
         el = new EventList(true);
      }
      
      waitFor(frame, el, -1, locate().getDefaultTimeout());
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget </b>
    * 
    * @param    frame
    *           The frame for which this request came or <code>null</code>
    *           if the currently focused frame should be used.
    * @param    el
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    * @param    focusWidget
    *           Widget which is to be in focus or <code>null</code> for no
    *           widget.
    */
   public static void waitFor(CommonFrame      frame,
                              EventList        el, 
                              GenericWidget<?> focusWidget)
   {
      waitFor(frame, el, focusWidget, locate().getDefaultTimeout());
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR PAUSE n </b>
    * 
    * @param    frame
    *           The frame for which this request came or <code>null</code>
    *           if the currently focused frame should be used.
    * @param    el
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    * @param    seconds
    *           Maximum time to wait for event.
    */
   public static void waitFor(CommonFrame frame, EventList el, double seconds)
   {
      waitFor(frame, el, -1, seconds);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR PAUSE n </b>
    * 
    * @param    frame
    *           The frame for which this request came or <code>null</code>
    *           if the currently focused frame should be used.
    * @param    el
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    * @param    seconds
    *           Maximum time to wait for event.
    */
   public static void waitFor(CommonFrame frame,
                              EventList   el, 
                              NumberType  seconds)
   {
      waitFor(frame, el, -1, seconds.isUnknown() ? -1 : seconds.doubleValue());
   }
   
   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget PAUSE n </b>
    * 
    * @param    frame
    *           The frame for which this request came or <code>null</code>
    *           if the currently focused frame should be used.
    * @param    el
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    * @param    focusWidget
    *           Widget which is to be in focus or <code>null</code> for no
    *           widget.
    * @param    seconds
    *           Maximum time to wait for event.
    */
   public static void waitFor(CommonFrame      frame,
                              EventList        el, 
                              GenericWidget<?> focusWidget,
                              double           seconds)
   {
      int id = (focusWidget == null) ? -1 : focusWidget.getId();
      waitFor(frame, el, id, seconds);
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget PAUSE n </b>
    * 
    * @param    frame
    *           The frame for which this request came or <code>null</code>
    *           if the currently focused frame should be used.
    * @param    el
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    * @param    focusWidget
    *           Widget which is to be in focus or <code>null</code> for no
    *           widget.
    * @param    seconds
    *           Maximum time to wait for event.
    */
   public static void waitFor(CommonFrame      frame,
                              EventList        el, 
                              GenericWidget<?> focusWidget,
                              NumberType       seconds)
   {
      waitFor(frame, el, focusWidget, seconds.isUnknown() ? -1 : seconds.doubleValue());
   }

   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS FRAME frame </b>
    * 
    * @param    frame
    *           The frame for which this request came or <code>null</code>
    *           if the currently focused frame should be used.
    * @param    el
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    * @param    focusFrame
    *           Frame which is to be in focus.
    */
   public static void waitFor(CommonFrame   frame,
                              EventList     el, 
                              CommonFrame   focusFrame)
   {
      waitFor(frame, el, focusFrame.asWidget());
   }
   
   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget PAUSE n </b>
    * 
    * @param    frame
    *           The frame for which this request came or <code>null</code>
    *           if the currently focused frame should be used.
    * @param    el
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    * @param    focusWidget
    *           Widget which is to be in focus or <code>null</code> for no
    *           widget.
    * @param    seconds
    *           Maximum time to wait for event.
    */
   public static void waitFor(CommonFrame  frame,
                              EventList    el, 
                              CommonFrame  focusWidget,
                              double       seconds)
   {
      int id = (focusWidget == null) ? -1 : focusWidget.getFrameId();
      waitFor(frame, el, id, seconds);
   }
   
   /**
    * Waits for specified list of events.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>WAIT-FOR FOCUS widget PAUSE n </b>
    * where fractional seconds are used (as millis).
    * 
    * @param    frame
    *           The frame for which this request came or <code>null</code>
    *           if the currently focused frame should be used.
    * @param    el
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    * @param    focusWidgetId
    *           Widget ID which is to be in focus or -1 for no widget.
    * @param    seconds
    *           Maximum time to wait for event.
    */
   public static void waitFor(CommonFrame   frame,
                              EventList     el, 
                              int           focusWidgetId,
                              double        seconds)
   {
      GenericFrame gfr = null;
      
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      // disable loop protection and increment interactions counter
      lt.markInteractive(true);
      
      // use the active frame if no frame is specified
      if (frame == null)
      {
         gfr = getCurrentFrame();
      }      
      else
      {
         if (frame instanceof GenericFrame)
         {
            gfr = (GenericFrame) frame;
         }
         else
         {
            gfr = lt.identifyFrame(frame.getFrameId());
         }
      }
      
      EventList wl = lt.prepareEventList(gfr, el);
      
      // WAIT-FOR does not change ENTERED state of widgets
      ScreenBuffer[] sb = null;

      if (gfr != null)
      {
         sb = lt.getEditableServerScreenBuffers(gfr);
      }

      boolean push = el != null;
      
      if (push)
      {
         lt.waitStack.push(el);
      }
      
      // call the client
      try
      {
         sb = lt.client.waitFor(wl, focusWidgetId, toMilliseconds(seconds), sb);
         
         if (gfr != null && sb != null)
         {
            handleScreenBuffers(gfr, sb, false);
         }
      }
      catch (ConditionException ce)
      {
         processCondition(ce);

         // continue the unwinding of the stack
         throw ce;
      }
      finally
      {
         if (push)
         {
            lt.waitStack.pop();
         }
      }
   }

   /**
    * Return the topmost or active frame.
    *
    * @return   The active frame or <code>null</code> if there is no active
    *           frame.
    */
   public static GenericFrame getCurrentFrame()
   {
      LogicalTerminal lt = locate();
      
      // identify the active frame
      int frameId = lt.client.getActiveFrame();
      
      if (frameId == -1)
      {
         // nothing to do, "there is no spoon"
         return null;
      }
            
      GenericFrame frame = lt.identifyFrame(frameId);
      
      if (frame == null)
      {
         throw new RuntimeException("Invalid frame ID " + frameId);
      }
      
      return frame;
   }
   
   /**
    * Translates a string of text, which is a key function, into the key label
    * of a keyboard key, currently assigned for this function.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KBLABEL </b>
    * 
    * @param   function
    *          text, describing the key function
    * 
    * @return  text, describing the key label. If the key function is
    *          unrecognizable, it is returned translated to upper case
    */
   public static character kbLabel(character function)
   {
      if (function.isUnknown())
         return new character();

      return new character(Keyboard.kbLabel(function.getValue()));
   }

   /**
    * Translates a string of text, which is a key function, into the key label
    * of a keyboard key, currently assigned for this function.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KBLABEL </b>
    * 
    * @param   function
    *          text, describing the key function
    * 
    * @return  text, describing the key label. If the key function is
    *          unrecognizable, it is returned translated to upper case
    */
   public static character kbLabel(String function)
   {
      return new character(Keyboard.kbLabel(function));
   }

   /**
    * Evaluates a key label and returns the corresponding integer key code.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KEYCODE </b>
    * 
    * @param   label
    *          Label to evaluate.
    * 
    * @return  key code.
    */
   public static integer keyCode(character label)
   {
      if (label.isUnknown())
         return new integer();
      
      return new integer(Keyboard.keyCode(label.getValue()));
   }

   /**
    * Evaluates a key label and returns the corresponding integer key code.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KEYCODE </b>
    * 
    * @param   label
    *          Label to evaluate.
    * 
    * @return  key code
    */
   public static integer keyCode(String label)
   {
      return new integer(Keyboard.keyCode(label));
   }

   /**
    * Evaluates key code and returns event name assigned to specified key.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KEYFUNCTION </b>
    * 
    * @param   code
    *          Code to evaluate.
    * 
    * @return  event name string
    */
   public static character keyFunction(integer code)
   {
      if (code.isUnknown())
         return new character();
      
      return new character(Keyboard.keyFunction(code.intValue()));
   }

   /**
    * Evaluates key code and returns a label of specified key.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KEYLABEL </b>
    * 
    * @param   code
    *          Code to evaluate.
    * 
    * @return  key label string
    *  
    */
   public static character keyLabel(integer code)
   {
      if (code.isUnknown())
         return new character();
      
      return new character(Keyboard.keyLabel(code.intValue()));
   }

   /**
    * Evaluates key code and returns a label of specified key.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>KEYLABEL </b>
    * 
    * @param   code
    *          Code to evaluate.
    * 
    * @return  key label string
    *  
    */
   public static character keyLabel(int code)
   {
      return new character(Keyboard.keyLabel(code));
   }
   
   /**
    * Moves specified widget before the widget "before" in the tab order.
    * 
    * @param   fieldGroup
    *          Containing field-group.
    * @param   widget
    *          The widget to move.
    * @param   before            
    *          The widget before which the widget should be moved.
    *          
    * @return  TRUE - the widget was moved successfully, FALSE - otherwise.         
    */
   public static logical moveBeforeTabItem(FieldGroup    fieldGroup, 
                                           CommonWidget  widget,
                                           CommonWidget  before)
   {
      fieldGroup.moveWidgetBefore((GenericWidget<?>)widget, (GenericWidget<?>)before);
      
      updateTabItemList(fieldGroup);
      
      return new logical(true);
   }

   /**
    * Notify the client that the tab item list for the given field group is changed.
    * 
    * @param    fieldGroup
    *           The field group
    */
   public static void updateTabItemList(FieldGroup fieldGroup)
   {
      List<Integer> frameIds = new ArrayList<>();
      List<int[]> itemLists = new ArrayList<>();
      
      gatherTabItemLists(frameIds, itemLists, fieldGroup);
      int[] fids = Utils.integerCollectionToPrimitive(frameIds);
      int[][] items = new int[itemLists.size()][];
      for (int i = 0; i < itemLists.size(); i++)
      {
         items[i] = itemLists.get(i);
      }
      
      LogicalTerminal lt = locate();
      lt.client.updateTabItemList(fids, items);
   }
   
   /**
    * Gather all the information about the updated tab item lists, including all nested child frames.
    * 
    * @param    frameIds
    *           The list where to gather the frame IDs.
    * @param    itemLists
    *           The list where to gather the tab item lists, for each frame.
    * @param    fieldGroup
    *           The field group
    */
   private static void gatherTabItemLists(List<Integer> frameIds, List<int[]> itemLists, FieldGroup fieldGroup)
   {
      int frameId       = fieldGroup.frame.getFrameId();
      int[] tabItemList = fieldGroup.getTabItemList();
      
      frameIds.add(frameId);
      itemLists.add(tabItemList);

      if (tabItemList != null)
      {
         LogicalTerminal lt = locate();
         
         for(int widgetId : tabItemList)
         {
            GenericWidget<?> widget = lt.getWidgetForIdInt(widgetId);
            if (widget instanceof FrameWidget)
            {
               GenericFrame frame = ((FrameWidget) widget).getFrame();
               if (widget._isRealized() && frame != null && frame.hasFieldGroup())
               {
                  handle firstChild = frame.firstChild();
                  if (firstChild._isValid())
                  {
                     CommonWidget fg = firstChild.unwrapWidget();
                     if (fg instanceof FieldGroup)
                     {
                        gatherTabItemLists(frameIds, itemLists, (FieldGroup) fg);
                     }
                  }
               }
            }
         }
      }
   }
   
   /**
    * Moves specified widget after the given widget in the tab order.
    * 
    * @param   fieldGroup
    *          Containing field-group.
    * @param   widget
    *          The widget to move.
    * @param   after            
    *          The widget after which the widget should be placed.
    *          
    * @return  TRUE - the widget was moved successfully, FALSE - otherwise.
    */
   public static logical moveAfterTabItem(FieldGroup    fieldGroup, 
                                          CommonWidget  widget,
                                          CommonWidget  after)
   {
      fieldGroup.moveWidgetAfter((GenericWidget<?>)widget, (GenericWidget<?>)after);
      
      updateTabItemList(fieldGroup);
      
      return new logical(true);
   }

   /**
    * Returns a handle for the instance of the only window, which is always active.
    * <p>
    * This method corresponds to the Progress system handle: <br>
    * <b>ACTIVE-WINDOW </b>
    * 
    * @return  instance of the active window for this session or unknown value for batch-mode
    */
   public static handle activeWindow()
   {
      if (EnvironmentOps.isBatchMode().booleanValue())
      {
         return new handle();
      }  
            
      return locate().activeWindow.asWidgetHandle();
   }

   /**
    * Returns a handle for the instance of the only window, which is always current.
    * <p>
    * This method corresponds to the Progress system handle: <br>
    * <b>CURRENT-WINDOW </b>
    * 
    * @return  a handle to current window or unknown value for batch-mode
    */
   public static handle currentWindow()
   {
      if (EnvironmentOps.isBatchMode().booleanValue())
      {
         return new handle();
      }
      
      return locate().currentWindow.asWidgetHandle();
   }
   
   /**
    * Returns a handle for the instance of the default window.
    * <p>
    * This method corresponds to the Progress system handle: <br>
    * <b>DEFAULT-WINDOW </b>
    * 
    * @return  a handle to default window or unknown value for batch-mode
    */
   public static handle defaultWindow()
   {
      if (EnvironmentOps.isBatchMode().booleanValue())
      {
         return new handle();
      }
      
      return locate().defaultWindow.asWidgetHandle();
   }

   /**
    * Assigns CURRENT-WINDOW system handle to a new window within ACTIVE-WINDOW.
    * 
    * @param   window
    *          a window reference.
    */
   public static void setCurrentWindow(WindowWidget window)
   {
      if (isChui())
      {
          // in chui or batch mode, this is a no-op.
          return;
      }
      
      LogicalTerminal lt = locate();
      
      lt.activeWindow = lt.currentWindow = window;
      lt.state.currentWindow = window.getId();
      
      lt.currentWindowStack.remove(window);
      lt.currentWindowStack.add(window);
   }

   /**
    * Assigns CURRENT-WINDOW system handle to a new window.
    * 
    * @param   h
    *          a handle reference.
    */
   public static void setCurrentWindow(handle h)
   {
      if (h == null || !h._isValid() || h.isUnknown())
      {
         // proc can be full name with path of procedure file or relative.
         String proc = EnvironmentOps.getSourceName(1).getValue();
         String ms = proc + ": Unable to evaluate field for assignment";
         ErrorManager.recordOrThrowError(143, ms);
         
         return;
      }

      // sometimes the handle can refer to the frame inside the window to be set as current
      if (h.getResource() instanceof FrameWidget)
      {
         h = h.unwrapWidget().getWindow();
      }
      
      if (!(h.getResource() instanceof WindowWidget))
      {
         String ms = "Unable to assign widget  to CURRENT-WINDOW";
         ErrorManager.recordOrThrowError(3130, ms, false);
         
         return;
      }

      setCurrentWindow((WindowWidget)h.getResource());
   }

   /**
    * Assigns ACTIVE-WINDOW system handle. This assignment is silently ignored, so this method 
    * does nothing.
    * 
    * @param   window
    *          a window reference.
    */
   public static void setActiveWindow(WindowWidget window)
   {
      // ignore assignment   
   }

   /**
    * Assigns ACTIVE-WINDOW system handle. This assignment is silently ignored, so this method 
    * does nothing.
    * 
    * @param   h
    *          a handle reference.
    */
   public static void setActiveWindow(handle h)
   {
      // ignore assignment
   }

   /**
    * Assigns DEFAULT-WINDOW system handle . This assignment is silently ignored, so this method 
    * does nothing.
    * 
    * @param   window
    *          a window reference.
    */
   public static void setDefaultWindow(WindowWidget window)
   {
      // ignore assignment
   }

   /**
    * Assigns DEFAULT-WINDOW system handle. This assignment is silently ignored, so this method 
    * does nothing.
    * 
    * @param   h
    *          a handle reference.
    */
   public static void setDefaultWindow(handle h)
   {
      // ignore assignment
   }
   
   /**
    * Get CURRENT-WINDOW attribute for given procedure handle.
    * 
    * @param   phandle
    *          Procedure handle.
    *          
    * @return  Reference to current window widget.
    */
   public static handle currentWindow(handle phandle)
   {
      if (EnvironmentOps.isBatchMode().booleanValue())
      {
         return new handle();
      }
      
      return ProcedureManager.getCurrentWindow(phandle.get());
   }
   
   /**
    * Returns the instance of a field level widget which has focus.
    * <p>
    * This method corresponds to the Progress system handle: <br>
    * <b>FOCUS </b>
    * 
    * @return  reference to field level widget which has input focus
    */
   public static handle focus()
   {
      GenericWidget<?> widget = getWidgetForId(locate().client.getFocus());
      
      return (widget == null) ? new handle() : new handle(widget);
   }

   /**
    * Returns the handle of a field level widget with given id.
    * <p>
    * This maps the HWND built-in function, a FWD extension.
    *  
    * @param    id
    *           the id of the widget
    *           
    * @return   reference to field level widget
    */
   public static handle getWidgetHandleById(NumberType id)
   {
      return id == null || id.isUnknown() ? new handle() : getWidgetHandleById(id.intValue());
   }

   /**
    * Returns the handle of a field level widget with given id.
    * <p>
    * This maps the HWND built-in function, a FWD extension.
    *  
    * @param    id
    *           the id of the widget
    *           
    * @return   reference to field level widget
    */
   public static handle getWidgetHandleById(int id)
   {
      GenericWidget<?> widget = getWidgetForId(id);
      
      return (widget == null) ? new handle() : new handle(widget);
   }


   /**
    * Sets the format specification for the specified widget and redraws the
    * widget.
    *
    * @param    widget
    *           The widget to modify.
    * @param    format
    *           The format specification.
    */
   public static void setFormat(int widget, String format)
   {
      locate().client.setFormat(widget, format);
   }
   
   /**
    * Set the new decimal and group separators.
    * 
    * @param    decSep
    *           The new decimal separator.
    * @param    groupSep
    *           The new group separator.
    */
   public static void setNumericFormat(char decSep, char groupSep)
   {
      String newFormat = decSep + "" + groupSep;
      String currentFormat = NumberType.getDecimalSeparator() + "" + NumberType.getGroupSeparator();
      if (currentFormat.equals(newFormat))
      {
         // format doesn't change, be a no-op
         return;
      }
      
      LogicalTerminal lt = locate();
      
      if (newFormat.equals(lt.currentClientNumericFormat))
      {
         // we are changing back to the format which is already set on the client-side, no need to push it. 
         // just reset the pending format change from the server-side
         lt.currentClientNumericFormat = null;
         lt.newClientNumericFormat = null;
         
         return;
      }
      
      if (lt.currentClientNumericFormat == null)
      {
         lt.currentClientNumericFormat = currentFormat;
      }
      
      // save the new format, it will be pushed on the next client API call via ServerState
      lt.newClientNumericFormat = newFormat;
   }
   
   /**
    * Register the frame as a header or footer from the stream that is the
    * current redirected terminal output destination.  The terminal must be
    * non-interactive for this method to process properly.  Subsequent
    * pagination operations on the stream will cause the header or footer to
    * render as needed, without any further explicit request from the server.
    * 
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    */
   public static void registerPageElement(int frameId, boolean header)
   {
      locate().client.registerPageElement(frameId, header);
   }
   
   /**
    * Register the frame as a header or footer from the stream that is the
    * current redirected terminal output destination.  The terminal must be
    * non-interactive for this method to process properly.  Subsequent
    * pagination operations on the stream will cause the header or footer to
    * render as needed, without any further explicit request from the server.
    * 
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    * @param    pulse
    *           <code>true</code> to make possible pulse on close, <code>false</code>
    *           otherwise.
    */
   public static void registerPageElement(int frameId, boolean header, boolean pulse)
   {
      locate().client.registerPageElement(frameId, header, pulse);
   }
   
   /**
    * Register the frame as a header or footer with the given stream as the
    * redirected terminal output destination.  Subsequent pagination
    * operations on the stream will cause the header or footer to render
    * as needed, without any further explicit request from the server.
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    * @param    streamId
    *           The stream to associate with the frame as a page element.
    */
   public static void registerPageElement(int     frameId,
                                          boolean header,
                                          int     streamId)
   {
      locate().client.registerPageElement(frameId, header, streamId);
   }
   
   /**
    * Deregister (remove) the frame as a header or footer from the stream 
    * that is the current redirected terminal output destination.  The
    * terminal must be non-interactive for this method to process properly.
    * 
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    */
   public static void deregisterPageElement(int frameId, boolean header)
   {
      locate().client.deregisterPageElement(frameId, header);
   }
   
   /**
    * Deregister (remove) the frame as a header or footer from the given
    * stream.
    * <p>
    * This is called (among other places) during hide processing on the
    * server, which must be safe to be called at any time.  This is true,
    * even if we aren't redirected, if the frame is not visible and/or if
    * there was no page element ever registered for this frame (because it
    * was never viewed).
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    * @param    streamId
    *           The stream to associate with the frame as a page element.
    */
   public static void deregisterPageElement(int     frameId,
                                            boolean header,
                                            int     streamId)
   {
      locate().client.deregisterPageElement(frameId, header, streamId);
   }
   
   /**
    * Deregister (remove) the frame as a header or footer from all streams
    * in which it is registered.
    *
    * @param    frameId
    *           The id of the frame for this request.
    * @param    header
    *           <code>true</code> to register as a header, <code>false</code>
    *           to register as a footer.
    */
   public static void forceDeregisterPageElement(int frameId, boolean header)
   {
      locate().client.forceDeregisterPageElement(frameId, header);
   }
   
   /**
    * Returns the instance of a widget associated with the currently executing
    * trigger.
    * 
    * @return  reference to widget for which trigger is executed
    */
   public static handle _self()
   {
      LogicalTerminal lt = locate();
      
      LinkedList<WrappedResource> list = lt.triggerResource;
      
      return (list.isEmpty()) ? new handle() : new handle(lt.triggerResource.getFirst());
   }
   
   /**
    * Applies the named event to the specified resource.  If the resource is a valid widget, the
    * work is delegated to the widget's apply API.
    *
    * @param    h
    *           The resource to which the event is applied.
    * @param    event
    *           Progress event name
    */
   public static void apply(handle h, character event)
   {
      if (event.isUnknown())
      {
         // this is a no-op in case of unknown events
         return;
      }
      
      // find a code for the named event
      int eventCode = Keyboard.encodeEvent(event.toStringMessage());

      if (eventCode == -1)
      {
         LOG.log(Level.WARNING, "Could not recognize event [" + event.toStringMessage() + "].");
         return;
      }

      apply(h, eventCode);
   }

   /**
    * Applies the named event to the specified resource.  If the resource is a valid widget, the
    * work is delegated to the widget's apply API.
    *
    * @param    h
    *           The resource to which the event is applied.
    * @param    event
    *           Progress event name
    */
   public static void apply(handle h, String event)
   {
      if (event == null)
      {
         // this is a no-op in case of unknown events
         return;
      }
      
      // find a code for the named event
      int eventCode = Keyboard.encodeEvent(event);

      if (eventCode == -1)
      {
         LOG.log(Level.WARNING, "Could not recognize event [" + event + "].");
         return;
      }

      apply(h, eventCode);
   }
   
   /**
    * Applies the named event to the specified resource.  If the resource is a valid widget, the
    * work is delegated to the widget's apply API.
    *
    * @param    h
    *           The resource to which the event is applied.
    * @param    eventCode
    *           Progress event code
    */
   public static void apply(handle h, double eventCode)
   {
      if (h.isUnknown() || !h._isValid())
      {
         applyToHandleError(true);
         
         return;
      }

      WrappedResource res = h.getResource();
      
      apply(res, eventCode);
   }

   /**
    * Applies the named event to the specified resource.  If the resource is a valid widget, the
    * work is delegated to the widget's apply API.
    *
    * @param    h
    *           The resource to which the event is applied.
    * @param    eventCode
    *           Progress event code
    */
   public static void apply(handle h, integer eventCode)
   {
      if (eventCode.isUnknown())
      {
         // this is a no-op in case of unknown events
         return;
      }
      
      long evtCodeLong = eventCode.getValue();
      if (evtCodeLong == -1L)
      {
         // this is a no-op too in case of code is -1
         return;
      }
      
      apply(h, evtCodeLong);
   }
   
   /**
    * Applies the named event to the specified resource.  If the resource is a valid widget, the
    * work is delegated to the widget's apply API.
    *
    * @param    res
    *           The resource to which the event is applied.
    * @param    event
    *           Progress event code
    */
   public static void apply(WrappedResource res, character event)
   {
      if (event.isUnknown())
      {
         // this is a no-op in case of unknown events
         return;
      }
      
      // find a code for the named event
      int eventCode = Keyboard.encodeEvent(event.getValue());

      if (eventCode == -1)
      {
         LOG.log(Level.WARNING, "Could not recognize event [" + event.toStringMessage() + "].");
         return;
      }

      apply(res, eventCode);
   }

   /**
    * Applies the named event to the specified resource.  If the resource is a valid widget, the
    * work is delegated to the widget's apply API.
    *
    * @param    res
    *           The resource to which the event is applied.
    * @param    event
    *           Progress event code
    */
   public static void apply(WrappedResource res, String event)
   {
      // find a code for the named event
      int eventCode = Keyboard.encodeEvent(event);

      if (eventCode == -1)
      {
         LOG.log(Level.WARNING, "Could not recognize event [" + event + "].");
         return;
      }

      apply(res, eventCode);
   }
   
   /**
    * Applies the named event to the specified resource.  If the resource is a valid widget, the
    * work is delegated to the widget's apply API.
    *
    * @param    res
    *           The resource to which the event is applied.
    * @param    eventCode
    *           Progress event code
    */
   public static void apply(WrappedResource res, integer eventCode)
   {
      if (eventCode.isUnknown())
      {
         // this is a no-op in case of unknown events
         return;
      }
      
      long evtCodeLong = eventCode.getValue();
      if (evtCodeLong == -1L)
      {
         // this is a no-op too in case of code is -1
         return;
      }
      
      apply(res, evtCodeLong);
   }
   
   /**
    * Applies the named event to the specified resource.  If the resource is a valid widget, the
    * work is delegated to the widget's apply API.
    *
    * @param    res
    *           The resource to which the event is applied.
    * @param    eventCode
    *           Progress event code
    */
   public static void apply(WrappedResource res, double eventCode)
   {
      boolean valid = res.valid();
      
      if (valid)
      {
         // check if resource is either widget or procedure (only these can be used!)
         if (res instanceof GenericWidget)
         {
            GenericWidget<?> gw = (GenericWidget<?>) res;
            
            if (gw.getWindow()._isValid())
            {
               gw.apply(eventCode);
            
               return;
            }
         }
         
         if (res instanceof GenericFrame)
         {
            GenericFrame gf = (GenericFrame) res;
            
            if (gf.getFrameWidget().getWindow()._isValid())
            {
               gf.apply(eventCode);
            
               return;
            }
         }
         
         if (!HandleOps.isSystemHandle(res))
         {
            LogicalTerminal lt = locate();
            
            // merge all triggers
            EventList wl = (lt.cachedEventList == null 
                              ? lt.cachedEventList = TriggerManager.mergeTriggerRegistry(lt.scope) 
                              : lt.cachedEventList);
            
            long resId = handle.resourceId(res);
   
            if (Keyboard.isDevEvent((int) eventCode)                  && 
                !lt.isWaitEvent((int) eventCode, Keyboard.SE_ANY_KEY) &&
                !wl.isEventMatchForWidget(-1, (int) eventCode, Keyboard.SE_ANY_KEY))
            {
               // there is no trigger for this user event (U1..U10), nothing to do 
               return;
            }

            lt.client.apply(resId, (int) eventCode, wl);
            
            return;
         }
      }
      
      // fallback, target can not be found or is invalid
      applyToHandleError(!valid);
   }
   
   /**
    * Applies the named event to the widget currently in focus in the given
    * frame.
    *
    * @param frame
    *        instance of <code>GenericFrame</code> this request came for
    * @param event
    *        Progress event name
    */
   public static void apply(GenericFrame frame, String event)
   {
      // find a code for the named event
      int eventCode = Keyboard.encodeEvent(event);

      if (eventCode == -1)
      {
         LOG.log(Level.WARNING, "Could not recognize event [" + event + "].");
         return;
      }

      // apply the encoded event
      apply(frame, frame.getFrameId(), eventCode);
   }

   /**
    * Applies the encoded event to the widget currently in focus in the given
    * frame.
    *
    * @param frame
    *        instance of <code>GenericFrame</code> this request came for
    * @param eventCode
    *        Progress event code
    */
   public static void apply(GenericFrame frame, int eventCode)
   {
      apply(frame, frame.getFrameId(), eventCode);
   }

   /**
    * Applies the named event to the specified widget.
    *
    * @param frame
    *        instance of <code>GenericFrame</code> this request came for
    * @param widgetId
    *        widget to apply the event to
    * @param event
    *        Progress event name
    */
   public static void apply(GenericFrame frame, int widgetId, String event)
   {
      // find a code for the named event
      int eventCode = Keyboard.encodeEvent(event);

      if (eventCode == -1)
      {
         LOG.log(Level.WARNING, "Could not recognize event [" + event + "].");
         return;
      }

      // apply the encoded event
      apply(frame, widgetId, eventCode);
   }

   /**
    * Applies the encoded event to the specified widget.
    * 
    * @param frame
    *        instance of <code>GenericFrame</code> this request came for
    * @param widgetId
    *        widget to apply the event to
    * @param eventCode
    *        Progress event code
    */
   public static void apply(GenericFrame frame, int widgetId, int eventCode)
   {
      if (widgetId == -1)
      {
         // if this is not a valid widget, show error and return. although the
         // message is formated as an abend message, this is not a real abend
         ErrorManager.displayAbend(3190, "Could not find target of APPLY statement");
         return;
      }

      ScreenBuffer sb[] = null;
      
      LogicalTerminal lt = locate();

      // we pass down the screen buffers even if the frame == null      
      sb = lt.getEditableServerScreenBuffers(frame);

      // merge all triggers
      EventList wl = (lt.cachedEventList == null 
                        ? lt.cachedEventList = TriggerManager.mergeTriggerRegistry(lt.scope) 
                        : lt.cachedEventList);
      
      if (Keyboard.isDevEvent(eventCode)                   &&
          !lt.isWaitEvent(eventCode, Keyboard.SE_ANY_KEY)  &&
          !wl.isEventMatchForWidget(widgetId, eventCode, Keyboard.SE_ANY_KEY))
      {
         // there is no trigger for this user event (U1..U10), nothing to do 
         return;
      }
      
      sb = lt.client.apply(widgetId, eventCode, sb, wl, true);
      
      handleScreenBuffers(frame, sb, false);
   }

   /**
    * Applies the named event to the widget in focus in the active frame.
    *
    * @param event
    *        Progress event name
    */
   public static void apply(String event)
   {
      // find a code for the named event
      int eventCode = Keyboard.encodeEvent(event);

      if (eventCode == -1)
      {
         LOG.log(Level.WARNING, "Could not recognize event [" + event + "].");
         return;
      }

      // apply the encoded event, allow negative event codes
      applyWorker(eventCode, false);
   }

   /**
    * Applies the named event to the widget in focus in the active frame.
    *
    * @param event
    *        Progress event name
    */
   public static void apply(character event)
   {
      if (event.isUnknown())
      {
         // this is a no-op in case of unknown events
         return;
      }
      
      apply(event.toStringMessage());
   }

   /**
    * Applies the encoded event to the widget in focus in the active frame.
    *
    * @param eventCode
    *        Progress event code
    */
   public static void apply(int64 eventCode)
   {
      if (eventCode.isUnknown())
         throw new RuntimeException("event cannot be unknown");
      
      apply(eventCode.intValue());
   }

   /**
    * Applies the encoded event to the widget in focus in the active frame.
    *
    * @param eventCode
    *        Progress event code
    */
   public static void apply(double eventCode)
   {
      apply((int)eventCode);
   }

   /**
    * Applies the encoded event to the widget in focus in the active frame.
    *
    * @param eventCode
    *        Progress event code
    */
   public static void apply(long eventCode)
   {
      // apply the encoded event, ignore negative event codes
      applyWorker(eventCode, true);
   }
   
   /**
    * Called from the client-side, it executes the server-side event by calling 
    * {@link ServerEvent#invoke(int)}.
    * 
    * @param    id
    *           The id of the server event.
    *
    * @return   The response to a server-side code execution.
    */
   public static ServerReturnValue invokeServerEvent(int id)
   {
      ServerReturnValue srv = new ServerReturnValue();
      
      LogicalTerminal lt = locate();
      lt.invokeOnServer(srv, () -> { ServerEvent.invoke(id); return null; });
      
      return srv;
   }
   
   /**
    * Invokes a command on the client widget implementation.
    *
    * @param   command
    *          The command to invoke.
    * @param   args
    *          Arguments.
    *
    * @return  Return value.
    */
   public static Object invokeWidgetCommand(CommonWidget widget, Enum command, Object... args)
   {
      try
      {
         return getClient().invokeWidgetCommand(widget.getId(), command, args);
      }
      catch (UserErrorException ex)
      {
         ErrorManager.displayError(ex.getMessage());
         return null;
      }
   }
   
   /**
    * Check if the resource with the given id is still valid to be used in a WAIT-FOR statement.
    * It will return <code>false</code> if:
    * <ul>
    *    <li>there is no resource associated with the given id or the resource is unknown.</li>
    *    <li>the SENSITIVE attribute for the resource is <code>false</code>.</li>
    *    <li>the resource is not valid.</li>
    * </ul>
    * In all other cases, it will return <code>true</code>
    * 
    * @param    id
    *           The resource id.
    * 
    * @return   See above.
    */
   public static boolean validResource(long id)
   {
      handle h = handle.fromString(Long.toString(id));
      if (h == null || h.isUnknown())
      {
         return false;
      }
      
      WrappedResource res = h.getResource();

      // check the SENSITIVE attribute first
      if (res instanceof Sensitive)
      {
         return ((Sensitive) res).isSensitive().booleanValue();
      }
      
      // check if the resource is valid last
      return h._isValid();
   }
   
   /**
    * Post the server event to the client side. The event will be executed only on the next 
    * PROCESS-EVENTS or UI-blocking statement.
    * 
    * @param    event
    *           The server event.
    */
   public static void postServerEvent(ServerEvent event)
   {
      LogicalTerminal lt = locate();
      
      lt.client.postServerEvent(event);
   }
   
   /**
    * Remove from the server event queue all the events in the given set, identified by their id.
    * 
    * @param    eventIds
    *           The set of event IDs to be removed.
    */
   public static void removeServerEvents(Set<Integer> eventIds)
   {
      LogicalTerminal lt = locate();
      
      lt.client.removeServerEvents(Utils.integerCollectionToPrimitive(eventIds));
   }

   /**
    * Called by the P2J client side, it will inform the server that a server-event was generated.
    * Although the event originates from the P2J Client Side, socket events are server-side 
    * events.
    * 
    * @param    resourceId
    *           The resource ID for the socket to which the READ-RESPONSE event belongs.
    *           
    * @return   The ID of the generated server event or <code>-1</code> if a event could not be
    *           posted (i.e. the resource is not SENSITIVE).
    */
   public static int readResponseEvent(long resourceId)
   {
      handle h = handle.fromString(Long.toString(resourceId));
      if (!h._isValid()) 
      {
         return -1; // the resource was deleted before the event can be processed.
      }

      SocketImpl socket = (SocketImpl) h.getResource();
      
      return socket.readResponseEvent();
   }
   
   /**
    * Called by the P2J client side, it will inform the server that a server event was generated.
    * Although the event originates from the P2J Client Side, socket events are server-side 
    * events.
    * 
    * @param    resourceId
    *           The resource ID for the server socket to which the CONNECT event belongs.
    * @param    clientSocketId
    *           The resource ID of the socket created for the new connection accepted by the
    *           server socket.
    *           
    * @return   The ID of the generated server event or <code>-1</code> if a event could not be
    *           posted (i.e. the resource is not SENSITIVE).
    */
   public static int connectEvent(long resourceId, long clientSocketId)
   {
      handle h = handle.fromString(Long.toString(resourceId));
      if (!h._isValid()) 
      {
         return -1; // the resource was deleted before the event can be processed.
      }
      
      SocketListenerImpl serverSocket = (SocketListenerImpl) h.getResource();

      return serverSocket.connectEvent(clientSocketId);
   }
   
   /**
    * Create a socket and associated with the given ID.  The received ID is not a resource ID; 
    * this represents the ID of the associated {@link LowLevelSocketImpl} instance.
    *   
    * @param    id
    *           The ID of the associated {@link LowLevelSocketImpl} instance.
    *           
    * @return   The ID of the created {@link SocketImpl} resource.
    */
   public static long createSocket(int id)
   {
      return SocketImpl.createSocket(id);
   }
   
   /**
    * Check the directory for the UI {@link com.goldencode.p2j.ui.client.gui.theme.Theme}. If
    * any {@code Theme} is configured in registry (in any scope) it is returned, otherwise
    * {@code null} is returned so that the client will chose the default. 
    */
   public static String getUiTheme()
   {
      return Utils.getDirectoryNodeString(
            null, "theme/name", "Windows10Theme", Utils.DirScope.BOTH, null);
   }
   
   /**
    * Gets the values for uploaded file size limits from server.
    *
    * @return   The file size limits values.  Index 0 is for per file limit, 1 - total limit,
    *           2 - the file size threshold for writing data to the disk
    */
   public static int[] getUploadFileSizeLimits()
   {
      int [] result = new int[3];

      // first get the per file limit
      String fileLimit = Utils.getDirectoryNodeString(
            null, "webClient/uploadSingleFile", "20M", Utils.DirScope.BOTH, null);
      // abd get the per operation limit
      String totalFilesLimit = Utils.getDirectoryNodeString(
            null, "webClient/uploadTotalFiles", "100M", Utils.DirScope.BOTH, null);
      //the file size threshold for writing data to the disk
      String fileSizeThresholdForDiskWriting = Utils.getDirectoryNodeString(
               null, "webClient/fileSizeThresholdForDiskWriting", "2M", Utils.DirScope.BOTH, null);
      // convert string values into intgers
      result[0] = Utils.getSizeLimitFromString(fileLimit);
      result[1] = Utils.getSizeLimitFromString(totalFilesLimit);
      result[2] = Utils.getSizeLimitFromString(fileSizeThresholdForDiskWriting);
      
      // and return to client
      return result;
   }
   
   /**
    * Gets the value for sub menu show/hide delay in GUI mode.  0 value disables delay
    * functionality.
    *   
    * @param    defaultDelay
    *           The default value to be used if the current value is nod configured on server
    *           side directory config file.
    *
    * @return  The sub menu delay in milliseconds fo show/hide submenu or 0 to disable delay.
    */
   public static int getGuiSubMenuDelay(int defaultDelay)
   {
      return Utils.getDirectoryNodeInt(null, "theme/submenu-delay", defaultDelay,
                                       Utils.DirScope.BOTH, null);
   }
   
   /**
    * Set the theme preferred fonts.
    *
    * @param   themeFonts
    *          The fonts the current installed {@code Theme} on client prefers.
    * @param   themeFontAliases
    *          The font aliases currently installed {@code Theme} on client prefers.
    */
   public static void setThemePreferredFonts(Map<String, String> themeFonts,
                                             Map<String, String> themeFontAliases)
   {
      FontTable.setThemePreferredFonts(themeFonts, themeFontAliases);
   }
   
   /**
    * Read the entire font-table.
    * 
    * @return   The font table.  As 4GL does, this is 0-indexed.
    */
   public static FontDetails<?>[] readFontTable()
   {
      return FontTable.readFontTable();
   }

   /**
    * Read the font-table with the name of the legacy system fonts.
    *
    * @return  The font table.  As 4GL does, this is 0-indexed.
    */
   public static FontDetails<?>[] readSystemFontTable()
   {
      return FontTable.readSystemFontTable();
   }
   
   /**
    * Read the default fonts.
    *
    * @param   useIni
    *          <code>true</code> to load the default fonts from the ini file specified for the client.
    *          <code>false</code> to load the default fonts from the directory.
    *
    * @return  An array with the font details, with default font on index o and default fixed
    *          font on index 1.
    */
   public static FontDetails<?>[] readDefaultFonts(boolean useIni)
   {
      return FontTable.readDefaultFonts(useIni);
   }

   /**
    * Returns font data in a form of map where key is the lower-cased name as defined in server
    * directory "custom-font" container and map value is the raw file data of respective font
    * files.  For bold font entries (node "bold" set to true) the key has the form
    * '&lt;lower-cased font name&gt;#true', otherwise the key has the form
    * '&lt;lower-cased font name&gt;#false'. The caller must not modify the byte array map values.
    *
    * @return  See above.
    */
   public static Map<String, byte[]> getCustomFontData()
   {
      return FontTable.getCustomFontData();
   }

   /**
    * Emits the supplied COM event.
    *
    * @param   widgetId
    *          Id of the widget that will become the source of the emitted event.
    * @param   eventName
    *          Event name.
    * @param   args
    *          Event arguments.
    *
    * @return  The array of all output arguments produced by the event, if any.
    */
   public static BaseDataType[] emitComEvent(int widgetId, String eventName, BaseDataType... args)
   {
      LogicalTerminal lt = instance.get();
      Object widget = lt.widgetRegistry.get(widgetId);

      if (widget == null)
      {
         LOG.log(Level.WARNING, "COM event " + eventName + " cannot be emitted. Widget with id " +
                                widgetId + " was not found.");
         return null;
      }

      if (!(widget instanceof EmitsComEvents))
      {
         throw new RuntimeException("COM event " + eventName + " cannot be emitted. Widget [" +
                                    widget + "] is not capable to emit com events.");
      }

      return ((EmitsComEvents) widget).emitComEvent(eventName, args);
   }

   /**
    * Clients use this method to request node expansion status change.
    * 
    * @param treeWidgetId
    *        the tree widget ID
    * @param nodeId
    *        the node ID
    * @param expand
    *        the new expansion status
    */
   public static void requestNodeExpand(final int treeWidgetId, final int nodeId, final boolean expand)
   {
      LogicalTerminal lt = instance.get();
      Object widget = lt.widgetRegistry.get(treeWidgetId);

      if (widget == null)
      {
         LOG.log(Level.WARNING,
                  "treeNodeCollapsed: widget with id " + treeWidgetId + " was not found.");
         return;
      }

      ((TreeWidgetBase) widget).requestNodeExpand(nodeId, expand);
   }
   
   /**
    * Get the legacy metrics for the given text, using the font specified by the given key.
    * 
    * @param    text
    *           The text.
    * @param    key
    *           The font's key.
    *           
    * @return   The legacy metrics, with width on index 0 and height on index 1.
    */
   public static int[] getLegacyTextMetrics(String text, String key)
   {
      return FontTable.getLegacyTextMetrics(text, key);
   }
   
   /**
    * Gets the client local storage.
    * 
    * @return   The client local storage
    */
   public static ObjectStorage getClientStorage()
   {
      return locate().clientLocalStorage;
   }
   
   /**
    * Read the legacy text metrics for all the texts and each font defined in the font table, for
    * this user.
    *
    * @return  A map having as key the font and text and as value the metrics (0 index - width, 
    *          1 index - height).
    */
   public static Map<String, int[]> readLegacyTextMetrics()
   {
      // TODO: improve this as currently can transfer a large amount of data

      // WARNING: it is not completely clear yet if the legacy text metrics are needed or not; 
      // they were initially added to ensure that the text metrics match 100% with the legacy 
      // system, as in some cases (label, text and other widgets) they are used to compute the 
      // implicit widget size.  We can't have any deviation from this size, as it affects the 
      // implicit frame layout.
      
      return FontTable.readLegacyTextMetrics();
   }

   /**
    * Sets the new application icon for given window on the client side.
    * 
    * @param    windowId
    *           A valid window id.
    * @param    small
    *           Flag indicating usage of the small icon case(16x16 in size).
    * @param    iconName
    *           The icon name to be loaded on server side and sent to client.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public static boolean setWindowIcon(int windowId, boolean small, String iconName)
   {
      boolean result = false;
      
      // check the application jar file first
      InputStream is = getImageStreamFromApplication(iconName);
      // found the icon 
      if (is != null)
      {
         byte[] imgData = UiUtils.getData(is);
         if (imgData != null)
         {
            // route call to client side
            result = locate().client.setWindowIcon(windowId, imgData);
         }
      }
      else
      {
         result = locate().client.loadWindowIcon(windowId, small, iconName);
      }
      
      return result;
   }
   
   /**
    * Sets the new default application icon on the client side.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public static boolean setWindowDefaultIcon()
   {
      // first we need to get the name for possible override from directory file
      Directory dir = DirectoryManager.getInstance();
      String iconFileName =
         dir.getString(Directory.ID_RELATIVE, WindowConfig.APP_ICON_SECTION, null);

      // got the valid filename
      if (iconFileName != null && iconFileName.length() > 0)
      {
         // check the application jar file first
         InputStream is = getImageStreamFromApplication(iconFileName);
         // found the icon 
         if (is != null)
         {
            // get the icon bytes
            byte[] iconData = UiUtils.getData(is);
            // data is ready to be sent
            if (iconData != null)
            {
               return locate().client.setWindowDefaultIcon(
                         Window.IconDataType.ICON_CUSTOM_JAR_RESOURCE, iconData);
            }
         }
         else
         {
            // if we are here - no application jar icon or it can not be loaded try to search
            // in PROPATH file system
            return locate().client.setWindowDefaultIcon(
                            Window.IconDataType.ICON_CUSTOM_CLIENT_FILENAME, iconFileName);
         }
      }
      
      // if we are here - the last chance is to load embedded icon from the client side p2j.jar
      return locate().client.setWindowDefaultIcon(Window.IconDataType.ICON_DEFAULT, null);
   }
   
   /**
    * Sets the new image for given button or image widgets on the client side.
    * 
    * @param    w
    *           The widget.
    * @param    type
    *           Image type to load.  Up, Down or Insensitive for buttons, -1 for image.
    * @param    imageName
    *           The image name to be loaded on server side and sent to client.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public static boolean setImage(final GenericWidget w, int type, String imageName)
   {
      // simple check for given widget is neither button nor image
      if (!(w instanceof ButtonWidget || w instanceof ImageWidget))
      {
         return false;
      }

      final int id = w.getId();

      // button widget with prepackaged images should bypass the application jar usage
      // these images will be loaded from p2j.jar on the client side
      if (!(w instanceof ButtonWidget) || !ImageHelper.isPrepackaged(imageName))
      {
         // check the application jar file
         try (final InputStream is = getImageStreamFromApplication(imageName))
         {
            if (is != null)
            {
               // found the image 
               final byte[] imgData = UiUtils.getData(is);
               if (imgData != null)
               {
                  // route call to client side
                  return locate().client.setImage(id, type, imgData);
               }
               
               return false;
            }
         }
         catch (IOException e)
         {
            // no-op
         }
      }

      // ask the client to search the local file system for image
      try
      {
         if (locate().client.loadImage(id, type, imageName))
         {
            return true;
         }
         ErrorManager.displayWarning(206,
                  "The file (" + imageName + ") for LOAD-IMAGE on the IMAGE "
                           + w.name().toStringMessage() + " is invalid on the current system.",
                  false);
      }
      catch (ErrorConditionException e)
      {
         ErrorManager.recordOrShowWarning(e.getKeyCode(), e.getMessage(), false, false, false,
                  false);
      }

      return false;
   }

   /**
    * Loads the supplied image from the supplied file name and registers the image with
    * {@link com.goldencode.p2j.ui.client.ImageHelper#addGlobalImage(int,
    * com.goldencode.p2j.ui.client.gui.driver.ImageWrapper)}
    * on the client. When the image is loaded successfully the returned value is the unique id
    * of the loaded image,
    * 
    * @see      com.goldencode.p2j.ui.client.gui.driver.ImageWrapper#getUniqueId()
    *
    * @param    wId
    *           A valid widget id.
    * @param    imgFile
    *           Valid file name for the image to be loaded, must not be {@code null}.
    *
    * @return   unique image id if operation successful or <code>-1</code> otherwise.
    */
   public static int loadGlobalImage(int wId, String imgFile)
   {
      byte[] data = null;
      InputStream is = getImageStreamFromApplication(imgFile);
      if (is != null)
      {
         data = UiUtils.getData(is);
      }

      return locate().client.addGlobalImage(wId, imgFile, data);
   }
   
   /**
    * Removes all previously registered global images for a specific widget.
    *
    * @param    wId
    *           Widget id of the widget for which images should be cleared.
    */
   public static void removeGlobalImages(int wId)
   {
      locate().client.removeGlobalImages(wId);
   }
   
   /**
    * Removes selected global images for a specific widget.
    *
    * @param   widgetId
    *          Widget ID of the widget for which images should be cleared.
    * @param   uniqueIds
    *          image IDs to be cleared
    */
   public void removeGlobalImages(final int widgetId, final Collection<Integer> uniqueIds)
   {
      locate().client.removeGlobalImages(widgetId, uniqueIds);
   }
   
   /**
    * Load custom mouse pointer from a file
    * 
    * @param id
    *        widget id
    * @param ptrName
    *        the name of the pointer data
    * @return <code>true</code> if successful.
    */
   public static boolean loadMousePointer(int id, String ptrName)
   {
      MousePointer ptr = ptrName == null ? null : MousePointer.fromString(ptrName);
      if (ptrName == null || ptr != null)
      {
         // Valid pre-defined cursor
         if (ptr != null && ptr.isIgnored())
         {
            return true;
         }
         // set pre-defined cursor or reset to default
         return locate().client.setMousePointer(id, ptr);
      }
      
      // cursor from a file, try to load at the server side first
      InputStream is = getImageStreamFromApplication(ptrName);
      if (is != null)
      {
         byte[] data = UiUtils.getData(is);
         if (data != null)
         {
            return locate().client.setMousePointer(id, data, ptrName);
         }
      }
      
      return locate().client.loadMousePointer(id, ptrName);
   }
   
   /**
    * Load custom resource from application jar file.
    * 
    * @param   resource
    *          The name of the resource to be loaded on server side.
    *
    * @return  Stream object to load image data or {@code null} if data is not found.
    */
   public static InputStream getResourceStreamFromApplication(String resource)
   {
      return getResourceStreamFromApplication(resource, null);
   }
   
   /**
    * Load custom resource from application jar file.
    * 
    * @param   resource
    *          The name of the resource to be loaded on server side.
    * @param   extensions
    *          Optional list of extensions to apply to the resource filename. If this list is
    *          non-null, {@code resource} is assumed to be the file's base name, sans extension.
    *
    * @return  Stream object to load image data or {@code null} if data is not found.
    */
   public static InputStream getResourceStreamFromApplication(String resource,
                                                              List<String> extensions)
   {
      Utils.JarResource jarEntry = FileSystemOps.searchResourceJars(resource, extensions);
      
      if (jarEntry != null)
      {
         return jarEntry.getLoader().getResourceAsStream(jarEntry.getName());
      }
      
      return null;
   }
   
   /**
    * Applies the encoded event to the widget in focus in the active frame.
    *
    * @param eventCode
    *        Progress event code
    * @param ignore
    *        Ignore negative event codes to be processed normally.
    */
   private static void applyWorker(long eventCode, boolean ignore)
   {
      if (eventCode < 0)
      {
         if (eventCode == -2 && isChui())
         {
            LOG.log(Level.WARNING,
                    "Probably implementation is missing: APPLY -2 in ChUI should act as APPLY \"ENDKEY\".");
         }
         // ignore negative event numbers if not forced
         if (ignore)
         {
            return;
         }
      }

      LogicalTerminal lt = locate();

      GenericFrame frame = null;
      handle focus = focus();
      int widgetId = -1;
      
      if (focus.isUnknown())
      {
         // no target can be found, APPLY call will send the event to the
         // client, but no triggers should be executed; widgetId remains on -1
      }
      else if (focus.getResource() instanceof CommonWidget)
      {
         CommonWidget widget = (CommonWidget) focus.getResource();
         
         // implicit APPLY calls target the actual focus
         widgetId = widget.getId();
         frame = widget.getFrame();
      }
      else
      {
         // TODO: focus is not a widget. what to do ?
      }
      
      // merge all triggers
      EventList wl = (lt.cachedEventList == null 
                        ? lt.cachedEventList = TriggerManager.mergeTriggerRegistry(lt.scope) 
                        : lt.cachedEventList);

      if (Keyboard.isDevEvent((int) eventCode)                   && 
          !lt.isWaitEvent((int) eventCode, Keyboard.SE_ANY_KEY)  &&
          !wl.isEventMatchForWidget(-1, (int) eventCode, Keyboard.SE_ANY_KEY))
      {
         // there is no trigger for this user event (U1..U10), nothing to do 
         return;
      }
      
      ScreenBuffer[] in = (frame == null) ? null : lt.getEditableServerScreenBuffers(frame);
      ScreenBuffer[] sb = lt.client.apply(widgetId, (int)eventCode, in, wl, false);
      handleScreenBuffers(frame, sb, false);
   }

   /**
    * Gets the image stream from application jar file or {@code null} if the resource is not found.
    * 
    * @param   image
    *          The image name to be loaded on server side.
    *
    * @return  Stream object to load image data or {@code null} if data is not found.
    */
   private static InputStream getImageStreamFromApplication(String image)
   {
      List<String> extensions = null;
      
      if (Utils.getExt(image).length() == 0)
      {
         boolean caseSens = EnvironmentOps.getLegacyCaseSensitive();
         extensions = caseSens ? ImageExt.EXT_VARIANTS : ImageExt.EXT_BASES;
      }
      
      return getResourceStreamFromApplication(image, extensions);
   }
   
   /**
    * Load the specified environment as resource.  The environment is stored in an INI file
    * embedded as a resource inside application jar file.
    * 
    * @param    env
    *           The INI file source.
    * 
    * @return   The environment file bytes, or {@code null} if the environment file could not be
    *           loaded.
    */
   public static byte[] loadEnvironment(String env)
   {
      byte data[] = null;
      
      try
      {
         URL resource = Resources.getResource(env);
         data = Resources.toByteArray(resource);
      }
      catch (Exception e)
      {
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, "Could not load INI file from source: " + env, e);
         }         
      }
      
      return data;
   }
   
   /**
    * Performs CHOOSE processing.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>CHOOSE COLOR </b>
    *
    * @param   frame
    *          instance of <code>GenericFrame</code> this request came for
    * @param   c
    *          CHOOSE statement options
    * @param   cs
    *          color specification for use with CHOOSE
    */
   public static void choose(GenericFrame frame, Choose c, ColorSpec cs)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      // disable loop protection and increment interactions counter
      lt.markInteractive();
      
      c.setColor(cs);
      
      try
      {
         ChooseData inc = c.getData(); 
         inc.setScreenBuffers(lt.getEditableServerScreenBuffers(frame));
         lt.frameRef = frame;
         
         EventList wait = new EventList(true);
         wait.addGoOn(inc.getGoOn());
         wait.setOperation(EventList.CHOOSE);
         
         lt.waitStack.push(wait);
         
         EventList wl = lt.prepareEventList(frame, wait);
         
         ChooseData res = lt.client.choose(frame.getFrameId(), inc, wl);
         
         Accessor kv = c.getKeysVar();
         
         // make the list of keys available to the caller if requested
         if (kv != null)
         {
            kv.set(new character(res.getKeys()));
         }
         
         // update the screen buffer and lastkey (this is the key that was
         // used to exit the choose)
         ScreenBuffer[] sbRes = res.getScreenBuffers();
         handleScreenBuffers(frame, sbRes, !inc.isFieldMode());

         if (sbRes != null)
         {
            // update the current iteration of the current frame
            processDownConfigs(sbRes[0].getAndClearDownConfigs());
         }
      }
      catch (ConditionException ce)
      {
         // make sure that the lastkey is properly maintained
         int kc = ce.getKeyCode();
         
         // anything other than -1 is a valid key press AND even -1 is a
         // valid indicator of CTRL-C if this was a STOP 
         if (kc != -1 || ce instanceof StopConditionException)
         {
            KeyReader.setLastKey(kc);
         }
         
         // continue the unwinding of the stack
         throw ce;
      }
      finally
      {
         lt.waitStack.pop();
      }
   }
   
   /**
    * Clears the data and colors for all fill-in fields of the frame.
    * For down frames, also clears side labels.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>CLEAR FRAME</b>
    * 
    * @param   frame
    *          instance of <code>GenericFrame</code> this request came for
    * @param   noPause
    *          <code>true</code> if NO-PAUSE option was used
    */
   public static void clear(GenericFrame frame, boolean noPause)
   {
      LogicalTerminal lt = locate();
      
      ScreenBuffer[] sb = lt.client.clear(frame.getFrameId(), noPause);
      
      handleScreenBuffers(sb);

      if (sb != null && frame == lt.frameRef)
      {
         String sv = sb[0].getFrameValue();
         lt.frameValue = (sv == null ? "" : sv);
      }
   }
      
   /**
    * Clears the data and colors for all fill-in fields of the default frame.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>CLEAR</b>
    */
   public static void clear()
   {
      clear(false);
   }
      
   /**
    * Clears the data and colors for all fill-in fields of the default frame.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>CLEAR</b>
    *
    * @param   noPause
    *          <code>true</code> if NO-PAUSE option was used
    */
   public static void clear(boolean noPause)
   {
      LogicalTerminal lt = locate();

      // identify the active frame
      int frameId = lt.client.getActiveFrame();
      
      if (frameId == -1)
      {
         // nothing to do, "there is no spoon"
         return;
      }
      
      GenericFrame frame = lt.identifyFrame(frameId);
      
      if (frame == null)
      {
         throw new RuntimeException("Invalid frame ID " + frameId);
      }
      
      // let the frame maintain its own state before calling back to the
      // LT for service by the client export
      frame.clear(noPause);
   }
   
   /**
    * Cleanup the pending shared resources, as external program instantiation failed.
    */
   public static void cleanupPending()
   {
      LogicalTerminal lt = locate();
      synchronized (lt.registryLock)
      {
         // remove data regarding pending menus that were not pushed to client 
         for (Integer menuId : lt.pendingMenus)
         {
            lt.menuRegistry.remove(menuId);
            lt.widgetRegistry.remove(menuId);
         }
         // remove 2nd level menu descriptions
         for (MenuDescription desc : lt.level2descriptions)
         {
            lt.menuRegistry.remove(desc.getRootMenuID());
            lt.widgetRegistry.remove(desc.getRootMenuID());
         }
      }
      
      lt.level2definitions.clear();
      lt.level2descriptions.clear();
      lt.pendingMenus.clear();
   }
   
   /**
    * Clears all occurrences for the down frame and resets the current display
    * position to the top of the frame.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>CLEAR ALL </b>
    * 
    * @param   frame
    *          instance of <code>GenericFrame</code> this request came for
    * @param   noPause
    *          <code>true</code> if NO-PAUSE option was used
    */
   public static void clearAll(GenericFrame frame, boolean noPause)
   {
      LogicalTerminal lt = locate();
      
      ScreenBuffer[] sb = lt.client.clearAll(frame.getFrameId(), noPause);
      
      if (sb[0].getBaseId() != -1)
      {
         handleScreenBuffers(sb);
         if (frame == lt.frameRef)
         {
            String sv = sb[0].getFrameValue();
            lt.frameValue = (sv == null ? "" : sv);
         }
      }
      
      processDownConfigs(sb[0].getAndClearDownConfigs());
   }
      
   /**
    * Clears all occurrences for the active down frame and resets the current
    * display position to the top of the frame.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>CLEAR ALL </b>
    */
   public static void clearAll()
   {
      clearAll(false);
   }
      
   /**
    * Clears all occurrences for the active down frame and resets the current
    * display position to the top of the frame.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>CLEAR ALL</b>
    *
    * @param   noPause
    *          <code>true</code> if NO-PAUSE option was used
    */
   public static void clearAll(boolean noPause)
   {
      LogicalTerminal lt = locate();

      // identify the active frame
      int frameId = lt.client.getActiveFrame();
      
      if (frameId == -1)
      {
         // nothing to do, "there is no spoon"
         return;
      }
      
      GenericFrame frame = lt.identifyFrame(frameId);
      
      if (frame == null)
      {
         throw new RuntimeException("Invalid frame ID " + frameId);
      }
      
      // let the frame maintain its own state before calling back to the
      // LT for service by the client export
      frame.clearAll(noPause);
   }
   
   /**
    * Delete the line of text in which the targeted editor's cursor currently resides.  Leave
    * the cursor at the leftmost position of the same line number (which will be the line
    * previously below this one if the cursor is not already on the last line) or at the
    * leftmost position of the last line (if there are no lines below this one).
    *
    * @param    id
    *           Widget id of the target editor.
    *
    * @return   <code>true</code> if the line was deleted, 
    *           null if no widget with given id is found
    */
   public static Boolean deleteEditorLine(int id)
   {
      return locate().client.deleteEditorLine(id);
   }
   
   /**
    * Copies the currently selected text from an Editor, Fill-In or ComboBox to Clibpoard.
    *
    * @param   id
    *          Widget id of the target editor.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public static boolean editCopySelection(int id)
   {
      return locate().client.editCopySelection(id, false);
   }
   
   /**
    * Cuts the currently selected text from an Editor, Fill-In or ComboBox to Clibpoard.
    *
    * @param   id
    *          Widget id of the target editor.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public static boolean editCutSelection(int id)
   {
      return locate().client.editCopySelection(id, true);
   }
   
   /**
    * Paste the Clipboard content into an Editor, Fill-In or ComboBox starting with current
    * cursor position.
    *
    * @param   id
    *          Widget id of the target editor.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public static boolean editPaste(int id)
   {
      return locate().client.editPaste(id);
   }
   
   /**
    * Checks if the Clipboard has data to paste from.
    *
    * @param   id
    *          Widget id of the target editor.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public static boolean isEditCanPaste(int id)
   {
      return locate().client.isEditCanPaste(id);
   }
   
   /**
    * Checks if the widget can handle text undo request.
    *
    * @param   id
    *          Widget id of the target editor.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public static boolean isEditCanUndo(int id)
   {
      return locate().client.isEditCanUndo(id);
   }
   
   /**
    * Undo the last changes in text editable widget if possible.
    *
    * @param   id
    *          Widget id of the target editor.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public static boolean editUndo(int id)
   {
      return locate().client.editUndo(id);
   }
   
   /**
    * Removes the currently selected text from an Editor, Fill-In or ComboBox.
    *
    * @param   id
    *          Widget id of the target editor.
    */
   public static void removeEditSelection(int id)
   {
      locate().client.removeEditSelection(id);
   }
   
   /**
    * Selects text in the targeted editor between start and end points.
    * 
    * @param    id
    *           Widget id of the target editor.
    * @param    start
    *           Starting offset.
    * @param    end
    *           Ending offset.
    *
    * @return   <code>true</code> if the operation succeeded.
    */
   public static Boolean setEditorSelection(int id, int start, int end)
   {
      return locate().client.setEditorSelection(id, start, end);
   }
   
   /**
    * Sets the cursor offset of the targeted editor.
    *
    * @param   id
    *          Widget id of the target editor.
    *
    * @param   offset
    *          the new offset value. Use -1 to move the caret at the end (MOVE-TO-EOF).
    *
    * @return  {@code true} if the cursor offset is valid and set, {@code null} if no widget 
    *          with given {@code id} is found.
    */
   public static Boolean setEditorCursorOffset(int id, int offset)
   {
      return locate().client.setEditorCursorOffset(id, offset);
   }
      
   /**
    * Sets the cursor position in the current line of the targeted editor
    *
    * @param    id
    *           Widget id of the target editor.
    * @param    pos
    *           the new cursor position value
    *           
    * @return   <code>true</code> if the cursor position is valid and set, 
    *           null if no widget with given id is found
    */
   public static Boolean setEditorCursorChar(int id, int pos)
   {
      return locate().client.setEditorCursorChar(id, pos);
   }
      
   /**
    * Sets the cursor line of the targeted editor
    *
    * @param    id
    *           Widget id of the target editor.
    * @param    line
    *           the new cursor line value
    *           
    * @return   <code>true</code> if the cursor line is valid and set, 
    *           null if no widget with given id is found
    */
   public static Boolean setEditorCursorLine(int id, int line)
   {
      return locate().client.setEditorCursorLine(id, line);
   }
   
   /**
    * Inserts string at the current cursor position of the targeted editor.
    *
    * @param    id
    *           Widget id of the target editor.
    * @param    txt
    *           The string to be inserted.
    *
    * @return  {@code true} if the string is inserted, {@code null} if no widget with given
    *          {@code id} is found.
    */
   public static Boolean insertEditorString(int id, String txt)
   {
      return locate().client.insertEditorString(id, txt);
   }
   
   /**
    * Inserts the content of a file at the current cursor position of the targeted editor.
    *
    * @param   id
    *          Widget id of the target editor.
    * @param   filePath
    *          The file to be inserted. Full or relative path may be used. In
    *          latter case, the {@code PROPATH} is used to find the file.
    *
    * @return  {@code true} if the file content is inserted, {@code null} if no widget with given
    *          {@code id} is found.
    */
   public static Boolean insertEditorFile(int id, String filePath)
   {
      return locate().client.insertEditorFile(id, filePath);
   }
   
   /**
    * Replace selection of the targeted editor with a given string.
    *
    * @param    id
    *           Widget id of the target editor.
    * @param    txt
    *           the replacement string
    *
    * @return   <code>true</code> if the selection is replaced, 
    *           null if no widget with given id is found
    */
   public static Boolean replaceEditorSelectionText(int id, String txt)
   {
      return locate().client.replaceEditorSelectionText(id, txt);
   }

   /**
    * Replace given substring of the targeted editor with a given string
    *
    * @param    id
    *           Widget id of the target editor.
    * @param    from
    *           the string to be replaced
    * @param    to
    *           the replacement string
    * @param    flag
    *           the replacement mode flag as described in 4GL EDITOR:REPLACE documentation 
    * 
    * @return   <code>true</code> if the replacement was performed, 
    *           null if no widget with given id is found
    */
   public static Boolean replaceEditorString(int id, String from, String to, int flag)
   {
      return locate().client.replaceEditorString(id, from, to, flag);
   }
   
   /**
    * Searches for a specified string starting from the current text caret position.
    *
    * @param   id
    *          Widget id of the target editor.
    * @param   pattern
    *          The string to be search. No wildcard allowed.
    * @param   mode
    *          a combinations of flags that specifies the type of search to be performed. The
    *          following values are accepted:
    *          <table summary="">
    *              <tr><th>Type of search</th><th>Flag value</th>
    *              <tr><td>EditorWidget.FIND_NEXT_OCCURRENCE</td><td>1</td></tr>
    *              <tr><td>EditorWidget.FIND_PREV_OCCURRENCE</td><td>2</td></tr>
    *              <tr><td>EditorWidget.FIND_CASE_SENSITIVE</td><td>4</td></tr>
    *              <tr><td>EditorWidget.FIND_WRAP_AROUND</td><td>16</td></tr>
    *              <tr><td>EditorWidget.FIND_SELECT</td><td>32</td></tr>
    *          </table>
    *
    * @return  if the text is found and caret moved, the method returns {@code yes}.
    */
   public static Boolean searchInEditor(int id, String pattern, int mode)
   {
      return locate().client.searchInEditor(id, pattern, mode);
   }

   /**
    * Disables input for all specified field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DISABLE ALL </b> <br>
    * <b>DISABLE ALL EXCEPT </b>
    *
    * @param   frame
    *          instance of <code>GenericFrame</code> this request came for
    * @param   widgetId
    *          array of field level widget IDs to be disabled or
    *          <code>null</code>, which means all of them.
    */
   public static void disable(GenericFrame frame, int[] widgetId)
   {
      disable(frame, widgetId, true);
   }
      
   /**
    * Disables input for all specified field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DISABLE ALL </b> <br>
    * <b>DISABLE ALL EXCEPT </b>
    * 
    * @param   frame
    *          instance of <code>GenericFrame</code> this request came for
    * @param   widgetId
    *          array of field level widget IDs to be disabled or
    *          <code>null</code>, which means all of them.
    * @param   resetEntered
    *          If <code>true</code> then <b>ENTERED</b> flag is reset for the
    *          widgets which are being disabled.
    */
   public static void disable(GenericFrame frame, int[] widgetId, boolean resetEntered)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      if (resetEntered)
      {
         // reset ENTERED flag for widgets which are explicitly disabled; no need to dynamically
         // update headers and other content since we are calling getFrameBuffer() below
         frame.getFrameBufferRaw().resetEntered(widgetId);
      }
      
      // refresh editable frames list
      lt.removeEditableFrameForWidgets(frame, widgetId);
      
      int[] nextTabItemList = frame.getTabItemsList();
      
      lt.client.enable(frame.getFrameId(),
                       widgetId,
                       lt.getEditableServerScreenBuffers(frame),
                       false,
                       // TODO check frame window usage
                       targetWindowWorker(frame, null, false),
                       nextTabItemList); 
   }

   /**
    * Disables input for all specified field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DISABLE ALL </b> <br>
    * <b>DISABLE ALL EXCEPT </b>
    * 
    * @param   frame
    *          instance of <code>GenericFrame</code> this request came for
    * @param   widgetId
    *          array of field level widget IDs to be disabled or
    *          <code>null</code>, which means all of them.
    */
   public static void disable(GenericFrame frame, int[] widgetId, handle hWin)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      // reset ENTERED flag for widgets which are explicitly disabled; no need to dynamically
      // update headers and other content since we are calling getFrameBuffer() below
      frame.getFrameBufferRaw().resetEntered(widgetId);
      
      // refresh editable frames list
      lt.removeEditableFrameForWidgets(frame, widgetId);
      
      int[] nextTabItemList = frame.getTabItemsList();
      
      ScreenBuffer[] sb = lt.client.enable(frame.getFrameId(),
                                           widgetId,
                                           lt.getEditableServerScreenBuffers(frame),
                                           false,
                                           targetWindowWorker(frame, hWin, false),
                                           nextTabItemList);

      handleScreenBuffers(sb);
   }

   /**
    * Enables input for all specified field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>ENABLE ALL </b> <br>
    * <b>ENABLE ALL EXCEPT </b>
    * 
    * @param   frame
    *          instance of <code>GenericFrame</code> this request came for
    * @param   widgetId
    *          array of field level widget IDs to be enabled or
    *          <code>null</code>, which means all of them.
    */
   public static void enable(GenericFrame frame, int[] widgetId)
   {
      // reset ENTERED flag for widgets which are explicitly disabled; no need to dynamically
      // update headers and other content since we are calling getFrameBuffer() below
      frame.getFrameBufferRaw().resetEntered(widgetId);
      
      LogicalTerminal lt = locate();
      
      lt.frameViewed(frame);
      frame.setHadView();

      // refresh editable frames list
      lt.addEditableFrameForWidgets(frame, widgetId);
      
      int[] nextTabItemList = frame.updateTabItemsList(widgetId, UpdatesOp.ENABLE);
      
      ScreenBuffer[] sb = lt.client.enable(frame.getFrameId(),
                                           widgetId,
                                           lt.getEditableServerScreenBuffers(frame),
                                           true,
                                           targetWindowWorker(frame, null, false),
                                           nextTabItemList);
      
      handleScreenBuffers(sb);
   }

   /**
    * Enables input for all specified field-level widgets of the frame.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>ENABLE ALL IN WINDOW </b> <br>
    * <b>ENABLE ALL EXCEPT IN WINDOW</b>
    * 
    * @param   frame
    *          instance of <code>GenericFrame</code> this request came for
    * @param   widgetId
    *          array of field level widget IDs to be enabled or
    *          <code>null</code>, which means all of them.
    * @param   hWin
    *          Handle of target window.
    */
   public static void enable(GenericFrame frame, int[] widgetId, handle hWin)
   {
      // reset ENTERED flag for widgets which are explicitly disabled; no need to dynamically
      // update headers and other content since we are calling getFrameBuffer() below
      frame.getFrameBufferRaw().resetEntered(widgetId);
      
      LogicalTerminal lt = locate();
      
      lt.frameViewed(frame);
      frame.setHadView();

      // refresh editable frames list
      lt.addEditableFrameForWidgets(frame, widgetId);
      
      int[] nextTabItemList = frame.updateTabItemsList(widgetId, UpdatesOp.ENABLE);
      
      lt.client.enable(frame.getFrameId(),
                       widgetId,
                       lt.getEditableServerScreenBuffers(frame),
                       true, 
                       targetWindowWorker(hWin, false),
                       nextTabItemList);
   }
   
   /**
    * Enables or disables the specified widget without changing the tab order,
    * which is how widget:SENSITIVE=YES/NO works.
    * 
    * @param    widget
    *           widget to enable or disable
    * @param    sensitive
    *           <code>TRUE</code> enables the widget, <code>FALSE</code>
    *           disables it.
    */
   public static void setSensitive(GenericWidget<?> widget, boolean sensitive)
   {
      LogicalTerminal lt = locate();
      
      int            id       = widget.getId();
      GenericFrame   frame    = widget.getFrame();
      ScreenBuffer[] frameBuf = (frame == null) ? null : lt.getEditableServerScreenBuffers(frame);
      int            frameId  = (frame == null) ? 0 : frame.getFrameId(); 
      int          windowId = (frame == null) ? WindowConfig.RESOLVE_WINDOW : frame.getWindowId();
      
      // refresh editable frames cache if required, but when sensitive flag is changing 
      if (widget.isSensitive().booleanValue() != sensitive)
      {
         if (sensitive)
         {
            lt.addEditableFrameForWidgets(frame, new int[] {id});
         }
         else
         {
            lt.removeEditableFrameForWidgets(frame, new int[] {id});
         }
      }
      
      int[] nextTabItemList = frame.getTabItemsList();
      
      lt.client.enable(frameId, id, frameBuf, sensitive, windowId, nextTabItemList); 
   }
   
   /**
    * Check driver type: CHUI or not
    * 
    * @return   <code>TRUE</code> if the CHUI, <code>FALSE</code> otherwise 
    */
   public static boolean isChui() 
   {
      ClientParameters cp = StandardServer.getClientParameters();
      
      if (cp != null)
      {
         return cp.chui;
      }
      
      if (LOG.isLoggable(Level.WARNING))
      {
         LOG.logp(Level.WARNING, "LogicalTerminal", "isChui", 
           "No client parameters are set: calling the client-side to determine the driver type " +
           "is expensive!");
      }
      
      return locate().client.isChui();
   }
   
   /**
    * Check the client type: embedded or not.
    * 
    * @return   <code>TRUE</code> if the client is running in embedded mode, 
    *           <code>FALSE</code> otherwise 
    */
   public static boolean isEmbedded() 
   {
      LogicalTerminal lt = locate();
      
      if (lt.embedded == null)
      {
         lt.embedded = lt.client.isEmbedded();
      }
      
      return lt.embedded;
   }
   
   /**
    * Shows the specified frame or widget on screen.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW widget
    * 
    * @param    widgetId
    *           The frame ID or widget ID to show.
    */
   public static void view(int widgetId)
   {
      LogicalTerminal lt = locate();
      
      int          id       = lt.lookupFrameIdFromWidgetId(widgetId);
      GenericFrame frame    = lt.identifyFrame(id);
      ScreenBuffer[] frameBuf = (frame == null) ? null : lt.getEditableServerScreenBuffers(frame); 
      int          windowId = (frame == null) ? WindowConfig.RESOLVE_WINDOW : frame.getWindowId();
      
      lt.frameViewed(frame);
      FieldGroup fg;
      if (frame != null)
      {
         fg = frame.getFieldGroup();
      }
      else
      {
         fg = null;
      }
      view(widgetId, frameBuf, null, true, windowId,
               fg != null ? fg.getTabItemList() : null);
   }

   /**
    * Shows all fields of the specified frame on screen.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW widgets
    * 
    * @param    frame
    *           The frame to make visible.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    */
   public static void view(GenericFrame frame, int[] widgetIds)
   {
      LogicalTerminal lt = locate();
      
      lt.frameViewed(frame);
      
      view(frame.getFrameId(),
           lt.getEditableServerScreenBuffers(frame),
           widgetIds, 
           true,
           frame.getWindowId(),
           frame.getTabItemsList());
   }

   /**
    * Shows all fields of the specified frame on screen.
    * <p>
    * This method corresponds to these Progress statements:
    * VIEW widgets
    * 
    * @param    frame
    *           The frame to make visible.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    hWin
    *           Handle of target window.
    */
   public static void view(GenericFrame frame, int[] widgetIds, handle hWin)
   {
      LogicalTerminal lt = locate();
      
      lt.frameViewed(frame);
      
      view(frame.getFrameId(),
           lt.getEditableServerScreenBuffers(frame),
           widgetIds, 
           true,
           // TODO check frame window usage
           targetWindowWorker(frame, hWin, false),
           frame.getTabItemsList());
   }
   
   /**
    * Shows all fields of the specified frame on screen.
    * <p>
    * The terminal will be temporarily redirected to the given stream for the
    * duration of this method call.  This allows the caller to avoid 
    * bracketing this call with {@link #switchOutput} invocations, as the
    * same result is achieved internal to this method's operation.
    * 
    * @param    frame
    *           The frame to make visible.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    streamId
    *           The stream to be used as the temporary terminal redirection
    *           (only for the duration of this call).
    * @param    register
    *           <code>true</code> if the frame associated with this request
    *           must be registered as a page element with the given stream. 
    * @param    pageTop
    *           <code>true</code> if the page element registration is for
    *           a page top element, <code>false</code> for a page bottom
    *           element.  This parameter is ignored if <code>register</code>
    *           is <code>false</code>.
    * @param    viewDisp
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    */
   public static void view(GenericFrame frame,
                           int[]        widgetIds,
                           int          streamId,
                           boolean      register,
                           boolean      pageTop,
                           boolean      viewDisp)
   {
      LogicalTerminal lt = locate();
      
      lt.frameViewed(frame);
      
      WidgetConfig[] wcfgs = lt.client.view(frame.getFrameId(),
                                            lt.getEditableServerScreenBuffers(frame),
                                            widgetIds,
                                            streamId,
                                            register,
                                            pageTop,
                                            viewDisp,
                                            // TODO check frame window usage
                                            targetWindowWorker(frame, null, false),
                                            frame.getTabItemsList());

      processDownConfigs(wcfgs);
   }
   
   /**
    * Support for the DISPLAY followed by DOWN combination.
    * 
    * @param    frame
    *           The frame this request came for.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    streamId
    *           The stream to be used as the temporary terminal redirection
    *           (only for the duration of this call).
    * @param    register
    *           <code>true</code> if the frame associated with this request
    *           must be registered as a page element with the given stream. 
    * @param    pageTop
    *           <code>true</code> if the page element registration is for
    *           a page top element, <code>false</code> for a page bottom
    *           element.  This parameter is ignored if <code>register</code>
    *           is <code>false</code>.
    */
   public static void displayAndDown(GenericFrame frame,
                                     int[]         widgetIds,
                                     int           streamId,
                                     boolean       register,
                                     boolean       pageTop)
   {
      LogicalTerminal lt = locate();
      
      lt.frameViewed(frame);
      
      FieldGroup fg = frame.getFieldGroup();
      
      lt.client.displayAndDown(frame.getFrameId(),
                               lt.getEditableServerScreenBuffers(frame),
                               widgetIds,
                               streamId,
                               register,
                               pageTop,
                               // TODO check frame window usage
                               targetWindowWorker(frame, null, false),
                               fg != null ? fg.getTabItemList() : null);
   }
   
   /**
    * Shows all fields of the specified frame on screen.
    * <p>
    * The terminal will be temporarily redirected to the given stream for the
    * duration of this method call.  This allows the caller to avoid 
    * bracketing this call with {@link #switchOutput} invocations, as the
    * same result is achieved internal to this method's operation.
    * 
    * @param    frame
    *           The frame to make visible.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    streamId
    *           The stream to be used as the temporary terminal redirection
    *           (only for the duration of this call).
    * @param    register
    *           <code>true</code> if the frame associated with this request
    *           must be registered as a page element with the given stream. 
    * @param    pageTop
    *           <code>true</code> if the page element registration is for
    *           a page top element, <code>false</code> for a page bottom
    *           element.  This parameter is ignored if <code>register</code>
    *           is <code>false</code>.
    * @param    viewDisp
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    * @param    hWin
    *           Handle of target window.
    */
   public static void view(GenericFrame frame,
                           int[]        widgetIds,
                           int          streamId,
                           boolean      register,
                           boolean      pageTop,
                           boolean      viewDisp,
                           handle       hWin)
   {
      LogicalTerminal lt = locate();
      
      lt.frameViewed(frame);
      
      WidgetConfig[] wcfgs = lt.client.view(frame.getFrameId(),
                                            lt.getEditableServerScreenBuffers(frame),
                                            widgetIds,
                                            streamId,
                                            register,
                                            pageTop,
                                            viewDisp,
                                            // TODO check frame window usage
                                            targetWindowWorker(frame, hWin, false),
                                            frame.getTabItemsList());
      processDownConfigs(wcfgs);
   }
   
   /**
    * Shows all fields of the specified frame on screen.
    * <p>
    * This method corresponds to these Progress statements:
    * DISPLAY widgets
    * 
    * @param    frame
    *           The frame for which this request came.
    * @param    frameBuf
    *           The screen buffer array to send to the client, does not need to be
    *           the current screen buffer of the frame.
    * @param    widgetId
    *           An array of field level widget IDs to be shown or
    *           <code>null</code>, which means all widgets.
    */
   public static void view(GenericFrame   frame,
                           ScreenBuffer[] frameBuf,
                           int[]          widgetId)
   {
      LogicalTerminal lt = locate();
      
      lt.frameViewed(frame);
      
      view(frame.getFrameId(),
           frameBuf,
           widgetId,
           false,
           frame.getWindowId(),
           frame.getTabItemsList());
   }

   /**
    *
    *
    * @param    frame
    *           The frame to be edited.
    * @param    widgetIds
    *           array of field level widget IDs to be enabled or
    *           <code>null</code>, which means all of them.
    * @param    wait
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    * @param    doIO
    *           tells to perform the requested operation if <code>true</code>.
    *           otherwise prepares the frame for editing session only.
    * @param    hWin
    *           Handle of target window.
    */
   public static void promptFor(GenericFrame frame,
                                int[]        widgetIds,
                                EventList    wait,
                                boolean      doIO,
                                handle       hWin)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      // disable loop protection and increment interactions counter
      lt.markInteractive(true);
      
      boolean push = wait != null;
      if (push)
      {
         lt.waitStack.push(wait);
      }
      EventList wl = lt.prepareEventList(frame, wait);
      
      // reset ENTERED flag for widgets which are explicitly disabled; no need to dynamically
      // update headers and other content since we are calling getFrameBuffer() below
      frame.getFrameBufferRaw().resetEntered(widgetIds);
     
      ScreenBuffer[] sb = lt.getEditableServerScreenBuffers(frame);
      
      // call the client
      try
      {
         lt.frameViewed(frame);
         
         int[] tabItemList = frame.updateTabItemsList(widgetIds, UpdatesOp.PROMPT_FOR);
         
         sb = lt.client.promptFor(frame.getFrameId(),
                                  widgetIds,
                                  sb,
                                  toMilliseconds(lt.getDefaultTimeout()),
                                  wl,
                                  doIO,
                                  targetWindowWorker(frame, hWin, false),
                                  tabItemList);
         handleScreenBuffers(frame, sb, false);
      }
      catch (ConditionException ce)
      {
         processCondition(ce);

         // continue the unwinding of the stack
         throw ce;
      }
      finally
      {
         if (push)
         {
            lt.waitStack.pop();
         }
      }
   }
   
   /**
    * Positions the cursor on a new line in a down frame.
    * In redirected mode, flushes the buffered output and, optionally,
    * inserts newlines into the stream. Corresponds to both conditional
    * and unconditional DOWN statements
    * 
    * @param    frame
    *           The frame for which this request came.
    * @param    lines
    *           Number and direction of iterations of the data to move. If
    *           positive, the move is <code>down</code>, if negative the
    *           move is <code>up</code>.  0 is ignored.
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    */
   public static void down(GenericFrame frame, int lines, boolean cond)
   {
      down(frame, lines, cond, -1);
   }

   /**
    * Positions the cursor on a new line in a down frame.
    * In redirected mode, flushes the buffered output and, optionally,
    * inserts newlines into the stream. Corresponds to both conditional
    * and unconditional DOWN statements
    * 
    * @param    frame
    *           The frame for which this request came.
    * @param    lines
    *           Number and direction of iterations of the data to move. If
    *           positive, the move is <code>down</code>, if negative the
    *           move is <code>up</code>.  0 is ignored.
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    * @param    sid
    *           The stream for which pending PUT or EXPORT output must be
    *           ended and written to the stream. -1 means that this is not
    *           necessary.
    */
   public static void down(GenericFrame frame,
                           int          lines,
                           boolean      cond,
                           int          sid)
   {
      if (lines == 0 && sid == -1)
         return;
      
      WidgetConfig[] wcfgs = locate().client.down(frame.getFrameId(), lines, cond, sid);

      processDownConfigs(wcfgs);
   }

   /**
    * Positions the cursor on a new line in a down frame.  When 
    * <code>lines</code> is negative, this operates as a move of the cursor
    * <code>up</code>.
    * <p>
    * The terminal will be temporarily redirected to the given stream for the
    * duration of this method call.  This allows the caller to avoid 
    * bracketing this call with {@link #switchOutput} invocations, as the
    * same result is achieved internal to this method's operation.
    * <p>
    * This method corresponds to these Progress statements: <br>
    * <b>DOWN n </b>
    * <b>UP n </b>
    * 
    * @param    frame
    *           The frame for which this request came.
    * @param    lines
    *           Number and direction of iterations of the data to move. If
    *           positive, the move is <code>down</code>, if negative the
    *           move is <code>up</code>.  0 is ignored.
    * @param    streamId
    *           The stream to be used as the temporary terminal redirection
    *           (only for the duration of this call).
    * @param    cond
    *           <code>true</code> means conditional down; lines is assumed
    *           to be 1
    * @param    endOut                
    *           <code>true</code> to force any pending PUT or EXPORT output
    *           to be ended for the given stream (so long as that stream is
    *           not the terminal).
    */
   public static void down(GenericFrame frame,
                           int          lines,
                           int          streamId, 
                           boolean      cond,
                           boolean      endOut)
   {
      if (lines == 0 && !endOut)
         return;
      
      WidgetConfig[] wcfgs = locate().client.down(frame.getFrameId(), lines, streamId, cond, endOut);
      
      processDownConfigs(wcfgs);
   }
   
   /**
    * Makes the frame or widget invisible.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>HIDE widget </b>
    * 
    * @param   widgetId
    *          instance of <code>GenericFrame</code> this request came for
    * @param   noPause
    *          <code>true</code> if NO-PAUSE option was used
    * @param   hidden
    *          If <code>true</code>, set the HIDDEN attribute.
    */
   public static boolean hide(int widgetId, boolean noPause, boolean hidden)
   {
      return locate().client.hide(widgetId, noPause, hidden);
   }

   /**
    * Makes the frame or widget invisible.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>HIDE widget </b>
    * 
    * @param   widgetId
    *          instance of <code>GenericFrame</code> this request came for
    * @param   noPause
    *          <code>true</code> if NO-PAUSE option was used
    * @param   hWin
    *          Handle of target window.
    */
   public static void hide(int widgetId, boolean noPause, handle hWin)
   {
      // hWin is ignored. widget window is used.
      locate().client.hide(widgetId, noPause, false);
   }

   /**
    * Specifies the field to be put in focus during the next input operation.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>NEXT-PROMPT </b>
    * 
    * @param   frame
    *          instance of <code>GenericFrame</code> this request came for
    * @param   widgetId
    *          widget to be put in focus
    */
   public static void nextPrompt(GenericFrame frame, int widgetId) 
   {
      locate().client.nextPrompt(frame.getFrameId(), widgetId);
   }

   /**
    * Scrolls the lines of the down frame up or down.
    * <p>
    * This method corresponds to this Progress statement: <br>
    * <b>SCROLL </b>
    * 
    * @param   frame
    *          instance of <code>GenericFrame</code> this request came for
    * @param   fromCurrent
    *          <code>true</code> if FROM-CURRENT option was used
    * @param   upDown
    *          <code>true</code> if scrolling up
    */
   public static void scroll(GenericFrame frame, boolean fromCurrent, 
                             boolean upDown)
   {
      locate().client.scroll(frame.getFrameId(), fromCurrent, upDown);
   }

   /**
    * Push the specified widget attribute change to the client-side.
    * 
    * @param    frameVisible
    *           The current (after config change) visible frame state.
    * @param    wid
    *           The widget ID.
    * @param    fname
    *           The config field name.
    * @param    value
    *           The new config field value.
    */
   public static void pushWidgetAttr(boolean frameVisible, int wid, String fname, Object value)
   {
      LogicalTerminal lt = locate();
      
      lt.client.pushWidgetAttr(frameVisible, wid, fname, value);
   }

   /**
    * Push the specified widget attribute changes to the client-side.
    * 
    * @param    frameVisible
    *           The current (after config change) visible frame state.
    * @param    wid
    *           The widget ID.
    * @param    fnames
    *           The list of config field names.
    * @param    values
    *           The list of new config field values.
    */
   public static void pushWidgetAttr(boolean  frameVisible, 
                                     int      wid, 
                                     String[] fnames, 
                                     Object[] values)
   {
      LogicalTerminal lt = locate();
      
      lt.client.pushWidgetAttr(frameVisible, wid, fnames, values);
   }
   
   /**
    * Place the frame.
    */
   public static void placeFrame(int frameId)
   {
      LogicalTerminal lt = locate();
      
      lt.client.placeFrame(frameId);
   }
   
   /**
    * Delete and detach a widget from a frame, at runtime.  The widget can be a dynamic widget or a full
    * static frame.
    * 
    * @param    frame
    *           The parent frame.
    * @param    widget
    *           The widget.
    */
   public static void deleteDynamicWidget(GenericFrame frame, GenericWidget widget)
   {
      if (widget instanceof BrowseWidget || 
          widget instanceof BrowseColumnWidget || 
          widget instanceof TreeWidgetBase)
      {
         // this has some special behavior, as the browse is made from more than one widget...
         frame.pushScreenDefinition(true);
      }
      
      LogicalTerminal lt = locate();
      
      lt.client.deleteDynamicWidget(frame.asWidget().getId(), widget.getId());
   }
   
   /**
    * Attach a widget to a frame, at runtime.  The widget can be a dynamic widget or a full static frame.
    * 
    * @param    frame
    *           The parent frame.
    * @param    widget
    *           The widget to attach.
    */
   public static void attachRuntimeWidget(GenericFrame frame, GenericWidget widget)
   {
      if (widget instanceof BrowseWidget || 
          widget instanceof BrowseColumnWidget || 
          widget instanceof TreeWidgetBase)
      {
         // this has some special behavior, as the browse is made from more than one widget...
         frame.pushScreenDefinition(true);
         return;
      }
      
      if (widget._dynamic())
      {
         WidgetConfig cfg = widget.config();
         ConfigManager mgr = ConfigManager.getInstance();
         WidgetConfig active = mgr.getActiveConfig(cfg.id);
         if (active == null)
         {
            mgr.addWidgetConfig(cfg);
         }
         else
         {
            assert(cfg.getClass().equals(active.getClass()));
         }
         
         widget.finishConfigProcessing();
      }
      
      LogicalTerminal lt = locate();
      
      lt.client.attachRuntimeWidget(frame.asWidget().getId(), widget.getId(), widget.config());
   }
   
   /**
    * Pushes the screen definition instance down to the client.
    * 
    * @param   frameDef
    *          an array of <code>ScreenDefinition</code> objects
    */
   public static void pushScreenDefinition(ScreenDefinition[] frameDef)
   {
      pushScreenDefInt(frameDef);
   }

   /**
    * Pushes the menu description instance down to the client.
    * 
    * @param   menuDescr
    *          an array of <code>MenuDescription</code> objects
    */
   public static void pushMenuDescription(MenuDescription[] menuDescr)
   {
      pushMenuDescrInt(menuDescr);
   }

   /**
    * Queue a screen definition for a future push down to the client using
    * {@link #processDeferredPush}.
    * 
    * @param   frameDef
    *          an array of <code>ScreenDefinition</code> objects
    */
   public static void deferPush(ScreenDefinition frameDef)
   {
      // this is called during init in the business logic and that means that
      // we haven't had the chance to mark the next partition yet because
      // scopeStart() won't be called until the business logic initializer
      // is complete
      LogicalTerminal lt = locate();
      lt.level2definitions.add(frameDef);
      lt.tm.registerPendingScopeable(lt);
   }

   /**
    * Queue a menu description for a future push down to the client using
    * {@link #processDeferredPush}.
    * 
    * @param   menuDescr
    *          an array of <code>MenuDescription</code> objects
    */
   public static void deferPush(MenuDescription menuDescr)
   {
      // this is called during init in the business logic and that means that
      // we haven't had the chance to mark the next partition yet because
      // scopeStart() won't be called until the business logic initializer
      // is complete
      LogicalTerminal lt = locate();
      lt.level2descriptions.add(menuDescr);
      lt.tm.registerPendingScopeable(lt);
   }
   
   /**
    * Push the entire list of queued screen definitions down to the client.
    * The definitions must have been previously queued via {@link #deferPush}.
    */
   public static void processDeferredPush()
   {
      processDeferredPush(false);
   }
   
   /**
    * Push the entire list of queued screen definitions down to the client.
    * The definitions must have been previously queued via {@link #deferPush}.
    * 
    * @param    fromLevel2
    *           Flag indicating if we need to collect the level2 definitions.
    */
   public static void processDeferredPush(boolean fromLevel2)
   {
      LogicalTerminal lt = locate();
      lt.tm.registerBlockScopeable(lt);

      if (fromLevel2)
      {
         lt.definitions.markPartition();
         if (!lt.level2definitions.isEmpty())
         {
            lt.definitions.addAll(lt.level2definitions);
         }
         lt.level2definitions.clear();
      }
      
      if (lt.definitions.size() > 0)
      {
         pushScreenDefInt(lt.definitions.toArray(EXEMPLAR));
         lt.definitions.clear();
      }

      if (fromLevel2)
      {
         lt.descriptions.markPartition();
         if (!lt.level2descriptions.isEmpty())
         {
            lt.descriptions.addAll(lt.level2descriptions);
         }
         lt.level2descriptions.clear();
      }

      if (lt.descriptions.size() > 0)
      {
         pushMenuDescrInt(lt.descriptions.toArray(new MenuDescription[]{}));
         lt.descriptions.clear();
      }
   }

   /**
    * Register a trigger within the current scope.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    * @param    proc
    *           The internal or external procedure's name to be executed.
    */
   public static void registerTriggerPersistent(EventList events,
                                                String    proc)
   {
      registerTriggerPersistent(events, proc, null, (BaseDataType[]) null);
   }
   
   /**
    * Register a trigger within the current scope.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    * @param    proc
    *           The internal or external procedure's name to be executed.
    * @param    params
    *           The parameters to be passed to the procedure (INPUT only).
    *           Each parameter is evaluated at trigger definition; the 
    *           parameters do not get reevaluated on trigger invocation.
    */
   public static void registerTriggerPersistent(EventList       events,
                                                String          proc,
                                                BaseDataType... params)
   {
      registerTriggerPersistent(events, proc, null, params);
   }

   /**
    * Register a trigger within the current scope.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    * @param    proc
    *           The internal or external procedure's name to be executed.
    * @param    inst
    *           The instance where the procedure is searched.
    */
   public static void registerTriggerPersistentIn(EventList events,
                                                  String    proc,
                                                  handle    inst)
   {
      registerTriggerPersistent(events, proc, inst, (BaseDataType[]) null);
   }

   /**
    * Register a trigger within the current scope.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    * @param    proc
    *           The internal or external procedure's name to be executed.
    * @param    inst
    *           The instance where the procedure is searched.
    * @param    params
    *           The parameters to be passed to the procedure (INPUT only).
    *           Each parameter is evaluated at trigger definition; the 
    *           parameters do not get reevaluated on trigger invocation.
    */
   public static void registerTriggerPersistentIn(EventList       events,
                                                  String          proc,
                                                  handle          inst,
                                                  BaseDataType... params)
   {
      registerTriggerPersistent(events, proc, inst, params);
   }
   
   /**
    * Register a trigger within the current scope.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    * @param    proc
    *           The internal or external procedure's name to be executed.
    * @param    inst
    *           The instance where the procedure is searched.
    * @param    params
    *           The parameters to be passed to the procedure (INPUT only).
    *           Each parameter is evaluated at trigger definition; the 
    *           parameters do not get reevaluated on trigger invocation.
    */
   private static void registerTriggerPersistent(EventList       events,
                                                 final String    proc,
                                                 final handle    inst,
                                                 BaseDataType... params)
   {
      if (!validEventList(events))
      {
         return;
      }
      
      // only INPUT parameters are allowed, so no need to emit the parameter
      // modes
      
      if (params == null)
         params = new BaseDataType[0];
      
      // each parameter is evaluated at trigger definition; the parameters do
      // not get reevaluated on trigger invocation.
      final Object[] args = new Object[params.length];
      for (int i = 0; i < params.length; i++)
      {
         args[i] = params[i].duplicate();
      }
      
      // create a custom trigger definition for this case
      Object outer = locate().pm._thisProcedure();
      TriggerDefinition td = new TriggerDefinition(() -> new Trigger()
      {
         @Override
         public void body()
         {
            // the search is always done under the scope of the procedure
            // where the event had happened.
            try
            {
               if (inst == null)
               {
                  ControlFlowOps.invoke(proc, args);
               }
               else
               {
                  ControlFlowOps.invokeIn(proc, inst, args);
               }
            }
            catch (ConditionException exc)
            {
               // don't eat fatal exceptions
               if (exc instanceof QuitConditionException ||
                   exc instanceof StopConditionException)
               {
                  throw exc;
               }
               
               // other conditions raised during trigger invocation are
               // silently ignored, and the event is consumed
               BlockManager.returnConsume();
            }
         }
      }, outer, false);
      
      // register the trigger 
      registerTrigger(events, td, true);
   }

   /**
    * Cleanup all triggers created in the specified external program.
    * 
    * @param    referent
    *           The external program which is being deleted.
    */
   @SuppressWarnings("unchecked")
   public static void cleanupTriggers(Object referent)
   {
      if (TransactionManager.isHeadless())
      {
         return;
      }

      LogicalTerminal lt = locate();
      
      // invalidate the cache
      lt.cachedEventList = null;

      // go through each scope and remove the triggers
      for (int i = 0; i < lt.scope.size(); i++)
      {
         LinkedList<EventList> events = (LinkedList<EventList>) lt.scope.getScopeAt(i);

         Iterator<EventList> itr = events.iterator();
         while (itr.hasNext())
         {
            EventList el = itr.next();
            int triggerId = el.getTriggerId();
            
            if (lt.isPersistentTrigger(triggerId))
            {
               continue;
            }
            
            TriggerDefinition td = lt.triggerRegistry.get(triggerId);
            if (td == null || td.getExternalProgram() != referent)
            {
               continue;
            }
            
            // the trigger's context is no longer alive, remove it
            itr.remove();

            lt.deregisterTrigger(triggerId, true);
         }
      }
   }
   
   /**
    * Cleanup all triggers registered for the specified resource or widget ID, assuming these are
    * being deleted.
    * 
    * @param    resourceId
    *           The resource ID.
    * @param    widgetId
    *           The widget ID
    */
   @SuppressWarnings("unchecked")
   public static void cleanupTriggers(Long resourceId, int widgetId)
   {
      LogicalTerminal lt = locate();

      if (lt.tm.isHeadless())
      {
         return;
      }
      
      resourceId = (resourceId == null ? ResourceIdHelper.INVALID_RESOURCE : resourceId);
      
      // invalidate the cache
      lt.cachedEventList = null;
      
      // go through each scope and remove the triggers
      for (int i = 0; i < lt.scope.size(); i++)
      {
         LinkedList<EventList> events = (LinkedList<EventList>) lt.scope.getScopeAt(i);
         
         Iterator<EventList> itr = events.iterator();
         while (itr.hasNext())
         {
            EventList el = itr.next();
            int triggerId = el.getTriggerId();
            if (triggerId != -1 && el.cleanup(lt, resourceId, widgetId))
            {
               itr.remove();
               
               lt.deregisterTrigger(triggerId, true);
            }
         }
      }
   }
   
   /**
    * Register a specific trigger instance within the current scope.  In this mode the same
    * instance of the trigger is used with every invocation. This may only be used if there is
    * NO NEED for support of variable definitions that are local to the trigger.  In nested
    * trigger invocations, such variable definitions must instead be implemented as inner 
    * class members for the recursive access to match the 4GL behavior.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    * @param    contain
    *           The containing object (all triggers are implemented as
    *           non-static inner classes, so they must have a reference to
    *           the containing class). Must not be <code>null</code>.
    * @param    trigger
    *           The specific instance of the trigger. Must not be <code>null</code>.
    */
   public static void registerTrigger(EventList events, Object contain, Supplier<Trigger> trigger)
   {
      registerTrigger(events,
                      new TriggerDefinition(trigger, contain, false), 
                      false);
   }
   
   /**
    * Register a specific trigger instance within the current scope.  In this mode the same
    * instance of the trigger is used with every invocation. This may only be used if there is
    * NO NEED for support of variable definitions that are local to the trigger.  In nested
    * trigger invocations, such variable definitions must instead be implemented as inner 
    * class members for the recursive access to match the 4GL behavior.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    * @param    contain
    *           The containing object (all triggers are implemented as
    *           non-static inner classes, so they must have a reference to
    *           the containing class). Must not be <code>null</code>.
    * @param    trans
    *           <code>true</code> if this trigger should start a full
    *           transaction. <code>false</code> if the trigger should be
    *           a sub-transaction.
    * @param    trigger
    *           The specific instance of the trigger. Must not be <code>null</code>.
    */
   public static void registerTrigger(EventList         events,
                                      Object            contain,
                                      boolean           trans,
                                      Supplier<Trigger> trigger)
   {
      registerTrigger(events,
                      new TriggerDefinition(trigger, contain, trans), 
                      false);
   }
   
   /**
    * Register a specific trigger instance within the current scope.  In this mode the same
    * instance of the trigger is used with every invocation. This may only be used if there is
    * NO NEED for support of variable definitions that are local to the trigger.  In nested
    * trigger invocations, such variable definitions must instead be implemented as inner 
    * class members for the recursive access to match the 4GL behavior.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    * @param    contain
    *           The containing object (all triggers are implemented as
    *           non-static inner classes, so they must have a reference to
    *           the containing class). Must not be <code>null</code>.
    * @param    trigger
    *           The specific instance of the trigger. Must not be <code>null</code>.
    */
   public static void registerTrigger(EventList events,
                                      Object    contain,
                                      Trigger   trigger)
   {
      registerTrigger(events,
                      new TriggerDefinition(() -> trigger, contain, false), 
                      false);
   }
   
   /**
    * Register a specific trigger instance within the current scope.  In this mode the same
    * instance of the trigger is used with every invocation. This may only be used if there is
    * NO NEED for support of variable definitions that are local to the trigger.  In nested
    * trigger invocations, such variable definitions must instead be implemented as inner 
    * class members for the recursive access to match the 4GL behavior.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    * @param    contain
    *           The containing object (all triggers are implemented as
    *           non-static inner classes, so they must have a reference to
    *           the containing class). Must not be <code>null</code>.
    * @param    trans
    *           <code>true</code> if this trigger should start a full
    *           transaction. <code>false</code> if the trigger should be
    *           a sub-transaction.
    * @param    trigger
    *           The specific instance of the trigger. Must not be <code>null</code>.
    */
   public static void registerTrigger(EventList events,
                                      Object    contain,
                                      boolean   trans,
                                      Trigger   trigger)
   {
      registerTrigger(events,
                      new TriggerDefinition(() -> trigger, contain, trans), 
                      false);
   }

   /**
    * Registers a trigger within the current scope.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    * @param    td
    *           The definition of the trigger that allows deferred and
    *           repeated instantiation of a fresh instance of the defined
    *           trigger.
    * @param    persistent
    *           TODO
    */
   public static void registerTrigger(EventList         events,
                                      TriggerDefinition td,
                                      boolean           persistent)
   {
      if (!validEventList(events))
      {
         return;
      }
      
      LogicalTerminal lt = locate();
      int tid = Utils.uniqueId();
      lt.triggerRegistry.put(new Integer(tid), td);
      lt.triggerChangeCount++;
      if (persistent)
      {
         lt.persistentTriggers.add(tid);
      }

      // invalidate the cache
      lt.cachedEventList = null;

      TriggerManager.register(lt.scope, events, tid);
      
      int[][] list = events.getRowDisplayEvents();
      if (list != null)
      {
         // if there is at least one ROW-DISPLAY event, send it to the client
         lt.client.registerRowDisplayEvents(list[0], list[1]);
      }
   }
   
   /**
    * Check if the given trigger is persistent (was added to the 
    * {@link #persistentTriggers} set.
    * 
    * @param    triggerId
    *           The trigger's ID.
    *           
    * @return   <code>true</code> if the given trigger is persistent.
    */
   public boolean isPersistentTrigger(int triggerId)
   {
      return persistentTriggers.contains(triggerId);
   }

   /**
    * Check if the top (most recent) trigger is a persistent trigger.
    *
    * @return <code>true</code> if the top (most recent) trigger is a persistent trigger.
    */
   public static boolean isTopTriggerPersistent()
   {
      LinkedList<Boolean> triggers = locate().triggerPersistent;
      return !triggers.isEmpty() && triggers.getFirst();
   }

   /**
    * Remove the ROW-DISPLAY triggers from the specified list.
    *
    * @param    triggers
    *           The triggers to be removed, if they are associated to a
    *           ROW-DISPLAY event. May be <code>null</code>.
    */
   public static void removeRowDisplayEvents(int[] triggers)
   {
      if (triggers != null && triggers.length > 0)
      {
         locate().client.removeRowDisplayEvents(triggers);
      }
   }
   
   /**
    * Deregisters a non-persistent trigger within the current scope for 
    * specified events and widgets.
    * 
    * @param    events
    *           List of events and widgets for which this trigger executes.
    */
   public static void deregisterTrigger(EventList events)
   {
      LogicalTerminal lt = locate();

      if (!lt.isPersistentTrigger(events.getTriggerId()))
      {
         // invalidate the cache
         lt.cachedEventList = null;

         TriggerManager.deregister(lt.scope, events);
      }
   }

   /**
    * Deregisters a non-persistent trigger entirely by its ID.
    * 
    * @param    triggerId
    *           The trigger ID to deregister.
    */
   public void deregisterTrigger(int triggerId)
   {
      deregisterTrigger(triggerId, false);
   }

   /**
    * Deregisters a trigger entirely by its ID, depending on its state and
    * on the <code>force</code> flag.
    * 
    * @param    triggerId
    *           The trigger ID to deregister.
    * @param    force
    *           flag indicating that the trigger should be deregistered,
    *           regardless of its state, persistent or non-persistent
    */
   public void deregisterTrigger(int triggerId, boolean force)
   {
      if (force || !isPersistentTrigger(triggerId))
      {
         triggerRegistry.remove(triggerId);
         persistentTriggers.remove(triggerId);
         triggerChangeCount++;
      }
   }
   
   /**
    * Get the trigger definition for the specified trigger ID.
    * 
    * @param    triggerId
    *           The trigger ID.
    *           
    * @return   The trigger's definition.
    */
   public TriggerDefinition getTriggerDefinition(int triggerId)
   {
      return triggerRegistry.get(triggerId);
   }

   /**
    * Signals a trigger to the application.  If a trigger is registered for
    * this event and source widget combination, then it will instantiated
    * and then executed.
    * <p>
    * Use this method to trigger directly from server, for example when the trigger is initiated by a
    * legacy property assignment.
    * <p>
    * This is an initial implementation, with some issues to be resolved. One of such issues is the
    * server-client state synchronization for example when new triggers get registered or unregistered
    * during the trigger invocation.
    *
    * @param    eventId
    *           The event which was generated.
    * @param    sourceWidgetId
    *           The widget which generated the event.
    *
    * @return   A container including the <code>ScreenBuffer</code>
    *           representing the latest screen values of all widgets and the
    *           consume flag.
    */
   public static ServerReturnValue trigger(int eventId, int sourceWidgetId)
   {
      TriggerMatch match = findTrigger(eventId, sourceWidgetId);
      if (match == null || match.triggerId == -1)
      {
         return null;
      }

      return trigger(match.triggerId, eventId, sourceWidgetId, -1, -1, -1, null);
   }
   
   /**
    * Find a trigger associated with an event and widget application.  If a trigger is registered for
    * this event and source widget combination, then it will has a valid trigger id. Otherwise, the 
    * trigger id in the wrapper will be null or {@code -1}.
    *
    * @param    eventId
    *           The event which was generated.
    * @param    sourceWidgetId
    *           The widget which generated the event.
    *
    * @return   A wrapper holding information for the trigger which was searched.
    */
   public static TriggerMatch findTrigger(int eventId, int sourceWidgetId)
   {
      LogicalTerminal lt = locate();
      EventList wl = (lt.cachedEventList == null 
                        ? lt.cachedEventList = TriggerManager.mergeTriggerRegistry(lt.scope) 
                        : lt.cachedEventList);

      // although the event list is available on the server in the current context, the lookup is forwarded
      // to the client as the lookup logic may use the widget tree to determine the result
      return LogicalTerminal.getClient().lookupEvent(wl,
                                                     eventId,
                                                     -1,
                                                     sourceWidgetId,
                                                     true,
                                                     false);
   }

   /**
    * Signals a trigger to the application.  If a trigger is registered for
    * this event and source widget combination, then it will instantiated
    * and then executed.
    * 
    * @param    triggerId
    *           The trigger to be called.
    * @param    eventId
    *           The event which was generated.
    * @param    sourceId
    *           The widget which generated the event.
    * @param    otherId
    *           The companion widget id (for ENTER/LEAVE).
    * @param    resourceId
    *           The resource which generated the event.
    * @param    matchedWidgetId
    *           Id of the widget that actually matched the trigger.
    * @param    sb
    *           The editable frames <code>ScreenBuffer</code> array
    *
    * @return   A container including the <code>ScreenBuffer</code> 
    *           representing the latest screen values of all widgets and the
    *           consume flag.
    */
   public static ServerReturnValue trigger(int          triggerId,
                                           int          eventId,
                                           int          sourceId,
                                           int          otherId,
                                           long         resourceId,
                                           int          matchedWidgetId,
                                           ScreenBuffer[] sb)
   {
      LogicalTerminal lt = locate();

      // identify the widget/procedure
      WrappedResource res = null;
      
      if (resourceId != -1)
      {
         // fallback to procedure
         handle h = handle.fromResourceId(resourceId);
         res = (h == null || !h._isValid()) ? null : h.getResource(); 
      }
      else
      {
         res = getWidgetForId(sourceId);
      }
      
      if (res == null)
      {
         if (LOG.isLoggable(Level.WARNING))
         {
            String noRes = "TRIGGER ERROR source not found for trigger ID %d, event ID %d " +
                           "(%s), source ID %d, resource ID %d";
                           
            String ename = Keyboard.eventName(eventId);
            
            LOG.log(Level.WARNING, 
                    String.format(noRes, triggerId, eventId, ename, sourceId, resourceId));
         }
         
         return null;
      }
      
      // post the updated ScreenBuffer to the frame
      GenericFrame frame = null;
      
      if (sb != null)
      {
         frame = lt.identifyFrame(sb[0].getBaseId());
      }
      
      if (frame != null)
      {
         // and handling the screen buffers starting with default one if any
         handleScreenBuffers(frame, sb, false);

         boolean clearLastKey = false;
         // to avoid get for id duplication
         GenericWidget<?> widOther = getWidgetForId(otherId);

         // in some cases LASTKEY must report -1 and LAST-EVENT:EVENTTYPE "PROGRESS":
         // (1) when LEAVE due to a mouse event with BUTTON widget as target
         if (eventId == Keyboard.SE_LEAVE &&
             Keyboard.isMouseEvent(sb[0].getKeyCode()) &&
             otherId >= WidgetId.WIDGET_ID_RANGE_BEGIN &&
             (widOther instanceof ButtonWidget ||
              (widOther instanceof BrowseColumnWidget && sourceId == otherId)))
         {
            clearLastKey = true;
         }
         // (2) when ENTRY on a BUTTON due to a mouse event
         else if (eventId == Keyboard.SE_ENTRY &&
                  Keyboard.isMouseEvent(sb[0].getKeyCode()) &&
                  res instanceof ButtonWidget)
         {
            clearLastKey = true;
         }
         // (3) when CHOOSE on a BUTTON
         else if (eventId == Keyboard.SE_CHOOSE &&
                  res instanceof ButtonWidget)
         {
            clearLastKey = true;
         }

         if (clearLastKey)
         {
            KeyReader.setLastKey(-1);
         }
      }
      
      String spec = null;
      
      // init logging details
      if (LOG.isLoggable(Level.FINER))
      {
         String ename = Keyboard.eventName(eventId);
         String fname = (frame == null) ? "none" : frame._getName();
         String wname = null;
         String cname = res.getClass().getName();
         
         if (res instanceof GenericWidget)
         {
            GenericWidget<?> wid = (GenericWidget<?>) res;
            
            wname = wid.name().toStringMessage();
         }
         
         String src = String.format("%s type %s", wname == null ? "unknown" : wname, cname);
         
         spec = String.format("TRIGGER %%s EVENT %s ON %s (frame %s)", ename, src, fname);
      }
      
      // identify the trigger for this event
      TriggerDefinition def = lt.triggerRegistry.get(new Integer(triggerId));
                                            
      if (def == null)
      {
         if (LOG.isLoggable(Level.FINER))
         {
            LOG.log(Level.FINER, String.format(spec, "DEFINITION MISSING for"));
         }
         
         return null;
      }
      
      // check if the external procedure still exists
      if (!lt.pm.hasReferent(def.getExternalProgram()))
      {
         // log that we are invoking a trigger in a deleted procedure
         return null;
      }
      
      // create a new instance of the trigger (this MUST be done for EVERY
      // invocation because it allows any of the trigger's data members to
      // be initialized to the proper default values), this is why we store
      // trigger instantiation metadata instead of a trigger instance
      Trigger trig = def.getTrigger();
      
      if (trig == null)
      {
         if (LOG.isLoggable(Level.FINER))
         {
            LOG.log(Level.FINER, String.format(spec, "INSTANTIATION FAILURE for"));
         }
         
         return null;
      }
      
      // save the previous consume flag
      boolean consume    = false;
      boolean oldConsume = lt.consumeEvent;
      
      // reset the consume flag
      lt.consumeEvent = false;
      
      // prepare the result
      TriggerReturnValue trv = new TriggerReturnValue();
      trv.consume = false;
      trv.buffer  = null;
      
      ServerReturnValue srv = new ServerReturnValue();
      srv.payload = trv;

      if (LOG.isLoggable(Level.FINER))
      {
         LOG.log(Level.FINER, String.format(spec, "BEGIN"));
      }
      
      // avoid recursion by looking up the widget in the trigger stack
      // and the event in the synced event stack and bypass it if the pair
      // matches
      
      String eventName = Keyboard.eventName(eventId);
      for (int i = 0; i < lt.triggerResource.size(); i ++)
      {
         WrappedResource triggerRes = lt.triggerResource.get(i);
         GenericWidget<?> triggerWidget = triggerRes instanceof GenericWidget ?
                                       (GenericWidget<?>) triggerRes : null;

         if (!triggerRes.equals(res) &&
             !(triggerWidget != null && WidgetId.match(triggerWidget.getId(), matchedWidgetId)) &&
             !(matchedWidgetId != -1 && lt.triggerMatchedWidgets.get(i).equals(matchedWidgetId)))
         {
            continue;
         }

         int triggerEventId = lt.triggerEvent.get(i).intValue();
         
         // recursion of the widget; first the simple check for equal codes
         if (triggerEventId == eventId)
         {
            if (LOG.isLoggable(Level.FINER))
            {
               LOG.log(Level.FINER, String.format(spec, "RECURSION BYPASS for"));
            }         

            // the pair matches entirely; ignore trigger
            return srv;
         }
         
         String triggerEventName = Keyboard.eventName(triggerEventId);
         
         // also compare event names; for example consider TAB event
         // which can have two different codes - ASCII 9 and app code -187
         if (triggerEventName != null && eventName != null &&
             !triggerEventName.isEmpty() && !eventName.isEmpty() && 
             triggerEventName.equals(eventName))
         {
            if (LOG.isLoggable(Level.FINER))
            {
               LOG.log(Level.FINER, String.format(spec, "RECURSION BYPASS for"));
            }         

            // the pair matches entirely; ignore trigger
            return srv;
         }
      }
      
      // call the trigger
      lt.triggerResource.addFirst(res);
      lt.triggerEvent.addFirst(new Integer(eventId));
      lt.triggerMatchedWidgets.addFirst(matchedWidgetId);
      lt.triggerPersistent.addFirst(lt.isPersistentTrigger(triggerId));

      try
      {
         SelfManager.pushSelf(_self());
         
         if (frame != null)
            frame.triggerEnter();
         
         // the TARGET-PROCEDURE needs to be the external program where the trigger was defined.
         Object target = def.getExternalProgram();
         
         Class<?> cls = target.getClass();
         lt.pm.pushCalleeInfo(cls, false, target, target, 
                              ProcedureManager.TRIGGER_PROCEDURE, null,
                              false, false);
         
         lt.setEventInfo(new EventInfo(eventName, sourceId, otherId));

         // save the frame statements for each live frame
         Map<Integer, UIStatement> frameStmts = lt.saveFrameStatements();
         try
         {
            lt.invokeOnServer(srv, () -> { trig.run(); return null; });
         }
         finally
         {
            lt.clearEventInfo();
            
            lt.restoreFrameStatements(frameStmts);
         }
      }

      finally
      {
         SelfManager.popSelf();
         
         lt.pm.popCalleeInfo();
         
         if (frame != null)
            frame.triggerLeave();
         
         lt.triggerResource.removeFirst();
         lt.triggerEvent.removeFirst();
         lt.triggerMatchedWidgets.removeFirst();
         lt.triggerPersistent.removeFirst();

         // restore the old consume flag
         consume = lt.consumeEvent;
         lt.consumeEvent = oldConsume;
      }
      
      // prepare the result
      trv.consume = consume;
      trv.buffer  = (frame != null) ? lt.getEditableServerScreenBuffers(frame) : null;
      
      if (LOG.isLoggable(Level.FINER))
      {
         LOG.log(Level.FINER,
                 String.format(spec, "END") + " - NO-APPLY = " + (consume ? "TRUE" : "FALSE"));
      }

      // TODO: make the data part of the payload
      lt.flushEnqueuedWidgetAttrs();

      return srv;
   }
   
   /**
    * Signals a validation event to the application.
    * <p>
    * This signal is used to trigger validation execution on the server.
    * 
    * @param   sourceId
    *          widget ID of the widget generated the trigger.
    * @param   sb
    *          updated <code>ScreenBuffer</code> array
    * 
    * @return  <code>ValidationResult</code> representing the validation  
    *          result, error message and screen value changes.
    */
   public static ValidationResult validate(int sourceId, ScreenBuffer[] sb)
   {
      String validationError = null;
      boolean result = false;
      ScreenBuffer[] diff = null;
      LogicalTerminal lt = locate();
      
      // identify the widget
      GenericWidget<?> widget = getWidgetForId(sourceId);
      
      if (widget == null)
      {
         // bogus widget, return something safe
         return new ValidationResult(true, null, null);
      }
      
      // post the updated ScreenBuffer to the frame
      GenericFrame frame = widget.getFrame();
      // post the updated ScreenBuffers to the respective frames
      handleScreenBuffers(frame, sb, false);
      
      // get the proposed widget value
      BaseDataType val = (BaseDataType) sb[0].getWidgetValue(sourceId);
      
      Class<?> type = ((ControlEntity<?>) widget).getDataClass();
      
      // use the existing value if no proposed value is present (unchanged)
      if (val == null)
      {
         // doesn't need to update dynamic data (headers...) to just get the value out
         Object value = frame.getFrameBufferRaw().getWidgetValue(sourceId);
         
         val = (BaseDataType) frame.convertScreenValue(type, value, widget._getFormat()); 
      }
      
      if (val == null)
      {
         val = (BaseDataType) frame.getter(sourceId, type, true, false);                              
      }
      
      // validate the change
      validationError = widget.validate(val, false);

      // create and return the validation result instance
      if (validationError == null)
         result = true;
      
      diff = lt.getEditableServerScreenBuffers(frame);
      
      return new ValidationResult(result, validationError, diff); 
   }
      
   /**
    * Locates and executes the HELP program (applhelp.p) if it exists in the current propath. The
    * first program found of that name will be executed (if multiple exist). This should only be
    * called if there was no registered trigger fired for this specific HELP event.
    *
    * @param   lastKey
    *          The most recent key event code 
    * @param   buffer
    *          The screen buffer array associated with the FOCUS as before HELP is invoked and
    *          other editable frames.
    */
   public static void help(int lastKey, ScreenBuffer[] buffer)
   {
      KeyReader.setLastKey(lastKey);
      
      character program = FileSystemOps.searchPath(APPLHELP);
      
      if (program.isUnknown())
      {
         message("No application help is available");
         return;
      }
      
      LogicalTerminal lt = locate();
      
      try
      {
         lt.helpActive = true;
         if (buffer != null)
         {
            lt.hlpFrameFieldId = buffer[0].getValueWidgetId();
            lt.hlpFrameRef = lt.identifyFrame(buffer[0].getBaseId());
            
            // let the other FRAME- states update, too
            handleScreenBuffers(lt.hlpFrameRef, buffer, false);
         }
         else
         {
            lt.hlpFrameFieldId = lt.frameFieldId;
            lt.hlpFrameRef = lt.frameRef;
         }

         ControlFlowOps.invoke(APPLHELP, new Object[0]);
      }
      finally
      {
         lt.hlpFrameFieldId = -1;
         lt.hlpFrameRef = null;
         lt.helpActive = false;
      }
   }

   /**
    * Obtains the latest screen buffer associated with the given frame.
    *
    * @param   frameId
    *          The frame for which the screen buffer should be returned.
    *
    * @return  <code>ScreenBuffer</code> array representing the current screen 
    *          values of all widgets in the given frame and other editable frames.
    */
   public static ScreenBuffer[] refresh(int frameId)   
   {
      LogicalTerminal lt = locate();
      
      // identify the first entry in screen buffer array
      GenericFrame frame = lt.identifyFrame(frameId);
      
      if (frame == null)
      {
         throw new RuntimeException("Invalid frame ID " + frameId);
      }
      
      return lt.getEditableServerScreenBuffers(frame);
   }
   
   /**
    * Suppresses the default behavior for the current UI event so that it has
    * the effect of <b>RETURN NO-APPLY</b>.
    */
   public static void consumeEvent()
   {
      locate().consumeEvent = true;
   }

   /**
    * Checks if the given widget ID is in use by some widget or frame.
    *
    * @param    widgetId
    *           The widget ID to check.
    *
    * @return   <code>TRUE</code> if ID is in use, <code>FALSE</code> otherwise.
    */
   public static boolean isWidgetIdInUse(int widgetId)
   {
      LogicalTerminal lt = locate();
      
      synchronized (lt.registryLock)
      {
         return lt.widgetRegistry.containsKey(widgetId);
      }
   }
   
   /**
    * Obtains the widget handle in the frame's tab order next to the given widget.
    *
    * @param    frame
    *           The frame which defines the tab item order
    *
    * @param    widget
    *           The given widget
    *
    * @return   A valid widget handle if the next tab item exists, otherwise returns unknown.
    */
   public static handle nextTabItem(GenericFrame frame, GenericWidget<?> widget)
   {
      return frame.getFieldGroup().nextTabItem(widget);
   }
   
   /**
    * Obtains the previous widget handle in the frame's tab order to the given widget.
    *
    * @param    frame
    *           The frame which defines the tab item order
    *
    * @param    widget
    *           The given widget
    *
    * @return   A valid widget handle if the previous tab item exists, otherwise returns unknown.
    */
   public static handle previousTabItem(GenericFrame frame, GenericWidget<?> widget)
   {
      return frame.getFieldGroup().previousTabItem(widget);
   }
   
   /**
    * Adds the frame to the internal registry.
    *
    * @param    frame
    *           The instance of a <code>GenericFrame</code> to add to the
    *           registry.  Must not be <code>null</code> and the frame ID
    *           of this frame must be unique.
    *           
    * @return   ID assigned to the frame.
    */
   static int registerFrame(GenericFrame frame)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      synchronized (lt.registryLock)
      {
         // create the frame definition
         int id = WidgetId.nextID().asInt();

         // add it
         lt.frameRegistry.put(id, frame);
         
         // refresh widget registry
         lt.widgetRegistry.put(id, frame);

         return id;
      }
   }
   
   /**
    * Adds the menu to the internal registry.
    *
    * @param    menu
    *           The instance of a <code>MenuWidget</code> to add to the
    *           registry.  Must not be <code>null</code> and the menu ID
    *           of this menu must be unique.
    * @param    pending
    *           Flag indicating if this is a pending registration.
    *           
    * @return   ID assigned to the menu.
    */
   static int registerMenu(MenuContainerWidget<?> menu, boolean pending)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      synchronized (lt.registryLock)
      {
         // create the frame definition
         int id = WidgetId.nextID().asInt();
         
         // add it
         lt.menuRegistry.put(id, menu);
         
         // refresh widget registry
         lt.widgetRegistry.put(id, menu);
         
         if (pending)
         {
            // keep a reference to it in case it need to be cleaned up before scope starts
            lt.pendingMenus.add(id);
         }
         
         return id;
      }
   }
   
   /**
    * Deletes the frame from the internal registry.
    * 
    * @param   frameId
    *          The ID of frame which should be removed from registry.
    */
   static void deregisterFrame(int frameId)
   {
      LogicalTerminal lt = locate();
      
      // scopeable registration must be done in any location where a frame state (requiring scope support) is
      // being 'touched' - this includes frame de-registration, even if the current block is not explicitly
      // using frames
      lt.tm.registerBlockScopeable(lt);

      lt.deregisterFrameInt(frameId);
   }
   
   /**
    * Register a frame as being dead.  It will be sent to the client-side on the next trip (if it matches the criteria).
    *
    * @param   frameId
    *          The ID of the dead frame.
    */
   static void addDeadFrameId(int frameId)
   {
      locate().deadFrameIds.add(frameId);
   }
   
   /**
    * Deletes the menu from the internal registry.
    * 
    * @param   menuId
    *          The ID of menu which should be removed from registry.
    */
   static void deregisterMenu(int menuId)
   {
      locate().deregisterMenuInt(menuId);
   }
   
   /**
    * Replace a frame in the internal registry.  It is expected that a frame
    * previously has been registered with the same frame ID as the given
    * frame.
    * <p>
    * This method is intended to support shared frames, where the
    * <code>GenericFrame</code> instance backing a shared frame in one scope
    * must be replaced temporarily with a new instance to back a "matching"
    * frame in a more deeply nested scope.  This ensures that all frame and
    * widget lookups within the nested scope find the proper objects in the
    * registry.
    * 
    * @param   frame
    *          Replacement frame instance.
    * 
    * @return  Frame previoulsy registered under same ID.
    */
   static GenericFrame replaceFrame(GenericFrame frame)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      Integer id = frame.getFrameId();
      
      synchronized (lt.registryLock)
      {
         if (!lt.widgetRegistry.containsKey(id))
         {
            throw new IllegalStateException(
               "Expected frame registry to contain frame with ID:  " + id);
         }
         
         // getting the previous version of the frame to be replaced
         GenericFrame oldFrame = (GenericFrame)lt.widgetRegistry.get(id);
         
         // change frame registry and get the old value
         lt.frameRegistry.put(id, frame);
         
         // refresh widget registry
         // first remove old frame content
         if (oldFrame != null)
         {
            lt.deregisterFrameContentInt(oldFrame);
         }

         // and replace frame itself
         // the new widgets will be added later at frame initialization step
         lt.widgetRegistry.put(id, frame);

         return oldFrame;
      }
   }

   /**
    * Release shared frame reference from the inactive shared frame pool.
    * 
    * @param   frameId
    *          ID of the frame to release.
    */
   static void releaseSharedFrame(int frameId)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      synchronized (lt.registryLock)
      {
         // refresh widget registry
         lt.widgetRegistry.remove(frameId);

         // not deregistering content because it can be reused further
      }      
   }

   /**
    * Register the shared frame.
    * 
    * @param   frame
    *          The shared frame to add back to the registry.
    */
   static void registerSharedFrame(GenericFrame frame)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      synchronized (lt.registryLock)
      {
         // refresh widget registry
         lt.widgetRegistry.put(frame.getFrameId(), frame);
      }
   }
   
   /**
    * Refresh the server-side widget configs so that their state is in sync with the current
    * frame row, as determined by client-side.
    * 
    * @param    wcfgs
    *           The widget configs for the current frame row.
    */
   private static void processDownConfigs(WidgetConfig[] wcfgs)
   {
      if (wcfgs == null)
      {
         return;
      }

      ConfigManager cfg = ConfigManager.getInstance();
      Set<WidgetId> processed = new HashSet<>();
      
      for (int i = 0; i < wcfgs.length; i++)
      {
         WidgetConfig wcfg = wcfgs[i];
         
         // protection - each widget ID must appear only once!
         if (processed.contains(wcfg.id))
         {
            throw new IllegalStateException("Widget ID " + wcfg.id + 
                                            " has already been processed.");
         }
         
         processed.add(wcfg.id);
         
         // standardize the widget ID - server-side has no "down body" widgets
         WidgetId wid = new WidgetId(wcfg.id.asInt());
         
         WidgetConfig active = cfg.getActiveConfig(wid);
         if (active != null)
         {
            active.applyConfig(wcfg);
         }
      }
   }

   /**
    * Prepares the array of the screen buffers to pass to the client when calling the client
    * export methods or returning from server export to be able the screen buffer synchronization.
    *
    * @param   frDefault
    *          The default frame is used as first element in array.
    *
    * @return  The array of the screen buffers to sync.
    *
    */
   private ScreenBuffer[] getEditableServerScreenBuffers(GenericFrame frDefault)
   {
      ArrayList<ScreenBuffer> sbArrayList = new ArrayList<ScreenBuffer>();
      // add the buffer from default frame if not null as the first entry
      if (frDefault != null)
      {
         sbArrayList.add(frDefault.getFrameBuffer(true));
      }
      // check to see if we need more screen buffers to add
      synchronized (registryLock)
      {
         Iterator<GenericFrame> frIter = editableFrameRegistry.keySet().iterator();
         // the ID of the current frame
         int defFrameId = frDefault != null ? frDefault.getFrameWidget().getId() : -1;
         while (frIter.hasNext())
         {
            // get next frame from editable frame list
            GenericFrame gfr = frIter.next();
            // create screen buffer for the given frame if it is not the default one
            if (defFrameId != gfr.getFrameWidget().getId())
            {
               ScreenBuffer moreScreenBuff = gfr.getFrameBuffer(false);
               // and add it to the list for further processing
               if (moreScreenBuff != null)
               {
                  sbArrayList.add(moreScreenBuff);
               }
            }
         }
      }
      // convert array list to array and return to the caller
      return sbArrayList.isEmpty() ? null
                                   : sbArrayList.toArray(new ScreenBuffer[sbArrayList.size()]);
   }
   
   /**
    * Possibly add the frame with given widget ID to the editable frames list if it is not already
    * there.
    * 
    * @param    widFrame
    *           instance of <code>GenericFrame</code> this request came for
    * @param    widIdToCheck
    *           The widget ID to check.
    */
   public void addEditableFrameForWidgets(GenericFrame widFrame, int[] widIdToCheck)
   {
      // check for NPE
      if (widFrame == null || widIdToCheck == null)
      {
         return;
      }
      
      // adding the frame for given widget if it is not already in cache
      synchronized (registryLock)
      {
         // getting the ID set from frame that already in the list
         RoaringBitmap widIdSet = editableFrameRegistry.get(widFrame);
         
         // new frame to add
         if (widIdSet == null)
         {
            widIdSet = new RoaringBitmap();
            editableFrameRegistry.put(widFrame, widIdSet);
         }
         
         // and add the widget ID to the set for respective frame
         for (int widId : widIdToCheck)
         {
            widIdSet.add(widId);
         }
      }
   }

   /**
    * Possibly remove the frame with given widget ID from the editable frames list if there is no
    * more enabled focused widgets in a frame left.
    * 
    * @param    widFrame
    *           instance of <code>GenericFrame</code> this request came for
    * @param    widIdToCheck
    *           The widget ID to check.
    */
   public void removeEditableFrameForWidgets(GenericFrame widFrame, int[] widIdToCheck)
   {
      // check for NPE
      if (widFrame == null || widIdToCheck == null)
      {
         return;
      }
      
      // removing the frame for given widget if it is in cache
      synchronized (registryLock)
      {
         // check if the frame is in the list
         RoaringBitmap widIdSet = editableFrameRegistry.get(widFrame);
         // if none - do nothing
         if (widIdSet != null)
         {
            // frame in the list, removing current widget ID array from it's set
            for (int widId : widIdToCheck)
            {
               widIdSet.remove(widId);
            }
            // check if this was the last widget in a frame
            if (widIdSet.isEmpty())
            {
               // remove the frame from the list
               editableFrameRegistry.remove(widFrame);
            }
         }
      }
   }

   /**
    * Internal worker for {@link #deregisterFrame(int)}.
    * 
    * @param   frameId
    *          The ID of frame which should be removed from registry.
    */
   private void deregisterFrameInt(int frameId)
   {
      synchronized (registryLock)
      {
         Integer id = new Integer(frameId);
         
         GenericFrame frame =  frameRegistry.remove(id);
         
         // deregisters all static widgets including header widget in a given frame
         // within the same lock
         if (frame != null && !frame.isShared())
         {
            widgetRegistry.remove(id);
            deregisterFrameContentInt(frame);
         }
         
         // process delete notifications for all widgets part of the frame
         if (frame != null)
         {
            frame.frameDelete();
            // remove from editable frames cache
            editableFrameRegistry.remove(frame);
         }

         if (scopeTrack != null)
         {
            if (scopeTrack.hasFrames())
            {
               scopeTrack.frames.remove(id);
            }
            if (scopeTrack.hasDeepFrames())
            {
               scopeTrack.deepFrames.remove(id);
            }
            
            if (scopeTrack.prev != null)
            {
               if (scopeTrack.prev.hasFrames())
               {
                  scopeTrack.prev.frames.remove(id);
               }
               if (scopeTrack.prev.hasDeepFrames())
               {
                  scopeTrack.prev.deepFrames.remove(id);
               }
            }
         }
      }
   }
   
   /**
    * Internal worker for {@link #deregisterMenu(int)}.
    * 
    * @param   menuId
    *          The ID of menu which should be removed from registry.
    */
   private void deregisterMenuInt(int menuId)
   {
      synchronized (registryLock)
      {
         Integer id = new Integer(menuId);
         
         MenuContainerWidget<?> menu =  menuRegistry.remove(id);
         
         // deregisters all static widgets in a given menu within the same lock
         if (menu != null)
         {
            widgetRegistry.remove(id);
         }
      }
   }
   
   /**
    * Internal worker for content cleanup.
    * 
    * @param   frameToProcess
    *          The frame to remove all containing widgets from registry.
    */
   private void deregisterFrameContentInt(GenericFrame frameToProcess)
   {
      int[] wId = frameToProcess.getStaticWidgetIds(true);
         
      if (wId == null)
      {
         // nothing to do
         return;
      }
      
      synchronized (registryLock)
      {
         for (int i = 0; i < wId.length; i++)
         {
            // remove widget from registry
            widgetRegistry.remove(wId[i]);
         }
      }
   }
   
   /**
    * Adds new dynamic frame to the application context.  It will be removed when respective
    * frame widget is removing via {@link FrameWidget#resourceDelete()}.
    * 
    * 
    * @param   dynFrame
    *          The dynamic frame widget to be added to registry.
    * @param   idFrame
    *          The widget ID to be used for this frame
    */
   static void addDynamicFrame(GenericFrame dynFrame, int idFrame)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      synchronized (lt.registryLock)
      {
         // add dynamic frame to frame registry
         lt.frameRegistry.put(idFrame, dynFrame);
         
         // refresh widget registry
         lt.widgetRegistry.put(idFrame, dynFrame);
      }
   }
   
   /**
    * Internal worker for {@link #applyChanges(Externalizable)}.
    * 
    * @param   frameId
    *          The ID of frame which should be removed from registry.
    * @param   down
    *          New DOWN value.         
    */
   private void refreshFrameDown(int frameId, int down)
   {
      synchronized (registryLock)
      {
         GenericFrame frame = frameRegistry.get(new Integer(frameId));
         
         if (frame != null)
            frame.refreshDown(down);
      }
   }
   
   /**
    * Obtains the frame associated with the given frame ID.
    *
    * @param    frameId
    *           The frame ID of the frame to identify.
    *
    * @return   An instance of <code>GenericFrame</code> or <code>null</code>.
    */
   private GenericFrame identifyFrame(int frameId)
   {
      synchronized (registryLock)
      {
         return frameRegistry.get(frameId);
      }
   }
   
   /**
    * Notify scope that frame is viewed.
    * <p>
    * This also registeres {@link LogicalTerminal} for scope notifications at the current block.
    *  
    * @param   frame
    *          Frame which is about to be viewed.
    */
   private void frameViewed(GenericFrame frame)
   {
      tm.registerBlockScopeable(this);
      
      scopeTrack.frameViewed(frame);
   }

   /**
    * Clear map of viewed frames.
    * <p>
    * This also registeres {@link LogicalTerminal} for scope notifications at the current block.
    */
   private void forceHiding()
   {
      tm.registerBlockScopeable(this);
      
      scopeTrack.forceHiding();
   }
   
   /**
    * Save the {@link UIStatement} running for each registered frame.
    * 
    * @return   A map with each frame's UI statement.
    */
   private Map<Integer, UIStatement> saveFrameStatements()
   {
      Map<Integer, UIStatement> res = new HashMap<>();
      
      synchronized (registryLock)
      {
         for (GenericFrame frame : frameRegistry.values())
         {
            res.put(frame.getFrameId(), frame.getStatement());
            
            frame.resetStatement();
         }
      }
      
      return res;
   }
   
   /**
    * Restore the UI statement for each registered frame to the saved value.
    * 
    * @param    state
    *           The state, as computed by {@link #saveFrameStatements()}.
    */
   private void restoreFrameStatements(@SuppressWarnings("hiding") Map<Integer, UIStatement> state)
   {
      for (Integer frameId : state.keySet())
      {
         GenericFrame frame = identifyFrame(frameId);

         if (frame != null)
         {
            frame.resetStatement();
            frame.setStatement(state.get(frameId));
         }
      }
   }
   
   /**
    * Retrieve the ID of the frame containing the widget with a given ID.
    *
    * @param    widgetId
    *           The widget ID.
    *
    * @return   The frame ID if widget is found or -1 if not.
    */
   public static int frameIdFromWidgetId(int widgetId)
   {
      return locate().lookupFrameIdFromWidgetId(widgetId);
   }

   /**
    * Returns all parent frames up the root frame starting from the provided widget.
    * The frames are sorted from the most inner up to the root frame. If no frames
    * are found, an empty list is returned. If the given widget is itself a frame,
    * then it will be included in the resulting list.
    *
    * @param    widgetId
    *           A widget id.
    *
    * @return   See above, if no frames are found, returns an empty list.
    */
   public static List<Integer> getParentFrames(int widgetId)
   {
      LogicalTerminal lt = locate();
      synchronized (lt.registryLock)
      {
         // get the widget with given ID
         Object widObject = lt.widgetRegistry.get(widgetId);

         if (widObject == null)
         {
            return Collections.emptyList();
         }

         List<Integer> result = new ArrayList<>();
         while (widObject != null)
         {
            int wid = -1;

            // we can store either GenericWidget object
            if (widObject instanceof GenericWidget)
            {
               // getting frame ID from widget config
               wid = ((GenericWidget<?>) widObject).config().frameId;
            }
            else // or GenericFrame
            {
               // getting frame ID from class method
               wid = ((GenericFrame) widObject).getFrameId();
            }

            if (wid != -1)
            {
               widObject = lt.widgetRegistry.get(wid);
               if (widObject != null)
               {
                  result.add(wid);
               }
            }
            else
            {
               widObject = null;
            }
         }

         return result;
      }
   }
   /**
    * Find a widget inside the given container under the given mouse pointer if there is a one.
    * 
    * @param    widgetId
    *           The widget ID of the given container
    * @param    x
    *           The x-coordinate of the given mouse pointer
    * @param    y
    *           The y-coordinate of the given mouse pointer
    * 
    * @return   The target widget id under the given mouse pointer or null
    */
   public static Integer findWidgetByCoordinates(int widgetId, int x, int y)
   {
      return locate().client.findWidgetByCoordinates(widgetId, x, y);
   }

   /**
    * Retrieve the ID of the frame containing the widget with a given ID.
    *
    * @param    widgetId
    *           The widget ID.
    *
    * @return   The frame ID if widget is found or -1 if not.
    */
   private int lookupFrameIdFromWidgetId(int widgetId)
   {
      synchronized (registryLock)
      {
         // get the widget with given ID
         Object widObject = widgetRegistry.get(widgetId);

         if (widObject == null)
         {
            return -1;
         }
         
         // id to return
         int resultId = -1;
         
         // we can store either GenericWidget object
         if (widObject instanceof GenericWidget)
         {
            // getting frame ID from widget config
            resultId = ((GenericWidget<?>) widObject).config().frameId;
         }
         else // or GenericFrame
         {
            // getting frame ID from class method
            resultId = ((GenericFrame)widObject).getFrameId();
         }
      
         // verify the frame exists in widget registry
         return (resultId != -1 && widgetRegistry.containsKey(resultId)) ? resultId : -1;
      }
   }
   
   /**
    * Prepare text for pause message.
    * 
    * @param   text
    *          The text to be prepared in pause message.
    * 
    * @return prepared text.
    */
   private static String pauseText(character text)
   {
      return (text == null || text.isUnknown()) ? null : text.toStringMessage();
   }

   /**
    * Gets the instance of the LogicalTerminal for the current context. If it
    * is the first time call for the session, a new instance is created.
    * 
    * @return  the instance of <code>LogicalTerminal</code> for this session.
    */
   private static LogicalTerminal locate()
   {
      return instance.get();
   }
   
   /**
    * Retrieve the widget associated with a given ID.
    * 
    * @param    widgetId
    *           The widget ID.
    * 
    * @return   The widget which corresponds to the given widget ID or
    *           <code>null</code> if the widget ID is invalid.
    */
   public static GenericWidget<?> getWidgetForId(int widgetId)
   {
      return locate().getWidgetForIdInt(widgetId);
   }
   
   /**
    * Gets the widget's resource id.
    * 
    * @param    widgetId
    *           The widget id.
    * 
    * @return   The resource id.
    */
   public static Long getResourceIdByWidgetId(int widgetId)
   {
      GenericWidget<?> widget = getWidgetForId(widgetId);
      
      if (widget != null)
      {
         return widget.id();
      }
      
      return null;
   }
   
   /**
    * Retrieve the widget or header widget associated with a given ID.
    *
    * @param    widgetId
    *           The widget ID.
    * 
    * @return   The widget which corresponds to the given widget ID, which may
    *           be a frame, or <code>null</code> if the widget ID is invalid.
    */
   protected GenericWidget<?> getWidgetForIdInt(int widgetId)
   {
      synchronized (registryLock)
      {
         if (widgetId == WidgetId._DEFAULT_WINDOW_ID)
         {
            return (WindowWidget) defaultWindow().getResource();
         }
         // TODO: else if (widgetId < WidgetId.DEFAULT_WINDOW_ID)
         // there are a lot of pseudo widgets with ID == -1, like labels so including -1 in
         // log region cause the server.log file to be too big so we need another criteria
         else if (widgetId < -1)
         {
            // log for error
            LOG.warning("Attempt to request widget for invalid ID!");
            return null;
         }
         
         Object regValue = widgetRegistry.get(widgetId);
         
         // return value from widget registry, it can be either GenericWidget or GenericFrame
         return (regValue == null || regValue instanceof GenericWidget)
                ? (GenericWidget<?>) regValue 
                : ((GenericFrame)regValue).asWidget();
      }
   }
   
   /**
    * Finds the first visible valid window according their creation order if there exists a target
    * window, otherwise returns the default window.
    * 
    * @return The first visible window.
    */
   private WindowWidget getFirstVisibleWindow()
   {
      synchronized (registryLock)
      {
         return this.windowRegistry.values().stream().filter(
                  (win) -> win._isVisible() && win.valid()).findFirst().orElse(this.defaultWindow);
      }
   }
   
   /**
    * Cleanup the dynamic frames, as the client is resetting.
    */
   public static void cleanupDynamicFrames()
   {
      LogicalTerminal lt = locate();
      List<GenericFrame> allFrames = new ArrayList<>(lt.frameRegistry.values());
      for (GenericFrame frame : allFrames)
      {
         FrameWidget fw = frame.getFrameWidget();
         
         if (fw._dynamic())
         {
            fw.delete();
         }
      }
      
      // make a bogus client-side call to force the delete on client-side, too
      lt.client.isChui();
   }

   /**
    * Mark the specified widgets to be destroyed, from the client-side.
    * 
    * @param    widgetIds
    *           A list of widget IDs.
    */
   public static void destroyWidgets(int[] widgetIds)
   {
      LogicalTerminal lt = locate();
      lt.addMarkEntry(new MarkEntry(widgetIds, MarkEntry.DESTROY_WIDGETS));
   }
   
   /**
    * Notifies the logical terminal that a frame is going out of scope.
    * 
    * @param    frame
    *           The frame requiring cleanup.
    */
   public static void destroyFrame(GenericFrame frame)
   {
      int id = frame.getFrameId();
      
      LogicalTerminal lt = locate();
      
      if (frame.isDialogBox() || !frame._isVisible() || frame.asWidget().config.overlay)
      {
         // this must be synchronous since it must hide right now
         lt.client.destroyFrame(id);
      }
      else
      {
         // defer the processing as these are just state changes that can't
         // be visible to the user until the next client export call
         lt.addMarkEntry(new MarkEntry(new int[] {id}, MarkEntry.DESTROY));
      }
   }

   /**
    * Returns the column position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-COL
    * </pre>
    * 
    * @param    frame
    *           The frame for this request.
    * 
    * @return   The column position for the upper-left corner of the frame.
    */
   public static decimal getFrameCol(GenericFrame frame)
   {
      return new decimal(locate().client.getFrameCol(frame.getFrameId()));
   }

   /**
    * Returns the number of iterations in a down frame.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-DOWN
    * </pre>
    * 
    * @param    frame
    *           The frame for this request.
    * 
    * @return   The number of data iterations in the frame.
    */
   public static integer getFrameDown(GenericFrame frame)
   {
      return new integer(locate().client.getFrameDown(frame.getFrameId()));
   }
   
   /**
    * Returns the current logical line number in a down frame.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-LINE
    * </pre>
    * 
    * @param    frame
    *           The frame for this request.
    * @param    pendingUpDown
    *           The pending down value for this frame.
    * 
    * @return   1-based ordinal of the current data iteration of the frame.
    */
   public static integer getFrameLine(GenericFrame frame, int pendingUpDown)
   {
      return new integer(locate().client.getFrameLine(frame.getFrameId(),
                                                      pendingUpDown));
   }
   
   /**
    * Returns the row position of the upper-left corner of the frame
    * within its window (in coordinates that are relative to the containing
    * window).
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-ROW
    * </pre>
    * 
    * @param    frame
    *           The frame for this request.
    * 
    * @return   The row position for the upper-left corner of the frame.
    */
   public static decimal getFrameRow(GenericFrame frame)
   {
      return new decimal(locate().client.getFrameRow(frame.getFrameId()));
   }
   
   /**
    * Returns the number of screen lines available for frame display (does
    * not include the message or status lines).
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * SCREEN-LINES
    * </pre>
    *
    * @return   The number of available screen lines.
    */
   public static integer getScreenLines()
   {
      LogicalTerminal lt = locate();
      
      if (lt.screenLines  == null)
         lt.screenLines = new integer(lt.client.getScreenLines());
      
      return lt.screenLines;
   }
    
   /**
    * Returns the number of screen lines available for message display.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * MESSAGE-LINES
    * </pre>
    *
    * @return   The number of available lines in the message area.
    */
   public static integer getMessageLines()
   {
      return new integer(locate().client.getMessageLines());
   }
   
   /**
    * Returns all registered menus.
    * 
    * @return  collection of MenuWidget.
    */
   public static Collection<MenuWidget> getAllMenus()
   {
      List<MenuWidget> result = new ArrayList<>();
      
      for(MenuContainerWidget<?> m : locate().menuRegistry.values())
      {
         if (m instanceof MenuWidget)
         {
            result.add((MenuWidget) m);
         }
      }
      
      return result;
   }
  
   /**
    * Returns the terminal type. This usually corresponds to the value set as
    * the <code>TERM</code> environment variable on UNIX or Linux unless set
    * via {@link #setTerminalType(String)}. On Windows it depends upon
    * whether the terminal is in graphical or character mode. If it is in
    * character mode, the result depends on the monitor type.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>TERMINAL</b>
    * 
    * @return   Terminal type for in this session.
    */
   public static character getTerminalType()
   {
      LogicalTerminal logicalTerminal = locate();
      
      if (logicalTerminal.terminalType == null)
      {
         // cache the terminal type
         logicalTerminal.terminalType = logicalTerminal.client.getTerminalType();
      } 
      
      return new character(logicalTerminal.terminalType);
   }
  
   /**
    * Sets the terminal type.  This will cause the terminal to potentially
    * change its configuration if the new type is different from the old
    * type.
    * <p>
    * This processing also clears the input buffer and resets any key
    * mappings to their default values.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>TERMINAL</b>
    * 
    * @param   term
    *          Terminal type name.  If <code>null</code>, this resets the
    *          terminal without changing the type.
    */
   public static void setTerminalType(character term)
   {
      if (!term.isUnknown())
      {
         setTerminalType(term.getValue());
      }
   }

   /**
    * Sets the terminal type.  This will cause the terminal to potentially
    * change its configuration if the new type is different from the old
    * type.
    * <p>
    * This processing also clears the input buffer and resets any key
    * mappings to their default values.
    * <p>
    * This method corresponds to the Progress statement: <br>
    * <b>TERMINAL</b>
    * 
    * @param   term
    *          Terminal type name.  If <code>null</code>, this resets the
    *          terminal without changing the type.
    */
   public static void setTerminalType(String term)
   {
      // this part must be done on the server and the client
      Keyboard.resetKeyboard();
      
      LogicalTerminal logicalTerminal = locate();

      // clean the cached value
      logicalTerminal.terminalType = null;
      
      // the client handles the rest
      logicalTerminal.client.setTerminalType(term);
   }

   /**
    * Resets the terminal but does not change the current terminal type.
    * <p>
    * This method corresponds to the Progress statement:<br>
    * <b>TERMINAL = TERMINAL</b>
    */
   public static void resetTerminal()
   {
      setTerminalType((String) null);
   }

   /**
    * Returns the field name where the cursor is or was most recently.
    * <p>
    * This method corresponds to the Progress <code>FRAME-FIELD</code> built-in function.
    * 
    * @return   The field name.
    */
   public static character getFrameField()
   {
      return new character(locate().frameField);
   }
  
   /**
    * Returns the backing table that was used to define the widget where the cursor is or was
    * most recently.
    * <p>
    * This method corresponds to the Progress <code>FRAME-FILE</code> built-in function.
    * 
    * @return   The table name or empty string if there is no enabled widget or if that widget
    *           was not associated with a database field.
    */
   public static character getFrameFile()
   {
      return new character(locate().frameFile);
   }
  
   /**
    * Returns the backing database that was used to define the widget where the cursor is or was
    * most recently.
    * <p>
    * This method corresponds to the Progress <code>FRAME-DB</code> built-in function.
    * 
    * @return   The database name or empty string if there is no enabled widget or if that widget
    *           was not associated with a database field.
    */
   public static character getFrameDb()
   {
      return new character(locate().frameDb);
   }
  
   /**
    * Returns the frame name where the cursor is or was most recently.
    * <p>
    * This method corresponds to the Progress built-in function:
    * <p>
    * <pre>
    * FRAME-NAME
    * </pre>
    * 
    * @return   The frame name or "" if the cursor was not in a named frame or if the widget is
    *           not enabled for input.
    */
   public static character getFrameName()
   {
      return new character(locate().frameName);
   }
  
   /**
    * Returns the field index within the array of fields where the cursor is
    * or was the most recently.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>FRAME-INDEX </b>
    * 
    * @return  field index
    */
   public static integer getFrameIndex()
   {
      return locate().frameIndex;
   }
   
   /**
    * Get the frame instance from the registry.
    * 
    * @return   See above.
    */
   public static GenericFrame getFrame(int id)
   {
      return locate().frameRegistry.get(id);
   }
   
   /**
    * Checks whether a GO event is pending for this editing block.
    * <p>
    * This method corresponds to the Progress built-in function: <br>
    * <b>GO-PENDING </b>
    * 
    * @return  yes if an event has been applied that caused pending GO for this
    *          editing block.
    */
   public static logical isGoPending()
   {
      return new logical(locate().client.isGoPending());
   }

   /**
    * Getter for the <b>NUM-SELECTED-ROWS</b> attribute.
    * 
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for
    * 
    * @return  current value of NUM-SELECTED-ROWS attribute
    */
   public static int getNumSelectedRows(BrowseWidget bw)
   {
      return locate().client.getNumSelectedRows(bw.getId());
   }

   /**
    * Retrieve the 0-based row index of the <i>n</i>th selected row in the
    * given browse widget.
    * 
    * @param   bw
    *          Browse widget associated with the request.
    * @param   n
    *          0-based position of the row in the list of selected rows. <code>-1</code> for the
    *          last selected row.
    *
    * @return  0-based row index associated with the <i>n</i>th selected row in the
    *          browse, or -1 if the associated Browse widget could not be
    *          identified on the client.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If <code>selIndex</code> is invalid or if there are no "available" rows.
    */
   public static int getSelectedRowIndex(BrowseWidget bw, int n)
   {
      return locate().client.getSelectedRowIndex(bw.getId(), n);
   }

   /**
    * Check that the current row can be deleted with DELETE-RESULT-LIST-ENTRY: the row should be
    * existing and in the view.
    *
    * @param   bw
    *          Browse widget associated with the request.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If the current row is not existing or not in the view.
    */
   public static void checkDeleteEntryRowIndex(BrowseWidget bw)
   {
      locate().client.checkDeleteEntryRowIndex(bw.getId());
   }

   /**
    * Retrieve the row indexes of the selected rows in the given browse widget.
    *
    * @param   bw
    *          Browse widget associated with the request.
    *
    * @return  Array of 0-based row indexes of the selected rows in the browse.
    *
    * @throws  com.goldencode.p2j.util.ErrorConditionException
    *          If <code>selIndex</code> is invalid or if there are no "available" rows.
    */
   public static int[] getSelectedRowIndexes(BrowseWidget bw)
   {
      return locate().client.getSelectedRowIndexes(bw.getId());
   }

   /**
    * Implements the <b>SELECT-ALL()</b> <b>BROWSE</b> widget method.
    * 
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for
    */
   public static void selectAll(BrowseWidget bw)
   {
      locate().client.selectAll(bw.getId());
   }
  
   /**
    * Implements the <b>DESELECT-ROWS() </b> <b>BROWSE </b> widget method.
    * 
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for
    */
   public static void deselectRows(BrowseWidget bw)
   {
      locate().client.deselectRows(bw.getId());
   }

   /**
    * Notify the client that columns have been added to the browse.
    *
    * @param bw
    *        instance of <code>BrowseWidget</code> this request came for.
    * @param refreshData
    *        <code>true</code> to refresh row data, <code>false</code> to only update UI.
    */
   public static void notifyColumnsUpdated(BrowseWidget bw, boolean refreshData)
   {
      locate().client.notifyColumnsUpdated(bw.getId(), refreshData);
   }

   /**
    * Gets a value that indicates whether the browse contains one or more rows (including new
    * rows).
    *
    * @param bw
    *        instance of <code>BrowseWidget</code> this request came for.
    *
    * @return <code>true</code> if the browse contains one or more rows.
    */
   public static boolean hasRows(BrowseWidget bw)
   {
      return locate().client.hasRows(bw.getId());
   }

   /**
    * Scrolls a browse (if necessary) to bring the currently selected row into view. If the browse
    * supports multiple selections, then SCROLL-TO-CURRENT-ROW() brings the most recently selected
    * row into view.<p>
    * The position of the scrolled row is the first row in the browse viewport, unless the current
    * row is already visible. In this case, the current row remains in the original position.
    *
    * @param  bw
    *         Instance of <code>BrowseWidget</code> this request came for.
    *
    * @throws ErrorConditionException
    *         if a numbered error occurs.
    */
   public static void scrollToCurrentRow(BrowseWidget bw)
   {
      locate().client.scrollToCurrentRow(bw.getId());
   }

   /**
    * Scrolls a browse (if necessary) to bring the specified selected row into view.
    *
    * @param  bw
    *         Instance of <code>BrowseWidget</code> this request came for.
    * @param  selectedRowIndex
    *         Selected row within the browse. A numbered list of selected rows, starting at 1 is
    *         maintained. When the <code>SCROLL-TO-SELECTED-ROW(n)</code> method is encountered,
    *         the list is iterated to find the <i>n</i>th selected row.
    *
    * @throws ErrorConditionException
    *         if a numbered error occurs.
    */
   public static void scrollToSelectedRow(BrowseWidget bw, int selectedRowIndex)
   {
      locate().client.scrollToSelectedRow(bw.getId(), selectedRowIndex);
   }

   /**
    * Refresh the current browse row using the provided data.
    *
    * @param bw
    *        Instance of <code>BrowseWidget</code> this request came for.
    * @param row
    *        Container with the new data for the current row. If an {@link BrowseRow#data} element
    *        is <code>null</code>, corresponding cell shouldn't be refreshed.
    */
   public static void refreshCurrentRow(BrowseWidget bw, BrowseRow row)
   {
      locate().client.refreshCurrentRow(bw.getId(), row);
   }

   /**
    * Performs client-side actions after a new row has been added to the backing result set
    * using CREATE-RESULT-LIST-ENTRY. Updates row IDs in the current view.
    *
    * @param  bw
    *         Instance of <code>BrowseWidget</code> this request came for.
    * @param  newRowId
    *         0-based index of the newly created row in the browse model. Replaces the old
    *         negative row id of the current new row.
    */
   public static void rowEntryCreated(BrowseWidget bw, int newRowId)
   {
      locate().client.rowEntryCreated(bw.getId(), newRowId);
   }

   /**
    * Performs client-side actions after a row has been deleted from the backing result set
    * using DELETE-RESULT-LIST-ENTRY. Updates row IDs in the current view.
    *
    * @param  bw
    *         Instance of <code>BrowseWidget</code> this request came for.
    * @param  deletedRowId
    *         0-based index of the deleted row in the browse model.
    */
   public static void rowEntryDeleted(BrowseWidget bw, int deletedRowId)
   {
      locate().client.rowEntryDeleted(bw.getId(), deletedRowId);
   }

   /**
    * Refresh the browse cache and redisplay the browse.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for
    * @param   deleteCurrentRowMode
    *          <code>true</code> if this function is used for update on DELETE-CURRENT-ROW.
    *          <code>false</code> if this function is used for update on REFRESH.
    */
   public static void refresh(BrowseWidget bw, boolean deleteCurrentRowMode)
   {
      locate().client.refresh(bw.getId(), deleteCurrentRowMode);
   }

   /**
    * Refresh the browse after some of its rows were deleted.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    * @param   rowIndex
    *          0-based index of the deleted row. <code>-1</code> if all of the selected rows were
    *          deleted.
    */
   public static void refreshAfterDeletion(BrowseWidget bw, int rowIndex)
   {
      locate().client.refreshAfterDeletion(bw.getId(), rowIndex);
   }

   /**
    * Implements the SELECT-NEXT-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row after the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer. The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record after the last
    * selected row becomes the selected row.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    *
    * @return  <code>true</code> on success.
    */
   public static logical selectNextRow(BrowseWidget bw)
   {
      return locate().client.selectNextRow(bw.getId());
   }

   /**
    * Implements the SELECT-PREV-ROW() widget method.<p>
    * Deselects all currently selected rows in a browse and selects the row before the deselected
    * row.<p>
    * This method also repositions the query to the new row and copies the record into the
    * database buffer.The browse automatically scrolls to the selected row if it is out of
    * view.<p>
    * This method is intended for use with a browse that supports the selection of a single row at
    * a time (MULTIPLE attribute is set to FALSE). If more than one row is selected when you
    * execute this method, all of the selected rows are deselected and the record before the last
    * selected row becomes the selected row.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    *
    * @return  <code>true</code> on success.
    */
   public static logical selectPreviousRow(BrowseWidget bw)
   {
      return locate().client.selectPreviousRow(bw.getId());
   }

   /**
    * Implements the <b>SELECT-FOCUSED-ROW() </b> <b>BROWSE </b> widget
    * method.
    * 
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for
    */
   public static void selectFocusedRow(BrowseWidget bw)
   {
      locate().client.selectFocusedRow(bw.getId());
   }

   /**
    * Implements the <b>DESELECT-FOCUSED-ROW() </b> <b>BROWSE </b> widget
    * method.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for
    */
   public static void deselectFocusedRow(BrowseWidget bw)
   {
      locate().client.deselectFocusedRow(bw.getId());
   }

   /**
    * Return <code>true</code> if the row that has focus is selected.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    *
    * @return  <code>true</code> if the row that has focus is selected.
    */
   public static boolean isFocusedRowSelected(BrowseWidget bw)
   {
      return locate().client.isFocusedRowSelected(bw.getId());
   }

   /**
    * Implements the DESELECT-SELECTED-ROW() widget method.<p> Deselects the <i>n</i>th selected
    * row in a browse.<p>
    * Progress maintains a numbered list of selected rows, starting at 1. When the
    * DESELECT-SELECTED-ROW(n) method is encountered, Progress searches this list to find
    * the <i>n</i>th selected row.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    * @param   rowIndex
    *          1-based index of the row (in the list of selected rows) to deselect.
    *
    * @return  <code>true</code> if the row is successfully deselected.
    */
   public static boolean deselectSelectedRow(BrowseWidget bw, int rowIndex)
   {
      return locate().client.deselectSelectedRow(bw.getId(), rowIndex);
   }

   /**
    * Implements the SELECT-ROW() widget method.<p>
    * Selects the specified row if it is currently in the browse viewport. In a single-select
    * browse, the previously selected row is deselected. No rows are deselected in a
    * multiple-select browse.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    * @param   rowIndex
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if the row with the given index exist in the browse viewport.
    */
   public static boolean selectRow(BrowseWidget bw, int rowIndex)
   {
      return locate().client.selectRow(bw.getId(), rowIndex);
   }

   /**
    * Implements the <b>IS-ROW-SELECTED()</b> <b>BROWSE </b> widget method.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    * @param   rowIndex
    *          1-based row index in the browse viewport.
    *
    * @return  <code>true</code> if a specified row in the browse viewport is currently selected.
    */
   public static boolean isRowSelected(BrowseWidget bw, int rowIndex)
   {
      return locate().client.isRowSelected(bw.getId(), rowIndex);
   }

   /**
    * Implements the FOCUSED-ROW attribute getter.<p>
    * 
    * @param bw
    *        browse widget 
    *
    * @return the 1-based index of the focused row in the viewport.
    */
   public static int getFocusedRow(BrowseWidget bw)
   {
      return locate().client.getFocusedRow(bw.getId());
   }

   /**
    * Get number of rows in the browse viewport set either by DOWN or SIZE options.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    *
    * @return  number of rows in the browse viewport.
    */
   public static int getBrowsePageRowCount(BrowseWidget bw)
   {
      return locate().client.getBrowsePageRowCount(bw.getId());
   }

   /**
    * Get the number of entries visible in the browse.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    *
    * @return number of entries visible in the browse.
    */
   public static int getBrowseNumEntries(BrowseWidget bw)
   {
      return locate().client.getBrowseCacheSize(bw.getId());
   }

   /**
    * Execute pending refresh when REFRESHABLE is turned on. Includes pending query repositions
    * and reopenings.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    * @param   queryReopened
    *          <code>true</code> if there is a pending query reopening.
    * @param   repositionBeforeReopen
    *          <code>true</code> if the last pending reposition was <i>before</i> the last pending
    *          reopening.
    * @param   repositionedRow
    *          0-based index of the last row the query was repositioned to. <code>-1</code> if
    *          there is no repositioned row.
    */
   public static void executePendingRefresh(BrowseWidget bw,
                                            boolean queryReopened,
                                            boolean repositionBeforeReopen,
                                            int repositionedRow)
   {
      locate().client.executePendingRefresh(bw.getId(),
                                            queryReopened,
                                            repositionBeforeReopen,
                                            repositionedRow);
   }

   /**
    * Refresh scroll row in the browse and redisplay. Unless specific parameters are passed, the
    * current row becomes the scroll row.
    *
    * @param   bw
    *          Instance of <code>BrowseWidget</code> this request came for.
    * @param   leaveOnRowInView
    *          If the current row is in the viewport, do not update the scroll row.
    * @param   preserveViewportRow
    *          Try to preserve the current viewport row (i.e. the row at the same position in
    *          the viewport becomes the current row after refresh).
    * @param   repositioned
    *          If the browse is refreshed because the query was repositioned.
    * @param   repositionError
    *          For reposition mode - if an error has happened during reposition.
    * @param   targetRepositionRow
    *          Row used as the target for reposition. If it was deleted, reposition with fetching
    *          may end up on a different row.
    */
   public static void refreshScrollRow(BrowseWidget bw,
                                       boolean leaveOnRowInView,
                                       boolean preserveViewportRow,
                                       boolean repositioned,
                                       boolean repositionError,
                                       int targetRepositionRow)
   {
      locate().client.refreshScrollRow(bw.getId(),
                                       leaveOnRowInView,
                                       preserveViewportRow,
                                       repositioned,
                                       repositionError,
                                       targetRepositionRow);
   }

   /**
    * Notify the browse that the underlying query has been repositioned..
    *
    * @param   bw
    *          Instance of <code>BrowseWidget</code> this request came for.
    * @param   repositionError
    *          For reposition mode - if an error has happened during reposition.
    * @param   targetRepositionRow
    *          For reposition mode - row used as the target for reposition. If it was deleted,
    *          reposition with fetching may end up on a different row.
    */
   public static void queryRepositioned(BrowseWidget bw,
                                       final boolean repositionError,
                                       final int targetRepositionRow)
   {
      locate().client.queryRepositioned(bw.getId(), repositionError, targetRepositionRow);
   }

   /**
    * Notify browse widget at client side that query was closed.
    * 
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for
    */
   public static void queryClosed(BrowseWidget bw)
   {
      locate().client.queryClosed(bw.getId());
   }

   /**
    * Calculate DOWN value for the browse basing on the given row height.
    *
    * @param  bw
    *         Instance of <code>BrowseWidget</code> this request came for.
    * @param  rowHeightPixels
    *         Row height in pixels.
    *
    * @return calculated DOWN value.
    */
   public static int calcBrowseDown(BrowseWidget bw, int rowHeightPixels)
   {
      return locate().client.calcBrowseDown(bw.getId(), rowHeightPixels);
   }

   /**
    * Push "intermediate" rows (i.e. the rows that were already fetched in the process of getting
    * new rows by
    * {@link BrowseWidget#getRows(int, int, boolean, int, ScreenBuffer[], RowDisplayStrategy)})
    * to the client, so these rows can be drawn if a pause occurs in ROW-DISPLAY trigger.
    *
    * @param  bw
    *         Instance of <code>BrowseWidget</code> this request came for.
    * @param  rows
    *         Set of rows to push.
    */
   public static void pushIntermediateRows(BrowseWidget bw, BrowseRow[] rows)
   {
      locate().client.pushIntermediateRows(bw.getId(), rows);
   }
   
   /**
    * Returns <code>true</code> if the current row is a new row and is not present in the
    * database.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    *
    * @return  <code>true</code> if the current row is a new row and <code>false</code> otherwise.
    */
   public static boolean isNewRow(BrowseWidget bw)
   {
      return locate().client.isNewRow(bw.getId());
   }

   /**
    * Returns <code>true</code> if the current row is modified.
    *
    * @param   bw
    *          instance of <code>BrowseWidget</code> this request came for.
    *
    * @return  <code>true</code> if the current row is modified.
    */
   public static boolean isCurrentRowModified(BrowseWidget bw)
   {
      return locate().client.isCurrentRowModified(bw.getId());
   }

   /**
    * Get value of <code>column.Y</code> attribute for the columns of this browse: vertical offset of the
    * current row from the top of the browse, in pixels for GUI or in chars for ChUI. The value is the same
    * for all columns in the browse.
    *
    *  @param   bw
    *           Instance of <code>BrowseWidget</code> this request came for.
    *
    * @return value of <code>column.Y</code> attribute for the columns of this browse.
    */
   public static int getBrowseColumnY(BrowseWidget bw)
   {
      return locate().client.getBrowseColumnY(bw.getId());
   }

   /**
    * Inserts a blank row in an updateable browse before or after the last selected row.
    * The blank row is a placeholder for a new record to be added through the browse.
    *
    * @param   bw
    *          Instance of <code>BrowseWidget</code> this request came for.
    * @param   after
    *          <code>true</code> if the row is added after the last selected row.
    *          <code>false</code> if the row is added before the last selected row.
    *
    * @return  <code>true</code> if the row was successfully added.
    */
   public static boolean insertRow(BrowseWidget bw, final boolean after)
   {
      return locate().client.insertRow(bw.getId(), after);
   }

   /**
    * Implements the client side of the MOVE-COLUMN() widget method.<p>
    * Repositions a column in a browse widget.
    *
    * @param  bw
    *         Instance of <code>BrowseWidget</code> this request came for.
    * @param  srcIndex
    *         0-based position of the column to be moved.
    * @param  destIndex
    *         0-based position to which the column is moved.
    */
   public static void moveColumnClient(BrowseWidget bw, int srcIndex, int destIndex)
   {
      locate().client.moveColumnClient(bw.getId(), srcIndex, destIndex);
   }

   /**
    * Run client-side part of the code which should be executed after an enhanced configuration is applied to
    * the browse.
    *
    * @param  bw
    *         Instance of <code>BrowseWidget</code> this request came for.
    */
   public static void applyEnhancedConfigClient(BrowseWidget bw)
   {
      locate().client.applyEnhancedConfigClient(bw.getId());
   }

   /**
    * Get enhanced browse configurations stored on the client side.
    *
    * @param  keys
    *         Array of the keys of the target configurations to get.
    *
    * @return array containing the target configurations. It has the same size as the array of
    *         {@code keys}. If there is no configuration found for a key, the corresponding array
    *         element is {@code null}.
    */
   public static EnhancedBrowseConfig[] getClientEnhancedConfigs(EhBrowseConfigKey[] keys)
   {
      return locate().client.getClientEnhancedConfigs(keys);
   }

   /**
    * Implements the server-side MOVE-COLUMN() widget method.<p>
    * Repositions a column in the browse widget.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    * @param  src
    *         0-based position of the column to be moved (column ordinal).
    * @param  dest
    *         0-based position to which the column is moved (column ordinal).
    */
   public static void moveColumnServer(int browseId, int src, int dest)
   {
      try
      {
         getBrowseById(browseId).moveColumn(new int64(src + 1), new int64(dest + 1), false);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
      }
   }

   /**
    * Notify server that browse layout has been performed on the client side.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    */
   public static void notifyBrowseLayoutPerformed(int browseId)
   {
      try
      {
         getBrowseById(browseId).notifyBrowseLayoutPerformed();
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
      }
   }

   /**
    * Set dynamic sorting for the browse.
    *
    * @param browseId
    *        ID of the browse which should handle the request.
    * @param columnOrdinals
    *        Array of the ordinals of the sort columns. Inside the array ordinals are positioned
    *        from the column which defines the most coarse sorting to the most granular.
    * @param ascending
    *        Array of the sorting directions of the sort columns. <code>true</code> for ascending,
    *        <code>false</code> for descending. Inside the array directions are positioned
    *        from the column which defines the most coarse sorting to the most granular.
    */
   public static void setDynamicSorting(int browseId, int[] columnOrdinals, boolean[] ascending)
   {
      try
      {
         getBrowseById(browseId).setDynamicSorting(columnOrdinals, ascending);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
      }
   }

   /**
    * Set dynamic filtering for the browse.
    *
    * @param browseId
    *        ID of the browse which should handle the request.
    * @param columnOrdinals
    *        Array of the ordinals of the filtered columns.
    * @param values
    *        Array of the filter values for the filtered columns..
    */
   public static void setDynamicFiltering(int browseId,
                                          int[] columnOrdinals,
                                          BaseDataType[] values)
   {
      try
      {
         getBrowseById(browseId).setDynamicFiltering(columnOrdinals, values);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
      }
   }
   
   /**
    * Receive a {@code MessageBlaster} notification back from client. This is caused by the
    * subscription of the {@code MessageBlaster} to a OS-level event. The message is dispatched
    * to the appropriate {@code MessageBlaster} instance, if such exists. Otherwise the message
    * is discarded.
    *
    * @param   id
    *          The id of the {@code MessageBlaster} instance. 
    * @param   msg
    *          The message id. Should be one already added to {@code msgList}. 
    * @param   wParam
    *          Message parameter 1.
    * @param   lParam
    *          Message parameter 2.
    * @param   lplRetVal
    *          Message parameter / return value.
    */
   public static void messageBlasterEvent(long id, int msg, int wParam, int lParam, int lplRetVal)
   {
      MsgBlaster mb = MsgBlaster.getInstanceById(id);
      if (mb != null) 
      {
         mb.incomingMessage(msg, wParam, lParam, lplRetVal);
      }
   }
   
   /**
    * Retrieve all local paths whitelisted as resource-base for html web browser.
    * 
    * @return   The whitelist of resource-base.
    */
   public static Map<String, URL> readHtmlBrowserWhitelist()
   {
      return HtmlBrowserWhitelist.readWhitelist();
   }

   /**
    * Returns the client metrics for current session.
    *
    * @return  Client metrics instance.
    */
   public static ClientMetrics getClientMetrics()
   {
      return new ClientMetrics(locate().clientMetrics);
   }

   /**
    * Sets the client metrics in the current session.
    *
    * @param   metrics
    *          The metrics instance.
    */
   public static void setClientMetrics(ClientMetrics metrics)
   {
      locate().clientMetrics = metrics;
   }

   /**
    * Each frame tracks whether there has been any iteration or explicit
    * requests for down or up that are conditional.  Any pending up or
    * down must be honored before certain client calls such as
    * <code>view</code>.  This method provides a facility for a frame to
    * store a "one-shot" update that will be synchronized with the client
    * at the next message.  This must only be done JUST BEFORE a subsequent
    * call to one of the FRAME-SPECIFIC client calls that require honoring
    * this conditional value. This requirement is necessary to avoid
    * tracking a list of all frame IDs and associated pending up/down values.
    * Since this requirement is highly coupled to the FRAME-SPECIFIC calls,
    * this should not be a problem.
    *
    * @param    upDown
    *           -1 for up or 1 for down.
    * @param    frameId
    *           The frame for which the update must be made.  This is really
    *           just here for safety in case the state sync is triggered
    *           before we expect.
    */
   public static void setConditionalUpDown(int upDown, int frameId)
   {
      LogicalTerminal lt = locate();
      lt.state.conditionalUpDown  = upDown;
      lt.state.conditionalFrameId = frameId;
   }
   
   /**
    * Enable hard format check for specified frame.
    * 
    * @param    id
    */
   public static void setHardFormatCheck(int id)
   {
      locate().state.formatCheckId = id;
   }
   
   /**
    * Calls the client side to implement SCROLL-TO-ITEM() method for selection-list widget.
    *
    * @param    slId
    *           Widget ID of the selection list to make scrolling.
    * @param    index
    *           The 1-based index of an item to scroll.
    *
    * @return   <code>TRUE</code> in success, <code>FALSE</code> otherwise.
    */
   public static boolean scrollToItem(int slId, int index)
   {
      // client side will work with 0-based index
      return locate().client.scrollToItem(slId, index - 1);
   }
   
   /**
    * Process a notification that a new block scope is about to be entered.
    * <p>
    * In the Logical Terminal, a new top level block scope causes creation of
    * a new scope in the dictionary of registered triggers.
    * <p>
    * If some other class of objects requires scoping, it can be added here.
    */
   @Override
   public void scopeStart()
   {
      // scopeable registration is done in different places - if changes are made, ensure that scopeable 
      // support is registered properly.
      
      // register at each scope, it is important to instantiate the scope
      // processor BEFORE the state.interactions counter is reset below
      tm.registerFinalizable(new ScopeProcessor(), false);

      // only if a block has the ENDKEY property, does the
      // interactions value get updated normally; all EDITING blocks and
      // any DO blocks/loops that don't have an explicit ON ENDKEY phrase
      // do not have the ENDKEY property and thus will bypass this code
      if (tm.hasEndkeyProp())
      {
         // save the current interactions value on our stack (to restore
         // when this new block is popped)
         blocks.push(state.interactions);
         
         state.interactions = 0;
      }
      
      // add a new scope for each top-level block
      if (tm.isTopLevelBlock())
      {
         // triggers get notifications at all top level blocks
         TriggerManager.scopeStart(scope);

         definitions.markPartition();
         
         // since the first scopeStart() for an external procedure is called
         // after the initializer is complete AND since the frame creation
         // is done in the business logic initializer AND since the frame
         // creation causes the deferred screen defs to be created, we have
         // a "2 level effect" where we have an extra cache to temporarily
         // hold the screen defs until this scopeStart() could be called
         if (level2definitions.size() > 0)
         {
            // add these into our main queue
            definitions.addAll(level2definitions);
            level2definitions.clear();
         }
         
         descriptions.markPartition();
         
         // since the first scopeStart() for an external procedure is called after the initializer
         // is complete AND since the frame creation is done in the business logic initializer AND
         // since the frame creation causes the deferred screen defs to be created, we have a 
         // "2 level effect" where we have an extra cache to temporarily hold the screen defs
         // until this scopeStart() could be called
         if (level2descriptions.size() > 0)
         {
            // add these into our main queue
            descriptions.addAll(level2descriptions);
            level2descriptions.clear();
         }
         // pendingMenus contains only menus that possible need deregistration when their external
         // procedure instantiation fails. If we reached this point, the set is not useful anymore
         // and must be discarded
         pendingMenus.clear();
         
         // track external procedure nesting
         if (tm.isExternalBlock())
         {
            locate().nesting++;
         }
      }
   }
   
   /**
    * Process a notification that a scope is about to be exited.
    * <p>
    * In the Logical Terminal, the current scope in the dictionary of
    * registered triggers is destroyed, but only when exiting a top level
    * block.
    */
   @Override
   public void scopeFinished()
   {
      // remove the scope for each top-level block
      if (tm.isTopLevelBlock())
      {
         definitions.removePartition();
         descriptions.removePartition();
         
         if (tm.isExternalBlock())
         {
            locate().nesting--;
         }
         
         if (TriggerManager.scopeFinished(scope))
         {
            // invalidate the cache
            cachedEventList = null;
         }
      }
   }

   /**
    * Notification that the instantiating procedure where the scope was opened was deleted.
    * <p>
    * This implementation is a no-op.
    */
   @Override
   public void scopeDeleted()
   {
      // no-op
   }
   
   /**
    * Get the {@link ScopeId} for the instance.
    * 
    * @return   {@link ScopeId#LOGICAL_TERMINAL}.
    */
   @Override
   public ScopeId getScopeId()
   {
      return ScopeId.LOGICAL_TERMINAL;
   }
   
   /**
    * Called by the protocol driver to obtain the latest state that should
    * be sent to the other side of the connection.  This is called just 
    * before a "real" message is sent and any changes provided will be
    * piggybacked into that message.
    * <p>
    * To enable incremental changes, it is guaranteed that this method will
    * be called every time any message is sent to the other side of the
    * connection.  In addition, this will only be called once per message.
    * <p>
    * In a system in which control flow is "sent" to the other side of the 
    * connection (a type of "conversation" or single-threaded synchronous
    * mode), it is further guaranteed that any changes that are made at the
    * other side after this is called will result in a call to this side's
    * {@link #applyChanges} method before this method is ever called again
    * (when control flow returns to this side of the connection). This allows
    * one to ensure that on the call to this method, any "dirty" flags can be
    * safely cleared and the state of this side will be updated with
    * the results of the other side's changes before this method must ever
    * again check to see if there are changes to send.
    *
    * @return   The payload of state changes to be sent to the other side or
    *           <code>null</code> if no changes need to be sent.
    */
   @Override
   public Externalizable getChanges()
   {      
      ServerState sstate = new ServerState(state);

      state.reset();
      
      // set the new numeric format to be set at the client-side
      sstate.numericFormat = newClientNumericFormat;
      
      // reset the cached values
      newClientNumericFormat = null;
      currentClientNumericFormat = null;
      
      // copy current external call nesting
      sstate.nesting = nesting;
      
      // clear dirty flags
      ColorTable.clearDirtyFlag(sstate.environments);
      
      sstate.deadFrames = null;
      
      if (!deadFrameIds.isEmpty())
      {
         List<Integer> deadFrames = new ArrayList<>(3);
         
         // copy all frames in dead procedures
         for (Integer frameId : deadFrameIds)
         {
            GenericFrame frame = frameRegistry.get(frameId);
            if (frame == null)
            {
               continue;
            }
            
            FrameWidget fw = frame.getFrameWidget();

            if (!fw._dynamic() && !fw.instantiatingProcedure()._isValid() && !frame.isInsideSetup())
            {
               deadFrames.add(frame.getFrameWidget().getId());
            }
         }
         
         deadFrameIds.clear();
         
         if (!deadFrames.isEmpty())
         {
            sstate.deadFrames = Utils.integerCollectionToPrimitive(deadFrames);
         }
      }
      
      sstate.suppressWarnings = tm.errHlp.isSuppressWarnings();

      // send and reset the modified widget attributes queue
      sstate.widgetAttrIds = widgetAttrIds.isEmpty() ? null : widgetAttrIds.toArray();
      sstate.widgetAttrValues = widgetAttrValues.isEmpty() ? null : widgetAttrValues.toArray();

      // com.carrotsearch.hppc.IntArrayList just cleans an existing buffer with Arrays.fill
      widgetAttrIds.clear();

      // ArrayList is more expensive, it clears by iterating every element of the backing buffer
      widgetAttrValues = new ArrayList<>();
      
      return sstate;
   }

   /**
    * Obtains the pressed keyboard key codes. 
    *
    * @return see above
    */
   public static int[] getKeyboardState() 
   {
      return locate().keyboardState;
   }
   
   /**
    * Set keyboard state.
    * 
    * @param keyboardState
    *        array of pressed key codes
    */
   public static void setKeyboardState(final int[] keyboardState)
   {
      locate().keyboardState = keyboardState;
   }

  /**
    * Called by the protocol driver upon reading a message that contains state
    * changes that need to be applied to this side of the connection. This is
    * called just after a new message is read from the connection and only if
    * that message had piggybacked state changes attached. This is guaranteed
    * to be called before the containing message is enqueued for further
    * "real" processing on this node.
    * 
    * @param   changes
    *          The payload of state changes that must be locally applied.
    */
   @Override
   public void applyChanges(Externalizable changes)
   {
      if (changes instanceof ClientState)
      {
         ClientState cs = (ClientState) changes;

         keyboardState = cs.keyboardState;
         
         if (cs.deregister != null)
         {
            // scopeable registration must be done in any location where a frame state (requiring scope 
            // support) is being 'touched' - this includes frame de-registration, even if the current block 
            // is not explicitly using frames

            tm.registerBlockScopeable(this);

            for (int i : cs.deregister)
            {
               deregisterFrameInt(i);
            }
         }
         
         if (cs.downIDs != null)
         {
            for (int i = 0; i < cs.downIDs.length; i++)
            {
               refreshFrameDown(cs.downIDs[i], cs.down[i]);
            }
         }
         
         if (cs.wrapped != null && scopeTrack != null)
         {
            tm.registerBlockScopeable(this);
            
            for (int id : cs.wrapped)
            {
               scopeTrack.frameWrapped(id);
            }
         }
         
         if (cs.hadPause)
         {
            tm.hadPause();
         }

         if (cs.functionKey != null && cs.functionKey != KeyInput.CHAR_UNDEFINED)
         {
            this.functionKey = cs.functionKey;
         }
         /*else 
         {
            this.functionKey = KeyInput.CHAR_UNDEFINED;
         }*/
         
         this.eventX = cs.eventX;
         this.eventY = cs.eventY;
         
         if (cs.virtualLastKey != null)
         {
            KeyReader.setLastKey(cs.virtualLastKey);
         }
         
         millis = 0;
      
         if (cs.configUpdates != null)
         {
            ServerConfigManager mgr = ConfigManager.getInstance();
            
            for (WidgetConfigUpdates cfgUpdate : cs.configUpdates)
            {
               GenericWidget<?> widget = getWidgetForId(cfgUpdate.widgetId);

               mgr.applyConfigUpdates(widget, cfgUpdate);
               
               if (widget instanceof FrameWidget && (!widget._isVisible() || widget._isHidden()))
               {
                  // attempt frame delete
                  widget.getFrame().frameDelete();
               }
            }
         }
         
         if (cs.activeWindowId > 0)
         {
            if (cs.activeWindowId == WidgetId._DEFAULT_WINDOW_ID)
            {
               activeWindow = defaultWindow;
            }
            else
            {
               activeWindow = windowRegistry.get(cs.activeWindowId);
            }
         }
      }
   }
   
   /**
    * Add instance of <code>MarkEntry</code> to list.
    * <p>
    * Note that if {@link MarkEntry} which should be added to the list is
    * {@link MarkEntry#CLOSE_SCOPE}, then we check if list still contains
    * {@link MarkEntry#OPEN_SCOPE} for this frame. If this is so, then all
    * entries which belong to this frame at current scope nesting can be
    * safely removed since there were no client calls between opening and
    * closing scope and frame state is not changed. This optimisation becomes
    * vital in cases when frame scope is opened/closed many times in loop
    * before triggering actual frame call. Such a scenario may result to
    * generation of very large list of operations which are effectively no-op
    * but consume significant amount of network traffic/memory and can even
    * crash client.
    * 
    * @param   me
    *          Instance of MarkEntry to add into state list of frame marking
    *          entries.
    */
   void addMarkEntry(MarkEntry me)
   {
      if (state.markList == null)
      {
         state.markList = new ArrayList<>();
      }
      else if (me.mode == MarkEntry.CLOSE_SCOPE)
      {
         if (cleanupList(me.frames[0]))
            return;
      } 
      else if (state.markList.size() > 0 && 
               (me.mode == MarkEntry.HIDE || 
                me.mode == MarkEntry.HIDE_NO_PAUSE))
      {
         if (checkDuplicate(me))
            return;
      }
      
      state.markList.add(me);
   }

   /**
    * Check if last entry in list is identical to specified one.
    * 
    * @param   me
    *          Entry to compare with.
    *          
    * @return  <code>true</code> if entries are equal.
    */
   private boolean checkDuplicate(MarkEntry me)
   {
      MarkEntry entry = state.markList.get(state.markList.size() - 1);

      if (entry.mode != me.mode || 
          entry.frames == null  || 
          me.frames == null     || 
          entry.frames.length != me.frames.length)
      {
         return false;
      }
      
      for (int i = 0; i < entry.frames.length; i++)
      {
         if (entry.frames[i] != me.frames[i])
            return false;
      }
      
      return true;
   }

   /**
    * Check for presence of {@link MarkEntry#OPEN_SCOPE} entry in the list and
    * if found then cleanup list of {@link MarkEntry} instances from
    * operations related to specified frame at current level of scope nesting.
    * 
    * @param   frameId
    *          Frame ID.
    *          
    * @return  <code>true</code> if list was cleaned up.
    */
   private boolean cleanupList(int frameId)
   {
      boolean remove = false;
      
      for (MarkEntry entry : state.markList)
      {
         if (entry.mode == MarkEntry.OPEN_SCOPE &&
             entry.frames != null               && 
             entry.frames.length == 1           && 
             entry.frames[0] == frameId)
         {
            // we found than OPEN_SCOPE is still in the list and this does
            // mean that frame operations didn't reach client yet so we can 
            // safely remove all entries related to this frame
            remove = true;
            break;
         }
      }
      
      if (remove)
      {
         for (int i = state.markList.size() - 1; i >= 0; i--)
         {
            MarkEntry entry = state.markList.get(i);
            
            if (entry.frames == null || entry.frames.length != 1)
               continue;

            if (entry.frames[0] != frameId || entry.mode == MarkEntry.DESTROY)
               continue;
            
            state.markList.remove(i);

            if (entry.mode == MarkEntry.OPEN_SCOPE)
               break;
         }
      }
      return remove;
   }
   
   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and close/
    * cleanup resources for the pseudo-terminal associated with
    * <code>oldId</code> (unless <code>oldId</code> is -1 in which case 
    * nothing needs to be closed).
    *
    * @param   oldId
    *          The previously redirected stream for which the pseudo-terminal
    *          is to be closed or -1 for no close/cleanup action.
    * @param   newId
    *          The stream for which a pseudo-terminal is to be setup for
    *          redirection or -1 to remove redirections and activate the
    *          interactive terminal.
    */
   public static void redirectOutput(int oldId, int newId)
   {
      if (newId == oldId)
      {
         return;
      }

      locate().client.redirectOutput(oldId, newId);
   }
   
   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and return
    * the id of the current output destination.  This may be another
    * pseudo-terminal or it may be -1 if the interactive terminal is in 
    * use.
    * <p>
    * If redirection has not yet been established for the given stream,
    *
    * @param    newId
    *           The stream for which a pseudo-terminal is to be setup for
    *           redirection or -1 to remove redirections and activate the
    *           interactive terminal.
    * @param    clean
    *           <code>true</code> if this stream will need deferred cleanup.
    *
    * @return   The id of the previous terminal or -1 if the previous 
    *           terminal was interactive.   
    */
   public static int switchOutput(int newId, boolean clean)
   {
      return switchOutput(newId, clean, false);
   }
   
   /**
    * Redirect terminal output to the stream represented by <code>newId</code>
    * (or to the interactive terminal if <code>newId</code> is -1) and return
    * the id of the current output destination.  This may be another
    * pseudo-terminal or it may be -1 if the interactive terminal is in 
    * use.
    * <p>
    * If redirection has not yet been established for the given stream,
    *
    * @param    newId
    *           The stream for which a pseudo-terminal is to be setup for
    *           redirection or -1 to remove redirections and activate the
    *           interactive terminal.
    * @param    clean
    *           <code>true</code> if this stream will need deferred cleanup.
    * @param    endOut
    *           <code>true</code> to force any pending PUT or EXPORT output
    *           to be ended for the given stream (so long as that stream is
    *           not the terminal).
    *
    * @return   The id of the previous terminal or -1 if the previous 
    *           terminal was interactive.   
    */
   public static int switchOutput(int newId, boolean clean, boolean endOut)
   {
      return locate().client.switchOutput(newId, clean, endOut);
   }
   
   /**
    * Called when a stream is closed to ensure that any terminal resources
    * associated with that id are cleaned up.  While this will be called
    * for any output stream close, actual cleanup processing is only 
    * triggered when the given stream id was used at least once via the
    * {@link #switchOutput} interface.
    * <p>
    * This should never be called when the redirection is currently active.
    *
    * @param    id
    *           The stream needing cleanup.
    */
   public static void cleanupPseudoTerminal(int id)
   {
      locate().client.cleanupPseudoTerminal(id);
   }
   
   /**
    * Sets the relogin flag.
    *
    * @param    relogin
    *           new state of the flag
   */
   public static void setRelogin(boolean relogin)
   {
      locate().relogin = relogin;
   }
   
   /**
    * Perform pause before end of application.
    *
    * @param   isQuit
    *          Set to <code>true</code> if QUIT statement is processed.
    */
   public void pauseBeforeEnd(boolean isQuit)
   {
      if (!isQuit && !locate().relogin)
      {
         client.pauseBeforeEnd();
      }
   }
   
   //---------------------------------------------------------------------------
   // Browse-related entries.
   //---------------------------------------------------------------------------

   /**
    * Get configurations of editors (fill-ins, toggle-boxes or combo-boxes) used for in-browse
    * editing.
    *
    * @param    browseId
    *           ID of the browse which should handle the request.
    *
    * @return   Array of configurations of editor widgets (fill-ins, toggle-boxes or combo-boxes)
    *           which are used for in-browse editing. An array element is configuration of the
    *           editor widget for the corresponding editable column or <code>null</code> if the
    *           column is not editable. Return value is <code>null</code> if the browse is
    *           read-only.
    */
   public static WidgetConfig[] getEditorsConfigs(int browseId)
   {
      WidgetConfig[] arrRes = null;
      try
      {
         arrRes = getBrowseById(browseId).getEditorsConfigs();
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         arrRes = new WidgetConfig[0];
      }
      return arrRes;
   }

   /**
    * Get the number of rows iterated by the query (for preselect queries - matches result size,
    * for adaptive queries - matches the size of the part of the result set the query has already
    * iterated over).
    *
    * @param    browseId
    *           ID of the browse which should handle the request.
    *
    * @return  number of iterated rows in the result set.
    */
   public static int getRowCount(int browseId)
   {
      int countRes = 0;
      try
      {
         countRes = getBrowseById(browseId).getRowCount();
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         countRes = 0;
      }
      return countRes;
   }

   /**
    * Get number of available rows starting from the given one.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   startRow
    *          Initial row to start count rows from. 
    * 
    * @return  Number of available rows.
    */
   public static int getRowCount(int browseId, int startRow)
   {
      int countRes = 0;
      try
      {
         countRes = getBrowseById(browseId).getRowCount(startRow);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         countRes = 0;
      }
      return countRes;
   }

   /**
    * Retrieves several subsequent rows from the database.
    *
    * @param    browseId
    *           ID of the browse which should handle the request.
    * @param    startRowIndex
    *           Index of the first row to retrieve (0-based).
    * @param    count
    *           Maximum count of rows to return. <code>-1</code> for no limit.
    * @param    forward
    *           <code>true</code> if rows are retrieved in the forward direction,
    *           <code>false</code> if rows are retrieved in the backward direction.
    * @param    rowDisplayTriggerId
    *           Trigger ID of the ROW-DISPLAY trigger. <code>-1</code> if there is no ROW-DISPLAY
    *           trigger for this browse.
    * @param    sb
    *           Screen buffer for the ROW-DISPLAY trigger. <code>null</code> if there is no
    *           ROW-DISPLAY trigger for this browse.
    * @param    displayStrategy
    *           Specifies type of action for which rows are fetched and defines how rows are
    *           iterated / displayed.
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           Note that if rows are fetched in the backward direction, the resulting rows are
    *           returned in the straight order starting from the row with the least row number.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   public static BrowseRow[] getRows(int browseId,
                                     int startRowIndex,
                                     int count,
                                     boolean forward,
                                     int rowDisplayTriggerId,
                                     ScreenBuffer[] sb,
                                     RowDisplayStrategy displayStrategy)
   {
      BrowseRow[] arrRes = null;
      try
      {
         arrRes = getBrowseById(browseId).getRows(startRowIndex,
                                                  count,
                                                  forward,
                                                  rowDisplayTriggerId,
                                                  sb,
                                                  displayStrategy);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         arrRes = new BrowseRow[0];
      }
      return arrRes;
   }

   /**
    * This function is used when search by key is performed in a browse. Retrieves the set of
    * rows which should be in view after search is completed. There are complex rules which
    * determine the top row of the view depending on how the <i>current</i> row is positioned
    * relatively to the current view and how the <i>found</i> row is positioned relatively to the
    * current view and to the current row.
    *
    * @param    browseId
    *           ID of the browse which should handle the request.
    * @param    searchChar
    *           Character to be searched for. Search is performed using only the first character
    *           (in the visual data representation) in the current column. Specifying a character
    *           puts this function in the search mode. This parameter is <code>null</code> for
    *           non-search mode.
    * @param    searchColumn
    *           For search mode: 0-based index of the column to be searched.
    * @param    viewportTopRow
    *           For search mode: 0-based index (in the result set) of the top row in the current
    *           view.
    * @param    viewportBottomRow
    *           For search mode: 0-based index (in the result set) of the bottom row in the
    *           current view.
    * @param    rowDisplayTriggerId
    *           Trigger ID of the ROW-DISPLAY trigger. <code>-1</code> if there is no ROW-DISPLAY
    *           trigger for this browse.
    * @param    sb
    *           Screen buffer for the ROW-DISPLAY trigger. <code>null</code> if there is no
    *           ROW-DISPLAY trigger for this browse.
    *
    * @return   Array of rows. Each row is array of cell values, but the first element of the row
    *           array is 0-based index of the row in the browse model.
    *           If no query is specified or the query is closed, <code>null</code> is returned.
    */
   public static BrowseRow[] getRows(int browseId,
                                     char searchChar,
                                     int searchColumn,
                                     int viewportTopRow,
                                     int viewportBottomRow,
                                     int rowDisplayTriggerId,
                                     ScreenBuffer[] sb)
   {
      BrowseRow[] arrRes = null;
      try
      {
         arrRes = getBrowseById(browseId).getRows(searchChar,
                                                  searchColumn,
                                                  viewportTopRow,
                                                  viewportBottomRow,
                                                  rowDisplayTriggerId,
                                                  sb);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         arrRes = new BrowseRow[0];
      }
      return arrRes;
   }
   
   /**
    * Updates selected cells in the given row.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          Index of the row to update.
    * @param   cellIndices
    *          Cells for update indexes.
    * @param   cellValues
    *          Cells for update new values.
    * @param   onlyNotifyRowReleased
    *          Only notify server that the row has been released. Data is not updated.
    * 
    * @throws  DataValidationException
    *          New values don't passed database validation rules.
    */
   public static void updateRow(int            browseId,
                                int            rowIndex,
                                int[]          cellIndices, 
                                BaseDataType[] cellValues,
                                boolean        onlyNotifyRowReleased)
   throws DataValidationException
   {
      try
      {
         getBrowseById(browseId).updateRow(rowIndex, cellIndices, cellValues,
                                           onlyNotifyRowReleased);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
      }
   }

   /**
    * Inform server side browse about current row in the browse.
    * 
    * @param   browseId
    *          ID of the browse which should handle the request.
    * @param   rowIndex
    *          Index of the current row.
    *          
    * @return  New current row.
    */
   public static int setCurrentRow(int browseId, int rowIndex)
   {
      int rowRes = 0;
      try
      {
         rowRes = getBrowseById(browseId).setCurrentRow(rowIndex, true);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         rowRes = 0;
      }
      return rowRes;
   }

   /**
    * Get current row in the query.
    *
    * @param   browseId
    *          ID of the browse which should handle the request.
    *          
    * @return  The current row index.
    */
   public static int getCurrentRow(int browseId)
   {
      int rowRes = 0;
      try
      {
         rowRes = getBrowseById(browseId).getCurrentRow();
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         rowRes = 0;
      }
      return rowRes;
   }

   /**
    * Determines if the specified browse row is "available", i.e. it is not deleted.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    * @param  rowIndex
    *         0-base index of the target row in the browse model.
    *
    * @return <code>true</code> if the row is "available", i.e. it is not deleted.
    */
   public static boolean isRowAvailable(int browseId, int rowIndex)
   {
      boolean bRes = false;
      try
      {
         bRes = getBrowseById(browseId).isRowAvailable(rowIndex);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         bRes = false;
      }
      return bRes;
   }

   /**
    * Run browse report.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    * @param  outputFileName
    *         Name of the output file.
    * @param  outputFormat
    *         Format of the output format.
    * @param  sb
    *         The screen buffer array data.
    *
    * @return <code>true</code> on success.
    */
   public static boolean runReport(int browseId,
                                   String outputFileName,
                                   MediaType outputFormat, 
                                   ScreenBuffer[] sb)
   {
      boolean bRes = false;
      try
      {
         LogicalTerminal.handleScreenBuffers(sb);
         bRes = getBrowseById(browseId).runReport(outputFileName, outputFormat);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         bRes = false;
      }
      return bRes;
   }

   /**
    * Get save targets for an enhanced browse configuration, which are available for the current
    * user.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    *
    * @return see above.
    */
   public static EnhancedBrowseSaveTarget[] getEnhancedBrowseSaveTargets(int browseId)
   {
      EnhancedBrowseSaveTarget[] arrRes = null;
      try
      {
         arrRes = getBrowseById(browseId).getEnhancedBrowseSaveTargets();
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         arrRes = new EnhancedBrowseSaveTarget[0];
      }
      return arrRes;
   }

   /**
    * For enhanced filtering: get data types of the fields mapped using <code>DB-FILTER-FIELD-MAPPING</code>.
    * They are used in column filters. E.g. if a column has <code>DB-FILTER-FIELD-MAPPING</code> set to
    * "some-table.some-date-field", data type from this column would be "date".
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    *
    * @return array of data types, each array element corresponds a browse column. If a column doesn't have
    *         <code>DB-FILTER-FIELD-MAPPING</code> set, corresponding array element is <code>null</code>.
    */
   public static String[] getColumnFiltersMappedDataTypes(int browseId)
   {
      String[] arrRes;
      try
      {
         arrRes = getBrowseById(browseId).getColumnFiltersMappedDataTypes();
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         arrRes = new String[0];
      }
      return arrRes;
   }

   /**
    * Get enhanced browse configuration reset targets available for the target browse and the
    * current user. It is a list of levels which have existing enhanced browse configurations
    * defined for the target browse, filtered according to the user's access rights.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    *
    * @return see above.
    */
   public static EnhancedBrowseSaveTarget[] getEnhancedBrowseResetTargets(int browseId)
   {
      EnhancedBrowseSaveTarget[] arrRes = null;
      try
      {
         arrRes = getBrowseById(browseId).getEnhancedBrowseResetTargets();
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
         arrRes = new EnhancedBrowseSaveTarget[0];
      }
      return arrRes;
   }

   /**
    * Reset (delete) an enhanced configuration at the certain level for this browse.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    * @param  resetTarget
    *         Defines the level at which the configuration will be reset.
    *
    * @return key of the configuration that has been deleted or <code>null</code> if there is
    *         no configuration at this level.
    */
   public static EhBrowseConfigKey resetEnhancedConfig(int browseId,
                                                       EnhancedBrowseSaveTarget resetTarget)
   {
      EhBrowseConfigKey res = null;
      try
      {
         res = getBrowseById(browseId).resetEnhancedConfig(resetTarget);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
      }
      return res;
   }

   /**
    * Apply the given enhanced browse configuration to the given browse.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    * @param  config
    *         Enhanced browse configuration to apply.
    *
    * @return <code>true</code> on success.
    */
   public static boolean applyEnhancedConfig(int browseId, EnhancedBrowseConfig config)
   {
      boolean bRes = false;
      try
      {
         bRes = getBrowseById(browseId).applyEnhancedConfig(config);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
      }
      return bRes;
   }

   /**
    * Save the current enhanced browse configuration using the specified save target. If the
    * configuration is supposed to be stored on the client side then on the server side it is only
    * cached.
    *
    * @param  browseId
    *         ID of the browse which should handle the request.
    * @param  saveTarget
    *         Save target (if configuration is applied to "all browses" / "specific browse" and
    *         "all users" / "current user").
    *
    * @return the configuration that has been saved or cached. <code>null</code> on error.
    */
   public static EnhancedBrowseConfig saveOrCacheEnhancedConfig(int browseId,
                                                              EnhancedBrowseSaveTarget saveTarget)
   {
      EnhancedBrowseConfig res = null;
      try
      {
         res = getBrowseById(browseId).saveOrCacheEnhancedConfig(saveTarget);
      }
      catch (MissingWidgetException exc)
      {
         // the browse widget is already gone, no need to crash application
         LOG.warning(exc.getMessage());
      }
      return res;
   }

   /**
    * Marks the end of the editing block when triggered by any abnormal exit
    * from the block on the server.  This method does NOT need to be called
    * if the exit is naturally triggered via {@link #continueEditing}. This
    * method allows the client's state to be cleaned up in the abnormal
    * termination case.
    */
   public static void stopEditingMode()
   {
      ScreenBuffer[] sb = locate().client.stopEditingMode();

      handleScreenBuffers(sb);
   }
   
   /**
    * Conditionally blocks until a next UI event is available.
    */
   public static void waitForNextKey()
   {
      locate().client.waitForNextKey();
   }
   
   /**
    * Checks to see whether a pending GO event is about to end the editing block.
    *
    * @param    hadNext
    *           <code>true</code> if the executing business logic executed a <code>NEXT</code>
    *           statement causing the editing block to iterate. This will result in clearing
    *           any current <code>GO-PENDING</code> condition.  This is passed as a parameter
    *           here so that a separate trip down to the client is not needed to process the
    *           editing block iteration state update.
    *
    * @return   <code>true</code> if there is no <code>GO-PENDING</code> and the editing mode
    *           continues.
    */
   public static boolean continueEditing(boolean hadNext)
   {
      ContinueEditingResult res = locate().client.continueEditing(hadNext);

      GenericFrame gfr = LogicalTerminal.getInstance().identifyFrame(res.editableBuffers[0].getBaseId());
      if (gfr != null)
      {
         handleScreenBuffers(gfr, res.editableBuffers, false);
      }

      return !res.goPending;
   }
   
   /**
    * Deferred resetting of the down frame content.
    *
    * @param   frame
    *          Frame to reset.
    */
   public static void resetFrame(GenericFrame frame)
   {
      if (!frame.getFrameWidget().isDown())
         return;
      
      locate().addMarkEntry(new MarkEntry(new int[] {frame.getFrameId()},
                                          MarkEntry.RESET_FRAME));
   }
   
   /**
    * Deferred signalling of the frame entering a new scope.
    * <p>
    * This signal is used by down frames for resetting both the contents and
    * the cursor to the initial state so that the frame can be properly reused
    * if it enters new scope.
    *
    * @param   frame
    *          Frame to signal.
    * @param   type
    *          Signal type.
    */
   public static void signalFrameScope(GenericFrame frame, int type)
   {
      locate().addMarkEntry(new MarkEntry(new int[] {frame.getFrameId()},
                                          type));
   }
   
   /**
    * Trip processing based on an interactive function being invoked by the
    * business logic.  This will disable infinite loop protection in the
    * <code>TransactionManager</code> and may increment a counter of the
    * interactive requests have occurred in the current iteration of the 
    * current block.
    *
    * @param    update
    *           <code>true</code> to increment the number of interactions in
    *           this block.
    */
   private void markInteractive(boolean update)
   {
      // turn off infinite loop protection
      TransactionManager.disableLoopProtection(null);
      
      if (update)
      {
         // track the number of interactions in this block
         state.interactions++;
         tm.incrementInteractions();
      }      
   }
   
   /**
    * Trip processing based on an interactive function being invoked by the
    * business logic.  This will disable infinite loop protection in the
    * <code>TransactionManager</code> and will increment a counter of the
    * interactive requests have occurred in the current iteration of the 
    * current block IF the current block is not <code>EDITING</code>.
    */
   private void markInteractive()
   {
      markInteractive(tm.getBlockType() != BlockType.EDITING);
   }
   
   /**
    * Suspends processing for the specified number of seconds or until the user presses any key,
    * whichever comes first. The type-ahead buffer is cleared. The specified (or no) message
    * appears on the status line.
    *
    * @param    seconds
    *           The number of seconds to pause. If specified as a -1 number this method waits
    *           indefinitely unless in batch mode. In latter case the method returns immediately.
    * @param    text
    *           If {@code null} the default message is used, otherwise the user-specified message
    *           will be used.
    * @param    noMessage
    *           {@code true} if NO-MESSAGE is in effect. This suppresses the message display, even
    *           if an explicit message text was passed.
    * @param    hWin
    *           target window for IN WINDOW clause.
    */
   private static void pauseWorker(double seconds, String text, boolean noMessage, handle hWin)
   {
      int millis = toMilliseconds(seconds);
      
      if (millis == -1 && isInBatchMode())
      {
         return;
      }

      if (millis > 0)
      {
         notifyRowDisplayDrawListener();
      }
      
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      try
      {
         // there is no need to notify the TransactionManager about this
         // pause (in this method) since the client will report the fact
         // that a pause occurred in its state sync and the TM will be
         // notified via the applyChanges() method of this class
         int keycode = lt.client.pause(millis, text, noMessage, targetWindowWorker(hWin, true));
         
         if (keycode != -2)
         {
            KeyReader.setLastKey(keycode);
         }
      }
      catch (ConditionException ce)
      {
         // make sure that the lastkey is properly maintained
         int kc = ce.getKeyCode();
         
         // anything other than -1 is a valid key press AND even -1 is a
         // valid indicator of CTRL-C if this was a STOP 
         if (kc != -1 || ce instanceof StopConditionException)
         {
            KeyReader.setLastKey(kc);
         }
         
         // continue the unwinding of the stack
         throw ce;
      }      
   }

   /**
    * Request indexes of selected items in RadioSet, ComboBox or SelectionList
    * widgets.
    * 
    * @param   widgetId
    *          Widget ID.
    * 
    * @return  array of 0-based indexes of selected items or <code>null</code>
    *          if specified widget does not exists or does not support such a
    *          request.
    */
   public static int[] getSelection(int widgetId)
   {
      return locate().client.getSelection(widgetId);
   }
   
   /**
    * Delete item from {@link ControlSetEntity}.
    * 
    * @param widgetId
    *        Widget ID. Must point to a {@link ControlSetEntity}.
    * @param idx
    *        1-based index of the item
    * @return <code>true</code> if successful
    * 
    */
   public static boolean deleteFromControlSet(int widgetId, int idx)
   {
      return locate().client.deleteFromControlSet(widgetId, idx);
   }
   
   /**
    * Get server state object.
    * 
    * @return  see above.
    */
   public static ServerState getServerState()
   {
      return locate().state;
   }

   /**
    * Register frame at current scope.
    * 
    * @param   frame
    *          instance of <code>GenericFrame</code> to register.
    */
   public static void registerFrameInScope(GenericFrame frame)
   {
      LogicalTerminal lt = locate();
      
      if (lt.scopeTrack != null)
      {
         lt.scopeTrack.register(frame);
      }
   }

   /**
    * Calls the Microsoft Windows Help engine to display Windows Help topics,
    * and the HTML Help engine to display HTML Help topics.
    *
    * @param   fileName
    *          Path to the help file.
    * @param   action
    *          Help action. See {@link SystemHelpAction} for more information.
    */
   public static void systemHelp(String fileName, SystemHelpAction action)
   {
      UnimplementedFeature.missing("SYSTEM HELP not implemented");
   }

   /**
    * Get the value of CLIENT-TYPE attribute, one of the following values:
    * <ul>
    *    <li>Standard ABL client: 4GLCLIENT</li>
    *    <li>WebClient: WEBCLIENT</li>
    *    <li>AppServer agent: APPSERVER</li>
    *    <li>WebSpeed agent: WEBSPEED</li>
    *    <li>Other special-purpose clients: Unknown value (?)</li>
    * </ul>
    * <p>
    * Currently, this returns either <code>APPSERVER</code> if this is an appserver session or
    * <code>CLIENT</code> in all other cases.
    * 
    * @return   See above.
    */
   public static character getClientType()
   {
      return new character(AppServerManager.isRemote() ? "APPSERVER" : "4GLCLIENT");
   }

   /**
    * Get the value of the WINDOW-SYSTEM attribute, which is the version on which the client was 
    * started. The parameter is computed from one of these values (order by highest to lowest
    * precedence):
    * <ol>
    *    <li>The value read from the directory <code>window-system</code> node, searched using the
    *        {@link Directory#ID_RELATIVE} method.</li>
    *    <li>The value configured at client-side, by the <code>client:driver:window-system</code>
    *        key.</li>
    * </ol>
    * 
    * @return   The value for <code>SESSION:WINDOW-SYSTEM</code>.
    */
   public static character getWindowSystem()
   {
      LogicalTerminal lt = locate();
      
      // use the cached value if previously found
      if (lt.windowSystem == null)
      {
         // try the directory first
         Directory dir = DirectoryManager.getInstance();
         String    win = dir.getString(Directory.ID_RELATIVE, "window-system", null);
      
         if (win == null)
         {
            // fallback to client-specified value
            ClientParameters cp = StandardServer.getClientParameters();
            
            if (cp != null)
            {
               win = cp.windowSystem;
            }
         }
         
         // if the client code is not correct or the client-parameters is not passed, then
         // it is possible this will set null
         lt.windowSystem = win;
      }
      
      // if param is null, unknown value will be the result
      return new character(lt.windowSystem);
   }

   /**
    * The value of state-string determines the wait state and the type of wait cursor displayed.
    * <p>
    * Valid return values are: 
    * <ul>
    *    <li>GENERAL - Displays the system-busy cursor</li>
    *    <li>COMPILER - Displays the ABL compiler cursor</li>
    *    <li>CUSTOM - a custom value, representing a valid cursor</li>
    *    <li>empty string, if none of the above or not a GUI client</li>
    * </ul>
    * <p>
    * This only works on GUI clients.
    * 
    * @return   One of the above described text values.
    */
   public static character getWaitState()
   {
      return new character(locate().waitState);
   }

   /**
    * The value of state-string determines the wait state and the type of wait cursor displayed. 
    * The valid values are: 
    * <ul>
    *    <li>GENERAL - Displays the system-busy cursor</li>
    *    <li>COMPILER - Displays the ABL compiler cursor</li>
    *    <li>custom - a custom value, representing a valid cursor</li>
    * </ul>
    * It does applies only on GUI clients.
    * 
    * @param    state
    *           The new wait-state value.
    * 
    * @return   <code>true</code> if the operation was successful. Returns <code>unknown</code>
    *           and is a no-op if not in GUI mode.
    */
   public static logical setWaitState(String state)
   {
      String stateName = "CUSTOM";
      String ptrName;
      if (state == null || "".equals(state.trim()))
      {
         stateName = "";
         ptrName = null; 
      }
      else
      {
         ptrName = StringHelper.safeTrimTrailing(state);
         if ("COMPILER".equalsIgnoreCase(ptrName))
         {
            stateName = "COMPILER";
            ptrName = "COMPILER-WAIT"; 
         }
         else if ("GENERAL".equalsIgnoreCase(ptrName)) 
         {
            stateName = "GENERAL";
            ptrName = "WAIT"; 
         }
      }
      
      MousePointer ptr = ptrName == null ? null : MousePointer.fromString(ptrName);
      if (ptr != null && ptr.isIgnored())
      {
         return new logical(false);
      }
      
      LogicalTerminal lt = locate();
      
      if (ptrName == null || ptr != null) // pre-defined cursor or reset
      {
         if (lt.client.setWaitState(ptr))
         {
            lt.waitState = stateName;
            return new logical(true);
         }
         return new logical(false);
      }
      
      // cursor from the file
      // try to load at the server side first
      InputStream is = getImageStreamFromApplication(ptrName);
      if (is != null)
      {
         byte[] data = UiUtils.getData(is);
         if (data != null)
         {
            if (lt.client.setWaitState(data))
            {
               lt.waitState = stateName;
               return new logical(true);
            }
         }
      }
      
      if (lt.client.setWaitState(ptrName))
      {
         lt.waitState = stateName;
         return new logical(true);
      }
      return new logical(false);
   }

   
   /**
    * The value of state-string determines the wait state and the type of wait cursor displayed. 
    * The valid values are: 
    * <ul>
    *    <li>GENERAL - Displays the system-busy cursor</li>
    *    <li>COMPILER - Displays the ABL compiler cursor</li>
    *    <li>custom - a custom value, representing a valid cursor</li>
    * </ul>
    * It does applies only on GUI clients.
    * 
    * @param    state
    *           The new wait-state value.
    * 
    * @return   <code>true</code> if the operation was successful. Returns <code>unknown</code>
    *           and is a no-op if not in GUI mode.
    */
   public static logical setWaitState(character state)
   {
      return setWaitState(state == null || state.isUnknown() ? "" : state.toStringMessage());
   }

   /**
    * Resets wait-state previously set by {@link LogicalTerminal#setWaitState(String)}.
    * <p>
    * This method should not be called from the server.
    * <p>
    * The purpose of the method is to provide a callback mechanism for the client when wait-state
    * is reset due to UI flow (i.e. when a modal window is shown). This is also the reason the method
    * doesn't perform a down-call to the client.
    */
   public static void resetWaitState()
   {
      locate().waitState = "";
   }

   /**
    * Utility method for convenient error reporting at client side.
    * 
    * @param    res
    *           Array of error entries.
    */
   static void message(ErrorEntry ... res)
   {
      for (ErrorEntry entry : res)
      {
         message(ErrorManager.buildErrorText(entry.num, entry.text));
      }
   }
   
   /**
    * Send screen definition to client and reset 'suppress redraw' flag.
    * 
    * @param   screenDef
    *          Array of definitions.
    */
   private static void pushScreenDefInt(ScreenDefinition[] screenDef)
   {
      LogicalTerminal lt = locate();
      
      for (ScreenDefinition def : screenDef)
      {
         GenericFrame frame = lt.frameRegistry.get(def.getRootFrameID());
         if (frame != null)
         {
            frame.finishConfigProcessing();
         }
      }
      
      WidgetConfig[] wcfgs = lt.client.pushScreenDefinition(screenDef);
      processDownConfigs(wcfgs);
      
      for (ScreenDefinition def : screenDef)
      {
         def.setSuppressRedraw(false);
      }
   }

   /**
    * Send menu description to client and reset 'suppress redraw' flag.
    * 
    * @param   menuDescr
    *          Array of definitions.
    */
   private static void pushMenuDescrInt(MenuDescription[] menuDescr)
   {
      LogicalTerminal lt = locate();
      
      for (MenuDescription def : menuDescr)
      {
         MenuContainerWidget<?> menu = lt.menuRegistry.get(def.getRootMenuID());
         
         if (menu != null)
         {
            menu.finishConfigProcessing();
         }
      }
      
      lt.client.pushMenuDescription(menuDescr);
      
      for (MenuDescription def : menuDescr)
      {
         def.setSuppressRedraw(false);
      }
   }
   
   /**
    * Honors any changes to the screen buffer received from the client, by
    * merging these into the proper frame's state.
    * 
    * @param   gfr
    *          The frame associated with the screen buffer.
    * @param   sb
    *          The screen buffer to process.
    * @param   allow
    *          <code>true</code> to allow the switching of screen buffers if
    *          the frame is a non-redirected down frame and the frame-line is
    *          detected as changing.  Otherwise this switching will be
    *          disabled.
    */
   private static void handleScreenBuffer(GenericFrame gfr,
                                          ScreenBuffer sb,
                                          boolean      allow)
   {
      if (gfr == null || sb == null)
         return;
      
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      // do regular screen buffer parsing
      gfr.setScreenBuffer(sb, allow);
      KeyReader.setLastKey(sb.getKeyCode(), sb.getKeyState());
      
      lt.frameRef = gfr;
      
      // get frame-field ID
      int id = sb.getValueWidgetId();
      
      // reset frame variables
      lt.frameIndex   = new integer(0);
      lt.frameValue   = "";
      lt.frameName    = "";
      lt.frameField   = "";
      lt.frameFieldId = id;
      lt.frameFile    = "";
      lt.frameDb      = "";
      
      // set new frame variables
      if (id >= 0)
      {
          GenericWidget<?> valWidget = gfr.getWidgetForId(id);
          
          if (valWidget instanceof ControlEntity)
          {
             lt.frameIndex = ((ControlEntity<?>) valWidget).getIndex();
             lt.frameValue = sb.getFrameValue();
             lt.frameName  = gfr._getName();
             lt.frameField = gfr.getOriginalName(valWidget);
             lt.frameFile  = valWidget._getTable();
             lt.frameDb    = valWidget._getDbName();
          }
      }
   }
   
   /**
    * Honors any changes to the screen buffer received from the client, by merging these into the
    * proper frame's state.  The FRAME-VALUE, FRAME-FIELD, FRAME-NAME are taking from the default
    * screen buffer(index 0 in array).  The rest of the screen buffers are handling normally.
    * 
    * @param   gfr
    *          The frame associated with the screen buffer.
    * @param   sb
    *          The screen buffer array to process.
    * @param   allow
    *          <code>true</code> to allow the switching of screen buffers if
    *          the frame is a non-redirected down frame and the frame-line is
    *          detected as changing.  Otherwise this switching will be
    *          disabled.
    */
   private static void handleScreenBuffers(GenericFrame gfr,
                                           ScreenBuffer sb[],
                                           boolean      allow)
   {
      // check for NPE
      if (sb == null)
      {
         return;
      }

      // the first buffer is used to grab the frame values
      handleScreenBuffer(gfr, sb[0], allow);
      // the rest are handling as usual
      handleScreenBuffers(sb, 1);
   }
   
   /**
    * Honors any changes to the screen buffer received from the client, by merging these into 
    * the proper frame's state.
    * <p>
    * This does not update the FRAME-VALUE, FRAME-FIELD, FRAME-NAME functions.
    * 
    * @param   sb
    *          The screen buffer to process.
    */
   private static void handleScreenBuffer(ScreenBuffer sb)
   {
      LogicalTerminal lt = locate();

      if (sb != null)
      {
         GenericFrame gfr = lt.identifyFrame(sb.getBaseId());
         
         if (gfr != null)
         {
            // do regular screen buffer parsing
            gfr.setScreenBuffer(sb, false);
         }
      }
   }
   
   /**
    * Honors any changes to the screen buffer array received from the client, by merging these
    * into the proper frame's state.  Processing from the first item to the end of array.
    * <p>
    * This does not update the FRAME-VALUE, FRAME-FIELD, FRAME-NAME functions.
    * 
    * @param   sb
    *          The screen buffer array to process.
    */
   private static void handleScreenBuffers(ScreenBuffer[] sb)
   {
      handleScreenBuffers(sb, 0);
   }
   
   /**
    * Honors any changes to the screen buffer array received from the client, by merging these
    * into the proper frame's state.
    * <p>
    * This does not update the FRAME-VALUE, FRAME-FIELD, FRAME-NAME functions.
    * 
    * @param   sb
    *          The screen buffer array to process.
    * @param   indexStart
    *          The array index to start handling -1 means starting with 0.
    */
   private static void handleScreenBuffers(ScreenBuffer[] sb, int indexStart)
   {
      // check for NPE
      if (sb == null)
      {
         return;
      }

      // check if we need to start processing from begin or has a shift
      int start = indexStart < 0 ? 0 : indexStart;
      
      for (int i = start; i < sb.length; i++)
      {
         handleScreenBuffer(sb[i]);
      }
   }
   
   /**
    * Prepare a new event list using the current triggers and the given
    * wait list as a basis.
    *
    * @param    frame 
    *           The frame to be edited.
    * @param    wait
    *           List of events to wait for or <code>null</code> for the 
    *           default.
    *
    * @return   The merged event list.
    */
   private EventList prepareEventList(GenericFrame frame, EventList wait)
   {
      // merge all triggers
      EventList wl = TriggerManager.mergeTriggerRegistry(scope);
      
      // copy exit condition events and GO-ON list
      wl.merge(wait, true);

      // set operation type
      wl.setOperation(wait.getOperation());

      // gather the validation information
      if (frame != null)
      {
         for (GenericWidget<?> gw : frame.getValidationSet())
         {
            wl.setValidatable(gw);
         }
      }
      
      return wl;
   }
   
   /**
    * Set the last key and maintain the interactions counter as needed. 
    *
    * @param    ce
    *           The exception to process.
    */
   private static void processCondition(ConditionException ce)
   {
      // make sure that the lastkey is properly maintained
      int kc = ce.getKeyCode();
      
      // anything other than -1 is a valid key press AND even -1 is a
      // valid indicator of CTRL-C if this was a STOP 
      if (kc != -1 || ce instanceof StopConditionException)
      {
         KeyReader.setLastKey(kc);
      }
   }
   
   /** 
    * Callback mechanism to manage frame hiding, interactions counter and other
    * state.  Heavily uses block notifications to trip processing.
    */
   private class ScopeProcessor 
   implements ConditionListener
   {
      /** Frames registered at current scope. */
      private HashMap<Integer, GenericFrame> frames = null;
      
      /** Frames registered at previous scope. */
      private HashMap<Integer, GenericFrame> deepFrames = null;
      
      /** Reference to previous scope in the chain. */
      private ScopeProcessor prev = null;

      /**
       * Set to <code>true</code> when first non-empty list of frames was
       * marked for hiding.
       */
      private boolean marked = false;

      /** Track if any ENDKEY/ERROR condition was raised in this block. */
      private boolean hadException = false;

      /** Track if any exception was raised in this block. */
      private boolean deferReset = false;

      /**
       * List of frames which were marked for hiding last time.
       */
      private ArrayList<Integer> hidden = null;
      
      /** Viewed in current iteration frames. */
      private RoaringBitmap viewed = null;
      
      /** Track frame hiding in current scope. */
      private boolean canHide = false;
      
      /** Mark exception from general hiding rule.*/
      private boolean isException = false;

      /** Maximal number of frames viewed in nested blocks. */
      private int innerViewed = 0;
      
      /** Frame in scope has been wrapped.*/
      private boolean hadWrap = false;

      /** Number of viewed frames in outer scope.*/
      private int savedInnerViewed = 0;

      /** Mark external procedure scope. */
      private final boolean realTopLevel;

      /** Mark inherited or temporary external procedure scope. */
      private boolean topLevel;

      /** Set 'canHide' flag on the next iteration. */
      private boolean forceCanHide = false;
      
      /** A snapshot of the interactions counter at block entry. */
      private int interactionsAtEntry = 0;
      
      /**
       * Create new instance.
       */
      ScopeProcessor()
      {
         interactionsAtEntry = state.interactions;
         
         this.realTopLevel = tm.isTopLevelBlock();
         
         this.prev = scopeTrack;
         
         if (prev != null)
         {
            canHide          = prev.canHide;
            savedInnerViewed = prev.innerViewed;
            topLevel         = prev.topLevel || realTopLevel;
            isException      = prev.isException && topLevel;
         }
         else
         {
            topLevel = realTopLevel;
         }
         
         scopeTrack = this;
      }
      
      /**
       * Clear map of viewed frames.
       */
      public void forceHiding()
      {
         forceCanHide = true;
      }

      /**
       * Notify scope that frame is viewed.
       * 
       * @param   frame
       *          Frame which is about to be viewed.
       */
      public void frameViewed(GenericFrame frame)
      {
         if (frame == null)
            return;
         
         try
         {
            // any frame from current or outer scope should hide other frames.
            // TODO: requires deeper investigation 
            if (!hasFrames() || !frames.containsKey(frame.getFrameId()))
            {
               // frame can hide others only once, either if it is not yet 
               // viewed after start of frame scope or if scope is iterated 
               if (hasFrames() || !frame.isCanHide())
               {
                  return;
               }

               LogicalTerminal.signalFrameScope(frame, 
                                                (topLevel) ?
                                                MarkEntry.CAN_HIDE_OTHERS_EXT :
                                                MarkEntry.CAN_HIDE_OTHERS);
               
               if (frame.isViewed() && !frame.getFrameWidget().isDown())
               {
                  getDeepFrames().put(frame.getFrameId(), frame);
               }
               
               if (topLevel)
               {
                  clearTopLevelFlag();
               }
               
               return;
            }

            if (hasViewed())
            {
               viewed.add(frame.getFrameId());
               return;
            }

            getViewed().add(frame.getFrameId());

            if (canHide || isException)
            {
               LogicalTerminal.signalFrameScope(frame, 
                                                (topLevel) ?
                                                MarkEntry.CAN_HIDE_OTHERS_EXT :
                                                MarkEntry.CAN_HIDE_OTHERS);

               canHide = false;
               isException = false;
               
               if (topLevel)
                  clearTopLevelFlag();
            }
         }
         finally
         {
            frame.setHadView();
         }
      }

      /**
       * Notify that frame is wrapped during DOWN or UP processing.
       * 
       * @param   id
       */
      public void frameWrapped(Integer id)
      {
         if (hasFrames() && frames.containsKey(id))
         {
            // simulate exceptional case
            hadWrap  = true;
         }
         else if (prev != null)
            prev.frameWrapped(id);
      }
      
      /**
       * Process end of block scope.
       * 
       * @see com.goldencode.p2j.util.Finalizable#finished()
       */
      @Override
      public void finished()
      {
         if (prev != null)
         {
            // merge current frames into top level block
            if (hasFrames())
            {
               prev.getDeepFrames().putAll(frames);
            }
            if (hasDeepFrames())
            {
               prev.getDeepFrames().putAll(deepFrames);
            }

            // provide outer block with details about viewed frames
            int count = (hadWrap) ? 2 : 0;
            
            if (!hadWrap && hasViewed() && hasFrames())
            {
               for (Integer id : viewed)
               {
                  GenericFrame frame = frames.get(id);

                  if (frame == null || 
                      frame.isRedirected() || 
                      frame.getFrameWidget().isDown())
                  {
                     continue;
                  }

                  count++;
               }
            }
                        
            if (innerViewed > 1 || count > 1)
               prev.innerViewed = (forceCanHide) ? 1 : 2;
            else
               prev.innerViewed = count + savedInnerViewed;
            
            frames = null;
            deepFrames = null;
         }

         scopeTrack = prev;
         
         BlockType btype = tm.getBlockType();
         
         if (isDoBlock(btype) && marked)
         {
            canHide = false;
         }
         
         if (isSpecialBlock(btype) && hadException && hasHidden())
         {
            deferMarkFrames(hidden, MarkEntry.UNHIDE);
         }
         
         // only if a block has the ENDKEY property, does the
         // interactions value get updated normally; all EDITING blocks and
         // any DO blocks/loops that don't have an explicit ON ENDKEY phrase
         // do not have the ENDKEY property and thus will bypass this code
         if (tm.hasEndkeyProp())
         {
            boolean rollup = tm.isEnclosingRollup();
            int     num    = blocks.pop().intValue();
            
            // certain enclosing blocks ignore rollup processing
            if (rollup)
            {
               // restore previous block value for interactions
               state.interactions += num;
            }
            
            // if an exception occurred clear the current block's
            // interactions counter
            if (deferReset)
            {
               // the local LT interactions counter has already been reset, so
               // only the TM counter must be handled here BUT we must also
               // avoid further modification to state.interactions at this
               // point
               tm.resetInteractions();
            }
            else
            {
               // we don't have to reset interactions to zero, but we may need
               // to restore the previous interactions counter so that we don't
               // get out of balance
               if (!rollup)
               {
                  // the rollup case was already handled above, this just
                  // handles the non-rollup case (which is limited to a nested
                  // trigger within an UPDATE OR a validation expression that
                  // calls a user-defined function within an UPDATE)
                  
                  // we aren't rolling up the value so it doesn't add to the
                  // counter, rather we just restore it to the old value
                  state.interactions = num;
               }
            }
         }
         
         if (prev != null && hasViewed())
         {
            prev.canHide     = canHide;
            prev.isException = false;
         }
      }
      
      /**
       * This implementation does nothing.
       */
      @Override
      public void deleted()
      {
         // no-op
      }
      
      /**
       * Register frame at current scope.
       * 
       * @param   frame
       *          Frame to register.
       */
      public void register(GenericFrame frame)
      {
         getFrames().put(new Integer(frame.getFrameId()), frame);
      }

      /**
       * Process iteration of block nesting scope. All regular frames present
       * at current scope are marked for hiding. All frames at nested scopes
       * are marked for hiding.
       */
      @Override
      public void iterate()
      {
         process(false);
      }

      /**
       * Process retrying of the block nesting scope. All frames present at
       * current scope are marked for hiding. All frames at nested scopes are
       * marked for hiding.
       */
      @Override
      public void retry()
      {
         process(true);
      }
      
      /**
       * Provides a notification that a {@link ConditionException} or an
       * unexpected <code>Throwable</code> was encountered for the current
       * block.
       * <p>
       * This method will reset the interactions counter for any block that
       * has the <code>ENDKEY</code> property. This is done to ensure that
       * if the block is exited, that the current block's interactions will
       * not be added to the containing block's interactions.  If the block
       * is retried or iterated, then the counter is going to be reset
       * anyway so this has no negative effect.
       * <p>
       * If the exception was for <code>ERROR</code> or <code>ENDKEY</code>
       * then a flag will be set to allow exit processing to be modified
       * accordingly.
       * 
       * @param    condition
       *           The condition thrown during block processing.
       */
      @Override
      public void condition(BlockManager.Condition condition)
      {
         // only if a block has the ENDKEY property, does the
         // interactions value get updated normally; all EDITING blocks and
         // any DO blocks/loops that don't have an explicit ON ENDKEY phrase
         // do not have the ENDKEY property and thus will bypass this code
         if (tm.hasEndkeyProp())
         {
            // reset interactions
            state.interactions = 0;
            deferReset = true;
         }
         
         if (condition == BlockManager.Condition.ENDKEY || 
             condition == BlockManager.Condition.ERROR)
         {
            hadException = true;

            if (hasFrames())
            {
               for (GenericFrame frame : frames.values())
               {
                  frame.setCondition();
               }
            }
         }
      }

      /**
       * Clear top level procedure flag.
       */
      private void clearTopLevelFlag()
      {
         ScopeProcessor ptr = this;
         
         do
         {
            ptr.topLevel = false;
            ptr = ptr.prev;
         }
         while (ptr != null && !ptr.realTopLevel);
      }
      
      /**
       * Handle block iteration and retry. Manage state including collection of
       * frames to mark for hide.
       * 
       * @param   isRetry
       *          Allows differentiation of behavior for retry and iterate.
       */
      private void process(boolean isRetry)
      {
         ArrayList<Integer> hide = null;
         
         if (hasHidden())
         {
            hidden.clear();
         }

         if (!isRetry)
         {
            canHide = (innerViewed < 2);
            isException = !canHide; 
         }
         else
         {
            canHide = true;
         }
         
         innerViewed = 0;
         if (hasViewed())
         {
            viewed.clear();
         }
         
         if (hasDeepFrames())
         {
            hide = new ArrayList<Integer>();
            hide.addAll(deepFrames.keySet());
         }

         // in case of RETRY handle deep frames separately
         if (isRetry && hide != null)
         {
            deferMarkFrames(hide, MarkEntry.HIDE);
            
            marked  = true;
            getHidden().addAll(hide);
            hide.clear();
         }
         
         BlockType btype = tm.getBlockType();
         
         if (!isDoBlock(btype) && hasFrames())
         {
            for (Map.Entry<Integer, GenericFrame> entry : frames.entrySet())
            {
               Integer id = entry.getKey();
               GenericFrame frame = entry.getValue();
               
               if (!frame.getFrameWidget().isDown() || isRetry)
               {
                  if (hide == null)
                  {
                     hide = new ArrayList<Integer>();
                  }
                  hide.add(id);
               }

               // in redirected mode, if this frame is page-elem we must 
               // deregister it from the stream headers
               if (frame.isRedirected() || frame.isOutRedirected()) 
               {
                  Stream out = null;
                  if (frame.isRedirected())
                  {
                     out = UnnamedStreams.output();
                  }
                  if (frame.isOutRedirected())
                  {
                     out = frame.getRedirOut();
                  }
                  if (out != null && frame.isPageElement(out))
                  {
                     boolean pageTop = frame.getPageTop();
                     int streamId = frame.getRemoteStreamId(out);
                     LogicalTerminal.deregisterPageElement(frame.getFrameId(), 
                                                           pageTop, 
                                                           streamId);
                  }
               }
            }
         }
            
         if (hide != null)
         {
            deferMarkFrames(hide, (isRetry && hadException) ? MarkEntry.HIDE_NO_PAUSE : MarkEntry.HIDE);
            
            marked  = true;
            getHidden().addAll(hide);
         }

         hadException = false;
         
         // only if a block has the ENDKEY property, does the
         // interactions value get updated normally; all EDITING blocks and
         // any DO blocks/loops that don't have an explicit ON ENDKEY phrase
         // do not have the ENDKEY property and thus will bypass this code
         if (tm.hasEndkeyProp())
         {
            // reset interactions
            state.interactions = 0;
         }
         else
         {
            // TODO: should this occur for iterate() as well?
            if (isRetry)
            {
               // reset the interactions counter to the same value it had on
               // entry
               state.interactions = interactionsAtEntry;
            }
         }
         
         // all blocks track this
         tm.resetInteractions();
      }
      
      /**
       * Setup the deferral processing to mark frames for hiding.
       * 
       * @param   list
       *          Array of frame IDs.
       * @param   mode
       *          Code of the performed operation. 
       */
      private void deferMarkFrames(ArrayList<Integer> list, int mode)
      {
         MarkEntry me = new MarkEntry();
         
         me.frames = Utils.integerCollectionToPrimitive(list);
         me.mode   = mode;
         
         addMarkEntry(me);
      }
      
      /**
       * Check current block scope for type of block.
       * 
       * @param    type
       *           The current block's type.
       * 
       * @return  <code>true</code> if current block scope is a DO block.
       */
      private boolean isDoBlock(BlockType type)
      {
         return (type == BlockType.DO           ||
                 type == BlockType.FINALLY      ||
                 type == BlockType.DO_WHILE     ||
                 type == BlockType.DO_TO        ||
                 type == BlockType.DO_TO_WHILE);
      }
      
      /**
       * Check type of current block and return <code>true</code> if this is
       * some kind of FOR or REPEAT loop.  This excludes non-looping FOR
       * blocks.
       *
       * @param    type
       *           The current block's type.
       * 
       * @return   <code>true</code> if current block is a <code>FOR</code>
       *           loop or <code>REPEAT</code> block.
       */
      private boolean isSpecialBlock(BlockType type)
      {
         return (type == BlockType.FOR_LOOP          ||
                 type == BlockType.FOR_LOOP_TO       ||
                 type == BlockType.FOR_LOOP_TO_WHILE ||
                 type == BlockType.FOR_LOOP_WHILE    ||
                 type == BlockType.REPEAT            ||
                 type == BlockType.REPEAT_TO         ||
                 type == BlockType.REPEAT_WHILE      ||
                 type == BlockType.REPEAT_TO_WHILE);
      }
      
      /**
       * Check if the {@link #viewed} set exists and is not empty.
       * 
       * @return   See above.
       */
      private boolean hasViewed()
      {
         return viewed != null && !viewed.isEmpty();
      }
      
      /**
       * Check if the {@link #frames} map exists and is not empty.
       * 
       * @return   See above.
       */
      private boolean hasFrames()
      {
         return frames != null && !frames.isEmpty();
      }
      
      /**
       * Check if the {@link #deepFrames} map exists and is not empty.
       * 
       * @return   See above.
       */
      private boolean hasDeepFrames()
      {
         return deepFrames != null && !deepFrames.isEmpty();
      }
      
      /**
       * Check if the {@link #hidden} list exists and is not empty.
       * 
       * @return   See above.
       */
      private boolean hasHidden()
      {
         return hidden != null && !hidden.isEmpty();
      }
      
      /**
       * Get the {@link #viewed} set and create it if it does not exist.
       * 
       * @return   See above.
       */
      private RoaringBitmap getViewed()
      {
         return viewed == null ? (viewed = new RoaringBitmap()) : viewed;
      }
      
      /**
       * Get the {@link #deepFrames} map and create it if it does not exist.
       * 
       * @return   See above.
       */
      private HashMap<Integer, GenericFrame> getDeepFrames()
      {
         return deepFrames == null ? (deepFrames = new HashMap<>()) : deepFrames;
      }
      
      /**
       * Get the {@link #frames} map and create it if it does not exist.
       * 
       * @return   See above.
       */
      private HashMap<Integer, GenericFrame> getFrames()
      {
         return frames == null ? (frames = new HashMap<>()) : frames;
      }
      
      /**
       * Get the {@link #hidden} list and create it if it does not exist.
       * 
       * @return   See above.
       */
      private ArrayList<Integer> getHidden()
      {
         return hidden == null ? (hidden = new ArrayList<>()) : hidden;
      }
      
      /**
       * Prepare human readable representations of the instance.
       *
       * @see  java.lang.Object#toString()
       */
      @Override
      public String toString()
      {
         return "Frames: " + (frames == null ? "n/a" : frames) + 
                " Deep: " + (deepFrames == null ? "n/a" : deepFrames);
      }
   }

   /**
    * Return new instance of Window widget.
    * 
    * @return  window widget reference.
    */
   public static WindowWidget createWindow()
   {
      //TODO: window should register itself somehow in terminal? 
      return new WindowWidget(true);
   }

   /**
    * Get current default frame.
    * 
    * @return  current default frame.
    */
   public static GenericFrame defaultFrame()
   {
      UnimplementedFeature.missing("DEFAULT-FRAME method? not implemented");
      return new GenericFrame();
   }
   
   /**
    * HEIGHT-CHARS attribute for SESSION handle.
    * 
    * @return  the height in character units.
    */
   public static decimal getHeightChars()
   {
      int wndId = locate().currentWindow.getId();
      int height = getClient().getDisplayHeight(wndId);
      return new decimal(coordinates().heightFromNative(height));
   }

   /**
    * HEIGHT-CHARS attribute is read-only for SESSION handle. Calls 
    * {@link SessionUtils#readOnlyError} to display appropriate error message.
    * 
    * @param   h
    *          Not used.
    */
   public static void setHeightChars(NumberType h)
   {
      SessionUtils.readOnlyError("HEIGHT-CHARS");
   }

   /**
    * HEIGHT-CHARS attribute is read-only for SESSION handle. Calls 
    * {@link SessionUtils#readOnlyError} to display appropriate error message.
    * 
    * @param   h
    *          Not used.
    */
   public static void setHeightChars(double h)
   {
      SessionUtils.readOnlyError("HEIGHT-CHARS");
   }
   
   /**
    * WIDTH-CHARS attribute for SESSION handle.
    * 
    * @return  the width in character units.
    */
   public static decimal getWidthChars()
   {
      int wndId = locate().currentWindow.getId();
      int width = getClient().getDisplayWidth(wndId);
      return new decimal(coordinates().widthFromNative(width));
   }

   /**
    * WIDTH-CHARS attribute is read-only for SESSION handle. Calls 
    * {@link SessionUtils#readOnlyError} to display appropriate error message.
    * 
    * @param   h
    *          Not used.
    */
   public static void setWidthChars(NumberType h)
   {
      SessionUtils.readOnlyError("WIDTH-CHARS");
   }

   /**
    * WIDTH-CHARS attribute is read-only for SESSION handle. Calls 
    * {@link SessionUtils#readOnlyError} to display appropriate error message.
    * 
    * @param   h
    *          Not used.
    */
   public static void setWidthChars(double h)
   {
      SessionUtils.readOnlyError("WIDTH-CHARS");
   }

   /**
    * Get the WIDTH Progress attribute. The value is equal to the current component char
    * width.
    * 
    * This attribute is an alias for <tt>WIDTH-CHARS attribute</tt>, this method shall
    * never be called directly, its solely purpose is to assert the call
    * <tt>CAN-QUERY(widget, "WIDTH")</tt> works properly. 
    *
    * @return    The integer value of the WIDTH attribute.
    */
   public static decimal getWidth()
   {
      return getWidthChars();
   }

   /**
    * Set the WIDTH Progress attribute.
    *
    * This attribute is an alias for <tt>WIDTH-CHARS attribute</tt>, this method shall
    * never be called directly, its solely purpose is to assert the call
    * <tt>CAN-SET(widget, "WIDTH")</tt> works properly. 
    *
    * @param    widthChars
    *           The integer value for the new WIDTH value.
    */
   public static void setWidth(NumberType widthChars)
   {
      setWidthChars(widthChars);
   }

   /**
    * Get the HEIGHT Progress attribute. The value is equal to the current component char
    * height.
    * 
    * This attribute is an alias for <tt>HEIGHT-CHARS attribute</tt>, this method shall
    * never be called directly, its solely purpose is to assert the call
    * <tt>CAN-QUERY(widget, "HEIGHT")</tt> works properly. 
    *
    * @return    The integer value of the HEIGHT attribute.
    */
   public static decimal getHeight()
   {
      return getHeightChars();
   }

   /**
    * Set the HEIGHT Progress attribute.
    *
    * This attribute is an alias for <tt>HEIGHT-CHARS attribute</tt>, this method shall
    * never be called directly, its solely purpose is to assert the call
    * <tt>CAN-SET(widget, "HEIGHT")</tt> works properly. 
    *
    * @param    heightChars
    *           The integer value for the new HEIGHT value.
    */
   public static void setHeight(NumberType heightChars)
   {
      setHeightChars(heightChars);
   }

   /**
    * WIDTH-PIXELS attribute for SESSION handle.
    * 
    * @return  the width in pixel units.
    */
   public static integer getWidthPixels()
   {
      int wndId = locate().currentWindow.getId();
      int width = getClient().getDisplayWidth(wndId);
      
      // TODO: add native-pixel conversion to CoordinatesConversion
      if (isChui())
      {
         return new integer(coordinates().columnToPixels(width));
      }
      return new integer(width);
   }

   /**
    * WIDTH-PIXELS attribute is read-only for SESSION handle. Calls 
    * {@link SessionUtils#readOnlyError} to display appropriate error message.
    *
    * @param    widthPixels
    *           Not used.
    */
   public static void setWidthPixels(NumberType widthPixels)
   {
      SessionUtils.readOnlyError("WIDTH-PIXELS");
   }

   /**
    * WIDTH-PIXELS attribute is read-only for SESSION handle. Calls 
    * {@link SessionUtils#readOnlyError} to display appropriate error message.
    *
    * @param    widthPixels
    *           Not used.
    */
   public static void setWidthPixels(int widthPixels)
   {
      SessionUtils.readOnlyError("WIDTH-PIXELS");
   }

   /**
    * HEIGHT-PIXELS attribute for SESSION handle.
    * 
    * @return  the height in pixel units.
    */
   public static integer getHeightPixels()
   {
      int wndId = locate().currentWindow.getId();
      int height = getClient().getDisplayHeight(wndId);
      
      // TODO: add native-pixel conversion to CoordinatesConversion
      if (isChui())
      {
         return new integer(coordinates().rowToPixels(height));
      }
      return new integer(height);
   }

   /**
    * HEIGHT-PIXELS attribute is read-only for SESSION handle. Calls 
    * {@link SessionUtils#readOnlyError} to display appropriate error message.
    *
    * @param    heightPixels
    *           Not used.
    */
   public static void setHeightPixels(NumberType heightPixels)
   {
      SessionUtils.readOnlyError("HEIGHT-PIXELS");
   }

   /**
    * HEIGHT-PIXELS attribute is read-only for SESSION handle. Calls 
    * {@link SessionUtils#readOnlyError} to display appropriate error message.
    *
    * @param    heightPixels
    *           Not used.
    */
   public static void setHeightPixels(int heightPixels)
   {
      SessionUtils.readOnlyError("HEIGHT-PIXELS");
   }
   
   /**
    * WORK-AREA-X attribute for SESSION handle.
    * 
    * @return  the attribute value
    */
   public static integer getWorkAreaX()
   {
      int wndId = locate().currentWindow.getId();
      int x = getClient().getDisplayWorkAreaX(wndId);
      
      // TODO: add native-pixel conversion to CoordinatesConversion
      if (isChui())
      {
         return new integer(coordinates().columnToPixels(x));
      }
      return new integer(x);
   }
   
   /**
    * WORK-AREA-X attribute is read-only for SESSION handle. Calls 
    * {@link SessionUtils#readOnlyError} to display appropriate error message.
    * 
    * @param   x
    *          Not used.
    */
   public static void setWorkAreaX(NumberType x)
   {
      SessionUtils.readOnlyError("WORK-AREA-X");
   }
   
   /**
    * WORK-AREA-Y attribute for SESSION handle.
    * 
    * @return  the attribute value
    */
   public static integer getWorkAreaY()
   {
      int wndId = locate().currentWindow.getId();
      int y = getClient().getDisplayWorkAreaY(wndId);
      
      // TODO: add native-pixel conversion to CoordinatesConversion      
      if (isChui())
      {
         return new integer(coordinates().rowToPixels(y));
      }
      return new integer(y);
   }
   
   /**
    * WORK-AREA-Y attribute is read-only for SESSION handle. Calls 
    * {@link SessionUtils#readOnlyError} to display appropriate error message.
    * 
    * @param   y
    *          Not used.
    */
   public static void setWorkAreaY(NumberType y)
   {
      SessionUtils.readOnlyError("WORK-AREA-Y");
   }
   
   /**
    * WORK-AREA-WIDTH-PIXELS attribute for SESSION handle.
    * 
    * @return  the attribute value
    */
   public static integer getWorkAreaWidthPixels()
   {
      int wndId = locate().currentWindow.getId();
      int width = getClient().getDisplayWorkAreaWidth(wndId);
      
      // TODO: add native-pixel conversion to CoordinatesConversion
      if (isChui())
      {
         return new integer(coordinates().columnToPixels(width));
      }
      return new integer(width);
   }

   /**
    * WORK-AREA-WIDTH-PIXELS attribute is read-only for SESSION handle. 
    * Calls {@link SessionUtils#readOnlyError} to display appropriate error message.
    * 
    * @param   widthPixels
    *          Not used.
    */
   public static void setWorkAreaWidthPixels(NumberType widthPixels)
   {
      SessionUtils.readOnlyError("WORK-AREA-WIDTH-PIXELS");
   }
   
   /**
    * WORK-AREA-HEIGHT-PIXELS attribute for SESSION handle.
    * 
    * @return  the attribute value
    */
   public static integer getWorkAreaHeightPixels()
   {
      int wndId = locate().currentWindow.getId();
      int height = getClient().getDisplayWorkAreaHeight(wndId);
      
      // TODO: add native-pixel conversion to CoordinatesConversion      
      if (isChui())
      {
         return new integer(coordinates().columnToPixels(height));
      }
      return new integer(height);
   }

   /**
    * WORK-AREA-HEIGHT-PIXELS attribute is read-only for SESSION handle. 
    * Calls {@link SessionUtils#readOnlyError} to display appropriate error message.
    * 
    * @param   heightPixels
    *          Not used.
    */
   public static void setWorkAreaHeightPixels(NumberType heightPixels)
   {
      SessionUtils.readOnlyError("WORK-AREA-HEIGHT-PIXELS");
   }
   
   /**
    * Gets the display type (GUI or TTY).  Corresponds with the
    * <code>SESSION:DISPLAY-TYPE</code> system handle attribute.
    * 
    * @return  Display type.  Currently, only TTY is supported.
    */
   public static character getDisplayType()
   {
      if (isChui())
      {
         return new character("TTY");
      }
      return new character("GUI");
   }
   
   /**
    * Gets the current desktop settings.
    * 
    * @return   The current desktop settings.
    */
   public static DesktopSettings getDesktopSettings()
   {
      return DesktopSettingsManager.getDesktopSettings();
   }
   
   /**
    * Gets the keyboard layouts settings.
    * 
    * @return   The keyboard layouts settings.
    */
   public static KeyboardLayoutsSettings getKeyboardLayoutsSettings()
   {
      return KeyboardLayoutsSettingsManager.getKeyboardLayoutsSettings();
   }

   /**
    * Retrieve the number of milliseconds to wait when loading a HTML web page.
    * This works as a timeout.
    * 
    * @return   An integer representing the HTML browser load timeout.
    */
   public static Integer getHTMLBrowserLoadTimeout()
   {
      return HtmlBrowserWidget.getLoadTimeout();
   }
   
   /**
    * Get THREE-D attribute value for SESSION handle.
    * 
    * @return  the attribute value
    */
   public static logical is3D()
   {
      return new logical(getClient().is3D());
   }
   
   /**
    * Set THREE-D attribute value for SESSION handle.
    * 
    * @param   value
    *          The value to set.
    */
   public static void set3D(logical value)
   {
      // TODO: unknown
      getClient().set3D(value.getValue());
   }
   
   /**
    * Get the state of SESSION:V6DISPLAY attribute. 
    * 
    * @return  the state of SESSION:V6DISPLAY attribute.
    */
   public static logical isV6Display()
   {
      return new logical(locate().v6display);
   }
   
   /**
    * Set the state of SESSION:SESSION:V6DISPLAY attribute.
    * 
    * @param   on
    *          The new state for SESSION:SESSION:V6DISPLAY attribute.
    */
   public static void setV6Display(logical on)
   {
      locate().v6display = on.booleanValue();
   }
   
   /**
    * Set the state of SESSION:SESSION:V6DISPLAY attribute.
    * 
    * @param   on
    *          The new state for SESSION:SESSION:V6DISPLAY attribute.
    */
   public static void setV6Display(boolean on)
   {
      locate().v6display = on;
   }

   /**
    * Gets the global SESSION:DISABLE-REDRAW attribute.
    * This attribute is a FWD extension, not real 4GL!
    *
    * @return  current value of DISABLE-REDRAW attribute.
    */
   public static logical isDisableRedraw()
   {
      return isChui() ? defaultWindow().unwrapWidget().isDisableRedraw()
                      : new logical(locate().disableRedraw);
   }
   
   /**
    * Sets the global SESSION:DISABLE-REDRAW attribute.
    * This attribute is a FWD extension, not real 4GL!
    *
    * @param   value
    *          Value to set to DISABLE-REDRAW attribute.
    */
   public static void setDisableRedraw(logical value)
   {
      if (value == null || value.isUnknown())
      {
         return;
      }

      setDisableRedraw(value.getValue());
   }
   
   /**
    * Sets the global SESSION:DISABLE-REDRAW attribute.
    * This attribute is a FWD extension, not real 4GL!
    *
    * @param   value
    *          Value to set to DISABLE-REDRAW attribute.
    */
   public static void setDisableRedraw(boolean value)
   {
      if (isChui())
      {
         // there is only one window in ChUI, use that
         defaultWindow().unwrapWidget().setDisableRedraw(value);
         return;
      }

      LogicalTerminal lt = locate();
      
      if (lt.disableRedraw == value)
      {
         return;
      }
      
      lt.disableRedraw = value;
      lt.client.setDisableRedraw(value);
   }
   
   /**
    * Disable or enable screen refresh (drawing) for the window associated with the
    * specified widget.  On enable, all pending output will be displayed. This is
    * useful to avoid flickering and improve performance during intermediate updates
    * that don't need to be seen by the user.
    *
    * @param    widget
    *           The widget whose containing window will have refresh disabled or
    *           enabled.
    * @param    disable
    *           <code>true</code> to disable refresh, <code>false</code> to enable.
    *
    * @return   Reports if the operation succeeded.
    */
   public static logical disableRedraw(handle widget, logical disable)
   {
      if (widget == null  || !widget._isValid() || widget.isUnknown() ||
          disable == null || disable.isUnknown())
      {
         return new logical(false);
      }
      
      return disableRedraw(widget, disable.booleanValue());
   }
   
   /**
    * Disable or enable screen refresh (drawing) for the window associated with the
    * specified widget.  On enable, all pending output will be displayed. This is
    * useful to avoid flickering and improve performance during intermediate updates
    * that don't need to be seen by the user.
    *
    * @param    widget
    *           The widget whose containing window will have refresh disabled or
    *           enabled.
    * @param    disable
    *           <code>true</code> to disable refresh, <code>false</code> to enable.
    *
    * @return   Reports if the operation succeeded.
    */
   public static logical disableRedraw(handle widget, boolean disable)
   {
      // dereference the resource
      WrappedResource wr = widget.getResource();
      
      // find the window for the given widget
      WindowWidget win = lookupWindow(wr);
      
      if (win == null)
      {
         return new logical(false);
      }
      
      win.setDisableRedraw(disable);
      
      return new logical(true);
   }
   
   /**
    * Disable or enable screen refresh (drawing) for the window associated with the
    * specified widget.  On enable, all pending output will be displayed. This is
    * useful to avoid flickering and improve performance during intermediate updates
    * that don't need to be seen by the user.
    * <p>
    * If a non-zero value is given as widgetId, that widget will have redraw disabled and will be
    * saved as the "locked" window. If called with a widgetId of 0, then the current locked 
    * window will have drawing re-enabled.
    * <p>
    * There can only be one locked window at any time, calling this with a non-zero widgetId when
    * there is already a locked window will do nothing.  Likewise, calling this with a 0 when
    * there is no locked window will do nothing.
    *
    * @param    widgetId
    *           The widget whose containing window will have refresh disabled or
    *           enabled.
    *
    * @return   Reports if the operation succeeded.
    */
   public static logical disableRedraw(int64 widgetId)
   {
      if (widgetId == null || widgetId.isUnknown())
      {
         return new logical(false);
      }
      
      return new logical(disableRedrawWorker(widgetId.intValue()));
   }
   
   /**
    * Disable or enable screen refresh (drawing) for the window associated with the
    * specified widget.  On enable, all pending output will be displayed. This is
    * useful to avoid flickering and improve performance during intermediate updates
    * that don't need to be seen by the user.
    * <p>
    * If a non-zero value is given as widgetId, that widget will have redraw disabled and will be
    * saved as the "locked" window. If called with a widgetId of 0, then the current locked 
    * window will have drawing re-enabled.
    * <p>
    * There can only be one locked window at any time, calling this with a non-zero widgetId when
    * there is already a locked window will do nothing.  Likewise, calling this with a 0 when
    * there is no locked window will do nothing.
    *
    * @param    widgetId
    *           The widget whose containing window will have refresh disabled or
    *           enabled.
    * @param    rc
    *           Output parameter that captures the return code which will be 1 if the operation
    *           succeeded and 0 on any failure. 
    *
    * @return   Reports if the operation succeeded.
    */
   public static logical disableRedraw(int64 widgetId, int64 rc)
   {
      // if rc is a field reference, then the BlockManager infra is needed to process the output
      // parameter properly (see TransactionManager.deregisterOutputParameterAssigner())
      return BlockManager.function("disable-redraw", logical.class, new Block((Body) () -> 
      {
         if (widgetId == null || widgetId.isUnknown())
         {
            rc.assign(0);
            BlockManager.returnNormal(new logical(false));
            return;
         }

         boolean result = disableRedrawWorker(widgetId.intValue());
         
         rc.assign(result ? 1 : 0);
         
         BlockManager.returnNormal(new logical(result));
      }));
   }   
   
   /**
    * Disable or enable screen refresh (drawing) for the window associated with the
    * specified widget.  On enable, all pending output will be displayed. This is
    * useful to avoid flickering and improve performance during intermediate updates
    * that don't need to be seen by the user.
    * <p>
    * If a non-zero value is given as widgetId, that widget will have redraw disabled and will be
    * saved as the "locked" window. If called with a widgetId of 0, then the current locked 
    * window will have drawing re-enabled.
    * <p>
    * There can only be one locked window at any time, calling this with a non-zero widgetId when
    * there is already a locked window will do nothing.  Likewise, calling this with a 0 when
    * there is no locked window will do nothing.
    *
    * @param    widgetId
    *           The widget whose containing window will have refresh disabled or
    *           enabled. 
    *
    * @return   Reports if the operation succeeded.
    */
   public static logical disableRedraw(int widgetId)
   {
      return new logical(disableRedrawWorker(widgetId));
   }
   
   /**
    * Disable or enable screen refresh (drawing) for the window associated with the
    * specified widget.  On enable, all pending output will be displayed. This is
    * useful to avoid flickering and improve performance during intermediate updates
    * that don't need to be seen by the user.
    * <p>
    * If a non-zero value is given as widgetId, that widget will have redraw disabled and will be
    * saved as the "locked" window. If called with a widgetId of 0, then the current locked 
    * window will have drawing re-enabled.
    * <p>
    * There can only be one locked window at any time, calling this with a non-zero widgetId when
    * there is already a locked window will do nothing.  Likewise, calling this with a 0 when
    * there is no locked window will do nothing.
    *
    * @param    widgetId
    *           The widget whose containing window will have refresh disabled or
    *           enabled. 
    * @param    rc
    *           Output parameter that captures the return code which will be 1 if the operation
    *           succeeded and 0 on any failure. 
    *
    * @return   Reports if the operation succeeded.
    */
   public static logical disableRedraw(int widgetId, int64 rc)
   {
      return disableRedraw(new int64(widgetId), rc);
   }
   
   /**
    * Set CURRENT-WINDOW for the given procedure handle.
    * 
    * @param   phandle
    *          Procedure handle.
    * @param   cWinHandle
    *          handle to the Window reference.
    */
   public static void setProcCurrentWindow(handle phandle, handle cWinHandle)
   {
      if (cWinHandle == null || !cWinHandle._isValid() || cWinHandle.isUnknown()) 
      {
         String msg = "**Unable to assign UNKNOWN value to attribute CURRENT-WINDOW " +
                      "on PROCEDURE WIDGET";
         // no space after ** - so we don't use prefix
         ErrorManager.recordOrShowError(4083, msg, false, false, false);
         
         return;
      }
      
      if (!(cWinHandle.getResource() instanceof WindowWidget))
      {
         String msg = "**Unable to set attribute CURRENT-WINDOW for PROCEDURE widget";
         // no space after ** - so we don't use prefix
         ErrorManager.recordOrShowError(4078, msg, false, false, false);
         
         return;
      }
      
      ProcedureManager.setCurrentWindow(phandle.get(), cWinHandle);
      
      // TODO: in GUI mode we may need some custom implementation
   }

   /**
    * Get color from given color spec or from default messages color.
    *
    * @param   cs
    *          Color specification.
    * @param   defaultColorspec
    *          Logical terminal instance.
    *
    * @return  color.
    */
   private static Color color(ColorSpec cs, ColorSpec defaultColorspec)
   {
      return cs != null ? cs.convert() : defaultColorspec.convert();
   }

   /** 
    * Get code of last event function (LAST-EVENT:FUNCTION)
    * 
    * @return  actionCode 
    */
   public static int getFunctionKey() 
   {
      return locate().functionKey;
   }
   
   /**
    * Implementation of the LAST-EVENT system handle's X attribute.
    *
    * @return  The pixel location of the mouse cursor relative to the left edge of the display.
    */
   public static Integer getEventX() 
   {
      return locate().eventX;
   }
   
   /**
    * Implementation of the LAST-EVENT system handle's Y attribute.
    *
    * @return  The pixel location of the mouse cursor relative to the left edge of the display.
    */
   public static Integer getEventY() 
   {
      return locate().eventY;
   }

   /**
    * Implementation of the LAST-EVENT system handle's COLUMN attribute.
    *
    * @return  The character location of the mouse cursor relative to the left edge of the display or
    *          {@code null} if no column value has been stored.
    */
   public static Double getEventColumn() 
   {
      LogicalTerminal lt = locate();
      
      return (lt.eventX == null) ? null : LogicalTerminal.coordinates.columnFromNative(lt.eventX);
   }
   
   /**
    * Implementation of the LAST-EVENT system handle's ROW attribute.
    *
    * @return  The character location of the mouse cursor relative to the left edge of the display or
    *          {@code null} if no column value has been stored.
    */
   public static Double getEventRow() 
   {
      LogicalTerminal lt = locate();
      
      return (lt.eventY == null) ? null : LogicalTerminal.coordinates.rowFromNative(lt.eventY);
   }

   /**
    * Open the hyperlink in the browse column by raising the event previously set for this column
    * using the HYPERLINK attribute (FWD extension). <code>hyperlinkText</code> is used as the
    * event parameter. I.e. opening of a hyperlink is an equivalent of <p>
    * <code>PUBLISH eventName (hyperlinkText).</code>
    *
    * @param columnWidgetId
    *        ID of the target browse column widget.
    * @param hyperlinkText
    *        Text of the hyperlink to be opened.
    */
   public static void openHyperlink(int columnWidgetId, String hyperlinkText)
   {
      BrowseColumnWidget columnWidget = (BrowseColumnWidget) getWidgetForId(columnWidgetId);

      if (columnWidget == null)
      {
         throw new RuntimeException("Can't locate BrowseColumnWidget with ID " + columnWidgetId);
      }

      columnWidget.openHyperlink(hyperlinkText);
   }

   /**
    * Register a widget which ID in widget registry map.
    * 
    * @param    widget
    *           Widget object to register.
    * @param    widgetId
    *           The integer ID to associate with new widget.
    */
   public static void registerWidget(GenericWidget<?> widget, int widgetId)
   {
      // simple check for validity
      if (widget == null || widgetId <= WidgetId._DEFAULT_WINDOW_ID)
      {
         // log for errors
         if (widget == null)
         {
            LOG.warning("Attempt to register invalid widget or widget ID!");
         }
         if (widgetId <= WidgetId._DEFAULT_WINDOW_ID)
         {
            LOG.warning("Attempt to register incorrect widget ID!");
         }
         
         return;
      }
      
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      
      synchronized (lt.registryLock)
      {
         // add new widget to registry
         lt.widgetRegistry.put(widgetId, widget);
         
         if (widget instanceof MenuItemWidget)
         {
            lt.pendingMenus.add(widgetId);
         }
      }
   }
   
   /**
    * Deregister a widget which ID from widget registry map.
    * 
    * @param    widgetId
    *           The integer ID of widget to remove from the map.
    */
   public static void deregisterWidget(int widgetId)
   {
      // simple check for validity
      if (widgetId <= WidgetId._DEFAULT_WINDOW_ID)
      {
         // log for error
         LOG.warning("Attempt to deregister incorrect widget ID(" + widgetId + ")!");
         return;
      }
      
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      synchronized (lt.registryLock)
      {
         // remove widget from registry
         lt.widgetRegistry.remove(widgetId);
      }
   }
   
   /**
    * Calling the client side to activate already created menu.
    * 
    * @param    menuWid
    *           The integer ID of the menu to activate.
    * @param    x
    *           The new X position of the menu, unknown value to keep the default value.
    * @param    y
    *           The new Y position of the menu, unknown value to keep the default value.
    */
   public static void activatePopUpMenu(MenuWidget menuWid, int64 x, int64 y)
   {
      int wId = menuWid != null ? menuWid.getId() : 0;
      
      if (wId != 0)
      {
         LogicalTerminal lt = locate();
      
         lt.client.activatePopUpMenu(wId, x, y);
      }
   }
   
   /**
    * Register a window widget which was dynamically created.
    * 
    * @param    config
    *           Window configuration.
    * @param    window
    *           The new created window widget.
    */
   static void registerWindow(WindowConfig config, WindowWidget window)
   {      
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      synchronized (lt.registryLock)
      {
         // Window IDs are requested from central session point 
         int id = WidgetId.nextID().asInt();
         
         // set window id and register
         config.id = new WidgetId(id);
         lt.windowRegistry.put(id, window);

         // assign environment
         BaseSysTable.registerWindow(config);

         // refresh widget registry
         lt.widgetRegistry.put(id, window);
         
         // register the window config
         ConfigManager.getInstance().addWidgetConfig(config);

         // push window down to client side
         LogicalTerminal.pushWindow(config);
      }
   }
   
   /**
    * Find the direct children of the specified parent window.
    * 
    * @param    parent
    *           The parent window.
    *           
    * @return   See above.
    */
   static List<WindowWidget> findDirectChildren(WindowWidget parent)
   {
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();
      List<WindowWidget> result = new ArrayList<>();
      
      synchronized (lt.registryLock)
      {
         for (WindowWidget ww : lt.windowRegistry.values())
         {
            if (ww.config.parentId == parent.getId())
            {
               result.add(ww);
            }
         }
      }
      
      return result;
   }
   
   /**
    * Deregister a window widget which was dynamically created, as it is now deleted.
    * 
    * @param    window
    *           The deleted window widget.
    */
   static void deregisterWindow(WindowWidget window)
   {      
      // locate the Logical Terminal for this session
      LogicalTerminal lt = locate();

      synchronized (lt.registryLock)
      {
         int wid = window.getId();
         
         lt.currentWindowStack.remove(window);
         
         if (lt.currentWindow.getId() == wid)
         {
            WindowWidget firstVisibleWindow = lt.getFirstVisibleWindow();
            
            // choose the current window
            WindowWidget targetWindow;
            
            if (lt.currentWindowStack.isEmpty())
            {
               if (lt.defaultWindow._isVisible())
               {
                  // assign DEFAULT-WINDOW to be CURRENT-WINDOW if the current window should be deleted.
                  targetWindow = lt.defaultWindow;
               }
               else
               {
                  targetWindow = firstVisibleWindow;
               }
            }
            else
            {
               targetWindow = Lists.reverse(lt.currentWindowStack).stream().filter(win ->
               {
                  return win.isVisible().booleanValue();
               }).findFirst().orElse(firstVisibleWindow);
            }
            
            setCurrentWindow(targetWindow);
         }
         
         // this needs to be done first!
         lt.client.removeWindow(wid);
         
         lt.windowRegistry.remove(wid);
         
         // refresh widget registry
         lt.widgetRegistry.remove(wid);
      }
   }
   
   /**
    * Get the {@link ClientExports} for this session.
    * 
    * @return   See above.
    */
   public static ClientExports getClient()
   {
      return locate().client;
   }
   
   /**
    * Get RGB value for specified index from color table associate to a Window.
    * 
    * @param   envName
    *          Environment name.
    * @param   index
    *          Color table index.
    *          
    * @return  RGB value in Java format.
    */
   public static int getColor(String envName, int index)
   {
      return ColorTable.getColor(envName, index);
   }

   /**
    * Get system colors table.
    * 
    * @return  System colors.
    */
   public static ColorTableEntry[] readSystemColorTable()
   {
      return ColorTable.getSystemColors();      
   }

   /**
    * Get color table of the specified environment.
    * 
    * @param   envName
    *          Environment name. Could be <code>null</code> if is default environment. 
    *          
    * @return  see above.
    */
   public static EnvironmentColorTable readColorTable(String envName)
   {
      return ColorTable.getColorTable(envName);
   }
   
   /**
    * Synchronize environment color table state.
    * 
    * @param   ect
    *          Color table to synchronize
    */
   public static void synch(EnvironmentColorTable ect)
   {
      ServerState state = locate().state;
      
      if (state.environments == null)
      {
         state.environments = new LinkedList<>();
      }
      
      state.environments.add(ect);
   }

   /**
    * Accesses the SESSION:TOOLTIPS attribute for this session.
    *
    * @return   <code>true</code> if tooltips are available for this session
    *           <code>false</code> otherwise.
    */
   public static boolean getSessionTooltips()
   {
      return SessionUtils.getSessionTooltips().booleanValue();
   }
   
   /**
    * Get the Progress stack trace
    * 
    * @return the Progress stack trace
    */
   public static String[] getProgressStackTrace()
   {
      List<String> stack = ProcedureManager.getStackTrace();
      return stack.toArray(new String[stack.size()]);
   }

   /**
    * This API is used when P2J client runs in embedded mode (i.e. the P2J screens are embedded
    * into another customer-specific application).  It allows the customer-specific to execute
    * converted 4GL code.  Only external programs, procedures and functions can be invoked.  The
    * result of this invocation (and any additional OUTPUT parameters, errors, etc) will be 
    * returned back to the customer-specific application.
    * 
    * @param    request
    *           The details about the request.
    *           
    * @return   The response for this request.
    */
   public static ServerReturnValue invoke(InvocationRequestPayload request)
   {
      if (!isEmbedded())
      {
         String msg = "Attempted to execute a remote invoke request in a non-embedded client.";

         throw new SilentUnwindException(new IllegalStateException(msg));
      }

      ServerReturnValue srv = new ServerReturnValue();
      LogicalTerminal lt = locate();

      // delegate the call to ControlFlowOps
      InvocationResponsePayload res = 
         (InvocationResponsePayload) lt.invokeOnServer(srv, () -> ControlFlowOps.invoke(request));
      
      srv.payload = res;
      
      return srv;
   }

   /**
    * Invoke an internal procedure with the specified name.
    * 
    * @param    procedure
    *           The procedure name.
    * @param    params
    *           The parameters.
    * 
    * @return   A response holding the RETURN-VALUE for this procedure call.
    */
   public static ServerReturnValue invokeProcedure(String procedure, Object... params)
   {
      ServerReturnValue srv = new ServerReturnValue();
      
      LogicalTerminal lt = locate();

      // delegate the call to ControlFlowOps
      Object res = lt.invokeOnServer(srv, () -> 
      {
         // simulate an RUN ... IN THIS-PROCEDURE so that an internal procedure is always resolved
         
         ControlFlowOps.invokeIn(procedure, lt.pm.thisProcedure(), params);
         
         return ControlFlowOps.getReturnValue();
      });
      
      srv.payload = res;
      
      return srv;
   }
   
   /**
    * Invoke an internal procedure with the specified name.
    * 
    * @param    procedure
    *           The internal procedure name.
    * @param    inProcHandle
    *           the parent procedure handle converted with {@link handle#toString()}
    * @param    params
    *           The parameters.
    * 
    * @return   A response holding the RETURN-VALUE for this procedure call.
    */
   public static ServerReturnValue invokeInProcedure(String procedure, String inProcHandle, Object... params)
   {
      ServerReturnValue srv = new ServerReturnValue();
      
      LogicalTerminal lt = locate();

      // delegate the call to ControlFlowOps
      Object res = lt.invokeOnServer(srv, () -> 
      {
         ControlFlowOps.invokeIn(procedure, handle.fromString(inProcHandle), params);
         
         return ControlFlowOps.getReturnValue();
      });
      
      srv.payload = res;
      
      return srv;
   }
   
   /**
    * This is a special API which is used to delegate a call from converted 4GL code 
    * (via _P2J_REMOTE_CALL_ functions) to the remote side.  Usually, this is used when P2J client
    * runs in embedded mode, i.e. the P2J screens are embedded in another customer-specific 
    * application.  The call will be remoted into the customer-specific application.
    * <p>
    * This API is expected to block waiting for a response.
    * 
    * @param    request
    *           The request data incoming from 4GL.
    * 
    * @return   The response from the customer-specific application, which will be sent back to 
    *           the converted 4GL code.
    */
   public static UnformattedPayload remoteCall(UnformattedPayload request)
   {
      return locate().client.remoteCall(request);
   }
   
   /**
    * Send a PUBLISH notification to a remote application. 
    * 
    * @param    res
    *           The external resource which needs to be notified of this PUBLISH event.
    * @param    event
    *           The published event name.
    * @param    param
    *           The parameters for this event.
    * @param    delayed
    *           Flag indicating if the PUBLISH must be executed immediately (when <code>true</code>)
    *           or posted to the client's event queue, otherwise.
    */
   public static void publish(ExternalResource res, String event, String[] param, boolean delayed)
   {
      locate().client.publish(res.id().longValue(), event, param, delayed);
   }
   
   /**
    * Delete the resource associated with the specified handle.
    * <p>
    * Only external programs ran by the embedded side and {@link ExternalResource external resources} 
    * can be deleted via this API.
    * 
    * @param    handleId
    *           The handle ID of the resource.
    * @param    noError
    *           <code>true</code> to process using silent error mode.
    */
   public static void deleteHandle(long handleId, boolean noError)
   {
      if (!isEmbedded())
      {
         String msg = "Attempted to execute a delete handle request in a non-embedded client.";
         throw new SilentUnwindException(new IllegalStateException(msg));
      }

      Runnable block = () ->
      {
         try
         {
            handle h = handle.fromResourceId(handleId);
   
            if (!h.isExternalResource() && 
                h._isValid()            && 
                !ProcedureManager.isRemotePersistentProcedure(h.get()))
            {
               String msg = "Attempting to delete a handle which was not obtained by the " +
                            "remote, customer-specific application!";
               if (LOG.isLoggable(Level.SEVERE))
               {
                  LOG.severe(msg);
               }
               
               throw new SilentUnwindException(new IllegalStateException(msg));
            }
   
            HandleOps.delete(h);
         }
         finally
         {
            ProcedureManager.deregisterRemotePersistentProcedure(handleId);
         }
      };
      
      if (noError)
      {
         ErrorManager.silent(block);
      }
      else
      {
         block.run();
      }
   }

   /**
    * Create a new subscription to the given event for a specified external resource. 
    * This resource will receive a notification when the event is published.
    *
    * @param    global
    *           Flag indicating if the subscription is for global (cross-session) events.
    * @param    subscriber
    *           The resource to notify when the specified event is published. If <code>null</code>,
    *           a {@link ExternalResource} will be built and its ID return.
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The handle representing the program or resource that is the source of the event. 
    *           If specified, the named event will only cause a notification if the event source 
    *           matches this handle. <code>null</code> represents the <code>ANYWHERE</code> option
    *           which means that the events will not be filtered by source.
    * @param    noError
    *           <code>true</code> to process using silent error mode.
    * 
    * @return   The subscriber's ID.
    */
   public static long subscribeExternal(boolean global, 
                                        Long    subscriber, 
                                        String  event, 
                                        Long    publisher,
                                        boolean noError)
   {
      handle hsubscriber = (subscriber == null ? null : handle.fromResourceId(subscriber));
      handle hpublisher = (publisher == null ? null : handle.fromResourceId(publisher));
      
      handle[] result = new handle[1];
      
      // not sure if this is needed, but it matches the way the original code was written
      result[0] = hsubscriber;

      Runnable block = () ->
      {
         result[0] = NamedEventManager.subscribeExternal(global, hsubscriber, event, hpublisher);
      };
      
      if (noError)
      {
         ErrorManager.silent(block);
      }
      else
      {
         block.run();
      }
      
      return result[0].getResourceId();
   }

   /**
    * Delete a subscription to the given event for the specified external resource.
    * <p>
    * All processing is done with silent error mode implicitly enabled.
    *
    * @param    global
    *           Flag indicating if the search is done in the global (cross-session) event register.
    * @param    subscriber
    *           The resource whose subscription is to be deleted.  Must not be <code>null</code>.
    * @param    event
    *           The name of the event which is being published. If this is <code>null</code>, then
    *           subscriptions to all events will be deleted.
    * @param    publisher
    *           The procedure or resource handle representing the source of the event.
    *           If specified, the subscription will only be deleted if the event source matches
    *           this handle.  <code>null</code> represents the <code>ALL</code> option which means
    *           that the events will not be filtered by source. When <code>null</code>, it means
    *           that the publisher was not set.
    */
   public static void unsubscribeExternal(boolean global, 
                                          Long    subscriber, 
                                          String  event,
                                          Long    publisher)
   {
      handle hsubscriber = handle.fromResourceId(subscriber);
      handle hpublisher = (publisher == null ? null : handle.fromResourceId(publisher));

      NamedEventManager.unsubscribeExternal(global, hsubscriber, event, hpublisher);
   }

   /**
    * Notify all subscribers (to the given event) that the event has occurred.
    * <p>
    * All processing is done with silent error mode implicitly enabled.
    * <p>
    * The subscriptions are fired in the order they were subscribed. Although 4GL
    * documentation specifically states, that "If a named event has multiple subscribers, 
    * the order in which the AVM notifies subscribers is undefined.", the actual
    * 4GL implementation (OpenEdge 10.2B) does fire the events in the order the subscriptions
    * get subscribed.
    *
    * @param    event
    *           The name of the event which is being published.
    * @param    publisher
    *           The procedure or resource handle representing the program that is the source of  
    *           the event. Must not be <code>null</code>.
    * @param    args
    *           The parameters passed to the subscriber. May be <code>null</code> to represent no
    *           parameters. When <code>null</code>, it means that no parameters were specified for
    *           this call.
    */
   public static void publishExternal(String event, long publisher, String[] args)
   {
      handle hpublisher = handle.fromResourceId(publisher);

      NamedEventManager.publishExternal(event, hpublisher, args);
   }

   /**
    * Create an {@link ExternalResource external resource} to be used by a remote, non P2J, side.
    * 
    * @return   The resource id.
    */
   public static long createExternal()
   {
      handle hres = handle.buildExternalResource();
      
      return handle.resourceId(hres.getResource());
   }
   
   /**
    * Refresh the server-side buffer for the given frame.
    * 
    * @param    buffer
    *           The buffer array data to be refreshed.  May be <code>null</code>.
    */
   public static void refreshBuffers(ScreenBuffer[] buffer)
   {
      if (buffer == null)
      {
         return;
      }

      final GenericWidget<?> widget = getWidgetForId(buffer[0].getBaseId());
      if (widget == null)
      {
         return;
      }

      GenericFrame frame = widget.getFrame();

      handleScreenBuffers(frame, buffer, false);
   }

   /**
    * Determines target window for different window-aware operations such as MESSAGE, PAUSE 
    * or others. Forces hWin usage, ignoring frames window. If hWin is null, frame window is used.
    * 
    * @param   frame
    *          A frame used in current operation (for example ENABLE, VIEW and so on).
    * @param   hWin
    *          Target window for IN WINDOW clause. <code>null</code> means get default window.
    * @param   ignoreThisProc
    *          <code>true</code> if concrete operation ignores THIS-PROCEDURE:CURRENT-WINDOW.
    * 
    * @return  ID of target window that is hWin resource id.  
    *          If hWin is a valid handle or "default" window handle.
    *          If default window is also unrealized then {@link WindowConfig#RESOLVE_WINDOW}.
    */
   public static int targetWindowWorker(GenericFrame frame, 
                                        handle       hWin,
                                        boolean      ignoreThisProc)
   {
      // If hWin is null (IN WINDOW clause is not present) - try to use frame window.
      if (hWin == null && frame.getWindowId() != WindowConfig.RESOLVE_WINDOW)
      {
         hWin = new handle(getWindowById(frame.getWindowId()));
      }
      
      WindowWidget wnd = targetWindowWorkerInt(hWin, ignoreThisProc, false);
      
      return wnd == null ? WindowConfig.RESOLVE_WINDOW : wnd.getId();
   }
   
   /**
    * Determines target window for different window-aware operations such as MESSAGE, PAUSE 
    * or others.
    * 
    * @param   hWin
    *          Target window for IN WINDOW clause. <code>null</code> means get default window.
    * @param   ignoreThisProc
    *          <code>true</code> if concrete operation ignores THIS-PROCEDURE:CURRENT-WINDOW.
    * 
    * @return  ID of target window that is hWin resource id.  
    *          If hWin is a valid handle or "default" window handle.
    *          If default window is also unrealized then {@link WindowConfig#RESOLVE_WINDOW}.
    */
   public static int targetWindowWorker(handle  hWin, boolean ignoreThisProc)
   {
      WindowWidget wnd = targetWindowWorkerInt(hWin, ignoreThisProc, true);
      return wnd == null ? WindowConfig.RESOLVE_WINDOW : wnd.getId();
   }
   
   /**
    * Determines target window for different window-aware operations such as MESSAGE, PAUSE 
    * or others.
    * 
    * @param   hWin
    *          Target window for IN WINDOW clause. <code>null</code> means get default window.
    * @param   ignoreThisProc
    *          <code>true</code> if concrete operation ignores THIS-PROCEDURE:CURRENT-WINDOW.
    * @param   checkRealized
    *          <code>true</code> means necessity to check if target window is realized. 
    *          If window is realized - it is used as a result.
    *          <code>false</code> means that there is not necessity to check if window is visible.
    *          For example, if <code>hWin = null &amp;&amp; !ignoreThisProc</code> then 
    *          THIS-PROCEDURE:CURRENT-WINDOW is returned in spite of whether it realized or not.
    * 
    * @return  target window that is hWin resource id.  
    *          If hWin is a valid handle or "default" window handle.
    *          If default window is also unrealized then null.
    */
   private static WindowWidget targetWindowWorkerInt(handle  hWin, 
                                                     boolean ignoreThisProc, 
                                                     boolean checkRealized)
   {
      if (hWin == null)
      {
         // return "default" window dependent on ignoreThisProc flag.
         
         LogicalTerminal terminal = locate();
         if (!ignoreThisProc)
         {
            Object referent = terminal.pm._thisProcedure();
            if (referent == null) 
            {
               return null;
            }
            handle procCurWin = ProcedureManager.getCurrentWindow(referent);
            if (!checkRealized || isWindowRealized((WindowWidget) procCurWin.getResource()))
            {
               return (WindowWidget) procCurWin.getResource();
            }
         }
         
         if (!checkRealized || isWindowRealized(terminal.currentWindow))
         {
            return terminal.currentWindow;
         }
         
         if (!checkRealized || isWindowRealized(terminal.defaultWindow))
         {
            return terminal.defaultWindow;
         }
         
         return null;
      }
      if (!hWin._isValid() || hWin.isUnknown())
      {
         String msg = "Could not find widget handle for window from expression";
         ErrorManager.recordOrThrowError(3138, msg);
         return null;
      } 
      else if (!(hWin.getResource() instanceof WindowWidget))
      {
         String type = hWin.unwrapType().getResourceType().getTypeName();
         String msg = "Wrong type of object handle (" + type + ")  -- must be a window handle";
         ErrorManager.recordOrThrowError(3139, msg);
         return null;
      }
      else
      {
         return (WindowWidget)hWin.getResource(); 
      }
   }
   
   /**
    * Obtain the containing window for the given resource.
    *
    * @param    wr
    *           This must be a window (will return teh reference to itself), a frame or a
    *           widget.
    *
    * @return   The window for this widget.
    */
   private static WindowWidget lookupWindow(WrappedResource wr)
   {
      WindowWidget win = null;
      
      if (wr instanceof WindowWidget)
      {
         win = (WindowWidget) wr;
      }
      else if (wr instanceof FrameWidget)
      {
         GenericFrame frame = ((FrameWidget) wr).getFrame();
         
         int winId = frame.getWindowId();
         
         if (winId != WindowConfig.RESOLVE_WINDOW)
         {
            win = getWindowById(winId);
         }
      }
      else if (wr instanceof GenericWidget)
      {
         GenericWidget<?> wid = (GenericWidget<?>) wr;
         
         handle hwin = wid.getWindow();
         
         if (hwin != null && hwin._isValid() && !hwin.isUnknown())
         {
            win = (WindowWidget) hwin.unwrapWindow();
         }
      }
      
      return win;
   }
   
   /**
    * Determines if specified window is realized (viewed) at this moment.
    * 
    * @param   winHandle
    *          A handle of window to check. 
    * 
    * @return <code>true</code> if realized, <code>false</code> otherwise.
    */
   private static boolean isWindowRealized(WindowWidget winHandle)
   {
      // TODO implement here or somewhere else
      return false;
   }
   
   /**
    * Lookup dynamic or static window by its id.
    * 
    * @param    id
    *           a window id.
    *           
    * @return   WindowWidget
    *           window found, otherwise {@code null}.
    *           
    */
   protected static WindowWidget lookupWindowById(final int id)
   {
      LogicalTerminal lt = locate();

      if (lt.activeWindow.getId() == id)
      {
         return lt.activeWindow;
      }
      
      if (lt.currentWindow.getId() == id)
      {
         return lt.currentWindow;
      }
      
      if (lt.defaultWindow.getId() == id)
      {
         return lt.defaultWindow;
      }
      
      // try dynamic windows
      return lt.windowRegistry.get(id);
   }
   
   /**
    * Get dynamic or static window by its id or throw an exception if not found.
    * 
    * @param    id
    *           a window id.
    *           
    * @return   WindowWidget
    *           the window found
    *           
    * @throws   IllegalArgumentException
    *           If there is not window with provided id.
    */
   protected static WindowWidget getWindowById(final int id)
   {
      final WindowWidget w = lookupWindowById(id);
      if (w != null)
      {
         return w;
      }
      
      throw new IllegalArgumentException("No window found with specified id=" + id);      
   }
   
   /**
    * Returns the coordinate conversion implementation based on {@linkplain LogicalTerminal}
    * of the current session context.  
    * 
    * @return   {@linkplain CoordinatesConversion} reference.
    */
   public static CoordinatesConversion coordinates()
   {
      return coordinates;
   }
   
   /**
    * Show errors related to APPLY ... TO handle problems.
    *  
    * @param    invalidHandle
    *           When <code>true</code> error 3135 is also shown; otherwise, only the 3190 error
    *           is shown.
    */
   private static void applyToHandleError(boolean invalidHandle)
   {
      final String msg1 = "Invalid handle.  Not initialized or points to a deleted object";
      final int code1 = 3135;
            
      LogicalTerminal lt = locate();
      handle thisp = lt.pm.thisProcedure();
      String pname = lt.pm.getRelativeName(thisp.get());
      
      // legacy line can't be determined, use 0 instead
      String msg2 = "Could not find target of APPLY statement (%s:0) (3190)";
      msg2 = String.format(msg2, pname);
      int code2 = 3190;
      
      final int[] errIds = invalidHandle ? new int[] { code1, code2 } : new int[] { code2 };
      final String[] msgs = invalidHandle ? new String[] { msg1, msg2 } : new String[] { msg2 };

      ErrorManager.recordOrShowError(errIds, msgs, false, false, false, true);
   }

   /**
    * Disable or enable screen refresh (drawing) for the window associated with the
    * specified widget.  On enable, all pending output will be displayed. This is
    * useful to avoid flickering and improve performance during intermediate updates
    * that don't need to be seen by the user.
    * <p>
    * If a non-zero value is given as widgetId, that widget will have redraw disabled and will be
    * saved as the "locked" window. If called with a widgetId of 0, then the current locked 
    * window will have drawing re-enabled.
    * <p>
    * There can only be one locked window at any time, calling this with a non-zero widgetId when
    * there is already a locked window will do nothing.  Likewise, calling this with a 0 when
    * there is no locked window will do nothing.
    *
    * @param    widgetId
    *           The widget whose containing window will have refresh disabled or
    *           enabled. 
    *
    * @return   Reports if the operation succeeded.
    */
   private static boolean disableRedrawWorker(int widgetId)
   {
      LogicalTerminal lt = locate();
      
      synchronized (lt.registryLock)
      {
         if (widgetId == 0)
         {
            if (lt.locked == null)
            {
               // nothing to do
               return false;
            }
            
            // re-enable
            lt.locked.setDisableRedraw(false);
            
            // clear state
            lt.locked = null;
         }
         else
         {
            if (lt.locked != null)
            {
               // the WIN32 API only allows locking one window at a time
               return false;
            }
            
            GenericWidget<?> widget = lt.getWidgetForIdInt(widgetId);
            WindowWidget     win    = lookupWindow(widget);
            
            if (win != null)
            {
               // save state
               lt.locked = win;
               
               // disable
               lt.locked.setDisableRedraw(true);
            }
         }
      }
      
      return true;
   }
   
   /**
    * Shows all fields of the specified frame on screen.
    * 
    * @param    widgetId
    *           The frame ID or widget this request came for.
    * @param    frameBuf
    *           The screen buffer array with current values of widgets for editable frames.
    * @param    widgetIds
    *           The array of field level widget IDs to be shown or
    *           <code>null</code>, which means all of them.
    * @param    viewDisp
    *           <code>true</code> if the call is for VIEW, <code>false</code>
    *           for DISPLAY.
    * @param    windowId
    *           Id of target window or {@link WindowConfig#RESOLVE_WINDOW} for 
    *           current or default window.
    */
   private static void view(int            widgetId, 
                            ScreenBuffer[] frameBuf, 
                            int[]          widgetIds,
                            boolean        viewDisp, 
                            int            windowId,
                            int[]        tabItemList)
   {
      LogicalTerminal lt = locate();
      
      ScreenBuffer[] sb =
         lt.client.view(widgetId, frameBuf, widgetIds, viewDisp, windowId, tabItemList);
      
      handleScreenBuffers(sb);
      
      if (sb != null)
      {
         processDownConfigs(sb[0].getAndClearDownConfigs());
      }
      
      GenericWidget<?> widget = widgetId > 0 ? lt.getWidgetForIdInt(widgetId) : null;
      if (widget instanceof FrameWidget)
      {
         // if this is a frame, move it to top
         widget.moveInChain(true);
      }
   }

   /**
    * Check if all resources and widgets in the specified event list are valid.
    * 
    * @param    events
    *           The event list.
    *           
    * @return   <code>true</code> if all resources are valid.
    */
   private static boolean validEventList(EventList events)
   {
      LogicalTerminal lt = locate();
      
      boolean[] ok = new boolean[] { true };
      // validate the resource IDs first
      LinkedList<EventDefinition> defs = new LinkedList<>();
      events.gather(defs, new HashSet<>());
      for (EventDefinition ed : defs)
      {
         RoaringBitmap wids = new RoaringBitmap();
         ed.collectWidgetIds(wids);
         wids.forEach((IntConsumer) (wid -> 
         {
            GenericWidget<?> w = lt.getWidgetForIdInt(wid);
            if (w == null)
            {
               ok[0] = false;
            }
         }));
         
         if (!ok[0])
         {
            break;
         }
         
         Roaring64Bitmap rids = new Roaring64Bitmap();
         ed.collectResourceIds(rids);
         rids.forEach((LongConsumer) (rid -> 
         {
            handle h = handle.fromResourceId(rid);
            if (!h._isValid())
            {
               ok[0] = false;
            }
         }));
         
         if (!ok[0])
         {
            break;
         }
      }
      
      if (!ok[0])
      {
         String[] msgs =
         {
           "Invalid handle.  Not initialized or points to a deleted object.",
           "Could not attach trigger in " + lt.pm.getRelativeName(lt.pm._thisProcedure()) + "."
         };
         int[] nums = { 3135, 3198 };
         ErrorManager.recordOrThrowError(nums, msgs, false, false, false);
         return false;
      }
      
      return true;
   }
   
   /**
    * Read file into widgets on the client side.
    *
    * @param    wId
    *           A valid editor widget id.
    * @param    fileName
    *           The file name to be loaded on client side.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public static boolean editorReadFile(int wId, String fileName)
   {
      if (fileName == null)
      {
         return false;
      }
      boolean result = locate().client.editorReadFile(wId, fileName);
      return result;
   }

   /**
    * Write data from widget into the file the client side.
    *
    * @param    wId
    *           A valid editor widget id.
    * @param    fileName
    *           The file name to be saved on client side.
    *
    * @return   <code>TRUE</code> if operation successful <code>FALSE</code> otherwise.
    */
   public static boolean editorSaveFile(int wId, String fileName)
   {
      if (fileName == null)
      {
         return false;
      }
      boolean result = locate().client.editorSaveFile(wId, fileName);
      return result;
   }

   /**
    * Utility method for debugging, can be used to include information from the server-side
    * to the client-side log
    *
    * @param id
    *        widget id
    * @param message
    *       String parameter
    */
   public static void debug(int id, String message)
   {
      locate().client.debug(id, message);
   }

   /**
    * Move the widget to the top in z-order
    * 
    * @param id
    *          widget id
    *          
    * @return <code>Boolean.TRUE</code> if operation was successful
    *         null if no widget with given id is found
    */
   public static Boolean moveToTop(int id)
   {
      return locate().client.moveToTop(id);
   }

   /**
    * Move the widget to the bottom in z-order
    * 
    * @param id
    *          widget id
    *          
    * @return <code>Boolean.TRUE</code> if operation was successful
    *         null if no widget with given id is found
    */
   public static Boolean moveToBottom(int id)
   {
      return locate().client.moveToBottom(id);
   }

   /**
    * Set the state of SESSION:IMMEDIATE-DISPLAY attribute.
    * 
    * @param   on
    *          The new state for SESSION:IMMEDIATE-DISPLAY attribute.
    */
   public static void setImmediateDisplay(boolean on)
   {
      LogicalTerminal lt = locate();
      lt.client.setImmediateDisplay(on);
      lt.isImmediateDisplay = on;
   }

   /**
    * Set the state of SESSION:IMMEDIATE-DISPLAY attribute.
    * 
    * @param   on
    *          The new state for SESSION:IMMEDIATE-DISPLAY attribute.
    */
   public static void setImmediateDisplay(logical on)
   {
      setImmediateDisplay(on == null || on.isUnknown() ? false : on.booleanValue());
   }

   /**
    * Get the state of SESSION:IMMEDIATE-DISPLAY attribute. 
    * 
    * @return  the state of SESSION:IMMEDIATE-DISPLAY attribute.
    */
   public static logical isImmediateDisplay()
   {
      return new logical(locate().isImmediateDisplay);
   }

   /**
    * Set the state of SESSION:DATA-ENTRY-RETURN attribute.
    * 
    * @param   on
    *          The new state for SESSION:DATA-ENTRY-RETURN attribute.
    */
   public static void setDataEntryReturn(boolean on)
   {
      LogicalTerminal lt = locate();
      lt.client.setDataEntryReturn(on);
      lt.isDataEntryReturn = Boolean.valueOf(on);
   }

   /**
    * Get the state of SESSION:DATA-ENTRY-RETURN attribute. 
    * 
    * @return  the state of SESSION:DATA-ENTRY-RETURN attribute.
    */
   public static logical _isDataEntryReturn()
   {
      Boolean on = locate().isDataEntryReturn; 
      return on == null ? new logical(isChui()) : new logical(on);
   }

   /**
    * Set the state of SESSION:APPL-ALERT-BOXES attribute.
    *
    * @param   on
    *          The new state for SESSION:APPL-ALERT-BOXES attribute.
    */
   public static void setUsesAlertBoxes(boolean on)
   {
      locate().isUsingAlertBoxes = on;
   }
   
   /**
    * Set the state of SESSION:APPL-ALERT-BOXES attribute.
    *
    * @param   on
    *          The new state for SESSION:APPL-ALERT-BOXES attribute.
    */
   public static void setUsesAlertBoxes(logical on)
   {
      setUsesAlertBoxes(on == null || on.isUnknown() ? false : on.booleanValue());
   }

   /**
    * Get the state of SESSION:APPL-ALERT-BOXES attribute. 
    *
    * @return  the state of SESSION:APPL-ALERT-BOXES attribute.
    */
   public static logical isUsingAlertBoxes()
   {
      return new logical(locate().isUsingAlertBoxes);
   }
   
   /**
    * Obtain the default system palette for current theme from client. Calls
    * {@code ThemeManager.getCurrentTheme().getDefaultSystemPalette()} remotely, on the client.
    *
    * @return  the default system palette for current theme from client.
    */
   public static Map<Integer, ColorRgb> getThemeSystemPalette()
   {
      return locate().client.getThemeSystemPalette();
   }
   
   /**
    * Write message and stack trace to the client log
    * 
    * @param message
    *        the message
    * @param stack
    *        the Progress stack trace
    */
   public static void clientLog(String message, String[] stack)
   {
      locate().client.clientLog(message, stack);
   }

   /**
    * Set listener (browse) which listens to specific events that can appear in ROW-DISPLAY
    * trigger (e.g. pause or message box events), which cause listener to be redrawn with the new
    * rows.
    */
   public static void setRowDisplayDrawListener(RowDisplayDrawListener listener)
   {
      rowDisplayDrawListener.set(listener);
   }
   
   /**
    * Returns the current user name for session.
    * 
    * @return   The current OS logged in user name.
    */
   public static String getUserName()
   {
      return locate().client.getUserName();
   }

   /**
    * Returns the currently connected client's machine name for session.
    * 
    * @return   The current system logged in for the given session.
    */
   public static String getHostName()
   {
      return locate().client.getHostName();
   }
   
   /**
    * Returns the currently connected client's machine name for session.
    * 
    * @return   The current system logged in for the given session.
    */
   public static long getHostByName(String name)
   {
      return locate().client.getHostByName(name);
   }
   
   /**
    * Invoke the specified code on the server-side, maintaining the trigger registry: any new
    * triggers are sent to the client side to be added, and dead triggers are sent to be removed.
    * 
    * @param    srv
    *           The instance where to save the trigger differences.
    * @param    code
    *           The code to be executed.
    *           
    * @return   Any value returned by the given code.
    */
   private Object invokeOnServer(ServerReturnValue srv, Supplier<Object> code)
   {
      // save the triggers in the surviving scopes, to be compared with the new list after
      // the code has finished
      EventList prevTriggers = (cachedEventList == null 
                                 ? cachedEventList = TriggerManager.mergeTriggerRegistry(scope) 
                                 : cachedEventList);
      long prevTriggerChangeCount = triggerChangeCount;
      Object res = code.get();
      if (triggerChangeCount != prevTriggerChangeCount)
      {
         EventList currentTriggers = (cachedEventList == null 
                                       ? cachedEventList = TriggerManager.mergeTriggerRegistry(scope) 
                                       : cachedEventList);
         TriggerManager.computeTriggerDifferences(srv, prevTriggers, currentTriggers);
      }
      
      return res;
   }

   /**
    * If an event such as pause or message box occurs in ROW-DISPLAY trigger, this functions
    * notifies corresponding browse so it is redrawn with the new rows.
    */
   private static void notifyRowDisplayDrawListener()
   {
      RowDisplayDrawListener listener = rowDisplayDrawListener.get();
      if (listener != null)
      {
         listener.pushIntermediateRows();
      }
   }

   /**
    * Get browse widget by widget ID.
    *
    * @param  browseId
    *         Browse widget ID.
    *
    * @return browse widget.
    *
    * @throws MissingWidgetException
    *         if the browse with such ID cannot be found.
    */
   private static BrowseWidget getBrowseById(int browseId)
   throws MissingWidgetException
   {
      BrowseWidget browse = (BrowseWidget) getWidgetForId(browseId);

      if (browse == null)
      {
         throw new MissingWidgetException("Can't locate BrowseWidget for ID = " + browseId);
      }

      return browse;
   }
   
   /**
    * Refresh widget. 
    * Some widgets like PROGRESS-BAR export a method like <handle>:Refresh( ).
    * 
    * @param   widgetId
    *          Widget ID.
    */
   public static void refreshWidget(int widgetId)   
   {
      locate().client.refreshWidget(widgetId);
   }

   /**
    * Start an OLE drag/drop event with the given control as the source.
    * 
    * @param   widgetId
    *          Widget ID.
    */
   public static void oleDrag(int widgetId)   
   {
      locate().client.oleDrag(widgetId);
   }
   
   /**
    * Remove all objects from clipboard.
    */
   public static void oleClear()
   {
      locate().client.oleClear();
   }

   /**
    * Get object of specified format.
    * 
    * @param   format
    *          clipboard format.
    * @param   id
    *          Source widget ID.
    *          
    * @return  data at specified format.
    */
   public static byte[] oleGetData(int format, int id)
   {
      return locate().client.oleGetData(format, id);
   }

   /**
    * Put object of specified format.
    * 
    * @param   value
    *          data value
    * @param   format
    *          clipboard format.
    */
   public static void oleSetData(byte[] value, int format)
   {
      locate().client.oleSetData(value, format);
   }

   /**
    * Check if specified format is supported.
    * 
    * @param   format
    *          clipboard format.
    *          
    * @return <code>true</code> if supported, <code>false</code> if not.
    */
   public static boolean oleGetFormat(int format)
   {
      return locate().client.oleGetFormat(format);
   }

   /**
    * Enqueues widget attribute assignment.
    *
    * @param   widgetId
    *          Widget id.
    * @param   fname
    *          Field name.
    * @param   fid
    *          Field id.
    * @param   value
    *          New value.
    */
   public void enqueuWidgetAttr(int widgetId, String fname, int fid, Object value)
   {
      VALUES_COUNTER.update(fname, 1);

      widgetAttrIds.add(widgetId);
      widgetAttrIds.add(fid);
      widgetAttrValues.add(value);
   }

   /**
    * Flushes enqueued widget attributes. This method only calls @flushEnqueuedWidgetAttrs@ on the client
    * (which is a no-op) to trigger {@link #getChanges()}, which itself handles the enqueued widget attributes.
    */
   public void flushEnqueuedWidgetAttrs()
   {
      if (widgetAttrIds.isEmpty())
      {
         return;
      }

      SIMPLE_COUNTER.update(1);

      // this will trigger getChanges, the actual flush will happen there
      client.flushEnqueuedWidgetAttrs();
   }

   /**
    * Check if the given event is in the list of WAIT-FOR termination {@link #waitStack events}.
    * 
    * @param    eventCodes
    *           The event codes.
    *           
    * @return   <code>true</code> if the event is in the {@link #waitStack}.
    */
   private boolean isWaitEvent(int... eventCodes)
   {
      if (waitStack.isEmpty())
      {
         return false;
      }
      
      Iterator<EventList> iter = waitStack.iterator();
      while (iter.hasNext())
      {
         EventList el = iter.next();
         if (el.isEventMatchForWidget(-1, eventCodes))
         {
            return true;
         }
         
         for (int i = 0; i < eventCodes.length; i++)
         {
            String lbl = Keyboard.eventName(eventCodes[i]);
            if (el.isGoEvent(lbl))
            {
               return true;
            }
         }
      }
      
      return false;
   }
   
   /**
    * Exception raised when a widget cannot be found.
    */
   private static class MissingWidgetException
   extends Exception
   {
      /**
       * Constructor with a message.
       *
       * @param message
       *        Exception message.
       */
      public MissingWidgetException(String message)
      {
         super(message);
      }
   }
}
